#ifndef WEFG_H
#   define WEFG_H

#include "Wtypes.h"
#include "lobs3d.h"
#include "pirat.h"
#include "lobs3d.h"


#ifdef USE_ESFLIB
#   include "esflib3.h" 
#else
#   include "espack3.h" 
#endif


/* G3D */
#include "g3d.h" 

/* BBT */
#include "bbt.h" 

#define ListOf(type) LIST 


struct WEFG_node_t;


/* Define as double or float; float could be useful if memory
   is of premium import
 */
#ifndef SAVE_MEM_BY_FLOATS
#  define SAVE_MEM_BY_FLOATS 0
#endif
 /* type of entry in the shape function table (float or double) */
#if SAVE_MEM_BY_FLOATS
typedef float WEFG_sf_tab_entry_type_t;
#else
typedef double WEFG_sf_tab_entry_type_t;
#endif

/*
   A row in the shape function table.
 */
typedef struct WEFG_sf_tab_row_t {
  struct WEFG_node_t        *node; /* EFG node */
  WEFG_sf_tab_entry_type_t   phi; /* shape function value */
  WEFG_sf_tab_entry_type_t   dphix, dphiy, dphiz; /* derivatives wrt reference coordinates */
}              WEFG_sf_tab_row_t;


/*
   Shape function table.
 */
typedef struct WEFG_sf_tab_t {
  double                time_stamp; /* time stamp */
  int                   tot_of_entries; /* total of entries */
  int                   num_of_entries; /* current number of entries */
  WEFG_sf_tab_row_t    *rows; /* rows; one for each node */
}              WEFG_sf_tab_t;



/*
   Vertex of the integration volume
 */
typedef struct WEFG_vertex_t {
  W_index_t                     id; /* unique integer */
  W_loc_t                       loc; /* location */
  ListOf(WEFG_hex_volume_t *)   incident_hex_volumes; /* list of incident integration volumes */
  struct WEFG_node_t           *efgn; /* EFG node or NULL */
  WEFG_sf_tab_t                 sftab; /* shape function table */
  W_matstate_t                  matstate; /* material state */
}              WEFG_vertex_t;





#define MAX_VERT_PER_FACE 4     /* Supported: triangles, quadrilaterals */
/*
   Boundary face.
 */
typedef struct WEFG_b_face_t {
  int            num_vert; /* number of vertices */
  WEFG_vertex_t *v[MAX_VERT_PER_FACE]; /* vertices */
  W_bool_t       planar; /* is planar? */
  G3D_box_t      bounding_box; /* bounding box */
}              WEFG_b_face_t;



/*
   Which boundary face should be chosen when checking
   for intersections of segments with faces?
 */
typedef enum WEFG_which_b_face_enum {
  WEFG_ANY_B_FACE              = 0, /* any face */
  WEFG_B_FACE_CLOSEST_TO_START = 1, /* closest to the start of the segment */
  WEFG_B_FACE_CLOSEST_TO_END   = 2 /* closest to the end of the segment */
}            WEFG_which_b_face_enum;





/*
   EFG node data.
 */
typedef struct WEFG_node_data_t {
  W_coords_3d_t            ref_coords; /* reference coordinates */
  W_coords_3d_t            displ_t; /* current displacement */
  W_coords_3d_t            displ_t_minus_dt; /* displacement at the end of the previous time step */
  W_velocity_3d_t          velocity_t; /* current velocity */
  W_acceleration_3d_t      acceleration_t; /* current acceleration */
  W_index_t                num_of_eqns; /* number of equations */
  W_index_vector_t         eq_numbers; /* equation numbers */
}              WEFG_node_data_t;

/*
   EFG node type
 */
typedef enum WEFG_efg_node_type_t {
  WEFG_TRUE_EFG_NODE = 0, /* true EFG node */
  WEFG_EFG_IF_NODE   = 1, /* EFG node at the vertex of an interface
                           integration volume */
  WEFG_FE_IF_NODE    = 2 /* EFG node acting as a front-end to an FE
                          node at the vertex of an interface integration
                          volume */
}  WEFG_efg_node_type_t;

/*
   EFG node
 */
typedef struct WEFG_node_t {
  int                      id; /* unique integer */
  WEFG_efg_node_type_t     type; /* type of node */
  float                    support_size; /* support size */
  WEFG_node_data_t   ndata; /* EFG node: non-NULL for WEFG_TRUE_EFG_NODE and WEFG_EFG_IF_NODE */
  W_node_t          *fen; /* FE node: non-NULL for WEFG_FE_IF_NODE */
#ifdef X_XFACE
  GraphicObj              *node_marker;
#endif
}              WEFG_node_t;


#define XDir 0
#define YDir 1
#define ZDir 2

/*
   Integration point.
 */
typedef struct WEFG_integration_point_t {
  WEFG_sf_tab_t                 sftab; /* shape function table */
  W_matstate_t                  matstate; /* material state */
}              WEFG_integration_point_t;



/*
   Quadrature points in an integration volume
 */
typedef struct WEFG_quadr_points_t {
  W_gauss_integration_rule_t    *qr; /* Gauss rule */
  WEFG_integration_point_t     **efg_ip; /* integration points */
}              WEFG_quadr_points_t;


/*
   System vector
 */
typedef struct WEFG_sys_vector_t {
  W_index_t      dim;           /* It is 1-based!!! */
  W_vector_t     v; /* vector */
}              WEFG_sys_vector_t;
  

/*
   Concentrated load.
 */
typedef struct WEFG_conc_load_t {
  W_index_t              id; /* unique integer */
  W_loc_t                loc; /* location */
  W_vect_t               amplitudes; /* amplitudes */
  W_index_t              time_func_num; /* time function number */
  WEFG_sf_tab_t          sftab; /* shape function table */
}              WEFG_conc_load_t;

/*
   Body load
 */
typedef struct WEFG_body_load_t {
  W_bool_t           is_set; /* was specified? */
  W_bool_t           is_time_independent; /* is time independent? */
  W_bool_t           is_constant; /* is constant in space? */
  char              *body_load_expr[3]; /* expression for components */
  W_vect_t           const_body_load; /* cached vector for a constant
                                         load
                                       */
  WEFG_sys_vector_t  time_independent_body_load; /* cached system vector
                                                    for a time independent
                                                    body load
                                                  */
}              WEFG_body_load_t;



/*
   Type of distributed load.
 */
typedef enum { WEFG_DISTRIB_LOAD_DEAD_PRESSURE = 0, /* fixed in reference configuration */
               WEFG_DISTRIB_LOAD_PRESSURE      = 1, /* follower */
               WEFG_DISTRIB_LOAD_DEAD_SHEAR    = 2, /* fixed in reference configuration */
               WEFG_DISTRIB_LOAD_SHEAR         = 3  /* follower */
             }  WEFG_distrib_load_type_t;

  
/*
   Distributed load.
 */
typedef struct WEFG_distrib_load_t {
  W_index_t                  id; /* unique integer */
  WEFG_distrib_load_type_t   type; /* type */
  struct {
    W_loc_t              loc; /* location in 3-D space */
    W_float_t            x, y; /* projection into the local plane of load patch */
    W_float_t            traction; /* traction */
  }                          vdata[4]; /* vertex data */
  W_index_t                  time_func_num; /* time function number */
  int                        suggested_qr; /* quadrature order */
  W_vect_t                   u, v, normal; /* local csys on the patch */
  WEFG_quadr_points_t       *qp; /* quadrature points for the patch */
  W_vect_t                   dir_vect; /* only for shear loads -- */
                                       /* gives with the normal the plane */
                                       /* to which the shear is parallel; */
                                       /* the vector /u/ is then computed */
                                       /* such that the shear is parallel */
                                       /* with it */
  double                     time_stamp;
}              WEFG_distrib_load_t;

/*
   Contact load type
 */
typedef enum { WEFG_CONTACT_LOAD_UNILATERAL = 0, /* unilateral */
               WEFG_CONTACT_LOAD_STICKY     = 1  /* acts both ways */
             }  WEFG_contact_load_type_t;

/*
   Contact load.
 */
typedef struct WEFG_contact_load_t {
  W_index_t                  id;
  WEFG_contact_load_type_t   type;
  double                     contact_press_scaling;
  struct {
    W_loc_t                loc; /* 3-D location */
    W_float_t              x, y; /* projection into the local plane of load patch */
  WEFG_sf_tab_t            sftab; /* shape fucntion table */
  }                          vdata[4]; /* vertex data */
  int                        suggested_qr; /* quadrature order */
  W_vect_t                   u, v, normal; /* local csys on the patch */
  WEFG_quadr_points_t       *qp; /* quadrature points for the patch */
  G3D_plane_t                contact_plane; /* plane of contact */
  double                     time_stamp; /* time stamp */
}              WEFG_contact_load_t;

  
typedef char WEFG_short_bool_t;

/*
   Integration volume -- hexahedron.
 */
typedef struct WEFG_hex_volume_t {
  W_index_t                      id; /* unique integer */
  WEFG_vertex_t                 *vertex[8]; /* vertices */
#ifdef X_XFACE
  GraphicObj                    *face_grep[6];
#endif
  WEFG_short_bool_t              face_on_boundary[6]; /* face is on the external boundary? */
  WEFG_short_bool_t              is_intersected_by_crack_surf; /* is interescted by crack? */
  int                            quad_order; /* quadrature order */
  WEFG_quadr_points_t           *qp; /* quadrature points */
  W_float_t                      volume; /* volume of integration cell */
  ListOf(WEFG_hex_volume_t *)    offspring; /* list of integration volumes
                                               which are the offspring of
                                               this one (NULL for integration
                                               volume which has not been
                                               subdivided)
                                             */
  double                         time_stamp; /* time stamp */
}              WEFG_hex_volume_t;

/*
   Interface data (used by functions related to the shape function
   evaluation).
 */
typedef struct WEFG_if_data_t {
  WEFG_hex_volume_t *h; /* integration volume */
  double xi, eta, theta; /* parametric coordinates
                          of evaluation point in the integration volume */
}              WEFG_if_data_t;


  
typedef void *W_ptr_t;



struct WEFG_crack_surf_methods_t;

typedef void *WEFG_crack_surf_data_t;

struct WEFG_superelem_t;

/*
   Generic crack surface.
 */
typedef struct WEFG_crack_surf_t {
  struct WEFG_superelem_t             *se; /* superelement */
  int                                  id; /* unique integer */
  WEFG_crack_surf_data_t               data; /* data for the crack instance */
  struct WEFG_crack_surf_methods_t    *methods; /* methods for the crack instance  */
  W_bool_t                             have_intersection; /* cached private data */
  W_bool_t                             changed_shape; /* did crack change shape? */
}              WEFG_crack_surf_t;

/*
   Position of evaluation point wrt crack surface.
 */
typedef enum WEFG_pt_pos_wrt_crack_surf_enum {
  WEFG_NEGATIVE_NORMAL     = -1, /* on the `negative normal' side */
  WEFG_ON_SURF             =  0, /* on the surface */
  WEFG_POSITIVE_NORMAL     =  1, /* on the `positive normal' side  */
  WEFG_AT_FRONT            =  2, /* at the front */
  WEFG_POS_WRT_SURF_UNSPEC =  1000 /* indeterminate */
} WEFG_pt_pos_wrt_crack_surf_enum;


typedef WEFG_pt_pos_wrt_crack_surf_enum
(*WEFG_node_pos_wrt_crack_surf_proc)(WEFG_crack_surf_t *cs,
                                          WEFG_node_t *efgn,
                                          double physical_tol);


/*
   Packet for passing data to the ESFLib callbacks.
 */
typedef struct WEFG_eval_for_packet_t {
  struct WEFG_superelem_t                 *se; /* superelement */
  WEFG_pt_pos_wrt_crack_surf_enum          pt_wrt_crack_surf_classification; /* position of evaluation point wrt to crack surface */
  WEFG_crack_surf_t                       *crack_surf; /* crack surface if above position specified; NULL otherwise */
  W_vect_t                                 crack_surf_normal; /* normal to the crack surface at the evaluation point or near it */
}              WEFG_eval_for_packet_t;


typedef W_bool_t
      (*WEFG_segment_intersects_crack_surf_proc)(WEFG_eval_for_packet_t *efp,
                                                 G3D_segment_t *s,
                                                 WEFG_node_t *efgn,
                                                 WEFG_crack_surf_t *cs,
                                                 double  tol);

typedef W_bool_t
      (*WEFG_hex_intersected_by_crack_surf_proc)(WEFG_crack_surf_t *cs,
                                                 WEFG_hex_volume_t *h);

typedef WEFG_pt_pos_wrt_crack_surf_enum
(*WEFG_point_pos_wrt_crack_surf_proc)(WEFG_crack_surf_t *cs,
                                      W_loc_t *pt, double physical_tol);

typedef void (*WEFG_modify_crack_surf_grep_proc)(struct WEFG_superelem_t *se,
                                                 double displ_scale,
                                                 WEFG_crack_surf_t *cs);

typedef void (*WEFG_propagate_proc)(WEFG_crack_surf_t *cs);


typedef void (*WEFG_write_hexvue_postproc_proc)(WEFG_crack_surf_t *cs,
                                                ckit_iobuf_t IOB,
                                                char *mode);

typedef void (*WEFG_write_tecplot_postproc_proc)(WEFG_crack_surf_t *cs,
                                                 ckit_iobuf_t IOB,
                                                 char *mode,
                                                 void *optdata);

typedef void (*WEFG_update_stress_proc)(struct WEFG_superelem_t *se,
					WEFG_crack_surf_t *cs);


typedef void (*WEFG_write_ve3d_loclist_proc)(WEFG_crack_surf_t *cs,
                                             ckit_iobuf_t IOB);

typedef void (*WEFG_write_ve3d_polys_proc)(WEFG_crack_surf_t *cs,
                                           ckit_iobuf_t IOB);
typedef void *(*WEFG_new_sampler_func)(WEFG_crack_surf_t *cs,
				       W_index_t pntidx, W_loc_t *loc);

typedef void (*WEFG_sampler_store_proc)(WEFG_crack_surf_t *cs,
					W_solution_history_t *h,
					void *data,
					double time);

typedef int (*WEFG_plane_generating_vector_func)
                  (WEFG_crack_surf_t *cs, 
                    double atx, double aty, double atz ,
                    double rcx, double rcy, double rcz ,
                    double *bx, double *by, double *bz );

typedef int (*WEFG_wrap_around_point_func)(WEFG_crack_surf_t *cs, 
                                           double atx, double aty, double atz,
                                           double rcx, double rcy, double rcz,
                                           double bx, double by, double bz,
                                           double *wapx, double *wapy, double *wapz);


/*
   Methods a specific crack instance has to implement.
 */
typedef struct WEFG_crack_surf_methods_t {
  WEFG_segment_intersects_crack_surf_proc    segment_intersects_crack_surf;
  WEFG_plane_generating_vector_func          plane_generating_vector;
  WEFG_wrap_around_point_func                wrap_around_point;
  WEFG_node_pos_wrt_crack_surf_proc          node_pos_wrt_crack_surf;
  WEFG_propagate_proc                        propagate;
#ifdef X_XFACE
  WEFG_modify_crack_surf_grep_proc           modify_crack_surf_grep;
#endif
  WEFG_write_hexvue_postproc_proc            write_hexvue_postproc;
  WEFG_write_ve3d_loclist_proc               write_ve3d_loclist;
  WEFG_write_ve3d_polys_proc                 write_ve3d_polys;
  WEFG_write_tecplot_postproc_proc           write_tecplot_postproc;
  WEFG_update_stress_proc                    update_stress;
  WEFG_new_sampler_func                      new_sampler;
  WEFG_sampler_store_proc                    sampler_store;
  WEFG_hex_intersected_by_crack_surf_proc    hex_intersected_by_crack_surf;
}              WEFG_crack_surf_methods_t;


typedef  WEFG_pt_pos_wrt_crack_surf_enum
   (*WEFG_on_which_face_of_crack_surf_proc)
      (WEFG_crack_surf_t *cs,
       W_loc_t *pt_on_surf,
       W_vect_t *normal_to_surf,
       W_ptr_t data);


/*
   Symmetry plane (not yet functional).
 */
typedef struct WEFG_symm_plane_t {
  int                direction_of_normal; /* x,y or z */
  W_loc_t            center;    /* center of the plane */
  G3D_plane_t        plane;     /* computed from the above data */
  double             tolerance; /* tolerance, under which a node is assumed */
                                /* to be located on the symmetry plane */
}              WEFG_symm_plane_t;



#include "Wmat3d.h"

/*
   EFG shape function statistics.
 */
typedef struct WEFG_efg_sf_stat_t {
  double  sum_nneighbors;
  int     nevals;
  int     min_nneighbors;
  int     max_nneighbors;
  double  avg_nneighbors;
}              WEFG_efg_sf_stat_t;


/*
   Triangle in a region of invalidated connectivities.
 */
typedef struct WEFG_region_tri_t {
  W_bool_t     is_valid; /* are the triangle valid? */
  G3D_tri_t    tri; /* triangle */
  G3D_box_t    bbox; /* cached bounding box */
  G3D_plane_t  plane; /* cached plane */
}              WEFG_region_tri_t;

/*
   Region of invalidated connectivities.
 */
typedef struct WEFG_region_t {
  W_bool_t           is_initialized; /* are region data valid? */
  G3D_box_t          bbox; /* cached bounding box */
  WEFG_region_tri_t  triangles[2]; /* quads are decomposed into triangles */
}              WEFG_region_t;


/*
   Integration point reference.
 */
typedef struct WEFG_ip_ref_package_t {
  WEFG_integration_point_t  *ip; /* integration point */
  W_loc_t                    loc; /* 3-D location */
  WEFG_hex_volume_t         *enclosing_hex; /* enclosing integration volume */
}              WEFG_ip_ref_package_t;

/*
   Interpolation package.
 */
typedef struct WEFG_ip_ipol_package_t {
  WEFG_ip_ref_package_t  *ip_ref;
  double weight;
}              WEFG_ip_ipol_package_t;

/*
   Field package.
 */
typedef struct WEFG_field_package_t {
  union {
    double tensor[3][3];
    double vector[3];
    double scalar;
    void  *pointer;
  } value;
  double weight;
}              WEFG_field_package_t;


typedef double (*WEFG_tensor_prod_comp_func)(double r);
typedef double (*WEFG_tensor_prod_comp_der_func)(double r);


struct WEFG_superelem_t;

typedef W_bool_t (*WEFG_update_field_pkg_proc)(void *options,
					       struct WEFG_superelem_t *se,
					       WEFG_ip_ref_package_t *ipref,
					       double weight,
					       WEFG_field_package_t *result);

	
/*
   Region of specified support size.
 */
typedef struct WEFG_support_size_region_t {
  G3D_box_t    box; /* box */
  double       support_size; /* support size */
}              WEFG_support_size_region_t;

/*
   List of regions of invalidated connectivities
 */
typedef struct WEFG_invalid_regions_t {
  W_bool_t                   triangles_are_reliable;
  ListOf(WEFG_region_t *)    regions;
}              WEFG_invalid_regions_t;

/*
   Type of EFG weight function.
 */
typedef enum WEFG_weight_func_type_enum {
  WEFG_QUARTIC_SPLINE_SPHERICAL_WEIGHT_FUNC     = 0,
  WEFG_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC        = 1,
  WEFG_SMOOTH_QUADRATIC_TENSOR_PROD_WEIGHT_FUNC = 2,
  WEFG_QUADRATIC_SPHERICAL_WEIGHT_FUNC          = 3,
  WEFG_SMOOTH_QUADRATIC_SPHERICAL_WEIGHT_FUNC   = 4,
  WEFG_GAUSSIAN_SPHERICAL_WEIGHT_FUNC           = 5
}            WEFG_weight_func_type_enum;

/*
   Order of complete polynomial basis.
 */
typedef enum WEFG_order_enum {
  WEFG_COMPLETE_LINEAR     = 1,
  WEFG_COMPLETE_QUADRATIC  = 2
}            WEFG_order_enum;


#define MIN_NUM_NEIGHBORS           4
#define SUPPORT_SIZE_MULT_FACTOR    1.1
#define DEFAULT_QUAD_ORDER          2
/*
   EFG superelement.
 */
typedef struct WEFG_superelem_t {
  W_domain_t                           *domain;
  /* Generate EFG nodes at this spacing. Input. */
  double                                node_spacing; 
  /* Merging distance. Either input, or computed as support_size/10. */
  double                                merging_distance; 
  /* Shortest dist. between EFG nodes Computed. */
  double                                shortest_distance;
  /* The number of nodes per hex volume side.  Generates anisotropic */
  /* spacing of EFG nodes, in case the hex is not a cube.  Specified */
  /* node spacing takes precedence. */
  int                                   num_nodes_per_hex_side;
  /* Compute support sizes if set. */
  W_bool_t                              compute_support_size;
  /* The support size may be enforced in boxes */
  ListOf(WEFG_support_size_region_t *)  support_size_region_list;
  /* Largest support size allowed. Input. */
  double                                max_support_size; 
  /* Support size for those nodes which are not located
     in one of the support size regions.  Fallback.  Input. */
  double                                support_size;
  /* Support size multiplier for all EFG nodes.  Input. */
  double                                support_size_multiplier;
  /* Suggested number of neighbors.  Input. */
  int                                   num_neighbors_suggested;
  /* Smallest number of neighbors.  Input. */
  int                                   num_neighbors_min;
  /* Weight function type */
  WEFG_weight_func_type_enum            weight_func_type;
  /* Weight function type */
  WEFG_order_enum                       basis_order;
  /* Precision tolerance used by the shape function library */
  double                                sflib_prec_tol;
  /* The solver used to solve the moment equations */
  ESFLIB_A_solver_enum_t                A_solver;
  /* Quadrature order in ALL hex volumes */
  int                                   quadrature_order;
# ifdef ENABLE_TWO_POINT_RULE
  /* Is the two-point quadrature enabled?  */
  W_bool_t                              enable_2pt_quadrature;
# endif
  /* List of all vertices in the EFG superelement */
  dict_dictionary                       verts;
  /* Search structure for vertices */
  pirat_front_end_t                    *verts_ss;
  /* Estimate of the number of vertices (used to set up the vertex search */
  /* structure). */
  int                                   vertex_ss_bucket_size;
  /* List of all top-level hex volumes in the EFG superelement.  Note: */
  /* The top-level hex volumes contribute the boundary faces which can */
  /* affect the visibility computation in shape func evaluation.  They */
  /* can be subdivided further for the purpose of numerical integration. */
  /* Their offspring shows the boundary faces in the graphical output, */
  /* but do not list these faces for the purpose of visibility computation. */
  ListOf(WEFG_hex_volume_t *)           hex_volumes;
  /* Search structure for hex volumes. */
  lobs_3d_front_end                     hex_ss;
  /* List of all the cracks */
  ListOf(WEFG_crack_surf_t *)           cracks;
  /* List of boundary faces.  See comment concerning the hex volumes above. */
  ListOf(WEFG_b_face_t *)               b_faces;
  /* Is the superelement domain convex? */
  W_bool_t                              is_convex;
  /* Is the superelement domain convex with cracks? */
  W_bool_t                              is_convex_with_cracks;
  /* The nodes are to be generated independently of the cells? */
  W_bool_t                              generate_nodes_in_volume;
  /* ...inside this volume. */
  W_loc_t                               lower_corner, upper_corner;
  /* Generate nodes at the vertices? */
  W_bool_t                              generate_nodes_at_vertices;
  /* Generate nodes in the interior of the integration cells, not at their */
  /* boundaries and vertices */
  W_bool_t                              generate_nodes_inside_only;
  /* List of all EFG nodes. */
  dict_dictionary                       efgnodes;
  /* Search structure for EFG nodes. */
  lobs_3d_front_end                     efg_node_ss;
  /* Estimate of the number of vertices (used to set up the efg node search */
  /* structure). */
  int                                   efg_node_ss_bucket_size;
  /* List of concentrated loads. */
  ListOf(WEFG_conc_load_t *)            conc_loads;
  /* List of distributed loads. */
  ListOf(WEFG_distrib_load_t *)         distrib_loads;
  /* List of contact loads. */
  ListOf(WEFG_contact_load_t *)         contact_loads;
  /* Body load (valid in the whole EFG domain) */
  WEFG_body_load_t                      body_load;
  /* Material id.  The same for all hex volumes constituting this */
  /* superelement */
  W_mat_3d_t                           *mat;
  /* Number of equations at the EFG nodes. */
  W_index_t                             num_of_equations;
  /* Mass matrix. */
  WEFG_sys_vector_t                     mass_matrix;
  /* Is the mass being assembled to internal DOF's only? */
  W_bool_t                              internal_mass_only;
  /* Should the mass matrix be kept constant during crack growth? */
  W_bool_t                              time_independent_mass;
  /* Total mass of the EFG superelement */
  W_float_t                             total_mass;
  /* Effective loads */
  WEFG_sys_vector_t                     effective_loads;
  /* Update vector. */
  WEFG_sys_vector_t                     update_vector;
  /* External load work. */
  W_float_t                             ext_load_work;
  /* Search tolerance -- no two efg nodes or vertices are expected */
  /* to be closer than that */
  double                                tolerance;
  /* List of symmetry planes */
  ListOf(WEFG_symm_plane_t *)           symm_planes;
  /* Compute stresses at the vertices and other output points */
  W_bool_t                              compute_stress_at_output_pts;
  /* Compute equiv. plastic def. at the vertices and other output points */
  W_bool_t                              compute_equiv_pl_def_at_output_pts;
  /* EFG shape functions statistics */
  WEFG_efg_sf_stat_t                    efg_sf_stat;
  /* Time stamp for shape function validity; all points at which */
  /* the time stamp is earlier need to recompute if they are */
  /* in the inclusion */
  double                                invalid_after_time;
  /* The bounding boxes define a region in which the shape functions */
  /* may become invalid after the time invalid_after_time */
  WEFG_invalid_regions_t                invalid_regions;
  /* Search structure for integration points */
  pirat_front_end_t                    *ip_ss;
  /* Estimate of the integration points (used to set up the search */
  /* structure). */
  int                                   ip_ss_bucket_size;
  /* Store the integration points in a search structure */
  W_bool_t                              store_ip_in_search_structure;
  /* Force recomputation of all shape functions */
  W_bool_t                              force_all_sf;
  /* Draw rays point-node for shape function tables. */
  W_bool_t                              draw_rays;
  /* Rayleigh mass damping coefficient */
  double                                mass_damping_coeff;
#ifdef USE_ESFLIB       
  /* Element-free Galerkin shape function library */
  ESFLIB_lib_t                          esflib;
#endif
  /* General constraint of nodes */
  int                                   general_constraint[3];
  /* Use wrap-around technique of making the shape functions C^0 */
  W_bool_t                              use_wrap_around;
  /* Scramble (randomize) orientation of the 0-6 diagonal of all hexahedra;
     This is useful when the two-point diagonal integration rule is used.
   */
  W_bool_t                              scramble_diagonals;
  /* The specification of connectivities */
  dict_dictionary                       coupling_fen_conn;
  /* The specification of connectivities */
  dict_dictionary                       driven_fen_conn;
}              WEFG_superelem_t;




/* Graphics */
#define EFG_SURFACE_LAYER      WX_ELEM_LAYER
#define EFG_NODE_LAYER         WX_NODE_LAYER
#define EFG_LOAD_LAYER         WX_LOAD_LAYER


#endif
