#ifndef WMAT1D_H
#   define WMAT1D_H

#include "Wmat3d.h"             /* depend on the 3D stress interface */

typedef void *W_mat_1d_state_t;

struct W_mat_1d_ip_ctx_t;

typedef enum W_mat_1d_kinem_var_type_enum {
  MAT_1D_ENG_STRAIN           =  0, /* du/dX */
  MAT_1D_GREEN_LAGRANGE       =  1, /* du/dX + 1/2(du/dX)^2 */
  MAT_1D_DISPL_GRAD           =  2, /* du/dX (X=reference coordinate) */
  MAT_1D_PREV_DISPL_GRAD      =  3  /* du/dX (X=reference coordinate); previous config */
}            W_mat_1d_kinem_var_type_enum;



typedef enum W_mat_1d_mat_state_var_enum {
  MAT_1D_EQUIV_PLASTIC_DEF         = 0,
  MAT_1D_DEFORM_ENERGY_DENSITY     = 1
}            W_mat_1d_mat_state_var_enum;


typedef enum W_mat_1d_mat_const_enum {
  MAT_1D_YOUNG_MODULUS                  = 0,
  MAT_1D_POISSON                        = 1,
  MAT_1D_TENSILE_STRENGTH               = 4,
  MAT_1D_DIL_WAVE_SPEED_1D              = 6,
  MAT_1D_KINEM_HARD_MOD                 = 7,
  MAT_1D_ISOTR_HARD_MOD                 = 8,
  MAT_1D_YIELD_STRESS                   = 9
}            W_mat_1d_mat_const_enum;


  
typedef W_bool_t (*WM1D_get_kinem_var_proc)(struct W_mat_1d_ip_ctx_t *ctx,
                                            W_mat_1d_kinem_var_type_enum  type,
                                            double *kinvar);

typedef struct W_mat_1d_ip_ctx_t {
  double                     time;
  double                     time_step;
  void                      *ctx_data;
  WM1D_get_kinem_var_proc    get_kinem_var; 
}              W_mat_1d_ip_ctx_t; 


struct W_mat_1d_t;

typedef W_mat_1d_state_t (*WM1D_get_mat_state_proc)(struct W_mat_1d_t *m,
                                                    W_loc_t *at_point);

typedef W_bool_t 
   (*WM1D_get_mat_const_proc)(struct W_mat_1d_t *m,
                               W_mat_1d_mat_const_enum type,
                               double *c);

typedef W_bool_t 
   (*WM1D_update_stress_proc)(struct W_mat_1d_t           *m,
                              W_mat_1d_state_t             ms,
                              W_mat_state_update_form_enum update_form,
                              W_mat_1d_ip_ctx_t           *ip_ctx,
                              W_mat_3d_stress_type_enum    stress_type,
                              double                      *stress,
                              W_mat_3d_stress_type_enum    *out_stress_type
                              );


typedef W_bool_t 
    (*WM1D_get_mat_state_var_proc)(struct W_mat_1d_t           *m,
                                   W_mat_1d_state_t             ms,
                                   W_mat_1d_ip_ctx_t           *ip_ctx,
                                   W_mat_1d_mat_state_var_enum  var_type,
                                   double                      *var
                                   );

typedef struct WM1D_mat_state_pkg_t {
  W_mat_1d_state_t        ms;
  double                  weight;
}              WM1D_mat_state_pkg_t;


typedef W_bool_t 
    (*WM1D_get_stiff_coeff_proc)(struct W_mat_1d_t           *m,
                            W_mat_1d_state_t             ms,
                            W_mat_1d_ip_ctx_t           *ip_ctx,
                                 double *stiff_coeff
                            );

  

typedef int (*WM1D_get_mcode_func) (struct W_mat_1d_t *m);

typedef struct W_mat_1d_methods_t {
  WM1D_get_mcode_func                     get_mcode;
  WM1D_get_mat_state_proc                 get_mat_state;
  WM1D_get_mat_const_proc                 get_mat_const;
  WM1D_update_stress_proc                 update_stress;
  WM1D_get_mat_state_var_proc             get_mat_state_var;
  WM1D_get_stiff_coeff_proc               get_stiff_coeff;
}              W_mat_1d_methods_t;
  




typedef struct W_mat_1d_t {
  W_float_t                       vol_mass;
  W_mat_1d_methods_t             *methods;
  void                           *data;
#ifdef X_XFACE
  EPixel                          color;
  EFringeTable                    fringes;
#endif
}              W_mat_1d_t;


typedef W_mat_1d_t * (*W_mat1d_reader_func) (W_domain_t *d);

void 
WM1D_initialize_readers (void);


/*
   Register a mat1d reader.
   To be able to use specific implementation
   of the general triaxial material (mat1d),
   its code needs to be linked to the
   executable, and its reader needs to be registered, so that the generic
   mat1d reader can look up the type-specific reader.
   */
void
WM1D_register_reader (int m1dcode /* unique integer among all elements */,
                      W_mat1d_reader_func reader /* the reader function */
                      );

W_mat_1d_t *
mat1d_reader (W_domain_t *d);

W_mat_1d_t *
WM1D_new_material(W_mat_1d_methods_t *methods,
                  void *data,
                  double vol_mass);

W_mat_1d_state_t 
WM1D_get_mat_state(W_mat_1d_t *m, W_loc_t *at_point);

W_mat_1d_state_t 
WM1D_interpolate_mat_state(W_mat_1d_t *m,
			   W_loc_t *at_point,
			   int npkgs,
			   WM1D_mat_state_pkg_t *pkgs);

W_bool_t 
WM1D_get_mat_const(W_mat_1d_t *m, W_mat_1d_mat_const_enum type, double *c);

/*
   Update stress in a material state and return it.
 */
W_bool_t /* TRUE if successful; FALSE otherwise, in which case an error
            message is generated
          */
WM1D_update_stress (W_mat_1d_t                  *m/* mat1d */,
                    W_mat_1d_state_t             ms, /* material state */
                    W_mat_state_update_form_enum update_form, /* update is final or temporary? */
                    W_mat_1d_ip_ctx_t           *ip_ctx,/* integration point context; used to generate kinematic quantities which the material state
                                                           might need in order to compute the requested stress */
                    W_mat_3d_stress_type_enum    stress_type,/* type of stress tensor */
                 double                       *stress,/* stress tensor */
                 W_mat_3d_stress_type_enum   *out_stress_type/* type of actually returned stress tensor; ideally, output type == requested type, but
                                                                sometimes the material state fails to compute the requested tensor for lack of information
                                                             */
                    );

W_bool_t 
WM1D_get_mat_state_var(W_mat_1d_t                  *m,
                       W_mat_1d_state_t             ms,
                       W_mat_1d_ip_ctx_t           *ip_ctx,
                       W_mat_1d_mat_state_var_enum  var_type,
                       double                      *var
                       );

double 
WM1D_get_volume_mass(W_mat_1d_t *m);

int
WM1D_get_mcode (W_mat_1d_t *m);

W_bool_t
WM1D_get_stiff_coeff (W_mat_1d_t                  *m/* mat1d */,
                    W_mat_1d_state_t             ms, /* material state */
                    W_mat_1d_ip_ctx_t           *ip_ctx,/* integration point context; used to generate kinematic quantities which the material state
                                                           might need in order to compute the requested stress */
                      double *stiff_coeff
                      );

#endif /* WMAT1D_H */
