#include "Wconfig.h"
#include "Wxf.h"
#include "Wcdi.h"
#include "Wutil.h"
 
/* RCSHEADER "$Id: cdi.c,v 1.2 2004/09/06 01:17:58 cvs Exp $" */

/* private methods */
 
static void
Warburton_advance_step (W_cd_integrator_t *i/* integrator */)
{
  /* Calculate effective loads at time t */

  (*i->calc_eff_loads) (i->domain);
   
  /* Solve for displacements at time t+dt */

  (*i->solve_for) (i->domain);
  
  /* Evaluate velocities and accelarations */

  (*i->update_config) (i->domain);

  i->t += i->dt;

  /* Change time step */
  if (i->finalize_step != NULL)
    i->dt = (*i->finalize_step) (i->domain);
}

static void
Newmark_advance_step (W_cd_integrator_t *i)
{
  /* Evaluate velocities and accelarations */

  (*i->update_config) (i->domain);
 
  /* Calculate effective loads at time t */

  (*i->calc_eff_loads) (i->domain);
   
  /* Solve for displacements at time t+dt */

  (*i->solve_for) (i->domain);

  i->t += i->dt;

  /* Change time step */
  if (i->finalize_step != NULL)
    i->dt = (*i->finalize_step) (i->domain);
}

/* public interface */
/*
   New Warburton integrator.
   Create a new Warburton central-difference integrator.
 */
W_cd_integrator_t * /* integrator */
W_cd_new_Warburton_integrator (void                    *domain, /* domain */
                               W_float_t               t0, /* initial time */
                               W_float_t               dt, /* time step */
                               W_cdi_calc_eff_loads_proc calc_eff_loads, /* procedure to calculate effective loads */
                               W_cdi_solve_for_proc      solve_for, /* procedure to solve for the primary unknowns */
                               W_cdi_update_config_proc  update_config, /* procedure to update configuration */
                               W_cdi_finalize_step_proc  finalize_step, /* procedure to finalize step */
                               char *optstring
     )
{  
  W_cd_integrator_t *i;
  
  /* Create integrator */
  i = w_new(W_cd_integrator_t);
  CHECK_MALLOC(i);

  /* Domain */
  i->domain = domain;

  /* parameter */
  i->gamma = 0.5;
  
  /* Select time step dt */
  i->t  = t0;
  i->dt = dt;
  
  /* Initialize procedures */
  i->calc_eff_loads = calc_eff_loads;
  i->solve_for      = solve_for;
  i->update_config  = update_config;
  i->finalize_step  = finalize_step;
  i->advance        = Warburton_advance_step;
  
  return i;
}


/*
   New Newmark integrator.
   Create a new Newmark central-difference integrator.
 */
W_cd_integrator_t *
W_cd_new_Newmark_integrator (void *domain, /* domain */
     W_float_t               t0, /* initial time */
     W_float_t               dt, /* time step */
     W_cdi_calc_eff_loads_proc calc_eff_loads, /* procedure to calculate effective loads */
     W_cdi_solve_for_proc      solve_for, /* procedure to solve for the primary unknowns */
     W_cdi_update_config_proc  update_config, /* procedure to update configuration */
                             W_cdi_finalize_step_proc  finalize_step, /* procedure to finalize step */
                             char *optstring
     )
{  
  W_cd_integrator_t *i;
  
  /* Create integrator */
  i = w_new(W_cd_integrator_t);
  CHECK_MALLOC(i);

  /* Domain */
  i->domain = domain;
  
  /* parameter */
  i->gamma = 0.5;

  /* Select time step dt */
  i->t  = t0;
  i->dt = dt;
  
  /* Initialize procedures */
  i->calc_eff_loads = calc_eff_loads;
  i->solve_for      = solve_for;
  i->update_config  = update_config;
  i->finalize_step  = finalize_step;
  i->advance        = Newmark_advance_step;

  if (optstring) {
    int pos;
    tokens_parser parser = tokens_new_parser ();
    if (tokens_parse_line (parser, optstring)) {
      if ((pos = tokens_have_keyword (parser, 1, "GAMMA")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          i->gamma = tokens_token_as_double (parser, pos+1);
        }
      }
    }
    tokens_delete_parser (parser);
  }
  
  return i;
}

/*
  Advance the integrator.
   Advance the central-difference integrator in time by a single time step.
 */
void
W_cd_advance (W_cd_integrator_t *i /* integrator */)
{
  assert(i != NULL);
  assert(i->advance != NULL);
  (*i->advance)(i);
}

/*
  Get current time.
   Get the current time from the integrator.
 */
W_float_t /* time */
W_cd_get_time (W_cd_integrator_t *i/* integrator */)
{
  return i->t;
}

/*
  Get time step.
   Get the time step from the integrator.
 */
W_float_t                       /* time step */
W_cd_get_dt (W_cd_integrator_t *i/* integrator */)
{
  return i->dt;
}

double
W_cd_get_gamma (W_cd_integrator_t *i) { return i->gamma; }

void
W_cd_set_gamma (W_cd_integrator_t *i, double gamma) { i->gamma = gamma; }
