#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Welem.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wmat1d.h"
#include "Wmat1deng.h"
#include "Wmat1dstvk.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wcmass.h"

/* RCSHEADER "$Id: cmass.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct priv_elem_t {
  W_node_t         *node; /* node ID's */
  W_float_t         mass;       /* mass */
}              priv_elem_t;


static int
elem_elcode (W_elem_t *el)
{
  return CMASS_ELCODE;
}


#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

#ifdef X_XFACE
static EPixel color = 0;
/*
  Modify the graphic representation of the element (a 3D line).
 */
static void
modify_elem_grep (W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  WCRec pnts[1];
  W_coords_3d_t refc[1], u[1];
  priv_elem_t *t = (priv_elem_t *)el->data;

  GET_FROM_NODE (t->node, WN_get_ref_coords, refc[0]);
  GET_FROM_NODE (t->node, WN_get_displ, u[0]);

  VECMAC_3_VEC (pnts[0], =, refc[0], +, displ_scale * u[0]);

  if (el->grep == NULL) {
    el->grep = CreateMarker3D (pnts);
    if (color == 0) {
      BOOLEAN suc;
      color = ColorGetPixelFromString ("white", &suc);
      if (!suc) { REPORT_WARNING (el->domain->errh, "Color not available"); }
    }
    EASValsSetColor (color);
    EASValsSetMType (FILLED_CIRCLE_MARKER);
    EASValsSetMSize (8);
    EGWithMaskChangeAttributes (COLOR_MASK | MSIZE_MASK | MTYPE_MASK,
                                el->grep);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
  } else {
    EGModifyGraphicsByHandle(el->grep, 1, &pnts[0]);
  }
}

#endif

/*
  Returns a unmodifiable string with the element name.
 */
static char *
type_name (W_elem_t *el)
{
  static char buf[] = "cmass";
  return buf;
}

/*
  Assembles nodal forces by which the element acts on its nodes.
  All effects (restoring and damping forces) must be included.
 */
static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  el->deformation_energy = 0;   /* no restoring forces */
}


/*
  Compute the total mass of the element.
 */
static W_float_t
total_mass (W_elem_t *el)
{
  priv_elem_t *t= (priv_elem_t *)el->data;

  return t->mass;
}


/*
  Assemble mass to the nodes.
 */
static void
assemble_mass (W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  WN_assemble_mass (t->node, t->mass);
}


/*
  Compute the suggested time step.
 */
static W_float_t 
suggested_time_step (W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  return FLT_MAX;
}


static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                        W_mass_type_enum mass_type
                        )
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  double k[9] = {0,0,0,0,0,0,0,0,0};
  W_index_t idcs[3];
  W_bool_t constrained[3];

  /* mass matrix */
  k[0] = k[4] = k[8] = t->mass * mass_mult;

  WN_eqn_numbers_and_constraints (t->node, &idcs[0], &constrained[0]);
  (*assemble_eff_stiff_callback)  (el->domain->newmark_integrator,
                                   FLES_ADD, 3, idcs, k, constrained);
}


/*
  Assembles nodal forces by which the element acts on its nodes.
  All effects (restoring and damping forces) must be included.
 */
static void
assemble_residual_forces (W_elem_t *el,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_acceleration_3d_t a[1];

  GET_FROM_NODE (t->node, WN_get_acceleration, a[0]);
  VECMAC_SCALAR (a[0], *=, -(t->mass));
  WN_assemble_force (t->node, &a[0]);

  el->deformation_energy = 0;
}

static W_float_t
kinetic_energy (W_elem_t *el, W_mass_type_enum mass_type)
{
  priv_elem_t *e = (priv_elem_t *)el->data;
  W_velocity_3d_t v;
  GET_FROM_NODE (e->node, WN_get_velocity, v);
  return (e->mass * VECMAC_DOT (v, v)) / 2;
}

/*
  Element method dispatch table.
 */
static W_elem_methods_t methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  assemble_residual_forces,
  assemble_eff_stiffness,
  kinetic_energy
};

/*
  Read data for point mass.
  Reads element data from stream as long as there is no 
  END_OF_BLOCK tag in the input line.  Creates elements,
  and adds them to the domain.
  */
W_bool_t                        /* TRUE if successful */
cmass_reader (W_domain_t *d     /* domain */
              )
{
  const char *input_line;
  long elem_num, nodeid;
  priv_elem_t *t;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_vect_t loc;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      t = w_new(priv_elem_t);
      CHECK_MALLOC(t);
      TOKENS_PARSE(input_line);
      elem_num    = TOKEN_INT(1);
      nodeid      = TOKEN_INT(2);
      t->node     = WD_get_node_by_idx (d, nodeid);
      if (t->node == NULL) {
        sprintf (W_buf (), "No such node: %ld", nodeid);
        REPORT_FATAL_ERROR (d->errh, W_buf ());
      }
      t->mass = TOKEN_DBL(3);
#ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif
      WD_add_element(d, elem_num,
                     WE_new_element(d, &methods, t
#                    ifdef PVM_NEED_GHOST_ELEMENTS
                                    , is_ghost
#                    endif
                     ));
    }
  }
  return FALSE;
}
