#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dstvk.h"
#include "Wmat3drel.h"
#include "Wmat3dpf.h"
#include "Wmat3dhalq.h"
#include "Wmat3dneohk.h"
#include "Wmat3dunrot.h"
#include "Wmat3dssj2ep.h"
#include "Wmat3durd.h"

typedef struct ip_ctx_data_t {
  double x, y, z;
}              ip_ctx_data_t;

static double def_grad[3][3];
static double prev_def_grad[3][3];
static double rel_def_grad[3][3];
static double midstep_def_grad[3][3];
static double cauchy_green[3][3];
static double green_lagrange[3][3];
static double midstep_rate_of_def[3][3];
static double midstep_spin[3][3];
static double polar_R[3][3];
static double polar_U[3][3];
static double rel_polar_R[3][3];
static double rel_polar_U[3][3];
static double ln_rel_polar_U[3][3];
static double def_grad_dot[3][3];

#define conjugate(expr) (expr)

#define UNIAXIAL_STRETCHING 

#ifdef UNIAXIAL_STRETCHING
/*
  u=[sym('a*x*t') sym('0') sym('0')];

  displ_grad=[diff(u(1),'x'), diff(u(1),'y'), diff(u(1),'z');
              diff(u(2),'x'), diff(u(2),'y'), diff(u(2),'z');
              diff(u(3),'x'), diff(u(3),'y'), diff(u(3),'z')];

  def_grad=displ_grad+diag(ones(3,1));

  def_grad_dot= diff(def_grad, 't');

  ccode (def_grad)
  ccode (def_grad_dot)
*/

static void
compute_def_grads (double t, double dt, double x, double y, double z)
{
  const double a = 0.01;
  
  def_grad[0][0] = a*t+1.0;      def_grad[0][1] = 0.0;          def_grad[0][2] = 0.0;
  def_grad[1][0] = 0.0;          def_grad[1][1] = 1.0;          def_grad[1][2] = 0.0;
  def_grad[2][0] = 0.0;          def_grad[2][1] = 0.0;          def_grad[2][2] = 1.0;

#define def_grad prev_def_grad
#define t (t-dt)
  def_grad[0][0] = a*t+1.0;      def_grad[0][1] = 0.0;          def_grad[0][2] = 0.0;
  def_grad[1][0] = 0.0;          def_grad[1][1] = 1.0;          def_grad[1][2] = 0.0;
  def_grad[2][0] = 0.0;          def_grad[2][1] = 0.0;          def_grad[2][2] = 1.0;
#undef  def_grad   
#undef t
 
  def_grad_dot[0][0] = a;        def_grad_dot[0][1] = 0.0;      def_grad_dot[0][2] = 0.0;
  def_grad_dot[1][0] = 0.0;      def_grad_dot[1][1] = 0.0;      def_grad_dot[1][2] = 0.0;
  def_grad_dot[2][0] = 0.0;      def_grad_dot[2][1] = 0.0;      def_grad_dot[2][2] = 0.0;
}
#endif

static void
compute_kinem_quantities (W_mat_3d_ip_ctx_t *ctx)
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;
  double t = ctx->time;
  double dt = ctx->time_step;
  double x = ipctx->x;
  double y = ipctx->y;
  double z = ipctx->z;

  /* current and previous deformation gradient */
  compute_def_grads (t, dt, x, y, z);

  /* Cauchy-Green tensor */
  MAT3DUTIL_MULT_ATB_3X3 (cauchy_green, def_grad, def_grad);

  /* Green-Lagrange strain tensor */
  {
    double unit[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
    MAT3DUTIL_COMB_AB_3X3 (green_lagrange, 0.5, cauchy_green, - 0.5, unit);
  }

  /* relative deformation gradient */
  {
    double inv_prev_def_grad[3][3];
    double det_prev_def_grad = MAT3DUTIL_DET_3X3(prev_def_grad);
    if (det_prev_def_grad <= 0) {
      sprintf(W_buf(), "Failed to invert def grad");
      REPORT_FATAL_ERROR(NULL, W_buf());
    }
    
    MAT3DUTIL_INVERT_3X3(det_prev_def_grad, prev_def_grad, inv_prev_def_grad);
    
    MAT3DUTIL_MULT_AB_3X3(rel_def_grad, def_grad, inv_prev_def_grad);
  }

  /* midstep deformation gradient */
  MAT3DUTIL_COMB_AB_3X3 (midstep_def_grad, 0.5, def_grad, 0.5, prev_def_grad);

  /* midstep rate of deformation and spin */
  {
    int i, j;
    double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
    double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
    double s[3][3], inv_s[3][3], q[3][3], r[3][3];
    double inv_dt = 1 / dt;
    
    MAT3DUTIL_ADD_AB_3X3 (s, m, rel_def_grad);
    MAT3DUTIL_SCALE_3X3 (s, 0.5);
    if (!W_mat3dutil_invert_3x3(s, inv_s)) {
      sprintf(W_buf(), "Failed to invert def grad");
      REPORT_FATAL_ERROR(NULL, W_buf());
    }
    MAT3DUTIL_ADD_AB_3X3 (q, rel_def_grad, M);
    MAT3DUTIL_MULT_AB_3X3 (r, inv_s, q);
    
    MAT3DUTIL_SYMM  (midstep_rate_of_def, r);
    MAT3DUTIL_ASYMM (midstep_spin, r);
    
    MAT3DUTIL_SCALE_3X3 (midstep_rate_of_def, inv_dt);
    MAT3DUTIL_SCALE_3X3 (midstep_spin, inv_dt);
  }

  /* polar decomposition */
  if (!W_polar_decomposition (def_grad, polar_R, polar_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }

  /* relative polar decomposition */
  if (!W_polar_decomposition (rel_def_grad, rel_polar_R, rel_polar_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }

  /* logarithm of the relative deformation tensor */
  W_ln_U (rel_polar_U, ln_rel_polar_U);
}

static W_bool_t 
get_kinem_var(W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3])
{
  compute_kinem_quantities (ctx);
  switch (type) {
  case MAT_3D_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, def_grad);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    MAT3DUTIL_COPY_3X3(kinvar, def_grad_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, rel_def_grad);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, prev_def_grad);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, midstep_def_grad);
    break;
  case MAT_3D_CAUCHY_GREEN:
    MAT3DUTIL_COPY_3X3(kinvar, cauchy_green);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    MAT3DUTIL_COPY_3X3(kinvar, midstep_rate_of_def);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    MAT3DUTIL_COPY_3X3(kinvar, midstep_spin);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    MAT3DUTIL_COPY_3X3(kinvar, green_lagrange);
    break;
  case MAT_3D_POLAR_R:
    MAT3DUTIL_COPY_3X3(kinvar, polar_R);
    break;
  case MAT_3D_POLAR_U:
    MAT3DUTIL_COPY_3X3(kinvar, polar_U);
    break;
  case MAT_3D_POLAR_REL_R:
    MAT3DUTIL_COPY_3X3(kinvar, rel_polar_R);
    break;
  case MAT_3D_POLAR_REL_U:
    MAT3DUTIL_COPY_3X3(kinvar, rel_polar_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    MAT3DUTIL_COPY_3X3(kinvar, ln_rel_polar_U);
    break;
  default:
    {
      static W_bool_t complained = FALSE;
      if (!complained) {
        sprintf (W_buf(), "Unknown kinematic tensor type: %d",     (int)type);
        REPORT_WARNING (NULL, W_buf());
        complained = TRUE;
      }
    }
    return FALSE;
  }

  return TRUE;
}

static void
test (void)
{
  const double st = 0;          /* start time */
  const double ft = 1.0;        /* finish time */
  const int nsteps = 1;
  double dt = (ft - st) / nsteps;
  double t;
  double stress[3][3]; 
  W_vect_t loc = {0,0,0};
  ip_ctx_data_t ip_ctx_data;
  const W_mat_3d_stress_type_enum stress_type = MAT_3D_2ND_PK;
  W_mat_3d_stress_type_enum out_stress_type;
  W_mat_3d_t *m = new_mat_3d_stv_k (2.1e9, 0.15, 0, 0, 7.850);
  W_mat_3d_state_t  *matst = WM3D_get_mat_state (m, &loc);
  W_mat_3d_ip_ctx_t ip_ctx;
  
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
  ip_ctx_data.x = 
  ip_ctx_data.y = 
  ip_ctx_data.z = 0;

  for (t = st+dt; t <= ft; t += dt) {
    ip_ctx.time          = t;
    ip_ctx.time_step     = dt;
    if (!WM3D_update_stress (m, matst, W_MAT_STATE_UPDATE_FINAL,
                             &ip_ctx, stress_type, stress, &out_stress_type))
      REPORT_FATAL_ERROR (NULL, "Update failed");
    if (out_stress_type != stress_type)
      REPORT_FATAL_ERROR (NULL, "Did not get requested stress");
    printf ("%g %g\n", t, stress[0][0]);
  }
}

int
main (int argc, char **argv)
{
  test ();
}
