#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "Wdomain.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGcr.h"
#include "WEFGtc.h"
#include "Wmat3d.h"
#include "WEFGsf.h"
#include "WEFGss.h"
#include "Whistry.h"




#define USE_BBT 1


/* Auxiliary crack surface */
typedef struct auxiliary_crack_surf_t {
  WEFG_crack_surf_t                    *crack_surf; /* generic crack surface */
  G3D_quad_t                            quad; /* quadrilateral defining the
                                               auxiliary crack */
  G3D_box_t                             bbox; /* bounding box of the crack */
}              auxiliary_crack_surf_t;



static W_bool_t 
segment_intersects_crack_surf(WEFG_eval_for_packet_t *efp,
                              G3D_segment_t *s, WEFG_node_t *efgn,
                              WEFG_crack_surf_t *cs,
                              double  tol);
static WEFG_pt_pos_wrt_crack_surf_enum
node_pos_wrt_crack_surf(WEFG_crack_surf_t *cs,
                        WEFG_node_t *efgn,
                        double physical_tol);
static W_bool_t 
hex_intersected_by_crack_surf(WEFG_crack_surf_t *cs, 
                              WEFG_hex_volume_t *h
                              );



static WEFG_crack_surf_methods_t methods = {
  segment_intersects_crack_surf,
  NULL,
  NULL,
  node_pos_wrt_crack_surf,
  NULL,
#ifdef X_XFACE
  NULL,
#endif
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  hex_intersected_by_crack_surf
};




/*
   Auxiliary crack reader.
   See User's guide for details.
 */
WEFG_crack_surf_t *
read_auxiliary_crack_surf(WEFG_superelem_t *se, /* EFG superelement */
                          W_domain_t *d /* input stream */
                          )
{
  const char *input_line;
  WEFG_crack_surf_t *cs = NULL;
  auxiliary_crack_surf_t *ac = NULL;
  int kp;
  double infl;

  while ((input_line = W_get_input_line(d)) != NULL) {
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD("END")) {
      if (TOKENS_HAVE_KEYWORD("AUXILIARY")) {
        return cs;
      } else
        REPORT_FATAL_ERROR(d->errh, "AUXILIARY does not end by END AUXILIARY");
    } else {
      if        (TOKENS_HAVE_KEYWORD("ID")) {
        int id;
        id = TOKEN_INT(2);
        ac = w_new(auxiliary_crack_surf_t);
        CHECK_MALLOC(ac);
	if ((kp = tokens_have_keyword(d->parser, 1, "QUAD")) > 0) {
          ac->quad.loc[0].x = TOKEN_DBL(kp+1);
          ac->quad.loc[0].y = TOKEN_DBL(kp+2);
          ac->quad.loc[0].z = TOKEN_DBL(kp+3);
          kp += 3;
          ac->quad.loc[1].x = TOKEN_DBL(kp+1);
          ac->quad.loc[1].y = TOKEN_DBL(kp+2);
          ac->quad.loc[1].z = TOKEN_DBL(kp+3);
          kp += 3;
          ac->quad.loc[2].x = TOKEN_DBL(kp+1);
          ac->quad.loc[2].y = TOKEN_DBL(kp+2);
          ac->quad.loc[2].z = TOKEN_DBL(kp+3);
          kp += 3;
          ac->quad.loc[3].x = TOKEN_DBL(kp+1);
          ac->quad.loc[3].y = TOKEN_DBL(kp+2);
          ac->quad.loc[3].z = TOKEN_DBL(kp+3);

          G3D_INIT_BOX(ac->bbox, ac->quad.loc[0]);
          G3D_UPDT_BOX(ac->bbox, ac->quad.loc[1]);
          G3D_UPDT_BOX(ac->bbox, ac->quad.loc[2]);
          G3D_UPDT_BOX(ac->bbox, ac->quad.loc[3]);
          infl = (G3D_BOX_RANGE(ac->bbox, x) +
                  G3D_BOX_RANGE(ac->bbox, y) +
                  G3D_BOX_RANGE(ac->bbox, z)) / 100;
          G3D_INFLATE_BOX(ac->bbox, infl);
        } else {
          REPORT_FATAL_ERROR(d->errh, "No geometry data");
        }
        cs = efgcr_new_crack_surf(se, id, ac, &methods);
        ac->crack_surf = cs;
      }
    }
  }
  REPORT_FATAL_ERROR(d->errh, "AUXILIARY does not end by END AUXILIARY");
  return NULL;
}

/*
   Segment intersects crack surface?
   Does a segment intersect the crack surface?

   The tolerance tol is normalized by the segment length --> eps. If intersection detected
 for |lambda| < eps or |1-lambda| < eps, the intersection point is considered
 to lie on the crack surface (lambda being the parameterization of the segment;
 lambda=0 corresponds to the evaluation point, lambda=1 corresponds to the node).
 */
static W_bool_t 
segment_intersects_crack_surf(WEFG_eval_for_packet_t *efp, /* evaluation packet */
                              G3D_segment_t *s, /* segment */
                              WEFG_node_t *efgn, /* EFG node */
                              WEFG_crack_surf_t *cs, /* crack surface */
                              double  tol /* tolerance; see writeup */
                              )
{
  auxiliary_crack_surf_t *ac = (auxiliary_crack_surf_t *)cs->data;
  G3D_vect_t ab, ip;
  double lambda;
  double eps, ab_length;
  G3D_box_t segbox;

  G3D_INIT_BOX(segbox, s->start);
  G3D_UPDT_BOX(segbox, s->end);

  if (!G3D_BOXES_OVERLAP(segbox, ac->bbox))
    return FALSE;
  
  /* Segment start is the evaluation point, end is the EFG node */
  VECMAC_3_VEC(ab, =, s->end, -, s->start);

  ab_length = VECMAC_NORM(ab);
  if (ab_length > 0)
    eps       = tol / ab_length;             /* gives d_lambda */
  else
    eps       = 0.00001;

  if (G3D_intersect_quad_and_line(&ip, &lambda, &s->start, &ab,
                                  &ac->quad, eps)) {
    return (lambda >= eps && lambda <= 1 - eps);
  } 
  
  return FALSE;
}

/*
  What is the position of the last found intersection.
  What is the position of the last found intersection with the crack surface?

  warning:

  This function is a quick and dirty solution really, it just returns
  `I don't know'.
 */
static WEFG_pt_pos_wrt_crack_surf_enum
node_pos_wrt_crack_surf(WEFG_crack_surf_t *cs,
                        WEFG_node_t *efgn,
                        double physical_tol)
{
  /* RAW */
  return WEFG_POS_WRT_SURF_UNSPEC;
}

/*
   Is integration volume intersected by the crack?
   Is integration volume intersected by the crack?  Check for all hex edges.
 */
static W_bool_t 
hex_intersected_by_crack_surf(WEFG_crack_surf_t *cs,  /* crack surface */
                              WEFG_hex_volume_t *h /* integration volume */
                              )
{
  static int first = TRUE;
  if (first)
    REPORT_WARNING(cs->se->domain->errh, "not yet implemented");
  first = FALSE;
  return FALSE;
}
  
#endif /* EFG_WASP  */
  
 
