#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "Wdomain.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGtc.h"
#include "WEFGac.h"

/* RCSHEADER "$Id: efgcr.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

/*
   Read crack data.
   Read crack data from the input stream.  See the User's manual for details
   on the syntax.
 */
void
read_crack (WEFG_superelem_t *se, /* superelement */
            W_domain_t *d
            ) 
{
  const char *input_line;

  while ((input_line = W_get_input_line (d)) ) {
    TOKENS_PARSE (input_line);
    if (TOKENS_HAVE_KEYWORD ("END")) {
      if (TOKENS_HAVE_KEYWORD ("CRACK")) 
        return;
      else
        REPORT_FATAL_ERROR (se->domain->errh, "CRACK does not end by END CRACK");
    }  else  {
      WEFG_crack_surf_t *cs;

      cs = NULL;
      if (TOKENS_HAVE_KEYWORD ("TRIANGULATED")) {
        cs = read_triangulated_crack_surf (se, d);
      }  else if (TOKENS_HAVE_KEYWORD ("AUXILIARY")) {
        cs = read_auxiliary_crack_surf (se, d);
      }  else  {
        REPORT_FATAL_ERROR (se->domain->errh, "Unknown crack type");
      } 
      if (cs) 
        attach_crack_surf (se, cs);  
    } 
  } 
  REPORT_FATAL_ERROR (se->domain->errh, "CRACK does not end by END CRACK");
} 



/*
   Create a new crack surface.
   Create a generic crack surface.
 */
WEFG_crack_surf_t * /* crack surface */
efgcr_new_crack_surf (WEFG_superelem_t           *se, /* EFG superelement */
                      int                         id, /* unique integer among all cracks */
                      WEFG_crack_surf_data_t      data, /* instance data */
                      WEFG_crack_surf_methods_t  *methods /* instance methods */
                      ) 
{ 
  WEFG_crack_surf_t *cs;

  cs = w_new (WEFG_crack_surf_t);
  CHECK_MALLOC (cs);

  cs->se                = se;
  cs->have_intersection = FALSE;
  cs->id                = id;
  cs->data              = data;
  cs->methods           = methods;
  cs->changed_shape     = FALSE;

  return cs;
} 


 
/*
   Segment intersects crack surface?
   Does a segment intersect the crack surface?

   The tolerance tol is normalized by the segment length --> eps. If intersection detected
 for |lambda| < eps or |1-lambda| < eps, the intersection point is considered
 to lie on the crack surface (lambda being the parameterization of the segment;
 lambda=0 corresponds to the evaluation point, lambda=1 corresponds to the node).

 requirements:

 The /segment_intersects_crack_surf/ must be defined by the crack surface
 instance.
 */
W_bool_t 
efgcr_segment_intersects_crack_surf (WEFG_eval_for_packet_t *efp, /* evaluation packet */
                                     G3D_segment_t *s, /* segment */
                                     WEFG_node_t *efgn, /* EFG node at the end of the segment */
                                     WEFG_crack_surf_t *cs, /* generic crack surface */
                                     double  tol /* tolerance in physical units
                                                  */
                                     ) 
{
  return
 (cs->have_intersection =
 (*cs->methods->segment_intersects_crack_surf) (efp, s, efgn, cs, tol));
} 


/*
   Position of node wrt crack surface.
   What is the position of a node with respect to the crack surface?

   The tolerance tol is normalized by the segment length --> eps. If intersection detected
 for |lambda| < eps or |1-lambda| < eps, the intersection point is considered
 to lie on the crack surface (lambda being the parameterization of the segment;
 lambda=0 corresponds to the evaluation point, lambda=1 corresponds to the node).

 
 requirements:

 The /segment_intersects_crack_surf/ must be defined by the crack surface
 instance.
 */
WEFG_pt_pos_wrt_crack_surf_enum
efgcr_node_pos_wrt_crack_surf (WEFG_crack_surf_t *cs, /* generic crack surface */
                              WEFG_node_t *efgn, /* EFG node */
                              double physical_tol /* physical tolerance */
                               ) 
{
  return (*cs->methods->node_pos_wrt_crack_surf) (cs, efgn, physical_tol);
} 




#ifdef X_XFACE
/* modify graphic representation.
   Modify the graphic representation of the crack to account for any growth,
   change in displacement scale etc.
 */
void
efgcr_modify_crack_surf_grep (WEFG_superelem_t *se, /* EFG superelement */
                              double displ_scale,
                              WEFG_crack_surf_t *cs /* generic crack surface */
                              ) 
{
  if (cs->methods->modify_crack_surf_grep) 
    (*cs->methods->modify_crack_surf_grep) (se, displ_scale, cs);
} 
#endif

/* Propagate crack.
   Propagate crack.  See specific instances of this method
   for details on how the crack is propagated.

   
 requirements:

 The /segment_intersects_crack_surf/ must be defined by the crack surface
 instance eventhough the instance does not support crack propagation.
 */
void
efgcr_propagate (WEFG_crack_surf_t *cs /* generic crack */
                 ) 
{
  if (cs->methods->propagate) 
    (*cs->methods->propagate) (cs);
} 


/* write HEXVUE file.
   write a HEXVUE post-processing file for the crack.
 */
void
efgcr_write_hexvue_postproc_file (WEFG_crack_surf_t *cs, /* generic crack */
                                 ckit_iobuf_t IOB, /* I/O buffer */
                                  char *mode /* write in mode */
                                  ) 
{
  if (cs->methods->write_hexvue_postproc) 
    (*cs->methods->write_hexvue_postproc) (cs, IOB, mode);
} 

/* write TECPLOT file.
   write a TECPLOT post-processing file for the crack.
 */
void 
efgcr_write_tecplot_postproc_file (WEFG_crack_surf_t *cs, /* generic crack */
                                  ckit_iobuf_t IOB, /* I/O buffer */
                                   char *mode, /* write in mode */
                                  void *optdata /* optional data */
                                   ) 
{
  if (cs->methods->write_tecplot_postproc) 
    (*cs->methods->write_tecplot_postproc) (cs, IOB, mode, optdata);
} 

/* update stress.
 Update stress in all points on the crack surface.
 */
void
efgcr_update_stress (WEFG_superelem_t *se, /* EFG superelement */
                     WEFG_crack_surf_t *cs/* generic crack */) 
{
  if (cs->methods->update_stress) 
    (*cs->methods->update_stress) (se, cs);
} 

/* write VE3D loclist. */
void
efgcr_write_ve3d_loclist (WEFG_crack_surf_t *cs, /* generic crack */
                          ckit_iobuf_t IOB
                          ) 
{
  if (cs->methods->write_ve3d_loclist) 
    (*cs->methods->write_ve3d_loclist) (cs, IOB);
} 

/* write VE3D polygons. */
void
efgcr_write_ve3d_polys (WEFG_crack_surf_t *cs, /* generic crack */
                        ckit_iobuf_t IOB
                        ) 
{
  if (cs->methods->write_ve3d_polys) 
    (*cs->methods->write_ve3d_polys) (cs, IOB);
} 




typedef struct crack_history_client_data_t  {
  WEFG_crack_surf_t     *cs;
  void                  *data;
}               crack_history_client_data_t;


/* create a new crack tip sampler
   Create a new sampler for a crack front point /loc/, which is referred to
   by /pntidx/.
 */
void * /* sampler or NULL if crack instance does not support sampling */
efgcr_new_sampler (WEFG_crack_surf_t *cs, /* generic crack */
                   W_index_t pntidx, /* sampler id */
                   W_loc_t *loc /* location; must be on crack surface */
                   ) 
{
  crack_history_client_data_t *cd;
  void *data;
  
  if (cs->methods->new_sampler) {
    data = (*cs->methods->new_sampler) (cs, pntidx, loc);
    if (data) {
      cd = w_new (crack_history_client_data_t);
      CHECK_MALLOC (cd);
      cd->cs   = cs;
      cd->data = data;
      return cd;
    } 
  }  
  return NULL;
} 



/*
   Store sampler data.
 */
void 
efgcr_sampler_store (W_elem_t *e, /* generic element */
		    W_solution_history_t *h, /* history buffer */
		    void *client, /* client data */
		    double time
                     ) 
{
  crack_history_client_data_t *cd = (crack_history_client_data_t *) client;
  WEFG_crack_surf_t *cs = cd->cs;

  if (cs->methods->sampler_store) 
    (*cs->methods->sampler_store) (cs, h, cd->data, time);
} 



/* wrap-around point.
   compute the wrap-around point.  The meaning of the parameters is documented
   in the ESFLib manual.
 */
int /* TRUE if wrap-around point valid; FALSE otherwise */
efgcr_wrap_around_point (WEFG_crack_surf_t *cs, /* generic crack */
                        double atx, /* evaluation point coordinate */
                         double aty, /* evaluation point coordinate */
                         double atz, /* evaluation point coordinate */
                        double rcx, /* EFG node coordinate */
                         double rcy, /* EFG node coordinate */
                         double rcz, /* EFG node coordinate */
                        double bx, /* plane-generating vector coordinate */
                         double by, /* plane-generating vector coordinate */
                         double bz, /* plane-generating vector coordinate */
                        double *wapx, /* wrap-around point coordinate */
                         double *wapy, /* wrap-around point coordinate */
                         double *wapz /* wrap-around point coordinate */
                         ) 
{
  if (cs->methods->wrap_around_point) 
    return (*cs->methods->wrap_around_point) (cs, atx, aty, atz, rcx, rcy, rcz,
                                             bx, by, bz, wapx, wapy, wapz);
  else
    return 0;
} 



/*
   Plane-generating vector.
   Compute plane-generating vector for the crack surface
   at the point atx,...,atz.  The meaning of the parameters is documented
   in the ESFLib manual.
 */
int /* TRUE if vector coordinates valid; FALSE otherwise */
efgcr_plane_generating_vector (WEFG_crack_surf_t *cs, 
                               double atx, /* evaluation point coordinate */
                               double aty, /* evaluation point coordinate */
                               double atz, /* evaluation point coordinate */
                               double rcx, /* EFG node coordinate */
                         double rcy, /* EFG node coordinate */
                         double rcz, /* EFG node coordinate */
                        double *bx, /* output: plane-generating vector coordinate */
                         double *by, /* output: plane-generating vector coordinate */
                         double *bz /* output: plane-generating vector coordinate */
                               ) 
{
  if (cs->methods->plane_generating_vector) {
    return (*cs->methods->plane_generating_vector) (cs, atx, aty, atz,
                                                   rcx, rcy, rcz,
                                                   bx, by, bz);
  }  else  { 
    *bx = *by = *bz = 0;
    return 0;
  } 
} 



/*
   Integration volume intersected by crack surface?
   Is the integration volume intersected by the crack surface?
 */
W_bool_t 
efgcr_hex_intersected_by_crack_surf (WEFG_crack_surf_t *cs, /* generic crack */
                                    WEFG_hex_volume_t *h /* integration volume */
                                    ) 
{
  if (cs->methods->hex_intersected_by_crack_surf) 
    return (*cs->methods->hex_intersected_by_crack_surf) (cs, h);
  else
    return FALSE;
} 

#endif /*
EFG_WASP */
