#include "Wconfig.h"

#ifdef EFG_WASP

#include "WEFG.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFGutil.h"
#include "WEFGgc.h"
#include "WEFGgt.h"
#include "WEFGgq.h"
#include "WEFGpriv.h"
#include "WEFGcr.h"

/* RCSHEADER "$Id: efggc.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

static void
check_planarity(WEFG_b_face_t *f);

/*
   Create a copy of the boundary face on input.
 */
WEFG_b_face_t * /* face or NULL (no memory) */
copy_b_face(WEFG_b_face_t *f /* face to be copied */
           )
{
  WEFG_b_face_t *newf;
  int i;

  newf = w_new(WEFG_b_face_t);
  CHECK_MALLOC(newf);

  newf->num_vert = f->num_vert;
  if (newf->num_vert > 4)
    REPORT_FATAL_ERROR(NULL, "Cannot handle face with more than 4 vertices");
  
  for (i = 0; i < newf->num_vert; i++) {
    newf->v[i] = f->v[i];
  }
  check_planarity(f);
  return newf;
}


static int
quad_is_planar(WEFG_b_face_t *f);

static void
check_planarity(WEFG_b_face_t *f)
{
  if        (f->num_vert == 3) {
    f->planar = TRUE;
  } else if (f->num_vert == 4) {
    f->planar = quad_is_planar(f);
  }
}


#define WARP_FRACT 0.2

/*
   Is the quadrilateral face on input `planar?'
 */
static int /* TRUE if planar; FALSE otherwise */
quad_is_planar(WEFG_b_face_t *f /* boundary face */
               )
{
  double l;
  W_vect_t v, n;
  G3D_plane_t P;
  
  G3D_plane_consts_3(&P, &f->v[0]->loc, &f->v[1]->loc, &f->v[2]->loc);
  VECMAC_3_VEC(v, =, f->v[3]->loc, -, f->v[1]->loc);
  l = VECMAC_NORM(v);
  n = P.n;
  VECMAC_NORMALIZE(n);
  return (fabs(VECMAC_DOT(n, v)) > WARP_FRACT * l);
}


static W_bool_t
segment_intersects_face(G3D_segment_t *s, WEFG_b_face_t *f, double eps);



/*
   Visibility criterion check.
   Is the EFG node /efgp/ hidden from an evaluation point by the boundary?

   
   The tolerance tol is normalized by the segment length --> eps. If intersection detected
 for |lambda| < eps or |1-lambda| < eps, the intersection point is considered
 to lie on the crack surface (lambda being the parameterization of the segment;
 lambda=0 corresponds to the evaluation point, lambda=1 corresponds to the node).

 algorithm:

 If the superelement is convex (that is specified as a hint), nothing is
 checked and FALSE is returned immediately.

 If the superelement is `convex with cracks', possible intersections
 of the segment with boundary faces are ignored, and the only checks
 done are intersections with crack surfaces; otherwise, if the superelement
 is non-convex, intersections are first sought with the boundary faces, and
 only then the crack surfaces are checked.

 If no intersections have been detected, the node is visible and the function
 returns FALSE;
 */
W_bool_t /* TRUE if EFG node `invisible'; FALSE otherwise */
node_hidden_by_boundary (SFL_generic_ptr_t eval_for,
                         double seg_start_x, /* x-coordinate of the evaluation point */
                         double seg_start_y, /* y-coordinate of the evaluation point */
                         double seg_start_z , /* z-coordinate of the evaluation point */
                         SFL_efg_node_id_t efgp, /* EFG node */
                         double seg_end_x, /* x-coordinate of the EFG node */
                         double seg_end_y,/* y-coordinate of the EFG node */
                         double seg_end_z ,/* z-coordinate of the EFG node */
                         double  eps /* tolerance in physical units; see
                                      DESCRIPTION */
                         )
{
  WEFG_eval_for_packet_t *efp = (WEFG_eval_for_packet_t *)eval_for;
  WEFG_superelem_t *se = efp->se;
  WEFG_node_t *efgn = (WEFG_node_t *)efgp;
  WEFG_b_face_t *f;
  G3D_segment_t s;
  WEFG_crack_surf_t *cs;

  /* If the body is convex, nothing needs to be checked */
  if (se->is_convex) return FALSE;
  
  s.start.x = seg_start_x;
  s.start.y = seg_start_y;
  s.start.z = seg_start_z;
  s.end.x   = seg_end_x;
  s.end.y   = seg_end_y;
  s.end.z   = seg_end_z;
  
  /* NOTE: It is assumed that the faces constituting the boundary */
  /* are convex polygons (triangles and reasonably shaped quadrilaterals */

  /* If the body is convex with cracks, there is no need to check for */
  /* intersections with the boundary faces. */

  if (!se->is_convex_with_cracks) {
    loop_through_list(se->b_faces, f, WEFG_b_face_t *) {
      if (segment_intersects_face(&s, f, eps))
        return TRUE;
    }
  }

  /* Now check for intersections with the crack faces */

  loop_through_list(se->cracks, cs, WEFG_crack_surf_t *) {
    if (efgcr_segment_intersects_crack_surf(efp, &s, efgn, cs, eps))
      return TRUE;
  }

  return FALSE;
}


/*
   Compute the plane-generating vector.

   Compute the plane-generating vector: if the superelement is convex,
   no plane generating vector can be correct; otherwise, check which
   crack surface detected the last intersection and let it handle
   the request.
 */
int /* TRUE if vector on output valid */
efggc_plane_generating_vector(SFL_generic_ptr_t eval_for, /* evaluation package */
                              double atx, /* evaluation point coordinate */
                              double aty, /* evaluation point coordinate */
                              double atz, /* evaluation point coordinate */
                              double rcx, /* EFG node coordinate */
                              double rcy, /* EFG node coordinate */
                              double rcz, /* EFG node coordinate */
                              double *bx, /* output: plane-generating vector coordinate */
                              double *by, /* output: plane-generating vector coordinate */
                              double *bz /* output: plane-generating vector coordinate */
                              )
{
  WEFG_eval_for_packet_t *efp = (WEFG_eval_for_packet_t *)eval_for;
  WEFG_superelem_t *se = efp->se;
  WEFG_crack_surf_t *cs;

  *bx = *by = *bz = 0;
  /* If the body is convex, no plane generating vector can be correct */
  if (se->is_convex) return FALSE;

  /* Let the crack surface which detected the last intersection
     handle the request
   */
  loop_through_list(se->cracks, cs, WEFG_crack_surf_t *) {
    if (cs->have_intersection)
      return efgcr_plane_generating_vector(cs,
                                           atx, aty, atz,
                                           rcx, rcy, rcz,
                                           bx, by, bz);
  }  
  return FALSE;
}


/*
   Use wrap-around?
   Should wrap-around smoothing of EFG shape functions be used?

   caveats:

   If there are some cracks defined, the function returns TRUE without
   further considerations.
 */
int 
efggc_use_cpp(SFL_generic_ptr_t eval_for, SFL_efg_node_id_t efgp)
{
  WEFG_eval_for_packet_t *efp = (WEFG_eval_for_packet_t *)eval_for;
  WEFG_superelem_t *se = efp->se;

  if (!se->use_wrap_around) {
    REPORT_FATAL_ERROR(se->domain->errh, "Why are we here?");
  }
  return (se->cracks != NULL && list_length(se->cracks) > 0); 
}

/*
   Compute the wrap-around point.
   Compute the wrap-around point for the crack which detected
   the last intersection.
 */
int 
efggc_wrap_around_point(SFL_generic_ptr_t eval_for, /* evaluation package */
                        double atx, /* evaluation point coordinate */
                         double aty, /* evaluation point coordinate */
                         double atz, /* evaluation point coordinate */
                        double rcx, /* EFG node coordinate */
                         double rcy, /* EFG node coordinate */
                         double rcz, /* EFG node coordinate */
                        double bx, /* plane-generating vector coordinate */
                         double by, /* plane-generating vector coordinate */
                         double bz, /* plane-generating vector coordinate */
                        double *wapx, /* wrap-around point coordinate */
                         double *wapy, /* wrap-around point coordinate */
                         double *wapz /* wrap-around point coordinate */
                        )
{
  WEFG_eval_for_packet_t *efp = (WEFG_eval_for_packet_t *)eval_for;
  WEFG_superelem_t *se = efp->se;
  WEFG_crack_surf_t *cs; 

  *wapx = *wapy = *wapz = -FLT_MAX; /* invalid point */
  loop_through_list(se->cracks, cs, WEFG_crack_surf_t *) {
    if (cs->have_intersection)
      return efgcr_wrap_around_point(cs,
                                     atx, aty, atz,
                                     rcx, rcy, rcz,
                                     bx, by, bz,
                                     wapx, wapy, wapz);
  }  
  return FALSE;
}




#define QP(DIR)                                    \
   (0.25 * (  f->v[0]->loc.DIR + f->v[1]->loc.DIR  \
            + f->v[2]->loc.DIR + f->v[3]->loc.DIR))

/*
   Does segment intersect boundary face?
   Does segment intersect boundary face?
 */
W_bool_t /* TRUE if intersection point on output valid */
efggc_segment_intersection_with_face(G3D_segment_t *s, /* segment */
                                     WEFG_b_face_t *f, /* face */
                                     W_bool_t line_intersection, /* compute
                                                                  line-polygon
                                                                  intersection rather than segment-polygon intersection */
                                     W_loc_t *ip, /* output: intersection point */
                                     double *lambda /* output:
                                                     parameter in the unit
                                                     range to intersection
                                                     */
                                     )
{
  W_vect_t ab;
  G3D_quad_t q;
  G3D_tri_t t;
  double tol, ab_length;
  G3D_box_t segbb;
#define FPT(n) f->v[n]->loc
#define TPT(n) t.loc[n]
#define QPT(n) q.loc[n]
#define ASS(bb, p0, p1, dir)                            \
  if (p0->dir <= p1->dir) {                             \
    bb.lower.dir = p0->dir; bb.upper.dir = p1->dir;     \
  } else {                                              \
    bb.lower.dir = p1->dir; bb.upper.dir = p0->dir;     \
  }
#define SEGBB(seg, bb)                                   \
  {                                                      \
    register W_loc_t *p0 = &seg->start, *p1 = &seg->end; \
                                                         \
    ASS(bb, p0, p1, x);                                  \
    ASS(bb, p0, p1, y);                                  \
    ASS(bb, p0, p1, z);                                  \
  }
#define BOXES_OVERLAP(r1, r2)                                         \
         (   (r1.lower.x <= r2.upper.x) && (r1.upper.x >= r2.lower.x) \
          && (r1.lower.y <= r2.upper.y) && (r1.upper.y >= r2.lower.y) \
          && (r1.lower.z <= r2.upper.z) && (r1.upper.z >= r2.lower.z) \
          )
  
  /* (i)   Find intersection of the segment with the plane of the */
  /*       face.  If there is no intersection, return FALSE; otherwise */
  /* (ii)  Test whether the intersection plane is inside the face */
  /*       (in its plane). */

  if (!line_intersection) {
    /* first check bounding boxes, but only for SEGMENT intersections, not */
    /* for LINE intersections. */
    SEGBB(s, segbb);
    if (!(BOXES_OVERLAP(segbb, f->bounding_box)))
      return FALSE;
  }
  
  VECMAC_3_VEC(ab, =, s->end, -, s->start);
  ab_length = VECMAC_NORM(ab);
  tol = ab_length / 1000;

  if        (f->num_vert == 3) {
    TPT(0) = FPT(0); TPT(1) = FPT(1); TPT(2) = FPT(2);
    return G3D_intersect_tri_and_line(ip, lambda, &s->start, &ab, &t, tol);
  } else if (f->num_vert == 4) {
    QPT(0) = FPT(0); QPT(1) = FPT(1); QPT(2) = FPT(2); QPT(3) = FPT(3);
    return G3D_intersect_quad_and_line(ip, lambda, &s->start, &ab, &q, tol);
  } else {
    REPORT_FATAL_ERROR(NULL, "unexpected face type");
    return FALSE;
  }
}


/*
   Does segment intersect boundary face?
   Does a segment intersect a boundary face?

   
   The tolerance eps is normalized by the segment length, L. If
   intersection detected for |lambda| < eps/L or |1-lambda| < eps/L, the
   intersection point is considered to lie on the crack surface
   (lambda being the parameterization of the segment; lambda=0
   corresponds to the evaluation point, lambda=1 corresponds to the
   node).

 */
static W_bool_t
segment_intersects_face(G3D_segment_t *s, /* segment */
                        WEFG_b_face_t *f, /* face */
                        double eps /* physical tolerance */
                        )
{
  W_vect_t ab, ip;
  W_float_t lambda;
  double ab_length;
  
  VECMAC_3_VEC(ab, =, s->end, -, s->start);
  ab_length = v3_length(&ab);

  if (!efggc_segment_intersection_with_face(s, f, FALSE, &ip, &lambda))
    return FALSE;
  
  eps /= ab_length;             /* gives d_lambda */
  return (lambda >= eps && lambda <= 1-eps);
}


/*
   Compute face normal.
   Compute the boundary face normal.
 */
void
efggc_face_normal (WEFG_b_face_t *f, /* face */
                   W_vect_t *n /* output: unit normal vector */
                   )
{
  G3D_plane_t plane;
  
  if        (f->num_vert == 3) {
    G3D_plane_consts_3 (&plane,
                        &f->v[0]->loc,
                        &f->v[1]->loc,
                        &f->v[2]->loc);
    *n = plane.n;
    VECMAC_NORMALIZE ((*n));
  } else if (f->num_vert == 4) {
    G3D_plane_consts_4 (&plane,
                        &f->v[0]->loc,
                        &f->v[1]->loc,
                        &f->v[2]->loc,
                        &f->v[3]->loc);
    *n = plane.n;
    VECMAC_NORMALIZE ((*n));
  } else {
    REPORT_FATAL_ERROR (NULL, "unexpected face type");
    VECMAC_SCALAR ((*n), =, 0);
  }
}



#endif /* EFG_WASP  */
