#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "Wdomain.h"
#include "Wmat3d.h"
#include "Wmat3dutil.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGii.h"
#include "WEFGss.h"
#include "Wnode.h"
#include "g3d.h"
#include "g3dhex.h"


/* RCSHEADER "$Id: efgii.c,v 1.4 2003/11/25 16:55:47 cvs Exp $" */


#ifndef COMPUTE_AUX_SIFS
#   define COMPUTE_AUX_SIFS FALSE
#endif
#  define STRESS         firstPK
#  define DISPL_GRAD     displ_grad
#  define DISPL_GRAD_DOT displ_grad_dot
#  define VELOCITY       vela
#  define ACCELERATION   acca




static void
hex_ii_contrib (WEFG_hex_volume_t *h,
	      WEFG_superelem_t *se);
static void
hex_ii_contrib_2x2x2_ipol (WEFG_hex_volume_t *h, WEFG_superelem_t *se);

static void
determine_bounding_box (void);
static void
q_func (void);
static void
where (void);
static void
r_func (void);
static void
theta_func (int mode);
static void
aux_fields (void);

#define MODE_I   0
#define MODE_II  1
#define MODE_III 2

/*
   Data container shared by all the functions in this file.
 */
typedef struct interaction_integral_opts_t  {
  WEFG_superelem_t               *se; /* EFG superelement */
  W_loc_t                         at; /* location at which the interaction
                                       integral is being evaluated */
  W_vect_t                        forward; /* `forward' unit vector */
  W_vect_t                        tangent; /* unit vector tangent to
                                            the crack front */
  W_vect_t                        normal; /* unit vector normal to the
                                           crack surface */
  double                          hsize_Xbar_plus;	/* forward size of
                                                         virtual extension
                                                         domain */
  double                          hsize_Xbar_minus;	/* forward size of
                                                         virtual extension
                                                         domain */
  double                          hsize_Ybar_plus;	/* tangent size of
                                                         virtual extension
                                                         domain */
  double                          hsize_Ybar_minus;	/* tangent size of
                                                         virtual extension
                                                         domain */
  double                          hsize_Zbar_plus;	/* normal size of
                                                         virtual extension
                                                         domain */
  double                          hsize_Zbar_minus;	/* normal size of
                                                         virtual extension
                                                         domain  */
  G3D_box_t                       bounding_box; /* bounding box
                                                 of the virtual extension
                                                 domain in the global coordinate system */
  double                          transf_mat[3][3]; /* transformation
                                                     matrix between
                                                     the local coordinate
                                                     system (forward, tangent, normal vector) and
                                                     the global axes */
  W_loc_t                         currpt; /* current point */
  W_loc_t                         currpt_bar; /* current point in the local
                                               coordinate system */
  double                          r; /* distance of current point
                                        from the front vertex in the
                                        forward/normal plane
                                      */
  double                          theta; /* angle to the current point
                                            in the
                                        forward/normal plane
                                          */
  double                          sin_theta; /* cached sin (theta) */
  double                          cos_theta; /* cached cos (theta) */
  double                          sin_theta_half; /* cached sin (theta/2) */
  double                          cos_theta_half; /* cached cos (theta/2) */
  double                          q; /* virtual extension function */
  double                          q_der[3]; /* derivatives of the
                                             virtual extension function
                                             */
  double                          r_func_val; /* function in the
                                                 2-D expressions for
                                                 the SIF's which depends
                                                 on r only; denoted R */
  double                          r_func_sub_r_val; /* dR/dr */
  double                          theta_func_val[3]; /* function
                                                        in the
                                                 2-D expressions for
                                                 the SIF's which depends
                                                 on theta only; denoted T
                                                      */
  double                          theta_func_sub_theta_val[3]; /* dT/dtheta */
  double                          aux_displ[3]; /* auxiliary displacement */
  double                          aux_displ_grad[3][3]; /* gradient of the
                                                         auxiliary displacement */
  double                          aux_displ_grad_dot[3][3]; /* time derivative
                                                             of the gradient of the
                                                         auxiliary displacement */
  double                          aux_strain[3][3]; /* auxiliary strain tensor */
  double                          aux_stress[3][3]; /* auxiliary stress tensor */
  double                          aux_vel[3]; /* auxiliary velocity */
  double                          aux_acc[3]; /* auxiliary acceleration  */
  int                             mode; /* mode (0,1,2 stands for K_I,
                                         K_II, K_III) */
  double                          rho; /* mass density */
  double                          E; /* Young's modulus */
  double                          nu; /* Poisson ratio */
  double                          mu; /* shear modulus */
  WEFG_tensor_prod_comp_func      f; /* virtual extension function
                                      is a tensor-product function
                                      in the local coordinate system
                                      at the vertex with component f */
  WEFG_tensor_prod_comp_der_func  f_r; /* derivative of the component */
  double                          ii_KI; /* to cache results */
  double                          ii_KII; /* to cache results */
  double                          ii_KIII; /* to cache results */
#if COMPUTE_AUX_SIFS
  double                          ii_aux_KI; /* to cache results */
  double                          ii_aux_KII; /* to cache results */
  double                          ii_aux_KIII; /* to cache results */
#endif
  int                             subsplit_for_2x2x2_ipol; /* should
                                                            integration cells
                                                            be subdivided
                                                            for the integration
                                                            of the interaction
                                                            integrals? */
  W_bool_t                        include_dynamic_effects;
}               interaction_integral_opts_t;




static interaction_integral_opts_t iiopts;

/* 
   EValuate stress intensity factors from interaction integrals.
   EValuate stress intensity factors from interaction integrals.
 */
void
efgii_sifs_volume (WEFG_superelem_t *se, /* EFG superelement */
                   W_loc_t *at_point, /* location at the front */
                   W_vect_t *forward, /* unit vector forward  */
                   W_vect_t *tangent, /* unit vector tangent to the front */
                   W_vect_t *normal, /* unit vector normal to the crack surface */
                   double hsize_forward_plus, /* forward size of
                                                         virtual extension
                                                         domain */
                   double hsize_forward_minus,/* forward size of
                                                         virtual extension
                                                         domain */
                   double hsize_tangent_plus,/* tangent size of
                                                         virtual extension
                                                         domain */
                   double hsize_tangent_minus,/* tangent size of
                                                         virtual extension
                                                         domain */
                   double hsize_normal_plus,/* normal size of
                                                         virtual extension
                                                         domain */
                   double hsize_normal_minus,/* normal size of
                                                         virtual extension
                                                         domain */
                   WEFG_tensor_prod_comp_func func, /* virtual extension function
                                      is a tensor-product function
                                      in the local coordinate system
                                      at the vertex with component f */
                   WEFG_tensor_prod_comp_der_func func_der, /* virtual extension function
                                      is a tensor-product function
                                      in the local coordinate system
                                      at the vertex with component f; this is the derivative
                                      of the component function wrt the argument */
                   double E_star, /* effective Young's modulus */
                   int subsplit, /* should
                                                            integration cells
                                                            be subdivided
                                                            for integration
                                                            of the interaction
                                                            integrals?  */
                   W_bool_t include_dynamic_effects, /* include dynamics effects?  If FALSE,
                                                      static K_i are output */
                   double *KI_dynamic, /* output: opening SIF */
                   double *KII_dynamic, /* output: shearing SIF */
                   double *KIII_dynamic, /* output: tearing SIF */
                   double *KI_equiv_dynamic, /* output: equivalent opening SIF  */
                   double *theta_dynamic /* output: angle in the forward/normal plane */
                   ) 
{
  WEFG_hex_volume_t *h;
  W_loc_t origin =  {0,0,0} ;

  /* the options */
  iiopts.se                      = se;
  iiopts.at                      = *at_point;
  /* NOTE:  forward = xbar, normal = ybar, tangent = -zbar */
  iiopts.forward                 = *forward;
  iiopts.tangent                 = *tangent;
  iiopts.normal                  = *normal;
  iiopts.hsize_Xbar_plus         = hsize_forward_plus;
  iiopts.hsize_Ybar_plus         = hsize_normal_plus;
  iiopts.hsize_Zbar_plus         = hsize_tangent_minus;
  iiopts.hsize_Xbar_minus        = hsize_forward_minus;
  iiopts.hsize_Ybar_minus        = hsize_normal_minus;
  iiopts.hsize_Zbar_minus        = hsize_tangent_plus;
  iiopts.transf_mat[0][0]        = iiopts.forward.x;
  iiopts.transf_mat[1][0]        = iiopts.forward.y;
  iiopts.transf_mat[2][0]        = iiopts.forward.z;
  iiopts.transf_mat[0][1]        = iiopts.normal.x;
  iiopts.transf_mat[1][1]        = iiopts.normal.y;
  iiopts.transf_mat[2][1]        = iiopts.normal.z;
  iiopts.transf_mat[0][2]        = -iiopts.tangent.x;
  iiopts.transf_mat[1][2]        = -iiopts.tangent.y;
  iiopts.transf_mat[2][2]        = -iiopts.tangent.z;
  iiopts.currpt_bar              = origin;
  iiopts.rho                     = WM3D_get_volume_mass (iiopts.se->mat);
  if (!WM3D_get_mat_const (iiopts.se->mat, MAT_3D_YOUNG_MODULUS, &iiopts.E)) 
    REPORT_FATAL_ERROR (iiopts.se->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const (iiopts.se->mat, MAT_3D_POISSON, &iiopts.nu)) 
    REPORT_FATAL_ERROR (iiopts.se->domain->errh, "No Poisson constant");
  iiopts.mu                      = iiopts.E / 2 / (1 + iiopts.nu);
  iiopts.f                       = func;
  iiopts.f_r                     = func_der;
  iiopts.subsplit_for_2x2x2_ipol = subsplit;
  iiopts.include_dynamic_effects = include_dynamic_effects;
  
  determine_bounding_box ();
  
  /* evaluate the domain volume integral */
  iiopts.ii_KI   = 0;
  iiopts.ii_KII  = 0;
  iiopts.ii_KIII = 0;
#if COMPUTE_AUX_SIFS
  iiopts.ii_aux_KI   = 0;
  iiopts.ii_aux_KII  = 0;
  iiopts.ii_aux_KIII = 0;
#endif  
  if (WOPT_OPT (se->domain->options, j_dom_eval) == W_PER_CELL_IPOL_PTS) {
    loop_through_list (se->hex_volumes, h, WEFG_hex_volume_t *) {
      if (h->is_intersected_by_crack_surf) 
        hex_ii_contrib (h, se);
      else
        hex_ii_contrib_2x2x2_ipol (h, se);
    } 
  }  else  { /* default */
    loop_through_list (se->hex_volumes, h, WEFG_hex_volume_t *) {
      hex_ii_contrib (h, se);
    } 
  } 

#if COMPUTE_AUX_SIFS
  iiopts.ii_aux_KI   = iiopts.ii_aux_KI * iiopts.E / (1 - iiopts.nu) / 2;
  iiopts.ii_aux_KII  = iiopts.ii_aux_KII * iiopts.E / (1 - iiopts.nu) / 2;
  iiopts.ii_aux_KIII = iiopts.ii_aux_KIII * iiopts.mu;
#endif

  *KI_dynamic   = iiopts.ii_KI   * E_star / 2;
  *KII_dynamic  = iiopts.ii_KII  * E_star / 2;
  *KIII_dynamic = iiopts.ii_KIII * iiopts.mu;

  W_compute_theta (*KI_dynamic, *KII_dynamic, theta_dynamic);
  *KI_equiv_dynamic = W_equiv_KI (*KI_dynamic, *KII_dynamic, *theta_dynamic);
} 

#define ALL_REF_COORDS                                                  \
  h->vertex[0]->loc.x, h->vertex[0]->loc.y, h->vertex[0]->loc.z,        \
  h->vertex[1]->loc.x, h->vertex[1]->loc.y, h->vertex[1]->loc.z,        \
  h->vertex[2]->loc.x, h->vertex[2]->loc.y, h->vertex[2]->loc.z,        \
  h->vertex[3]->loc.x, h->vertex[3]->loc.y, h->vertex[3]->loc.z,        \
  h->vertex[4]->loc.x, h->vertex[4]->loc.y, h->vertex[4]->loc.z,        \
  h->vertex[5]->loc.x, h->vertex[5]->loc.y, h->vertex[5]->loc.z,        \
  h->vertex[6]->loc.x, h->vertex[6]->loc.y, h->vertex[6]->loc.z,        \
  h->vertex[7]->loc.x, h->vertex[7]->loc.y, h->vertex[7]->loc.z

#define ALL_GAUSS_LOCS ip[LOOP_I].Loc, ip[LOOP_J].Loc, ip[LOOP_K].Loc

#define GP_WGT(gp) (gp[LOOP_I].Wgt * gp[LOOP_J].Wgt * gp[LOOP_K].Wgt) 

#define N_LOC(h, v, DIR, n) \
   v.DIR = (  n[0] * h->vertex[0]->loc.DIR      \
            + n[1] * h->vertex[1]->loc.DIR      \
            + n[2] * h->vertex[2]->loc.DIR      \
            + n[3] * h->vertex[3]->loc.DIR      \
            + n[4] * h->vertex[4]->loc.DIR      \
            + n[5] * h->vertex[5]->loc.DIR      \
            + n[6] * h->vertex[6]->loc.DIR      \
            + n[7] * h->vertex[7]->loc.DIR) 



/*
   Compute bounding box of the integration volume.
   The integration volume is centered at the vertex on the front.
   It is oriented along the local coordinate axes given by the forward,
   tangent and normal vector.
 */
static void
determine_bounding_box (void) 
{
  W_loc_t pt;
  W_loc_t Xbar, Ybar, Zbar;

  G3D_INIT_BOX (iiopts.bounding_box, iiopts.at);

#define INITV(SIGNF, SIGNT, SIGNN, WHEREF, WHERET, WHEREN) \
   {                                                                      \
    VECMAC_3_SCALAR (Xbar, =,                                             \
 (SIGNF iiopts.hsize_Xbar_##WHEREF * iiopts.transf_mat[0][0]),  \
 (SIGNF iiopts.hsize_Xbar_##WHEREF * iiopts.transf_mat[1][0]),  \
 (SIGNF iiopts.hsize_Xbar_##WHEREF * iiopts.transf_mat[2][0])); \
    VECMAC_3_SCALAR (Ybar, =,                                             \
 (SIGNT iiopts.hsize_Ybar_##WHERET * iiopts.transf_mat[0][1]),  \
 (SIGNT iiopts.hsize_Ybar_##WHERET * iiopts.transf_mat[1][1]),  \
 (SIGNT iiopts.hsize_Ybar_##WHERET * iiopts.transf_mat[2][1])); \
    VECMAC_3_SCALAR (Zbar, =,                                             \
 (SIGNN iiopts.hsize_Zbar_##WHEREN * iiopts.transf_mat[0][2]),  \
 (SIGNN iiopts.hsize_Zbar_##WHEREN * iiopts.transf_mat[1][2]),  \
 (SIGNN iiopts.hsize_Zbar_##WHEREN * iiopts.transf_mat[2][2])); \
                                                                         \
    VECMAC_3_SCALAR (pt, =,                                               \
		  iiopts.at.x + Xbar.x + Ybar.x + Zbar.x,                \
		  iiopts.at.y + Xbar.y + Ybar.y + Zbar.y,                \
		  iiopts.at.z + Xbar.z + Ybar.z + Zbar.z);               \
  } 
  
  INITV (-, -, -, minus, minus, minus);
  G3D_UPDT_BOX (iiopts.bounding_box, pt);
  INITV (+, -, -, plus, minus, minus);
  G3D_UPDT_BOX (iiopts.bounding_box, pt);
  INITV (+, +, -, plus, plus, minus);
  G3D_UPDT_BOX (iiopts.bounding_box, pt);
  INITV (-, +, -, minus, plus, minus);
  G3D_UPDT_BOX (iiopts.bounding_box, pt);
 
  INITV (-, -, +, minus, minus, plus);
  G3D_UPDT_BOX (iiopts.bounding_box, pt);
  INITV (+, -, +, plus, minus, plus);
  G3D_UPDT_BOX (iiopts.bounding_box, pt);
  INITV (+, +, +, plus, plus, plus);
  G3D_UPDT_BOX (iiopts.bounding_box, pt);
  INITV (-, +, +, minus, plus, plus);
  G3D_UPDT_BOX (iiopts.bounding_box, pt);
} 

#define CBND(hbbox, hex, which, func, N) \
   {                                                              \
    hbbox.which.x = func (hbbox.which.x, hex->vertex[N]->loc.x);  \
    hbbox.which.y = func (hbbox.which.y, hex->vertex[N]->loc.y);  \
    hbbox.which.z = func (hbbox.which.z, hex->vertex[N]->loc.z);  \
  } 
#define DEFINE_HBBOX(hex, hbbox) \
   {                                                                  \
    hbbox.lower.x = hbbox.upper.x = hex->vertex[0]->loc.x;           \
    hbbox.lower.y = hbbox.upper.y = hex->vertex[0]->loc.y;           \
    hbbox.lower.z = hbbox.upper.z = hex->vertex[0]->loc.z;           \
    CBND (hbbox, hex, lower, min, 1);                                 \
    CBND (hbbox, hex, lower, min, 2);                                 \
    CBND (hbbox, hex, lower, min, 3);                                 \
    CBND (hbbox, hex, lower, min, 4);                                 \
    CBND (hbbox, hex, lower, min, 5);                                 \
    CBND (hbbox, hex, lower, min, 6);                                 \
    CBND (hbbox, hex, lower, min, 7);                                 \
    CBND (hbbox, hex, upper, max, 1);                                 \
    CBND (hbbox, hex, upper, max, 2);                                 \
    CBND (hbbox, hex, upper, max, 3);                                 \
    CBND (hbbox, hex, upper, max, 4);                                 \
    CBND (hbbox, hex, upper, max, 5);                                 \
    CBND (hbbox, hex, upper, max, 6);                                 \
    CBND (hbbox, hex, upper, max, 7);                                 \
  } 


#define BOXES_OVERLAP(r1, r2) 				     	\
 ((r1.lower.x <= r2.upper.x) && (r1.upper.x >= r2.lower.x) 	\
    && (r1.lower.y <= r2.upper.y) && (r1.upper.y >= r2.lower.y) 	\
    && (r1.lower.z <= r2.upper.z) && (r1.upper.z >= r2.lower.z)) 


#define TRANSF_V_3(VP_, A_, V_) \
  {                                                                      \
   VP_.x  = (A_[0][0]*V_.x + A_[1][0]*V_.y + A_[2][0]*V_.z);            \
   VP_.y  = (A_[0][1]*V_.x + A_[1][1]*V_.y + A_[2][1]*V_.z);            \
   VP_.z  = (A_[0][2]*V_.x + A_[1][2]*V_.y + A_[2][2]*V_.z);            \
} 

static void
eval_ii_contrib (WEFG_integration_point_t *cip, /* integration point */
                W_mat_3d_ip_ctx_t *ip_ctx, /* integration context */
                double ii_contrib[3] /* input/output:
                                      interaction integral values */
                 
#if COMPUTE_AUX_SIFS
                , double ii_aux_contrib[3]
#endif
                ) ;



/*
   Evaluate contribution to the interaction integral.
   Evaluate contribution to the interaction integral from the
   integration volume.
 */
static void
hex_ii_contrib (WEFG_hex_volume_t *h, /* integration volume */
                WEFG_superelem_t *se /* EFG superelement */
               ) 
{
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  double jdetw, J[3][3];
  double Nmat[8];
  double ii_contrib[3];
#if COMPUTE_AUX_SIFS
  double ii_aux_contrib[3];
#endif
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int qo;
  G3D_box_t hbox;
  WEFG_hex_volume_t *ho;
  FLAG_2PT_RULE (se, h);
#define CURRPT() 				        \
   {					        	\
    N_LOC (h, iiopts.currpt, x, Nmat);			\
    N_LOC (h, iiopts.currpt, y, Nmat);			\
    N_LOC (h, iiopts.currpt, z, Nmat);			\
  } 
#undef ALL_GAUSS_LOCS 
#define ALL_GAUSS_LOCS gp[LOOP_I].Loc, gp[LOOP_J].Loc, gp[LOOP_K].Loc
#undef GP_WGT
#define GP_WGT(gp) (gp[LOOP_I].Wgt * gp[LOOP_J].Wgt * gp[LOOP_K].Wgt) 
 
   DEFINE_HBBOX (h, hbox);
   if (!BOXES_OVERLAP (hbox, iiopts.bounding_box)) 
     return;
   
   if (h->offspring) {
     loop_through_list (h->offspring, ho, WEFG_hex_volume_t *) {
       hex_ii_contrib (ho, se);
     } 
   }  else  { /* No offspring */
     
     ip_ctx_data.se       = se;
     ip_ctx.time          = WD_time (se->domain);
     ip_ctx.time_step     = WD_time_step (se->domain);
     ip_ctx.ctx_data      = &ip_ctx_data;
     ip_ctx.get_kinem_var = efgss_get_kinem_var;
     
     qo = h->qp->qr->order;
     ip = h->qp->efg_ip;
     gp = h->qp->qr->ip;
     IPA_LOOP (qo) {
       USE_QUADR_PT (se) {
         G3D_hex_N (Nmat, ALL_GAUSS_LOCS);  
         G3D_hex_J (J, ALL_GAUSS_LOCS, ALL_REF_COORDS);
         cip               = IPA_ENTRY (ip);
         ip_ctx_data.sftab = cip->sftab;
         EFGSS_AT_NEW_IP ();
         INCL_EXCL_2PT_CASE (  { jdetw = 4 * GP_WGT (gp) * G3D_hex_detJ (J); } ,
                               { jdetw = GP_WGT (gp) * G3D_hex_detJ (J); } 
                               );
         CURRPT ();
         if (G3D_PT_INSIDE_BOX (iiopts.bounding_box, iiopts.currpt)) {
           eval_ii_contrib (cip, &ip_ctx,
                            ii_contrib
#                         if COMPUTE_AUX_SIFS
                            , ii_aux_contrib
#                         endif
                            );
           iiopts.ii_KI   += ii_contrib[MODE_I] * jdetw;
           iiopts.ii_KII  += ii_contrib[MODE_II] * jdetw;
           iiopts.ii_KIII += ii_contrib[MODE_III] * jdetw;
#         if COMPUTE_AUX_SIFS
           iiopts.ii_aux_KI    += ii_aux_contrib[MODE_I] * jdetw;
           iiopts.ii_aux_KII   += ii_aux_contrib[MODE_II] * jdetw;
           iiopts.ii_aux_KIII  += ii_aux_contrib[MODE_III] * jdetw;
#         endif
         }  /* if (G3D_PT_INSIDE_BOX */
       } 
     }  IPA_LOOP_END ();
     
   }  /* No offspring */
} 



/*
   Evaluate coordinates of the current integration point.
   Compute the coordinates in the local coordinate system.
   Compute r and theta, and also the cached data.
 */
static void
where (void) 
{
  W_loc_t d;
  VECMAC_3_VEC (d, =, iiopts.currpt, -, iiopts.at);
  iiopts.currpt_bar.x = VECMAC_DOT (iiopts.forward, d);
  iiopts.currpt_bar.y = VECMAC_DOT (iiopts.normal, d);
  iiopts.currpt_bar.z = -VECMAC_DOT (iiopts.tangent,  d);
  
  iiopts.r = sqrt (iiopts.currpt_bar.x * iiopts.currpt_bar.x +
                  iiopts.currpt_bar.y * iiopts.currpt_bar.y);
  iiopts.theta = atan2 (iiopts.currpt_bar.y, iiopts.currpt_bar.x);
  iiopts.sin_theta      = iiopts.currpt_bar.y / iiopts.r;
  iiopts.cos_theta      = iiopts.currpt_bar.x / iiopts.r;
  iiopts.sin_theta_half = sin (iiopts.theta / 2);
  iiopts.cos_theta_half = cos (iiopts.theta / 2);
} 


/*
   Evaluate interaction integral contribution.
   Evaluate interaction integral contribution at an integration point.
   Compute the location and all the cached data, and the virtual extension
   function value (plus derivatives) .  Finally evaluate the constituent
   products and add them up updating the /ii_contrib/ array.
 */ 
static void
eval_ii_contrib (WEFG_integration_point_t *cip, /* integration point */
                W_mat_3d_ip_ctx_t *ip_ctx, /* integration context */
                double ii_contrib[3] /* input/output:
                                      interaction integral values */
                 
#if COMPUTE_AUX_SIFS
                , double ii_aux_contrib[3]
#endif
                ) 
{
  double displ_grad[3][3];
  double displ_grad_dot[3][3];
  double firstPK[3][3];
  double tmp[3][3];
  W_vect_t vec, vel, acc;
  double *acca, *vela, temp;
  int i, j;
  double W, K;
  W_mat_3d_stress_type_enum out_type;
  int mode;

  where ();
  q_func ();

  ii_contrib[0] = ii_contrib[1] = ii_contrib[2] = 0;
#if COMPUTE_AUX_SIFS
  ii_aux_contrib[0] = ii_aux_contrib[1] = ii_aux_contrib[2] = 0;
#endif
  
  if (iiopts.q == 0) 
    return; /* evaluation point is outside of q support */

  /* Get displacement gradient at point */
  if (! (ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, displ_grad)) {
    REPORT_ERROR (NULL, "No displacement gradient");
    return;
  } 
  displ_grad[0][0] -= 1;
  displ_grad[1][1] -= 1;
  displ_grad[2][2] -= 1;
  
  if (iiopts.include_dynamic_effects) {
    /* Get F_dot (material time derivative of the
       displacement gradient at point */
    if (! (ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD_DOT, displ_grad_dot)) {
      REPORT_ERROR (NULL, "No displacement gradient");
      return;
    } 
  } else {
    MAT3DUTIL_ZERO_3X3 (displ_grad_dot);
  }
  
  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress (iiopts.se->mat, cip->matstate, W_MAT_STATE_UPDATE_TEMP,
		       ip_ctx, MAT_3D_1ST_PK, firstPK,
		       &out_type)) {
    REPORT_ERROR (iiopts.se->domain->errh, "Failed to compute stress");
    return;
  } 

  /* Deformation energy (strain energy) at point */
  if (!WM3D_get_mat_state_var (iiopts.se->mat, cip->matstate,
			      ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY, &W)) {
    REPORT_ERROR (iiopts.se->domain->errh, "Failed to retrieve deformation energy");
    return;
  } 

  /* Kinetic energy */
  if (iiopts.include_dynamic_effects) {
    efgsf_eval_curr_vel (&cip->sftab, &vec.x, &vec.y, &vec.z);
    TRANSF_V_3 (vel, iiopts.transf_mat, vec);
    efgsf_eval_curr_acc (&cip->sftab, &vec.x, &vec.y, &vec.z);
    TRANSF_V_3 (acc, iiopts.transf_mat, vec);
    K = 0.5 * iiopts.rho * VECMAC_DOT (vel, vel);
  } else {
    K = 0;
    VECMAC_SCALAR (vel, =, 0);
    VECMAC_SCALAR (acc, =, 0);
  }
  acca = &acc.x;
  vela = &vel.x;

  /* Transform the data to the barred coordinate system */
  MAT3DUTIL_MULT_ATB_3X3 (tmp, /* = */ iiopts.transf_mat, /* * */ displ_grad);
  MAT3DUTIL_MULT_AB_3X3 (displ_grad, /* = */ tmp, /* * */ iiopts.transf_mat);

  MAT3DUTIL_MULT_ATB_3X3 (tmp, /* = */ iiopts.transf_mat, /* * */ firstPK);
  MAT3DUTIL_MULT_AB_3X3 (firstPK, /* = */ tmp, /* * */ iiopts.transf_mat);

  MAT3DUTIL_MULT_ATB_3X3 (tmp, /* = */iiopts.transf_mat, /* * */displ_grad_dot);
  MAT3DUTIL_MULT_AB_3X3 (displ_grad_dot, /* = */tmp, /* * */iiopts.transf_mat);

  FOR (mode, 3) {
    r_func ();
    theta_func (mode);
    aux_fields ();
    /* (i) */ 
    temp = MAT3DUTIL_DBL_PROD_3X3 (STRESS, iiopts.aux_strain) * iiopts.q_der[0];
    ii_contrib[mode] -= temp;
#if COMPUTE_AUX_SIFS
    ii_aux_contrib_static[mode] -=
      MAT3DUTIL_DBL_PROD_3X3 (iiopts.aux_stress, iiopts.aux_strain) 
        * iiopts.q_der[0];
#endif
    FOR (i, 3) {
      /* (ii) */
      temp = iiopts.rho * (iiopts.aux_vel[i] * VELOCITY[i]) * iiopts.q_der[0];
      ii_contrib[mode] += temp;
#if COMPUTE_AUX_SIFS
      ii_aux_contrib[mode] +=
        iiopts.rho * (iiopts.aux_vel[i] * iiopts.aux_vel[i]) * iiopts.q_der[0];
#endif
    } 
    FOR (i, 3) {
      FOR (j, 3) {
        /* (iii) and (iv) */
        temp = (iiopts.aux_stress[i][j] * DISPL_GRAD[i][0] +
                STRESS[i][j] * iiopts.aux_displ_grad[i][0]
                ) * iiopts.q_der[j];
        ii_contrib[mode] += temp;
#if COMPUTE_AUX_SIFS
        ii_aux_contrib[mode] +=
          (iiopts.aux_stress[i][j] * iiopts.aux_displ_grad[i][0] +
           iiopts.aux_stress[i][j] * iiopts.aux_displ_grad[i][0]
           ) * iiopts.q_der[j];
#endif
      } 
    } 
    /* (v), (vi), (vii) and (viii) */
    FOR (i, 3) {
      ii_contrib[mode] +=
        iiopts.rho * (  iiopts.aux_acc[i] * DISPL_GRAD[i][0]
                        + ACCELERATION[i] * iiopts.aux_displ_grad[i][0]
                        - iiopts.aux_vel[i] * DISPL_GRAD_DOT[i][0]
                        - VELOCITY[i] * iiopts.aux_displ_grad_dot[i][0]
                        ) * iiopts.q;
#if COMPUTE_AUX_SIFS
      ii_aux_contrib[mode] +=
        iiopts.rho * (  iiopts.aux_acc[i] * iiopts.aux_displ_grad[i][0]
                      + iiopts.aux_acc[i] * iiopts.aux_displ_grad[i][0]
                      - iiopts.aux_vel[i] * iiopts.aux_displ_grad_dot[i][0]
                      - iiopts.aux_vel[i] * iiopts.aux_displ_grad_dot[i][0]
                      ) * iiopts.q;
#endif
    } 
  } 
  
  return;
} 








/*
   Evaluate the virtual extension function.
   Evaluate the virtual extension function at the current point.
 */
static void
q_func (void) 
{
  double xs = (iiopts.currpt_bar.x > 0
               ? iiopts.hsize_Xbar_plus
               : iiopts.hsize_Xbar_minus);
  double ys = (iiopts.currpt_bar.y > 0
               ? iiopts.hsize_Ybar_plus
               : iiopts.hsize_Ybar_minus);
  double zs = (iiopts.currpt_bar.z > 0
               ? iiopts.hsize_Zbar_plus
               : iiopts.hsize_Zbar_minus);
  double xbar = iiopts.currpt_bar.x / xs;
  double ybar = iiopts.currpt_bar.y / ys;
  double zbar = iiopts.currpt_bar.z / zs;
  
  /* Virtual displacement along the forward vector */
  iiopts.q = (iiopts.f (xbar) * iiopts.f (ybar) * iiopts.f (zbar));

  /* Derivatives */
  iiopts.q_der[0] = (iiopts.f_r (xbar) / xs * iiopts.f (ybar) * iiopts.f (zbar));
  iiopts.q_der[1] = (iiopts.f (xbar) * iiopts.f_r (ybar) / ys * iiopts.f (zbar));
  iiopts.q_der[2] = (iiopts.f (xbar) * iiopts.f (ybar) * iiopts.f_r (zbar) / zs);
} 


/* function in the
 2-D expressions for
 the SIF's which depends
 on r only; denoted R */
static void
r_func (void) 
{
#define ONE_OVER_SQRT_2_PI 0.39894228040143267794
  iiopts.r_func_val =
    0.5 / iiopts.mu * sqrt (iiopts.r) * ONE_OVER_SQRT_2_PI;
  iiopts.r_func_sub_r_val =
    0.25 / iiopts.mu * ONE_OVER_SQRT_2_PI / sqrt (iiopts.r);
} 

/*
   evaluate T (theta) function. the T (theta) function appears
   in the 2-D expressions for the SIF's, and depends on theta only.
   It is different for each mode.
   */
static void
theta_func (int mode) 
{
  double st2 = iiopts.sin_theta_half;
  double ct2 = iiopts.cos_theta_half;
  double k   = 3 - 4 * iiopts.nu;

  switch (mode) {
  case MODE_I:
    iiopts.theta_func_val[0] =  ct2 * (k - 1 + 2 * st2 * st2);
    iiopts.theta_func_val[1] =  st2 * (k + 1 - 2 * ct2 * ct2);
    iiopts.theta_func_val[2] =  0;
    iiopts.theta_func_sub_theta_val[0] =
      -0.5 * st2 * (k - 1 + 2 * st2 * st2 - 4 * ct2 * ct2);
    iiopts.theta_func_sub_theta_val[1] =
       0.5 * ct2 * (k + 1 - 2 * ct2 * ct2 + 4 * st2 * st2);
    iiopts.theta_func_sub_theta_val[2] = 0;
    break;
  case MODE_II:
    iiopts.theta_func_val[0] =  st2 * (k + 1 + 2 * ct2 * ct2);
    iiopts.theta_func_val[1] = -ct2 * (k - 1 - 2 * st2 * st2);
    iiopts.theta_func_val[2] =  0;
    iiopts.theta_func_sub_theta_val[0] =
      0.5 * ct2 * (k + 1 + 2 * ct2 * ct2 - 4 * st2 * st2);
    iiopts.theta_func_sub_theta_val[1] = 
      0.5 * st2 * (k - 1 - 2 * st2 * st2 + 4 * ct2 * ct2);
    iiopts.theta_func_sub_theta_val[2] = 0;
    break;
  case MODE_III:
    iiopts.theta_func_val[0] = 0;
    iiopts.theta_func_val[1] = 0;
    iiopts.theta_func_val[2] = 4 * st2;
    iiopts.theta_func_sub_theta_val[0] = 0;
    iiopts.theta_func_sub_theta_val[1] = 0;
    iiopts.theta_func_sub_theta_val[2] = 2 * ct2;
    break;  
  } 
} 
 
/*
   EValuate the auxiliary fields.
   EValuate the auxiliary fields.  The mode is taken into account
   through the T (theta) function.
 */
static void
aux_fields (void) 
{
  int i;
  double r   = iiopts.r;
  double st  = iiopts.sin_theta;
  double ct  = iiopts.cos_theta;
  double g   = iiopts.r_func_val;
  double g_r = iiopts.r_func_sub_r_val;
  double E   = iiopts.E;
  double nu  = iiopts.nu;
  double mu  = iiopts.mu;
  double lambda = E * nu / (1 + nu) / (1 - 2*nu);
  double treps;
  
  FOR (i, 3) {
    iiopts.aux_displ[i] = iiopts.r_func_val * iiopts.theta_func_val[i];
    iiopts.aux_vel[i]   = 0;
    iiopts.aux_acc[i]   = 0;
  } 

  MAT3DUTIL_ZERO_3X3 (iiopts.aux_displ_grad);
  MAT3DUTIL_ZERO_3X3 (iiopts.aux_displ_grad_dot);
  FOR (i, 3) {
    iiopts.aux_displ_grad[i][0] +=
      ct * g_r * iiopts.theta_func_val[i] -
        st / r * g * iiopts.theta_func_sub_theta_val[i];
    iiopts.aux_displ_grad[i][1] +=
      st * g_r * iiopts.theta_func_val[i] +
        ct / r * g * iiopts.theta_func_sub_theta_val[i];
    iiopts.aux_displ_grad[i][2] = 0;
  } 
  MAT3DUTIL_COMB_ATB_3X3 (iiopts.aux_strain,
                         0.5, iiopts.aux_displ_grad,
                         0.5, iiopts.aux_displ_grad);
  treps = MAT3DUTIL_TRACE_3X3 (iiopts.aux_strain);
  iiopts.aux_stress[0][0] = lambda * treps + 2 * mu * iiopts.aux_strain[0][0];
  iiopts.aux_stress[0][1] =                  2 * mu * iiopts.aux_strain[0][1];
  iiopts.aux_stress[0][2] =                  2 * mu * iiopts.aux_strain[0][2];
  iiopts.aux_stress[1][0] =                  2 * mu * iiopts.aux_strain[1][0];
  iiopts.aux_stress[1][1] = lambda * treps + 2 * mu * iiopts.aux_strain[1][1];
  iiopts.aux_stress[1][2] =                  2 * mu * iiopts.aux_strain[1][2];
  iiopts.aux_stress[2][0] =                  2 * mu * iiopts.aux_strain[2][0];
  iiopts.aux_stress[2][1] =                  2 * mu * iiopts.aux_strain[2][1];
  iiopts.aux_stress[2][2] = lambda * treps + 2 * mu * iiopts.aux_strain[2][2];
} 


/* compute the mode-I auxiliary stress field. */
static void
mode_I_stress (void) 
{
  double r   = iiopts.r;
  double st2 = iiopts.sin_theta_half;
  double ct2 = iiopts.cos_theta_half;
  double st32= sin (3 * iiopts.theta / 2);
  double ct32= cos (3 * iiopts.theta / 2);
  double nu  = iiopts.nu;
  double KI_aux = 1;

  MAT3DUTIL_ZERO_3X3 (iiopts.aux_stress);
  iiopts.aux_stress[0][0]
    = KI_aux / sqrt (2 * M_PI * r) * ct2 * (1 - st2 * st32);
  iiopts.aux_stress[1][1]
    = KI_aux / sqrt (2 * M_PI * r) * ct2 * (1 + st2 * st32);
  iiopts.aux_stress[0][1] = iiopts.aux_stress[1][0]
    = KI_aux / sqrt (2 * M_PI * r) * st2 * ct2 * ct32;
  iiopts.aux_stress[0][2] = iiopts.aux_stress[2][0]
    = 0;
  iiopts.aux_stress[1][2] = iiopts.aux_stress[2][1]
    = 0;
  iiopts.aux_stress[2][2]
    = nu * (iiopts.aux_stress[0][0] + iiopts.aux_stress[1][1]);
} 







static void
fill_ip_2x2x2_pkg (double xi, double eta, double theta,
		  WEFG_integration_point_t *cip, W_mat_3d_ip_ctx_t *ip_ctx);
 

typedef struct ipol_2x2x2_pkg_t  {
  double          displ_grad[3][3];
  double          displ_grad_dot[3][3];
  W_vect_t        vel, acc;
  double          firstPK[3][3];
  double          W;
  double          K;
}               ipol_2x2x2_pkg_t;


static ipol_2x2x2_pkg_t ipol_2x2x2_pkgs[8];

/*
   Fill package for a 2x2x2 re-interpolation.

   If the crack does not intersect an integration volume (cell), the
   re-interpolation of from a 2x2x2 quadrature is carried out to enhance
   the accuracy of the interaction integrals.  The integration volumes
 (cells) are then subdivided, usually into 4x4x4 sub-cells and the
   fields re-interpolated from the 2x2x2 quadrature points are then
   integrated on the sub-cells.
 */
static void
fill_ip_2x2x2_pkg (double xi, /* parametric coordinate of the integration point */
                  double eta, /* parametric coordinate of the integration point */
                  double theta, /* parametric coordinate of the integration point */
		  WEFG_integration_point_t *cip, /* integration point */
                  W_mat_3d_ip_ctx_t *ip_ctx /* integration point context */
                  ) 
{
  double displ_grad[3][3];
  double displ_grad_dot[3][3];
  double firstPK[3][3];
  double tmp[3][3];
  W_vect_t vel, acc, vec;
  double W, K;
  W_mat_3d_stress_type_enum out_type;
  int entry = -1;
  
  if (xi < 0 && eta < 0 && theta < 0) {entry = 0;
  }  else if (xi > 0 && eta < 0 && theta < 0) {entry = 1;
  }  else if (xi > 0 && eta > 0 && theta < 0) {entry = 2;
  }  else if (xi < 0 && eta > 0 && theta < 0) {entry = 3;
  }  else if (xi < 0 && eta < 0 && theta > 0) {entry = 4;
  }  else if (xi > 0 && eta < 0 && theta > 0) {entry = 5;
  }  else if (xi > 0 && eta > 0 && theta > 0) {entry = 6;
  }  else if (xi < 0 && eta > 0 && theta > 0) {entry = 7;
  } 

  /* Get displacement gradient at point */
  if (! (ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, displ_grad)) {
    REPORT_ERROR (NULL, "No displacement gradient");
    return;
  } 
  displ_grad[0][0] -= 1;
  displ_grad[1][1] -= 1;
  displ_grad[2][2] -= 1;
  
  /* Get F_dot (material time derivative of the
     displacement gradient at point */
  if (! (ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD_DOT, displ_grad_dot)) {
    REPORT_ERROR (NULL,"No displacement gradient");
    return;
  } 
  
  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress (iiopts.se->mat, cip->matstate, W_MAT_STATE_UPDATE_TEMP,
		       ip_ctx, MAT_3D_1ST_PK, firstPK,
		       &out_type)) {
    REPORT_ERROR (iiopts.se->domain->errh, "Failed to compute stress");
    return;
  } 

  /* Deformation energy (strain energy) at point */
  if (!WM3D_get_mat_state_var (iiopts.se->mat, cip->matstate,
			      ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY,
			      &W)) {
    REPORT_ERROR (iiopts.se->domain->errh, "Failed to retrieve deformation energy");
    return;
  } 

  /* Kinetic energy */
  efgsf_eval_curr_vel (&cip->sftab, &vec.x, &vec.y, &vec.z);
  TRANSF_V_3 (vel, iiopts.transf_mat, vec);
  efgsf_eval_curr_acc (&cip->sftab, &vec.x, &vec.y, &vec.z);
  TRANSF_V_3 (acc, iiopts.transf_mat, vec);
  K = 0.5 * iiopts.rho * VECMAC_DOT (vel, vel);

  /* Transform the data to the barred coordinate system */
  MAT3DUTIL_MULT_ATB_3X3 (tmp, /* = */ iiopts.transf_mat, /* * */ displ_grad);
  MAT3DUTIL_MULT_AB_3X3 (displ_grad, /* = */ tmp, /* * */ iiopts.transf_mat);

  MAT3DUTIL_MULT_ATB_3X3 (tmp, /* = */ iiopts.transf_mat, /* * */ firstPK);
  MAT3DUTIL_MULT_AB_3X3 (firstPK, /* = */ tmp, /* * */ iiopts.transf_mat);

  MAT3DUTIL_MULT_ATB_3X3 (tmp, /* = */iiopts.transf_mat, /* * */displ_grad_dot);
  MAT3DUTIL_MULT_AB_3X3 (displ_grad_dot, /* = */tmp, /* * */iiopts.transf_mat);

  /* Store into the package */
  MAT3DUTIL_COPY_3X3 (ipol_2x2x2_pkgs[entry].displ_grad, displ_grad);
  MAT3DUTIL_COPY_3X3 (ipol_2x2x2_pkgs[entry].displ_grad_dot, displ_grad_dot);
  MAT3DUTIL_COPY_3X3 (ipol_2x2x2_pkgs[entry].firstPK, firstPK);
  ipol_2x2x2_pkgs[entry].vel = vel;
  ipol_2x2x2_pkgs[entry].acc = acc;
  ipol_2x2x2_pkgs[entry].W = W;
  ipol_2x2x2_pkgs[entry].K = K;
} 

/*
   Evaluate contribution to the interaction integrals
   from 2x2x2 re-interpolation on a sub-cell.
 */
static void
eval_ii_contrib_2x2x2_ipol (double xi, /* parametric coordinate */
                           double eta, /* parametric coordinate */
                           double theta, /* parametric coordinate */
                          double ii_contrib[3] /* input/output:
                                                  interaction integral  */
#if COMPUTE_AUX_SIFS
                           , double ii_aux_contrib[3]
#endif
                           );


static void
hex_ii_contrib_2x2x2_ipol (WEFG_hex_volume_t *h, WEFG_superelem_t *se) 
{
  WEFG_integration_point_t **ip, *cip;
  W_gauss_integration_point_t *gp;
  double jdetw, J[3][3];
  double Nmat[8];
  double ii_contrib[3];
#if COMPUTE_AUX_SIFS
  double ii_aux_contrib[3];
#endif
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int qo;
  G3D_box_t hbox;
  WEFG_hex_volume_t *ho;
  FLAG_2PT_RULE (se, h);
  W_gauss_integration_rule_t *gr;
  W_loc_t subhcorner[8];
#undef NLOC
#undef ALL_REF_COORDS
#define NLOC(v, DIR, n) \
   v.DIR = (  n[0] * subhcorner[0].DIR      \
            + n[1] * subhcorner[1].DIR      \
            + n[2] * subhcorner[2].DIR      \
            + n[3] * subhcorner[3].DIR      \
            + n[4] * subhcorner[4].DIR      \
            + n[5] * subhcorner[5].DIR      \
            + n[6] * subhcorner[6].DIR      \
            + n[7] * subhcorner[7].DIR) 
#define PT(pt, Nmat) \
   {                                            \
    NLOC (pt, x, Nmat);                         \
    NLOC (pt, y, Nmat);                         \
    NLOC (pt, z, Nmat);                         \
  } 
#define SUBHCORN(h, which, Nmat) \
   {                                            \
    N_LOC (h, subhcorner[which], x, Nmat);      \
    N_LOC (h, subhcorner[which], y, Nmat);      \
    N_LOC (h, subhcorner[which], z, Nmat);      \
  } 
#define SUBHCORNERS(h, xi, eta, theta, incr) \
   {                                                     \
    G3D_hex_N (Nmat, xi, eta, theta);                    \
    SUBHCORN (h, 0, Nmat);                               \
    G3D_hex_N (Nmat, xi+incr, eta, theta);               \
    SUBHCORN (h, 1, Nmat);                               \
    G3D_hex_N (Nmat, xi+incr, eta+incr, theta);          \
    SUBHCORN (h, 2, Nmat);                               \
    G3D_hex_N (Nmat, xi, eta+incr, theta);               \
    SUBHCORN (h, 3, Nmat);                               \
    G3D_hex_N (Nmat, xi, eta, theta+incr);               \
    SUBHCORN (h, 4, Nmat);                               \
    G3D_hex_N (Nmat, xi+incr, eta, theta+incr);          \
    SUBHCORN (h, 5, Nmat);                               \
    G3D_hex_N (Nmat, xi+incr, eta+incr, theta+incr);     \
    SUBHCORN (h, 6, Nmat);                               \
    G3D_hex_N (Nmat, xi, eta+incr, theta+incr);          \
    SUBHCORN (h, 7, Nmat);                               \
  } 
#define ALL_SUBHCORNER_COORDS                               \
  subhcorner[0].x, subhcorner[0].y, subhcorner[0].z,        \
  subhcorner[1].x, subhcorner[1].y, subhcorner[1].z,        \
  subhcorner[2].x, subhcorner[2].y, subhcorner[2].z,        \
  subhcorner[3].x, subhcorner[3].y, subhcorner[3].z,        \
  subhcorner[4].x, subhcorner[4].y, subhcorner[4].z,        \
  subhcorner[5].x, subhcorner[5].y, subhcorner[5].z,        \
  subhcorner[6].x, subhcorner[6].y, subhcorner[6].z,        \
  subhcorner[7].x, subhcorner[7].y, subhcorner[7].z
#undef CURRPT
#define CURRPT() PT (iiopts.currpt, Nmat) 
#undef ALL_GAUSS_LOCS 
#define ALL_GAUSS_LOCS gp[LOOP_I].Loc, gp[LOOP_J].Loc, gp[LOOP_K].Loc
#undef GP_WGT
#define GP_WGT(gp) (gp[LOOP_I].Wgt * gp[LOOP_J].Wgt * gp[LOOP_K].Wgt) 
 
  DEFINE_HBBOX (h, hbox);
  if (!BOXES_OVERLAP (hbox, iiopts.bounding_box)) 
    return;

  if (h->offspring) {
    loop_through_list (h->offspring, ho, WEFG_hex_volume_t *) {
      hex_ii_contrib_2x2x2_ipol (ho, se);
    } 
  }  else  { /* No offspring */
    
    ip_ctx_data.se       = se;
    ip_ctx.time          = WD_time (se->domain);
    ip_ctx.time_step     = WD_time_step (se->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = efgss_get_kinem_var;
    
    qo = h->qp->qr->order;
    if (qo != 2) {
      REPORT_FATAL_ERROR (iiopts.se->domain->errh, "Interpolated field not applicable");
    } 
    INC_2PT_RULE (if (use_2pt_rule) {
      REPORT_FATAL_ERROR (iiopts.se->domain->errh, "Interpolated field not applicable");
    } );
    ip = h->qp->efg_ip;
    gp = h->qp->qr->ip;
    IPA_LOOP (qo) {
      G3D_hex_N (Nmat, ALL_GAUSS_LOCS);  
      cip               = IPA_ENTRY (ip);
      ip_ctx_data.sftab = cip->sftab;
      EFGSS_AT_NEW_IP ();
      fill_ip_2x2x2_pkg (ALL_GAUSS_LOCS, cip, &ip_ctx);
    }  IPA_LOOP_END ();

    gr  = WG_get_integration_rule (2);
    gp  = gr->ip;
     { /* sub-split block */
      double xi, eta, theta;
      double incr = 2.0 / iiopts.subsplit_for_2x2x2_ipol;
      double upper = 1.001 - incr;
      
      for (xi = -1; xi < upper; xi += incr) {
        for (eta = -1; eta < upper; eta += incr) {
          for (theta = -1; theta < upper; theta += incr) {
            SUBHCORNERS (h, xi, eta, theta, incr);
            IPA_LOOP (2) {
              G3D_hex_N (Nmat, ALL_GAUSS_LOCS);
              CURRPT ();
              if (G3D_PT_INSIDE_BOX (iiopts.bounding_box, iiopts.currpt)) {
                G3D_hex_J (J, ALL_GAUSS_LOCS, ALL_SUBHCORNER_COORDS);
                jdetw = GP_WGT (gp) * G3D_hex_detJ (J);
                eval_ii_contrib_2x2x2_ipol (ALL_GAUSS_LOCS,
                                           ii_contrib
#                                          if COMPUTE_AUX_SIFS
                                           , ii_aux_contrib
#                                          endif 
                                           );
                iiopts.ii_KI   += ii_contrib[MODE_I] * jdetw;
                iiopts.ii_KII  += ii_contrib[MODE_II] * jdetw;
                iiopts.ii_KIII += ii_contrib[MODE_III] * jdetw;
#               if COMPUTE_AUX_SIFS
                iiopts.ii_aux_KI    += ii_aux_contrib[MODE_I] * jdetw;
                iiopts.ii_aux_KII   += ii_aux_contrib[MODE_II] * jdetw;
                iiopts.ii_aux_KIII  += ii_aux_contrib[MODE_III] * jdetw;
#               endif
              }  /* if (G3D_PT_INSIDE_BOX */
            }  IPA_LOOP_END ();
          } 
        } 
      } 
    }  /* sub-split block */

  }  /* No offspring */
} 

/*
   Evaluate interaction integral contribution.
   Evaluate interaction integral contribution at an integration point.
   Compute the location and all the cached data, and the virtual extension
   function value (plus derivatives) .  Finally evaluate the constituent
   products and add them up updating the /ii_contrib/ array.
 */ 
static void
eval_ii_contrib_2x2x2_ipol (double xi, /* parametric coordinate */
                           double eta, /* parametric coordinate */
                           double theta, /* parametric coordinate */
                          double ii_contrib[3] /* input/output:
                                                  interaction integral  */
#if COMPUTE_AUX_SIFS
                           , double ii_aux_contrib[3]
#endif
                           )
{
  double displ_grad[3][3];
  double displ_grad_dot[3][3];
  double firstPK[3][3];
#define SQRT_3 1.73205080756887729352
  double modxi    = SQRT_3 * xi;
  double modeta   = SQRT_3 * eta;
  double modtheta = SQRT_3 * theta;
  double Nmat[8];
  double W, K, temp;
  double *acca, *vela;
  W_vect_t vel, acc;
  int i, j;
  int mode;

  where ();
  q_func ();

  ii_contrib[0] = ii_contrib[1] = ii_contrib[2] = 0;
#if COMPUTE_AUX_SIFS
  ii_aux_contrib[0] = ii_aux_contrib[1] = ii_aux_contrib[2] = 0;
#endif
  
  if (iiopts.q == 0) 
    return; /* evaluation point is outside of q support */

  /* Interpolate */
  G3D_hex_N (Nmat, modxi, modeta, modtheta);
#define COMP_T1(T, I, J) 				\
   { T[I][J] = (						\
	       Nmat[0] * ipol_2x2x2_pkgs[0].T[I][J] +	\
	       Nmat[1] * ipol_2x2x2_pkgs[1].T[I][J] +	\
	       Nmat[2] * ipol_2x2x2_pkgs[2].T[I][J] +	\
	       Nmat[3] * ipol_2x2x2_pkgs[3].T[I][J] +	\
	       Nmat[4] * ipol_2x2x2_pkgs[4].T[I][J] +	\
	       Nmat[5] * ipol_2x2x2_pkgs[5].T[I][J] +	\
	       Nmat[6] * ipol_2x2x2_pkgs[6].T[I][J] +	\
	       Nmat[7] * ipol_2x2x2_pkgs[7].T[I][J]	\
	       );					\
  } 
#define COMP_T(T) 						\
   {								\
    COMP_T1 (T, 0, 0); COMP_T1 (T, 0, 1); COMP_T1 (T, 0, 2);	\
    COMP_T1 (T, 1, 0); COMP_T1 (T, 1, 1); COMP_T1 (T, 1, 2);	\
    COMP_T1 (T, 2, 0); COMP_T1 (T, 2, 1); COMP_T1 (T, 2, 2);	\
  } 
#define COMP_S(S) 				\
   { S = (					\
	 Nmat[0] * ipol_2x2x2_pkgs[0].S +	\
	 Nmat[1] * ipol_2x2x2_pkgs[1].S +	\
	 Nmat[2] * ipol_2x2x2_pkgs[2].S +	\
	 Nmat[3] * ipol_2x2x2_pkgs[3].S +	\
	 Nmat[4] * ipol_2x2x2_pkgs[4].S +	\
	 Nmat[5] * ipol_2x2x2_pkgs[5].S +	\
	 Nmat[6] * ipol_2x2x2_pkgs[6].S +	\
	 Nmat[7] * ipol_2x2x2_pkgs[7].S		\
	 );					\
  } 
#define COMP_V1(V, DIR) \
   { V.DIR = (                                   \
           Nmat[0] * ipol_2x2x2_pkgs[0].V.DIR + \
           Nmat[1] * ipol_2x2x2_pkgs[1].V.DIR + \
           Nmat[2] * ipol_2x2x2_pkgs[2].V.DIR + \
           Nmat[3] * ipol_2x2x2_pkgs[3].V.DIR + \
           Nmat[4] * ipol_2x2x2_pkgs[4].V.DIR + \
           Nmat[5] * ipol_2x2x2_pkgs[5].V.DIR + \
           Nmat[6] * ipol_2x2x2_pkgs[6].V.DIR + \
           Nmat[7] * ipol_2x2x2_pkgs[7].V.DIR   \
           );                                   \
  } 
#define COMP_V(V) { COMP_V1 (V, x); COMP_V1 (V, y); COMP_V1 (V, z); } 

  COMP_T (displ_grad);
  COMP_T (firstPK);
  COMP_S (W);

  if (iiopts.include_dynamic_effects) {
    COMP_T (displ_grad_dot);
    COMP_V (vel);
    COMP_V (acc);
    COMP_S (K);
  } else {
    MAT3DUTIL_ZERO_3X3 (displ_grad_dot);
    VECMAC_SCALAR (vel, =, 0);
    VECMAC_SCALAR (acc, =, 0);
    K = 0;
  }

  acca = &acc.x;
  vela = &vel.x;

  FOR (mode, 3) {

    r_func ();
    theta_func (mode);
    aux_fields ();

    temp = MAT3DUTIL_DBL_PROD_3X3 (STRESS, iiopts.aux_strain) * iiopts.q_der[0];
    ii_contrib[mode] -= temp;

#if COMPUTE_AUX_SIFS
    ii_aux_contrib[mode] -=
      MAT3DUTIL_DBL_PROD_3X3 (iiopts.aux_stress, iiopts.aux_strain) 
        * iiopts.q_der[0];
#endif
    FOR (i, 3) {
      ii_contrib[mode] +=
        iiopts.rho * (iiopts.aux_vel[i] * VELOCITY[i]) * iiopts.q_der[0];
#if COMPUTE_AUX_SIFS
      ii_aux_contrib[mode] +=
        iiopts.rho * (iiopts.aux_vel[i] * iiopts.aux_vel[i]) * iiopts.q_der[0];
#endif
    } 
    FOR (i, 3) {
      FOR (j, 3) {
        temp = (iiopts.aux_stress[i][j] * DISPL_GRAD[i][0] +
                STRESS[i][j] * iiopts.aux_displ_grad[i][0]
                ) * iiopts.q_der[j];
        ii_contrib[mode] += temp;
#if COMPUTE_AUX_SIFS
        ii_aux_contrib[mode] +=
          (iiopts.aux_stress[i][j] * iiopts.aux_displ_grad[i][0] +
           iiopts.aux_stress[i][j] * iiopts.aux_displ_grad[i][0]
           ) * iiopts.q_der[j];
#endif
      } 
    } 
    FOR (i, 3) {
      ii_contrib[mode] +=
        iiopts.rho * (  iiopts.aux_acc[i] * DISPL_GRAD[i][0]
                      + ACCELERATION[i] * iiopts.aux_displ_grad[i][0]
                      - iiopts.aux_vel[i] * DISPL_GRAD_DOT[i][0]
                      - VELOCITY[i] * iiopts.aux_displ_grad_dot[i][0]
                      ) * iiopts.q;
#if COMPUTE_AUX_SIFS
      ii_aux_contrib[mode] +=
        iiopts.rho * (  iiopts.aux_acc[i] * iiopts.aux_displ_grad[i][0]
                      + iiopts.aux_acc[i] * iiopts.aux_displ_grad[i][0]
                      - iiopts.aux_vel[i] * iiopts.aux_displ_grad_dot[i][0]
                      - iiopts.aux_vel[i] * iiopts.aux_displ_grad_dot[i][0]
                      ) * iiopts.q;
#endif
    } 
  } 

  return;
} 

#endif /* EFG_WASP  */
