#include "Wconfig.h"

#ifdef EFG_WASP

#include "WEFG.h"
#include "WEFGsf.h"
#include "Wdomain.h"
#include "Welem.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"

/* RCSHEADER "$Id: efgn.c,v 1.2 2004/09/06 01:17:58 cvs Exp $" */

static void 
fix_if_on_symm_plane(WEFG_node_t *en, WEFG_symm_plane_t *sp);

/*
   Number equations.
 */
W_index_t
efgn_number_eqns(WEFG_node_t *en, /* EFG node */
                 W_index_t eqn, /* last number (pass 0 to start)  */
                 ListOf(WEFG_symm_plane_t *) symm_planes /* list of symmetry planes (may be NULL) */
                 )
{
  W_index_t i, actual_num_of_eqns = 0;

  if (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    /* Makes sense only for EFG nodes associated to the internal */
    /* degrees of freedom; the FE_IF_NODEs are associated to global dofs. */
    if (symm_planes != NULL) {
      WEFG_symm_plane_t *sp;
      loop_through_list(symm_planes, sp, WEFG_symm_plane_t *) {
        fix_if_on_symm_plane(en, sp);
      }
    }
    for (i = 0; i < en->ndata.num_of_eqns; i++)
      if (en->ndata.eq_numbers[i] == FREE_DOF) { 
        en->ndata.eq_numbers[i] = ++eqn;
        actual_num_of_eqns = i+1;
      }
    en->ndata.num_of_eqns = actual_num_of_eqns;
  }
  return eqn;
}

/* fix node on symmetry plane
   fix node on symmetry plane.

   warning:

   not yet functional
 */
static void
fix_if_on_symm_plane(WEFG_node_t *en, WEFG_symm_plane_t *sp)
{
  double dist;
  W_vect_t v;

  v3_sub(&en->ndata.ref_coords, &sp->center, &v);
  dist = fabs(v3_dot(&v, &sp->plane.n));
  if (dist < sp->tolerance) {
    if (sp->direction_of_normal == 'X')
      en->ndata.eq_numbers[0] = FIXED_DOF;
    if (sp->direction_of_normal == 'Y')
      en->ndata.eq_numbers[1] = FIXED_DOF;
    if (sp->direction_of_normal == 'Z')
      en->ndata.eq_numbers[2] = FIXED_DOF;
  }
}



/* create a new EFG node.
   create a new EFG node.
 */
WEFG_node_t * /* node or NULL if not enough memory */
new_efg_node (WEFG_superelem_t *se, /* EFG superelement */
              W_loc_t *loc, /* location */
              W_float_t support_size /* support size (interpreted by the
                                        selected weight function)
                                      */
              )
{
  WEFG_node_t *en;
  static W_index_t id = 0;
  W_loc_t zerov = {0,0,0};

  en = w_new(WEFG_node_t);
  CHECK_MALLOC(en);

  en->id                         = ++id;
  en->type                       = WEFG_TRUE_EFG_NODE;
  en->support_size               = support_size;
  if (en->support_size <= 0) {
    sprintf (W_buf (), "node %d: support size = %g <= 0",
             (int)en->id, en->support_size);
    REPORT_WARNING (se->domain->errh, W_buf ());
  }
  en->ndata.ref_coords        = *loc;
  en->ndata.displ_t           = zerov;
  en->ndata.displ_t_minus_dt  = zerov;
  en->ndata.velocity_t        = zerov; 
  en->ndata.acceleration_t    = zerov;
  en->ndata.num_of_eqns       = 3;
  en->ndata.eq_numbers
    = w_new_array (en->ndata.num_of_eqns, W_index_t);
  CHECK_MALLOC(en->ndata.eq_numbers);
  en->ndata.eq_numbers[0]     = se->general_constraint[0];
  en->ndata.eq_numbers[1]     = se->general_constraint[1];
  en->ndata.eq_numbers[2]     = se->general_constraint[2];
  en->fen = 0; /* no FE node */
#ifdef X_XFACE
  en->node_marker = NULL;
#endif

  return en;
}

/*
   Change node type.
   Change node type to `FE interface node'.
   Insert a reference to a FE node into the EFG node.
 */
void
efgn_change_to_fe_if_node(WEFG_node_t *en, W_node_t *fen)
{
  en->type     = WEFG_FE_IF_NODE;
  en->fen = fen;
}

/*
   Change node type.
   Change node type to `node driving an FE node'.
 */
void
efgn_change_to_driving_node (WEFG_node_t *en, W_node_t *fen)
{
  en->fen = fen;
}

/*
   Change node type.
   Change node type to `EFG interface node'.
 */
void
efgn_change_to_efg_if_node(WEFG_node_t *en)
{
  en->type = WEFG_EFG_IF_NODE;
}


/*
   Apply initial conditions.
   Apply initial conditions (velocity).
 */
void
efgn_apply_init_cond(WEFG_node_t *en, /* EFG node */
                     W_vect_t *velocity, /* velocity */
                     W_float_t time_step /* time step */
                     )
{
  if        (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    en->ndata.velocity_t.x = velocity->x;
    en->ndata.velocity_t.y = velocity->y;
    en->ndata.velocity_t.z = velocity->z;
    en->ndata.displ_t_minus_dt.x
      = - en->ndata.velocity_t.x * time_step;
    en->ndata.displ_t_minus_dt.y =
      - en->ndata.velocity_t.y * time_step;
    en->ndata.displ_t_minus_dt.z =
      - en->ndata.velocity_t.z * time_step;
  } else if (EFGN_IS_FE_IF(en)) {
    double dt = time_step;
    W_vect_t v = *velocity, a, pd;
    W_node_t *n = en->fen;
    W_bool_t apply[6], is_fixed[6];
    WN_is_fixed (n, is_fixed);
    { int i; FOR (i, 6) apply[i] = ! is_fixed[i]; }
    if (is_fixed[0]) v.x = 0;
    if (is_fixed[1]) v.y = 0;
    if (is_fixed[2]) v.z = 0;
    WN_get_acceleration (n, &a.x, &a.y, &a.z);
    VECMAC_3_VEC (pd, =, -dt * v, +, dt*dt * a);
    WN_set_velocity (n, &v, apply);
    WN_set_prev_displ (n, &pd, apply);
  } else {
    REPORT_FATAL_ERROR(NULL, "bad EFG node type");
  }
}

/*
   Access
   reference coordinates.
   Access
   reference coordinates
   of the node.
 */
void
efgn_ref_coords(SFL_efg_node_id_t efgn, /* EFG node */
                double *x, /* output: x-coordinate */
                double *y, /* output: y-coordinate */
                double *z /* output: z-coordinate */
                )
{
  WEFG_node_t *en = (WEFG_node_t *)efgn;
  
  if        (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    *x = en->ndata.ref_coords.x;
    *y = en->ndata.ref_coords.y;
    *z = en->ndata.ref_coords.z;
  } else if (EFGN_IS_FE_IF(en)) {
    WN_get_ref_coords (en->fen, x, y, z);
  } else {
    REPORT_FATAL_ERROR(NULL, "bad EFG node type");
  }
}

/*
   Access
    coordinates from the previous time.
   Access
   coordinates of the node from the previous time
 */
void
efgn_prev_coords(WEFG_node_t *en,
                 double *x, /* output: x-coordinate */
                double *y, /* output: y-coordinate */
                double *z /* output: z-coordinate */
                 )
{
  if        (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    *x = en->ndata.ref_coords.x + en->ndata.displ_t_minus_dt.x;
    *y = en->ndata.ref_coords.y + en->ndata.displ_t_minus_dt.y;
    *z = en->ndata.ref_coords.z + en->ndata.displ_t_minus_dt.z;
  } else if (EFGN_IS_FE_IF(en)) {
    WN_get_prev_coords (en->fen, x, y, z);
  } else {
    REPORT_FATAL_ERROR(NULL, "bad EFG node type");
  }
}

/*
   Access
   current coordinates.
   Access
   current coordinates of the node.
 */
void
efgn_curr_coords(WEFG_node_t *en,
                 double *x, /* output: x-coordinate */
                double *y, /* output: y-coordinate */
                double *z /* output: z-coordinate */
                 )
{
  if        (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    *x = en->ndata.ref_coords.x + en->ndata.displ_t.x;
    *y = en->ndata.ref_coords.y + en->ndata.displ_t.y;
    *z = en->ndata.ref_coords.z + en->ndata.displ_t.z;
  } else if (EFGN_IS_FE_IF(en)) {
    WN_get_coords (en->fen, x, y, z);
  } else {
    REPORT_FATAL_ERROR(NULL, "bad EFG node type");
  }
}

/*
   Access
   current velocity.
   Access
   current velocity of the node.
 */
void
efgn_curr_vel (WEFG_node_t *en,
               double *x, /* output: x-component */
                double *y, /* output: y-component */
                double *z /* output: z-component */
               )
{
  if        (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    *x = en->ndata.velocity_t.x;
    *y = en->ndata.velocity_t.y;
    *z = en->ndata.velocity_t.z;
  } else if (EFGN_IS_FE_IF(en)) {
    WN_get_velocity (en->fen, x, y, z);
  } else {
    REPORT_FATAL_ERROR(NULL, "bad EFG node type");
  }
}


/*
   Access
   current acceleration.
   Access
   current acceleration of the node.
 */
void
efgn_curr_accel(WEFG_node_t *en,
                double *x, /* output: x-component */
                double *y, /* output: y-component */
                double *z /* output: z-component */
                )
{
  if        (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    *x = en->ndata.acceleration_t.x;
    *y = en->ndata.acceleration_t.y;
    *z = en->ndata.acceleration_t.z;
  } else if (EFGN_IS_FE_IF(en)) {
    WN_get_acceleration (en->fen, x, y, z);
  } else {
    REPORT_FATAL_ERROR(NULL, "bad EFG node type");
  }
}


/*
   Perform Warburton update.
   Perform Warburton's central difference update on the EFG node.
   See Krysl, Belytschko, Computers & Structures (1998).
 */
void
efgn_Warburton_update(WEFG_node_t *en, /* EFG node */
                      WEFG_superelem_t *se, /* EFG superelement */
                      W_float_t dt /* time step */
                      )
{
  W_float_t dt2 = dt * dt, X, A, c_t_minus_dt, dt_inv = 1 / dt;
  W_index_t ne;

  if ((EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en))) {
    WEFG_node_data_t *n = &en->ndata;
    if ((ne = n->eq_numbers[W_DX]) > 0) {
      A = GET_FROM_UPDATE_VECT(se, ne);
      X = (-n->displ_t_minus_dt.x + 2 * n->displ_t.x + dt2 * A);
      c_t_minus_dt           = n->displ_t_minus_dt.x;
      n->displ_t_minus_dt.x  = n->displ_t.x;
      n->displ_t.x           = X;
      n->acceleration_t.x    = A;
      n->velocity_t.x        = (n->displ_t.x - c_t_minus_dt) * dt_inv * 0.5;
    } 
    if ((ne = n->eq_numbers[W_DY]) > 0) {
      A = GET_FROM_UPDATE_VECT(se, ne);
      X = (-n->displ_t_minus_dt.y + 2 * n->displ_t.y + dt2 * A);
      c_t_minus_dt           = n->displ_t_minus_dt.y;
      n->displ_t_minus_dt.y  = n->displ_t.y;
      n->displ_t.y           = X;
      n->acceleration_t.y    = A;
      n->velocity_t.y        = (n->displ_t.y - c_t_minus_dt) * dt_inv * 0.5;
    }
    if ((ne = n->eq_numbers[W_DZ]) > 0) {
      A = GET_FROM_UPDATE_VECT(se, ne);
      X = (-n->displ_t_minus_dt.z + 2 * n->displ_t.z + dt2 * A);
      c_t_minus_dt           = n->displ_t_minus_dt.z;
      n->displ_t_minus_dt.z  = n->displ_t.z;
      n->displ_t.z           = X;
      n->acceleration_t.z    = A;
      n->velocity_t.z        = (n->displ_t.z - c_t_minus_dt) * dt_inv * 0.5;
    }
  }
}

/*
   Perform Newmark update.
   Perform Newmark's central difference update on the EFG node.
   See Krysl, Belytschko, Computers & Structures (1998).
 */
void
efgn_Newmark_update(WEFG_node_t *en, /* EFG node */
                      WEFG_superelem_t *se, /* EFG superelement */
                      W_float_t dt /* time step */
                    )
{
  W_float_t dt2 = dt * dt, A;
  W_index_t ne;

  if ((EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en))) {
    WEFG_node_data_t *n = &en->ndata;
    if ((ne = n->eq_numbers[W_DX]) > 0) {
      A = GET_FROM_UPDATE_VECT(se, ne);
      n->velocity_t.x       += dt * 0.5 * (n->acceleration_t.x + A);
      n->displ_t_minus_dt.x  = n->displ_t.x;
      n->displ_t.x          += dt * n->velocity_t.x + 0.5 * dt2 * A;
      n->acceleration_t.x    = A;
    } 
    if ((ne = n->eq_numbers[W_DY]) > 0) {
      A = GET_FROM_UPDATE_VECT(se, ne);
      n->velocity_t.y       += dt * 0.5 * (n->acceleration_t.y + A);
      n->displ_t_minus_dt.y  = n->displ_t.y;
      n->displ_t.y          += dt * n->velocity_t.y + 0.5 * dt2 * A;
      n->acceleration_t.y    = A;
    }
    if ((ne = n->eq_numbers[W_DZ]) > 0) {
      A = GET_FROM_UPDATE_VECT(se, ne);
      n->velocity_t.z       += dt * 0.5 * (n->acceleration_t.z + A);
      n->displ_t_minus_dt.z  = n->displ_t.z;
      n->displ_t.z          += dt * n->velocity_t.z + 0.5 * dt2 * A;
      n->acceleration_t.z    = A;
    }
  }
}


#ifdef X_XFACE

static WCRec pnts[2] = {{0, 0, 0}, {0, 0, 0}};
static BOOLEAN first = TRUE;
static EPixel efgn_color = 0;
static EPixel efgn_fe_if_color = 0;
static EPixel efgn_efg_if_color = 0;

void
efgn_modify_grep(WEFG_superelem_t *se, double displ_scale, WEFG_node_t *n)
{
  W_loc_t rl, cl;
  double x, y, z;
  
  if (first) {
    efgn_color = ColorGetPixelFromString("magenta", &first);
    efgn_fe_if_color = ColorGetPixelFromString("cyan", &first);
    efgn_efg_if_color = ColorGetPixelFromString("yellow", &first);
  }
  efgn_ref_coords(n, &x, &y, &z); rl.x = x; rl.y = y; rl.z = z;
  efgn_curr_coords(n, &x, &y, &z); cl.x = x; cl.y = y; cl.z = z;
  pnts[0].x = (rl.x + displ_scale * (cl.x - rl.x));
  pnts[0].y = (rl.y + displ_scale * (cl.y - rl.y));
  pnts[0].z = (rl.z + displ_scale * (cl.z - rl.z));
  if (n->node_marker == NULL) {
    n->node_marker = CreateMarker3D(&pnts[0]);
    EMNoDrawAddGraphicsToModel(ESIModel (), n->node_marker);
    EASValsSetMType(X_LETTER_MARKER);
    if (EFGN_IS_TRUE_EFG(n))
      EASValsSetColor(efgn_color);
    else if (EFGN_IS_EFG_IF(n))
      EASValsSetColor(efgn_efg_if_color);
    else if (EFGN_IS_FE_IF(n))
      EASValsSetColor(efgn_fe_if_color);
    EASValsSetMSize(4);
    EASValsSetLayer(EFG_NODE_LAYER);
    EGWithMaskChangeAttributes(  COLOR_MASK | LAYER_MASK
                               | MTYPE_MASK | MSIZE_MASK,
                               n->node_marker);
  } else {
    EGModifyGraphicsGeometry(n->node_marker, &pnts[0]);
  }
}

#endif /* ifdef X_XFACE */

/*
   Assemble force.
   Assemble force at the EFG node.
   The force is assembled either to the superelement array of
   effective loads, for a true EFG node or an `interface EFG' node, or
   to the FE node, for an `interface FE' node.
 */
void
efgn_assemble_force(WEFG_node_t *en, /* EFG node */
                    WEFG_superelem_t *se, /* EFG superelement */
                    W_vect_t *f /* force vector */
                    )
{
  if (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    register W_index_t AEL_ne;                              
    /*    fprintf(stderr, "assembling to true efg: %g %g %g\n", f->x, f->y, f->z);*/
    if ((AEL_ne = en->ndata.eq_numbers[W_DX]) > 0) { 
      se->effective_loads.v[AEL_ne] += f->x;
    }                                              
    if ((AEL_ne = en->ndata.eq_numbers[W_DY]) > 0) { 
      se->effective_loads.v[AEL_ne] += f->y;
    }                                              
    if ((AEL_ne = en->ndata.eq_numbers[W_DZ]) > 0) { 
      se->effective_loads.v[AEL_ne] += f->z;
    }                                              
  } else if (EFGN_IS_FE_IF(en)) {
        fprintf(stderr, "assembling to fe if: %g %g %g\n", f->x, f->y, f->z);
    WN_assemble_force(en->fen, f);
  } else {
    REPORT_FATAL_ERROR(NULL, "bad EFG node type");
  }
}

/*
   Assemble mass.
   Assemble mass at the EFG node.
   The mass is assembled either to the superelement mass matrix,
   for a true EFG node or an `interface EFG' node, or
   to the FE node, for an `interface FE' node.
 */
void
efgn_assemble_mass (WEFG_node_t *en, /* EFG node */
                    WEFG_superelem_t *se, /* EFG superelement */
                    double M /* mass */
                    )
{
  if (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    register W_index_t ATM_ne;                         
    if ((ATM_ne = en->ndata.eq_numbers[W_DX]) > 0)
      se->mass_matrix.v[ATM_ne] += (M);       
    if ((ATM_ne = en->ndata.eq_numbers[W_DY]) > 0)
      se->mass_matrix.v[ATM_ne] += (M);       
    if ((ATM_ne = en->ndata.eq_numbers[W_DZ]) > 0)
      se->mass_matrix.v[ATM_ne] += (M);
    if (en->fen) WN_assemble_mass(en->fen, M);
  } else if (EFGN_IS_FE_IF(en)) {
    if (!se->internal_mass_only)
      WN_assemble_mass(en->fen, M);
  } else {
    REPORT_FATAL_ERROR(NULL, "bad EFG node type");
  }  
}


/*
   Compute kinetic energy.
   Compute kinetic energy associated with a EFG node.
 */
double
efgn_kinetic_energy(WEFG_node_t *en, /* EFG node */
                    WEFG_superelem_t *se, /* EFG superelement */
                    double dt /* time step */
                    )
{
  W_float_t e = 0, v;
  W_index_t ne;
#define FROM_VELOCITY 1

  if ((ne = en->ndata.eq_numbers[W_DX]) > 0) {
    v =
#     if FROM_VELOCITY
      en->ndata.velocity_t.x;
#     else
      (en->ndata.displ_t.x - en->ndata.displ_t_minus_dt.x) / dt;
#     endif
    e += v * v * se->mass_matrix.v[ne];
  }
  if ((ne = en->ndata.eq_numbers[W_DY]) > 0) {
    v =
#     if FROM_VELOCITY
      en->ndata.velocity_t.y;
#     else
      (en->ndata.displ_t.y - en->ndata.displ_t_minus_dt.y) / dt;
#     endif
    e += v * v * se->mass_matrix.v[ne];
  }
  if ((ne = en->ndata.eq_numbers[W_DZ]) > 0) {
    v =
#     if FROM_VELOCITY
      en->ndata.velocity_t.z;
#     else
      (en->ndata.displ_t.z - en->ndata.displ_t_minus_dt.z) / dt;
#     endif
    e += v * v * se->mass_matrix.v[ne];
  }

  return 0.5 * e;
}
  

/* access support size.
    access support size associated with the EFG node.

    caveat:

    While there are three values on output, the node really recognizes
    only isotropic supports.  Hence, all three values on output are
    the same.
 */
void
efgn_support_size (SFL_efg_node_id_t efgp,
                   double *sx, /* output: support size in the global x direction */
                   double *sy, /* output: support size in the global y direction */
                   double *sz /* output: support size in the global z direction */
                   )
{
  WEFG_node_t *en = (WEFG_node_t *)efgp;
  if (en->support_size == 0)
    REPORT_FATAL_ERROR(NULL, "zero support_size");
  *sx = *sy = *sz = en->support_size;
}

/*
   Assemble damping force.
   Assemble damping force at the EFG node.
   The damping force is assembled either to the superelement array of
   effective loads, for a true EFG node or an `interface EFG' node, or
   to the FE node, for an `interface FE' node.
 */
void
efgn_assemble_damping(WEFG_node_t *en, /* EFG node */
		      WEFG_superelem_t *se, /* EFG superelement */
                      double damping_coeff, /* damping coefficient */
		      double dt /* time step */
                      )
{
  double c = damping_coeff;
  
  if (EFGN_IS_TRUE_EFG(en) || EFGN_IS_EFG_IF(en)) {
    register W_index_t AEL_ne;                              
    if ((AEL_ne = en->ndata.eq_numbers[W_DX]) > 0) { 
      se->effective_loads.v[AEL_ne] -=
	c * se->mass_matrix.v[AEL_ne] * en->ndata.velocity_t.x;
    }                                              
    if ((AEL_ne = en->ndata.eq_numbers[W_DY]) > 0) { 
      se->effective_loads.v[AEL_ne] -=
	c * se->mass_matrix.v[AEL_ne] * en->ndata.velocity_t.y;
    }                                              
    if ((AEL_ne = en->ndata.eq_numbers[W_DZ]) > 0) { 
      se->effective_loads.v[AEL_ne] -=
	c * se->mass_matrix.v[AEL_ne] * en->ndata.velocity_t.z;
    }                                              
  }
} 


/*
   Print info.
   Print EFG node info.
 */
void
efgn_print_info(WEFG_node_t *en, /* EFG node */
                FILE *fp, /* stream */
                int indent, /* initial indent */
                WEFG_superelem_t *se /* EFG superelement */
                )
{
#define PRINT_VECTOR(name, member)                      \
  {                                                     \
    W_INDENT(fp, indent);                               \
    fprintf(fp,  name " = {%g, %g, %g}\n",              \
            en->ndata.member.x,                     \
            en->ndata.member.y,                     \
            en->ndata.member.z);                    \
  }

  W_INDENT(fp, indent);
  fprintf(fp, "Node # %d\n", en->id);
  
  indent += 3;
  W_INDENT(fp, indent);
  fprintf(fp, "Node type = %s\n",
          (EFGN_IS_TRUE_EFG(en)
           ? "TRUE EFG node"
           : (EFGN_IS_EFG_IF(en)
              ? "INTERFACE EFG node"
              : "FE/INTERFACE node"))
          );
  W_INDENT(fp, indent);
  fprintf(fp, "Support size = %g\n", en->support_size);
  if (EFGN_IS_TRUE_EFG(en) ||
      EFGN_IS_EFG_IF(en)) {
    PRINT_VECTOR("Ref. location  ", ref_coords);
    PRINT_VECTOR("Displacement   ", displ_t);
    PRINT_VECTOR("Prev. displ.   ", displ_t_minus_dt);
    PRINT_VECTOR("Velocity       ", velocity_t);
    PRINT_VECTOR("Acceleration   ", acceleration_t);
    W_INDENT(fp, indent);
    fprintf(fp, "# of equations = %ld\n", en->ndata.num_of_eqns);
    {
      double mass = 0;
      if (en->ndata.eq_numbers[W_DX] > 0)
        mass = se->mass_matrix.v[en->ndata.eq_numbers[W_DX]];
      if (en->ndata.eq_numbers[W_DY] > 0)
        mass = se->mass_matrix.v[en->ndata.eq_numbers[W_DY]];
      if (en->ndata.eq_numbers[W_DZ] > 0)
        mass = se->mass_matrix.v[en->ndata.eq_numbers[W_DZ]];
      W_INDENT(fp, indent);
      fprintf(fp, "Associated mass = %g\n", mass);
    }
  } else if (EFGN_IS_FE_IF(en)) {
    W_INDENT(fp, indent);
    fprintf(fp, "Tied to FE node # %ld\n", en->fen->id);
  } else {
    REPORT_FATAL_ERROR(NULL, "Unknown EFG node type");
  }
}

#endif /* EFG_WASP  */
