#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "Wdomain.h"
#include "Wmat3d.h"
#include "Wnode.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFG.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGsf.h"




static SFL_sf_table_t sf = NULL;

#ifndef DRAW_X_TO_XI_RAYS 
#define DRAW_X_TO_XI_RAYS 0
#endif

#if DRAW_X_TO_XI_RAYS
static void
draw_ray(W_loc_t *from, W_loc_t *to)
{
  static GraphicObj *ray = NULL;
  WCRec pts[2] = {{0,0,0},{0,0,0}};

  if (ray == NULL) {
    int suc = 0;
    ray = CreateLine3D(pts);
    EASValsSetColor(ColorGetPixelFromString("yellow", &suc));
    EGWithMaskChangeAttributes(COLOR_MASK, ray);
  }
  if (to != NULL) {
    pts[0] = *from; pts[1] = *to;
    EGModifyGraphicsGeometry(ray, pts);
    EMDrawGraphics(ESIModel(), ray);
  }
}
#endif


/*
   Create a shape function table.
   Create a EFG-superelement-specific shape function table at point.
   
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
efgsf_create_sftab (WEFG_superelem_t *se, /* EFG superelement */
                    WEFG_sf_tab_t *sftab, /* input/output: shape function
                                           table */
                    SFL_sf_table_t sf, /* input: ESFLib shape function table */
                    W_loc_t *at, /* location */
                    W_bool_t allow_draw_rays /* graphical debugging */
                    )
{
  int k, ne;
  double phi, dphi[3];
  double time_stamp = se->invalid_after_time;

  sftab->num_of_entries = SFL_num_of_entries(sf);
  if (sftab->rows != NULL) {
    if (sftab->tot_of_entries < sftab->num_of_entries) {
      w_free(sftab->rows); sftab->rows = NULL;
    }
  }
  if (sftab->rows == NULL) {
    sftab->tot_of_entries = sftab->num_of_entries;
    sftab->rows  = w_new_array(sftab->tot_of_entries, WEFG_sf_tab_row_t);
    CHECK_MALLOC(sftab->rows);
  }

  ne = SFTAB_NUM_OF_ENTRIES((*sftab));
  FOR(k, ne) {
    if (!SFL_inspect_row(sf, k, &phi, dphi)) {
      REPORT_ERROR(se->domain->errh, (char *)SFL_explain_error(sf));
    }
    SFTAB_PHI((*sftab), k)    = phi;
    SFTAB_DPHI_X((*sftab), k) = dphi[0];
    SFTAB_DPHI_Y((*sftab), k) = dphi[1];
    SFTAB_DPHI_Z((*sftab), k) = dphi[2];
    SFTAB_NODE((*sftab), k)   = SFL_efg_point_at(sf, k);
  }

#if DRAW_X_TO_XI_RAYS
  if (se->draw_rays && allow_draw_rays) {
    FOR(k, ne) {
      W_loc_t enrc;
      EFGN_REF_COORDS (SFTAB_NODE((*sftab), k), enrc.x, enrc.y, enrc.z);
      draw_ray(&enrc, at);
    }
  }
#endif

  sftab->time_stamp = time_stamp;
  
  return TRUE;
}



/*
   Create shape function table.
   Create shape function table specific to the EFG shape function library used.
 */
SFL_sf_table_t /* table */
efgsf_sf_table (WEFG_superelem_t *se /* EFG superelement */
                )
{
  if (sf == NULL) { 
#   ifdef USE_ESFLIB
    sf = ESFL3_new_shape_func(se->esflib);
    if (sf == NULL) REPORT_ERROR(se->domain->errh, (char *)ESFL3_explain_error(se->esflib));
#   else
    sf = ESP_3_new_shape_func();
    if (sf == NULL) REPORT_ERROR(se->domain->errh, ESP_3_explain_error());
#   endif
  }

  return sf;
}

/*
   Evaluate current location.
   Evaluate current location of material point at which
   the shape function table on input had been constructed.
 */
void
efgsf_eval_curr_loc(WEFG_sf_tab_t *sftab, /* shape function table */
                    double *ptx, /* output: x-coordinate */
                    double *pty, /* output: y-coordinate */
                    double *ptz /* output: z-coordinate */
                    )
{
  int k, ne;
  double phi, x, y, z;
  WEFG_node_t *efgn;

  ne = SFTAB_NUM_OF_ENTRIES((*sftab));
  *ptx = *pty = *ptz = 0;
  FOR(k, ne) {
    efgn = SFTAB_NODE((*sftab), k);
    EFGN_CURR_COORDS(efgn, x, y, z);
    phi  = SFTAB_PHI((*sftab), k);
    *ptx += phi * x;
    *pty += phi * y;
    *ptz += phi * z;
  }
}




/*
   Evaluate current velocity.
   Evaluate current velocity of material point at which
   the shape function table on input had been constructed.
 */
void
efgsf_eval_curr_vel(WEFG_sf_tab_t *sftab, /* shape function table */
                    double *vx, /* output: x-component */
                    double *vy, /* output: y-component */
                    double *vz /* output: z-component */
                    )
{
  int k, ne;
  double phi, x, y, z;
  WEFG_node_t *efgn;

  ne = SFTAB_NUM_OF_ENTRIES((*sftab));
  *vx = *vy = *vz = 0;
  FOR(k, ne) {
    efgn = SFTAB_NODE((*sftab), k);
    EFGN_CURR_VELOC(efgn, x, y, z);
    phi  = SFTAB_PHI((*sftab), k);
    *vx += phi * x;
    *vy += phi * y;
    *vz += phi * z;
  }
}

/*
   Evaluate current acceleration.
   Evaluate current acceleration of material point at which
   the shape function table on input had been constructed.
 */
void
efgsf_eval_curr_acc(WEFG_sf_tab_t *sftab, /* shape function table */
                    double *ax, /* output: x-component */
                    double *ay, /* output: y-component */
                    double *az /* output: z-component */
                    )
{
  int k, ne;
  double phi, x, y, z;
  WEFG_node_t *efgn;

  ne = SFTAB_NUM_OF_ENTRIES((*sftab));
  *ax = *ay = *az = 0;
  FOR(k, ne) {
    efgn = SFTAB_NODE((*sftab), k);
    EFGN_CURR_ACCEL(efgn, x, y, z);
    phi  = SFTAB_PHI((*sftab), k);
    *ax += phi * x;
    *ay += phi * y;
    *az += phi * z;
  }
}


static int current_node = 0;

/*
   Get the first EFG node.
   This is a callback for use by the EFG shape function library.
   Get the first EFG node from the node search structure which
   has been found during search.
 */
SFL_efg_node_id_t 
efgsf_first (ESFLIB_search_struct_t SS /* search structure */
             )
{
  WEFG_sf_tab_t *sftab = (WEFG_sf_tab_t *)SS;
  current_node = 0;
  return SFTAB_NODE((*sftab), current_node);
}

/*
   Get the next EFG node.
   This is a callback for use by the EFG shape function library.
   Get the next EFG node from the node search structure which
   has been found during search.
 */
SFL_efg_node_id_t 
efgsf_next (ESFLIB_search_struct_t SS /* search structure */
            )
{
  WEFG_sf_tab_t *sftab = (WEFG_sf_tab_t *)SS;
  if (++current_node < SFTAB_NUM_OF_ENTRIES((*sftab)))
    return SFTAB_NODE((*sftab), current_node);
  else
    return NULL;
}




#endif /* EFG_WASP  */
