#include "Wconfig.h"

#ifdef EFG_WASP

#include "Wtypes.h"
#include "Wdomain.h"
#include "Wmat3d.h"
#include "Wmat3dutil.h"
#include "Wnode.h"
#include "Wgaussp.h"
#include "Wxf.h" 
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFG.h"
#include "WEFGsf.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGss.h"


#define VALID(which) which##_is_valid

#define DCL_TENSOR(T)                           \
   static double T[3][3];                       \
   static W_bool_t VALID (T) = FALSE

DCL_TENSOR (F);
DCL_TENSOR (prev_F);
DCL_TENSOR (inv_F);
DCL_TENSOR (rel_F);
DCL_TENSOR (midstep_F);
DCL_TENSOR (inv_rel_F);
DCL_TENSOR (C);
DCL_TENSOR (E);
DCL_TENSOR (D);
DCL_TENSOR (midstep_D);
DCL_TENSOR (midstep_W);
DCL_TENSOR (F_dot);
DCL_TENSOR (L);
DCL_TENSOR (R);
DCL_TENSOR (U);
DCL_TENSOR (V);
DCL_TENSOR (rel_R);
DCL_TENSOR (rel_U);
DCL_TENSOR (rel_V);
DCL_TENSOR (ln_rel_U);


static char *caller_file = NULL;
static int   caller_line = 0;

/*
   Set caller.
   Set caller file name and line from which called.
   For debugging purposes.
 */
void
efgss_set_caller (char *file, int line)
{
  caller_file = file;
  caller_line = line;
}

  
/*
   Prepare for a new integration point.
   Prepare for a new integration point.  Clear flags for cached objects
   forcing thus their re-computation.
 */
void
efgss_at_new_ip (void)
{
#define INVALIDATE(which) VALID (which) = FALSE
  
  INVALIDATE (F);
  INVALIDATE (inv_F);
  INVALIDATE (prev_F);
  INVALIDATE (rel_F);
  INVALIDATE (midstep_F);
  INVALIDATE (inv_rel_F);
  INVALIDATE (C);
  INVALIDATE (E);
  INVALIDATE (D);
  INVALIDATE (midstep_D);
  INVALIDATE (midstep_W);
  INVALIDATE (F_dot);
  INVALIDATE (L);
  INVALIDATE (R);
  INVALIDATE (U);
  INVALIDATE (V);
  INVALIDATE (rel_R);
  INVALIDATE (rel_U);
  INVALIDATE (rel_V);
  INVALIDATE (ln_rel_U);
}


static void
compute_def_grad (ip_ctx_data_t *ipctx);
static void
compute_cauchy_green (ip_ctx_data_t *ipctx);
static void
compute_green_lagrange (ip_ctx_data_t *ipctx);
static void
compute_rel_def_grad (ip_ctx_data_t *ipctx);
static void
compute_midstep_def_grad (ip_ctx_data_t *ipctx);
static void
compute_rate_of_def (ip_ctx_data_t *ipctx);
static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx);
static void
compute_midstep_spin  (ip_ctx_data_t *ipctx);
static void
compute_def_grad_dot (ip_ctx_data_t *ipctx);
static void
compute_vel_grad (ip_ctx_data_t *ipctx);
static void
compute_inv_def_grad (ip_ctx_data_t *ipctx);
static void
compute_polar_decomp (ip_ctx_data_t *ipctx);
static void
compute_rel_polar_decomp (ip_ctx_data_t *ipctx);
static void
compute_ln_rel_U (ip_ctx_data_t *ipctx);


/*
   Get a kinematic variable.
   Get a kinematic variable (tensor) at an integration point.
 */
W_bool_t 
efgss_get_kinem_var (W_mat_3d_ip_ctx_t *ctx, /* integration point context */
                     W_mat_3d_kinem_var_type_enum  type, /* kinematic tensor type */
                     double kinvar[3][3] /* output: tensor components
                                            of the variable */
                     )
{
  ip_ctx_data_t *ipctx =  (ip_ctx_data_t *)ctx->ctx_data;

  switch  (type) {
  case MAT_3D_DEF_GRAD:
    if  (!VALID (F)) compute_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, F);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    if  (!VALID (F)) compute_def_grad_dot (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, F_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    if  (!VALID (rel_F)) compute_rel_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    if  (!VALID (rel_F)) compute_rel_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    if  (!VALID (midstep_F)) compute_midstep_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    if  (!VALID (C)) compute_cauchy_green (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, C);
    break;
  case MAT_3D_RATE_OF_DEF:
    if  (!VALID (D)) compute_rate_of_def (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, D);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    if  (!VALID (midstep_D)) compute_midstep_rate_of_def (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    if  (!VALID (midstep_W)) compute_midstep_spin (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_W);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    if  (!VALID (E)) compute_green_lagrange (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, E);
    break;
  case MAT_3D_POLAR_R:
    if  (!VALID (R)) compute_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, R);
    break;
  case MAT_3D_POLAR_U:
    if  (!VALID (U)) compute_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, U);
    break;
  case MAT_3D_POLAR_REL_R:
    if  (!VALID (rel_R)) compute_rel_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    if  (!VALID (rel_U)) compute_rel_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    if  (!VALID (ln_rel_U)) compute_ln_rel_U (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, ln_rel_U);
    break;
  default:
    REPORT_FATAL_ERROR (NULL, "Unknown kinematic tensor type");
    break;
  }

  return TRUE;
}


#define YELL() fprintf (stderr, "NaN  (%s, l.%d)\n", __FILE__, __LINE__)

/*
   Compute relative deformation gradient.
 */
static void
compute_rel_def_grad (ip_ctx_data_t *ipctx)
{
  WEFG_sf_tab_t sftab = ipctx->sftab;
  int k, n;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  WEFG_node_t *efgn;
  double x, y, z;
  double inv_prev_F[3][3];
  double det_prev_F;
  register double dphix, dphiy, dphiz;
  WEFG_sf_tab_row_t *sftrowp, sftrow;

  if  (!VALID (F)) compute_def_grad (ipctx);
   
  n = SFTAB_NUM_OF_ENTRIES (sftab); 

  /* Deformation gradient at the previous time */
  sftrowp = sftab.rows;
  FOR (k, n) {
    sftrow =  (*sftrowp); sftrowp++;
    efgn   = SFTAB_ROW_NODE (sftrow);
    EFGN_PREV_DISPLS  (efgn, x, y, z);
    dphix  = SFTAB_ROW_DPHI_X (sftrow);
    dphiy  = SFTAB_ROW_DPHI_Y (sftrow);
    dphiz  = SFTAB_ROW_DPHI_Z (sftrow);
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  prev_F[0][0] = f00 + 1;
  prev_F[0][1] = f01;
  prev_F[0][2] = f02;
  
  prev_F[1][0] = f10;
  prev_F[1][1] = f11 + 1;
  prev_F[1][2] = f12;
  
  prev_F[2][0] = f20;
  prev_F[2][1] = f21;
  prev_F[2][2] = f22 + 1;

  det_prev_F = MAT3DUTIL_DET_3X3 (prev_F);
  if  (det_prev_F <= 0) {
    sprintf (W_buf (), "Failed to invert def grad  (called from %s, l.%d)",
	    caller_file, caller_line);
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }

  MAT3DUTIL_INVERT_3X3 (det_prev_F, prev_F, inv_prev_F);

  MAT3DUTIL_MULT_AB_3X3 (rel_F, F, inv_prev_F);

  VALID (rel_F) = TRUE;
}



/*
   Compute the mid-step deformation gradient.
 */
static void
compute_midstep_def_grad (ip_ctx_data_t *ipctx)
{
  double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} }, s[3][3], inv_s[3][3];
  
  if  (!VALID (rel_F)) compute_rel_def_grad (ipctx);

  MAT3DUTIL_ADD_AB_3X3 (s, m, rel_F);
  MAT3DUTIL_SCALE_3X3 (s, 0.5);
  if  (!W_mat3dutil_invert_3x3 (s, inv_s)) {
    sprintf (W_buf (), "Failed to invert  (called from %s, l.%d)",
	    caller_file, caller_line);
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }
  MAT3DUTIL_MULT_AB_3X3 (midstep_F, inv_s, rel_F);

  VALID (midstep_F) = TRUE;
}

/*
   Compute the velocity gradient.
 */
static void
compute_vel_grad (ip_ctx_data_t *ipctx)
{
  if  (!VALID (F)) compute_def_grad (ipctx);
  if  (!VALID (F_dot)) compute_def_grad_dot (ipctx);
  if  (!VALID (inv_F)) compute_inv_def_grad (ipctx);

  MAT3DUTIL_MULT_AB_3X3 (L, F_dot, inv_F);
  
  VALID (L) = TRUE;
}

/*
   Compute rate of deformation.
 */
static void
compute_rate_of_def (ip_ctx_data_t *ipctx)
{
  int i, j;
  
  if  (!VALID (L)) compute_vel_grad (ipctx);

  FOR (i, 3)
    FOR (j, 3) 
      D[i][j] = 0.5 *  (L[i][j] + L[j][i]);

  VALID (D) = TRUE;
}


static int gag = 0;


/* Compute current deformation gradient */
static void
compute_def_grad (ip_ctx_data_t *ipctx)
{
  WEFG_sf_tab_t sftab = ipctx->sftab;
  int k, n;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  WEFG_node_t *efgn;
  double x, y, z;
  register double dphix, dphiy, dphiz;
  WEFG_sf_tab_row_t *sftrowp, sftrow;

  n = SFTAB_NUM_OF_ENTRIES (sftab);

  /* Deformation gradient */
  sftrowp = sftab.rows;
  FOR (k, n) {
    sftrow =  (*sftrowp); sftrowp++;
    efgn   = SFTAB_ROW_NODE (sftrow);
    EFGN_CURR_DISPLS  (efgn, x, y, z);
    dphix  = SFTAB_ROW_DPHI_X (sftrow);
    dphiy  = SFTAB_ROW_DPHI_Y (sftrow);
    dphiz  = SFTAB_ROW_DPHI_Z (sftrow);
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  F[0][0] = f00 + 1;
  F[0][1] = f01;
  F[0][2] = f02;
  
  F[1][0] = f10;
  F[1][1] = f11 + 1;
  F[1][2] = f12;
  
  F[2][0] = f20;
  F[2][1] = f21;
  F[2][2] = f22 + 1;

  VALID (F) = TRUE;
}


/* Compute time derivative of deformation gradient */
static void
compute_def_grad_dot (ip_ctx_data_t *ipctx)
{
  WEFG_sf_tab_t sftab = ipctx->sftab;
  int k, n;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  WEFG_node_t *efgn;
  double x, y, z;
  register double dphix, dphiy, dphiz;
  WEFG_sf_tab_row_t *sftrowp, sftrow;

  n = SFTAB_NUM_OF_ENTRIES (sftab);

  /* Deformation gradient */
  sftrowp = sftab.rows;
  FOR (k, n) {
    sftrow =  (*sftrowp); sftrowp++;
    efgn   = SFTAB_ROW_NODE (sftrow);
    EFGN_CURR_VELOC  (efgn, x, y, z);
    dphix  = SFTAB_ROW_DPHI_X (sftrow);
    dphiy  = SFTAB_ROW_DPHI_Y (sftrow);
    dphiz  = SFTAB_ROW_DPHI_Z (sftrow);
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  }

  F_dot[0][0] = f00;
  F_dot[0][1] = f01;
  F_dot[0][2] = f02;
  
  F_dot[1][0] = f10;
  F_dot[1][1] = f11;
  F_dot[1][2] = f12;
  
  F_dot[2][0] = f20;
  F_dot[2][1] = f21;
  F_dot[2][2] = f22;

  VALID (F_dot) = TRUE;
}



/* Compute the Green-Lagrange tensor */
static void
compute_green_lagrange (ip_ctx_data_t *ipctx)
{
  if  (!VALID (C)) compute_cauchy_green (ipctx);

  /* Green-Lagrange strain tensor */
  E[0][0] = 0.5 *  (C[0][0] - 1);
  E[0][1] = 0.5 * C[0][1];
  E[0][2] = 0.5 * C[0][2];
  E[1][0] = 0.5 * C[1][0];
  E[1][1] = 0.5 *  (C[1][1] - 1);
  E[1][2] = 0.5 * C[1][2];
  E[2][0] = 0.5 * C[2][0];
  E[2][1] = 0.5 * C[2][1];
  E[2][2] = 0.5 *  (C[2][2] - 1);
  
  VALID (E) = TRUE;
}


/* Compute the Cauchy tensor */
static void
compute_cauchy_green (ip_ctx_data_t *ipctx)
{
  if  (!VALID (F)) compute_def_grad (ipctx);
  
  /* Cauchy-Green deformation tensor */
  MAT3DUTIL_MULT_ATB_3X3 (C, F, F);

  VALID (C) = TRUE;
}

/* Compute the inverse deformation gradient */
static void
compute_inv_def_grad (ip_ctx_data_t *ipctx)
{
  if  (!VALID (F)) compute_def_grad (ipctx);

  if  (!W_mat3dutil_invert_3x3 (F, inv_F)) {
    sprintf (W_buf (), "Failed to invert def grad  (called from %s, l.%d)",
	    caller_file, caller_line);
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }

  VALID (inv_F) = TRUE;
}

/* compute mid-step rate of deformation */
static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx)
{
  int i, j;
  double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
  double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
  double s[3][3], inv_s[3][3], q[3][3], r[3][3];
  double inv_dt = 1 / WD_time_step (ipctx->se->domain);
  
  if  (!VALID (rel_F)) compute_rel_def_grad (ipctx);

  MAT3DUTIL_ADD_AB_3X3 (s, m, rel_F);
  MAT3DUTIL_SCALE_3X3 (s, 0.5);
  if  (!W_mat3dutil_invert_3x3 (s, inv_s)) {
    sprintf (W_buf (), "Failed to invert def grad  (called from %s, l.%d)",
	    caller_file, caller_line);
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }
  MAT3DUTIL_ADD_AB_3X3 (q, rel_F, M);
  MAT3DUTIL_MULT_AB_3X3 (r, inv_s, q);

  FOR (i, 3)
    FOR (j, 3) 
      midstep_D[i][j] = 0.5 *  (r[i][j] + r[j][i]);
  MAT3DUTIL_SCALE_3X3 (midstep_D, inv_dt);
  
  VALID (midstep_D) = TRUE;
}



/* compute mid-step spin tensor */
static void
compute_midstep_spin (ip_ctx_data_t *ipctx)
{
  int i, j;
  double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
  double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
  double s[3][3], inv_s[3][3], q[3][3], r[3][3];
  double inv_dt = 1 / WD_time_step  (ipctx->se->domain);
  
  if  (!VALID (rel_F)) compute_rel_def_grad (ipctx);

  MAT3DUTIL_ADD_AB_3X3 (s, m, rel_F);
  MAT3DUTIL_SCALE_3X3 (s, 0.5);
  if  (!W_mat3dutil_invert_3x3 (s, inv_s)) {
    sprintf (W_buf (), "Failed to invert def grad  (called from %s, l.%d)",
	    caller_file, caller_line);
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }
  MAT3DUTIL_ADD_AB_3X3 (q, rel_F, M);
  MAT3DUTIL_MULT_AB_3X3 (r, inv_s, q);

  FOR (i, 3)
    FOR (j, 3) 
      midstep_W[i][j] = 0.5 *  (r[i][j] - r[j][i]);
  MAT3DUTIL_SCALE_3X3 (midstep_W, inv_dt);
  
  VALID (midstep_W) = TRUE;
}



/* compute polar decomposition */
static void
compute_polar_decomp (ip_ctx_data_t *ipctx)
{
  if  (!VALID (F)) compute_def_grad (ipctx);
  
  if (!W_polar_decomposition (F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID (R) = TRUE;
  VALID (U) = TRUE;
}

/* compute relative polar decomposition */
static void
compute_rel_polar_decomp (ip_ctx_data_t *ipctx)
{
  if  (!VALID (rel_F)) compute_rel_def_grad (ipctx);
  
  if (!W_polar_decomposition (rel_F, rel_R, rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID (rel_R) = TRUE;
  VALID (rel_U) = TRUE;
}


/* compute ln(U) */
static void
compute_ln_rel_U (ip_ctx_data_t *ipctx)
{
  if (!VALID(rel_U)) compute_rel_polar_decomp(ipctx);
  
  W_ln_U (rel_U, ln_rel_U);
  
  VALID(ln_rel_U) = TRUE;
}

#endif /* EFG_WASP  */
