#include "Wconfig.h"

#ifdef EFG_WASP

#include "WEFG.h"
#include "WEFGsf.h"
#include "Wdomain.h"
#include "Welem.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "WEFGutil.h"
#include "WEFGpriv.h"
#include "WEFGss.h"
#include "WEFGsf.h"
#include "WEFGcr.h"
#include "Whistry.h"


static W_loc_t  bi_unit_cube_vertices[8] = {
  {-1, -1, -1},
  { 1, -1, -1},
  { 1,  1, -1},
  {-1,  1, -1},
  {-1, -1,  1},
  { 1, -1,  1},
  { 1,  1,  1},
  {-1,  1,  1}
};



/*
   Find parametric coordinates of vertex
   in an incident interface integration volume.
 */
static void
find_if_hv(WEFG_vertex_t *v /* integration volume vertex */,
           WEFG_hex_volume_t **if_hv, /* output: integration volume */
           double *xi, /* output: parametric coordinate */
           double *eta, /* output: parametric coordinate */
           double *theta /* output: parametric coordinate */
           )
{
  WEFG_hex_volume_t *h;
  int i;
  
  if (v->incident_hex_volumes == NULL) {
    *if_hv = NULL;
  }
  loop_through_list(v->incident_hex_volumes, h, WEFG_hex_volume_t *) {
    FOR(i, 8) {
      if (h->vertex[i]->efgn != NULL) {
        if (h->vertex[i]->efgn == v->efgn) {
          *if_hv = h;
          *xi    = bi_unit_cube_vertices[i].x;
          *eta   = bi_unit_cube_vertices[i].y;
          *theta = bi_unit_cube_vertices[i].z;
          return;
        }
      }
    }
  }
  
}


/*
   Evaluate shape function table.
   Evaluate shape function table at an integration volume vertex.
   The procedure varies depending on whether the vertex
   is associated with an EFG node or not.  

   Erorrs:

   If the evaluation fails for any reason, a fatal error is reported
   specifying the reasons for the failure as returned by ESFLib.
 */
static void
eval_sftab(WEFG_superelem_t *se,
           WEFG_vertex_t *v /* integration volume vertex */
           )
{
  SFL_sf_table_t sf;
  WEFG_eval_for_packet_t efp;

  if (SFTAB_INVALID (v->sftab, se, &v->loc)) {
    efp.se                               = se;
    efp.pt_wrt_crack_surf_classification = WEFG_POS_WRT_SURF_UNSPEC;
    efp.crack_surf                       = NULL;
  
    sf = efgsf_sf_table(efp.se);

    if (v->efgn != NULL && (EFGN_IS_FE_IF(v->efgn))) {
      /* an EFG/FE node at a vertex of an interfacial hex volume  */
      WEFG_if_data_t if_data;
      find_if_hv(v, &if_data.h, &if_data.xi, &if_data.eta, &if_data.theta);
      if (!SFTAB_IS_EMPTY(v->sftab)) {
        USE_SFTAB_SEARCH_FUNC(se, v->sftab); 
      }
      if (!SFL_shape_func_at(sf, v->loc.x, v->loc.y, v->loc.z,
                               &efp, &if_data)) {
        SFL_ERROR(sf, v->loc.x, v->loc.y, v->loc.z, E_K_FATAL);
      }
    } else {
      /* Ordinary vertex */
      if (!SFL_shape_func_at(sf, v->loc.x, v->loc.y, v->loc.z,
                               &efp, NULL)) {
        SFL_ERROR(sf, v->loc.x, v->loc.y, v->loc.z, E_K_FATAL);
      }
    }
    efgsf_create_sftab (se, &v->sftab, sf, &v->loc, FALSE);
    USE_DEFAULT_SEARCH_FUNC(se);
    UPDATE_EFG_SF_STAT(se, SFTAB_NUM_OF_ENTRIES(v->sftab));
  }
  TOUCH_SFTAB(v->sftab, se);
}

/*
   Read integration volume vertices.
   Read all integration volume vertices from the input stream.
   See the User's guide for details.
 */
void
read_vertices(WEFG_superelem_t *se, W_domain_t *d)
{
  const char *input_line;
  WEFG_vertex_t *v;
  W_index_t id;
  W_loc_t loc;
  
  while ((input_line = W_get_input_line(d)) != NULL) {
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD("END")) {
      if (TOKENS_HAVE_KEYWORD("VERT"))
        return;
      else
        REPORT_FATAL_ERROR(se->domain->errh, "VERT does not end by END VERT");
    } else {
      id      = TOKEN_INT(1);
      loc.x   = TOKEN_DBL(2);
      loc.y   = TOKEN_DBL(3);
      loc.z   = TOKEN_DBL(4);
      v = efgv_new_vertex(id, &loc);
      efg_attach_vertex(se, v);
    }
  }
  REPORT_FATAL_ERROR(se->domain->errh, "VERT does not end by END VERT");
}


static int
check_location(W_loc_t *vl,  W_loc_t *nl);

/* compute the average length of the diagonal
    compute the average length of the diagonal of an integration volume.
 */
static double
hex_diagonal (WEFG_hex_volume_t *h /* integration volume */
              )
{
  double d = 0;
  G3D_vect_t v;

  VECMAC_3_VEC (v, =, h->vertex[0]->loc, -, h->vertex[6]->loc);
  d += VECMAC_NORM (v);
  VECMAC_3_VEC (v, =, h->vertex[1]->loc, -, h->vertex[7]->loc);
  d += VECMAC_NORM (v);
  VECMAC_3_VEC (v, =, h->vertex[2]->loc, -, h->vertex[4]->loc);
  d += VECMAC_NORM (v);
  VECMAC_3_VEC (v, =, h->vertex[3]->loc, -, h->vertex[5]->loc);
  d += VECMAC_NORM (v);
  return d / 4;
}



/*
   Read connectivities.
   Read connectivities from the input stream.
   See the User's guide for details.
 */
void
read_connectivities (WEFG_superelem_t *se, /* EFG superelement */
                     W_domain_t *d /* domain */
                     )
{
  const char *input_line;
  int nfe, nefg;
  WEFG_vertex_t *v;
  W_node_t *fe_node;
  int driven_conn = FALSE;
  
  while ((input_line = W_get_input_line(d)) != NULL) {
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD("END")) {
      if (TOKENS_HAVE_KEYWORD("CONNECT"))
        return;
      else
        REPORT_FATAL_ERROR(se->domain->errh, "CONNECT does not end by END CONNECT");
    } else {
      W_loc_t feloc;
      if (TOKENS_HAVE_KEYWORD("KIND") && TOKENS_HAVE_KEYWORD("DRIVEN")) {
        driven_conn = TRUE;
      } else if (TOKENS_HAVE_KEYWORD("ALL")) {
        dict_iterator di = dict_get_iterator();
        WEFG_vertex_t *v;
        while (v = dict_next_obj(se->verts, di)) {
          fe_node = WD_get_node_by_idx(d, v->id);
          if (fe_node == NULL)
            REPORT_FATAL_ERROR(se->domain->errh, "Bad fe node in connectivity");
          WN_get_ref_coords(fe_node, &feloc.x, &feloc.y, &feloc.z);
          if (!check_location(&v->loc, &feloc)) {
            sprintf(W_buf(), "Locations of vertex %d and fe node %d do not agree",
                    (int)v->id, (int)fe_node->id);
            REPORT_FATAL_ERROR(se->domain->errh, W_buf());
          }
          if (driven_conn) dict_add_obj_handle_name (se->driven_fen_conn, fe_node, v);
          else             dict_add_obj_handle_name (se->coupling_fen_conn, fe_node, v);
        }
        dict_del_iterator(di);
      } else {
        nefg = TOKEN_INT(1);
        nfe  = TOKEN_INT(2);
        v = efg_get_vertex (se, nefg);
        if (v == NULL) REPORT_FATAL_ERROR(se->domain->errh, "Bad vertex in connectivity");
        fe_node = WD_get_node_by_idx(d, nfe);
        if (fe_node == NULL)
          REPORT_FATAL_ERROR(se->domain->errh, "Bad fe node in connectivity");
        WN_get_ref_coords(fe_node, &feloc.x, &feloc.y, &feloc.z);
        if (!check_location(&v->loc, &feloc)) {
          sprintf(W_buf(), "Locations of vertex %d and fe node %d do not agree",
                  (int)v->id, (int)nfe);
          REPORT_FATAL_ERROR(se->domain->errh, W_buf());
        }
        if (driven_conn) dict_add_obj_handle_name (se->driven_fen_conn, fe_node, v);
        else             dict_add_obj_handle_name (se->coupling_fen_conn, fe_node, v);
      }
    }
  }
  REPORT_FATAL_ERROR(se->domain->errh, "CONNECT does not end by END CONNECT");
}
    


W_node_t *
efgv_get_coupling_fen_conn (WEFG_superelem_t *se, WEFG_vertex_t *v /* integration volume vertex */)
{
  W_node_t *n;
  n = (W_node_t *) dict_find_obj_with_handle_name (se->coupling_fen_conn, v);
  return n;
}

W_node_t *
efgv_get_driven_fen_conn (WEFG_superelem_t *se, WEFG_vertex_t *v /* integration volume vertex */)
{
  W_node_t *n;
  n = (W_node_t *) dict_find_obj_with_handle_name (se->driven_fen_conn, v);
  return n;
}



static int
check_location(W_loc_t *vl,  W_loc_t *nl)
{
  return (   vl->x == nl->x
          && vl->y == nl->y
          && vl->z == nl->z);
}
    
/* attach integration volume.
   attach integration volume incident on the vertex.
 */
void
attach_hex_vol(WEFG_vertex_t *v, /* integration volume vertex */
               WEFG_hex_volume_t *h /* incident integration volume */
               )
{
  if (v->incident_hex_volumes == NULL) v->incident_hex_volumes = make_list();
  add_to_tail(v->incident_hex_volumes, h);
}

/* iterate integration volumes.
    iterate integration volumes incident on a vertex.
 */
WEFG_hex_volume_t * /* next volume or NULL if there no more */
next_incident_hex_vol(WEFG_vertex_t *v /* integration volume vertex */,
                      WEFG_hex_volume_t *h /* pass NULL to get the first */
                      )
{
  if (v->incident_hex_volumes == NULL)
    return NULL;
  else
    return (WEFG_hex_volume_t *)get_list_next(v->incident_hex_volumes, h);
}


/*
   Create a vertex.
   Create an integration volume vertex.
 */
WEFG_vertex_t * /* vertex */
efgv_new_vertex(W_index_t id, /* unique integer as read from the input */
                W_loc_t *loc /* location */
                )
{
  static W_index_t the_id=0;
  WEFG_vertex_t *v;

  v = w_new(WEFG_vertex_t);
  CHECK_MALLOC(v);
  if (id == 0)
    v->id      = --the_id;
  else
    v->id      = id;
  v->loc                  = *loc;
  v->efgn                 = NULL;
  v->incident_hex_volumes = NULL;
  INIT_SFTAB(v->sftab);
  v->matstate             = NULL;

  return v;
}

/*
   Update stress state.
   Update stress state at an integration volume vertex.
 */
void
efgv_update_stress(WEFG_superelem_t *se, /* EFG superelement */
                   WEFG_vertex_t *v /* integration volume vertex */)
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3];
  W_mat_3d_stress_type_enum out_type;

  if (v->matstate == NULL) v->matstate = WM3D_get_mat_state(se->mat, &v->loc);
  if (SFTAB_IS_STALE(v->sftab, se)) eval_sftab(se, v);

  ip_ctx_data.se       = se;
  ip_ctx_data.sftab    = v->sftab;
  ip_ctx.time          = WD_time(se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;
  efgss_at_new_ip();
  if (!WM3D_update_stress(se->mat, v->matstate, W_MAT_STATE_UPDATE_FINAL,
                          &ip_ctx, MAT_3D_1ST_PK, P,
                          &out_type))
    REPORT_FATAL_ERROR(se->domain->errh, "Failed to compute stress");
}


/*
  Retrieve named material state variable.
 */
double
efgv_get_named_var (WEFG_superelem_t *se, /* EFG superelement */
                    WEFG_vertex_t *v /* integration volume vertex */,
                    char *var)
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double val = 0;

  if (v->matstate == NULL) v->matstate = WM3D_get_mat_state(se->mat, &v->loc);
  if (SFTAB_IS_STALE(v->sftab, se)) eval_sftab(se, v);

  ip_ctx_data.se       = se;
  ip_ctx_data.sftab    = v->sftab;
  ip_ctx.time          = WD_time(se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;
  efgss_at_new_ip();
  if (WM3D_get_mat_state_named_var (se->mat, v->matstate, &ip_ctx, var, &val)) {
    return val;
  } else return 0;
}



/*
   Retrieve stress.
   GEt stress stored in the material state at an integration volume vertex.
 */
void
efgv_get_stress(WEFG_superelem_t *se, /* EFG superelement */
                WEFG_vertex_t *v /* integration volume vertex */,
                W_output_stress_type_enum stress_type, /* stress type */
                double stress[3][3] /* output: stress */
                )
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  W_mat_3d_stress_type_enum out_type;

  if (v->matstate == NULL) v->matstate = WM3D_get_mat_state(se->mat, &v->loc);
  if (SFTAB_IS_STALE(v->sftab, se)) eval_sftab(se, v);

  ip_ctx_data.se       = se;
  ip_ctx_data.sftab    = v->sftab;
  ip_ctx.time          = WD_time (se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;
  efgss_at_new_ip();
  if      (stress_type == W_OUTPUT_STRESS_CAUCHY) {
    if (!WM3D_update_stress(se->mat, v->matstate, W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_CAUCHY, stress, &out_type))
      REPORT_FATAL_ERROR(se->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_CAUCHY)
      REPORT_FATAL_ERROR(se->domain->errh, "Stress type not available");
  } else if (stress_type == W_OUTPUT_STRESS_2ND_PK) {
    if (!WM3D_update_stress(se->mat, v->matstate, W_MAT_STATE_UPDATE_TEMP,
                         &ip_ctx, MAT_3D_2ND_PK, stress, &out_type))
      REPORT_FATAL_ERROR(se->domain->errh, "Failed to compute stress");
    if (out_type != MAT_3D_2ND_PK)
      REPORT_FATAL_ERROR(se->domain->errh, "Stress type not available");
  }
}

/*
   Retrieve equivalent plastic strain.
   GEt equivalent plastic strain
   stored in the material state at an integration volume vertex.
 */
void
efgv_get_equiv_pl_def(WEFG_superelem_t *se, /* EFG superelement */
                      WEFG_vertex_t *v /* integration volume vertex */,
                      double *equiv_pl_def /* output: equivalent plastic strain */
                      )
{
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;

  if (v->matstate == NULL) v->matstate = WM3D_get_mat_state(se->mat, &v->loc);
  if (SFTAB_IS_STALE(v->sftab, se)) eval_sftab(se, v);

  ip_ctx_data.se       = se;
  ip_ctx_data.sftab    = v->sftab;
  ip_ctx.time          = WD_time (se->domain);
  ip_ctx.time_step     = WD_time_step (se->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = efgss_get_kinem_var;
  efgss_at_new_ip();
  if (!WM3D_get_mat_state_var(se->mat, v->matstate,
                              &ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF, equiv_pl_def))
    REPORT_FATAL_ERROR(se->domain->errh, "Failed to retrieve equiv plastic def");
}


/* evaluate current location.
 evaluate current location of integration volume vertex */
void
vertex_current_loc(WEFG_superelem_t *se, /* EFG superelement */
                   WEFG_vertex_t *v /* integration volume vertex */,
                   W_loc_t *loc /* output: location */
                   )
{
  double x, y, z;

  if (v->matstate == NULL) v->matstate = WM3D_get_mat_state(se->mat, &v->loc);
  if (SFTAB_IS_STALE(v->sftab, se)) eval_sftab(se, v);

  efgsf_eval_curr_loc(&v->sftab, &x, &y, &z);
  loc->x = x; loc->y = y; loc->z = z;
}



/* evaluate current location.
 evaluate current location of integration volume vertex */
void
vertex_current_disp(WEFG_superelem_t *se, /* EFG superelement */
                   WEFG_vertex_t *v /* integration volume vertex */,
                   W_loc_t *loc /* output: location */
                   )
{
  double x, y, z;

  if (v->matstate == NULL) v->matstate = WM3D_get_mat_state(se->mat, &v->loc);
  if (SFTAB_IS_STALE(v->sftab, se)) eval_sftab(se, v);

  efgsf_eval_curr_loc(&v->sftab, &x, &y, &z);
  loc->x = x - v->loc.x; loc->y = y - v->loc.y; loc->z = z - v->loc.z;
}


/* evaluate current velocity.
 evaluate current velocity of integration volume vertex */
void
vertex_current_vel(WEFG_superelem_t *se,
                   WEFG_vertex_t *v /* integration volume vertex */,
                   W_loc_t *vel /* output: velocity */
                   )
{
  double x, y, z;

  if (v->matstate == NULL) v->matstate = WM3D_get_mat_state(se->mat, &v->loc);
  if (SFTAB_IS_STALE(v->sftab, se)) eval_sftab(se, v);

  efgsf_eval_curr_vel(&v->sftab, &x, &y, &z);
  vel->x = x; vel->y = y; vel->z = z;
}

/* evaluate current acceleration.
 evaluate current acceleration of integration volume vertex */
void
vertex_current_acc(WEFG_superelem_t *se, /* EFG superelement */
                   WEFG_vertex_t *v /* integration volume vertex */,
                   W_loc_t *acc /* output: acceleration */
                   )
{
  double x, y, z;

  if (v->matstate == NULL) v->matstate = WM3D_get_mat_state(se->mat, &v->loc);
  if (SFTAB_IS_STALE(v->sftab, se)) eval_sftab(se, v);

  efgsf_eval_curr_vel(&v->sftab, &x, &y, &z);
  acc->x = x; acc->y = y; acc->z = z;
}


/*
   access reference location.
   access reference location of integration volume vertex.
 */
void
vertex_ref_loc(WEFG_superelem_t *se, /* EFG superelement */
               WEFG_vertex_t *v /* integration volume vertex */,
               W_loc_t *loc /* output: reference locaiton */
               )
{
  loc->x = v->loc.x; loc->y = v->loc.y; loc->z = v->loc.z;
}




typedef struct efgv_mat_pt_history_client_data_t {
  W_index_t                        pntidx;
  W_loc_t                          loc;
  WEFG_vertex_t                   *v;
}              efgv_mat_pt_history_client_data_t;



void *
efgv_new_sampler(WEFG_superelem_t *se, W_index_t pntidx, W_loc_t *loc)
{
  efgv_mat_pt_history_client_data_t *cd;

  cd = w_new(efgv_mat_pt_history_client_data_t);
  CHECK_MALLOC(cd);

  cd->pntidx = pntidx;
  cd->loc    = *loc;
  cd->v      = efgv_new_vertex(0, &cd->loc);
  efg_attach_vertex(se, cd->v);
  { W_loc_t pt; vertex_current_loc(se, cd->v, &pt); }

  return (void *)cd;
}


static void
store (W_elem_t *e,
       W_history_writer_t *hw,
		efgv_mat_pt_history_client_data_t *cd,
		W_float_t time)
{
  WEFG_superelem_t *se = (WEFG_superelem_t *)e->data;
  double S[3][3];
  double equiv_pl_def, pressure;
  W_loc_t x, v, a;
#define STORE(SUFFIX, data)                             \
  {                                                     \
    char name[64];                                      \
    sprintf (name, "p%d%s", cd->pntidx, SUFFIX);        \
    WHR_store (hw, name, time, data);                   \
  }

  vertex_current_loc (se, cd->v, &x);
  vertex_current_vel (se, cd->v, &v);
  vertex_current_acc (se, cd->v, &a);

  STORE ("TDx", x.x - cd->loc.x);
  STORE ("TDy", x.y - cd->loc.y);
  STORE ("TDz", x.z - cd->loc.z);
  STORE ("TVx", v.x);
  STORE ("TVy", v.y);
  STORE ("TVz", v.z);
  STORE ("TAx", a.x);
  STORE ("TAy", a.y);
  STORE ("TAz", a.z);
  if (se->compute_stress_at_output_pts) {
    efgv_get_stress(se, cd->v, WOPT_OPT (hw->domain->options, output_stress_type), S);
    pressure  = (S[0][0] + S[1][1] + S[2][2]) / 3;
    if        (WOPT_OPT (hw->domain->options, output_stress_type)
               == W_OUTPUT_STRESS_CAUCHY) {
      STORE ("sxx", S[0][0]);
      STORE ("sxy", S[0][1]);
      STORE ("sxz", S[0][2]);
      STORE ("syy", S[1][1]);
      STORE ("syz", S[1][2]);
      STORE ("szz", S[2][2]);
      STORE ("press", pressure);
    } else if (WOPT_OPT (hw->domain->options, output_stress_type)
               == W_OUTPUT_STRESS_2ND_PK) {
      STORE ("Sxx", S[0][0]);
      STORE ("Sxy", S[0][1]);
      STORE ("Sxz", S[0][2]);
      STORE ("Syy", S[1][1]);
      STORE ("Syz", S[1][2]);
      STORE ("Szz", S[2][2]);
      STORE ("Press", pressure);
    } else {
      REPORT_ERROR(hw->domain->errh, "Unknown stress type for output");
    }
  }
  if (se->compute_equiv_pl_def_at_output_pts) {
    efgv_get_equiv_pl_def (se, cd->v, &equiv_pl_def);
    STORE ("EqPlS", equiv_pl_def);
  }
}

void 
efgv_sampler_store(W_elem_t *e,
		   W_solution_history_t *h,
		   void *data,
		   double time)
{
  efgv_mat_pt_history_client_data_t *cd =
    (efgv_mat_pt_history_client_data_t *)data;

  store (e, h->writer, cd, time);
}




void
efgv_write_tecplot(WEFG_vertex_t *v /* integration volume vertex */,
                   WEFG_superelem_t *se, FILE *stream,
                   W_index_t *count)
{
  G3D_loc_t cl;
  double equiv_pl_def, vonmises = 0;
  double S[3][3];
  
  fprintf(stream, "%g %g %g ", v->loc.x, v->loc.y, v->loc.z);
  
  if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_DISPL_MASK) {
    vertex_current_loc(se, v, &cl);
    fprintf(stream, "%g ", cl.x - v->loc.x);
    fprintf(stream, "%g ", cl.y - v->loc.y);
    fprintf(stream, "%g ", cl.z - v->loc.z);
  }
  if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_STRESS_MASK ||
      WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK) {
    double means;
    efgv_get_stress(se, v, WOPT_OPT (se->domain->options, output_stress_type), S);
    fprintf(stream, "%g ", S[0][0]);
    fprintf(stream, "%g ", S[0][1]);
    fprintf(stream, "%g ", S[0][2]);
    fprintf(stream, "%g ", S[1][1]);
    fprintf(stream, "%g ", S[1][2]);
    fprintf(stream, "%g ", S[2][2]);
    means = (S[0][0] + S[1][1] + S[2][2]) / 3;
    S[0][0] -= means; S[1][1] -= means; S[2][2] -= means;
    vonmises = sqrt(1.5
                    * (  S[0][0]*S[0][0] + S[0][1]*S[0][1] + S[0][2]*S[0][2]
                       + S[1][0]*S[1][0] + S[1][1]*S[1][1] + S[1][2]*S[1][2]
                       + S[2][0]*S[2][0] + S[2][1]*S[2][1] + S[2][2]*S[2][2]
                       )
                    );  }
  if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_EQPLS_MASK) {
    efgv_get_equiv_pl_def(se, v, &equiv_pl_def);
    fprintf(stream, "%g ", equiv_pl_def);
  }
  if (WOPT_OPT (se->domain->options, postfile_output_spec_deprec) & W_POSTFILE_VON_MISES_MASK) {
    fprintf(stream, "%g ", vonmises);
  }
  fprintf(stream, "\n");
  (*count)++;
}

void
efgv_Newmark_update (WEFG_vertex_t *v, WEFG_superelem_t *se)
{
  if (v->efgn && v->efgn->fen) {
    W_loc_t U, V, A;
    W_bool_t apply[3] = {1,1,1};
    vertex_current_disp (se, v, &U);
    vertex_current_vel (se, v, &V);
    vertex_current_acc (se, v, &A);
    WN_set_displ (v->efgn->fen, &U, apply);
    WN_set_velocity (v->efgn->fen, &V, apply);
    WN_set_acceleration (v->efgn->fen, &A, apply);
  }
}

void
efgv_assemble_driven_fen_force (WEFG_vertex_t *v, WEFG_superelem_t *se)
{
  W_vect_t pf, f;
  int i, ne = SFTAB_NUM_OF_ENTRIES(v->sftab);
  WN_get_force (v->efgn->fen, &f.x, &f.y, &f.z);
  if (VECMAC_NORM(f) > 0) {
    FOR(i, ne) {
      double phi = SFTAB_PHI(v->sftab, i);
      if (phi != 0) {
        VECMAC_2_VEC (pf, =, v->efgn->fen->mass * phi * f);
        efgn_assemble_force (v->efgn, se, &pf);
      }
    }
  }
}

static double
avg_distance_of_connected_vertices (WEFG_vertex_t *v)
{
  double dist = 0;
  W_vect_t d;
  int i, nd = 0;
  WEFG_hex_volume_t *h = 0;
  while ((h = next_incident_hex_vol(v, h))) {
    FOR(i, 8) {
      VECMAC_3_VEC(d, =, h->vertex[i]->loc, -, v->loc);
      dist += VECMAC_NORM(d); nd++;
    }
  }
  return dist / nd;
}

void
efgv_generate_efgn (WEFG_vertex_t *v, WEFG_superelem_t *se)
{
  G3D_box_t range;
  W_node_t *fen;
  WEFG_node_t *en;
  double avg_dist = avg_distance_of_connected_vertices(v);
  
  G3D_INIT_BOX(range, v->loc); 
  G3D_INFLATE_BOX(range, se->merging_distance); 
  
  en = efg_node_in_range (se, &range);
  if (en == NULL)
    en = efg_add_node (se, &v->loc);
  v->efgn = en;
  
  efg_set_max_support_size (se, avg_dist);
  
  if ((fen = efgv_get_driven_fen_conn (se, v)) != NULL) {
    /* driven FE node connection: the FE node is driven by the motion of the vertex */
    efgn_change_to_driving_node (v->efgn, fen);
  }
  efg_attach_node (se, v->efgn);
}


#endif /* EFG_WASP  */
