
#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Werrh.h"
#include "Wutil.h"
#include "Wmpi.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <timeP.h>

/* RCSHEADER "$Id: errh.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct W_error_handler_t {
  char        *err_log_name;
  FILE        *error_stream;
  char         buf[2048];
}              W_error_handler_t;

static void
open_error_stream (W_error_handler_t *eh)
{
  if (eh->err_log_name) {
    eh->error_stream = fopen (eh->err_log_name, "a");
  }
  if (!eh->error_stream) eh->error_stream = stderr;
}

static void
update_error_log (W_error_handler_t *eh, char *msg)
{
  if (eh) {
    if (eh->error_stream == NULL) open_error_stream (eh);
    fprintf (eh->error_stream, "%s", msg);
    fflush (eh->error_stream);
  } else {
    fprintf (stderr, "%s", msg);
    fflush (stderr);
  }
}

/*
  Create a new error handler.
  Create a new error handler.  If error log name not supplied, or if error log
  file could not be opened, error messages are printed to stderr.
 */
struct W_error_handler_t * /* non-NULL if successful */
WEH_new_handler (char *err_log_name /* error log name */
                 )
{
  W_error_handler_t *eh;
  eh = NEWP_OBJ (W_error_handler_t);
  if (!eh) return NULL;
  if (!err_log_name)  eh->err_log_name = "?";
  else {
    DUPLICATE_STRING (eh->err_log_name, err_log_name);
  }
  eh->error_stream = NULL;
  return eh;
}

/*
   Start an error message.
   Print the initial part of an error message.
   This should be followed by WEH_error_msg() and WEH_error_end().
 */
int /* ignored */
WEH_error_start (W_error_handler_t *eh, /* error handler */
                 char *file, /* file name where error occurred */
                 int line, /* line in the file /file/ */
                 int kind /* severity of the error; default is E_K_FATAL */
                 )
{
  char buf[1024];
  char time[64];
  W_get_date_and_time (time);
  sprintf (buf, "\n>>> %s\n", time);
  update_error_log (eh, buf);
  sprintf(buf, "File %s, line %d: ", file, line);
  update_error_log (eh, buf);
  switch (kind) {
  case E_K_WARNING:
    update_error_log (eh, "Warning\n");
    break;
  case E_K_ERROR:
    update_error_log (eh, "Error\n");
    break;
  case E_K_FATAL:
  default:
    update_error_log (eh, "Fatal error\n");
    break;
  }
  return TRUE;
}


/*
   Print an error message.
   Print the actual error message.  This should be preceded
   by WEH_error_start() and followed by WEH_error_end().
 */
int /* ignored */
WEH_error_msg (W_error_handler_t *eh, /* error handler */
             char *msg /* message */
             )
{
  update_error_log (eh, msg);
  return TRUE;
}


/*
   End an error message.
   This routine concludes the error notification sequence.
   It should be preceded by WEH_error_start() and WEH_error_msg().
 */
int
WEH_error_end (W_error_handler_t *eh, /* error handler */
               int kind /* severity of the error (or warning) */
               )
{
  update_error_log (eh, "\n<<<\n");

  switch (kind) {
  case E_K_FATAL:
    fprintf (stderr, "Fatal error; see %s\n", (eh ? eh->err_log_name : "stderr"));
    exit (EXIT_FAILURE);
    break;
  case E_K_ERROR:
    fprintf (stderr, "Error; see %s\n", (eh ? eh->err_log_name : "stderr"));
    exit (EXIT_FAILURE);
    break;
  case E_K_WARNING:
    fprintf (stderr, "Warning; see %s\n", (eh ? eh->err_log_name : "stderr"));
    break;
  }
  return TRUE;
}



/*
   Notify of an error.
   Produce an error message.  This function simply groups the
   functionality of W_error_start (), W_error_msg (), and W_error_end().
 */
int 
WEH_error (W_error_handler_t *eh,
           char *file, int line, int kind, char *explanation)
{
  WEH_error_start (eh, file, line, kind);
  WEH_error_msg (eh, explanation);
  WEH_error_end (eh, kind);
  return 1;
}



/*
  Format an error message.
  Format an error message for error that occurred while reading an input file.
 */
const char * /* string with an error message */
WEH_read_failure (W_error_handler_t *eh, /* error handler */
                  int curr_line_num, /* current line number of input file */
                  char *input_line /* contents of input line which caused failure */
                  )
{
  if (eh) {
    sprintf (eh->buf, "Bad input on line #%d:\n\"%s\"\n", curr_line_num, input_line);
    return (const char *) eh->buf;
  } else {
    static char buf[2048];
    sprintf (buf, "Bad input on line #%d:\n\"%s\"\n", curr_line_num, input_line);
    return (const char *) buf;
  }
}


/*
  Notify of an error.
  Notify of an error which occurred while parsing an input line.
 */
int
WEH_parse_error (W_error_handler_t *eh, /* error handler */
               char *file, /* file where error occurred */
               int line, /* line in the file */
               int tok_pos, /* token position */
               tokens_token_type expect /* expected token value */,
                 int curr_line_num /* current line number */
               )
{
  fprintf(stderr,
          "parse error in %s (%d)\nExpecting %s at pos %d on line %d\n",
          file, line, expect == TOKENS_NUMBER? "number": "string",
          tok_pos, curr_line_num);
  exit(1);
  return 1;
}

/*
  Format an error message.
  Format an error message for unrecognized input.
 */
const char *
WEH_unrecognized_input_error (W_error_handler_t *eh, /* error handler */
                              char *file, /* in which source file */
                              int line, /* at which line in source file */
                              int curr_line_num, /* line number of input file */
                              char *input_line /* contents of offending line */
                              )
{
  if (eh) {
    sprintf (eh->buf, 
             "Unrecognized input in %s (%d), input line %d:\n%s",
             file, line, curr_line_num, input_line);
    return eh->buf;
  } else {
    static char buf[2048];
    sprintf (buf, "Unrecognized input in %s (%d), input line %d\n%s",
             file, line, curr_line_num, input_line);
    return buf;
  }
}


/*
   Callback for the G3D library.
 */
void 
W_g3d_error_callback (void)
{
  WEH_error(NULL,
            G3D_error_in_file(),
            G3D_error_on_line(),
            E_K_ERROR,
            G3D_error_message());
}


#include <math.h>

/*
  math exception handler.
 */
int /* ignore */
matherr(struct exception *x /* exception */
        )
{
  char buf[1024];
  
  sprintf(buf, "Math exception in %s() (arg %g, retval %g)\n",
          x->name, x->arg1, x->retval); 
  REPORT_FATAL_ERROR(NULL, buf);
  return TRUE;
}
