#include "Wconfig.h"

#include "Wtypes.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Wcdi.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Wcontface.h"
#include "Wfhex.h"

/* RCSHEADER "$Id: fhex.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

#define GET_FROM_NODE(node, func, tovect) \
   func (node, &tovect.x, &tovect.y, &tovect.z) 


static int msidcs[2][2][2] =  {
   { {0,7} ,  {2,3} } ,
   { {4,5} ,  {6,1} } 
} ;

typedef struct hex_shape_func_t  {
  double n[8];
  double n_x[8];
  double n_y[8];
  double n_z[8];
  double J[3][3];
  double invJ[3][3];
  double detJ;
}               hex_shape_func_t;

typedef struct gp_data_t      {
  W_mat_3d_state_p   matst;
  double            *n_x;
  double            *n_y;
  double            *n_z;
  double             wdetJ;
}                   gp_data_t;

typedef struct priv_elem_t  {
  W_node_t          *node[8];
  W_material_t      *material;
  gp_data_t         *gpdata;
  W_bool_t           use_2pt_rule;
}               priv_elem_t;




static int
elem_elcode (W_elem_t *el);
static char *
type_name (W_elem_t *el);
static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form);
static void
assemble_mass (W_elem_t *el);
static W_float_t 
suggested_time_step (W_elem_t *el);


#ifdef X_XFACE
static void
modify_elem_grep (W_elem_t *el, double displ_scale, W_bool_t immed_draw) 
{
  priv_elem_t *h = (priv_elem_t *) el->data;
  int i;
  W_vect_t refc, u;
  WCRec pnts[8];

  for (i = 0; i < 8; i++) {
    GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc);
    GET_FROM_NODE (h->node[i], WN_get_displ, u);
    pnts[i].x = refc.x + displ_scale * u.x;
    pnts[i].y = refc.y + displ_scale * u.y;
    pnts[i].z = refc.z + displ_scale * u.z;
  } 

  if (el->grep == NULL) {
    el->grep = CreateHexahedron (pnts);
    EASValsSetColor (WM_get_color (h->material));
    EGWithMaskChangeAttributes (COLOR_MASK, el->grep);
    EMNoDrawAddGraphicsToModel (ESIModel (), el->grep);
  }  else  {
    EGModifyGraphicsGeometry (el->grep, pnts);
  } 
} 
#endif

static W_float_t
total_mass (W_elem_t *el);
static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB, int type,
                     char *mode, void *optdata);



static void
scramble_diagonal (priv_elem_t *h);
static void
fill_ip_2x2x2_pkg (double xi, double eta, double theta,
                  W_mat_3d_state_p matstate,
                  W_mat_3d_ip_ctx_t *ip_ctx,
                  W_mat_3d_t *mat,
                  W_mat_3d_stress_type_enum stress_type
                  );
static void
eval_at_extrapolated_pt (double xi, double eta, double theta,
                         double s[3][3], double *equiv_pl_strain,
                         double *von_mises);
static void
clear_2x2x2_pkgs (void);
static void
accum_unused_2x2x2_pkgs (double xi, double eta, double theta);
static void
finish_unused_2x2x2_pkgs (void);


static int 
num_of_faces (W_elem_t *el) 
{
  return 6;
} 

static int 
get_face_vertex_nodes (W_elem_t *el, int nface, int ndim, W_node_t *n[]) 
{
  priv_elem_t *h = (priv_elem_t *) el->data;
  int i;
  if (ndim >= 4) {
    FOR (i, 4) {
      n[i] = h->node[g3dhex_vertex_on_face[nface][i]];
    } 
  }
  return 4;
} 

static int
get_contact_faces (W_elem_t *el, int nface, WCF_contact_face_t *cf[], int maxcf) 
{
  priv_elem_t *h = (priv_elem_t *) el->data;
  int i;
  
  if (maxcf < 1) {
    FOR (i, maxcf) { cf[i] = NULL; } 
    return 1;
  } 
  
  cf[0] = WCF_new_contact_face (h->node[g3dhex_vertex_on_face[nface][0]],
                                h->node[g3dhex_vertex_on_face[nface][1]],
                                h->node[g3dhex_vertex_on_face[nface][2]],
                                h->node[g3dhex_vertex_on_face[nface][3]]);
  
  return 1;
} 

static W_elem_methods_t hex_methods =  {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  write_postproc_file,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  num_of_faces,
  get_face_vertex_nodes,
  get_contact_faces
} ;
  
static W_gauss_integration_rule_t *gr = NULL;
static W_gauss_integration_point_t *gp = NULL;




typedef struct ip_ctx_data_t  {
  priv_elem_t                 *h;
  double                *n_x, *n_y, *n_z;
}               ip_ctx_data_t;

static void
at_new_ip (void);
static W_bool_t 
get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3]);


#define CHECK_2PT_CASE(H, I, J, K) \
 ((!use_2pt_rule) \
     || ((I == 0 && J == 0 && K == 0) \
         || (I == 1 && J == 1 && K == 1)) ) 
#define USE_QUADR_PT(H, I, J, K) if (CHECK_2PT_CASE (H, I, J, K)) 
#define ALL_GP_LOCS(i,j,k) gp[i].Loc, gp[j].Loc, gp[k].Loc
#define GP_WGT(gp,I,J,K) (gp[I].Wgt * gp[J].Wgt * gp[K].Wgt) 
#define ALL_REF_COORDS(H) \
 H->node[0]->ref_coords.x, H->node[0]->ref_coords.y, H->node[0]->ref_coords.z,\
 H->node[1]->ref_coords.x, H->node[1]->ref_coords.y, H->node[1]->ref_coords.z,\
 H->node[2]->ref_coords.x, H->node[2]->ref_coords.y, H->node[2]->ref_coords.z,\
 H->node[3]->ref_coords.x, H->node[3]->ref_coords.y, H->node[3]->ref_coords.z,\
 H->node[4]->ref_coords.x, H->node[4]->ref_coords.y, H->node[4]->ref_coords.z,\
 H->node[5]->ref_coords.x, H->node[5]->ref_coords.y, H->node[5]->ref_coords.z,\
 H->node[6]->ref_coords.x, H->node[6]->ref_coords.y, H->node[6]->ref_coords.z,\
 H->node[7]->ref_coords.x, H->node[7]->ref_coords.y, H->node[7]->ref_coords.z


static hex_shape_func_t hex_shape_func;
  
#define N_LOC(h, v, DIR, n) \
   v.DIR = (  n[0] * h->node[0]->ref_coords.DIR      \
            + n[1] * h->node[1]->ref_coords.DIR      \
            + n[2] * h->node[2]->ref_coords.DIR      \
            + n[3] * h->node[3]->ref_coords.DIR      \
            + n[4] * h->node[4]->ref_coords.DIR      \
            + n[5] * h->node[5]->ref_coords.DIR      \
            + n[6] * h->node[6]->ref_coords.DIR      \
            + n[7] * h->node[7]->ref_coords.DIR) 

#define LOC(h, gp) \
   {                                             \
    N_LOC (h, gp, x, hex_shape_func.n);          \
    N_LOC (h, gp, y, hex_shape_func.n);          \
    N_LOC (h, gp, z, hex_shape_func.n);          \
  } 

/*
  Element reader.
  Reads element data for the pcont3 element type
  from stream as long as there is no
  END_OF_BLOCK tag in the input line. Creates elements.
 */
W_bool_t                        /* TRUE if successful */
fhex_reader (W_domain_t *d /* domain */) 
{
  const char *input_line;
  long elem_num;
  priv_elem_t *h;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_index_t nodeid, matid;
  int i, j, k;
  W_material_t *mat;
  W_mat_3d_t *m;
  W_bool_t use_2pt_rule = FALSE;

  if (gr == NULL) {
    gr = WG_get_integration_rule (2);
  } 

  while ((input_line = W_get_input_line (d)) ) {
    if (W_is_end_of_block (input_line)) {
      return TRUE;
    }  else  {
      TOKENS_PARSE (input_line);
      if (TOKENS_HAVE_KEYWORD ("USE_2PT_RULE")) {
        use_2pt_rule = TRUE;
      }  else if (TOKENS_HAVE_KEYWORD ("DO_NOT_USE_2PT_RULE")) {
        use_2pt_rule = FALSE;
      }  else  { /* ordinary line with element data */
        h = w_new (priv_elem_t);
        CHECK_MALLOC (h);
        elem_num             = TOKEN_INT (1);
        FOR (i, 8) {
          nodeid = TOKEN_INT (2+i);
          h->node[i] = WD_get_node_by_idx (d, nodeid);
          if (h->node[i] == NULL) {
            sprintf (W_buf (), "No such node: %ld", nodeid);
            REPORT_ERROR (d->errh, W_buf ());
          } 
        } 
        matid = TOKEN_INT (10);
        mat = WD_get_material_by_idx (d, matid);
        if (mat == NULL) {
          sprintf (W_buf (), "No such material: %ld", matid);
          REPORT_ERROR (d->errh, W_buf ());
        } 
        h->material     = mat;
        m = (W_mat_3d_t *) h->material->data;
        h->use_2pt_rule = use_2pt_rule;
        h->gpdata       = w_new_array (h->use_2pt_rule ? 2 : 8, gp_data_t);
        CHECK_MALLOC (h->gpdata);
        gp = gr->ip;
        if (h->use_2pt_rule) {
          W_loc_t loc;
          scramble_diagonal (h);
          i = j = k = 0;
          G3D_hex_N (hex_shape_func.n, ALL_GP_LOCS (i,j,k));
          LOC (h, loc);
          h->gpdata[msidcs[i][j][k]].matst = WM3D_get_mat_state (m, &loc);
          i = j = k = 1;
          G3D_hex_N (hex_shape_func.n, ALL_GP_LOCS (i,j,k));
          LOC (h, loc);
          h->gpdata[msidcs[i][j][k]].matst = WM3D_get_mat_state (m, &loc);
        }  else  {
          FOR (i, 2) {
            FOR (j, 2) {
              FOR (k, 2) {
                W_loc_t loc;
                G3D_hex_N (hex_shape_func.n, ALL_GP_LOCS (i,j,k));
                LOC (h, loc);
                h->gpdata[msidcs[i][j][k]].matst = WM3D_get_mat_state (m, &loc);
              } 
            } 
          } 
        } 
#ifdef PVM_NEED_GHOST_ELEMENTS
        is_ghost = FALSE;
        if (TOKENS_HAVE_KEYWORD ("GHOST")) 
          is_ghost = TRUE;
#endif      
        WD_add_element (d, elem_num,
                       WE_new_element (d, &hex_methods, h
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                      , is_ghost
#                                   endif
                                      ));
      }  /* ordinary data line */
    }  /* closing else */
  } 
  return FALSE;
} 



static int
elem_elcode (W_elem_t *el) 
{
  return FHEX_ELCODE;
} 


static char *
type_name (W_elem_t *el) 
{
  static char buf[] = "fhex";
  return buf;
} 








/* Returns nodal forces by which the  element acts on its nodal points. */

static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form) 
{
  priv_elem_t *h = (priv_elem_t *) el->data;
  W_mat_3d_t *m = (W_mat_3d_t *) h->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  double P[3][3], scale, ded;
  W_mat_3d_stress_type_enum out_type;
  W_vect_t f;
  W_bool_t use_2pt_rule = h->use_2pt_rule;
  int i, j, k, n;
  int gpix;
#define EVAL_SF(i,j,k,gpix)                                                                     \
   {                                                                                            \
     if (!h->gpdata[gpix].n_x) {                                                                \
       int _counter_i;                                                                          \
       G3D_hex_interp_opt (hex_shape_func.n,                                                    \
                           hex_shape_func.n_x,                                                  \
                           hex_shape_func.n_y,                                                  \
                           hex_shape_func.n_z,                                                  \
                           &hex_shape_func.detJ,                                                \
                           hex_shape_func.J,                                                    \
                           hex_shape_func.invJ,                                                 \
                           ALL_GP_LOCS (i,j,k),                                                 \
                           ALL_REF_COORDS (h));                                                 \
       h->gpdata[gpix].n_x = w_new_array (8, double); CHECK_MALLOC (h->gpdata[gpix].n_x);       \
       h->gpdata[gpix].n_y = w_new_array (8, double); CHECK_MALLOC (h->gpdata[gpix].n_y);       \
       h->gpdata[gpix].n_z = w_new_array (8, double); CHECK_MALLOC (h->gpdata[gpix].n_z);       \
       FOR (_counter_i, 8) {                                                                    \
         h->gpdata[gpix].n_x[_counter_i] = hex_shape_func.n_x[_counter_i];                      \
         h->gpdata[gpix].n_y[_counter_i] = hex_shape_func.n_y[_counter_i];                      \
         h->gpdata[gpix].n_z[_counter_i] = hex_shape_func.n_z[_counter_i];                      \
       }                                                                                        \
       h->gpdata[gpix].wdetJ = (use_2pt_rule ? 4 : 1) * GP_WGT (gp,i,j,k) * hex_shape_func.detJ;\
     }                                                                                          \
  } 
#define COMP_F(N)                                                                               \
   {                                                                                            \
    double dphix = ip_ctx_data.n_x[N], dphiy = ip_ctx_data.n_y[N], dphiz = ip_ctx_data.n_z[N];  \
    f.x = - (P[0][0]*dphix + P[0][1]*dphiy + P[0][2]*dphiz);                                    \
    f.y = - (P[1][0]*dphix + P[1][1]*dphiy + P[1][2]*dphiz);                                    \
    f.z = - (P[2][0]*dphix + P[2][1]*dphiy + P[2][2]*dphiz);                                    \
  } 
  
  ip_ctx_data.h        = h;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ip_ctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;

  el->deformation_energy = 0;
  gp = gr->ip;
  FOR (i, 2) {
    FOR (j, 2) {
      FOR (k, 2) {
        USE_QUADR_PT (h, i, j, k) {
          gpix = msidcs[i][j][k];
          EVAL_SF (i, j, k, gpix);
          ip_ctx_data.n_x = h->gpdata[gpix].n_x;
          ip_ctx_data.n_y = h->gpdata[gpix].n_y;
          ip_ctx_data.n_z = h->gpdata[gpix].n_z;
          at_new_ip ();
          if (!WM3D_update_stress (m, h->gpdata[gpix].matst, update_form,
                                  &ip_ctx, MAT_3D_1ST_PK, P, &out_type)) 
            REPORT_ERROR (el->domain->errh, "Failed to compute stress");
          if (!WM3D_get_mat_state_var (m, h->gpdata[gpix].matst,
                                       &ip_ctx, MAT_3D_DEFORM_ENERGY_DENSITY, &ded))
            REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
          scale = h->gpdata[gpix].wdetJ;
          el->deformation_energy += ded * scale;
          MAT3DUTIL_SCALE_3X3 (P, scale);
          FOR (n, 8) {
            COMP_F (n);
            WN_assemble_force (h->node[n], &f);
          } 
        }  /* USE_QUADR_PT */
      }  /* FOR (k, 2) */
    }  /* FOR (j, 2) */
  }  /* FOR (i, 2) */
} 

static void
assemble_mass (W_elem_t *el) 
{
  W_float_t m;
  priv_elem_t *h = (priv_elem_t *) el->data;
  int i;
  
  m = total_mass (el) / 8;

  for (i = 0; i < 8; i++) 
    WN_assemble_mass (h->node[i], m);
} 


static W_float_t 
suggested_time_step (W_elem_t *el) 
{
  priv_elem_t *h = (priv_elem_t *) el->data;
  W_mat_3d_t *m = (W_mat_3d_t *) h->material->data;
  W_coords_3d_t d;
  W_float_t E, nu;
  W_float_t rho, speed_of_sound;
  double shortest_side = FLT_MAX;

  if (!WM3D_get_mat_const (m, MAT_3D_YOUNG_MODULUS, &E)) 
    REPORT_FATAL_ERROR (el->domain->errh, "No Young's modulus");
  if (!WM3D_get_mat_const (m, MAT_3D_POISSON, &nu)) 
    REPORT_FATAL_ERROR (el->domain->errh, "No Poisson constant");
  
  rho      = WM3D_get_volume_mass (m);
  speed_of_sound = sqrt (E / (1 - 2*nu) / rho);
#define SUBTR(res, a, b) \
    {res.x = (a) .x - (b) .x; res.y = (a) .y - (b) .y; res.z = (a) .z - (b) .z;} 
#define COND_MIN(a, b) \
    { double the_B = b; if (the_B != 0) a = min (a, the_B); } 
  SUBTR (d, h->node[1]->ref_coords, h->node[0]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[2]->ref_coords, h->node[1]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[3]->ref_coords, h->node[2]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[0]->ref_coords, h->node[3]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[5]->ref_coords, h->node[4]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[6]->ref_coords, h->node[5]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[7]->ref_coords, h->node[6]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[4]->ref_coords, h->node[7]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[0]->ref_coords, h->node[4]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[1]->ref_coords, h->node[5]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[2]->ref_coords, h->node[6]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  SUBTR (d, h->node[3]->ref_coords, h->node[7]->ref_coords);
  COND_MIN (shortest_side, W_vector_3d_length (&d));
  
  return shortest_side / speed_of_sound;                 
} 


static W_float_t
total_mass (W_elem_t *el) 
{
  priv_elem_t *h = (priv_elem_t *) el->data;
  W_mat_3d_t *m = (W_mat_3d_t *) h->material->data;
  W_float_t volume_mass, J[3][3];
  int i, j, k;
  double volume = 0;

  volume_mass = WM3D_get_volume_mass (m);
  
  gp = gr->ip;
  for (i = 0; i < 2; i++) {
    for (j = 0; j < 2; j++) {
      for (k = 0; k < 2; k++) {
        G3D_hex_J (J, ALL_GP_LOCS (i,j,k), ALL_REF_COORDS (h));
        volume += GP_WGT (gp,i,j,k) * G3D_hex_detJ (J);
      } 
    } 
  } 

  return (volume_mass * volume);
} 








static void
write_postproc_file (W_elem_t *el, ckit_iobuf_t IOB,
                     int type, char *mode, void *optdata) 
{
  priv_elem_t *h = (priv_elem_t *) el->data;
  W_bool_t use_2pt_rule = h->use_2pt_rule;
  int i, j, k;
  W_vect_t refc[8], u[8];
  W_mat_3d_t *m = (W_mat_3d_t *) h->material->data;
  W_mat_3d_ip_ctx_t ip_ctx;
  ip_ctx_data_t ip_ctx_data;
  int pt, ival, gpix;
  int reqlen;
  G3D_loc_t *pts[8];
  static int buflen        = 0;
  static double *values[8] =  {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL} ;
  W_mat_3d_stress_type_enum outtype;
  W_output_stress_type_enum output_stress_type
    = WOPT_OPT (el->domain->options, output_stress_type);
  int postfile_output_spec
    = WOPT_OPT (el->domain->options, postfile_output_spec_deprec);

  outtype = (output_stress_type == W_OUTPUT_STRESS_CAUCHY
             ? MAT_3D_CAUCHY
             : (output_stress_type == W_OUTPUT_STRESS_2ND_PK
                ? MAT_3D_2ND_PK
                : MAT_3D_CAUCHY)
             );
             

  if (type == HEXVUE_POSTPROC_FILE) {
    double S[3][3] =  { {0,0,0} , {0,0,0} , {0,0,0} } ;
    double equiv_pl_def = 0, von_mises = 0;

    reqlen = 0;
    if (postfile_output_spec & W_POSTFILE_DISPL_MASK) 
      reqlen += 3;
    if (postfile_output_spec & W_POSTFILE_STRESS_MASK) 
      reqlen += 6;
    if (postfile_output_spec & W_POSTFILE_EQPLS_MASK) 
      reqlen++;
    if (postfile_output_spec & W_POSTFILE_VON_MISES_MASK) 
      reqlen++;
    
    if (buflen < reqlen) {
      FOR (i, 8) {
        values[i] = NEWP_REALLOC_OBJ_ARRAY (values[i], reqlen, double);
      } 
      buflen = reqlen;
    } 
  
    ip_ctx_data.h        = h;
    ip_ctx.time          = WD_time (el->domain);
    ip_ctx.time_step     = WD_time_step (el->domain);
    ip_ctx.ctx_data      = &ip_ctx_data;
    ip_ctx.get_kinem_var = get_kinem_var;

    clear_2x2x2_pkgs ();
    gp = gr->ip;
    FOR (i, 2) {
      FOR (j, 2) {
        FOR (k, 2) {
          USE_QUADR_PT (h, i, j, k) {
            gpix = msidcs[i][j][k];
            EVAL_SF (i, j, k, gpix);
            at_new_ip ();
            ip_ctx_data.n_x = h->gpdata[gpix].n_x;
            ip_ctx_data.n_y = h->gpdata[gpix].n_y;
            ip_ctx_data.n_z = h->gpdata[gpix].n_z;
            fill_ip_2x2x2_pkg (ALL_GP_LOCS (i,j,k),
                               h->gpdata[gpix].matst,   
                               &ip_ctx, m, outtype);
          }  else  {
            accum_unused_2x2x2_pkgs (ALL_GP_LOCS (i,j,k));
          } 
        } 
      } 
    } 
    if (use_2pt_rule) 
      finish_unused_2x2x2_pkgs ();

    FOR (i, 8) {
      GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc[i]);
      GET_FROM_NODE (h->node[i], WN_get_displ, u[i]);
    } 

    pt = 0;
    FOR (i, 2) {
      FOR (j, 2) {
        FOR (k, 2) {
          pts[pt]   = &refc[pt];
          ival = 0;
          if (postfile_output_spec & W_POSTFILE_DISPL_MASK) {
            values[pt][ival] = u[pt].x; ival++;
            values[pt][ival] = u[pt].y; ival++;
            values[pt][ival] = u[pt].z; ival++;
          } 

          eval_at_extrapolated_pt (g3dhex_vertex_param_coord[pt][0],
                                   g3dhex_vertex_param_coord[pt][1],
                                   g3dhex_vertex_param_coord[pt][2],
                                   S, &equiv_pl_def, &von_mises);
          
          if (postfile_output_spec & W_POSTFILE_STRESS_MASK) {
            values[pt][ival] = S[0][0]; ival++;
            values[pt][ival] = S[0][1]; ival++;
            values[pt][ival] = S[0][2]; ival++;
            values[pt][ival] = S[1][1]; ival++;
            values[pt][ival] = S[1][2]; ival++;
            values[pt][ival] = S[2][2]; ival++;
          } 
          if (postfile_output_spec & W_POSTFILE_EQPLS_MASK) {
            values[pt][ival] = equiv_pl_def; ival++;
          } 
          if (postfile_output_spec & W_POSTFILE_VON_MISES_MASK) {
            values[pt][ival] = von_mises; ival++;
          } 
          if (ival > reqlen) REPORT_FATAL_ERROR (el->domain->errh, "wrong");
          pt++;
        } 
      } 
    } 
    W_write_hexvue_hex (IOB, pts, reqlen, values, mode);
  }  else if (type == TECPLOT_I_POSTPROC_FILE) {
    /* RAW */
  }  /* type */
#undef  EVAL_SF
} 



#define VALID(which) which##_is_valid

#define DCL_TENSOR(T) \
   static double T[3][3];                       \
   static W_bool_t VALID (T) = FALSE
 
DCL_TENSOR (F);
DCL_TENSOR (prev_F);
DCL_TENSOR (inv_F);
DCL_TENSOR (rel_F);
DCL_TENSOR (midstep_F);
DCL_TENSOR (C);
DCL_TENSOR (E);
DCL_TENSOR (D);
DCL_TENSOR (midstep_D);
DCL_TENSOR (midstep_W);
DCL_TENSOR (F_dot);
DCL_TENSOR (L);
DCL_TENSOR (R);
DCL_TENSOR (U);
DCL_TENSOR (rel_R);
DCL_TENSOR (rel_U);
DCL_TENSOR (ln_rel_U);


static void
at_new_ip (void) 
{
#define INVALIDATE(which) VALID (which) = FALSE
  
  INVALIDATE (F);
  INVALIDATE (inv_F);
  INVALIDATE (prev_F);
  INVALIDATE (rel_F);
  INVALIDATE (midstep_F);
  INVALIDATE (C);
  INVALIDATE (E);
  INVALIDATE (D);
  INVALIDATE (midstep_D);
  INVALIDATE (midstep_W);
  INVALIDATE (F_dot);
  INVALIDATE (L);
  INVALIDATE (R);
  INVALIDATE (U);
  INVALIDATE (rel_R);
  INVALIDATE (rel_U);
  INVALIDATE (ln_rel_U);
} 


static void
compute_def_grad (ip_ctx_data_t *ipctx);
static void
compute_cauchy_green (ip_ctx_data_t *ipctx);
static void
compute_green_lagrange (ip_ctx_data_t *ipctx);
static void
compute_rel_def_grad (ip_ctx_data_t *ipctx);
static void
compute_midstep_def_grad (ip_ctx_data_t *ipctx);
static void
compute_rate_of_def (ip_ctx_data_t *ipctx);
static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx, double time_step);
static void
compute_def_grad_dot (ip_ctx_data_t *ipctx);
static void
compute_vel_grad (ip_ctx_data_t *ipctx);
static void
compute_inv_def_grad (ip_ctx_data_t *ipctx);
static void
compute_polar_decomp (ip_ctx_data_t *ipctx);
static void
compute_rel_polar_decomp (ip_ctx_data_t *ipctx);
static void
compute_ln_rel_U (ip_ctx_data_t *ipctx);



static W_bool_t 
get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
              W_mat_3d_kinem_var_type_enum  type,
              double kinvar[3][3]) 
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *) ctx->ctx_data;

  switch (type) {
  case MAT_3D_DEF_GRAD:
    if (!VALID (F)) compute_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, F);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    if (!VALID (F)) compute_def_grad_dot (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, F_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    if (!VALID (rel_F)) compute_rel_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    if (!VALID (prev_F)) compute_rel_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    if (!VALID (midstep_F)) compute_midstep_def_grad (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    if (!VALID (C)) compute_cauchy_green (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, C);
    break;
  case MAT_3D_RATE_OF_DEF:
    if (!VALID (D)) compute_rate_of_def (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, D);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    if (!VALID (midstep_D)) 
      compute_midstep_rate_of_def (ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    if (!VALID (midstep_W)) 
      compute_midstep_rate_of_def (ipctx, ctx->time_step);
    MAT3DUTIL_COPY_3X3 (kinvar, midstep_W);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    if (!VALID (E)) compute_green_lagrange (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, E);
    break;
  case MAT_3D_POLAR_R:
    if (!VALID (R)) compute_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, R);
    break;
  case MAT_3D_POLAR_U:
    if (!VALID (U)) compute_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, U);
    break;
  case MAT_3D_POLAR_REL_R:
    if (!VALID (rel_R)) compute_rel_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    if (!VALID (rel_U)) compute_rel_polar_decomp (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    if (!VALID (ln_rel_U)) compute_ln_rel_U (ipctx);
    MAT3DUTIL_COPY_3X3 (kinvar, ln_rel_U);
    break;
  default:
    REPORT_FATAL_ERROR (NULL, "Unknown kinematic tensor type");
    break;
  } 

  return TRUE;
} 

static void
compute_prev_def_grad (ip_ctx_data_t *ipctx) 
{
  priv_elem_t *h = ipctx->h;
  double *n_x = ipctx->n_x;
  double *n_y = ipctx->n_y;
  double *n_z = ipctx->n_z;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  double x, y, z;
  register double dphix, dphiy, dphiz;
  W_node_t *n;

  FOR (k, 8) {
    n = h->node[k];
    WN_get_prev_coords (n, &x, &y, &z);
    dphix  = n_x[k];
    dphiy  = n_y[k];
    dphiz  = n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  } 

  prev_F[0][0] = f00;
  prev_F[0][1] = f01;
  prev_F[0][2] = f02;
  
  prev_F[1][0] = f10;
  prev_F[1][1] = f11;
  prev_F[1][2] = f12;
  
  prev_F[2][0] = f20;
  prev_F[2][1] = f21;
  prev_F[2][2] = f22;

  VALID (prev_F) = TRUE;
} 

static void
compute_rel_def_grad (ip_ctx_data_t *ipctx) 
{
  priv_elem_t *h = ipctx->h;
  double inv_prev_F[3][3];
  double det_prev_F;

  if (!VALID (F)) compute_def_grad (ipctx);
  if (!VALID (prev_F)) compute_prev_def_grad (ipctx);
   
  det_prev_F = MAT3DUTIL_DET_3X3 (prev_F);
  if (det_prev_F <= 0) {
    sprintf (W_buf (), "Failed to invert def grad");
    REPORT_FATAL_ERROR (NULL, W_buf ());
  } 

  MAT3DUTIL_INVERT_3X3 (det_prev_F, prev_F, inv_prev_F);

  MAT3DUTIL_MULT_AB_3X3 (rel_F, F, inv_prev_F);

  VALID (rel_F) = TRUE;
} 

static void
compute_midstep_def_grad (ip_ctx_data_t *ipctx)
{
  priv_elem_t *h = ipctx->h;
  
  if (!VALID (F)) compute_def_grad (ipctx);
  if (!VALID (prev_F)) compute_prev_def_grad (ipctx);
   
  MAT3DUTIL_COMB_AB_3X3 (midstep_F, 0.5, F, 0.5, prev_F);

  VALID(midstep_F) = TRUE;
}

static void
compute_vel_grad (ip_ctx_data_t *ipctx) 
{
  if (!VALID (F)) compute_def_grad (ipctx);
  if (!VALID (F_dot)) compute_def_grad_dot (ipctx);
  if (!VALID (inv_F)) compute_inv_def_grad (ipctx);

  MAT3DUTIL_MULT_AB_3X3 (L, F_dot, inv_F);
  
  VALID (L) = TRUE;
} 


static void
compute_rate_of_def (ip_ctx_data_t *ipctx) 
{
  int i, j;
  
  if (!VALID (L)) compute_vel_grad (ipctx);

  FOR (i, 3) 
    FOR (j, 3) 
      D[i][j] = 0.5 * (L[i][j] + L[j][i]);

  VALID (D) = TRUE;
} 




static void
compute_def_grad (ip_ctx_data_t *ipctx) 
{
  priv_elem_t *h = ipctx->h;
  double *n_x = ipctx->n_x;
  double *n_y = ipctx->n_y;
  double *n_z = ipctx->n_z;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR (k, 8) {
    n = h->node[k];
    WN_get_coords (n, &x, &y, &z);
    dphix  = n_x[k];
    dphiy  = n_y[k];
    dphiz  = n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  } 

  F[0][0] = f00;
  F[0][1] = f01;
  F[0][2] = f02;
  
  F[1][0] = f10;
  F[1][1] = f11;
  F[1][2] = f12;
  
  F[2][0] = f20;
  F[2][1] = f21;
  F[2][2] = f22;

  VALID (F) = TRUE;
} 



static void
compute_def_grad_dot (ip_ctx_data_t *ipctx) 
{
  priv_elem_t *h = ipctx->h;
  double *n_x = ipctx->n_x;
  double *n_y = ipctx->n_y;
  double *n_z = ipctx->n_z;
  int k;
  double f00 = 0, f01 = 0, f02 = 0;
  double f10 = 0, f11 = 0, f12 = 0;
  double f20 = 0, f21 = 0, f22 = 0;
  W_node_t *n;
  double x, y, z;
  register double dphix, dphiy, dphiz;

  /* Deformation gradient */
  FOR (k, 8) {
    n = h->node[k];
    x = n->velocity_t.x;
    y = n->velocity_t.y;
    z = n->velocity_t.z;
    dphix  = n_x[k];
    dphiy  = n_y[k];
    dphiz  = n_z[k];
    f00 += dphix * x;
    f01 += dphiy * x;
    f02 += dphiz * x;
    f10 += dphix * y;
    f11 += dphiy * y;
    f12 += dphiz * y;
    f20 += dphix * z;
    f21 += dphiy * z;
    f22 += dphiz * z;
  } 

  F_dot[0][0] = f00;
  F_dot[0][1] = f01;
  F_dot[0][2] = f02;
  
  F_dot[1][0] = f10;
  F_dot[1][1] = f11;
  F_dot[1][2] = f12;
  
  F_dot[2][0] = f20;
  F_dot[2][1] = f21;
  F_dot[2][2] = f22;

  VALID (F_dot) = TRUE;
} 




static void
compute_green_lagrange (ip_ctx_data_t *ipctx) 
{
  if (!VALID (C)) compute_cauchy_green (ipctx);

  /* Green-Lagrange strain tensor */
  E[0][0] = 0.5 * (C[0][0] - 1);
  E[0][1] = 0.5 * C[0][1];
  E[0][2] = 0.5 * C[0][2];
  E[1][0] = 0.5 * C[1][0];
  E[1][1] = 0.5 * (C[1][1] - 1);
  E[1][2] = 0.5 * C[1][2];
  E[2][0] = 0.5 * C[2][0];
  E[2][1] = 0.5 * C[2][1];
  E[2][2] = 0.5 * (C[2][2] - 1);
  
  VALID (E) = TRUE;
} 



static void
compute_cauchy_green (ip_ctx_data_t *ipctx) 
{
  if (!VALID (F)) compute_def_grad (ipctx);
  
  /* Cauchy-Green deformation tensor */
  MAT3DUTIL_MULT_ATB_3X3 (C, F, F);

  VALID (C) = TRUE;
} 


static void
compute_inv_def_grad (ip_ctx_data_t *ipctx) 
{
  if (!VALID (F)) compute_def_grad (ipctx);

  if (!W_mat3dutil_invert_3x3 (F, inv_F)) {
    sprintf (W_buf (), "Failed to invert def grad");
    REPORT_FATAL_ERROR (NULL, W_buf ());
  } 

  VALID (inv_F) = TRUE;
} 


static void
compute_midstep_rate_of_def (ip_ctx_data_t *ipctx, double time_step) 
{
  int i, j;
  double m[3][3] =  {  {1, 0, 0} ,  {0, 1, 0} ,  {0, 0, 1}  } ;
  double M[3][3] =  {  {-1, 0, 0} ,  {0, -1, 0} ,  {0, 0, -1}  } ;
  double s[3][3], inv_s[3][3], q[3][3], r[3][3];
  double inv_dt = 1 / time_step;
  
  if (!VALID (rel_F)) compute_rel_def_grad (ipctx);

  MAT3DUTIL_ADD_AB_3X3 (s, m, rel_F);
  MAT3DUTIL_SCALE_3X3 (s, 0.5);
  if (!W_mat3dutil_invert_3x3 (s, inv_s)) {
    sprintf (W_buf (), "Failed to invert def grad");
    REPORT_FATAL_ERROR (NULL, W_buf ());
  } 
  MAT3DUTIL_ADD_AB_3X3 (q, rel_F, M);
  MAT3DUTIL_MULT_AB_3X3 (r, inv_s, q);

  FOR (i, 3) 
    FOR (j, 3) {
      midstep_D[i][j] = 0.5 * (r[i][j] + r[j][i]);
      midstep_W[i][j] = 0.5 * (r[i][j] - r[j][i]);
    } 
  MAT3DUTIL_SCALE_3X3 (midstep_D, inv_dt);
  MAT3DUTIL_SCALE_3X3 (midstep_W, inv_dt);
  
  VALID (midstep_D) = TRUE;
  VALID (midstep_W) = TRUE;
} 




static void
compute_polar_decomp (ip_ctx_data_t *ipctx) 
{
  if (!VALID (F)) compute_def_grad (ipctx);
  
  if (!W_polar_decomposition (F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID (R) = TRUE;
  VALID (U) = TRUE;
} 


static void
compute_rel_polar_decomp (ip_ctx_data_t *ipctx) 
{
  if (!VALID (rel_F)) compute_rel_def_grad (ipctx);
  
  if (!W_polar_decomposition (rel_F, rel_R, rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
  VALID (rel_R) = TRUE;
  VALID (rel_U) = TRUE;
} 



static void
compute_ln_rel_U (ip_ctx_data_t *ipctx) 
{
  double lneval[3];

  if (!VALID (rel_U)) compute_rel_polar_decomp (ipctx);
  
  W_ln_U (rel_U, ln_rel_U);

  VALID (ln_rel_U) = TRUE;
} 



static int
flip (void) 
{
  long rn;
  int two = 2;

  rn = (long) ((double) LONG_MAX * rand_0_to_1());

  return (( (rn / two) * 2) == rn);
} 


static int
one_of_four (void) 
{
  
  if (flip ()) 
    if (flip ()) return 1;
    else         return 2;
  else
    if (flip ()) return 3;
    else         return 4;
} 



static void
scramble_diagonal (priv_elem_t *h) 
{
  int which = one_of_four ();
#define FOUR_SHIFT(start) \
   {                                             \
    W_node_t *t = h->node[start];               \
    h->node[start  ] = h->node[start+1];        \
    h->node[start+1] = h->node[start+2];        \
    h->node[start+2] = h->node[start+3];        \
    h->node[start+3] = t;                       \
  } 
#define SHIFT_ONCE() { FOUR_SHIFT (0); FOUR_SHIFT (4); } 

  switch (which) {
  case 1:
    /* do nothing */
    break;
  case 2:
    /* rotate once */
    SHIFT_ONCE ();
    break;
  case 3:
    /* rotate twice */
    SHIFT_ONCE ();
    SHIFT_ONCE ();
    break;
  case 4:
    /* rotate thrice */
    SHIFT_ONCE ();
    SHIFT_ONCE ();
    SHIFT_ONCE ();
    break;
  default:
    break;
  } 
} 



 

typedef struct ipol_2x2x2_pkg_t  {
  double          stress[3][3];
  double          equiv_pl_strain;
  double          von_mises;
}               ipol_2x2x2_pkg_t;


static ipol_2x2x2_pkg_t ipol_2x2x2_pkgs[8];


static void
fill_ip_2x2x2_pkg (double xi, double eta, double theta,
                  W_mat_3d_state_p matstate,
                  W_mat_3d_ip_ctx_t *ip_ctx,
                  W_mat_3d_t *mat,
                  W_mat_3d_stress_type_enum stress_type
                  ) 
{
  double stress[3][3];
  double equiv_pl_strain, von_mises;
  W_mat_3d_stress_type_enum out_type;
  int entry = -1;
  
  if (xi < 0 && eta < 0 && theta < 0) {entry = 0;
  }  else if (xi > 0 && eta < 0 && theta < 0) {entry = 1;
  }  else if (xi > 0 && eta > 0 && theta < 0) {entry = 2;
  }  else if (xi < 0 && eta > 0 && theta < 0) {entry = 3;
  }  else if (xi < 0 && eta < 0 && theta > 0) {entry = 4;
  }  else if (xi > 0 && eta < 0 && theta > 0) {entry = 5;
  }  else if (xi > 0 && eta > 0 && theta > 0) {entry = 6;
  }  else if (xi < 0 && eta > 0 && theta > 0) {entry = 7;
  } 

  /* Get 1st Piola-Kirchhoff stress at point */
  if (!WM3D_update_stress (mat, matstate, W_MAT_STATE_UPDATE_TEMP,
                        ip_ctx, stress_type, stress,
                        &out_type)) {
    REPORT_ERROR (NULL, "Failed to compute stress");
    return;
  } 
  if (out_type != stress_type) 
    REPORT_WARNING (NULL, "output stress type mismatch");

  if (!WM3D_get_mat_state_var (mat, matstate,
                               ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF,
                               &equiv_pl_strain)) 
    REPORT_FATAL_ERROR (NULL, "Failed to retrieve equiv plastic def");

   {
    double means = (stress[0][0] + stress[1][1] + stress[2][2]) / 3;
    double s00 = stress[0][0] - means;
    double s11 = stress[1][1] - means;
    double s22 = stress[2][2] - means;
    
    von_mises
      = sqrt (1.5
             * (  s00*s00
                + stress[0][1]*stress[0][1]
                + stress[0][2]*stress[0][2]
                + stress[1][0]*stress[1][0]
                + s11*s11
                + stress[1][2]*stress[1][2]
                + stress[2][0]*stress[2][0]
                + stress[2][1]*stress[2][1]
                + s22*s22
                ) 
             );
  } 
  
  /* Store into the package */
  MAT3DUTIL_COPY_3X3 (ipol_2x2x2_pkgs[entry].stress, stress);
  ipol_2x2x2_pkgs[entry].equiv_pl_strain = equiv_pl_strain;
  ipol_2x2x2_pkgs[entry].von_mises       = von_mises;
} 


static void
eval_at_extrapolated_pt (double xi, double eta, double theta,
                         double stress[3][3], double *eps,
                         double *vm) 
{
#define SQRT_3 1.73205080756887729352
  double modxi    = SQRT_3 * xi;
  double modeta   = SQRT_3 * eta;
  double modtheta = SQRT_3 * theta;
  double Nmat[8];
  double equiv_pl_strain, von_mises;

  /* Interpolate */
  G3D_hex_N (Nmat, modxi, modeta, modtheta);

#define COMP_T1(T, I, J) 				\
   { T[I][J] = (						\
	       Nmat[0] * ipol_2x2x2_pkgs[0].T[I][J] +	\
	       Nmat[1] * ipol_2x2x2_pkgs[1].T[I][J] +	\
	       Nmat[2] * ipol_2x2x2_pkgs[2].T[I][J] +	\
	       Nmat[3] * ipol_2x2x2_pkgs[3].T[I][J] +	\
	       Nmat[4] * ipol_2x2x2_pkgs[4].T[I][J] +	\
	       Nmat[5] * ipol_2x2x2_pkgs[5].T[I][J] +	\
	       Nmat[6] * ipol_2x2x2_pkgs[6].T[I][J] +	\
	       Nmat[7] * ipol_2x2x2_pkgs[7].T[I][J]	\
	       );					\
  } 
#define COMP_T(T) 						\
   {								\
    COMP_T1 (T, 0, 0); COMP_T1 (T, 0, 1); COMP_T1 (T, 0, 2);	\
    COMP_T1 (T, 1, 0); COMP_T1 (T, 1, 1); COMP_T1 (T, 1, 2);	\
    COMP_T1 (T, 2, 0); COMP_T1 (T, 2, 1); COMP_T1 (T, 2, 2);	\
  } 
#define COMP_S(S) 				\
   { S = (					\
	 Nmat[0] * ipol_2x2x2_pkgs[0].S +	\
	 Nmat[1] * ipol_2x2x2_pkgs[1].S +	\
	 Nmat[2] * ipol_2x2x2_pkgs[2].S +	\
	 Nmat[3] * ipol_2x2x2_pkgs[3].S +	\
	 Nmat[4] * ipol_2x2x2_pkgs[4].S +	\
	 Nmat[5] * ipol_2x2x2_pkgs[5].S +	\
	 Nmat[6] * ipol_2x2x2_pkgs[6].S +	\
	 Nmat[7] * ipol_2x2x2_pkgs[7].S		\
	 );					\
  } 
#define COMP_V1(V, DIR) \
   { V.DIR = (                                   \
           Nmat[0] * ipol_2x2x2_pkgs[0].V.DIR + \
           Nmat[1] * ipol_2x2x2_pkgs[1].V.DIR + \
           Nmat[2] * ipol_2x2x2_pkgs[2].V.DIR + \
           Nmat[3] * ipol_2x2x2_pkgs[3].V.DIR + \
           Nmat[4] * ipol_2x2x2_pkgs[4].V.DIR + \
           Nmat[5] * ipol_2x2x2_pkgs[5].V.DIR + \
           Nmat[6] * ipol_2x2x2_pkgs[6].V.DIR + \
           Nmat[7] * ipol_2x2x2_pkgs[7].V.DIR   \
           );                                   \
  } 
#define COMP_V(V) { COMP_V1 (V, x); COMP_V1 (V, y); COMP_V1 (V, z); } 

  COMP_T (stress);
  COMP_S (equiv_pl_strain); *eps = equiv_pl_strain;
  COMP_S (von_mises); *vm = von_mises;

  return;
} 


static void
clear_2x2x2_pkgs (void) 
{
  int i;

  FOR (i, 8) {
    MAT3DUTIL_ZERO_3X3 (ipol_2x2x2_pkgs[i].stress);
    ipol_2x2x2_pkgs[i].equiv_pl_strain = 0;
    ipol_2x2x2_pkgs[i].von_mises       = 0;
  } 
} 


static void
accum_unused_2x2x2_pkgs (double xi, double eta, double theta) 
{
  int i;
  
  FOR (i, 8) {
    if (i != 0 && i != 6) {
      MAT3DUTIL_INCR_A_BY_B_3X3 (ipol_2x2x2_pkgs[i].stress,
                                 ipol_2x2x2_pkgs[0].stress);
      MAT3DUTIL_INCR_A_BY_B_3X3 (ipol_2x2x2_pkgs[i].stress,
                                 ipol_2x2x2_pkgs[6].stress);
      ipol_2x2x2_pkgs[i].von_mises += ipol_2x2x2_pkgs[0].von_mises;
      ipol_2x2x2_pkgs[i].von_mises += ipol_2x2x2_pkgs[1].von_mises;
      ipol_2x2x2_pkgs[i].equiv_pl_strain += ipol_2x2x2_pkgs[0].equiv_pl_strain;
      ipol_2x2x2_pkgs[i].equiv_pl_strain += ipol_2x2x2_pkgs[1].equiv_pl_strain;
    } 
  } 
} 



static void
finish_unused_2x2x2_pkgs (void) 
{
  int i;

  FOR (i, 8) {
    if (i != 0 && i != 6) {
      MAT3DUTIL_SCALE_3X3 (ipol_2x2x2_pkgs[i].stress, 0.5);
      ipol_2x2x2_pkgs[i].equiv_pl_strain *= 0.5;
      ipol_2x2x2_pkgs[i].von_mises *= 0.5;
    } 
  } 
} 
