/** 
 **   hardening.c
 **/

#include "Wconfig.h"
 
#include "Wtypes.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "Whistry.h"
#include "Wcdi.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Wnewmarki.h"
#include "Wcdi.h"
#include "Wcontface.h"
#include "Whex20.h"
#include "flestype.h"


/**
 **  Create mesh
 **/
 
static void 
create_mesh (hardening_solver_t *hsi)
{
  int mid;
  mat_t *u = (mat_t *)hsi->m->data;
  W_domain_t *d = u->d;

  int mymid = hsi->m->mat->id;
  W_elem_t *e = NULL;
  while(e = WD_get_next_element(d,e)){
    mid = WE_get_material_id(e);      
    if(mid == mymid) {
       switch(WE_elem_elcode(e))
       {
          case HEX20_ELCODE:
            dict_add_obj_handle_name (hsi->eld, make_hex20(e, hsi), e); 
	    break;
          default:
            REPORT_FATAL_ERROR (NULL, "Hardening solver could not create element"); 
	    break;
       }
    } /* if (mid == mymid) */
  } 
}


/**
 **  check if the node is in hsi->eqnd library already
 **  if not, add the node into hsi->eqnd
 **
 **  NOTE: int obj of hsi->eqnd doesn't necessarily correspond to node id
 **  ----
 **/

static W_index_t
find_eqn(hardening_solver_t *hsi,
         W_node_t           *n
        )
{
  int eqn;
  
  if ((eqn = (int) dict_find_obj_with_handle_name(hsi->eqnd, n)) == 0 ) {
     eqn =  dict_count(hsi->eqnd) + 1;
     dict_add_obj_handle_name (hsi->eqnd,  (dict_object)eqn, (dict_object)n ); /* eqn# of hsi->eqnd library starts from 1 */
  }
  return (eqn - 1); /* eqn# of priv_elem_t starts from 0 */
}


static hardening_solver_t *
make_hardening_solver(W_mat_3d_t *m)
{
  hardening_solver_t *hsi = w_new(hardening_solver_t);
  CHECK_MALLOC(hsi);
  hsi->m = m;
  hsi->fs = NULL;
  hsi->fs_dmeschach = NULL;

  hsi->eld = dict_create_dictionary ();
  CHECK_MALLOC(hsi->eld);
  hsi->eqnd =  dict_create_dictionary ();
  CHECK_MALLOC(hsi->eqnd);
  hsi->bdry_nodes =  dict_create_dictionary ();
  CHECK_MALLOC(hsi->bdry_nodes);
  hsi->matstd =  dict_create_dictionary ();
  CHECK_MALLOC(hsi->matstd);

  hsi->rhost = NULL;
  hsi->rhost_dt = NULL;
  hsi->curr_sys = 0;
  hsi->constrained = NULL;
  hsi->constraints = NULL;
  return hsi;
}

static int
tot_of_equations (void *iip)
{
  hardening_solver_t *hsi = (hardening_solver_t *)iip;
  return dict_count(hsi->eqnd);
}

static int
error_notify (fles_error_enum error_num,
              int line,
              char *file,
              char *detail,
              void *iip)
{
  char buf[1024];
  hardening_solver_t *hsi = (hardening_solver_t *)iip;
  W_mat_3d_t *m = hsi->m;
  mat_t *u = (mat_t *)m->data;
  W_domain_t *d = u->d;
  sprintf (buf, "%s (l.%d), %s", file, line, detail);
  REPORT_FATAL_ERROR (d->errh, buf);
  return 0;
}

typedef struct fef_ref_t {
  W_elem_t *e;
  int       nface;
  int       nn;
  W_node_t *n[9];
  int       ref;
}              fef_ref_t;

static LIST_OF (fef_ref_t *)
find_frl (W_node_t *n, dict_dictionary nfrd)
{
  return dict_find_obj_with_handle_name (nfrd, n);
}
 
 
static W_node_t *
min_id_fvn (W_node_t *fvn[9], int nfvn)
{
  int mini = 0, minid = INT_MAX, i;
 
  FOR (i, nfvn) {
    if (fvn[i]->id < minid) {
      minid = fvn[i]->id;
      mini  = i;
    }
  }
  return fvn[mini];
}
 
static fef_ref_t *
find_fr (LIST frl, int nfvn, W_node_t *fvn[9])
{
  int had = 0;
  int i, j;
  fef_ref_t *fr;
 
  loop_through_list (frl, fr, fef_ref_t *) {
    if (fr->nn == nfvn) {
      had = 0;
      FOR (i, nfvn) {
        FOR (j, nfvn) {
          if (fvn[i] == fr->n[j]) {
            had++;
            break;
          }
        }
      }
      if (had == nfvn) return fr;
    }
  }
  return NULL;

}

#define DICT_LOOP_START(dict, vartype, var)                                     \
  {                                                                             \
    vartype *var = NULL;                                                        \
    dict_iterator d_iter_ = dict_get_iterator ();                               \
    dict_reset_iterator (d_iter_);                                              \
    while ((var = (vartype *) dict_iter_objs (dict->d, d_iter_)) != NULL) {
#define DICT_LOOP_END                           \
    }                                           \
    dict_del_iterator (d_iter_);                \
  }

#define ELEM_LOOP_START(domain, var) DICT_LOOP_START(domain->elements, W_elem_t, var)
#define ELEM_LOOP_END                DICT_LOOP_END

/*
   Get list of contact faces.
 */
static LIST_OF (fef_ref_t *)
get_list_of_bdry_faces (W_domain_t *d)
{
  dict_dictionary nfrld;
  int i, j, nf, nfvn;
  LIST_OF (fef_ref_t *) frefl;
#define MAXNODES 9
  W_node_t *fvn[MAXNODES], *n;
  LIST_OF (fef_ref_t *) frl, tmpl;
  fef_ref_t *fr;

  /* accumulate FE faces */
  
  nfrld = dict_create_dictionary ();
  
  tmpl = make_list ();
  
  ELEM_LOOP_START (d, e) {
    nf = WE_num_of_faces (e);
    FOR (i, nf) {
      nfvn = WE_get_face_vertex_nodes (e, i, MAXNODES, fvn);
      if (nfvn > MAXNODES) {
        REPORT_FATAL_ERROR (d->errh,
                          "Can handle only boundary faces of less than 9 nodes");
      }
      n    = min_id_fvn (fvn, nfvn);
      frl  = find_frl (n, nfrld); /* frl -- the list of fr's whose minimum node id is the same as the id of n */
      if (!frl) {
        frl = make_list ();
        dict_add_obj_handle_name (nfrld, frl, n);
      }
      fr = find_fr (frl, nfvn, fvn); /* nfvn is the num of the vertex nodes of the face */
      if (!fr) {
        fr = w_new (fef_ref_t); CHECK_MALLOC (fr);
        fr->e     = e;
        fr->nface = i;
        fr->nn    = nfvn;
        FOR (j, fr->nn) { fr->n[j] = fvn[j]; }
        fr->ref   = 0;
        add_to_tail (frl, fr);
        add_to_tail (tmpl, fr);
      } 
      fr->ref++;
    }
  } ELEM_LOOP_END;

  /* create face list by inspecting the accumulated normals */
  
  frefl = make_list ();
  fr     = NULL;
  while ((fr = (fef_ref_t *) get_list_next (tmpl, fr))) {
    if (fr->ref == 1) {
      add_to_tail (frefl, fr);
    } 
  }

  /* cleanup */
  
  free_list (tmpl, NOT_NODES);
  {
    dict_iterator diter;
    diter = dict_get_iterator ();
    dict_reset_iterator (diter);
    while ((frl = (LIST) dict_iter_objs (nfrld, diter))) {
      free_list (frl, NOT_NODES);
    }
    dict_del_iterator (diter);
  }
  dict_destroy_dictionary (nfrld);

  return frefl;
}

 #define DISP_DICT_LOOP_START(dict, vartype, var)                                     \
   {                                                                             \
     vartype *var = NULL;                                                        \
     dict_iterator d_iter_ = dict_get_iterator ();                               \
     dict_reset_iterator (d_iter_);                                              \
     while ((var = (vartype *) dict_iter_objs (dict->d, d_iter_)) != NULL) {
 #define DISP_DICT_LOOP_END                           \
     }                                           \
     dict_del_iterator (d_iter_);                \
   }

/**
 **  set_fles() should be called first time the hsi is called
 **/

static void
set_fles(hardening_solver_t *hsi)
{
  int i, j; /*counters*/
  int num_nodes = dict_count(hsi->eqnd);  

#ifdef SOLVER_PETSC
  int method = PETSC_OPTS_NZ_PER_ROW;
  double nz_per_row = 0.01;
#endif
 
  mat_t *u = (mat_t *)(hsi->m->data); 
  W_node_t *n;
  int length, eqn, id;

  hsi->rhost = w_new_array(num_nodes, rhos_t);
  CHECK_MALLOC(hsi->rhost);
  hsi->rhost_dt = w_new_array(num_nodes, rhos_t);
  CHECK_MALLOC(hsi->rhost_dt);
  hsi->constrained = w_new_array(num_nodes, constrained_t);
  CHECK_MALLOC(hsi->constrained);

  FOR(i, num_nodes){
     FOR(j, N_SLIP_SYSTEMS){
      /* 2.8732e+7 to make g_0 100Mpa, as in classical model */
        hsi->rhost_dt[i][j] = 2.0e+7; 
        hsi->constrained[i][j] = FALSE; 
     }
  }

  
  /**
     apply initial condition and EBC
     check if the node is on the boundary, except the plane-strain bc,
     and set the constraints and homogeneous EBC's
  **/

  DISP_DICT_LOOP_START(u->d->disp_bcs, W_disp_bc_t, ps_bc) {
    if(ps_bc->applies_in_interval.on == 2e+7) {
       loop_through_list(ps_bc->applies_to, n, W_node_t *){
          if ((eqn = (int) dict_find_obj_with_handle_name(hsi->eqnd, (dict_object)n)) != 0) {
	    if ((id = (int) dict_find_obj_with_handle_name(hsi->bdry_nodes, (dict_object)n)) == 0) {
              id =  dict_count(hsi->bdry_nodes) + 1;
              dict_add_obj_handle_name (hsi->bdry_nodes, (dict_object)id , (dict_object)n);
              FOR(j, N_SLIP_SYSTEMS){
                 hsi->rhost_dt[eqn - 1][j] = 0.0; 
	         hsi->constrained[eqn - 1][j] = TRUE;
              }
            }
          }
       }
    }
  } DISP_DICT_LOOP_END;
  id = dict_count(hsi->bdry_nodes);

/*
  frl = get_list_of_bdry_faces (u->d);
  fr = NULL;
  while ((fr = (fef_ref_t *) get_list_next (frl, fr))) {
    FOR(i, 9) {
       if (fr->n[i] != NULL) {
          if ((eqn = dict_find_obj_with_handle_name(hsi->eqnd, fr->n[i])) != 0) {
	    if ((id = dict_find_obj_with_handle_name(hsi->bdry_nodes, fr->n[i])) == 0) {
              id =  dict_count(hsi->bdry_nodes) + 1;
             dict_add_obj_handle_name (hsi->bdry_nodes, id , fr->n[i]);
            FOR(j, N_SLIP_SYSTEMS){
                 hsi->rhost_dt[eqn - 1][j] = 0.0; 
	         hsi->constrained[eqn - 1][j] = TRUE;
               }
	    }
          } 
       } 
    } 
  }
  length = list_length(frl);
*/
 
  hsi->fs = fles_new_gep_solver (hsi, tot_of_equations, error_notify);
  /*  hsi->fs = fles_new_spetsc_solver (hsi, tot_of_equations, error_notify);*/
  /*  hsi->fs->set_option( hsi->fs, PETSC_OPTS_NZ_PER_ROW, (void *)&nz_per_row);*/

  hsi->fs_dmeschach = fles_new_dense_meschach_solver (hsi, tot_of_equations, error_notify);
}

static void
assemble_lhs_callback (hardening_solver_t *hsi,
                        fles_update_mode_enum update_mode, /* update mode;  */
                        int nr, /* number of rows and columns in square matrix to be assembled */
                        int *ri, /* array of equations numbers, which apply to both rows and columns
                                  of the matrix to be assembled */
                        double *values, /* square matrix to be assembled, passed
                                           as a linear array of nr*nr values */
                        W_bool_t *constrained /* array of nr elements; TRUE if degree-of-freedom
                                               constrained, FALSE otherwise */
                        )
{
  int j, k, l;
  W_bool_t need_to_rewrite = FALSE; /* do I need to modify the element matrix
                                       because of some constraints?
                                     */
 
  FOR (j, nr) { if (constrained[j]) { need_to_rewrite = TRUE; break; } }
 
  if (need_to_rewrite) {
    l = 0;
    FOR (j, nr) {
      FOR (k, nr) {
        if (j != k && (constrained[j] || constrained[k])) {
          values[l] = 0;
        }
        l++;
      }
    }
  }
 
  hsi->fs->assemble_mat (hsi->fs,
                         FLES_FALSE, update_mode, nr, ri, nr, ri, values);
  hsi->fs_dmeschach->assemble_mat (hsi->fs_dmeschach,
                         FLES_FALSE, update_mode, nr, ri, nr, ri, values);
}

static void
assemble_rhs_callback (hardening_solver_t *hsi, 
                       fles_update_mode_enum update_mode, /* update mode */
                       int nr,/* number of rows in vector to be assembled */
                       int *ri,/* array of equations numbers */
                       double *values /* vector to be assembled */
                       )
{
  hsi->fs->assemble_vect (hsi->fs,update_mode, nr, ri, values);
  hsi->fs_dmeschach->assemble_vect (hsi->fs_dmeschach, update_mode, nr, ri, values);
}


/**
 **   Save solution to hsi->rhost 
 **/
 
static void
save_rhost_dt(hardening_solver_t *hsi)
{
  int i, num;
  W_node_t *node = 0;
  double sol;
  fles_get_from_sol_func  get_from_sol = hsi->fs->get_from_sol;
  /*  fles_get_from_sol_func  get_from_sol = hsi->fs_dmeschach->get_from_sol;*/
  int eqns = hsi->fs->tot_of_equations(hsi->fs->client);
  mat_t *u = (mat_t *)hsi->m->data;

  while( node = (W_node_t *)W_dict_next_obj(u->d->nodes, node) )
  {
    i = (int) dict_find_obj_with_handle_name(hsi->eqnd, (dict_object)node);
    /* if the node is constrained, then do not save */
    if ((num = (int) dict_find_obj_with_handle_name(hsi->bdry_nodes, (dict_object)node)) == 0) 
    {
       sol = get_from_sol(hsi->fs, i-1);
       hsi->rhost_dt[i-1][hsi->curr_sys] = sol; /* get_from_sol(hsi->fs, i);*/
    }
    else
    {
       hsi->rhost_dt[i-1][hsi->curr_sys] = 0.0; /* set to preset values */
    }
  }
}

/** 
 **   Update yield stress at all the integration pts 
 **/

static void
update_yieldstress(hardening_solver_t *hsi) 
{
  mat3dcryspl_elem_t *e = 0;
  while(e = (mat3dcryspl_elem_t *)dict_next_obj(hsi->eld, e)){ 
     e->methods->update_ys(e);
  }
}

/**
 **  Post-processing
 **/

static void  
write_hexvue_file(hardening_solver_t *hsi)
{
  FILE *stream;
  ckit_iobuf_t IOB;
  static int num_files = 0;
  int n_output_values = N_SLIP_SYSTEMS;
  int i;

  mat3dcryspl_elem_t *e = 0;
  mat_t *u = (mat_t *)(hsi->m->data);

  char out[132];  
  sprintf(out, "%s_rhos_%d.xvu", WOPT_OPT (u->d->options, in_file_basename), num_files++);

  stream = ckit_io_open_stream (out, "w");
  if (stream == NULL) REPORT_ERROR (u->d->errh, (char *)ckit_io_error_explanation ());
  IOB = ckit_io_new_buffer (stream);
  if (IOB == NULL) REPORT_ERROR (u->d->errh, (char *)ckit_io_error_explanation ());

  W_get_date_and_time (out);
  ckit_io_put_line (IOB, "! HEXVUE file (C) 1995 Petr Krysl.");
  sprintf (W_buf (), "!     Generated by HARDENING SOLVER on %s.", out);
  ckit_io_put_line (IOB, W_buf ());
  sprintf (W_buf (), "%s; Time: %g (file %s, %s)",
          u->d->comment, WD_time(u->d), WOPT_OPT (u->d->options, in_file_name), out);
  ckit_io_put_line (IOB, W_buf ());
  /*  n_output_values = N_SLIP_SYSTEMS;*/
 
  sprintf (W_buf (), "%d ! # of values per vertex", n_output_values);
  ckit_io_put_line (IOB, W_buf ());
 
  for (i = 0; i < n_output_values; i++) {
    char buf[1024];
    sprintf (buf, "\"rho_%d\"", i+1);
    ckit_io_put_line (IOB, buf);
  }
 
  ckit_io_put_line (IOB, "! END OF HEADER");

  while(e = (mat3dcryspl_elem_t *)dict_next_obj(hsi->eld, e)){
     e->methods->write_hexvue_XTRA_file(e, IOB);
  }
  
  ckit_io_put_line (IOB, "END ! HEXVUE file.");
  ckit_io_delete_buffer (IOB);

  fclose (stream);
}

/**
typedef struct node_client_rec_t {
  W_index_t                   idx;
  W_solution_history_t        history;
}              node_client_rec_t;

typedef struct elem_client_rec_t {
  W_index_t                   idx;
  W_solution_history_t        history;
}              elem_client_rec_t;
**/


/**
 **  writing history files for rhos, xmgr files
 **/

/******************************************************************/
static void
RHO_update_history(hardening_solver_t *hsi)
{
  mat_t *u = (mat_t *)(hsi->m->data);
  W_history_writer_t *r = u->d->history_writer;

  FILE *stream;
  char filename[1024]; 
  static W_bool_t first_time = TRUE;

  int i,n;
  double time = WD_time(u->d);

  /*
//how to clean the old history file, because don't want to append to the old history files
//should add static "first_time" array, so that it is possible to add header in each file */

  W_node_t *node = NULL;
  while( node = (W_node_t *)W_dict_next_obj(u->d->nodes, node) )
  {
     sprintf (filename, "%s/p%drho.xmgr", r->history_dir, node->id);    
     if (first_time) 
     {
        if ( (stream = fopen(filename,"w")) == NULL )
           REPORT_ERROR(u->d->errh, "Could not open stream");
     }
     else
     {
        if ( (stream = fopen(filename,"a")) == NULL )
           REPORT_ERROR(u->d->errh, "Could not open stream");
     }

     fprintf(stream, "%g ", time);

     n = (int) dict_find_obj_with_handle_name(hsi->eqnd, (dict_object) node);
     FOR(i, N_SLIP_SYSTEMS) { fprintf(stream, " %g ", hsi->rhost_dt[n - 1][i]); }
     
     fprintf(stream, "\n");
     
     fclose(stream);

  }
  first_time = FALSE;
}
/****************************************************************/

/** 
 ** .m files for matlab postprocess
 **/

/****************************************************************
static void
RHO_update_history(hardening_solver_t *hsi)
{
  mat_t *u = (mat_t *)(hsi->m->data);
  W_domain_t *d = u->d;
  W_history_writer_t *r = d->history_writer;
  W_solution_history_t *h = NULL;
  history_client_data_t *cd = NULL;

  double *rho_data;
  double time = WD_time(d);

  elem_client_rec_t *ec = NULL;

  FILE *stream;
  char filename[1024];

  int i, j, pntidx;

  assert(r != NULL);   

  ec = (elem_client_rec_t *)get_list_next (r->elem_clients, ec);
  while (ec != NULL)
  {
    cd = (history_client_data_t *)(ec->history.client_data);

    pntidx = cd->pntidx;
    rho_data = hsi->rhost_dt[pntidx - 1];

    FOR(i, N_SLIP_SYSTEMS)
    {
       sprintf (filename, "%s/p%drho_%d.m", r->history_dir, pntidx, i+1);
       if ((stream = fopen(filename, "a")) == NULL)
           REPORT_ERROR(d->errh, "Could not open stream");

       fprintf(stream, "%g  %g\n", time, rho_data[i]);
       fclose(stream);
    }
    ec = (elem_client_rec_t *)get_list_next (r->elem_clients, ec);
  }
}
******************************************************************/

/**
 **	swap hsi->rhost and hsi->rhost_dt 
 **/

static void
swap_rhos(hardening_solver_t *hsi)
{
  rhos_t *rhos;

  rhos = hsi->rhost;
  hsi->rhost = hsi->rhost_dt;
  hsi->rhost_dt = rhos;
}


#define DBGPRHO(tit,num,what)  {                                  						 \
    int I_I_;                                     								 \
    fprintf (stderr, "%s = [ ", tit);                    							 \
    FOR (I_I_, num) {                        								 	 \
         fprintf (stderr, "%g,  %g,  %g,  %g ", what[I_I_][0], what[I_I_][1], what[I_I_][6], what[I_I_][7]);     \
       fprintf (stderr, ";\n");                          							 \
    }                                                   							 \
    fprintf (stderr, "]\n");                             							 \
  }

/** 
 **     solve the dislocatio density of all slip systems by loop 
 **     update the flow stress at the quadrature pts
 **/

static void
solve_for_hardening(void *client) /* client is a pointer to the address of hsi */
{
  char *dumplhs = "grain20.lhs";
  char *dumprhs = "grain20.rhs";
  char *dumpsol_gep = "gep.sol";
  char *dumpsol_dmeschach = "dmeschach.sol";

  hardening_solver_t *hsi = 0;
  int num_elem;
  int which_elem = 0;
 
  mat3dcryspl_elem_t *e = 0;
  W_mat_3d_t *m = 0;
  mat_t *u = 0;
  int i, j, num_nodes = 0; /* i,j are the counters */
 
  if (client == NULL) {
      REPORT_FATAL_ERROR (u->d->errh, "Failed in solve_for_hardening, client of the hardening_solver is NULL")
  }
 
/*  if (*client == NULL) {
    REPORT_FATAL_ERROR (u->d->errh, "Failed in solve_for_hardening, client of the hardening_solver is NULL")
  } */
 
  hsi = (hardening_solver_t *)client;
  num_elem = dict_count( hsi->eld );
  if( num_elem == 0 ) { create_mesh(hsi); } /* if hsi->eld is empty then create mesh */
 
  if (hsi->fs == 0){
     set_fles(hsi);
  }
 
  m = hsi->m;
  u = (mat_t *)(m->data);
  num_nodes = dict_count(hsi->eqnd);

  fprintf (stderr, "#    solve for hardening is called!\n");

  /* swap hsi->rhost_dt and hsi->rhost */
  swap_rhos(hsi);

  /* the loop for solving dislocation densities of all slip systems */

  for( i = 0; i<N_SLIP_SYSTEMS; i++)
  { 
     hsi->curr_sys = i; 

     hsi->fs->zero_rhs (hsi->fs);
     hsi->fs->zero_lhs (hsi->fs);

     hsi->fs_dmeschach->zero_rhs (hsi->fs_dmeschach);
     hsi->fs_dmeschach->zero_lhs (hsi->fs_dmeschach);

     which_elem = 1; 
     e = 0;
     while(e = (mat3dcryspl_elem_t *)dict_next_obj(hsi->eld, e)){
        e->methods->assemble_rhs(e);
        e->methods->assemble_lhs(e);
	which_elem++;
     }
 
     hsi->fs->finish_rhs (hsi->fs);
     hsi->fs->finish_lhs (hsi->fs, FLES_FINAL_FINISH);
 
     hsi->fs_dmeschach->finish_rhs (hsi->fs_dmeschach);
     hsi->fs_dmeschach->finish_lhs (hsi->fs_dmeschach, FLES_FINAL_FINISH);

     if (i == 0) {
       /*        hsi->fs->dump_lhs(hsi->fs, dumplhs);*/
        hsi->fs->dump_rhs(hsi->fs, dumprhs);
     }

     if (!hsi->fs->solve (hsi->fs))
         REPORT_FATAL_ERROR (u->d->errh, "Failed in solve_for_rhos")

     if (!hsi->fs_dmeschach->solve (hsi->fs_dmeschach))
         REPORT_FATAL_ERROR (u->d->errh, "Failed in dmeschach solver")

     if (i == 0) {
        hsi->fs->dump_sol(hsi->fs, dumpsol_gep);
        hsi->fs_dmeschach->dump_sol(hsi->fs_dmeschach, dumpsol_dmeschach);
     }
     save_rhost_dt(hsi); /* save solution to hsi->rhost_dt */

  } /* for (i; N_SLIP_SYSTEMS) */

  /* when program reaches here, the computation 
     of dislocation densities on all slip systems 
     is done 
   */

  /*go over the quadrature pts, update the flow stress
    go over every element, and its integration pts */

  /*update_yieldstress(hsi);
    done*/
  
  /* post-process */
  write_hexvue_file(hsi);

  /* rhos HIST data */
  RHO_update_history(hsi);
}


static double  
get_rho_state(hardening_solver_t  *hsi, 
              void                *matst, 
	      int                  i     /* i means which slip system is being taken care of */ 
             )
{
 /* which element matst lies */
  mat3dcryspl_elem_t *e = 0;
  double rho;

  /*  for plotting the dislocation densities, at the moment
      the start time in the input file must NOT start from 0
      for example, you can use 
        H "sequence 0.005 7 0.005" 
      instead of 
	H "sequence 0 7 0.005" 
      in the OPTION block of the input file *.in
      or the program will crash
   */

  /* return the element which holds the matst obj*/
  e = (mat3dcryspl_elem_t *)dict_find_obj_with_handle_name(hsi->matstd, matst); 
  rho = e->methods->get_rho(e, matst, i); 
  return rho;
}

static void
add_matst(hardening_solver_t *hsi, 
	  mat3dcryspl_elem_t *e, 
	  void *matst
	  )
{
  mat3dcryspl_elem_t *el;
  
  if ((el = dict_find_obj_with_handle_name(hsi->matstd, e)) == 0 ) {
     dict_add_obj_handle_name (hsi->matstd, e, matst );  
  }
}
