#include "Wconfig.h"
#include "Wxf.h"
#include "Whistry.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Wnode.h"
#include "Welem.h"
#include "Wnewmarki.h"

#include <fcntl.h>
#include <sys/stat.h>

/* RCSHEADER "$Id: histry.c,v 1.3 2003/08/02 21:04:14 cvs Exp $" */

#define HIST_DIR_NAME "WASP.HIST"
#define HIST_DIR_MODE (S_IREAD | S_IWRITE | S_IEXEC)


typedef struct var_hist_pair_t {
  double time;
  double val;
}              var_hist_pair_t;

typedef struct var_hist_buff_t {
  char                *hist_file_base_name;
  int                  buffdim;
  int                  buffidx;
  var_hist_pair_t     *pairs;
  W_bool_t             first_time;
}              var_hist_buff_t;


typedef struct node_client_rec_t {
  W_index_t                   idx;
  W_solution_history_t        history;
}              node_client_rec_t;


typedef struct elem_client_rec_t {
  W_index_t                   idx;
  W_solution_history_t        history;
}              elem_client_rec_t;


static void
strip_path (char *dest, char *src)
{
  char *p;
  
  p = strrchr(src, '/');
  if (p == NULL) sprintf(dest, "%s", src);
  else {
    sprintf(dest, "%s", ++p);
  }
}


static void
history_dir_name (W_history_writer_t *r)
{
  char *PWD = NULL;
  char buffer[MAXFILENAMECHAR];
  char msg[MAXFILENAMECHAR];
  char inp[MAXFILENAMECHAR];

  if ((PWD = getenv("PWD")) == NULL) {
    PWD = ".";
  }

  sprintf(buffer, "%s/%s", PWD, HIST_DIR_NAME);
  if (!W_file_exists(buffer)) {
    if (mkdir(buffer, HIST_DIR_MODE) == -1) {
      sprintf(msg, "Cannot create history directory %s!?", buffer);
      REPORT_FATAL_ERROR(r->domain->errh, msg);
    }
  }
  if (WOPT_OPT (r->domain->options, hist_file_name) != NULL)
    strip_path(inp, WOPT_OPT (r->domain->options, hist_file_name));
  else
    strip_path(inp, WOPT_OPT (r->domain->options, in_file_name));
  sprintf(buffer, "%s/%s/%s", PWD, HIST_DIR_NAME, inp);
  if (W_file_exists(buffer)) {
    if (!WOPT_OPT (r->domain->options, careless)) {
      if (rmdir(buffer) == -1) {
        sprintf(msg, "History directory %s not empty!?", buffer);
        REPORT_FATAL_ERROR(r->domain->errh, msg);
      }
      if (mkdir(buffer, HIST_DIR_MODE) == -1) {
        sprintf(msg, "Cannot create history directory %s!?", buffer);
        REPORT_FATAL_ERROR(r->domain->errh, msg);
      }
    }
  } else {
    if (mkdir(buffer, HIST_DIR_MODE) == -1) {
      sprintf(msg, "Cannot create history directory %s!?", buffer);
      REPORT_FATAL_ERROR(r->domain->errh, msg);
    }
  }
  
  DUPLICATE_STRING (r->history_dir, buffer);
}



/*
   Create a new history writer.
 */
W_history_writer_t * /* new history writer */
WHR_new_history_writer (W_domain_t *d /* domain */
                         )
{
  W_history_writer_t *r;

  r = w_new(W_history_writer_t);
  CHECK_MALLOC(r);

  r->domain                       = d;
  r->history_dir                  = NULL;
  r->var_hist_buff_d              = dict_create_dictionary (); CHECK_MALLOC (r->var_hist_buff_d);
  r->hist_buff_length             = WOPT_OPT (d->options, hist_buff_length);
  r->node_clients                 = make_list();
  r->elem_clients                 = make_list();
  r->domain_client                = FALSE;
  r->newmark_client               = FALSE;

  /* Set up the directories */
  history_dir_name (r);
  
  return r;
}

static var_hist_buff_t *
find_var_hist_buff (W_history_writer_t *hw, char *hist_file_base_name)
{
  var_hist_buff_t *vhb = dict_find_obj_with_string_name (hw->var_hist_buff_d, hist_file_base_name);
  if (!vhb) {
    vhb = w_new (var_hist_buff_t); CHECK_MALLOC (vhb);
    vhb->buffdim = hw->hist_buff_length + 1;
    vhb->buffidx = 0;
    vhb->pairs   = w_new_array (vhb->buffdim, var_hist_pair_t); CHECK_MALLOC (vhb->pairs);
    DUPLICATE_STRING (vhb->hist_file_base_name, hist_file_base_name);
    vhb->first_time = TRUE;
    dict_add_obj_string_name (hw->var_hist_buff_d, vhb, vhb->hist_file_base_name);
  }
  return vhb;
}


static void
store_to_var_hist_buff (var_hist_buff_t *vhb, double time, double val)
{
  int i = vhb->buffidx;
  vhb->pairs[i].time = time;
  vhb->pairs[i].val  = val;
  vhb->buffidx++;
}

static W_bool_t
have_place_in_var_hist_buff (var_hist_buff_t *vhb)
{
  return (vhb->buffidx < vhb->buffdim - 1);
}

static void
flush_var_hist_buff (var_hist_buff_t *vhb, W_history_writer_t *hw)
{
  char buf[1024];
  int i;
  FILE *stream;
  
  WHR_file_name (hw->history_dir, vhb->hist_file_base_name, buf);

  if (vhb->first_time) {
    WHR_write_header (buf, vhb->hist_file_base_name);
    vhb->first_time = FALSE;
  }

  if ((stream = fopen(buf, "a")) == NULL)
      REPORT_ERROR(hw->domain->errh, "Could not open stream");

  for (i = 0; i < vhb->buffidx; i++)
    fprintf(stream, "%g %g\n", vhb->pairs[i].time, vhb->pairs[i].val);
    
  fclose(stream);

  vhb->buffidx = 0;
}


static void
flush_all (W_history_writer_t *hw)
{
  dict_iterator di = dict_get_iterator ();
  var_hist_buff_t *vhb;
  if (!di) {
    REPORT_FATAL_ERROR (hw->domain->errh, "No iterator could be created");
  }
  dict_reset_iterator (di);
  while ((vhb = (var_hist_buff_t *) dict_iter_objs (hw->var_hist_buff_d, di))) {
    flush_var_hist_buff (vhb, hw);
  }
  dict_del_iterator (di);
}

/*
   Update history records for this history writer.
 */
void
WHR_update_history (W_history_writer_t *r, /* history writer */
                    W_bool_t should_flush /* should the buffers be flushed? */
                    )
{
  node_client_rec_t *nc = NULL;
  elem_client_rec_t *ec = NULL;
  
  assert(r != NULL);
  
  nc = (node_client_rec_t *)get_list_next (r->node_clients, nc);
  while (nc != NULL) {
    WN_update_history (WD_get_node_by_idx (r->domain, nc->idx),
                      &nc->history, WD_time(r->domain));
    nc = (node_client_rec_t *)get_list_next (r->node_clients, nc);
  }

  ec = (elem_client_rec_t *)get_list_next (r->elem_clients, ec);
  while (ec != NULL) {  
    WE_update_history (WD_get_elem_by_idx (r->domain, ec->idx),
                      &ec->history, WD_time (r->domain));
    ec = (elem_client_rec_t *)get_list_next (r->elem_clients, ec);
  }

  if (r->domain_client)
    WD_update_history (r->domain);

  if (r->newmark_client && r->domain->newmark_integrator)
    W_newmark_update_history (r->domain->newmark_integrator, r);

  if (should_flush)
    flush_all (r);
}


/*
   Add a node client to the history writer.
 */
W_bool_t
WHR_add_node_client (W_history_writer_t *r, /* history writer */
                     W_index_t idx, /* index of the client */
                     char *input_line /* input line from the input file */
                     )
{
  node_client_rec_t *nc = NULL;
  W_node_t *n;

  assert(r != NULL);

  n = WD_get_node_by_idx(r->domain, idx);

  if (n) {
    nc = w_new (node_client_rec_t);
    CHECK_MALLOC(nc);
    
    nc->idx                 = idx;
    nc->history.writer      = r;
    nc->history.client_data =
      WN_new_history_client_data (n, nc->idx, r, input_line);
    add_to_tail(r->node_clients, nc);
  } else {
    sprintf (W_buf (), "Node %d not used by domain; no history is output",
             (int) idx);
    REPORT_WARNING (r->domain->errh, W_buf ());
  }
  
  return TRUE;
}




/*
   Add an element history client to the history writer.
 */
W_bool_t
WHR_add_elem_client (W_history_writer_t *r,/* history writer */
                     W_index_t idx, /* index of the client */
                     char *input_line /* input line from the input file */
                     )
{
  elem_client_rec_t *ec = NULL;
  W_elem_t *el;

  assert(r != NULL);
  
  ec = w_new(elem_client_rec_t);
  CHECK_MALLOC(ec);

  ec->idx                 = idx;
  ec->history.writer      = r;
  el = WD_get_elem_by_idx(r->domain, ec->idx);
  
  if (el == 0) {
    sprintf (W_buf(), "Element %d not found", ec->idx);
    REPORT_FATAL_ERROR (NULL, W_buf());
  }
  ec->history.client_data
    = WE_new_history_client_data (el, ec->idx, r, input_line);
  add_to_tail(r->elem_clients, ec);
 
  return TRUE;
}



/*
   The reader of history clients.
 */
W_bool_t
W_history_clients_reader (W_domain_t *d)
{
  const char *input_line;
  long cnum;
  char ccode;

  if (d->history_writer == NULL)
    if (WOPT_OPT (d->options, no_output)) {
      return W_read_and_discard_block (d);
    } else {
      REPORT_FATAL_ERROR (d->errh, "no history writer");
      return FALSE;
    }
  
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      TOKENS_PARSE(input_line);
      ccode = TOKEN_STR(1)[0];
      switch (ccode) {
      case NODE_CLIENT:
        cnum = TOKEN_INT(2);
        WHR_add_node_client(d->history_writer, cnum, (char *)input_line);
        break;
      case ELEM_CLIENT:
        cnum = TOKEN_INT(2);
        WHR_add_elem_client(d->history_writer, cnum, (char *)input_line);
        break;
      case ENERGY_CLIENT:
      case DOMAIN_CLIENT:
        d->history_writer->domain_client = TRUE;
        break;
      case NEWMARK_CLIENT:
        d->history_writer->newmark_client = TRUE;
        break;
      default:
        REPORT_FATAL_ERROR(d->errh, "Unknown client");
        break;
      }
    }
  }
  return FALSE;
}


#define XMGR_DATA_FORMAT 1
#if XMGR_DATA_FORMAT
static int current_set = 0;
#define MAX_SET_NUM 29
static int set_color_table_idx = 0;
static int set_color_table[] =
{
  1, 2, 3, 4, 8, 10, 11, 12, 13, 14, 15,
  1, 2, 3, 4, 8, 10, 11, 12, 13, 14, 15,
  1, 2, 3, 4, 8, 10, 11, 12, 13, 14, 15,
  1, 2, 3, 4, 8, 10, 11, 12, 13, 14, 15,
  1, 2, 3, 4, 8, 10, 11, 12, 13, 14, 15
};
#define  MAX_SET_COLOR_TABLE_IDX 54 /* must correspond to the above! */
static int set_style_table_idx = 0;
static int set_style_table[] = {1, 2, 3, 4, 5};
#define  MAX_SET_STYLE_TABLE_IDX 4
#endif

/*
   Write a header to a disk history file.
 */
void
WHR_write_header (char *file, /* history file name */
                  char *ytitle /* label for the y-axis */
                  )
{
  FILE *stream;
  int fd;
  char buffer[MAXFILENAMECHAR];
#if XMGR_DATA_FORMAT
  char date[80];
#endif
  
#define PMODE 0755
#define POL(aLine) fprintf(stream, "%s", aLine)
  
  if ((fd = creat(file, PMODE)) == -1) { /* truncate the file  */
    sprintf(buffer, "History file %s could not be emptied.", file);
    REPORT_FATAL_ERROR(NULL, buffer);
  }
  close(fd);
  
  if ((stream = fopen(file, "a")) == NULL)
    REPORT_FATAL_ERROR(NULL, "Could not open history file");

#if XMGR_DATA_FORMAT
  POL("@    title \"WASP (C) 1994-1999 P.Krysl\"\n");
  POL("@g0  autoscale type SPEC\n");
  POL("@    title font 2\n");
  POL("@    title size 1.12\n");
  W_get_date_and_time(date);
  sprintf(buffer, "@    subtitle \"%s, %s\"\n", date, file);
  POL(buffer);
  POL("@    subtitle font 2\n");
  POL("@    subtitle size 0.8\n");
  POL("@    xaxis  label \"Time\"\n");
  POL("@    xaxis  ticklabel prec 6\n");
  POL("@    xaxis  ticklabel format general\n");
  POL("@    xaxis  ticklabel font 2\n");
  POL("@    xaxis  ticklabel char size 0.67\n");
  POL("@    xaxis  ticklabel start 0.000000\n");
  POL("@    xaxis  tick major grid on\n");
  POL("@    yaxis  label \""); POL(ytitle); POL("\"\n");
  POL("@    yaxis  ticklabel prec 6\n");
  POL("@    yaxis  ticklabel format general\n");
  POL("@    yaxis  ticklabel font 2\n");
  POL("@    yaxis  ticklabel char size 0.67\n");
  POL("@    yaxis  ticklabel start 0.000000\n");
  POL("@    yaxis  tick major grid on\n");
  sprintf(buffer, "@    s%d type xy\n", current_set);
  POL(buffer);
  sprintf(buffer, "@    s%d linewidth 2\n", current_set);
  POL(buffer);
  sprintf(buffer, "@    s%d color %d\n", current_set,
          set_color_table[set_color_table_idx]);
  POL(buffer);
  sprintf(buffer, "@    s%d linestyle %d\n", current_set,
          set_style_table[set_style_table_idx]);
  POL(buffer);
#endif
  
  fclose(stream);

#if XMGR_DATA_FORMAT
  current_set++;
  if (current_set > MAX_SET_NUM)
    current_set = 0;
  set_color_table_idx++;
  if (set_color_table_idx > MAX_SET_COLOR_TABLE_IDX)
    set_color_table_idx = 0;
  set_style_table_idx++;
  if (set_style_table_idx > MAX_SET_STYLE_TABLE_IDX)
    set_style_table_idx = 0;
#endif
}



void
WHR_file_name (char *history_dir,
               char *client_data_name,
               char *buf
               )
{
  char abuf[256];
  sprintf (abuf, "%s", client_data_name);
#if XMGR_DATA_FORMAT
  sprintf (buf, "%s/%s.xmgr", history_dir, abuf);
#else
  sprintf (buf, "%s/%s.gpl", history_dir, abuf);
#endif
}

int
WHR_store (W_history_writer_t *hw, char *hist_file_base_name, double time, double val)
{
  var_hist_buff_t *vhb = find_var_hist_buff (hw, hist_file_base_name);
  if (!vhb) return FALSE;
  else {
    if (!have_place_in_var_hist_buff (vhb))
      flush_var_hist_buff (vhb, hw);
    store_to_var_hist_buff (vhb, time, val);
    return TRUE;
  }
}
