#include <stdio.h>
#include <stdlib.h>
#include "getoptP.h"
#include "tokensP.h"
#include "newP.h"

static char *prog = NULL;

static void
error (char *s, char *f, int l)
{
  fprintf (stderr, "\n**** ERROR in %s (%s;l.%d): %s\n",
           (prog ? prog : "?"), f, l, s);
  exit (1);
}
#define ERROR(S) { error (S, __FILE__, __LINE__); }

typedef struct abscissa_ordinate_pair_t {
  double     abscissa;
  double     ordinate;
}              abscissa_ordinate_pair_t;

typedef struct tabular_data_t {
  int                       num_of_pairs;
  abscissa_ordinate_pair_t *pairs;
}              tabular_data_t;


static double
min_abscissa (tabular_data_t *td)
{
  if (td->num_of_pairs >= 1) {
    return td->pairs[0].abscissa;
  } else {
    return 0;
  }
}

static double
max_abscissa (tabular_data_t *td)
{
  if (td->num_of_pairs >= 1) {
    return td->pairs[td->num_of_pairs-1].abscissa;
  } else {
    return 0;
  }
}

static double
eval_ordinate (tabular_data_t *td,
               double abscissa)
{
  double dA, dO;
  int i, L = 0;
  int num_of_pairs = td->num_of_pairs;
  abscissa_ordinate_pair_t *pairs = td->pairs;

  if (num_of_pairs <= 1)
    return 0;
  
  dA = pairs[1].abscissa - pairs[0].abscissa;
  dO = pairs[1].ordinate - pairs[0].ordinate;

  i = 1;
  while (i < num_of_pairs) {
    L = i;
    if (abscissa <= pairs[i].abscissa)
      break;
    i++;
  }
  dA = pairs[L].abscissa - pairs[L-1].abscissa;
  dO = pairs[L].ordinate - pairs[L-1].ordinate;

  if (dA == 0)
    return (pairs[L].ordinate + pairs[L-1].ordinate) / 2;
  else
    return pairs[L-1].ordinate + (abscissa - pairs[L-1].abscissa) * dO / dA;
}


static tabular_data_t *
read_tabular_data (char *file)
{
  tabular_data_t *td;
  tokens_parser parser;
  typedef struct pair_t {
    double x, y;
  }              pair_t;
  pair_t *p;
  LIST pl;
  
  parser  = tokens_new_parser ();
  if (!parser) { ERROR ("No parser"); }

  if (!tokens_open_file (parser, (char *) file)) {
    char buf[1024];
    sprintf (buf, "Could not open file \"%s\"", file);
    ERROR (buf);
  }

  td = NEWP_OBJ (tabular_data_t);
  if (!td) {ERROR ("Not enough memory");}

  pl = make_list (); if (!pl) { ERROR ("Not enough memory"); }
  
  while (tokens_next_line (parser)) {
    if (tokens_total_of_tokens (parser) == 2 &&
        tokens_type_of_token (parser, 1) == TOKENS_NUMBER &&
        tokens_type_of_token (parser, 2) == TOKENS_NUMBER
        ) {
      p = NEWP_OBJ (pair_t); if (!p) { ERROR ("Not enough memory"); }
      p->x = tokens_token_as_double (parser, 1);
      p->y = tokens_token_as_double (parser, 2);
      add_to_tail (pl, p);
    }
  }

  td->num_of_pairs = list_length (pl);
  if (td->num_of_pairs < 1) {
    ERROR ("No data");
  }
  td->pairs = NEWP_OBJ_ARRAY (td->num_of_pairs, abscissa_ordinate_pair_t);
  if (! td->pairs) { ERROR ("Not enough memory"); }

  {
    int i = 0;
    loop_through_list (pl, p, pair_t *) {
      td->pairs[i].abscissa = p->x;
      td->pairs[i].ordinate = p->y;
      i++;
    }
  }
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
  free_list (pl, AND_NODES);

  return td;
}

typedef double (*eval_func) (void *data, double abscissa);


static double
l2 (void *data, eval_func eval, double from, double to, int nintervals)
{
  double da = (to - from) / nintervals;
  int i;
  double abscissa, o, po, sum;
  
  if (da == 0) return 0;
  if (da < 0) { double x = to; to = from; from = x; da = -da; }

  
  for (i = 0,
       sum = 0,
       abscissa = from,
       po = eval (data, abscissa); i < nintervals; i++) {
    abscissa += da;
    o = eval (data, abscissa);
    sum += da * (o + po) / 2;
    po = o;
  }
  return sqrt (sum);
}

static double
eval_l2 (void *data, double abscissa)
{
  double v = eval_ordinate ((tabular_data_t *) data, abscissa);
  return (v*v);
}

static double
l2_of_one (char *file, double up_to_time)
{
  tabular_data_t *td1;
  double mina, maxa;

  td1 = read_tabular_data (file); if (!td1) { ERROR ("No table"); }

  mina = min_abscissa (td1);
  maxa = max_abscissa (td1);
  maxa = min (maxa, up_to_time);

  return l2 (td1, eval_l2, mina, maxa, td1->num_of_pairs-1);
}


typedef struct two_tabular_data_t {
  tabular_data_t *td1, *td2;
}              two_tabular_data_t;

static double
eval_l2_diff (void *data, double abscissa)
{
  two_tabular_data_t *two = (two_tabular_data_t *)data;
  double v = (eval_ordinate (two->td1, abscissa) - eval_ordinate (two->td2, abscissa));
  return (v*v);
}

static double
l2_of_diff (char *file1, char *file2, double up_to_time)
{
  tabular_data_t *td1, *td2;
  double mina, maxa;
  two_tabular_data_t two;

  td1 = read_tabular_data (file1); if (!td1) { ERROR ("No table"); }
  td2 = read_tabular_data (file2); if (!td2) { ERROR ("No table"); }

  mina = min_abscissa (td1);
  mina = min (mina, min_abscissa (td2));
  maxa = max_abscissa (td1);
  maxa = max (maxa, max_abscissa (td2));
  maxa = min (maxa, up_to_time);

  two.td1 = td1;
  two.td2 = td2;
  
  return l2 (&two, eval_l2_diff, mina, maxa,
             max (td1->num_of_pairs, td2->num_of_pairs)-1);
}


static char *file1 = NULL;
static char *file2 = NULL;
static int relative = FALSE;
static double up_to_time = 1e30;

static void 
usage (int argc, char **argv) 
{
  char *option, *usage;

  fprintf (stderr, "usage: %s + options below\n", argv[0]);
  
  if (ckit_opt_first_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option ? option : "?"),
             (usage ? usage  : "?"));
  }  
  while (ckit_opt_next_recognized (&option, &usage)) {
    fprintf (stderr, "%s %s\n",
             (option ? option : "?"),
             (usage ? usage  : "?"));
  }  
  exit (0);
}  


static void
parse_cmd_line (int argc, char **argv) 
{
  int help = FALSE;
#define SET_BOOL(name) \
    {                                                             \
    name = TRUE;                                                \
  }  
#define SET_WITH_ARG(name, format) 

  
  GET_OPT_BOOL (argc, argv, "-relative", SET_BOOL (relative),
                "compute relative norm: r=|u1-u2|_2/|u1|_2");
  
  GET_OPT_BOOL (argc, argv, "-help", SET_BOOL (help),
                "prints help");
  
  GET_OPT_ARG (argc, argv, "-u1", char, file1,
              SET_WITH_ARG (file1, s),
              "file with data u1; L2 norm |u1|_2");

  GET_OPT_ARG (argc, argv, "-u2", char, file2,
              SET_WITH_ARG (file2, s),
              "file with data u2; L2 norm |u2|_2");

  GET_OPT_ARG (argc, argv, "-up_to_time", double, up_to_time,
              SET_WITH_ARG (up_to_time, g),
              "evaluate up to time");

  if (help) 
    usage (argc, argv);
}  


int
main (int argc, char **argv)
{
  tabular_data_t *td1, *td2;
  double mina, maxa, l2diff;
  two_tabular_data_t two;

  prog = argv[0];
  parse_cmd_line (argc, argv);
  
  if (file1 && !file2) {
    double u1n = l2_of_one (file1, up_to_time);
    printf ("%g\n", u1n);
  } else if (file1 && file2) {
    double dn, un1 = 1;
    dn = l2_of_diff (file1, file2, up_to_time);
    if (relative) un1 = l2_of_one (file1, up_to_time);
    printf ("%g\n", dn / un1);
  } else {
    usage (argc, argv);
  }

  return 0;
}
