#include "Wconfig.h"
#include "Wxf.h"
#define WMAT002_PRIVATE_HEADER
#include "Wmat002.h"
#include "Wdomain.h"
#include "Wmats.h"

/* RCSHEADER "$Id: mat002.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct mat_t {
  W_float_t    E;
  W_float_t    nu;
  W_float_t    yield_pos;
  W_float_t    yield_neg;
}              mat_t;

typedef struct mat002_matstate_t {
  W_float_t   stress;
  W_float_t   strain;
}              mat002_matstate_t;


static W_bool_t
mat002_get_uniaxial_elast_const(W_material_t *m,
                                 W_float_t *E,
                                 W_float_t *nu);

static W_matstate_t 
mat002_get_mat_state(W_material_t *m);

static W_bool_t
mat002_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           );

static W_bool_t
mat002_is_in_plastic_state(W_material_t *m,
                           W_matstate_t ms);

static W_float_t
mat002_get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                               W_float_t strain);


static W_material_methods_t mat002_methods = {
  mat002_get_uniaxial_elast_const,
  mat002_get_uniaxial_stress,
  mat002_get_uniaxial_def_energy,
  NULL,
  NULL,
  NULL,
  NULL,
  mat002_get_mat_state,
  mat002_is_in_plastic_state
};
  

/* Reads uniaxial linear elastic material data from stream */
/* as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates materials. */

W_bool_t 
mat002_reader(W_domain_t *d)
{
  const char *input_line;
  long mat_num;
  double E, nu, vol_mass, yield_pos, yield_neg;
  mat_t *u;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld %lf %lf %lf %lf %lf",
                 &mat_num, &E, &nu, &vol_mass, &yield_pos, &yield_neg) != 6) {
        BAD_INPUT();
      } else {
        u = w_new(mat_t);
        CHECK_MALLOC(u);
        u->E         = E; 
        u->nu        = nu;
        u->yield_pos = yield_pos;
        u->yield_neg = yield_neg;
        WD_add_material(d, mat_num, 
                        WM_new_material(MCODE_UNIAX_LE, mat_num,
                                        &mat002_methods, u, vol_mass));
      }
    }
  }
  return FALSE;
}


static W_bool_t
mat002_get_uniaxial_elast_const(W_material_t *m,
                                 W_float_t *E,
                                 W_float_t *nu)
{
  mat_t *u = (mat_t *)m->data;

  *E  = u->E;
  *nu = u->nu;
  return TRUE;
}



static W_matstate_t 
mat002_get_mat_state(W_material_t *m)
{ 
  mat002_matstate_t *ms;
  
  ms = w_new(mat002_matstate_t);
  CHECK_MALLOC(ms);

  ms->stress = ms->strain = 0;
  return (W_matstate_t)ms;
}



static W_bool_t
mat002_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           )
{
  mat_t *u = (mat_t *)m->data;
  mat002_matstate_t *m002ms = (mat002_matstate_t *)ms;
  W_float_t dsigma, sigma;

  if (   m002ms->stress > u->yield_neg
      && m002ms->stress < u->yield_pos) { /* elastic state */
    dsigma  = (strain - m002ms->strain) * u->E;
    sigma   = m002ms->stress + dsigma;
    sigma   = min(sigma, u->yield_pos);
    sigma   = max(sigma, u->yield_neg);
  } else {                                /* plastic state */
    if (strain_rate * m002ms->strain > 0) /* loading */
      sigma = m002ms->stress;
    else {                                /* unloading */
      dsigma  = (strain - m002ms->strain) * u->E;
      sigma   = m002ms->stress + dsigma;
      sigma   = min(sigma, u->yield_pos);
      sigma   = max(sigma, u->yield_neg);
    }
  }
  /* update material state */
  m002ms->strain = strain;
  *stress = m002ms->stress = sigma;
  return TRUE;
}



static W_bool_t
mat002_is_in_plastic_state(W_material_t *m,
                           W_matstate_t ms)
{
  mat_t *u = (mat_t *)m->data;
  mat002_matstate_t *m002ms = (mat002_matstate_t *)ms;
  
  if (   m002ms->stress <= u->yield_neg
      || m002ms->stress >= u->yield_pos) { 
    return TRUE;
  } else {       
    return FALSE;
  }
}




static W_float_t
mat002_get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                               W_float_t strain)
{
  mat_t *u = (mat_t *)m->data;
  mat002_matstate_t *m002ms = (mat002_matstate_t *)ms;
  
  return 0.5 * m002ms->stress * m002ms->stress / u->E;
}
