#include "Wconfig.h"
#include "Wxf.h"
#define WMAT003_PRIVATE_HEADER
#include "Wmat003.h"
#include "Wdomain.h"
#include "Wmats.h"

/* RCSHEADER "$Id: mat003.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct mat_t {
  W_float_t    E;
  W_float_t    nu;
  W_float_t    yield_stress_upp;
  W_float_t    yield_stress_low;
  int          hardening_type;
  W_float_t    hardening_modulus_upp;
  W_float_t    hardening_modulus_low;
}              mat_t;

typedef struct mat003_matstate_t {
  W_bool_t    is_plastic;
  W_float_t   stress;
  W_float_t   strain;
  W_float_t   dstrain;
  W_float_t   yield_stress_upp;
  W_float_t   yield_stress_low;
  W_float_t   def_energy;
}              mat003_matstate_t;




static W_bool_t
mat003_get_uniaxial_elast_const(W_material_t *m,
                                 W_float_t *E,
                                 W_float_t *nu);

static W_matstate_t 
mat003_get_mat_state(W_material_t *m);

static W_bool_t
mat003_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           );

static W_bool_t
mat003_is_in_plastic_state(W_material_t *m,
                           W_matstate_t ms);

static W_float_t
mat003_get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                               W_float_t strain);

static W_material_methods_t mat003_methods = {
  mat003_get_uniaxial_elast_const,
  mat003_get_uniaxial_stress,
  mat003_get_uniaxial_def_energy,
  NULL,
  NULL,
  NULL,
  NULL,
  mat003_get_mat_state,
  mat003_is_in_plastic_state
};
 


/* Reads uniaxial elastic/plastic material with hardening data from stream */
/* as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates materials. */

W_bool_t 
mat003_reader(W_domain_t *d)
{
  const char *input_line;
  long mat_num;
  double E, nu, vol_mass, yield_upp, yield_low, hmupp, hmlow;
  int htype;
  mat_t *u;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld %lf %lf %lf %lf %lf %d %lf %lf",
                 &mat_num, &E, &nu, &vol_mass, &yield_upp, &yield_low,
                 &htype, &hmupp, &hmlow) != 9) {
        BAD_INPUT();
      } else {
        u = w_new(mat_t);
        CHECK_MALLOC(u);
        u->E                     = E; 
        u->nu                    = nu;
        u->yield_stress_upp      = yield_upp;
        u->yield_stress_low      = yield_low;
        u->hardening_type        = htype;
        u->hardening_modulus_upp = hmupp;
        u->hardening_modulus_low = hmlow;
        WD_add_material(d, mat_num, 
                        WM_new_material(MCODE_UNIAX_LE, mat_num,
                                        &mat003_methods, u, vol_mass));
      }
    }
  }
  return FALSE;
}


static W_bool_t
mat003_get_uniaxial_elast_const(W_material_t *m,
                                 W_float_t *E,
                                 W_float_t *nu)
{
  mat_t *u = (mat_t *)m->data;

  *E  = u->E;
  *nu = u->nu;
  return TRUE;
}



static W_matstate_t 
mat003_get_mat_state(W_material_t *m)
{ 
  mat003_matstate_t *ms;
  mat_t *u = (mat_t *)m->data;
  
  ms = w_new(mat003_matstate_t);
  CHECK_MALLOC(ms);

  ms->stress = ms->strain = 0;
  ms->is_plastic = FALSE;
  ms->yield_stress_low = u->yield_stress_upp;
  ms->yield_stress_upp = u->yield_stress_low;
  return (W_matstate_t)ms;
}





static W_bool_t
mat003_is_in_plastic_state(W_material_t *m,
                           W_matstate_t ms)
{
  mat003_matstate_t *m003ms = (mat003_matstate_t *)ms;
  
  return (m003ms->is_plastic != 0);
}




static W_bool_t
mat003_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           )
{
  mat_t *u = (mat_t *)m->data;
  mat003_matstate_t *m003ms = (mat003_matstate_t *)ms;
  W_float_t dsigma, sigma, dsigmay, deps;

  deps = (strain - m003ms->strain);
  if (!m003ms->is_plastic) { /* elastic state */
    dsigma  = deps * u->E;
    if (deps > 0) { /* extending */
      dsigmay = (m003ms->yield_stress_upp - m003ms->stress);
      if (dsigma / dsigmay > 1) { /* yielding */
        dsigma  = (deps - dsigmay / u->E) * u->hardening_modulus_upp;
        sigma   = m003ms->yield_stress_upp + dsigma;
        m003ms->is_plastic = +1;
      } else {
        sigma   = m003ms->stress + deps * u->E;
      }
    } else {        /* shortening */
      dsigmay = (m003ms->yield_stress_low - m003ms->stress);
      if (dsigma / dsigmay > 1) { /* yielding */
        dsigma  = (deps - dsigmay / u->E) * u->hardening_modulus_low;
        sigma   = m003ms->yield_stress_low + dsigma;
        m003ms->is_plastic = -1;
      } else {
        sigma   = m003ms->stress + deps * u->E;
      }
    }
  } else {                    /* plastic state */
    if (m003ms->dstrain * m003ms->is_plastic > 0) /* loading */
      if (m003ms->stress > m003ms->yield_stress_upp)
        sigma = m003ms->stress + deps * u->hardening_modulus_upp;
      else
        sigma = m003ms->stress + deps * u->hardening_modulus_low;
    else {                      /* unloading */
      if (u->hardening_type == MAT003_ISOTROPIC_HARDENING) {
        if (m003ms->stress > m003ms->yield_stress_upp)
          m003ms->yield_stress_upp = m003ms->stress;
        else 
          m003ms->yield_stress_low = m003ms->stress;
      } else if (u->hardening_type == MAT003_KINEMATIC_HARDENING) {
        double d = m003ms->yield_stress_upp - m003ms->yield_stress_low;
        
        if (m003ms->stress > m003ms->yield_stress_upp) {
          m003ms->yield_stress_upp = m003ms->stress;
          m003ms->yield_stress_low = m003ms->yield_stress_upp - d;
        } else { 
          m003ms->yield_stress_low = m003ms->stress;
          m003ms->yield_stress_upp = m003ms->yield_stress_low + d;
        }
      }
      dsigma  = deps * u->E;
      sigma   = m003ms->stress + dsigma;
      m003ms->is_plastic = FALSE;
    }
  }
  /* update material state */
  m003ms->dstrain = strain - m003ms->strain;
  m003ms->strain = strain;
  *stress = m003ms->stress = sigma;
  return TRUE;
}


static W_float_t
mat003_get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                               W_float_t strain)
{
  mat_t *u = (mat_t *)m->data;
  mat003_matstate_t *m003ms = (mat003_matstate_t *)ms;
  
  return 0.5 * m003ms->stress * m003ms->stress / u->E;
}
