#include "Wconfig.h"
#include "Wxf.h"
#define WMAT004_PRIVATE_HEADER
#include "Wmat004.h"
#include "Wdomain.h"
#include "Wmats.h"

/* RCSHEADER "$Id: mat004.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct mat_t {
  W_float_t    E;
  W_float_t    nu;
  int          pss_code;
}              mat_t;

typedef struct mat004_matstate_t {
  int      pss_type;
}              mat004_matstate_t;


static W_bool_t
mat004_get_uniaxial_elast_const(W_material_t *m,
                                 W_float_t *E,
                                 W_float_t *nu);


static W_matstate_t 
mat004_get_mat_state(W_material_t *m);

static W_bool_t
mat004_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           );

static W_bool_t
mat004_get_pss_stress(W_material_t *m,
                      W_matstate_t ms,
                      W_float_t eps1, W_float_t eps2, W_float_t gamma,
                      W_float_t eps1_rate,
                      W_float_t eps2_rate,
                      W_float_t gamma_rate,
                      W_float_t *sigma1,  W_float_t *sigma2,
                      W_float_t *tau
                      );

static W_float_t
mat004_get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                               W_float_t strain);

static W_float_t
mat004_get_pss_def_energy(W_material_t *m, W_matstate_t ms,
                            W_float_t eps1, W_float_t eps2, W_float_t gamma);


static W_material_methods_t mat004_methods = {
  mat004_get_uniaxial_elast_const,
  mat004_get_uniaxial_stress,
  mat004_get_uniaxial_def_energy,
  mat004_get_pss_stress,
  mat004_get_pss_def_energy,
  NULL,
  NULL,
  mat004_get_mat_state,
  NULL
};


/* Reads data of linear isotropic elastic material data from stream */
/* as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates materials. */

W_bool_t 
mat004_reader(W_domain_t *d)
{
  const char *input_line;
  long mat_num, pss_code;
  double E, nu, rho;
  mat_t *u;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      TOKENS_PARSE(input_line);
      mat_num = TOKEN_INT (1);
      E = TOKEN_DBL (2);
      nu = TOKEN_DBL (3);
      pss_code = TOKEN_INT (4);
      rho = TOKEN_DBL (5);
      u = w_new(mat_t);
      CHECK_MALLOC(u);
      u->E        = E; 
      u->nu       = nu;
      u->pss_code = pss_code;
      WD_add_material(d, mat_num, 
                      WM_new_material(MCODE_BIAX_LE, mat_num, &mat004_methods, u, rho));
    }
  }
  return FALSE;
}


static W_bool_t
mat004_get_uniaxial_elast_const(W_material_t *m,
                                 W_float_t *E,
                                 W_float_t *nu)
{
  mat_t *u = (mat_t *)m->data;

  *E  = u->E;
  *nu = u->nu;
  return TRUE;
}


static W_matstate_t 
mat004_get_mat_state(W_material_t *m)
{
  mat004_matstate_t *ms;
  
  ms = w_new(mat004_matstate_t);
  CHECK_MALLOC(ms);

  ms->pss_type = PSSCODE_PLANE_STRESS; /* default */
  return (W_matstate_t)ms;
}



static W_bool_t
mat004_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           )
{
  mat_t *u = (mat_t *)m->data;
  
  *stress = u->E * strain;
  return TRUE;
}



static W_float_t
mat004_get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                               W_float_t strain)
{
  mat_t *u = (mat_t *)m->data;
  
  return 0.5 * u->E * strain * strain;
}


static W_bool_t
mat004_get_pss_stress(W_material_t *m,
                      W_matstate_t ms,
                      W_float_t eps1, W_float_t eps2, W_float_t gamma,
                      W_float_t eps1_rate,
                      W_float_t eps2_rate,
                      W_float_t gamma_rate,
                      W_float_t *sigma1,  W_float_t *sigma2,
                      W_float_t *tau
                      )
{
  mat_t *u = (mat_t *)m->data;
  mat004_matstate_t *m004ms = (mat004_matstate_t *)ms;
  double ee, G;
  
  if (m004ms->pss_type == PSSCODE_PLANE_STRAIN) {
    ee = u->E / ((1 + u->nu) * (1 - u->nu - u->nu));
    G  = u->E / (2 + u->nu + u->nu);
    *sigma1 = (  (1 - u->nu) * ee * eps1
               +      u->nu  * ee * eps2);
    *sigma2 = (       u->nu  * ee * eps1
               + (1 - u->nu) * ee * eps2);
    *tau    =  G * gamma;
  } else {           /* PSSCODE_PLANE_STRESS */
    ee = u->E / (1 - u->nu * u->nu);
    G  = u->E / (2 * (1 + u->nu));
    *sigma1 = (           ee * eps1
               +  u->nu * ee * eps2);
    *sigma2 = (   u->nu * ee * eps1
               +          ee * eps2);
    *tau    =  G * gamma;
  }
  return TRUE;
}


static W_float_t
mat004_get_pss_def_energy(W_material_t *m, W_matstate_t ms,
                            W_float_t eps1, W_float_t eps2, W_float_t gamma)
{
  mat_t *u = (mat_t *)m->data;
  mat004_matstate_t *m004ms = (mat004_matstate_t *)ms;
  double ee, G;
  double sed = 0;

  if (m004ms->pss_type == PSSCODE_PLANE_STRAIN) {
    ee = u->E / ((1 + u->nu) * (1 - u->nu - u->nu));
    G  = u->E / (2 + u->nu + u->nu);
    sed =  ((1 - u->nu) * ee * eps1 + u->nu  * ee * eps2) * eps1;
    sed += (u->nu  * ee * eps1 + (1 - u->nu) * ee * eps2) * eps2;
    sed += G * gamma * gamma;
  } else {           /* PSSCODE_PLANE_STRESS */
    ee = u->E / (1 - u->nu * u->nu);
    G  = u->E / (2 * (1 + u->nu));
    sed = (ee * eps1 +  u->nu * ee * eps2) * eps1;
    sed += (u->nu * ee * eps1 + ee * eps2) * eps2;
    sed += G * gamma *gamma;
  }
  return 0.5 * sed;
}

