
#include "Wconfig.h"
#define WMAT005_PRIVATE_HEADER
#include "Wmat005.h"
#include "Wdomain.h"
#include "Wmats.h"

/* RCSHEADER "$Id: mat005.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct mat_t {
  W_float_t    E;
  W_float_t    nu;
  W_float_t    C[6][6];
}              mat_t;

static void
set_up_C(W_float_t E, W_float_t nu, W_float_t C[6][6]);

static W_bool_t
mat005_get_uniaxial_elast_const(W_material_t *m,
                                 W_float_t *E,
                                 W_float_t *nu);


static W_matstate_t 
mat005_get_mat_state(W_material_t *m);

static W_bool_t
mat005_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           );

static W_float_t
mat005_get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                               W_float_t strain);

static W_float_t 
mat005_get_3d_def_energy(W_material_t *m, W_matstate_t ms, W_float_t eps[6]);


static W_bool_t 
mat005_get_3d_stress(W_material_t *m, W_matstate_t ms,
                     W_float_t eps[6], W_float_t eps_rate[6],
                     W_float_t sigma[6]);


static W_material_methods_t mat005_methods = {
  mat005_get_uniaxial_elast_const,
  mat005_get_uniaxial_stress,
  mat005_get_uniaxial_def_energy,
  NULL,
  NULL,
  mat005_get_3d_stress,
  mat005_get_3d_def_energy,
  mat005_get_mat_state,
  NULL
};
  


/*
  Reads uniaxial linear elastic material data from stream 
  as long as there is no 
  END_OF_BLOCK tag in the input line. Creates materials.
*/
W_bool_t                        /* TRUE if successful */
mat005_reader(W_domain_t *d     /* domain */
              )
{
  const char *input_line;
  long mat_num;
  double E, nu, vol_mass;
  mat_t *u;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld %lf %lf %lf",
                 &mat_num, &E, &nu, &vol_mass) != 4) {
        BAD_INPUT();
      } else {
        u = w_new(mat_t);
        CHECK_MALLOC(u);
        u->E        = E; 
        u->nu       = nu;
        set_up_C(u->E, u->nu, u->C);
        WD_add_material(d, mat_num, 
                        WM_new_material(MCODE_TRIAX_LE, mat_num,
                                        &mat005_methods, u, vol_mass));
      }
    }
  }
  return FALSE;
}


static void
set_up_C(W_float_t E, W_float_t nu, W_float_t C[6][6])
{
  double c = E * (1 - nu) / ((1 + nu) * (1 - 2 * nu));
  int i, j;

  for (i = 0; i < 6; i++)
    for (j = 0; j < 6; j++)
      C[i][j] = 0;
  
  C[0][0] = c;
  C[0][1] = c * nu / (1 - nu);
  C[0][2] = C[0][1];
  C[1][0] = C[0][1];
  C[1][1] = c;
  C[1][2] = C[0][1];
  C[2][0] = C[0][1];
  C[2][1] = C[0][1];
  C[2][2] = c;
  C[3][3] = c * (1 - 2 * nu) / 2 / (1 - nu);
  C[4][4] = C[3][3];
  C[5][5] = C[3][3];
}


static W_bool_t
mat005_get_uniaxial_elast_const(W_material_t *m,
                                 W_float_t *E,
                                 W_float_t *nu)
{
  mat_t *u = (mat_t *)m->data;

  *E  = u->E;
  *nu = u->nu;
  return TRUE;
}



static int dummy_mat005_state = 0;

static W_matstate_t 
mat005_get_mat_state(W_material_t *m)
{
  return (W_matstate_t)&dummy_mat005_state; /* to avoid returning NULL */
}



static W_bool_t
mat005_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           )
{
  mat_t *u = (mat_t *)m->data;
  
  *stress = u->E * strain;
  return TRUE;
}



static W_float_t
mat005_get_uniaxial_def_energy(W_material_t *m, W_matstate_t ms,
                               W_float_t strain)
{
  mat_t *u = (mat_t *)m->data;
  
  return 0.5 * u->E * strain * strain;
}


static W_float_t 
mat005_get_3d_def_energy(W_material_t *m, W_matstate_t ms, W_float_t eps[6])
{
  double stress[6] = {0,0,0,0,0,0};

  mat005_get_3d_stress(m, ms, eps, NULL, stress);
  return 0.5 * (  stress[0]*eps[0]
                + stress[1]*eps[1]
                + stress[2]*eps[2]
                + stress[3]*eps[3]
                + stress[4]*eps[4]
                + stress[5]*eps[5]);
}

static W_bool_t 
mat005_get_3d_stress(W_material_t *m, W_matstate_t ms,
                     W_float_t eps[6], W_float_t eps_rate[6],
                     W_float_t sigma[6])
{
  mat_t *u = (mat_t *)m->data;

  sigma[0] = u->C[0][0] * eps[0] + u->C[0][1] * eps[1] + u->C[0][2] * eps[2];
  sigma[1] = u->C[1][0] * eps[0] + u->C[1][1] * eps[1] + u->C[1][2] * eps[2];
  sigma[2] = u->C[2][0] * eps[0] + u->C[2][1] * eps[1] + u->C[2][2] * eps[2];
  sigma[3] = u->C[3][3] * eps[3];
  sigma[4] = u->C[4][4] * eps[4];
  sigma[5] = u->C[5][5] * eps[5];

  return TRUE;
}
 
