#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat1d.h"

/* RCSHEADER "$Id: mat1d.c,v 1.2 2002/03/21 01:32:35 cvs Exp $" */

typedef struct reader_intermediary_t {
  W_mat1d_reader_func reader;
}              reader_intermediary_t;

static dict_dictionary reader_intermediary_table = NULL;



static W_mat1d_reader_func
find_reader (int m1dcode)
{
  reader_intermediary_t *ri = NULL;
  
  ri = dict_find_obj_with_int_name (reader_intermediary_table, m1dcode);
  if (ri == NULL) {
    sprintf (W_buf (), "Reader for mat3d type %d not registered",
             m1dcode);
    REPORT_ERROR (NULL, W_buf ());
  }
  return ri->reader;
}


static void 
initialize_readers_for_builtins (void)
{
#define W_DCL_READER(reader) W_mat1d_reader_func reader
#define W_REGISTER_READER(m1dcode, reader) WM1D_register_reader (m1dcode, reader)
#include "Wmat1deng.h"
#include "Wmat1dstvk.h"
#include "Wmat1dj2pl.h"
#include "Wmat1dj2vp.h"
#include "Wmat1dgen3d.h"
   W_REGISTER_READER (MAT_1D_STV_K, read_mat_1d_stv_k);
   W_REGISTER_READER (MAT_1D_ENG,   read_mat_1d_eng);
   W_REGISTER_READER (MAT_1D_J2PL,  read_mat_1d_j2pl);
   W_REGISTER_READER (MAT_1D_J2VP,  read_mat_1d_j2vp);
   W_REGISTER_READER (MAT_1D_GEN3D, read_mat_1d_gen3d);
#undef  W_DCL_READER
#undef  W_REGISTER_READER
}

/*
  Top-level 1d material reader.
   The generic mat1d reader.  The reader for the particular
   material in the input is found from among the registered
   mat1d readers.  If no reader for the material type was
   registered, NULL is returned; otherwise
 */
W_mat_1d_t */* If no reader for the material type was
   registered, NULL is returned; otherwise the result of the reader to which
   the reading was delegated is returned. */
mat1d_reader (W_domain_t *d /* domain */)
{
  static int readers_initialized = FALSE;
  const char *input_line;
  long m1dcode;
  W_mat1d_reader_func reader = NULL;
  
  if (!readers_initialized) {
    initialize_readers_for_builtins ();
    readers_initialized = TRUE;
  }

  input_line = W_get_input_line(d);
  TOKENS_PARSE(input_line);
  m1dcode = TOKEN_INT(1);

  reader = find_reader (m1dcode);
  if (reader != NULL) {
    return (*reader) (d);
  } else {
    return NULL;
  }
}


/*
   Register a mat1d reader.
   To be able to use specific implementation
   of the general triaxial material (mat1d),
   its code needs to be linked to the
   executable, and its reader needs to be registered, so that the generic
   mat1d reader can look up the type-specific reader.
   */
void
WM1D_register_reader (int m1dcode /* unique integer among all elements */,
                      W_mat1d_reader_func reader /* the reader function */
                      )
{
  reader_intermediary_t *ri;
  
  if (reader_intermediary_table == NULL) {
    reader_intermediary_table = dict_create_dictionary ();
  }
  ri = (reader_intermediary_t *) dict_find_obj_with_int_name (reader_intermediary_table, m1dcode);
  if (ri != NULL) {
    sprintf (W_buf (), "Reader for element type %d already registered",
             m1dcode);
    REPORT_ERROR (NULL, W_buf ());
  }
  ri = w_new (reader_intermediary_t); CHECK_MALLOC (ri);
  ri->reader = reader;
  dict_add_obj_int_name (reader_intermediary_table, ri, m1dcode);
}
  
/*
   Create a new 3-D material instance.
   The type specific information is supplied in "methods" and "data".
 */
W_mat_1d_t * /* material */
WM1D_new_material(
                  W_mat_1d_methods_t *methods, /* table of type-specific
                                                  functions
                                                */
                  void *data, /* mat1d type-specific data */
                  double vol_mass /* volume mass density */
                  )
{
  W_mat_1d_t *m;

  m = w_new(W_mat_1d_t); CHECK_MALLOC(m);

  m->vol_mass = vol_mass;
  m->methods  = methods;
  m->data     = data;
#ifdef X_XFACE
  m->color    = 0x0; 
  m->fringes  = NULL;
#endif
  return m;
}


/*
   Get volume mass density.
 */
double /* volume mass density */
WM1D_get_volume_mass (W_mat_1d_t *m/* mat1d */)
{
  return (m->vol_mass);
}


/*
   Get material type code.
 */
int /* code or -131313 signalling an invalid material or
       some kind of failure */
WM1D_get_mcode (W_mat_1d_t *m/* mat1d */)
{
  if (m->methods->get_mcode) {
    return (*m->methods->get_mcode) (m);
  } else {
    return -131313;
  }
}
 


/*
   Get material constant.
 */
W_bool_t 
WM1D_get_mat_const (W_mat_1d_t *m/* mat1d */,
                    W_mat_1d_mat_const_enum type, /* type of constant */
                    double *c /* value */
                    )
{
  if (m->methods->get_mat_const) {
    return (*m->methods->get_mat_const)(m, type, c);
  } else {
    *c = 0;
    return FALSE; 
  }
} 


/*
   Update stress in a material state and return it.
 */
W_bool_t /* TRUE if successful; FALSE otherwise, in which case an error
            message is generated
          */
WM1D_update_stress (W_mat_1d_t                  *m/* mat1d */,
                    W_mat_1d_state_t             ms, /* material state */
                    W_mat_state_update_form_enum update_form, /* update is final or temporary? */
                    W_mat_1d_ip_ctx_t           *ip_ctx,/* integration point context; used to generate kinematic quantities which the material state
                                                           might need in order to compute the requested stress */
                    W_mat_3d_stress_type_enum    stress_type,/* type of stress tensor */
                 double                       *stress,/* stress tensor */
                 W_mat_3d_stress_type_enum   *out_stress_type/* type of actually returned stress tensor; ideally, output type == requested type, but
                                                                sometimes the material state fails to compute the requested tensor for lack of information
                                                             */
                    )
{
  if (m->methods->update_stress) {
    return (*m->methods->update_stress)(m, ms, update_form, ip_ctx,
                                        stress_type, stress, out_stress_type);
  } else {
    return FALSE;
  }
}


/*
   Get material state variable.
 */
W_bool_t /* TRUE if successful; FALSE otherwise */
WM1D_get_mat_state_var (W_mat_1d_t           *m/* mat1d */,
                        W_mat_1d_state_t             ms, /* material state */
                        W_mat_1d_ip_ctx_t           *ip_ctx, /* integration point context; used to generate kinematic quantities which the material state
might need in order to compute the requested variable
                                                              */
                        W_mat_1d_mat_state_var_enum  var_type, /* variable type */
                        double                      *var /* variable value is returned in this argument  */
                        )
{
  if (m->methods->get_mat_state_var) {
    return (*m->methods->get_mat_state_var)(m, ms, ip_ctx,
                                            var_type, var);
  } else {
    return FALSE;
  }
}


/*
   Get a new material state at a given point.
   */
W_mat_1d_state_t /* material state */
WM1D_get_mat_state (W_mat_1d_t *m/* mat1d */,
                    W_loc_t *at_point /* point */
                    )
{
  if (m->methods->get_mat_state)
    return m->methods->get_mat_state (m, at_point);
  else
    return NULL;
}

W_bool_t
WM1D_get_stiff_coeff (W_mat_1d_t                  *m/* mat1d */,
                    W_mat_1d_state_t             ms, /* material state */
                    W_mat_1d_ip_ctx_t           *ip_ctx,/* integration point context; used to generate kinematic quantities which the material state
                                                           might need in order to compute the requested stress */
                      double *stiff_coeff
                      )
{
  if (m->methods->get_mat_state)
    return m->methods->get_stiff_coeff (m, ms, ip_ctx, stiff_coeff);
  else
    return FALSE;
}
