 
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat1d.h"
#include "Wmat1deng.h"

/* RCSHEADER "$Id: mat1deng.c,v 1.2 2002/03/21 01:32:35 cvs Exp $" */

static int
get_mcode (W_mat_1d_t *m)
{
  return MAT_1D_ENG;
}


typedef struct eng_mat_t {
  double E, nu;
  double dil_wave_speed_1d;
  double tensile_strength;
}              eng_mat_t;


typedef struct matstate_t {
  double   deform_energy_density;
}              matstate_t;


static W_bool_t
get_mat_const (W_mat_1d_t *m, W_mat_1d_mat_const_enum type, double *c)
{
  eng_mat_t *u = (eng_mat_t *)m->data;

  switch (type) {
  case MAT_1D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_1D_POISSON:
    *c = u->nu;
    break;
  case MAT_1D_TENSILE_STRENGTH:
    *c = u->tensile_strength; 
    break;
  case MAT_1D_DIL_WAVE_SPEED_1D:
    *c = u->dil_wave_speed_1d;
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

  
/* Function updates the stress state, and returns the stress (with proper */
/* conversion into the requested type).  Note that the function get_stress() */
/* is identical with this function since no stress state is saved for this */
/* type of elastic material. */
static W_bool_t 
update_stress (W_mat_1d_t                  *m,
               W_mat_1d_state_t             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_1d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                      *stress,
               W_mat_3d_stress_type_enum    *out_stress_type
               )
{
  double E, S;
  eng_mat_t *u = (eng_mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  matstate_t input_matst;

  /* save the material state if the update is temporary */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;
  
  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_ENG_STRAIN, &E))
    return FALSE;

  S = u->E * E;

  matst->deform_energy_density = 0.5 * S * E;

  *stress = S;/* RAW need to take into account the stress-type request */

  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
    
  return TRUE;
}

static W_bool_t 
get_stress (W_mat_1d_t                  *m,
            W_mat_1d_state_t             ms,
            W_mat_1d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
            double                      *stress,
               W_mat_3d_stress_type_enum    *out_stress_type
            )
{
  double E, S;
  eng_mat_t *u = (eng_mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_ENG_STRAIN, &E))
    return FALSE;

  S = u->E * E;

  matst->deform_energy_density = 0.5 * S * E;

  *stress = S;/* RAW need to take into account the stress-type request */
    
  return TRUE;
}





static W_mat_1d_state_t 
get_mat_state (W_mat_1d_t *m, W_loc_t *at_point)
{
  matstate_t *ms;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->deform_energy_density = 0;

  return (W_mat_1d_state_t)ms;
}


static W_bool_t 
get_mat_state_var (W_mat_1d_t           *m,
                       W_mat_1d_state_t             ms,
                       W_mat_1d_ip_ctx_t           *ip_ctx,
                       W_mat_1d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_1D_EQUIV_PLASTIC_DEF:
    *var = 0; /* elastic material */
    return TRUE;
  case MAT_1D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}


static W_bool_t
get_stiff_coeff (W_mat_1d_t                  *m,
                 W_mat_1d_state_t             ms, 
                 W_mat_1d_ip_ctx_t           *ip_ctx,
                 double *stiff_coeff
                 )
{
  eng_mat_t *u = (eng_mat_t *)m->data;

  *stiff_coeff = u->E;

  return TRUE;
}


static W_mat_1d_methods_t mat1d_methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  get_stiff_coeff
};


W_mat_1d_t *
read_mat_1d_eng (W_domain_t *d)
{
  const char *input_line;
  double rho = 1;
  eng_mat_t *u;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      u = w_new(eng_mat_t); CHECK_MALLOC(u);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { u->E = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { u->nu = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { u->tensile_strength = TOKEN_DBL_NEXT(1); },
			    { u->tensile_strength = u->E / 100; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      u->dil_wave_speed_1d
        = sqrt (u->E / rho);
      
      return WM1D_new_material (&mat1d_methods, (void *)u, rho);
    }
  }
  return NULL;
}

