#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Wmats.h"
#include "Wmat3d.h"
#include "Wmat1d.h"
#include "Wmat1dgen3d.h"

/* RCSHEADER "$Id: mat1dgen3d.c,v 1.2 2002/03/21 01:32:35 cvs Exp $" */

static int
get_mcode (W_mat_1d_t *m)
{
  return MAT_1D_GEN3D;
}

typedef struct mat_t {
  W_mat_3d_t *m3d;
  double transverse_stress_tol;
}              mat_t;

typedef struct matstate_t {
  W_mat_3d_state_p  matst3d;
  double            transverse_stretch;
}              matstate_t;

static W_bool_t
get_mat_const (W_mat_1d_t *m, W_mat_1d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;
  switch (type) {
  case MAT_1D_YOUNG_MODULUS:
    return WM3D_get_mat_const (u->m3d, MAT_3D_YOUNG_MODULUS, c);
  case MAT_1D_POISSON:
    return WM3D_get_mat_const (u->m3d, MAT_3D_POISSON, c);
  case MAT_1D_DIL_WAVE_SPEED_1D:
    {
      double E, rho;
      WM3D_get_mat_const (u->m3d, MAT_3D_YOUNG_MODULUS, &E);
      *c = sqrt (E / m->vol_mass);
    }
    return TRUE;
  default:
    *c = 0;
    return FALSE;
  }
}
  
typedef struct ip_ctx_data_t {
  double F[3][3];
  double prev_F[3][3];
  double rel_F[3][3];
  double midstep_F[3][3];
  double C[3][3];
  double E[3][3];
  double midstep_D[3][3];
  double midstep_W[3][3];
  double R[3][3];
  double U[3][3];
  double rel_R[3][3];
  double rel_U[3][3];
  double ln_rel_U[3][3];
  double F_dot[3][3];
}              ip_ctx_data_t;

/*
  NOTE: this routine does not cache a validity flag with
  each kinematic tensor as the element routines do.  It should
  be implemented to compute quantities just-in-time if this
  proves a computational bottleneck.
 */
static void
compute_kinem_quantities (W_mat_3d_ip_ctx_t *ctx)
{
  ip_ctx_data_t *ipdata = (ip_ctx_data_t *)ctx->ctx_data;
  double t = ctx->time;
  double dt = ctx->time_step;

  /* Cauchy-Green tensor */
  MAT3DUTIL_MULT_ATB_3X3 (ipdata->C, ipdata->F, ipdata->F);

  /* Green-Lagrange strain tensor */
  {
    double unit[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
    MAT3DUTIL_COMB_AB_3X3 (ipdata->E, 0.5, ipdata->C, - 0.5, unit);
  }

  /* relative deformation gradient */
  {
    double inv_prev_F[3][3];
    double det_prev_F = MAT3DUTIL_DET_3X3(ipdata->prev_F);
    if (det_prev_F <= 0) {
      sprintf(W_buf(), "Failed to invert def grad");
      REPORT_FATAL_ERROR(NULL, W_buf());
    }
    
    MAT3DUTIL_INVERT_3X3(det_prev_F, ipdata->prev_F, inv_prev_F);
    
    MAT3DUTIL_MULT_AB_3X3(ipdata->rel_F, ipdata->F, inv_prev_F);
  }

  /* midstep deformation gradient */
  MAT3DUTIL_COMB_AB_3X3 (ipdata->midstep_F, 0.5, ipdata->F, 0.5, ipdata->prev_F);

  /* midstep rate of deformation and spin */
  {
    int i, j;
    double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
    double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
    double s[3][3], inv_s[3][3], q[3][3], r[3][3];
    double inv_dt = 1 / dt;
    
    MAT3DUTIL_ADD_AB_3X3 (s, m, ipdata->rel_F);
    MAT3DUTIL_SCALE_3X3 (s, 0.5);
    if (!W_mat3dutil_invert_3x3(s, inv_s)) {
      sprintf(W_buf(), "Failed to invert def grad");
      REPORT_FATAL_ERROR(NULL, W_buf());
    }
    MAT3DUTIL_ADD_AB_3X3 (q, ipdata->rel_F, M);
    MAT3DUTIL_MULT_AB_3X3 (r, inv_s, q);
    
    MAT3DUTIL_SYMM  (ipdata->midstep_D, r);
    MAT3DUTIL_ASYMM (ipdata->midstep_W, r);
    
    MAT3DUTIL_SCALE_3X3 (ipdata->midstep_D, inv_dt);
    MAT3DUTIL_SCALE_3X3 (ipdata->midstep_W, inv_dt);
  }

  /* polar decomposition */
  if (!W_polar_decomposition (ipdata->F, ipdata->R, ipdata->U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }

  /* relative polar decomposition */
  if (!W_polar_decomposition (ipdata->rel_F, ipdata->rel_R, ipdata->rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }

  /* logarithm of the relative deformation tensor */
  W_ln_U (ipdata->rel_U, ipdata->ln_rel_U);
}

static W_bool_t 
get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
               W_mat_3d_kinem_var_type_enum  type,
               double kinvar[3][3])
{
  ip_ctx_data_t *ipdata = (ip_ctx_data_t *)ctx->ctx_data;

  compute_kinem_quantities (ctx);

  switch (type) {
  case MAT_3D_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->F);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->F_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->C);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->midstep_W);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->E);
    break;
  case MAT_3D_POLAR_R:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->R);
    break;
  case MAT_3D_POLAR_U:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->U);
    break;
  case MAT_3D_POLAR_REL_R:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->ln_rel_U);
    break;
  default:
    {
      static W_bool_t complained = FALSE;
      if (!complained) {
        sprintf (W_buf(), "Unknown kinematic tensor type: %d",     (int)type);
        REPORT_WARNING (NULL, W_buf());
        complained = TRUE;
      }
    }
    return FALSE;
  }

  return TRUE;
}

static void
setup_deformation_gradients (ip_ctx_data_t *ipdata,
                             double prev_dudX, double prev_stretch,
                             double dudX, double stretch)
{
  MAT3DUTIL_ZERO_3X3 (ipdata->F);
  ipdata->F[0][0] = 1 + dudX;
  ipdata->F[1][1] = ipdata->F[2][2] = stretch;
  MAT3DUTIL_ZERO_3X3 (ipdata->prev_F);
  ipdata->prev_F[0][0] = 1 + prev_dudX;
  ipdata->prev_F[1][1] = ipdata->prev_F[2][2] = prev_stretch;
}

typedef struct ckit_fminbr_client_t {
  W_mat_3d_t               *m;
  W_mat_3d_state_p          matst;
  W_mat_3d_ip_ctx_t        *ip_ctx;
  double                    s_scale;
  W_mat_3d_stress_type_enum stress_type;
}              ckit_fminbr_client_t;

static double
eval_constraint (void *client, double stretch)
{
  ckit_fminbr_client_t *c = (ckit_fminbr_client_t *) client;
  ip_ctx_data_t *ipdata = (ip_ctx_data_t *)c->ip_ctx->ctx_data;
  double stress[3][3];
  W_mat_3d_stress_type_enum out_stress_type;
  double fval;
  /* modify deformation gradient */
  ipdata->F[1][1] = ipdata->F[2][2] = stretch;
  /* temporarily update the material state */
  if (!WM3D_update_stress (c->m, c->matst, W_MAT_STATE_UPDATE_TEMP,
                           c->ip_ctx, c->stress_type, stress, &out_stress_type))
    REPORT_FATAL_ERROR (NULL, "Update failed");
  /* evaluate the zero-in-plane-stress constraint */
  if (c->s_scale == 0) {
    double E, sigma_y;
    if (WM3D_get_mat_const (c->m, MAT_3D_YOUNG_MODULUS, &E)) {
      if (WM3D_get_mat_const (c->m, MAT_3D_INITIAL_YIELD_STRESS, &sigma_y)) {
        c->s_scale = sigma_y;
      } else {
        c->s_scale = fabs (E * (ipdata->F[0][0] - 1));
      }
    } else {
      REPORT_FATAL_ERROR (NULL, "No Young's modulus");
    }
  }
  fval = powi ((stress[1][1] + stress[2][2])/2 / c->s_scale, 2);
  return fval;
}

/* Function updates the stress state, and returns the stress (with proper 
   conversion into the requested type). */  
static W_bool_t 
update_stress (W_mat_1d_t                  *m,
               W_mat_1d_state_t             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_1d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                      *stress,
               W_mat_3d_stress_type_enum    *out_stress_type
               )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  matstate_t input_matst;
  double prev_dudX, dudX;
  double lo, hi;
  const double range_tol = 1e-16, fval_tol = powi (u->transverse_stress_tol, 2);
  ip_ctx_data_t ipdata;
  ckit_fminbr_client_t c;
  double stress3d[3][3];
  W_mat_3d_ip_ctx_t subst_ip_ctx;
  W_mat_3d_stress_type_enum  out_stress_type_1;
  double prev_transverse_stretch = matst->transverse_stretch;
  
  /* save the material state if the update is temporary */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;
  
  /* get kinematic variables */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_PREV_DISPL_GRAD, &prev_dudX))  
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_DISPL_GRAD, &dudX))
    return FALSE;

  /* set up substitute kinematic context */
  subst_ip_ctx.time          = ip_ctx->time;
  subst_ip_ctx.time_step     = ip_ctx->time_step;
  subst_ip_ctx.ctx_data      = &ipdata;
  subst_ip_ctx.get_kinem_var = get_kinem_var;
  setup_deformation_gradients (&ipdata, prev_dudX, prev_transverse_stretch,
                               dudX, matst->transverse_stretch);
  c.m = u->m3d;
  c.matst = matst->matst3d;
  c.ip_ctx = &subst_ip_ctx;
  c.s_scale = 0;
  if (ipdata.F[0][0] >= 1) { lo = 2 - ipdata.F[0][0]; hi = 1; }
  else                     { lo = 1; hi = 2 - ipdata.F[0][0]; }
  /* solve for the transverse stretch */
  lo -= 100*range_tol;
  hi += 100*range_tol;
  matst->transverse_stretch = ckit_fminbr (lo, hi,
                                           &c, eval_constraint,
                                           range_tol, fval_tol);
#if 0
  printf ("###############################################################\n");
  printf ("fval_tol=%g, lo=%g, hi=%g, stretch=%g\n", fval_tol, lo, hi, matst->transverse_stretch);
#endif

  setup_deformation_gradients (&ipdata, prev_dudX, prev_transverse_stretch,
                               dudX, matst->transverse_stretch);
  if (!WM3D_update_stress (u->m3d, matst->matst3d, update_form,
                           &subst_ip_ctx, MAT_3D_CAUCHY, stress3d, &out_stress_type_1))
    REPORT_FATAL_ERROR (NULL, "Update failed");
  if (out_stress_type_1 != MAT_3D_CAUCHY)
    REPORT_FATAL_ERROR (NULL, "Update failed");
  
  /* RAW need to transform according to the stress type */
#if 0
  printf ("sigma=%g, transv=%g\n", stress3d[0][0], stress3d[1][1]);
  printf ("###############################################################\n");
#endif
  *stress = stress3d[0][0];

  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
    
  return TRUE;
}

static W_bool_t 
get_stress (W_mat_1d_t                  *m,
            W_mat_1d_state_t             ms,
            W_mat_1d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
            double                      *stress,
              W_mat_3d_stress_type_enum    *out_stress_type
             )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double prev_dudX, dudX;
  double lo, hi;
  ip_ctx_data_t ipdata;
  double stress3d[3][3];
  W_mat_3d_ip_ctx_t subst_ip_ctx;
  W_mat_3d_stress_type_enum  out_stress_type_1;
  
  /* get kinematic variables */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_PREV_DISPL_GRAD, &prev_dudX))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_DISPL_GRAD, &dudX))
    return FALSE;

  /* set up substitute kinematic context */
  subst_ip_ctx.time          = ip_ctx->time;
  subst_ip_ctx.time_step     = ip_ctx->time_step;
  subst_ip_ctx.ctx_data      = &ipdata;
  subst_ip_ctx.get_kinem_var = get_kinem_var;
  setup_deformation_gradients (&ipdata, prev_dudX, matst->transverse_stretch,
                               dudX, matst->transverse_stretch);
  if (!WM3D_update_stress (u->m3d, matst->matst3d, W_MAT_STATE_UPDATE_TEMP,
                           &subst_ip_ctx, MAT_3D_CAUCHY, stress3d, &out_stress_type_1))
    REPORT_FATAL_ERROR (NULL, "Update failed");
  if (out_stress_type_1 != MAT_3D_CAUCHY)
    REPORT_FATAL_ERROR (NULL, "Update failed");
  
  /* RAW need to transform according to the stress type */
  *stress = stress3d[0][0];
    
  return TRUE;
}

static W_mat_1d_state_t 
get_mat_state (W_mat_1d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->matst3d = WM3D_get_mat_state (u->m3d, at_point);
  ms->transverse_stretch = 1;

  return (W_mat_1d_state_t)ms;
}

static W_bool_t 
get_mat_state_var (W_mat_1d_t           *m,
                       W_mat_1d_state_t             ms,
                       W_mat_1d_ip_ctx_t           *ip_ctx,
                       W_mat_1d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double dudX, prev_dudX;
  ip_ctx_data_t ipdata;
  W_mat_3d_ip_ctx_t subst_ip_ctx;

  /* get kinematic variables */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_PREV_DISPL_GRAD, &prev_dudX))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_DISPL_GRAD, &dudX))
    return FALSE;
  /* set up substitute kinematic context */
  subst_ip_ctx.time          = ip_ctx->time;
  subst_ip_ctx.time_step     = ip_ctx->time_step;
  subst_ip_ctx.ctx_data      = &ipdata;
  subst_ip_ctx.get_kinem_var = get_kinem_var;
  setup_deformation_gradients (&ipdata, prev_dudX, matst->transverse_stretch,
                               dudX, matst->transverse_stretch);

  switch(var_type) {
  case MAT_1D_EQUIV_PLASTIC_DEF:
    return WM3D_get_mat_state_var (u->m3d, matst->matst3d, &subst_ip_ctx,
                                   MAT_3D_EQUIV_PLASTIC_DEF, var);
  case MAT_1D_DEFORM_ENERGY_DENSITY:
    return WM3D_get_mat_state_var (u->m3d, matst->matst3d, &subst_ip_ctx,
                                   MAT_3D_DEFORM_ENERGY_DENSITY, var);
  default:
    return FALSE;
  }
}

static W_bool_t
get_stiff_coeff (W_mat_1d_t                  *m,
                 W_mat_1d_state_t             ms, 
                 W_mat_1d_ip_ctx_t           *ip_ctx,
                 double *stiff_coeff
                 )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  int success;
  double C[6][6];
  double dudX, prev_dudX;
  ip_ctx_data_t ipdata;
  W_mat_3d_ip_ctx_t subst_ip_ctx;

  /* get kinematic variables */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_PREV_DISPL_GRAD, &prev_dudX))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_DISPL_GRAD, &dudX))
    return FALSE;
  /* set up substitute kinematic context */
  subst_ip_ctx.time          = ip_ctx->time;
  subst_ip_ctx.time_step     = ip_ctx->time_step;
  subst_ip_ctx.ctx_data      = &ipdata;
  subst_ip_ctx.get_kinem_var = get_kinem_var;
  setup_deformation_gradients (&ipdata, prev_dudX, matst->transverse_stretch,
                               dudX, matst->transverse_stretch);

  success = WM3D_get_mat_stiff (u->m3d, matst->matst3d, &subst_ip_ctx,
                               MAT_3D_STIFF_EULERIAN, C);
  if (success) {
    *stiff_coeff = C[0][0];
    return TRUE;
  } else
    return FALSE;
}


static W_mat_1d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  get_stiff_coeff
};


W_mat_1d_t *
read_mat_1d_gen3d (W_domain_t *d)
{
  const char *input_line;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      mat_t *u;
      W_material_t *m;
      double transverse_stress_tol = 0.001;
      int matid = 0;
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("MAT3D", TRUE,
			    { matid = TOKEN_INT_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("TRANSVERSE_STRESS_TOL", FALSE,
			    { transverse_stress_tol = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      u = w_new(mat_t); CHECK_MALLOC(u);
      m = WD_get_material_by_idx (d, matid);
      if (m == NULL) {
        sprintf (W_buf (), "No such material %d", matid);
        REPORT_FATAL_ERROR (d->errh, W_buf ());
      }
      u->m3d = (W_mat_3d_t *)m->data;
      u->transverse_stress_tol = transverse_stress_tol;
      return WM1D_new_material (&methods, (void *)u, WM3D_get_volume_mass (u->m3d));
    }
  }
  return NULL;
}

