 
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat1d.h"
#include "Wmat1dj2pl.h"

/* RCSHEADER "$Id: mat1dj2pl.c,v 1.2 2002/03/21 01:32:35 cvs Exp $" */

#define SQRT_2_OVER_3 0.81649658092772603272

static int
get_mcode (W_mat_1d_t *m)
{
    return MAT_1D_J2PL;
}


typedef struct mat_t {
  double E, nu;
  double dil_wave_speed_1d;
  double initial_yield_stress;
  double K; /* kinematic hardening modulus */
  double H; /* isotropic hardening modulus */
}              mat_t;


typedef struct matstate_t {
  double   sigma_y; /* current yield stress */
  double   eps_pl;  /* plastic strain */
  double   equiv_pl_def; /* internal variable (equivalent plastic deformation) */
  double   alpha;   /* back-stress */
  double   deform_energy_density;
}              matstate_t;


static W_bool_t
get_mat_const (W_mat_1d_t *m, W_mat_1d_mat_const_enum type, double *c)
{
    mat_t *u = (mat_t *)m->data;

  switch (type) {
  case MAT_1D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_1D_POISSON:
    *c = u->nu;
    break;
  case MAT_1D_TENSILE_STRENGTH:
    *c = 0;
    break;
  case MAT_1D_DIL_WAVE_SPEED_1D:
    *c = u->dil_wave_speed_1d;
    break;
  case MAT_1D_KINEM_HARD_MOD:
    *c = u->K;
    break;
  case MAT_1D_ISOTR_HARD_MOD:
    *c = u->H;
    break;
  case MAT_1D_YIELD_STRESS:
    *c = u->initial_yield_stress;
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

  
/* Function updates the stress state, and returns the stress (with proper */
/* conversion into the requested type).  Note that the function get_stress() */
/* is identical with this function since no stress state is saved for this */
/* type of elastic material. */
static W_bool_t 
update_stress (W_mat_1d_t                  *m,
               W_mat_1d_state_t             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_1d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                      *stress,
               W_mat_3d_stress_type_enum    *out_stress_type
               )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  matstate_t input_matst;
  double K = u->K, H = u->H, E = u->E;
  double eps, eps_pl = matst->eps_pl; 
  double sigma_y = matst->sigma_y;
  double alpha = matst->alpha;
  double sigma, sigma_trial, ksi_trial, f_trial;

  /* save the material state if the update is temporary */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;
  
  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_ENG_STRAIN, &eps))
    return FALSE;

  sigma_trial = E * (eps - eps_pl); /* trial stress */
  ksi_trial   = sigma_trial - alpha; /* trial relative stress */
  f_trial     = fabs (ksi_trial) - sigma_y;

  if (f_trial <= 0) {           /* elastic step */
    sigma = sigma_trial;
  } else {
    double dgamma = f_trial / (E + K + H);
    int sign_ksi_trial = ksi_trial >= 0 ? +1 : -1;
    sigma = sigma_trial - dgamma * E * sign_ksi_trial;
    matst->eps_pl       += dgamma * sign_ksi_trial;
    matst->alpha        += dgamma * H * sign_ksi_trial;
    matst->sigma_y      += dgamma * K;
    matst->equiv_pl_def += SQRT_2_OVER_3 * dgamma;
  }

  matst->deform_energy_density = 0.5 * sigma * (eps - matst->eps_pl);

  *stress = sigma;/* RAW need to take into account the stress-type request */

  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
    
  return TRUE;
}

static W_bool_t 
get_stress (W_mat_1d_t                  *m,
            W_mat_1d_state_t             ms,
            W_mat_1d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum stress_type,
            double                       *stress,
               W_mat_3d_stress_type_enum *out_stress_type
            )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double sigma, eps;

  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_ENG_STRAIN, &eps))
    return FALSE;

  sigma = u->E * (eps - matst->eps_pl); 

  matst->deform_energy_density = 0.5 * sigma * (eps - matst->eps_pl);

  *stress = sigma; /* RAW need to take into account the stress-type request */
    
  return TRUE;
}

static W_mat_1d_state_t 
get_mat_state (W_mat_1d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->sigma_y               = u->initial_yield_stress;
  ms->eps_pl                = 0;
  ms->equiv_pl_def          = 0;
  ms->alpha                 = 0;
  ms->deform_energy_density = 0;

  return (W_mat_1d_state_t)ms;
}

static W_bool_t 
get_mat_state_var (W_mat_1d_t           *m,
                       W_mat_1d_state_t             ms,
                       W_mat_1d_ip_ctx_t           *ip_ctx,
                       W_mat_1d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_1D_EQUIV_PLASTIC_DEF:
    *var = matst->equiv_pl_def;
    return TRUE;
  case MAT_1D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}

static W_bool_t
get_stiff_coeff (W_mat_1d_t                  *m,
                 W_mat_1d_state_t             ms, 
                 W_mat_1d_ip_ctx_t           *ip_ctx,
                 double *stiff_coeff
                 )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double eps_pl = matst->eps_pl;
  double alpha = matst->alpha;
  double sigma_y = matst->sigma_y;
  double eps, sigma, sigma_trial, ksi_trial, f_trial;

  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_1D_ENG_STRAIN, &eps))
    return FALSE;

  sigma_trial = u->E * (eps - eps_pl); /* trial stress */
  ksi_trial   = sigma_trial - alpha; /* trial relative stress */
  f_trial     = fabs (ksi_trial) - sigma_y;

  if (f_trial <= 0) { /* elastic step */
    *stiff_coeff = u->E;
  } else {            /* plastic step */
    *stiff_coeff = u->E * (u->K + u->H) / (u->E + u->K + u->H);
  }

  return TRUE;
}


static W_mat_1d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  get_stiff_coeff
};


W_mat_1d_t *
read_mat_1d_j2pl (W_domain_t *d)
{
  const char *input_line;
  double rho = 1;
  mat_t *u;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      u = w_new(mat_t); CHECK_MALLOC(u);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { u->E = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { u->nu = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("SIGMA_Y", TRUE,
			    { u->initial_yield_stress = TOKEN_DBL_NEXT(1); },
			    { u->initial_yield_stress = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("K", FALSE,
			    { u->K = TOKEN_DBL_NEXT(1); },
			    { u->K = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("H", FALSE,
			    { u->H = TOKEN_DBL_NEXT(1); },
			    { u->H = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      u->dil_wave_speed_1d
        = sqrt (u->E / rho);
      
      return WM1D_new_material (&methods, (void *)u, rho);
    }
  }
  return NULL;
}

