#include "Wdomain.h"
#include "Wmat3dckinks.h"
    
#define N_SLIP_SYSTEMS 2
    
#define ISQRT2 0.70710678118654752440
#define ISQRT3 0.57735026918962576451
#define S(x,y,z) x,y,z
#define M(x,y,z) x,y,z
    
/* slip directions expressed in the a,b,c vectors */
static double ref_s[N_SLIP_SYSTEMS][3] = {
  {S(0,1,0)},   /* B2 */
  {S(0,0,1)}   /* B4 */
};
/* slip plane normals expressed in the a,b,c vectors */
static double ref_m[N_SLIP_SYSTEMS][3] = {
  {M(1,0,0)},   /* B2 */
  {M(1,0,0)}   /* B4 */
};
    
static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_CKINKS;
}
    
#include "hardening.h"

/* RCSHEADER "$Id: mat3dckinks.c,v 1.3 2004/04/27 20:10:10 cvs Exp $" */
  
typedef struct mat_t {
  W_domain_t *d;
  double c11, c44, c12; /* elastic relation S=C:E for a cubic symmetry */
  char *a[3];
  char *b[3]; /* orthogonal unit vectors along the principal cubic axes */
  double gamma_0; /* reference shear strain rate */
  double q; /* hardening behaviour parameter; q=1 .. isotropic (Taylor);
                      1 <= q <= 1.4 Kocks (1970) */
  double h_0; /* initial hardening modulus */
  double m_srs; /* strain-rate sensitivity exponent */
  double g_0; /* initial flow stress */
  double g_inf; /* saturation flow stress [Asaro (1983)] */
  double s[N_SLIP_SYSTEMS][3]; /* slip directions */
  double m[N_SLIP_SYSTEMS][3]; /* normals to slip planes */
  int max_inner_iter; /* max number of iterations allowed in the inner loop */
  double activity_tol; /* tolerance under which systems are considered inactive */
  int debug; /* debug level */
  double e_incr; /* increment of the Green-Lagrange tensor for the numerical
                    differentiation to obtain the constitutive stiffness matrix */
  W_bool_t elastic; /* this material is elastic; no plasticity is allowed */
  hardening_solver_t *hsi;

  double G; /* shear modulus */
  double J; /*coef in athermal dislocation storage term in the evolution of dislocation density*/
  double K; /*coef in thermally activated recovery term in the evolution of dislocation density*/ 
  double mb; /* magnitude of Burgers vector */
  double const_a; /* constant a in the expression for yield stress */
                  /* a small imperfection can be made by reducing const_a by 25% */
  double const_k; /* constant k in the expression for reference shear strain rate gamma_0 */
  /* do we need to add initial dislocation density here */
}              mat_t; 

typedef struct matstate_t {
  double   S[6];           /* stores current 2nd P-K stress: */
  double   tau[N_SLIP_SYSTEMS]; /* resolved shear stress on all slip systems */
  double   Fp[3][3]; /* current plastic deformation gradient */
  double   g[N_SLIP_SYSTEMS]; /* current flow stress on a slip system */
  double   gamma[N_SLIP_SYSTEMS]; /* current shear strain on a slip system */
  double   dgamma[N_SLIP_SYSTEMS]; /* current shear strain increment on a slip system */
  double   accum_gamma; /* accumulated plastic slip on all slip systems */
  double   deform_energy_density; /* deformation energy density */
  W_vect_t a, b, c;
}              matstate_t;


#define SIGN(a) (a > 0? +1 : -1)

static W_bool_t
get_mat_const(W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;
  double lambda = u->c12; /* this is a very crude approximation: only true for c11-c12=2*c44 */
  double G = 2 * u->c44; /* faked: this only makes sense for isotropic materials */
  double nu = lambda / 2 / (lambda + G);
  double E = 2 * (1 + nu) * G;
  double bulk_mod = E / 3 / (1 - 2*nu);

  switch (type) { 
  case MAT_3D_YOUNG_MODULUS:
    *c = E;
    break;
  case MAT_3D_POISSON:
    *c = nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = bulk_mod;
    break;
  default: /* quantities that are not supported */
    *c = 0;
    return FALSE;
  }
  return TRUE;
}


static W_mat_3d_state_p 
get_mat_state(W_mat_3d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;
  int i;
  int stat;
  double p;
  const double eps = 1e-9;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

    expeval_set_value("x", at_point->x);
    expeval_set_value("y", at_point->y);
    expeval_set_value("z", at_point->z);

    stat = expeval_eval (u->a[0], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(u->d->errh, buf);
      p = 0;
    } 
  ms->a.x = p;
    stat = expeval_eval (u->a[1], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(u->d->errh, buf);
      p = 0;
    } 
  ms->a.y = p;
    stat = expeval_eval (u->a[2], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(u->d->errh, buf);
      p = 0;
    } 
  ms->a.z = p;
  W_normalize_vect_3d (&ms->a); 

    stat = expeval_eval (u->b[0], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(u->d->errh, buf);
      p = 0;
    } 
  ms->b.x = p;
    stat = expeval_eval (u->b[1], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(u->d->errh, buf);
      p = 0;
    } 
  ms->b.y = p;
    stat = expeval_eval (u->b[2], &p);
    if (stat != EXPEVAL_E_OK) {
      char buf[1024];
      sprintf(buf, "%s", expeval_err_msgs[stat]);
      REPORT_WARNING(u->d->errh, buf);
      p = 0;
    } 
  ms->b.z = p;
  W_normalize_vect_3d (&ms->b); 

/* c-vector from cross product of ms->a and ms->b */
  ms->c.x = ms->a.y * ms->b.z - ms->a.z * ms->b.y;
  ms->c.y = ms->a.z * ms->b.x - ms->a.x * ms->b.z;
  ms->c.z = ms->a.x * ms->b.y - ms->a.y * ms->b.x; 
  W_normalize_vect_3d (&ms->c); 

  if (W_dot_prod_3d (&ms->a, &ms->b) > eps) {
  	REPORT_FATAL_ERROR (NULL, "Non-orthogonal orientation vectors");
  }
  if (W_dot_prod_3d (&ms->c, &ms->b) > eps) {
  	REPORT_FATAL_ERROR (NULL, "Non-orthogonal orientation vectors");
  }
  if (W_dot_prod_3d (&ms->c, &ms->a) > eps) {
  	REPORT_FATAL_ERROR (NULL, "Non-orthogonal orientation vectors");
  }
  if (fabs (W_vector_3d_length (&ms->a) - 1) > eps) {
  	REPORT_FATAL_ERROR (NULL, "Non-u orientation vector");
  }
  if (fabs (W_vector_3d_length (&ms->b) - 1) > eps) {
  	REPORT_FATAL_ERROR (NULL, "Non-unit orientation vector");
  }
  if (fabs (W_vector_3d_length (&ms->c) - 1) > eps) {
  	REPORT_FATAL_ERROR (NULL, "Non-unit orientation vector");
  }
    
  FOR (i, 6) ms->S[i] = 0;
  MAT3DUTIL_UNIT_3X3 (ms->Fp);
  FOR (i, N_SLIP_SYSTEMS) {
    ms->g[i] = u->g_0;
    ms->gamma[i] = 0;
    ms->dgamma[i] = 0;
  }
  ms->accum_gamma = 0;
  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_stress (W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           );

static void
calc_Dbar (double c11, double c12, double c44,
           W_vect_t *a, W_vect_t *b, W_vect_t *c,
           double Dbar[6][6]
           )
{
  double a1 = (*a).x, a2 = (*a).y, a3 = (*a).z;
  double b1 = (*b).x, b2 = (*b).y, b3 = (*b).z;
  double c1 = (*c).x, c2 = (*c).y, c3 = (*c).z;
  double c11c122c44 = (c11 - c12 - 2*c44);
  
  Dbar[0][0] = c12 +  2*c44 + c11c122c44 * ((a1*a1*a1*a1 + b1*b1*b1*b1 + c1*c1*c1*c1));
  Dbar[0][1] = Dbar[1][0] = c12 +   c11c122c44 * ((a1*a1*a2*a2 + b1*b1*b2*b2 + c1*c1*c2*c2));
  Dbar[0][2] = Dbar[2][0] = c12 +   c11c122c44 * ((a1*a1*a3*a3 + b1*b1*b3*b3 + c1*c1*c3*c3));
  Dbar[0][3] = Dbar[3][0] = ((  c11c122c44 * ((a1*a1*a1*a2 + b1*b1*b1*b2 + c1*c1*c1*c2))) + (  c11c122c44 * ((a1*a1*a2*a1 + b1*b1*b2*b1 + c1*c1*c2*c1))))/2;
  Dbar[0][4] = Dbar[4][0] = ((  c11c122c44 * ((a1*a1*a1*a3 + b1*b1*b1*b3 + c1*c1*c1*c3))) + (  c11c122c44 * ((a1*a1*a3*a1 + b1*b1*b3*b1 + c1*c1*c3*c1))))/2;
  Dbar[0][5] = Dbar[5][0] = ((  c11c122c44 * ((a1*a1*a2*a3 + b1*b1*b2*b3 + c1*c1*c2*c3))) + (  c11c122c44 * ((a1*a1*a3*a2 + b1*b1*b3*b2 + c1*c1*c3*c2))))/2;
  Dbar[1][1] = c12 +  2*c44 + c11c122c44 * ((a2*a2*a2*a2 + b2*b2*b2*b2 + c2*c2*c2*c2));
  Dbar[1][2] = Dbar[2][1] = c12 +   c11c122c44 * ((a2*a2*a3*a3 + b2*b2*b3*b3 + c2*c2*c3*c3));
  Dbar[1][3] = Dbar[3][1] = ((  c11c122c44 * ((a2*a2*a1*a2 + b2*b2*b1*b2 + c2*c2*c1*c2))) + (  c11c122c44 * ((a2*a2*a2*a1 + b2*b2*b2*b1 + c2*c2*c2*c1))))/2;
  Dbar[1][4] = Dbar[4][1] = ((  c11c122c44 * ((a2*a2*a1*a3 + b2*b2*b1*b3 + c2*c2*c1*c3))) + (  c11c122c44 * ((a2*a2*a3*a1 + b2*b2*b3*b1 + c2*c2*c3*c1))))/2;
  Dbar[1][5] = Dbar[5][1] = ((  c11c122c44 * ((a2*a2*a2*a3 + b2*b2*b2*b3 + c2*c2*c2*c3))) + (  c11c122c44 * ((a2*a2*a3*a2 + b2*b2*b3*b2 + c2*c2*c3*c2))))/2;
  Dbar[2][2] = c12 +  2*c44 + c11c122c44 * ((a3*a3*a3*a3 + b3*b3*b3*b3 + c3*c3*c3*c3));
  Dbar[2][3] = Dbar[3][2] = ((  c11c122c44 * ((a3*a3*a1*a2 + b3*b3*b1*b2 + c3*c3*c1*c2))) + (  c11c122c44 * ((a3*a3*a2*a1 + b3*b3*b2*b1 + c3*c3*c2*c1))))/2;
  Dbar[2][4] = Dbar[4][2] = ((  c11c122c44 * ((a3*a3*a1*a3 + b3*b3*b1*b3 + c3*c3*c1*c3))) + (  c11c122c44 * ((a3*a3*a3*a1 + b3*b3*b3*b1 + c3*c3*c3*c1))))/2;
  Dbar[2][5] = Dbar[5][2] = ((  c11c122c44 * ((a3*a3*a2*a3 + b3*b3*b2*b3 + c3*c3*c2*c3))) + (  c11c122c44 * ((a3*a3*a3*a2 + b3*b3*b3*b2 + c3*c3*c3*c2))))/2;
  Dbar[3][3] = (( 2*c44*(0.5) + c11c122c44 * ((a1*a2*a1*a2 + b1*b2*b1*b2 + c1*c2*c1*c2))) + ( 2*c44*(0.5) + c11c122c44 * ((a1*a2*a2*a1 + b1*b2*b2*b1 + c1*c2*c2*c1))))/2;
  Dbar[3][4] = Dbar[4][3] = ((  c11c122c44 * ((a1*a2*a1*a3 + b1*b2*b1*b3 + c1*c2*c1*c3))) + (  c11c122c44 * ((a1*a2*a3*a1 + b1*b2*b3*b1 + c1*c2*c3*c1))))/2;
  Dbar[3][5] = Dbar[5][3] = ((  c11c122c44 * ((a1*a2*a2*a3 + b1*b2*b2*b3 + c1*c2*c2*c3))) + (  c11c122c44 * ((a1*a2*a3*a2 + b1*b2*b3*b2 + c1*c2*c3*c2))))/2;
  Dbar[4][4] = (( 2*c44*(0.5) + c11c122c44 * ((a1*a3*a1*a3 + b1*b3*b1*b3 + c1*c3*c1*c3))) + ( 2*c44*(0.5) + c11c122c44 * ((a1*a3*a3*a1 + b1*b3*b3*b1 + c1*c3*c3*c1))))/2;
  Dbar[4][5] = Dbar[5][4] = ((  c11c122c44 * ((a1*a3*a2*a3 + b1*b3*b2*b3 + c1*c3*c2*c3))) + (  c11c122c44 * ((a1*a3*a3*a2 + b1*b3*b3*b2 + c1*c3*c3*c2))))/2;
  Dbar[5][5] = (( 2*c44*(0.5) + c11c122c44 * ((a2*a3*a2*a3 + b2*b3*b2*b3 + c2*c3*c2*c3))) + ( 2*c44*(0.5) + c11c122c44 * ((a2*a3*a3*a2 + b2*b3*b3*b2 + c2*c3*c3*c2))))/2;
}

#define VOP1(dest, op1, a, numel)               \
{                                               \
  int _j__, _N__ = numel;                       \
  FOR (_j__, _N__) {                            \
    dest[_j__] op1 a[_j__];                     \
  }                                             \
}
#define VOP1C(dest, op1, C, numel)               \
{                                               \
  int _j__, _N__ = numel;                       \
  FOR (_j__, _N__) {                            \
    dest[_j__] op1 C;                     \
  }                                             \
}
#define VOP2(dest, op1, a, op2, b, numel)       \
{                                               \
  int _j__, _N__ = numel;                       \
  FOR (_j__, _N__) {                            \
    dest[_j__] op1 a[_j__] op2 b[_j__];         \
  }                                             \
}
#define MOP1(dest, op1, a, numel)               \
{                                               \
  int _i__, _j__, _N__ = numel;                 \
  FOR (_i__, _N__) {                            \
     FOR (_j__, _N__) {                         \
       dest[_i__][_j__] op1 a[_i__][_j__];      \
     }                                          \
  }                                             \
}
#define MOP2(dest, op1, a, op2, b, numel)                       \
{                                                               \
  int _i__, _j__, _N__ = numel;                                 \
  FOR (_i__, _N__) {                                            \
     FOR (_j__, _N__) {                                         \
       dest[_i__][_j__] op1 a[_i__][_j__] op2 b[_i__][_j_];     \
     }                                                          \
  }                                                             \
}

static void
calc_Cebar1 (double F1[3][3], /* F_n+1   total deformation gradient at t_n+1*/
             double Fp1[3][3], /* F^p_n+1  current plastic deformation gradient at t_n+1 */
             double Cebar1[3][3] /* F^p_n+1  current plastic deformation gradient at t_n+1 */
             )
{
  double Fe1[3][3];
  double invFp1[3][3];
  double detFp1 = MAT3DUTIL_DET_3X3 (Fp1);
  if (detFp1 == 0) {
    REPORT_FATAL_ERROR (NULL, "Zero det(Fp1)");
  }
  MAT3DUTIL_INVERT_3X3 (detFp1, Fp1, invFp1);
  MAT3DUTIL_MULT_AB_3X3 (Fe1, F1, invFp1);
  MAT3DUTIL_MULT_ATB_3X3 (Cebar1, Fe1, Fe1);
}

static void
calc_Sbar (double Dbar[6][6],/* Sbar=Dbar:Ebar: 2nd PK stress on the intermediate configuration  */
        double Cebar1[3][3],/* elastic Cauchy-Green tensor */
        double Fp1[3][3], /* F^p_n+1  current plastic deformation gradient at t_n+1 */
        double Sbar[3][3] /* OUT: PK2 stress tensor */
        )
{
  double E[3][3];
  double Ev[6];
  double Sv[6];
  MAT3DUTIL_2_MAT (E, = 0.5 *, Cebar1);
  MAT3DUTIL_MAT_MOD_DIAG (E, -=, 0.5);
  Ev[0] = E[0][0];
  Ev[1] = E[1][1];
  Ev[2] = E[2][2];
  Ev[3] = E[0][1]+E[1][0];
  Ev[4] = E[0][2]+E[2][0];
  Ev[5] = E[2][1]+E[1][2];
  MAT3DUTIL_S_C_E (Sv,Dbar,Ev);
  MAT3DUTIL_V_TO_T (Sv, Sbar);
}
      
static void
calc_S (double Sbar[3][3], double Fp[3][3], double S[3][3])
{
  double t[3][3], inv_Fp[3][3];
  double det_Fp = MAT3DUTIL_DET_3X3 (Fp);
  MAT3DUTIL_INVERT_3X3 (det_Fp, Fp, inv_Fp);
  MAT3DUTIL_MULT_ABT_3X3 (t, Sbar, inv_Fp);
  MAT3DUTIL_MULT_AB_3X3  (S, inv_Fp, t);
}

static double
l2_norm_squared (double f[N_SLIP_SYSTEMS])
{
  double d = 0;
  int i;
  FOR (i, N_SLIP_SYSTEMS) { d += (f[i])*(f[i]); }
  return d;
}

static double
linf_norm (double f[N_SLIP_SYSTEMS])
{
  double d = 0;
  int i;
  FOR (i, N_SLIP_SYSTEMS) { if ( fabs(f[i]) > d ) d = fabs(f[i]); }
  return d;
}

static void
calc_Fp1 (double Fp[3][3],
          double dgamma[N_SLIP_SYSTEMS],
          double s[N_SLIP_SYSTEMS][3],
          double m[N_SLIP_SYSTEMS][3], double Fp1[3][3])
{
#define IMPL_FP 1
#if IMPL_FP /* Implicit time stepping of the Fp */
  int i;
  double Id_dgsm_sum[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
  double inv_Id_dgsm_sum[3][3];
  FOR (i, N_SLIP_SYSTEMS) {
    double dgsm[3][3], minus_dg =  - dgamma[i];
    MAT3DUTIL_TENS_PROD_V (dgsm, s[i], m[i]);
    MAT3DUTIL_SCALE_3X3 (dgsm, minus_dg);
    MAT3DUTIL_INCR_A_BY_B_3X3 (Id_dgsm_sum, dgsm);
  }
  MAT3DUTIL_MAT_MOD_DIAG (Id_dgsm_sum, +=, 1.0);
  if (!W_mat3dutil_invert_3x3 (Id_dgsm_sum, inv_Id_dgsm_sum)) {
    REPORT_FATAL_ERROR (NULL, "Id_dgsm_sum is singular");
  }
  MAT3DUTIL_MULT_AB_3X3 (Fp1, inv_Id_dgsm_sum, Fp);
#else /* Explicit time stepping of the Fp */
  int i;
  double Id_dgsm_sum[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
  FOR (i, N_SLIP_SYSTEMS) {
    double dgsm[3][3], minus_dg =  dgamma[i];
    MAT3DUTIL_TENS_PROD_V (dgsm, s[i], m[i]);
    MAT3DUTIL_SCALE_3X3 (dgsm, minus_dg);
    MAT3DUTIL_INCR_A_BY_B_3X3 (Id_dgsm_sum, dgsm);
  }
  MAT3DUTIL_MAT_MOD_DIAG (Id_dgsm_sum, +=, 1.0);
  MAT3DUTIL_MULT_AB_3X3 (Fp1, Id_dgsm_sum, Fp);
#endif
}

static void
calc_tau1 (double Cebar1[3][3],
           double S1[3][3],
           double s[N_SLIP_SYSTEMS][3], double m[N_SLIP_SYSTEMS][3],
           double tau1[N_SLIP_SYSTEMS] /* resolved shear stress */
           )
{
  int i;
  double Cebar1s[3];
  double S1m[3];
  FOR (i, N_SLIP_SYSTEMS) {
    MAT3DUTIL_MULT_A_3X3_V_3 (Cebar1s, Cebar1, s[i]);
    MAT3DUTIL_MULT_A_3X3_V_3 (S1m, S1, m[i]);
    tau1[i] = MAT3DUTIL_DOT_V (Cebar1s, S1m);
  }
}


static void
calc_g1 (double q, double h_0, double g_0, double g_inf,
         double g[N_SLIP_SYSTEMS],
         double accum_gamma, 
         double dgamma[N_SLIP_SYSTEMS],
         double g1[N_SLIP_SYSTEMS])
{
#define DELTA(i,j) (i==j?1:0)
  int i, j;
  double hgamma = 0, accum_gamma1 = accum_gamma, ch;
  FOR (i, N_SLIP_SYSTEMS) { accum_gamma1 += fabs(dgamma[i]); }
  ch = cosh (h_0 * accum_gamma1 / (g_inf - g_0));
  hgamma = h_0 / (ch * ch); /* = sech^2(h_0*gamma_sum/(tau_s - tau_0)) */
  FOR (i, N_SLIP_SYSTEMS) {
    double dg = 0;
    FOR (j, N_SLIP_SYSTEMS) {
      double h1ij = hgamma * (q + DELTA(i,j)*(1-q));
      dg += h1ij * fabs(dgamma[j]);
    }
    g1[i] = g[i] + dg;
  }
}

/*
static void
calc_g1 (double q, double h_0, double g_0, double g_inf,
         double g[N_SLIP_SYSTEMS],
         double accum_gamma,
         double dgamma[N_SLIP_SYSTEMS],
         double g1[N_SLIP_SYSTEMS])
{
  int i;
  FOR (i, N_SLIP_SYSTEMS) {
    g1[i] = g[i];
  }
}
*/

static void
calc_R (double q, double h_0, double g_0, double g_inf, double m_srs, double gamma_0,
        double Dbar[6][6],  double F1[3][3],
        double Fp[3][3],  double gamma[N_SLIP_SYSTEMS], double accum_gamma,
        double g[N_SLIP_SYSTEMS],  double s[N_SLIP_SYSTEMS][3],
        double m[N_SLIP_SYSTEMS][3],
        double dt,  double dgamma[N_SLIP_SYSTEMS],
        W_bool_t active[N_SLIP_SYSTEMS],
        double R[N_SLIP_SYSTEMS]
        )
{
  double Fp1[3][3];
  double Cebar1[3][3]; /* Green-Lagrange strain */
  double Sbar1[3][3]; /* 2nd PK */
  double h1[N_SLIP_SYSTEMS][N_SLIP_SYSTEMS]; /* current hardening matrix */
  double g1[N_SLIP_SYSTEMS];
  double tau1[N_SLIP_SYSTEMS];
  int i;
  calc_g1 (q, h_0, g_0, g_inf, g, accum_gamma, dgamma, g1);
  calc_Fp1 (Fp, dgamma, s, m, Fp1); 
  calc_Cebar1 (F1, Fp1, Cebar1);
  calc_Sbar (Dbar, Cebar1, Fp1, Sbar1);
  calc_tau1 (Cebar1, Sbar1, s, m, tau1);

/******
    fprintf(stderr, "calc_R:\n");
    fprintf (stderr, "tau = [ ");                                   
    fprintf (stderr, "%g %g, %g %g ", tau1[0], tau1[1], tau1[6], tau1[7]); 
    fprintf (stderr, "]\n");                                            

    fprintf (stderr, "g = [ ");                                   
    fprintf (stderr, "%g %g, %g %g ", g1[0], g1[1], g1[6], g1[7]); 
    fprintf (stderr, "]\n");                                            
*******/

  FOR (i, N_SLIP_SYSTEMS) {
    if (active[i]) {
      R[i] = SIGN(dgamma[i]) * pow(fabs(dgamma[i]/dt/gamma_0), m_srs) - tau1[i]/g1[i];
    } else {
      R[i] = 0;
    }
  }

/******
    fprintf (stderr, "dgamma = [ ");                                   
    fprintf (stderr, "%g, %g ", dgamma[0], dgamma[1]); 
    fprintf (stderr, "]\n");                                            
*******/

/*********
    fprintf (stderr, "R = [ ");                                   
    fprintf (stderr, "%g %g, %g %g ", R[0], R[1], R[6], R[7]); 
    fprintf (stderr, "]\n");                                            
********/
}

static void
calc_h_overhat1 (double h_0,
                 double accum_gamma,
                 double tau_s, double tau_0, /* tau_s Saturation strength
                                         tau_0 Critical resolved shear stress */
                 double q,
                 double dgamma[N_SLIP_SYSTEMS],
                 double h_overhat1[N_SLIP_SYSTEMS][N_SLIP_SYSTEMS]
                )
{
  int     i,j, k;
  double  h1;
  double  temp1 = h_0 * accum_gamma/(tau_s-tau_0);
  double  temp2 = h_0 * (-2 * pow(cosh(temp1),-2) * tanh(temp1)) * h_0/(tau_s-tau_0);
  double  temp3; 
  double  h = h_0*pow(cosh(temp1),-2);
  FOR (i, N_SLIP_SYSTEMS) {
     FOR (j, N_SLIP_SYSTEMS) { 
	h_overhat1[i][j] = 0;
     }
  }
	
  FOR (i, N_SLIP_SYSTEMS) {
     FOR (j, N_SLIP_SYSTEMS) {
         if (i == j) h1 = h;
         else        h1 = h * q;
 
         temp3 = 0;
         FOR (k, N_SLIP_SYSTEMS) {
             if (i == k) temp3 += dgamma[k];
             else        temp3 += q * dgamma[k];
         }
        h_overhat1[i][j] = h1 + temp2 * temp3;
     }
  }
 
}

static void
calc_Rbar1 (int N, /* N stands for the Nth slip system */
            double Cebar1[3][3],
            double Fp[3][3],
            double invFp1[3][3],
            double s[N_SLIP_SYSTEMS][3],
            double m[N_SLIP_SYSTEMS][3],
            double Rbar1[3][3]
           )
 
{
  double dgsm[3][3];
  double temp1[3][3];
  double temp2[3][3];
 
  MAT3DUTIL_TENS_PROD_V (dgsm, s[N], m[N]);
  MAT3DUTIL_MULT_AB_3X3 (temp1, Cebar1, dgsm);
  MAT3DUTIL_MULT_AB_3X3 (temp2, temp1, Fp);
  MAT3DUTIL_MULT_AB_3X3 (Rbar1, temp2, invFp1);
}
 
/* effective spatial plastic flow direction */
static void
calc_r1(double  Rbar1[3][3],
        double  Fe1[3][3],
        double  r1[3][3]
        )
 
{
  double temp[3][3];
  double invFe1[3][3];
  double detFe1 = MAT3DUTIL_DET_3X3 (Fe1);
  if (detFe1 == 0) {
    REPORT_FATAL_ERROR (NULL, "Zero det(Fe1)");
  }
  MAT3DUTIL_INVERT_3X3 (detFe1, Fe1, invFe1);
  MAT3DUTIL_MULT_AB_3X3 (temp, invFe1, Rbar1);
  MAT3DUTIL_MULT_AB_3X3 (r1, temp, invFe1);
 
}

/* effective spatial normal direction */
static void
calc_n1(int N,  /* N stands for the Nth slip system */
        double gamma_0,
        double m_srs,
        double tau1[N_SLIP_SYSTEMS],
        double g1[N_SLIP_SYSTEMS],
        double invFe1[3][3],
        double s[N_SLIP_SYSTEMS][3],
        double m[N_SLIP_SYSTEMS][3],
	double Cebar1[3][3],
        double n1[3][3]
        )
{
  int i,j;
  double dphidtau;
  double dphids[3][3];
  double temp1[3][3];
  double temp2[3];

/* calculate the derivative of Phi w.r.t. tau1 */

  if (tau1[N] > g1[N])  dphidtau = SIGN(tau1[N]) * gamma_0 * pow(tau1[N]/g1[N], 1/m_srs - 1)/g1[N]/m_srs;
  else                  dphidtau = 0;
 
/* calculate the derivative of tau1 w.r.t. 2nd PK stress tensor */
  MAT3DUTIL_MULT_A_3X3_V_3(temp2, Cebar1, s[N]);
  FOR (i, 3) {
      FOR (j, 3){
        dphids[i][j] = temp2[i] * m[N][j] ;
      }
  }
 
/* calculate the derivative of Phi w.r.t. 2nd PK stress tensor */
  MAT3DUTIL_SCALE_3X3(dphids, dphidtau);
 
  MAT3DUTIL_MULT_AB_3X3 (temp1, dphids, invFe1);
  MAT3DUTIL_MULT_AB_3X3 (n1, temp1, invFe1);
 
}

static void
calc_A1(double Fp[3][3],
        double Cebar1[3][3],
        double Fe1[3][3],
        double invFe1[3][3],
        double invFp1[3][3],
        double C[9][9],
        double gamma_0,
        double m_srs,
        double tau1[N_SLIP_SYSTEMS],
        double g1[N_SLIP_SYSTEMS],
        double h_overhat1[N_SLIP_SYSTEMS][N_SLIP_SYSTEMS],
	double s[N_SLIP_SYSTEMS][3],
	double m[N_SLIP_SYSTEMS][3],
        double dt,
        gep_dense_matrix_t *A1
        )
{
  int i,j;
  double r1[3][3], n1[3][3];
  double dphidg;
  double n1v[9];
  double r1v[9];
  double temp1, temp2[9];
  double Rbar1[3][3]; 

  FOR (i, N_SLIP_SYSTEMS) {
     FOR (j, N_SLIP_SYSTEMS) {
         GEP_ELEM((*A1), i, j) = 0;
         calc_n1 (i, gamma_0, m_srs, tau1, g1, invFe1, s, m, Cebar1, n1);
         calc_Rbar1 (j, Cebar1, Fp, invFp1, s, m, Rbar1);
         calc_r1 (Rbar1, Fe1, r1);
 
         MAT3DUTIL_T_TO_V9(n1, n1v);
         MAT3DUTIL_T_TO_V9(r1, r1v);
         MAT3DUTIL_MULT_A_9x9_V_9(temp2, C, r1v);
         MAT3DUTIL_MULT_V9_V9(temp1, n1v, temp2);

         if (tau1[i] > g1[i])  dphidg = -SIGN(tau1[i]) * gamma_0 * pow(tau1[i]/g1[i], 1/m_srs - 1) * tau1[i]/g1[i]/g1[i]/m_srs;
         else                  dphidg = 0;
 
         GEP_ELEM((*A1), i, j) -= dphidg * h_overhat1[i][j];
         GEP_ELEM((*A1), i, j) += temp1;
         if (i == j) {
             GEP_ELEM((*A1), i, j) += 1/dt;
         }
    }
  }
}



typedef struct {
  double q;
  double h_0;
  double g_0;
  double g_inf;
  double m_srs;
  double gamma_0;
  double (*Dbarp)[6];
  double (*F1p)[3];
  double (*Fpp)[3];
  double *gammap;
  double accum_gamma;
  double *gp;
  double (*sp)[3];
  double (*mp)[3];
  double dt;
  double *dgammap;
  W_bool_t *activep;
  double dir[N_SLIP_SYSTEMS];
}  fminbr_client;

static double
get_Rn (fminbr_client *c, double eta)
{
  double dgamma1[N_SLIP_SYSTEMS];
  double R[N_SLIP_SYSTEMS];
  double Rn2;
  VOP2 (dgamma1, =, (c->dgammap), -, eta * c->dir, N_SLIP_SYSTEMS); /* note the minus */
  calc_R (c->q, c->h_0, c->g_0, c->g_inf, c->m_srs, c->gamma_0,
          c->Dbarp, c->F1p, c->Fpp, c->gammap, c->accum_gamma, c->gp, c->sp, c->mp,
          c->dt, dgamma1, c->activep, R);
  Rn2 = l2_norm_squared (R);
  /*  fprintf (stderr, "eta=%g, Rn2=%g\n", eta, Rn2);*/
  return Rn2;
}

static double
line_search (double q, double h_0, double g_0, double g_inf, double m_srs, double gamma_0,
             double Dbar[6][6],  double F1[3][3],
             double Fp[3][3],  double gamma[N_SLIP_SYSTEMS], double accum_gamma,
             double g[N_SLIP_SYSTEMS],  double s[N_SLIP_SYSTEMS][3],
             double m[N_SLIP_SYSTEMS][3],
             double dt,  double dgamma[N_SLIP_SYSTEMS],
             W_bool_t active[N_SLIP_SYSTEMS],
             double R[N_SLIP_SYSTEMS],
             double Rn
             )
{
  double dir[N_SLIP_SYSTEMS];
  double eta = 1;
  double eta_tol = 0.000001;
  double Rn_tol = 0.1 * Rn;
  fminbr_client c;
  VOP1 (dir, =, R, N_SLIP_SYSTEMS);
  c.q = q; c.h_0 = h_0; c.g_0 = g_0; c.g_inf = g_inf;
  c.m_srs = m_srs; c.gamma_0 = gamma_0; c.Dbarp = Dbar; c.F1p = F1;
  c.Fpp = Fp; c.gammap = gamma; c.accum_gamma = accum_gamma; c.gp = g; c.sp = s; c.mp = m; c.dt = dt;
  c.dgammap = dgamma; c.activep = active; VOP1 (c.dir, =, R, N_SLIP_SYSTEMS);
  eta = ckit_fminbr (-1, 1.0+eta_tol, &c,
                     (double (*)(void *, double)) get_Rn, eta_tol, Rn_tol);
  return eta;
}

static void
calc_jac (double q,  double h_0,  double g_0, double g_inf, double m_srs,  double gamma_0,
          double Dbar[6][6],  double F1[3][3],
          double Fp[3][3],  double gamma[N_SLIP_SYSTEMS], double accum_gamma,
          double g[N_SLIP_SYSTEMS],
          double s[N_SLIP_SYSTEMS][3],  double m[N_SLIP_SYSTEMS][3],
          double dt,  double dgamma[N_SLIP_SYSTEMS],
          W_bool_t active[N_SLIP_SYSTEMS],
          gep_dense_matrix_t *jac /* OUT: jacobian */
          )
{
  int i, j;
  double eps[N_SLIP_SYSTEMS], e;
  double dgamma1[N_SLIP_SYSTEMS];
  double fm[N_SLIP_SYSTEMS];
  double fp[N_SLIP_SYSTEMS];
  VOP1 (dgamma1, =, dgamma, N_SLIP_SYSTEMS);
  /* compute eps for each system */
  FOR (i, N_SLIP_SYSTEMS) {
    if (active[i]) {
      if (dgamma[i] != 0) {
        eps[i] = fabs(dgamma[i]) / 1000;
      } else
        eps[i] = 1e-12;
    } else
      eps[i] = 0;
  }
  /* evaluate the numerical derivatives */
  FOR (i, N_SLIP_SYSTEMS) {
    if (active[i]) {
      dgamma1[i] = dgamma[i] + eps[i];
      calc_R (q, h_0, g_0, g_inf, m_srs, gamma_0, Dbar, F1, Fp, gamma, accum_gamma,
              g, s, m, dt, dgamma1, active, fp);
      dgamma1[i] = dgamma[i] - eps[i];
      calc_R (q, h_0, g_0, g_inf, m_srs, gamma_0, Dbar, F1, Fp, gamma, accum_gamma,
              g, s, m, dt, dgamma1, active, fm);
      FOR (j, N_SLIP_SYSTEMS) {
        if (active[j]) {
          GEP_ELEM((*jac), j, i) = (fp[j] - fm[j])/(2*eps[i]);
        } else {
          if (i == j) GEP_ELEM((*jac), i, j) = 1;
          else        GEP_ELEM((*jac), i, j) = 0;
        }
      }
      dgamma1[i] = dgamma[i];
    } else {
      FOR (j, N_SLIP_SYSTEMS) { GEP_ELEM((*jac), i, j) = 0; } GEP_ELEM((*jac), i, i) = 1;
    }
  }
}

#define DBGP_DGAMMA(tit,what)  {                                        \
    int I_I_;                                                           \
    fprintf (stderr, "%s = [ ", tit);                                   \
    FOR (I_I_, N_SLIP_SYSTEMS) { fprintf (stderr, "%g ", what[I_I_]); }	\
    fprintf (stderr, "]\n");                                            \
  }

#define DBGP(tit,what)  {                                                    \
    int I_I_;                                                           \
    fprintf (stderr, "%s = [ ", tit);                                   \
    FOR (I_I_, N_SLIP_SYSTEMS) { fprintf (stderr, "%g ", what[I_I_]); } \
    fprintf (stderr, "]\n");                                             \
  }

/* PRINT SLIP SYSTEM 1,2,7,8 ONLY */
#define DBGP1278(tit,what)  {                                               \
    fprintf (stderr, "%s = [ ", tit);                                   \
    fprintf (stderr, "%g %g, %g %g ", what[0], what[1], what[6], what[7]); \
    fprintf (stderr, "]\n");                                            \
  }
#define DBGPM(tit,what)  {                                   \
    int I_I_, J_J_;                                     \
    fprintf (stderr, "%s = [ ", tit);                    \
    FOR (I_I_, N_SLIP_SYSTEMS) {                        \
       FOR (J_J_, N_SLIP_SYSTEMS) {                     \
         fprintf (stderr, "%g ", what[I_I_][J_J_]);     \
       }                                                \
       fprintf (stderr, ";\n");                          \
    }                                                   \
    fprintf (stderr, "]\n");                             \
  }
#define DBGPM1(tit,what)  {                                   \
    int I_I_, J_J_;                                     \
    fprintf (stderr, "%s = [ ", tit);                    \
    FOR (I_I_, 3) {                        \
       FOR (J_J_, 3) {                     \
         fprintf (stderr, "%g ", what[I_I_][J_J_]);     \
       }                                                \
       fprintf (stderr, ";\n");                          \
    }                                                   \
    fprintf (stderr, "]\n");                             \
  }
#define DBGPMK(tit,what)  {                                   \
    int I_I_, J_J_;                                     \
    fprintf (stderr, "%s = [ ", tit);                    \
    FOR (I_I_, 24) {                        \
       FOR (J_J_, 24) {                     \
         fprintf (stderr, "%g ", what[I_I_][J_J_]);     \
       }                                                \
       fprintf (stderr, ";\n");                          \
    }                                                   \
    fprintf (stderr, "]\n");                             \
  }

/* -(exp(log(p)*m)-1)/m */
static W_bool_t
constraints_satisfied (double q, double h_0, double g_0, double g_inf,
                       double m_srs, double gamma_0,
                       double Dbar[6][6],  double F1[3][3],
                       double Fp[3][3],  double gamma[N_SLIP_SYSTEMS], double accum_gamma,
                       double g[N_SLIP_SYSTEMS],  double s[N_SLIP_SYSTEMS][3],
                       double m[N_SLIP_SYSTEMS][3],
                       double dt,  double dgamma[N_SLIP_SYSTEMS],
                       double activity_tol,
                       W_bool_t old_active[N_SLIP_SYSTEMS],
                       W_bool_t active[N_SLIP_SYSTEMS],
                       int *nactive_out
                       )
{
  int i, nviolated = 0;
  int nactive = 0;
  double Fp1[3][3];
  double Cebar1[3][3]; /* Green-Lagrange strain */
  double Sbar1[3][3]; /* 2nd PK */
  double h1[N_SLIP_SYSTEMS][N_SLIP_SYSTEMS]; /* current hardening matrix */
  double g1[N_SLIP_SYSTEMS];
  double tau1[N_SLIP_SYSTEMS];
  const double overstress_constraint = exp(log(activity_tol)*m_srs);

  /* compute actual resolved shear stresses */
  calc_g1 (q, h_0, g_0, g_inf, g, accum_gamma, dgamma, g1);
  calc_Fp1 (Fp, dgamma, s, m, Fp1); 
  calc_Cebar1 (F1, Fp1, Cebar1);
  calc_Sbar (Dbar, Cebar1, Fp1, Sbar1);
  calc_tau1 (Cebar1, Sbar1, s, m, tau1);

  /* if a system was inactive, its shear stress
     should be below a fraction of the flow stress: overstress_constraint*g */
  nviolated = 0;
  {
    double max_overstress = 0; int maxi;
    FOR (i, N_SLIP_SYSTEMS) {
      if ((!old_active[i]) && (fabs(tau1[i]) >= overstress_constraint * g1[i])) {
        nviolated++;
        if (fabs(tau1[i]/g1[i]) > max_overstress) {
          max_overstress = fabs(tau1[i]/g1[i]);
          maxi = i;
        }
      }
    }
    if (nviolated > 0) { 
      nactive = 0;
      FOR (i, N_SLIP_SYSTEMS) {
        /* add any system that was active before, or should be active now */
        if (old_active[i] || (fabs(tau1[i]/g1[i]) >= overstress_constraint)) {
          active[i] = TRUE; nactive++;
        }
      }
      *nactive_out = nactive;
      return FALSE; /* constraints not satisfied */
    }
  }
  
  /* no violations: set the active flag only for those systems that actually yield */
  nactive = 0;
  FOR (i, N_SLIP_SYSTEMS) {
    active[i] = (fabs(tau1[i]) >= overstress_constraint*g1[i]);
    if (active[i]) nactive++;
  }
  return TRUE; /* done */
}  

static void
upd (W_bool_t elastic,
     W_bool_t must_be_accurate,
     double time,
     int debug, /* should I print debugging information, and how much? */
     int max_inner_iter,
     double activity_tol,
     double s[N_SLIP_SYSTEMS][3], /* slip directions */
     double m[N_SLIP_SYSTEMS][3], /* normals to slip planes */
     double g_0, /* initial flow stress */
     double g_inf, 
     double q, /* hardening behaviour parameter */
     double h_0, /* initial hardening rate */
     double gamma_0, /* reference shear strain rate */
     double m_srs, /* strain-rate sensitivity exponent */
     double Dbar[6][6], /* S=C:E */
     double F1[3][3], /* F_n+1   total deformation gradient at t_n+1*/
     double Fp[3][3], /* F^p_n  plastic deformation gradient at t_n */
     double g[N_SLIP_SYSTEMS], /* flow stress on slip system at t_n */
     double tau1[N_SLIP_SYSTEMS], /* resolved shear stress on slip system at t_n */
     double gamma[N_SLIP_SYSTEMS], /* shear strain on slip system at t_n */
     double accum_gamma, /* accumulated slip */
     double dt, /* time increment */
     double Sv_out[6], /* OUT: PK2 */
     double Fp1_out[3][3],          /* OUT: plastic deformation gradient */
     double g1_out[N_SLIP_SYSTEMS], /* OUT: flow stresses */
     double *accum_gamma_out, /* OUT: accumulated slip */
     double gamma1_out[N_SLIP_SYSTEMS], /* OUT: shear strain rates */
     double dgamma_out[N_SLIP_SYSTEMS], /* OUT: shear strain rates */
     double *deform_energy_density_out /* OUT: deformation energy density */
     )
{
  double R[N_SLIP_SYSTEMS];
  int i, j, outer_iter, inner_iter;
  const double Rn2_tol = 1e-16;
  const double Rn2_rel_tol = 0.01;
  double rcond, z[N_SLIP_SYSTEMS];
  const double rcond_tol = 1e-20;
  GEP_MATRIX (jac);
  int jacpvt[N_SLIP_SYSTEMS];
  double dgamma[N_SLIP_SYSTEMS];
  double ddg[N_SLIP_SYSTEMS];
  double sl;
  W_bool_t active[N_SLIP_SYSTEMS];
  W_bool_t old_active[N_SLIP_SYSTEMS];
  int nactive;
  double eta, Rn2_in_first, prev_Rn2;

  GEP_STACKINIT(jac, N_SLIP_SYSTEMS);
#define CONVERGED ((inner_iter > 1) && ((Rn2 <= Rn2_tol)))
  
  outer_iter = 1;
  VOP1C (active, =, FALSE, N_SLIP_SYSTEMS); /* none are active */
  VOP1C (old_active, =, FALSE, N_SLIP_SYSTEMS); /* none are active */
  VOP1C (dgamma, =, 0, N_SLIP_SYSTEMS);
  nactive = 0;
  if (!elastic) {
    while (!constraints_satisfied (q, h_0, g_0, g_inf, m_srs, gamma_0, Dbar, F1, Fp,
                                   gamma, accum_gamma, g, s, m, dt, dgamma, activity_tol,
                                   old_active, active, &nactive))
      { /* Solve for the active slip system */
      restart:
        prev_Rn2 = 1e300;
        VOP1C (dgamma, =, 0, N_SLIP_SYSTEMS);
        VOP1C (ddg, =, 0, N_SLIP_SYSTEMS);
        inner_iter = 1;
        if (debug > 0) {
          int jj = 0;
          fprintf (stderr, "@@@@ %d active slip systems in %d-th outer iteration: ",
                   nactive, outer_iter);
          FOR (jj, N_SLIP_SYSTEMS) fprintf(stderr, "%d", active[jj]);
          fprintf(stderr, "\n");
        }

        do {   /* Solve for the increment of shear strains */
          double Rn2;
          calc_R (q, h_0, g_0, g_inf, m_srs, gamma_0, Dbar, F1, Fp,
                  gamma, accum_gamma, g, s, m, dt, dgamma, active, R);
          if (debug > 2) { DBGP("R", R); }
          Rn2 = l2_norm_squared (R); if (inner_iter == 1) Rn2_in_first = prev_Rn2 = Rn2;
          if ((Rn2 > prev_Rn2)) {
            double eta ;
            VOP1 (dgamma, -=, ddg, N_SLIP_SYSTEMS); /* undo */
            eta = line_search (q, h_0, g_0, g_inf, m_srs, gamma_0, Dbar, F1, Fp, gamma,
                               accum_gamma, g, s, m, dt, dgamma, active, ddg, prev_Rn2);
            VOP1 (dgamma, -=, eta*ddg, N_SLIP_SYSTEMS);
            calc_R (q, h_0, g_0, g_inf, m_srs, gamma_0, Dbar, F1, Fp,
                    gamma, accum_gamma, g, s, m, dt, dgamma, active, R);
            Rn2 = l2_norm_squared (R); 
          }
          if (debug > 0) { fprintf (stderr, "   @@@@  Iteration %d: %g @@@@\n", inner_iter, Rn2); }
          if (CONVERGED) break;
          calc_jac (q, h_0, g_0, g_inf, m_srs, gamma_0, Dbar, F1, Fp,
                    gamma, accum_gamma, g, s, m, dt, dgamma, active, &jac);
          gep_sgeco (&jac, jacpvt, &rcond, z);
          if (rcond < rcond_tol) {
            int kk; double regeps = g_0/gamma_0/dt;
            calc_jac (q, h_0, g_0, g_inf, m_srs, gamma_0, Dbar, F1, Fp,
                      gamma, accum_gamma, g, s, m, dt, dgamma, active, &jac);
            if (debug) fprintf (stderr, "rcond before regularization=%g\n", rcond);
            FOR (kk, N_SLIP_SYSTEMS) { GEP_ELEM(jac, kk, kk) += regeps; }
            gep_sgeco (&jac, jacpvt, &rcond, z);
            if (debug) fprintf (stderr, "rcond after regularization=%g\n", rcond);
          }
          gep_sgesl (&jac, jacpvt, R, GEP_JOB_A_X_B);
          VOP1 (ddg, =, -R, N_SLIP_SYSTEMS); /* POSITIVE incremental direction */
          VOP1 (dgamma, +=, ddg, N_SLIP_SYSTEMS); 
          if (debug > 2) { DBGP("dgamma", dgamma); }

          if (inner_iter > max_inner_iter) {
            if (must_be_accurate) {
              sprintf (W_buf(), "Converged to %g (initial Rn2=%g) in final update", Rn2, Rn2_in_first);
              REPORT_WARNING (NULL, W_buf());
            } else {
              if (Rn2 > Rn2_in_first / 100) {
                sprintf (W_buf(), "Converged to %g (initial Rn2=%g)", Rn2, Rn2_in_first);
                REPORT_WARNING (NULL, W_buf());
              }
            }
            break;
          }
          prev_Rn2 = Rn2;
          inner_iter++;
        } while (TRUE);
        outer_iter++;
        VOP1 (old_active, =, active, N_SLIP_SYSTEMS); 
      };
  } /* plasticity treatment */
  
    /* Return the result */
  {
    double S1[3][3], Sbar1[3][3]; /* 2nd PK */
    double Cebar1[3][3]; /* Green-Lagrange strain */
    int j;
    VOP2 (gamma1_out, =, gamma, +, dgamma, N_SLIP_SYSTEMS);
    *accum_gamma_out = accum_gamma;
    FOR (j, N_SLIP_SYSTEMS) { *accum_gamma_out += fabs(dgamma[j]); }
    FOR (j, N_SLIP_SYSTEMS) { dgamma_out[j] = dgamma[j]; }
    calc_Fp1 (Fp, dgamma, s, m, Fp1_out); 
    calc_Cebar1 (F1, Fp1_out, Cebar1);

    calc_Sbar (Dbar, Cebar1, Fp1_out, Sbar1);
    calc_tau1 (Cebar1, Sbar1, s, m, tau1); /* save tau to matstate */

    calc_g1 (q, h_0, g_0, g_inf, g, *accum_gamma_out, dgamma, g1_out);
    calc_S (Sbar1, Fp1_out, S1);
    MAT3DUTIL_T_TO_V (S1, Sv_out);
    {
      double E1[3][3];
      MAT3DUTIL_2_MAT (E1, = 0.5 *, Cebar1);
      MAT3DUTIL_MAT_MOD_DIAG (E1, -=, 0.5);
      *deform_energy_density_out
        = 0.5 * MAT3DUTIL_DBL_PROD_3X3 (Sbar1, E1);
    }
  }
}

static W_bool_t  
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  double F[3][3];
  double Dbar[6][6];
  double Sv[6]; 
  double gamma1[N_SLIP_SYSTEMS];
  double dgamma[N_SLIP_SYSTEMS];
  double g1[N_SLIP_SYSTEMS];
  double accum_gamma;
  double Fp1[3][3];
  double deform_energy_density;
  int i, k;
  double dt = ip_ctx->time_step;
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  W_bool_t result = TRUE;
  matstate_t input_matst;

  double s1[N_SLIP_SYSTEMS][3]; /* slip directions */
  double m1[N_SLIP_SYSTEMS][3]; /* normals to slip planes */

  if (!ip_ctx->get_kinem_var) return FALSE;  

  /* kinematic quantities */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
  
  /* save current material state */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  /* Calculate slip systems */
  /* could be cached with the material */
  FOR (i, N_SLIP_SYSTEMS) { 
     s1[i][0] = matst->a.x * ref_s[i][0] + matst->b.x * ref_s[i][1] + matst->c.x * ref_s[i][2];
     s1[i][1] = matst->a.y * ref_s[i][0] + matst->b.y * ref_s[i][1] + matst->c.y * ref_s[i][2];
     s1[i][2] = matst->a.z * ref_s[i][0] + matst->b.z * ref_s[i][1] + matst->c.z * ref_s[i][2];
     m1[i][0] = matst->a.x * ref_m[i][0] + matst->b.x * ref_m[i][1] + matst->c.x * ref_m[i][2];
     m1[i][1] = matst->a.y * ref_m[i][0] + matst->b.y * ref_m[i][1] + matst->c.y * ref_m[i][2];
     m1[i][2] = matst->a.z * ref_m[i][0] + matst->b.z * ref_m[i][1] + matst->c.z * ref_m[i][2];
  }
  calc_Dbar (u->c11, u->c12, u->c44, &matst->a, &matst->b, &matst->c, Dbar);
  
  upd (u->elastic, (update_form == W_MAT_STATE_UPDATE_FINAL),
       ip_ctx->time, u->debug, u->max_inner_iter, u->activity_tol,
       s1, m1, u->g_0, u->g_inf,
       u->q, u->h_0, u->gamma_0, u->m_srs,
       Dbar, F, matst->Fp, matst->g, matst->tau, matst->gamma, matst->accum_gamma, dt,
       Sv, Fp1, g1, &accum_gamma, gamma1, dgamma, &deform_energy_density);

    /* for ckink problem only */
 /*  FOR(k, N_SLIP_SYSTEMS)
   {
     if ((!u->elastic) && fabs(matst->tau[k]) >= 1.1 * matst->g[k])
     {
      char msg[1024];
      sprintf(msg, "tau_%d = %g, dgamma_%d = %g, gamma_%d = %g", k+1, matst->tau[k], k+1, dgamma[k], k+1, gamma1[k]);
      REPORT_WARNING (u->d->errh, msg);
     }
   }  */

  /* save state */
  matst->S[0] = Sv[0];
  matst->S[1] = Sv[1];
  matst->S[2] = Sv[2];
  matst->S[3] = Sv[3];
  matst->S[4] = Sv[4];
  matst->S[5] = Sv[5];
  MAT3DUTIL_2_MAT (matst->Fp, =, Fp1);
  FOR (i, N_SLIP_SYSTEMS) {
    matst->g[i] = g1[i];
    matst->gamma[i] = gamma1[i];
    matst->dgamma[i] = dgamma[i];
  }
  matst->accum_gamma = accum_gamma;
  matst->deform_energy_density = deform_energy_density;

  result = get_stress (m, ms, ip_ctx, stress_type, stress, out_stress_type);
   
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}


/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress (W_mat_3d_t                  *m,
            W_mat_3d_state_p             ms,
            W_mat_3d_ip_ctx_t           *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double                       stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  matstate_t *matst = (matstate_t *)ms;
  double S[3][3];
  W_bool_t result = TRUE;

  MAT3DUTIL_V_TO_T (matst->S, S);
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    {
      double F[3][3], P[3][3], invdetF, sigma[3][3];
      int i, j, k;
      if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
        result = FALSE; goto exit;
      }
      FOR(i, 3) {
        FOR(j, 3) {
          P[i][j] = 0; /* /F/ * /S/ */ 
          FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
        }
      }
      invdetF = 1 / MAT3DUTIL_DET_3X3(F);
      FOR(i, 3) {
        FOR(j, 3) {
          sigma[i][j] = 0; /* 1/detF /P/ * /F/^T */ 
          FOR(k, 3) sigma[i][j] += invdetF * P[i][k] * F[j][k];
        }
      }
      MAT3DUTIL_COPY_3X3(stress, sigma);
    }
    break;
  case MAT_3D_1ST_PK:
    {
      double F[3][3], P[3][3];
      int i, j, k;
      if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
        result = FALSE; goto exit;
      }
      FOR(i, 3) {
        FOR(j, 3) {
          P[i][j] = 0; /* /F/ * /S/ */ 
          FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
        }
      }
      MAT3DUTIL_COPY_3X3(stress, P);
    }
    break;
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, S);
    *out_stress_type = MAT_3D_2ND_PK;
    break;
  }
    
 exit:
  return result;
}


static W_bool_t 
get_mat_state_var (W_mat_3d_t           *m,
                   W_mat_3d_state_p             ms,
                   W_mat_3d_ip_ctx_t           *ip_ctx,
                   W_mat_3d_mat_state_var_enum  var_type,
                   double                      *var
                   )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0; /* RAW */
    return TRUE;
  default:
    return FALSE;
  }
}

static W_bool_t 
get_mat_state_named_var (W_mat_3d_t *m,
                         W_mat_3d_state_p ms,
                         W_mat_3d_ip_ctx_t *ip_ctx,
                         char *name,
                         double *var
                         )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

  if        (tokens_token_and_keyword_equiv (name, "fp_xx")) {
    *var = matst->Fp[0][0];
  } else if (tokens_token_and_keyword_equiv (name, "fp_xy")) {
    *var = matst->Fp[0][1];
  } else if (tokens_token_and_keyword_equiv (name, "fp_xz")) {
    *var = matst->Fp[0][2];
  } else if (tokens_token_and_keyword_equiv (name, "fp_yx")) {
    *var = matst->Fp[1][0];
  } else if (tokens_token_and_keyword_equiv (name, "fp_yy")) {
    *var = matst->Fp[1][1];
  } else if (tokens_token_and_keyword_equiv (name, "fp_yz")) {
    *var = matst->Fp[1][2];
  } else if (tokens_token_and_keyword_equiv (name, "fp_zx")) {
    *var = matst->Fp[2][0];
  } else if (tokens_token_and_keyword_equiv (name, "fp_zy")) {
    *var = matst->Fp[2][1];
  } else if (tokens_token_and_keyword_equiv (name, "fp_zz")) {
    *var = matst->Fp[2][2];
  } else if (tokens_token_and_keyword_equiv (name, "ln_U_1") ||
             tokens_token_and_keyword_equiv (name, "ln_U_2") ||
             tokens_token_and_keyword_equiv (name, "ln_U_3") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U") 
             ) {
    double F[3][3], R[3][3], U[3][3], ln_U[3][3];
    double evals[3], evecs[3][3], lnevals[3];
    *var = 0;
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    if (W_polar_decomposition (F, R, U)) {
      W_eigendecomposition (U, evecs, evals);
      if (tokens_token_and_keyword_equiv (name, "ln_U_1")) *var = log(evals[0]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_2")) *var = log(evals[1]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_3")) *var = log(evals[2]);
      if (tokens_token_and_keyword_equiv (name, "max_ln_U")) {
        *var = fabs(log(evals[0]));
        if (fabs(log(evals[1])) > *var) *var = fabs(log(evals[1]));
        if (fabs(log(evals[2])) > *var) *var = fabs(log(evals[2]));
      }
      goto exit;
    }
  } else {
    int i;
    char buf[128];
    FOR (i, N_SLIP_SYSTEMS) {
      sprintf (buf, "gamma_%d", i+1);
      if (tokens_token_and_keyword_equiv (name, buf)) {
        *var = matst->gamma[i];
        goto exit;
      }
    }
    FOR (i, N_SLIP_SYSTEMS) {
      sprintf (buf, "g_%d", i+1);
      if (tokens_token_and_keyword_equiv (name, buf)) {
        *var = matst->g[i];
        goto exit;
      }
    }
    FOR (i, N_SLIP_SYSTEMS) {
      sprintf (buf, "tau_%d", i+1);
      if (tokens_token_and_keyword_equiv (name, buf)) {


	*var = matst->tau[i];
        goto exit;
      }
    }

    *var = 0;
    return FALSE;
  }
 exit:
  return TRUE;
}

static void
update_history (W_mat_3d_t *m,
                W_mat_3d_state_p ms,
                W_mat_3d_ip_ctx_t *ip_ctx,
                W_solution_history_t *h, /* solution history */
                W_index_t pntidx,
                W_float_t time
                )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

#define STORE(HW, NAME, TIME, I, VAL)           \
  {                                             \
    char buf[128];                              \
    sprintf (buf, "p%d%s_%d", pntidx, NAME, I); \
    WHR_store (HW, buf, TIME, VAL);             \
  }
  int i;

  FOR (i, N_SLIP_SYSTEMS) {
    STORE (h->writer, "g", time, i+1, matst->g[i]);
    STORE (h->writer, "gamma", time, i+1, matst->gamma[i]);
    STORE (h->writer, "tau", time, i+1, matst->tau[i]);
  }
}

static W_mat_3d_state_p 
interpolate_mat_state(W_mat_3d_t *m,
		      W_loc_t *at_point,
		      int npkgs,
		      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    FOR(i, 6) {
      ms->S[i]      += pms->S[i] * WGHT(k);
    }
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}


static W_bool_t
get_mat_stiff_elastic (W_mat_3d_t *m/* mat3d */,
                       W_mat_3d_state_p ms,/* material state */
                       W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                                    used to generate kinematic
                                                    quantities which the material state
                                                    might need in order to compute
                                                    the requested stress */
                       W_mat_3d_stiff_enum stiff_type, /* stiffness type */
                       double D[6][6] /* material stiffness tensor in matrix form */
                       )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  double c11 = u->c11, c12 = u->c12, c44 = u->c44;
  double t0, s1, s2, Dbar[6][6];
  double invFp[3][3], detFp;
  double iFp00, iFp01, iFp02;
  double iFp10, iFp11, iFp12;
  double iFp20, iFp21, iFp22;
  double Dbar11, Dbar12, Dbar13, Dbar14, Dbar15, Dbar16, Dbar22, Dbar23,
    Dbar24, Dbar25, Dbar26, Dbar33, Dbar34, Dbar35, Dbar36, Dbar44,
    Dbar45, Dbar46, Dbar55, Dbar56, Dbar66;

  
  detFp = MAT3DUTIL_DET_3X3 (matst->Fp);
  if (detFp == 0) {
    REPORT_FATAL_ERROR (NULL, "Zero det(Fp)");
  }
  MAT3DUTIL_INVERT_3X3 (detFp, matst->Fp, invFp);
  iFp00 = invFp[0][0];    iFp01 = invFp[0][1];    iFp02 = invFp[0][2];
  iFp10 = invFp[1][0];    iFp11 = invFp[1][1];    iFp12 = invFp[1][2];
  iFp20 = invFp[2][0];    iFp21 = invFp[2][1];    iFp22 = invFp[2][2];
  
  /* Calculate Dbar such that Sbar = Dbar:Ebar */
  calc_Dbar (c11, c12, c44, &matst->a, &matst->b, &matst->c, Dbar);
  Dbar11=Dbar[0][0];
  Dbar12=Dbar[0][1];
  Dbar13=Dbar[0][2];
  Dbar14=Dbar[0][3];
  Dbar15=Dbar[0][4];
  Dbar16=Dbar[0][5];
  Dbar22=Dbar[1][1];
  Dbar23=Dbar[1][2];
  Dbar24=Dbar[1][3];
  Dbar25=Dbar[1][4];
  Dbar26=Dbar[1][5];
  Dbar33=Dbar[2][2];
  Dbar34=Dbar[2][3];
  Dbar35=Dbar[2][4];
  Dbar36=Dbar[2][5];
  Dbar44=Dbar[3][3];
  Dbar45=Dbar[3][4];
  Dbar46=Dbar[3][5];
  Dbar55=Dbar[4][4];
  Dbar56=Dbar[4][5];
  Dbar66=Dbar[5][5];

  /* Now pull back by Fp */;
  t0 = Dbar11*iFp00*iFp00*iFp00*iFp00+4.0*Dbar14*iFp00*iFp00*iFp00*iFp01+4.0*Dbar15*iFp00*iFp00*iFp00*iFp02+2.0*Dbar12*iFp00*iFp00*iFp01*iFp01+4.0*Dbar16*iFp00*iFp00*iFp01*iFp02+2.0*Dbar13*iFp00*iFp00*iFp02*iFp02+4.0*Dbar44*iFp00*iFp00*iFp01*iFp01+8.0*Dbar45*iFp00*iFp00*iFp01*iFp02+4.0*Dbar24*iFp00*iFp01*iFp01*iFp01+8.0*Dbar46*iFp00*iFp01*iFp01*iFp02+4.0*Dbar34*iFp00*iFp01*iFp02*iFp02+4.0*Dbar55*iFp02*iFp02*iFp00*iFp00+4.0*Dbar25*iFp02*iFp00*iFp01*iFp01+8.0*Dbar56*iFp02*iFp02*iFp00*iFp01+4.0*Dbar35*iFp02*iFp02*iFp02*iFp00+Dbar22*iFp01*iFp01*iFp01*iFp01+4.0*Dbar26*iFp01*iFp01*iFp01*iFp02+2.0*Dbar23*iFp01*iFp01*iFp02*iFp02+4.0*Dbar66*iFp02*iFp02*iFp01*iFp01+4.0*Dbar36*iFp01*iFp02*iFp02*iFp02+Dbar33*iFp02*iFp02*iFp02*iFp02;
  D[0][0] = t0;
  s1 = 2.0*Dbar25*iFp02*iFp00*iFp11*iFp11+2.0*Dbar34*iFp00*iFp01*iFp12*iFp12+2.0*Dbar25*iFp01*iFp01*iFp10*iFp12+2.0*Dbar16*iFp02*iFp01*iFp10*iFp10+2.0*Dbar16*iFp00*iFp00*iFp11*iFp12+2.0*Dbar34*iFp02*iFp02*iFp10*iFp11+Dbar13*iFp00*iFp00*iFp12*iFp12+2.0*Dbar36*iFp02*iFp02*iFp11*iFp12+2.0*Dbar35*iFp02*iFp02*iFp10*iFp12+Dbar12*iFp00*iFp00*iFp11*iFp11+2.0*Dbar24*iFp01*iFp01*iFp10*iFp11+2.0*Dbar35*iFp00*iFp12*iFp12*iFp02+2.0*Dbar36*iFp01*iFp12*iFp12*iFp02+2.0*Dbar24*iFp00*iFp11*iFp11*iFp01+2.0*Dbar15*iFp00*iFp00*iFp10*iFp12+2.0*Dbar14*iFp00*iFp10*iFp10*iFp01+2.0*Dbar15*iFp00*iFp10*iFp10*iFp02+2.0*Dbar14*iFp00*iFp00*iFp10*iFp11;
  t0 = s1+2.0*Dbar26*iFp01*iFp11*iFp11*iFp02+2.0*Dbar26*iFp01*iFp01*iFp11*iFp12+4.0*Dbar66*iFp01*iFp12*iFp11*iFp02+Dbar13*iFp02*iFp02*iFp10*iFp10+Dbar23*iFp01*iFp01*iFp12*iFp12+Dbar23*iFp02*iFp02*iFp11*iFp11+Dbar11*iFp00*iFp00*iFp10*iFp10+4.0*Dbar46*iFp01*iFp10*iFp11*iFp02+Dbar12*iFp01*iFp01*iFp10*iFp10+4.0*Dbar44*iFp00*iFp11*iFp10*iFp01+4.0*Dbar56*iFp01*iFp12*iFp10*iFp02+Dbar22*iFp01*iFp01*iFp11*iFp11+4.0*Dbar55*iFp00*iFp12*iFp10*iFp02+4.0*Dbar45*iFp00*iFp12*iFp10*iFp01+4.0*Dbar56*iFp00*iFp12*iFp11*iFp02+4.0*Dbar45*iFp00*iFp11*iFp10*iFp02+4.0*Dbar46*iFp00*iFp11*iFp12*iFp01+Dbar33*iFp02*iFp02*iFp12*iFp12;
  D[0][1] = D[1][0] = t0;
  s1 = 2.0*Dbar16*iFp01*iFp02*iFp20*iFp20+2.0*Dbar34*iFp00*iFp01*iFp22*iFp22+2.0*Dbar26*iFp01*iFp01*iFp21*iFp22+2.0*Dbar35*iFp00*iFp02*iFp22*iFp22+2.0*Dbar25*iFp01*iFp01*iFp20*iFp22+2.0*Dbar24*iFp01*iFp01*iFp20*iFp21+Dbar11*iFp00*iFp00*iFp20*iFp20+Dbar12*iFp00*iFp00*iFp21*iFp21+Dbar33*iFp02*iFp02*iFp22*iFp22+Dbar23*iFp02*iFp02*iFp21*iFp21+Dbar13*iFp02*iFp02*iFp20*iFp20+Dbar23*iFp01*iFp01*iFp22*iFp22+2.0*Dbar34*iFp02*iFp02*iFp20*iFp21+2.0*Dbar35*iFp02*iFp02*iFp20*iFp22+2.0*Dbar14*iFp00*iFp00*iFp20*iFp21+2.0*Dbar15*iFp00*iFp00*iFp20*iFp22+2.0*Dbar16*iFp00*iFp00*iFp21*iFp22+Dbar22*iFp01*iFp01*iFp21*iFp21;
  t0 = s1+Dbar12*iFp01*iFp01*iFp20*iFp20+2.0*Dbar25*iFp00*iFp02*iFp21*iFp21+Dbar13*iFp00*iFp00*iFp22*iFp22+2.0*Dbar36*iFp01*iFp02*iFp22*iFp22+2.0*Dbar26*iFp01*iFp02*iFp21*iFp21+2.0*Dbar36*iFp02*iFp02*iFp21*iFp22+2.0*Dbar24*iFp00*iFp01*iFp21*iFp21+2.0*Dbar14*iFp00*iFp01*iFp20*iFp20+4.0*Dbar45*iFp00*iFp02*iFp20*iFp21+4.0*Dbar46*iFp01*iFp02*iFp20*iFp21+4.0*Dbar55*iFp00*iFp02*iFp20*iFp22+4.0*Dbar66*iFp01*iFp02*iFp21*iFp22+4.0*Dbar56*iFp00*iFp02*iFp21*iFp22+2.0*Dbar15*iFp00*iFp02*iFp20*iFp20+4.0*Dbar45*iFp00*iFp01*iFp20*iFp22+4.0*Dbar44*iFp00*iFp01*iFp20*iFp21+4.0*Dbar46*iFp00*iFp01*iFp21*iFp22+4.0*Dbar56*iFp01*iFp02*iFp20*iFp22;
  D[0][2] = D[2][0] = t0;
  s1 = Dbar35*iFp02*iFp02*iFp02*iFp10+Dbar11*iFp00*iFp00*iFp00*iFp10+Dbar14*iFp00*iFp00*iFp00*iFp11+Dbar15*iFp00*iFp00*iFp00*iFp12+Dbar22*iFp01*iFp01*iFp01*iFp11+Dbar26*iFp01*iFp01*iFp01*iFp12+Dbar24*iFp01*iFp01*iFp01*iFp10+Dbar33*iFp02*iFp02*iFp02*iFp12+Dbar36*iFp02*iFp02*iFp02*iFp11+2.0*Dbar55*iFp02*iFp00*iFp00*iFp12+2.0*Dbar44*iFp00*iFp00*iFp01*iFp11+3.0*Dbar24*iFp00*iFp01*iFp01*iFp11+2.0*Dbar45*iFp00*iFp00*iFp01*iFp12+2.0*Dbar46*iFp00*iFp01*iFp01*iFp12+2.0*Dbar55*iFp02*iFp02*iFp00*iFp10+3.0*Dbar14*iFp00*iFp00*iFp10*iFp01+3.0*Dbar15*iFp00*iFp00*iFp10*iFp02+2.0*Dbar44*iFp00*iFp01*iFp01*iFp10+3.0*Dbar36*iFp01*iFp02*iFp02*iFp12+2.0*Dbar66*iFp01*iFp01*iFp02*iFp12+Dbar23*iFp02*iFp02*iFp11*iFp01+Dbar34*iFp02*iFp02*iFp10*iFp01;
  s2 = s1+Dbar16*iFp00*iFp00*iFp11*iFp02+Dbar12*iFp00*iFp00*iFp11*iFp01+Dbar13*iFp00*iFp00*iFp12*iFp02+Dbar13*iFp02*iFp02*iFp10*iFp00+Dbar16*iFp00*iFp00*iFp12*iFp01+Dbar25*iFp01*iFp01*iFp12*iFp00+Dbar12*iFp01*iFp01*iFp10*iFp00+Dbar25*iFp01*iFp01*iFp10*iFp02+Dbar23*iFp01*iFp01*iFp12*iFp02+Dbar34*iFp02*iFp02*iFp11*iFp00+2.0*Dbar56*iFp01*iFp02*iFp02*iFp10;
  t0 = s2+2.0*Dbar66*iFp01*iFp02*iFp02*iFp11+3.0*Dbar26*iFp01*iFp01*iFp11*iFp02+2.0*Dbar46*iFp01*iFp01*iFp02*iFp10+3.0*Dbar35*iFp02*iFp02*iFp00*iFp12+2.0*Dbar45*iFp02*iFp00*iFp00*iFp11+2.0*Dbar56*iFp02*iFp02*iFp00*iFp11+2.0*Dbar16*iFp01*iFp02*iFp10*iFp00+2.0*Dbar25*iFp02*iFp00*iFp11*iFp01+4.0*Dbar45*iFp00*iFp01*iFp10*iFp02+4.0*Dbar46*iFp00*iFp01*iFp11*iFp02+2.0*Dbar34*iFp01*iFp00*iFp12*iFp02+4.0*Dbar56*iFp02*iFp00*iFp12*iFp01;
  D[0][3] = D[3][0] = t0;
  s1 = 2.0*Dbar66*iFp01*iFp01*iFp02*iFp22+2.0*Dbar55*iFp00*iFp02*iFp02*iFp20+3.0*Dbar36*iFp01*iFp02*iFp02*iFp22+2.0*Dbar45*iFp00*iFp00*iFp02*iFp21+2.0*Dbar44*iFp00*iFp00*iFp01*iFp21+3.0*Dbar24*iFp00*iFp01*iFp01*iFp21+2.0*Dbar66*iFp01*iFp02*iFp02*iFp21+Dbar12*iFp00*iFp00*iFp21*iFp01+Dbar16*iFp00*iFp00*iFp22*iFp01+Dbar16*iFp00*iFp00*iFp21*iFp02+Dbar13*iFp00*iFp00*iFp22*iFp02+Dbar12*iFp01*iFp01*iFp20*iFp00+Dbar25*iFp01*iFp01*iFp20*iFp02+Dbar23*iFp01*iFp01*iFp22*iFp02+Dbar22*iFp01*iFp01*iFp01*iFp21+Dbar34*iFp02*iFp02*iFp20*iFp01+Dbar25*iFp01*iFp01*iFp22*iFp00+Dbar34*iFp02*iFp02*iFp21*iFp00+Dbar23*iFp02*iFp02*iFp21*iFp01+Dbar13*iFp02*iFp02*iFp20*iFp00+2.0*Dbar55*iFp00*iFp00*iFp02*iFp22+3.0*Dbar14*iFp00*iFp00*iFp20*iFp01;
  t0 = s1+3.0*Dbar26*iFp01*iFp01*iFp21*iFp02+3.0*Dbar35*iFp00*iFp02*iFp02*iFp22+2.0*Dbar56*iFp00*iFp02*iFp02*iFp21+2.0*Dbar45*iFp00*iFp00*iFp01*iFp22+2.0*Dbar46*iFp00*iFp01*iFp01*iFp22+2.0*Dbar44*iFp00*iFp01*iFp01*iFp20+2.0*Dbar56*iFp01*iFp02*iFp02*iFp20+Dbar35*iFp02*iFp02*iFp02*iFp20+Dbar33*iFp02*iFp02*iFp02*iFp22+2.0*Dbar46*iFp01*iFp01*iFp02*iFp20+Dbar36*iFp02*iFp02*iFp02*iFp21+3.0*Dbar15*iFp00*iFp00*iFp20*iFp02+Dbar26*iFp01*iFp01*iFp01*iFp22+Dbar11*iFp00*iFp00*iFp00*iFp20+Dbar24*iFp01*iFp01*iFp01*iFp20+2.0*Dbar34*iFp00*iFp01*iFp22*iFp02+4.0*Dbar45*iFp00*iFp01*iFp20*iFp02+2.0*Dbar25*iFp00*iFp02*iFp21*iFp01+2.0*Dbar16*iFp01*iFp02*iFp20*iFp00+Dbar15*iFp00*iFp00*iFp00*iFp22+4.0*Dbar46*iFp00*iFp01*iFp21*iFp02+4.0*Dbar56*iFp00*iFp02*iFp22*iFp01+Dbar14*iFp00*iFp00*iFp00*iFp21;
  D[0][4] = D[4][0] = t0;
  s2 = Dbar12*iFp00*iFp00*iFp21*iFp11+Dbar15*iFp00*iFp00*iFp20*iFp12+Dbar16*iFp00*iFp00*iFp21*iFp12+Dbar16*iFp00*iFp00*iFp22*iFp11+Dbar14*iFp00*iFp00*iFp21*iFp10+Dbar11*iFp00*iFp00*iFp20*iFp10+Dbar14*iFp00*iFp00*iFp20*iFp11+Dbar25*iFp01*iFp01*iFp20*iFp12+Dbar24*iFp01*iFp01*iFp21*iFp10+Dbar12*iFp01*iFp01*iFp20*iFp10+Dbar15*iFp00*iFp00*iFp22*iFp10+Dbar13*iFp00*iFp00*iFp22*iFp12+Dbar23*iFp01*iFp01*iFp22*iFp12;
  s1 = s2+Dbar25*iFp01*iFp01*iFp22*iFp10+Dbar26*iFp01*iFp01*iFp22*iFp11+Dbar22*iFp01*iFp01*iFp21*iFp11+Dbar26*iFp01*iFp01*iFp21*iFp12+Dbar24*iFp01*iFp01*iFp20*iFp11+Dbar35*iFp02*iFp02*iFp20*iFp12+Dbar13*iFp02*iFp02*iFp20*iFp10+Dbar34*iFp02*iFp02*iFp20*iFp11+Dbar34*iFp02*iFp02*iFp21*iFp10+Dbar36*iFp02*iFp02*iFp22*iFp11+Dbar33*iFp02*iFp02*iFp22*iFp12+Dbar23*iFp02*iFp02*iFp21*iFp11+Dbar36*iFp02*iFp02*iFp21*iFp12+Dbar35*iFp02*iFp02*iFp22*iFp10;
  s2 = s1+2.0*Dbar36*iFp01*iFp02*iFp22*iFp12+2.0*Dbar56*iFp00*iFp02*iFp22*iFp11+2.0*Dbar44*iFp00*iFp01*iFp21*iFp10+2.0*Dbar24*iFp00*iFp01*iFp21*iFp11+2.0*Dbar45*iFp00*iFp02*iFp21*iFp10+2.0*Dbar44*iFp00*iFp01*iFp20*iFp11+2.0*Dbar25*iFp00*iFp02*iFp21*iFp11+2.0*Dbar66*iFp01*iFp02*iFp22*iFp11+2.0*Dbar15*iFp00*iFp02*iFp20*iFp10+2.0*Dbar55*iFp00*iFp02*iFp20*iFp12+2.0*Dbar45*iFp00*iFp02*iFp20*iFp11+2.0*Dbar56*iFp01*iFp02*iFp22*iFp10+2.0*Dbar46*iFp00*iFp01*iFp22*iFp11;
  t0 = s2+2.0*Dbar66*iFp01*iFp02*iFp21*iFp12+2.0*Dbar46*iFp01*iFp02*iFp21*iFp10+2.0*Dbar34*iFp00*iFp01*iFp22*iFp12+2.0*Dbar14*iFp00*iFp01*iFp20*iFp10+2.0*Dbar45*iFp00*iFp01*iFp20*iFp12+2.0*Dbar56*iFp00*iFp02*iFp21*iFp12+2.0*Dbar55*iFp00*iFp02*iFp22*iFp10+2.0*Dbar16*iFp01*iFp02*iFp20*iFp10+2.0*Dbar35*iFp00*iFp02*iFp22*iFp12+2.0*Dbar46*iFp00*iFp01*iFp21*iFp12+2.0*Dbar45*iFp00*iFp01*iFp22*iFp10+2.0*Dbar56*iFp01*iFp02*iFp20*iFp12+2.0*Dbar46*iFp01*iFp02*iFp20*iFp11+2.0*Dbar26*iFp01*iFp02*iFp21*iFp11;
  D[0][5] = D[5][0] = t0;
  t0 = Dbar11*iFp10*iFp10*iFp10*iFp10+4.0*Dbar14*iFp10*iFp10*iFp10*iFp11+4.0*Dbar15*iFp10*iFp10*iFp10*iFp12+2.0*Dbar12*iFp10*iFp10*iFp11*iFp11+4.0*Dbar16*iFp10*iFp10*iFp11*iFp12+2.0*Dbar13*iFp10*iFp10*iFp12*iFp12+4.0*Dbar44*iFp11*iFp11*iFp10*iFp10+8.0*Dbar45*iFp10*iFp10*iFp11*iFp12+4.0*Dbar24*iFp11*iFp11*iFp11*iFp10+8.0*Dbar46*iFp10*iFp11*iFp11*iFp12+4.0*Dbar34*iFp10*iFp11*iFp12*iFp12+4.0*Dbar55*iFp10*iFp10*iFp12*iFp12+4.0*Dbar25*iFp10*iFp12*iFp11*iFp11+8.0*Dbar56*iFp10*iFp12*iFp12*iFp11+4.0*Dbar35*iFp10*iFp12*iFp12*iFp12+Dbar22*iFp11*iFp11*iFp11*iFp11+4.0*Dbar26*iFp11*iFp11*iFp11*iFp12+2.0*Dbar23*iFp11*iFp11*iFp12*iFp12+4.0*Dbar66*iFp11*iFp11*iFp12*iFp12+4.0*Dbar36*iFp11*iFp12*iFp12*iFp12+Dbar33*iFp12*iFp12*iFp12*iFp12;
  D[1][1] = t0;
  s1 = Dbar12*iFp10*iFp10*iFp21*iFp21+Dbar13*iFp10*iFp10*iFp22*iFp22+2.0*Dbar16*iFp10*iFp10*iFp21*iFp22+Dbar22*iFp11*iFp11*iFp21*iFp21+2.0*Dbar14*iFp10*iFp11*iFp20*iFp20+2.0*Dbar24*iFp10*iFp11*iFp21*iFp21+2.0*Dbar15*iFp10*iFp12*iFp20*iFp20+2.0*Dbar34*iFp10*iFp11*iFp22*iFp22+Dbar11*iFp10*iFp10*iFp20*iFp20+2.0*Dbar15*iFp10*iFp10*iFp20*iFp22+2.0*Dbar14*iFp10*iFp10*iFp20*iFp21+2.0*Dbar25*iFp11*iFp11*iFp20*iFp22+Dbar33*iFp12*iFp12*iFp22*iFp22+Dbar23*iFp12*iFp12*iFp21*iFp21+Dbar12*iFp11*iFp11*iFp20*iFp20+Dbar13*iFp12*iFp12*iFp20*iFp20+2.0*Dbar36*iFp12*iFp12*iFp21*iFp22+2.0*Dbar24*iFp11*iFp11*iFp20*iFp21;
  t0 = s1+2.0*Dbar36*iFp11*iFp12*iFp22*iFp22+2.0*Dbar26*iFp11*iFp12*iFp21*iFp21+2.0*Dbar34*iFp12*iFp12*iFp20*iFp21+2.0*Dbar35*iFp12*iFp12*iFp20*iFp22+2.0*Dbar26*iFp11*iFp11*iFp21*iFp22+2.0*Dbar35*iFp10*iFp12*iFp22*iFp22+2.0*Dbar16*iFp11*iFp12*iFp20*iFp20+4.0*Dbar46*iFp11*iFp12*iFp20*iFp21+4.0*Dbar56*iFp11*iFp12*iFp20*iFp22+4.0*Dbar66*iFp11*iFp12*iFp21*iFp22+2.0*Dbar25*iFp10*iFp12*iFp21*iFp21+4.0*Dbar44*iFp10*iFp11*iFp20*iFp21+4.0*Dbar56*iFp10*iFp12*iFp21*iFp22+4.0*Dbar55*iFp10*iFp12*iFp20*iFp22+4.0*Dbar45*iFp10*iFp12*iFp20*iFp21+4.0*Dbar46*iFp10*iFp11*iFp21*iFp22+4.0*Dbar45*iFp10*iFp11*iFp20*iFp22+Dbar23*iFp11*iFp11*iFp22*iFp22;
  D[1][2] = D[2][1] = t0;
  s1 = Dbar12*iFp10*iFp10*iFp11*iFp01+Dbar13*iFp10*iFp10*iFp12*iFp02+Dbar25*iFp11*iFp11*iFp12*iFp00+Dbar25*iFp11*iFp11*iFp10*iFp02+Dbar23*iFp11*iFp11*iFp12*iFp02+Dbar34*iFp12*iFp12*iFp11*iFp00+Dbar23*iFp12*iFp12*iFp11*iFp01+Dbar12*iFp11*iFp11*iFp10*iFp00+Dbar16*iFp10*iFp10*iFp12*iFp01+Dbar36*iFp12*iFp12*iFp12*iFp01+Dbar35*iFp12*iFp12*iFp12*iFp00+Dbar33*iFp12*iFp12*iFp12*iFp02+2.0*Dbar45*iFp10*iFp10*iFp11*iFp02+Dbar26*iFp11*iFp11*iFp11*iFp02+Dbar24*iFp11*iFp11*iFp11*iFp00+Dbar22*iFp11*iFp11*iFp11*iFp01+Dbar11*iFp10*iFp10*iFp10*iFp00+Dbar14*iFp10*iFp10*iFp10*iFp01+Dbar15*iFp10*iFp10*iFp10*iFp02+2.0*Dbar44*iFp10*iFp11*iFp11*iFp00+3.0*Dbar24*iFp10*iFp11*iFp11*iFp01+2.0*Dbar46*iFp10*iFp11*iFp11*iFp02;
  s2 = s1+3.0*Dbar35*iFp10*iFp12*iFp12*iFp02+2.0*Dbar56*iFp10*iFp12*iFp12*iFp01+2.0*Dbar66*iFp11*iFp12*iFp12*iFp01+2.0*Dbar55*iFp10*iFp12*iFp12*iFp00+3.0*Dbar36*iFp11*iFp12*iFp12*iFp02+Dbar16*iFp10*iFp10*iFp11*iFp02+Dbar34*iFp12*iFp12*iFp10*iFp01+Dbar13*iFp12*iFp12*iFp10*iFp00+2.0*Dbar56*iFp11*iFp12*iFp12*iFp00+2.0*Dbar46*iFp11*iFp11*iFp12*iFp00+3.0*Dbar26*iFp11*iFp11*iFp12*iFp01;
  t0 = s2+2.0*Dbar55*iFp10*iFp10*iFp12*iFp02+2.0*Dbar45*iFp10*iFp10*iFp12*iFp01+3.0*Dbar15*iFp10*iFp10*iFp12*iFp00+3.0*Dbar14*iFp10*iFp10*iFp11*iFp00+2.0*Dbar44*iFp10*iFp10*iFp11*iFp01+4.0*Dbar45*iFp10*iFp11*iFp12*iFp00+4.0*Dbar46*iFp10*iFp11*iFp12*iFp01+4.0*Dbar56*iFp10*iFp12*iFp11*iFp02+2.0*Dbar34*iFp10*iFp11*iFp12*iFp02+2.0*Dbar25*iFp10*iFp12*iFp11*iFp01+2.0*Dbar16*iFp11*iFp12*iFp10*iFp00+2.0*Dbar66*iFp11*iFp11*iFp12*iFp02;
  D[1][3] = D[3][1] = t0;
  s2 = Dbar35*iFp12*iFp12*iFp22*iFp00+Dbar26*iFp11*iFp11*iFp21*iFp02+Dbar23*iFp11*iFp11*iFp22*iFp02+Dbar26*iFp11*iFp11*iFp22*iFp01+Dbar13*iFp12*iFp12*iFp20*iFp00+Dbar34*iFp12*iFp12*iFp21*iFp00+Dbar14*iFp10*iFp10*iFp20*iFp01+Dbar36*iFp12*iFp12*iFp21*iFp02+Dbar25*iFp11*iFp11*iFp20*iFp02+Dbar23*iFp12*iFp12*iFp21*iFp01+Dbar15*iFp10*iFp10*iFp20*iFp02+Dbar12*iFp10*iFp10*iFp21*iFp01+Dbar14*iFp10*iFp10*iFp21*iFp00;
  s1 = s2+Dbar16*iFp10*iFp10*iFp21*iFp02+Dbar12*iFp11*iFp11*iFp20*iFp00+Dbar35*iFp12*iFp12*iFp20*iFp02+Dbar22*iFp11*iFp11*iFp21*iFp01+Dbar33*iFp12*iFp12*iFp22*iFp02+Dbar36*iFp12*iFp12*iFp22*iFp01+Dbar11*iFp10*iFp10*iFp20*iFp00+Dbar15*iFp10*iFp10*iFp22*iFp00+Dbar25*iFp11*iFp11*iFp22*iFp00+Dbar24*iFp11*iFp11*iFp20*iFp01+Dbar13*iFp10*iFp10*iFp22*iFp02+Dbar16*iFp10*iFp10*iFp22*iFp01+Dbar24*iFp11*iFp11*iFp21*iFp00+Dbar34*iFp12*iFp12*iFp20*iFp01;
  s2 = s1+2.0*Dbar45*iFp10*iFp11*iFp22*iFp00+2.0*Dbar66*iFp11*iFp12*iFp21*iFp02+2.0*Dbar14*iFp10*iFp11*iFp20*iFp00+2.0*Dbar34*iFp10*iFp11*iFp22*iFp02+2.0*Dbar15*iFp10*iFp12*iFp20*iFp00+2.0*Dbar55*iFp10*iFp12*iFp20*iFp02+2.0*Dbar45*iFp10*iFp12*iFp21*iFp00+2.0*Dbar56*iFp10*iFp12*iFp21*iFp02+2.0*Dbar56*iFp11*iFp12*iFp22*iFp00+2.0*Dbar36*iFp11*iFp12*iFp22*iFp02+2.0*Dbar25*iFp10*iFp12*iFp21*iFp01+2.0*Dbar66*iFp11*iFp12*iFp22*iFp01+2.0*Dbar35*iFp10*iFp12*iFp22*iFp02;
  t0 = s2+2.0*Dbar45*iFp10*iFp12*iFp20*iFp01+2.0*Dbar56*iFp10*iFp12*iFp22*iFp01+2.0*Dbar55*iFp10*iFp12*iFp22*iFp00+2.0*Dbar44*iFp10*iFp11*iFp20*iFp01+2.0*Dbar45*iFp10*iFp11*iFp20*iFp02+2.0*Dbar16*iFp11*iFp12*iFp20*iFp00+2.0*Dbar46*iFp11*iFp12*iFp21*iFp00+2.0*Dbar44*iFp10*iFp11*iFp21*iFp00+2.0*Dbar46*iFp11*iFp12*iFp20*iFp01+2.0*Dbar56*iFp11*iFp12*iFp20*iFp02+2.0*Dbar26*iFp11*iFp12*iFp21*iFp01+2.0*Dbar24*iFp10*iFp11*iFp21*iFp01+2.0*Dbar46*iFp10*iFp11*iFp22*iFp01+2.0*Dbar46*iFp10*iFp11*iFp21*iFp02;
  D[1][4] = D[4][1] = t0;
  s1 = Dbar22*iFp11*iFp11*iFp11*iFp21+Dbar24*iFp11*iFp11*iFp11*iFp20+Dbar35*iFp12*iFp12*iFp12*iFp20+Dbar15*iFp10*iFp10*iFp10*iFp22+Dbar14*iFp10*iFp10*iFp10*iFp21+Dbar33*iFp12*iFp12*iFp12*iFp22+Dbar11*iFp10*iFp10*iFp10*iFp20+Dbar36*iFp12*iFp12*iFp12*iFp21+Dbar26*iFp11*iFp11*iFp11*iFp22+Dbar16*iFp10*iFp10*iFp21*iFp12+Dbar13*iFp10*iFp10*iFp22*iFp12+Dbar12*iFp10*iFp10*iFp21*iFp11+Dbar34*iFp12*iFp12*iFp21*iFp10+Dbar34*iFp12*iFp12*iFp20*iFp11+3.0*Dbar24*iFp10*iFp11*iFp11*iFp21+2.0*Dbar46*iFp11*iFp11*iFp12*iFp20+3.0*Dbar26*iFp11*iFp11*iFp21*iFp12+2.0*Dbar44*iFp10*iFp10*iFp11*iFp21+2.0*Dbar44*iFp10*iFp11*iFp11*iFp20+3.0*Dbar35*iFp10*iFp12*iFp12*iFp22+2.0*Dbar56*iFp10*iFp12*iFp12*iFp21+2.0*Dbar55*iFp10*iFp10*iFp12*iFp22;
  s2 = s1+2.0*Dbar45*iFp10*iFp10*iFp12*iFp21+Dbar23*iFp12*iFp12*iFp21*iFp11+Dbar25*iFp11*iFp11*iFp20*iFp12+Dbar25*iFp11*iFp11*iFp22*iFp10+Dbar13*iFp12*iFp12*iFp20*iFp10+Dbar23*iFp11*iFp11*iFp22*iFp12+Dbar16*iFp10*iFp10*iFp22*iFp11+Dbar12*iFp11*iFp11*iFp20*iFp10+2.0*Dbar55*iFp10*iFp12*iFp12*iFp20+2.0*Dbar66*iFp11*iFp11*iFp12*iFp22+3.0*Dbar36*iFp11*iFp12*iFp12*iFp22;
  t0 = s2+3.0*Dbar15*iFp10*iFp10*iFp20*iFp12+2.0*Dbar45*iFp10*iFp10*iFp11*iFp22+2.0*Dbar66*iFp11*iFp12*iFp12*iFp21+3.0*Dbar14*iFp10*iFp10*iFp20*iFp11+2.0*Dbar46*iFp10*iFp11*iFp11*iFp22+2.0*Dbar56*iFp11*iFp12*iFp12*iFp20+2.0*Dbar16*iFp11*iFp12*iFp20*iFp10+4.0*Dbar46*iFp10*iFp11*iFp21*iFp12+2.0*Dbar34*iFp10*iFp11*iFp22*iFp12+4.0*Dbar56*iFp10*iFp12*iFp22*iFp11+4.0*Dbar45*iFp10*iFp11*iFp20*iFp12+2.0*Dbar25*iFp10*iFp12*iFp21*iFp11;
  D[1][5] = D[5][1] = t0;
  t0 = Dbar11*iFp20*iFp20*iFp20*iFp20+4.0*Dbar14*iFp20*iFp20*iFp20*iFp21+4.0*Dbar15*iFp20*iFp20*iFp20*iFp22+2.0*Dbar12*iFp20*iFp20*iFp21*iFp21+4.0*Dbar16*iFp20*iFp20*iFp21*iFp22+2.0*Dbar13*iFp20*iFp20*iFp22*iFp22+4.0*Dbar44*iFp20*iFp20*iFp21*iFp21+8.0*Dbar45*iFp20*iFp20*iFp21*iFp22+4.0*Dbar24*iFp20*iFp21*iFp21*iFp21+8.0*Dbar46*iFp20*iFp21*iFp21*iFp22+4.0*Dbar34*iFp20*iFp21*iFp22*iFp22+4.0*Dbar55*iFp20*iFp20*iFp22*iFp22+4.0*Dbar25*iFp20*iFp22*iFp21*iFp21+8.0*Dbar56*iFp20*iFp22*iFp22*iFp21+4.0*Dbar35*iFp20*iFp22*iFp22*iFp22+Dbar22*iFp21*iFp21*iFp21*iFp21+4.0*Dbar26*iFp21*iFp21*iFp21*iFp22+2.0*Dbar23*iFp21*iFp21*iFp22*iFp22+4.0*Dbar66*iFp21*iFp21*iFp22*iFp22+4.0*Dbar36*iFp22*iFp22*iFp22*iFp21+Dbar33*iFp22*iFp22*iFp22*iFp22;
  D[2][2] = t0;
  s2 = Dbar23*iFp21*iFp21*iFp12*iFp02+Dbar16*iFp20*iFp20*iFp02*iFp11+Dbar13*iFp20*iFp20*iFp02*iFp12+Dbar14*iFp20*iFp20*iFp00*iFp11+Dbar15*iFp20*iFp20*iFp00*iFp12+Dbar11*iFp20*iFp20*iFp00*iFp10+Dbar14*iFp20*iFp20*iFp01*iFp10+Dbar24*iFp21*iFp21*iFp10*iFp01+Dbar35*iFp22*iFp22*iFp12*iFp00+Dbar34*iFp22*iFp22*iFp11*iFp00+Dbar15*iFp20*iFp20*iFp02*iFp10+Dbar23*iFp22*iFp22*iFp11*iFp01+Dbar35*iFp22*iFp22*iFp10*iFp02;
  s1 = s2+Dbar22*iFp21*iFp21*iFp11*iFp01+Dbar12*iFp20*iFp20*iFp01*iFp11+Dbar16*iFp20*iFp20*iFp01*iFp12+Dbar25*iFp21*iFp21*iFp10*iFp02+Dbar26*iFp21*iFp21*iFp12*iFp01+Dbar34*iFp22*iFp22*iFp10*iFp01+Dbar13*iFp22*iFp22*iFp10*iFp00+Dbar36*iFp22*iFp22*iFp11*iFp02+Dbar12*iFp21*iFp21*iFp10*iFp00+2.0*Dbar45*iFp20*iFp21*iFp02*iFp10+2.0*Dbar46*iFp20*iFp21*iFp01*iFp12+2.0*Dbar15*iFp20*iFp22*iFp00*iFp10+2.0*Dbar34*iFp20*iFp21*iFp02*iFp12+2.0*Dbar55*iFp20*iFp22*iFp00*iFp12;
  s2 = s1+2.0*Dbar45*iFp20*iFp22*iFp00*iFp11+Dbar36*iFp22*iFp22*iFp12*iFp01+Dbar33*iFp22*iFp22*iFp12*iFp02+Dbar26*iFp21*iFp21*iFp11*iFp02+Dbar25*iFp21*iFp21*iFp12*iFp00+Dbar24*iFp21*iFp21*iFp11*iFp00+2.0*Dbar26*iFp21*iFp22*iFp11*iFp01+2.0*Dbar45*iFp20*iFp22*iFp01*iFp10+2.0*Dbar56*iFp21*iFp22*iFp12*iFp00+2.0*Dbar36*iFp21*iFp22*iFp12*iFp02+2.0*Dbar66*iFp21*iFp22*iFp12*iFp01+2.0*Dbar46*iFp21*iFp22*iFp11*iFp00+2.0*Dbar46*iFp20*iFp21*iFp02*iFp11;
  t0 = s2+2.0*Dbar56*iFp21*iFp22*iFp10*iFp02+2.0*Dbar66*iFp21*iFp22*iFp11*iFp02+2.0*Dbar46*iFp21*iFp22*iFp10*iFp01+2.0*Dbar16*iFp21*iFp22*iFp10*iFp00+2.0*Dbar35*iFp20*iFp22*iFp02*iFp12+2.0*Dbar25*iFp20*iFp22*iFp01*iFp11+2.0*Dbar14*iFp20*iFp21*iFp10*iFp00+2.0*Dbar56*iFp20*iFp22*iFp02*iFp11+2.0*Dbar56*iFp20*iFp22*iFp01*iFp12+2.0*Dbar55*iFp20*iFp22*iFp02*iFp10+2.0*Dbar45*iFp20*iFp21*iFp00*iFp12+2.0*Dbar44*iFp20*iFp21*iFp00*iFp11+2.0*Dbar24*iFp20*iFp21*iFp01*iFp11+2.0*Dbar44*iFp20*iFp21*iFp01*iFp10;
  D[2][3] = D[3][2] = t0;
  s1 = Dbar26*iFp21*iFp21*iFp21*iFp02+Dbar22*iFp21*iFp21*iFp21*iFp01+Dbar33*iFp22*iFp22*iFp22*iFp02+Dbar35*iFp22*iFp22*iFp22*iFp00+Dbar36*iFp22*iFp22*iFp22*iFp01+Dbar11*iFp20*iFp20*iFp20*iFp00+Dbar15*iFp20*iFp20*iFp20*iFp02+Dbar14*iFp20*iFp20*iFp20*iFp01+Dbar12*iFp20*iFp20*iFp21*iFp01+Dbar16*iFp20*iFp20*iFp21*iFp02+Dbar25*iFp21*iFp21*iFp22*iFp00+Dbar23*iFp21*iFp21*iFp22*iFp02+Dbar34*iFp22*iFp22*iFp21*iFp00+Dbar12*iFp21*iFp21*iFp20*iFp00+3.0*Dbar36*iFp21*iFp22*iFp22*iFp02+2.0*Dbar46*iFp22*iFp21*iFp21*iFp00+2.0*Dbar56*iFp22*iFp22*iFp21*iFp00+2.0*Dbar66*iFp22*iFp21*iFp21*iFp02+3.0*Dbar26*iFp21*iFp21*iFp22*iFp01+3.0*Dbar35*iFp22*iFp22*iFp20*iFp02+2.0*Dbar56*iFp22*iFp22*iFp20*iFp01+2.0*Dbar55*iFp22*iFp22*iFp20*iFp00;
  t0 = s1+3.0*Dbar15*iFp20*iFp20*iFp22*iFp00+2.0*Dbar55*iFp22*iFp20*iFp20*iFp02+2.0*Dbar45*iFp22*iFp20*iFp20*iFp01+Dbar23*iFp22*iFp22*iFp21*iFp01+Dbar16*iFp20*iFp20*iFp22*iFp01+Dbar13*iFp20*iFp20*iFp22*iFp02+Dbar25*iFp21*iFp21*iFp20*iFp02+Dbar34*iFp22*iFp22*iFp20*iFp01+Dbar13*iFp22*iFp22*iFp20*iFp00+2.0*Dbar44*iFp20*iFp20*iFp21*iFp01+2.0*Dbar45*iFp20*iFp20*iFp21*iFp02+2.0*Dbar44*iFp20*iFp21*iFp21*iFp00+3.0*Dbar24*iFp20*iFp21*iFp21*iFp01+2.0*Dbar46*iFp20*iFp21*iFp21*iFp02+3.0*Dbar14*iFp20*iFp20*iFp21*iFp00+2.0*Dbar66*iFp22*iFp22*iFp21*iFp01+Dbar24*iFp21*iFp21*iFp21*iFp00+4.0*Dbar45*iFp20*iFp21*iFp22*iFp00+4.0*Dbar56*iFp22*iFp20*iFp21*iFp02+2.0*Dbar25*iFp22*iFp20*iFp21*iFp01+2.0*Dbar34*iFp20*iFp21*iFp22*iFp02+2.0*Dbar16*iFp22*iFp21*iFp20*iFp00+4.0*Dbar46*iFp20*iFp21*iFp22*iFp01;
  D[2][4] = D[4][2] = t0;
  s1 = Dbar24*iFp21*iFp21*iFp21*iFp10+Dbar11*iFp20*iFp20*iFp20*iFp10+Dbar14*iFp20*iFp20*iFp20*iFp11+Dbar15*iFp20*iFp20*iFp20*iFp12+Dbar26*iFp21*iFp21*iFp21*iFp12+Dbar36*iFp22*iFp22*iFp22*iFp11+Dbar34*iFp22*iFp22*iFp20*iFp11+Dbar25*iFp21*iFp21*iFp20*iFp12+Dbar34*iFp22*iFp22*iFp21*iFp10+Dbar23*iFp22*iFp22*iFp21*iFp11+Dbar12*iFp21*iFp21*iFp20*iFp10+Dbar13*iFp22*iFp22*iFp20*iFp10+2.0*Dbar56*iFp20*iFp22*iFp22*iFp11+3.0*Dbar35*iFp22*iFp22*iFp20*iFp12+2.0*Dbar55*iFp20*iFp22*iFp22*iFp10+2.0*Dbar45*iFp20*iFp20*iFp22*iFp11+2.0*Dbar55*iFp20*iFp20*iFp22*iFp12+3.0*Dbar24*iFp20*iFp21*iFp21*iFp11+2.0*Dbar56*iFp22*iFp22*iFp21*iFp10+2.0*Dbar46*iFp20*iFp21*iFp21*iFp12+2.0*Dbar66*iFp22*iFp22*iFp21*iFp11+3.0*Dbar36*iFp22*iFp22*iFp21*iFp12;
  t0 = s1+Dbar16*iFp20*iFp20*iFp21*iFp12+3.0*Dbar15*iFp20*iFp20*iFp22*iFp10+Dbar16*iFp20*iFp20*iFp22*iFp11+Dbar13*iFp20*iFp20*iFp22*iFp12+2.0*Dbar46*iFp22*iFp21*iFp21*iFp10+2.0*Dbar44*iFp20*iFp20*iFp21*iFp11+2.0*Dbar45*iFp20*iFp20*iFp21*iFp12+2.0*Dbar66*iFp22*iFp21*iFp21*iFp12+2.0*Dbar44*iFp20*iFp21*iFp21*iFp10+Dbar23*iFp21*iFp21*iFp22*iFp12+3.0*Dbar14*iFp20*iFp20*iFp21*iFp10+Dbar12*iFp20*iFp20*iFp21*iFp11+3.0*Dbar26*iFp22*iFp21*iFp21*iFp11+Dbar25*iFp21*iFp21*iFp22*iFp10+2.0*Dbar34*iFp20*iFp21*iFp22*iFp12+2.0*Dbar16*iFp21*iFp22*iFp20*iFp10+Dbar35*iFp22*iFp22*iFp22*iFp10+4.0*Dbar45*iFp20*iFp21*iFp22*iFp10+Dbar22*iFp21*iFp21*iFp21*iFp11+4.0*Dbar56*iFp20*iFp22*iFp21*iFp12+2.0*Dbar25*iFp20*iFp22*iFp21*iFp11+Dbar33*iFp22*iFp22*iFp22*iFp12+4.0*Dbar46*iFp20*iFp21*iFp22*iFp11;
  D[2][5] = D[5][2] = t0;
  s1 = Dbar44*iFp01*iFp01*iFp10*iFp10+Dbar55*iFp00*iFp00*iFp12*iFp12+Dbar55*iFp02*iFp02*iFp10*iFp10+Dbar44*iFp00*iFp00*iFp11*iFp11+Dbar66*iFp01*iFp01*iFp12*iFp12+Dbar11*iFp00*iFp00*iFp10*iFp10+Dbar33*iFp02*iFp02*iFp12*iFp12+Dbar22*iFp01*iFp01*iFp11*iFp11+2.0*Dbar36*iFp01*iFp12*iFp12*iFp02+2.0*Dbar24*iFp00*iFp11*iFp11*iFp01+2.0*Dbar46*iFp00*iFp11*iFp11*iFp02+2.0*Dbar45*iFp00*iFp00*iFp11*iFp12+2.0*Dbar15*iFp00*iFp00*iFp10*iFp12+2.0*Dbar14*iFp00*iFp10*iFp10*iFp01+2.0*Dbar15*iFp00*iFp10*iFp10*iFp02+2.0*Dbar14*iFp00*iFp00*iFp10*iFp11+2.0*Dbar26*iFp01*iFp11*iFp11*iFp02+2.0*Dbar26*iFp01*iFp01*iFp11*iFp12+2.0*Dbar36*iFp02*iFp02*iFp11*iFp12+2.0*Dbar46*iFp01*iFp01*iFp10*iFp12+2.0*Dbar35*iFp02*iFp02*iFp10*iFp12+2.0*Dbar45*iFp01*iFp10*iFp10*iFp02;
  s2 = s1+2.0*Dbar24*iFp01*iFp01*iFp10*iFp11+2.0*Dbar56*iFp00*iFp12*iFp12*iFp01+2.0*Dbar56*iFp02*iFp02*iFp10*iFp11+2.0*Dbar35*iFp00*iFp12*iFp12*iFp02+2.0*Dbar44*iFp00*iFp11*iFp10*iFp01+2.0*Dbar13*iFp00*iFp10*iFp12*iFp02+2.0*Dbar16*iFp00*iFp10*iFp12*iFp01+2.0*Dbar55*iFp00*iFp12*iFp10*iFp02+2.0*Dbar45*iFp00*iFp12*iFp10*iFp01+2.0*Dbar46*iFp01*iFp10*iFp11*iFp02+2.0*Dbar34*iFp01*iFp10*iFp12*iFp02;
  t0 = s2+2.0*Dbar34*iFp00*iFp11*iFp12*iFp02+2.0*Dbar46*iFp00*iFp11*iFp12*iFp01+2.0*Dbar45*iFp00*iFp11*iFp10*iFp02+2.0*Dbar56*iFp00*iFp12*iFp11*iFp02+2.0*Dbar66*iFp01*iFp12*iFp11*iFp02+2.0*Dbar25*iFp00*iFp12*iFp11*iFp01+2.0*Dbar25*iFp01*iFp11*iFp10*iFp02+2.0*Dbar56*iFp01*iFp12*iFp10*iFp02+2.0*Dbar23*iFp01*iFp11*iFp12*iFp02+2.0*Dbar12*iFp00*iFp10*iFp11*iFp01+2.0*Dbar16*iFp00*iFp10*iFp11*iFp02+Dbar66*iFp02*iFp02*iFp11*iFp11;
  D[3][3] = t0;
  s2 = Dbar11*iFp00*iFp00*iFp20*iFp10+Dbar14*iFp00*iFp00*iFp20*iFp11+Dbar24*iFp01*iFp01*iFp21*iFp10+Dbar15*iFp00*iFp00*iFp22*iFp10+Dbar26*iFp01*iFp01*iFp22*iFp11+Dbar22*iFp01*iFp01*iFp21*iFp11+Dbar26*iFp01*iFp01*iFp21*iFp12+Dbar24*iFp01*iFp01*iFp20*iFp11+Dbar35*iFp02*iFp02*iFp20*iFp12+Dbar36*iFp02*iFp02*iFp22*iFp11+Dbar15*iFp00*iFp00*iFp20*iFp12+Dbar14*iFp00*iFp00*iFp21*iFp10+Dbar56*iFp10*iFp02*iFp02*iFp21+Dbar46*iFp12*iFp01*iFp01*iFp20+Dbar55*iFp12*iFp00*iFp00*iFp22+Dbar66*iFp12*iFp01*iFp01*iFp22+Dbar46*iFp10*iFp01*iFp01*iFp22+Dbar55*iFp10*iFp02*iFp02*iFp20;
  s1 = s2+Dbar45*iFp12*iFp00*iFp00*iFp21+Dbar45*iFp11*iFp00*iFp00*iFp22+Dbar44*iFp11*iFp00*iFp00*iFp21+Dbar66*iFp11*iFp02*iFp02*iFp21+Dbar56*iFp11*iFp02*iFp02*iFp20+Dbar44*iFp10*iFp01*iFp01*iFp20+Dbar35*iFp02*iFp02*iFp22*iFp10+Dbar56*iFp00*iFp02*iFp22*iFp11+2.0*Dbar24*iFp00*iFp01*iFp21*iFp11+Dbar44*iFp00*iFp01*iFp21*iFp10+Dbar55*iFp00*iFp02*iFp22*iFp10+Dbar56*iFp00*iFp02*iFp21*iFp12+Dbar45*iFp00*iFp01*iFp20*iFp12+Dbar44*iFp00*iFp01*iFp20*iFp11+Dbar45*iFp00*iFp02*iFp21*iFp10+2.0*Dbar14*iFp00*iFp01*iFp20*iFp10+Dbar45*iFp00*iFp02*iFp20*iFp11+Dbar55*iFp00*iFp02*iFp20*iFp12;
  s2 = 2.0*Dbar15*iFp00*iFp02*iFp20*iFp10+Dbar66*iFp01*iFp02*iFp22*iFp11+Dbar66*iFp01*iFp02*iFp21*iFp12+Dbar46*iFp00*iFp01*iFp22*iFp11+Dbar56*iFp01*iFp02*iFp22*iFp10+Dbar46*iFp01*iFp02*iFp21*iFp10+2.0*Dbar26*iFp01*iFp02*iFp21*iFp11+Dbar46*iFp01*iFp02*iFp20*iFp11+Dbar56*iFp01*iFp02*iFp20*iFp12+Dbar33*iFp02*iFp02*iFp22*iFp12+Dbar45*iFp00*iFp01*iFp22*iFp10+Dbar46*iFp00*iFp01*iFp21*iFp12+2.0*Dbar35*iFp00*iFp02*iFp22*iFp12+2.0*Dbar36*iFp01*iFp02*iFp22*iFp12+Dbar34*iFp12*iFp02*iFp20*iFp01+Dbar34*iFp12*iFp02*iFp21*iFp00+Dbar16*iFp12*iFp01*iFp20*iFp00+Dbar13*iFp12*iFp02*iFp20*iFp00;
  t0 = s2+Dbar34*iFp10*iFp01*iFp22*iFp02+Dbar25*iFp10*iFp02*iFp21*iFp01+Dbar25*iFp12*iFp00*iFp21*iFp01+Dbar36*iFp02*iFp02*iFp21*iFp12+Dbar25*iFp11*iFp01*iFp22*iFp00+Dbar23*iFp11*iFp01*iFp22*iFp02+Dbar12*iFp11*iFp01*iFp20*iFp00+Dbar23*iFp12*iFp02*iFp21*iFp01+Dbar34*iFp11*iFp00*iFp22*iFp02+Dbar16*iFp11*iFp02*iFp20*iFp00+Dbar16*iFp10*iFp00*iFp21*iFp02+Dbar16*iFp10*iFp00*iFp22*iFp01+Dbar13*iFp10*iFp00*iFp22*iFp02+Dbar25*iFp11*iFp01*iFp20*iFp02+Dbar12*iFp10*iFp00*iFp21*iFp01+2.0*Dbar45*iFp10*iFp01*iFp20*iFp02+2.0*Dbar56*iFp12*iFp00*iFp22*iFp01+2.0*Dbar46*iFp11*iFp00*iFp21*iFp02+s1;
  D[3][4] = D[4][3] = t0;
  s2 = Dbar24*iFp11*iFp11*iFp20*iFp01+Dbar56*iFp12*iFp12*iFp01*iFp20+Dbar46*iFp11*iFp11*iFp00*iFp22+Dbar46*iFp11*iFp11*iFp02*iFp20+Dbar44*iFp10*iFp10*iFp01*iFp21+Dbar44*iFp11*iFp11*iFp00*iFp20+Dbar66*iFp12*iFp12*iFp01*iFp21+Dbar45*iFp10*iFp10*iFp02*iFp21+2.0*Dbar46*iFp10*iFp01*iFp21*iFp12+Dbar56*iFp11*iFp12*iFp22*iFp00+Dbar55*iFp10*iFp12*iFp20*iFp02+Dbar45*iFp10*iFp12*iFp21*iFp00+Dbar45*iFp10*iFp11*iFp22*iFp00+Dbar66*iFp11*iFp12*iFp22*iFp01+2.0*Dbar35*iFp10*iFp12*iFp22*iFp02+2.0*Dbar36*iFp11*iFp12*iFp22*iFp02+Dbar45*iFp10*iFp12*iFp20*iFp01+Dbar56*iFp10*iFp12*iFp21*iFp02;
  s1 = s2+2.0*Dbar15*iFp10*iFp12*iFp20*iFp00+Dbar66*iFp11*iFp12*iFp21*iFp02+2.0*Dbar45*iFp11*iFp00*iFp20*iFp12+2.0*Dbar56*iFp10*iFp02*iFp22*iFp11+Dbar44*iFp10*iFp11*iFp20*iFp01+2.0*Dbar14*iFp10*iFp11*iFp20*iFp00+Dbar55*iFp10*iFp12*iFp22*iFp00+Dbar56*iFp10*iFp12*iFp22*iFp01+Dbar56*iFp11*iFp12*iFp20*iFp02+Dbar46*iFp11*iFp12*iFp21*iFp00+Dbar44*iFp10*iFp11*iFp21*iFp00+Dbar46*iFp11*iFp12*iFp20*iFp01+Dbar45*iFp10*iFp11*iFp20*iFp02+Dbar34*iFp12*iFp02*iFp20*iFp11+Dbar13*iFp12*iFp02*iFp20*iFp10+Dbar46*iFp10*iFp11*iFp22*iFp01+Dbar46*iFp10*iFp11*iFp21*iFp02+2.0*Dbar26*iFp11*iFp12*iFp21*iFp01;
  s2 = 2.0*Dbar24*iFp10*iFp11*iFp21*iFp01+Dbar23*iFp11*iFp01*iFp22*iFp12+Dbar25*iFp11*iFp01*iFp22*iFp10+Dbar34*iFp11*iFp00*iFp22*iFp12+Dbar23*iFp12*iFp02*iFp21*iFp11+Dbar34*iFp12*iFp02*iFp21*iFp10+Dbar16*iFp12*iFp01*iFp20*iFp10+Dbar25*iFp11*iFp01*iFp20*iFp12+Dbar16*iFp10*iFp00*iFp21*iFp12+Dbar12*iFp10*iFp00*iFp21*iFp11+Dbar12*iFp11*iFp01*iFp20*iFp10+Dbar16*iFp11*iFp02*iFp20*iFp10+Dbar25*iFp12*iFp00*iFp21*iFp11+Dbar13*iFp10*iFp00*iFp22*iFp12+Dbar25*iFp10*iFp02*iFp21*iFp11+Dbar24*iFp11*iFp11*iFp21*iFp00+Dbar45*iFp10*iFp10*iFp01*iFp22+Dbar56*iFp12*iFp12*iFp00*iFp21;
  t0 = s2+Dbar55*iFp12*iFp12*iFp00*iFp20+Dbar15*iFp10*iFp10*iFp20*iFp02+Dbar55*iFp10*iFp10*iFp02*iFp22+s1+Dbar16*iFp10*iFp00*iFp22*iFp11+Dbar35*iFp12*iFp12*iFp22*iFp00+Dbar66*iFp11*iFp11*iFp02*iFp22+Dbar15*iFp10*iFp10*iFp22*iFp00+Dbar33*iFp12*iFp12*iFp22*iFp02+Dbar26*iFp11*iFp11*iFp21*iFp02+Dbar35*iFp12*iFp12*iFp20*iFp02+Dbar14*iFp10*iFp10*iFp21*iFp00+Dbar34*iFp10*iFp01*iFp22*iFp12+Dbar26*iFp11*iFp11*iFp22*iFp01+Dbar36*iFp12*iFp12*iFp22*iFp01+Dbar36*iFp12*iFp12*iFp21*iFp02+Dbar11*iFp10*iFp10*iFp20*iFp00+Dbar22*iFp11*iFp11*iFp21*iFp01+Dbar14*iFp10*iFp10*iFp20*iFp01;
  D[3][5] = D[5][3] = t0;
  s1 = Dbar44*iFp21*iFp21*iFp00*iFp00+Dbar66*iFp22*iFp22*iFp01*iFp01+Dbar55*iFp22*iFp22*iFp00*iFp00+2.0*Dbar46*iFp20*iFp01*iFp01*iFp22+2.0*Dbar45*iFp20*iFp20*iFp01*iFp02+2.0*Dbar45*iFp21*iFp00*iFp00*iFp22+2.0*Dbar46*iFp21*iFp21*iFp00*iFp02+2.0*Dbar56*iFp22*iFp22*iFp00*iFp01+2.0*Dbar35*iFp02*iFp02*iFp20*iFp22+2.0*Dbar14*iFp00*iFp00*iFp20*iFp21+2.0*Dbar15*iFp00*iFp00*iFp20*iFp22+2.0*Dbar36*iFp01*iFp02*iFp22*iFp22+2.0*Dbar26*iFp01*iFp02*iFp21*iFp21+2.0*Dbar36*iFp02*iFp02*iFp21*iFp22+2.0*Dbar15*iFp00*iFp02*iFp20*iFp20+2.0*Dbar24*iFp00*iFp01*iFp21*iFp21+2.0*Dbar14*iFp00*iFp01*iFp20*iFp20+2.0*Dbar26*iFp01*iFp01*iFp21*iFp22+2.0*Dbar35*iFp00*iFp02*iFp22*iFp22+2.0*Dbar24*iFp01*iFp01*iFp20*iFp21+2.0*Dbar56*iFp20*iFp02*iFp02*iFp21+2.0*Dbar45*iFp00*iFp02*iFp20*iFp21;
  s2 = s1+2.0*Dbar25*iFp21*iFp01*iFp22*iFp00+2.0*Dbar34*iFp20*iFp01*iFp22*iFp02+2.0*Dbar34*iFp21*iFp00*iFp22*iFp02+2.0*Dbar16*iFp20*iFp00*iFp21*iFp02+2.0*Dbar13*iFp20*iFp00*iFp22*iFp02+2.0*Dbar16*iFp20*iFp00*iFp22*iFp01+2.0*Dbar46*iFp01*iFp02*iFp20*iFp21+2.0*Dbar25*iFp20*iFp02*iFp21*iFp01+2.0*Dbar46*iFp00*iFp01*iFp21*iFp22+2.0*Dbar56*iFp01*iFp02*iFp20*iFp22+2.0*Dbar44*iFp00*iFp01*iFp20*iFp21;
  t0 = s2+2.0*Dbar56*iFp00*iFp02*iFp21*iFp22+2.0*Dbar45*iFp00*iFp01*iFp20*iFp22+2.0*Dbar55*iFp00*iFp02*iFp20*iFp22+2.0*Dbar66*iFp01*iFp02*iFp21*iFp22+Dbar22*iFp01*iFp01*iFp21*iFp21+Dbar33*iFp02*iFp02*iFp22*iFp22+Dbar11*iFp00*iFp00*iFp20*iFp20+Dbar66*iFp21*iFp21*iFp02*iFp02+Dbar55*iFp20*iFp20*iFp02*iFp02+2.0*Dbar12*iFp20*iFp00*iFp21*iFp01+Dbar44*iFp20*iFp20*iFp01*iFp01+2.0*Dbar23*iFp21*iFp01*iFp22*iFp02;
  D[4][4] = t0;
  s2 = Dbar11*iFp20*iFp20*iFp00*iFp10+Dbar14*iFp20*iFp20*iFp01*iFp10+Dbar24*iFp21*iFp21*iFp10*iFp01+Dbar35*iFp22*iFp22*iFp12*iFp00+Dbar22*iFp21*iFp21*iFp11*iFp01+Dbar55*iFp20*iFp20*iFp02*iFp12+Dbar55*iFp22*iFp22*iFp00*iFp10+Dbar56*iFp22*iFp22*iFp00*iFp11+Dbar44*iFp21*iFp21*iFp00*iFp10+Dbar44*iFp20*iFp20*iFp01*iFp11+Dbar45*iFp20*iFp20*iFp01*iFp12+Dbar46*iFp21*iFp21*iFp02*iFp10+Dbar66*iFp21*iFp21*iFp02*iFp12+Dbar56*iFp22*iFp22*iFp01*iFp10+Dbar46*iFp21*iFp21*iFp00*iFp12+Dbar14*iFp20*iFp20*iFp00*iFp11+Dbar15*iFp20*iFp20*iFp00*iFp12+Dbar34*iFp22*iFp02*iFp21*iFp10;
  s1 = s2+Dbar23*iFp22*iFp02*iFp21*iFp11+Dbar34*iFp21*iFp00*iFp22*iFp12+Dbar36*iFp22*iFp22*iFp12*iFp01+Dbar33*iFp22*iFp22*iFp12*iFp02+Dbar35*iFp22*iFp22*iFp10*iFp02+2.0*Dbar56*iFp20*iFp02*iFp21*iFp12+2.0*Dbar45*iFp21*iFp00*iFp22*iFp10+2.0*Dbar36*iFp21*iFp22*iFp12*iFp02+Dbar45*iFp20*iFp22*iFp01*iFp10+2.0*Dbar46*iFp22*iFp01*iFp20*iFp11+Dbar56*iFp21*iFp22*iFp10*iFp02+Dbar66*iFp21*iFp22*iFp11*iFp02+2.0*Dbar26*iFp21*iFp22*iFp11*iFp01+Dbar55*iFp20*iFp22*iFp02*iFp10+Dbar56*iFp20*iFp22*iFp02*iFp11+2.0*Dbar14*iFp20*iFp21*iFp10*iFp00+2.0*Dbar35*iFp20*iFp22*iFp02*iFp12+Dbar44*iFp20*iFp21*iFp01*iFp10;
  s2 = Dbar45*iFp20*iFp21*iFp00*iFp12+Dbar44*iFp20*iFp21*iFp00*iFp11+2.0*Dbar24*iFp20*iFp21*iFp01*iFp11+Dbar56*iFp20*iFp22*iFp01*iFp12+Dbar46*iFp21*iFp22*iFp10*iFp01+Dbar46*iFp21*iFp22*iFp11*iFp00+Dbar66*iFp21*iFp22*iFp12*iFp01+Dbar56*iFp21*iFp22*iFp12*iFp00+Dbar25*iFp21*iFp01*iFp20*iFp12+Dbar12*iFp20*iFp00*iFp21*iFp11+Dbar16*iFp20*iFp00*iFp21*iFp12+Dbar25*iFp20*iFp02*iFp21*iFp11+Dbar16*iFp22*iFp01*iFp20*iFp10+Dbar66*iFp22*iFp22*iFp01*iFp11+Dbar25*iFp21*iFp01*iFp22*iFp10+Dbar55*iFp20*iFp22*iFp00*iFp12+Dbar45*iFp20*iFp22*iFp00*iFp11+2.0*Dbar15*iFp20*iFp22*iFp00*iFp10;
  t0 = s2+Dbar46*iFp20*iFp21*iFp02*iFp11+Dbar45*iFp20*iFp21*iFp02*iFp10+Dbar46*iFp20*iFp21*iFp01*iFp12+Dbar16*iFp21*iFp02*iFp20*iFp10+Dbar23*iFp21*iFp01*iFp22*iFp12+Dbar16*iFp20*iFp00*iFp22*iFp11+Dbar25*iFp22*iFp00*iFp21*iFp11+Dbar34*iFp20*iFp01*iFp22*iFp12+s1+Dbar12*iFp21*iFp01*iFp20*iFp10+Dbar13*iFp22*iFp02*iFp20*iFp10+Dbar34*iFp22*iFp02*iFp20*iFp11+Dbar13*iFp20*iFp00*iFp22*iFp12+Dbar45*iFp20*iFp20*iFp02*iFp11+Dbar24*iFp21*iFp21*iFp11*iFp00+Dbar26*iFp21*iFp21*iFp12*iFp01+Dbar15*iFp20*iFp20*iFp02*iFp10+Dbar36*iFp22*iFp22*iFp11*iFp02+Dbar26*iFp21*iFp21*iFp11*iFp02;
  D[4][5] = D[5][4] = t0;
  s1 = Dbar66*iFp21*iFp21*iFp12*iFp12+Dbar44*iFp21*iFp21*iFp10*iFp10+Dbar66*iFp22*iFp22*iFp11*iFp11+Dbar55*iFp20*iFp20*iFp12*iFp12+2.0*Dbar14*iFp10*iFp11*iFp20*iFp20+2.0*Dbar24*iFp10*iFp11*iFp21*iFp21+2.0*Dbar15*iFp10*iFp12*iFp20*iFp20+2.0*Dbar15*iFp10*iFp10*iFp20*iFp22+2.0*Dbar14*iFp10*iFp10*iFp20*iFp21+2.0*Dbar36*iFp12*iFp12*iFp21*iFp22+2.0*Dbar24*iFp11*iFp11*iFp20*iFp21+2.0*Dbar36*iFp11*iFp12*iFp22*iFp22+2.0*Dbar26*iFp11*iFp12*iFp21*iFp21+2.0*Dbar35*iFp12*iFp12*iFp20*iFp22+2.0*Dbar56*iFp22*iFp22*iFp10*iFp11+2.0*Dbar46*iFp21*iFp21*iFp10*iFp12+2.0*Dbar45*iFp22*iFp10*iFp10*iFp21+2.0*Dbar56*iFp21*iFp12*iFp12*iFp20+Dbar55*iFp22*iFp22*iFp10*iFp10+Dbar44*iFp20*iFp20*iFp11*iFp11+2.0*Dbar16*iFp20*iFp10*iFp22*iFp11+2.0*Dbar16*iFp20*iFp10*iFp21*iFp12;
  s2 = s1+2.0*Dbar13*iFp20*iFp10*iFp22*iFp12+2.0*Dbar23*iFp21*iFp11*iFp22*iFp12+2.0*Dbar25*iFp21*iFp11*iFp22*iFp10+2.0*Dbar46*iFp22*iFp11*iFp11*iFp20+2.0*Dbar34*iFp20*iFp11*iFp22*iFp12+2.0*Dbar34*iFp21*iFp10*iFp22*iFp12+Dbar11*iFp10*iFp10*iFp20*iFp20+2.0*Dbar12*iFp20*iFp10*iFp21*iFp11+2.0*Dbar25*iFp20*iFp12*iFp21*iFp11+2.0*Dbar45*iFp20*iFp20*iFp11*iFp12+2.0*Dbar56*iFp10*iFp12*iFp21*iFp22;
  t0 = s2+2.0*Dbar44*iFp10*iFp11*iFp20*iFp21+2.0*Dbar45*iFp10*iFp11*iFp20*iFp22+2.0*Dbar45*iFp10*iFp12*iFp20*iFp21+2.0*Dbar46*iFp10*iFp11*iFp21*iFp22+2.0*Dbar56*iFp11*iFp12*iFp20*iFp22+2.0*Dbar35*iFp10*iFp12*iFp22*iFp22+2.0*Dbar46*iFp11*iFp12*iFp20*iFp21+2.0*Dbar66*iFp11*iFp12*iFp21*iFp22+2.0*Dbar55*iFp10*iFp12*iFp20*iFp22+2.0*Dbar26*iFp11*iFp11*iFp21*iFp22+Dbar33*iFp12*iFp12*iFp22*iFp22+Dbar22*iFp11*iFp11*iFp21*iFp21;
  D[5][5] = t0;

  /* I'm handling request for Lagrangian stiffness only;
     if Eulerian, should transform first.
  */
  if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) return TRUE;
  else                                       return FALSE;
}

static W_bool_t
get_mat_stiff_elastic_9x9 (W_mat_3d_t *m /* mat3d */,
                       W_mat_3d_state_p ms,/* material state */
                       W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                                    used to generate kinematic
                                                    quantities which the material state
                                                    might need in order to compute
                                                    the requested stress */
                       W_mat_3d_stiff_enum stiff_type, /* stiffness type */
                       double D[9][9] /* material stiffness tensor in matrix form */
                       )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  double C[6][6];
  int i, j; 
/* get C[6][6] first, then expand to D[9][9] */
  get_mat_stiff_elastic(m, ms, ip_ctx, stiff_type, C);
  FOR (i, 9) {
      FOR(j,9) {
	if (i <=5  && j <= 5) { 
		D[i][j] = C[i][j];
	} else {
		if (i >5 && j<=5 ) D[i][j] = C[i-3][j];
		else {
		if (i <=5 && j>5 ) D[i][j] = C[i][j-3];
		else	       D[i][j] = C[i-3][j-3];
		     }	
	       }
      }
  }
  /* I'm handling request for Lagrangian stiffness only;
     if Eulerian, should transform first.
  */
  if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) return TRUE;
  else                                       return FALSE;
}

typedef struct {
  double F[3][3];
} custom_data_t;

static W_bool_t 
custom_get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
               W_mat_3d_kinem_var_type_enum  type,
               double kinvar[3][3])
{
  custom_data_t *custom_data = (custom_data_t *) ctx->ctx_data;
  switch (type) {
  case MAT_3D_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, custom_data->F);
    break;
  default:
    REPORT_FATAL_ERROR(NULL, "Unknown kinematic tensor type");
    break;
  }
  return TRUE;
}

static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* stiffness type */
               double C[6][6] /* material stiffness tensor in matrix form */
               )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  int i, j, k, l;
  double e = u->e_incr;
  double U[3][3], R[3][3], dU[3][3], U1[3][3];
  custom_data_t custom_data;
  double Sp[6], Sn[6];
  W_mat_3d_stress_type_enum   out_stress_type;
  int result = TRUE;
  int v_to_t[6][2] = {{0,0},{1,1},{2,2},{0,1},{0,2},{1,2}};
  int mult[6] = {1,1,1,2,2,2};
  double equiv_pl_def;
  W_mat_3d_ip_ctx_t custom_ip_ctx;
  GEP_MATRIX (um);
  int umpvt[6];
  GEP_STACKINIT(um, 6);
  

  if (u->e_incr == 0) /* numerical differentiation not allowed */ {
    W_bool_t result =  get_mat_stiff_elastic (m, ms, ip_ctx, stiff_type, C);
    return result;
  }
  
  /* zero out */
  FOR(i, 6) FOR (j, 6) C[i][j] = 0;

  /* here we access the current kinematic variables through the element defined
     kinematics callback
   */
  if (!ip_ctx->get_kinem_var) return FALSE;  
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, custom_data.F)) {
    result = FALSE; goto exit;
  }

  /* now replace the kinematic variable callback by our custom callback */
  custom_ip_ctx               = *ip_ctx;
  custom_ip_ctx.get_kinem_var = custom_get_kinem_var;
  custom_ip_ctx.ctx_data      = &custom_data;

  /* compute the polar decomposition of the elastic deformation gradient */
  if (!W_polar_decomposition (custom_data.F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
#define UM(i,j,v) {GEP_ELEM(um, i, j) = v;}
  UM(0,0,U[0][0]); UM(0,1,0);       UM(0,2,0);       UM(0,3,U[1][0]);             UM(0,4,U[2][0]);             UM(0,5,0);
  UM(1,0,0);       UM(1,1,U[1][1]); UM(1,2,0);       UM(1,3,U[0][1]);             UM(1,4,0);                   UM(1,5,U[2][1]);
  UM(2,0,0);       UM(2,1,0);       UM(2,2,U[2][2]); UM(2,3,0);                   UM(2,4,U[0][2]);             UM(2,5,U[1][2]);
  UM(3,0,U[0][1]); UM(3,1,U[1][0]); UM(3,2,0);       UM(3,3,(U[0][0]+U[1][1])/2); UM(3,4,U[2][1]/2);           UM(3,5,U[2][0]/2);
  UM(4,0,U[0][2]); UM(4,1,0);       UM(4,2,U[2][0]); UM(4,3,U[1][2]/2);           UM(4,4,(U[0][0]+U[2][2])/2); UM(4,5,U[1][0]/2);
  UM(5,0,0);       UM(5,1,U[1][2]); UM(5,2,U[2][1]); UM(5,3,U[0][2]/2);           UM(5,4,U[0][1]/2);           UM(5,5,(U[1][1]+U[2][2])/2);
  gep_sgefa (&um, umpvt);
  
#define GET_2ND_PK(S)                                                                   \
  {                                                                                     \
    double St[3][3];                                                                    \
    if (!WM3D_update_stress (m, ms, W_MAT_STATE_UPDATE_TEMP,                            \
                             &custom_ip_ctx, MAT_3D_2ND_PK, St, &out_stress_type)) {    \
      result = FALSE; goto exit;                                                        \
    }                                                                                   \
    if (out_stress_type != MAT_3D_2ND_PK) {result = FALSE; goto exit;}                  \
    S[0] = St[0][0];                                                                    \
    S[1] = St[1][1];                                                                    \
    S[2] = St[2][2];                                                                    \
    S[3] = St[0][1];                                                                    \
    S[4] = St[0][2];                                                                    \
    S[5] = St[1][2];                                                                    \
  }
#define COMP_DU(K,delta_E)                      \
  {                                             \
    double dUv[6]={0,0,0,0,0,0};                \
    dUv[K] = delta_E;                           \
    gep_sgesl (&um, umpvt, dUv, GEP_JOB_A_X_B); \
    MAT3DUTIL_V_TO_T (dUv, dU);                 \
  }

  FOR (k, 6) {
    COMP_DU(k, (+e));
    MAT3DUTIL_2_MAT (U1, =, U);
    MAT3DUTIL_2_MAT (U1, +=, dU);
    MAT3DUTIL_MULT_AB_3X3 (custom_data.F, R, U1);
    GET_2ND_PK (Sp);
    MAT3DUTIL_2_MAT (U1, =, U);
    MAT3DUTIL_2_MAT (U1, -=, dU);
    MAT3DUTIL_MULT_AB_3X3 (custom_data.F, R, U1);
    GET_2ND_PK (Sn);
    FOR (l, 6) {
      C[k][l] = (Sp[l] - Sn[l]) / (2 * e) / mult[l];
    }
  }

#if 1
  /* symmetrize */
  FOR (k, 6) {
    FOR (l, 6) {
      if (k != l) {
        double a = (C[k][l] + C[l][k]) / 2;
        C[k][l] = C[l][k] = a;
      }
    }
  }
#endif

 exit:

  if (result) {
    /* I'm handling request for Lagrangian stiffness only;
     if Eulerian, should transform first. */
    if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) return TRUE;
    else                                       return FALSE;
  } else
    return FALSE; /* have nothing to show for my effort :( */
}

 
static W_bool_t
get_mat_stiff_9x9 (W_mat_3d_t           *m/* mat3d */,
                   W_mat_3d_state_p     ms,/* material state */
                   W_mat_3d_ip_ctx_t    *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
                   W_mat_3d_stiff_enum  stiff_type, /* stiffness type */
                   /* double               Kmat[9][9] */
                   double               K[9][9]
                   )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
 
  double h_overhat1[N_SLIP_SYSTEMS][N_SLIP_SYSTEMS];

  double invA1[N_SLIP_SYSTEMS][N_SLIP_SYSTEMS];
  double rcond, z[N_SLIP_SYSTEMS];
  const double rcond_tol = 1e-20;
  GEP_MATRIX (A1m);
  int    A1pvt[N_SLIP_SYSTEMS];
  double R[N_SLIP_SYSTEMS];
  double unit[N_SLIP_SYSTEMS][N_SLIP_SYSTEMS];
 
  double Rbar1[3][3];
  double n1[3][3], r1[3][3];
  double n1v[9], r1v[9];
  double v1[9], v2[9];
  double s[N_SLIP_SYSTEMS][3], mm[N_SLIP_SYSTEMS][3];
  double K1[9][9], K2[9][9];
  double Kmat[9][9];
  int    i, j, k, alpha, beta;
 
  double h_0 = u->h_0;
  double accum_gamma = matst->accum_gamma;
  double tau_s = u->g_inf;
  double tau_0 = u->g_0;
  double q = u->q;
  double m_srs = u->m_srs;
  double gamma_0 = u->gamma_0;
  
  double C[9][9]; 
  double Dbar[6][6];
  double invFp1[3][3], invFp[3][3];
  double detFp, detFp1, detFe1;
  double Fp1[3][3];
  double Fp[3][3];
  double Fe1[3][3];
  double invFe1[3][3];
  double F1[3][3];
  double Cebar1[3][3];
 
  double tau1[N_SLIP_SYSTEMS];
  double Sbar1[3][3], S1[3][3];
  double ks[3][3];
  double temp[3][3], temp1[3][3];
 
  double g[N_SLIP_SYSTEMS];
  double g1[N_SLIP_SYSTEMS];
  double gamma[N_SLIP_SYSTEMS];
  double dgamma[N_SLIP_SYSTEMS];
  double s1[N_SLIP_SYSTEMS][3]; /* slip directions */
  double m1[N_SLIP_SYSTEMS][3]; /* normals to slip planes */
  double dt = ip_ctx->time_step;
  W_bool_t result = 0;

  GEP_STACKINIT(A1m, N_SLIP_SYSTEMS);
  /* zero out K[9][9] */
   FOR (i, 9) {
       FOR (j, 9) {
	  Kmat[i][j] = 0;
       }
   }
  /* kinematic quantities */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F1))
    return FALSE;
 
  /* Calculate slip systems */
  /* could be cached with the material */
  FOR (i, N_SLIP_SYSTEMS) { 
     s1[i][0] = matst->a.x * ref_s[i][0] + matst->b.x * ref_s[i][1] + matst->c.x * ref_s[i][2];
     s1[i][1] = matst->a.y * ref_s[i][0] + matst->b.y * ref_s[i][1] + matst->c.y * ref_s[i][2];
     s1[i][2] = matst->a.z * ref_s[i][0] + matst->b.z * ref_s[i][1] + matst->c.z * ref_s[i][2];
     m1[i][0] = matst->a.x * ref_m[i][0] + matst->b.x * ref_m[i][1] + matst->c.x * ref_m[i][2];
     m1[i][1] = matst->a.y * ref_m[i][0] + matst->b.y * ref_m[i][1] + matst->c.y * ref_m[i][2];
     m1[i][2] = matst->a.z * ref_m[i][0] + matst->b.z * ref_m[i][1] + matst->c.z * ref_m[i][2];
  }
  FOR (i, N_SLIP_SYSTEMS) {
      FOR (j, 3) {
      s[i][j] = s1[i][j];
      mm[i][j] = m1[i][j];
      }
  }
 
  FOR (i, N_SLIP_SYSTEMS) g[i] = matst->g[i];
  FOR (i, N_SLIP_SYSTEMS) gamma[i] = matst->gamma[i];
  FOR (i, N_SLIP_SYSTEMS) dgamma[i] = matst->dgamma[i];
  FOR (i, 3) {
      FOR(j, 3) {
          Fp[i][j] = matst->Fp[i][j];
      }
  }
 
/* Get invFp */
  detFp = MAT3DUTIL_DET_3X3 (matst->Fp);
  if (detFp == 0) {
    REPORT_FATAL_ERROR (NULL, "Zero det(Fp)");
  }
  MAT3DUTIL_INVERT_3X3 (detFp, matst->Fp, invFp);
 
/* Get Fp1 and invFp1 */
  calc_Fp1 (Fp, dgamma, s, mm, Fp1);
  detFp1 = MAT3DUTIL_DET_3X3 (Fp1);
  if (detFp1 == 0) {
    REPORT_FATAL_ERROR (NULL, "Zero det(Fp1)");
  }
  MAT3DUTIL_INVERT_3X3 (detFp1, Fp1, invFp1);
 
  calc_g1 (q, h_0, tau_0, tau_s, g, accum_gamma, dgamma, g1);
  calc_Cebar1(F1,Fp1, Cebar1);
  MAT3DUTIL_MULT_AB_3X3 (Fe1, F1, invFp1);
  detFe1 = MAT3DUTIL_DET_3X3 (Fe1);
  if (detFe1 == 0) {
    REPORT_FATAL_ERROR (NULL, "Zero det(Fe1)");
  }
  MAT3DUTIL_INVERT_3X3 (detFe1, Fe1, invFe1);
 
  calc_Dbar ( u->c11, u->c12, u->c44, &matst->a, &matst->b, &matst->c, Dbar);
  calc_Sbar (Dbar, Cebar1, Fp1, Sbar1);
  calc_S (Sbar1, Fp1, S1);
  calc_tau1 (Cebar1, Sbar1, s, mm, tau1);
  /* get dt */
 
  /* Get spatial elastic moduli matrix */
  result =  get_mat_stiff_elastic_9x9 (m, ms, ip_ctx, stiff_type, C);
 
  calc_h_overhat1(h_0, accum_gamma, tau_s, tau_0, q, dgamma, h_overhat1);
 
  calc_A1(Fp, Cebar1, Fe1, invFe1, invFp1, C, gamma_0, m_srs, tau1, g1, h_overhat1, s, mm, dt, &A1m);

  /* Need to know the inverse of A1 */
  FOR (i, N_SLIP_SYSTEMS) {
      FOR (j, N_SLIP_SYSTEMS) {
      if (i == j) unit[i][j] = 1;
      else unit[i][j] = 0;
      }
  }

  gep_sgeco(&A1m, A1pvt, &rcond, z);
  FOR (i, N_SLIP_SYSTEMS) {
     FOR (j, N_SLIP_SYSTEMS) {
         R[j] = unit[j][i];
     }
     gep_sgesl (&A1m, A1pvt, R, GEP_JOB_A_X_B);
     FOR (k, N_SLIP_SYSTEMS) {
         invA1[k][i] = R[k];
     }
  }

  FOR (alpha, N_SLIP_SYSTEMS) {
      FOR(beta, N_SLIP_SYSTEMS) {
         calc_Rbar1(alpha, Cebar1, Fp, invFp1, s, mm, Rbar1);
         calc_r1(Rbar1, Fe1, r1);
         calc_n1(beta, gamma_0, m_srs, tau1, g1, invFe1, s, mm, Cebar1, n1);
         MAT3DUTIL_T_TO_V9(n1, n1v);
         MAT3DUTIL_T_TO_V9(r1, r1v);
         MAT3DUTIL_MULT_A_9x9_V_9(v1, C, r1v);
         MAT3DUTIL_MULT_V_9_A_9x9(v2, n1v, C);
         MAT3DUTIL_TENS_PROD_V9(K1, v1, v2);
 
         /* get kirchhoff stress in the spatial coordinates ks = F1 * S1 * F1T */
         MAT3DUTIL_MULT_ABAT_3X3(ks, F1, S1);
         MAT3DUTIL_MULT_ABT_3X3(temp, ks, r1);
         MAT3DUTIL_ADD_ABT_3X3(temp1, temp, temp);
         MAT3DUTIL_T_TO_V9(temp1, v1);
         MAT3DUTIL_TENS_PROD_V9(K2, v1, v2);
 
          FOR (i, 9) {
              FOR (j, 9) {
                Kmat[i][j] -= invA1[alpha][beta] * K1[i][j];
                Kmat[i][j] -= invA1[alpha][beta] * K2[i][j];
              }
          }
    }
  }
 
   FOR (i, 9) {
      FOR (j, 9) {
          K[i][j] = C[i][j] + Kmat[i][j];
      }
  } 
  return TRUE;
}

static W_mat_3d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const,
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  get_mat_state_named_var,
  update_history,
  get_mat_stiff_9x9,
};
    
W_mat_3d_t *
read_mat_3d_ckinks (W_domain_t *d)
{
  const char *input_line;
  double c11=0, c12=0, c44=0, rho = 1;
  double q, h_0, gamma_0, m_srs, g_0, g_inf;
  double J, K, G, mb, const_a, const_k, D;/* For solving the dislocation density */
  char *a[3], ax[512], ay[512], az[512];
  char *b[3], bx[512], by[512], bz[512];
  int debug = 0;
  int max_inner_iter = 20;
  double e_incr = 0;
  double activity_tol = 0.01;
  W_bool_t elastic = 0;
  W_mat_3d_t *mat = 0;
    
  a[0] = ax; a[1] = ay; a[2] = az;
  b[0] = bx; b[1] = by; b[2] = bz;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("C11", TRUE,
    			    { c11 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c11 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("C12", TRUE,
    			    { c12 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c12 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("C44", TRUE,
    			    { c44 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c44 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("A", TRUE,
                            { DUPLICATE_STRING(a[0], TOKEN_STR_NEXT(1)); DUPLICATE_STRING(a[1], TOKEN_STR_NEXT(2)); DUPLICATE_STRING(a[2], TOKEN_STR_NEXT(3)); },
                            { a[0] = "1"; a[1] = "0"; a[2] = "0"; }
    			    );
      TOKENS_SET_BY_KEYWORD("B", TRUE,
                            { DUPLICATE_STRING(b[0], TOKEN_STR_NEXT(1)); DUPLICATE_STRING(b[1], TOKEN_STR_NEXT(2)); DUPLICATE_STRING(b[2], TOKEN_STR_NEXT(3)); },
                            { b[0] = "0"; b[1] = "1"; b[2] = "0"; }
    			    );
/*      
      TOKENS_SET_BY_KEYWORD("C", FALSE,
                            { c.x = TOKEN_DBL_NEXT(1); c.y = TOKEN_DBL_NEXT(2); c.z = TOKEN_DBL_NEXT(3);
                              W_normalize_vect_3d (&c); },
                            { W_cross_prod_3d (&c, &a, &b); W_normalize_vect_3d (&c); }
    			    );
*/
      TOKENS_SET_BY_KEYWORD("Q", FALSE,
    			    { q = TOKEN_DBL_NEXT(1); },
    			    { q = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("H_0", TRUE,
    			    { h_0 = TOKEN_DBL_NEXT(1); },
    			    { h_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("G_0", TRUE,
    			    { g_0 = TOKEN_DBL_NEXT(1); },
    			    { g_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("G_INF", TRUE,
    			    { g_inf = TOKEN_DBL_NEXT(1); },
    			    { g_inf = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("GAMMA_0", TRUE,
    			    { gamma_0 = TOKEN_DBL_NEXT(1); },
    			    { gamma_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("M_SRS", TRUE,
    			    { m_srs = TOKEN_DBL_NEXT(1); },
    			    { m_srs = 0.02; }
    			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
    			    { rho = TOKEN_DBL_NEXT(1); },
    			    { rho = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("DEBUG", FALSE,
    			    { debug = TOKEN_INT_NEXT(1); },
                            { debug = 0; /* no debugging */ }
    			    );
      TOKENS_SET_BY_KEYWORD("MAX_INNER_ITER", FALSE,
    			    { max_inner_iter = TOKEN_INT_NEXT(1); },
                            { max_inner_iter = 20;  }
    			    );
      TOKENS_SET_BY_KEYWORD("E_INCR", FALSE,
    			    { e_incr = TOKEN_DBL_NEXT(1); },
    			    { e_incr = 0; }/* numerical approx of tangent */
    			    );
      TOKENS_SET_BY_KEYWORD("ACTIVITY_TOL", FALSE,
    			    { activity_tol = TOKEN_DBL_NEXT(1); },
    			    { activity_tol = 0.01; }
    			    );
      TOKENS_SET_BY_KEYWORD("ELASTIC", FALSE,
                            { elastic = TOKEN_INT_NEXT(1); },
                            { elastic = 0; }
                            );

      {
        mat_t *u;
        int i;

        u = w_new(mat_t);
        CHECK_MALLOC(u);
        u->c11 = c11; 
        u->c12 = c12; 
        u->c44 = c44; 
        if (u->c11 < 0) REPORT_FATAL_ERROR(NULL, "Bad c11 (should be > 0)");
        if (u->c12 < 0) REPORT_FATAL_ERROR(NULL, "Bad c12 (should be > 0)");
        if (u->c44 < 0) REPORT_FATAL_ERROR(NULL, "Bad c44 (should be > 0)");
        u->a[0] = a[0]; u->a[1] = a[1]; u->a[2] = a[2];
        u->b[0] = b[0]; u->b[1] = b[1]; u->b[2] = b[2];

        u->gamma_0 = gamma_0;
        u->q = q;
        u->h_0 = h_0;
        u->m_srs = m_srs;
        u->g_0 = g_0;
        u->g_inf = g_inf;
        if (u->g_inf < u->g_0) {
          REPORT_FATAL_ERROR (NULL, "g_inf should be greater than g_0");
        }
        u->debug = debug;
        u->max_inner_iter = max_inner_iter;
        u->activity_tol = activity_tol;
        u->e_incr = e_incr;
        u->elastic = elastic;

        u->J = J;
        u->K = K;
        u->mb = mb;
        u->G = G;
        u->const_a = const_a;
        u->const_k = const_k;
        
    	u->d = d;
        mat = WM3D_new_material(&methods, (void *)u, rho); 
    
    	u->hsi = 0;

    	return mat;
    
      }
    }
  }
  return NULL;
}

