#include "Wmat3dcrysplbcc.h"

#define N_SLIP_SYSTEMS 24

#define ISQRT2 0.70710678118654752440
#define ISQRT3 0.57735026918962576451
#define S(x,y,z) x*ISQRT2,y*ISQRT2,z*ISQRT2
#define M(x,y,z) x*ISQRT3,y*ISQRT3,z*ISQRT3

/* slip directions expressed in the a,b,c vectors */
static double ref_s[N_SLIP_SYSTEMS][3] = {
  {S(0,-1,1)},   /* B2 */
  {S(1,0,-1)},   /* B4 */
  {S(-1,1,0)},   /* B5 */
  {S(1,0,1)},    /* A3 */
  {S(0,1,-1)},   /* A2 */
  {S(-1,-1,0)},  /* A6 */
  {S(0,1,1)},    /* C1 */
  {S(-1,0,-1)},  /* C3 */
  {S(1,-1,0)},   /* C5 */
  {S(-1,0,1)},   /* D4 */
  {S(0,-1,-1)},  /* D1 */
  {S(1,1,0)}     /* D6 */
};
/* slip plane normals expressed in the a,b,c vectors */
static double ref_m[N_SLIP_SYSTEMS][3] = {
  {M(1,1,1)},   /* B2 */
  {M(1,1,1)},   /* B4 */
  {M(1,1,1)},   /* B5 */
  {M(-1,1,1)},  /* A3 */
  {M(-1,1,1)},  /* A2 */
  {M(-1,1,1)},  /* A6 */
  {M(-1,-1,1)}, /* C1 */
  {M(-1,-1,1)}, /* C3 */
  {M(-1,-1,1)}, /* C5 */
  {M(1,-1,1)},  /* D4 */
  {M(1,-1,1)},  /* D1 */
  {M(1,-1,1)}   /* D6 */
};

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_CRYSPL_BCC;
}


#include "mat3dcrysplc.c"

W_mat_3d_t *
read_mat_3d_cryspl_bcc (W_domain_t *d)
{
  const char *input_line;
  double c11=0, c12=0, c44=0, rho = 1;
  double q, h_0, gamma_0, m_srs, g_0, g_inf;
  W_vect_t a={1,0,0};
  W_vect_t b={0,1,0};
  W_vect_t c={0,0,1};
  int debug = 0;
  int max_inner_iter = 20;
  double e_incr = 0;
  double activity_tol = 0.01;
  W_bool_t elastic = 0;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("C11", TRUE,
			    { c11 = TOKEN_DBL_NEXT(1); },
			    { REPORT_FATAL_ERROR (NULL, "c11 must be given"); }
			    );
      TOKENS_SET_BY_KEYWORD("C12", TRUE,
			    { c12 = TOKEN_DBL_NEXT(1); },
			    { REPORT_FATAL_ERROR (NULL, "c12 must be given"); }
			    );
      TOKENS_SET_BY_KEYWORD("C44", TRUE,
			    { c44 = TOKEN_DBL_NEXT(1); },
			    { REPORT_FATAL_ERROR (NULL, "c44 must be given"); }
			    );
      TOKENS_SET_BY_KEYWORD("A", TRUE,
                            { a.x = TOKEN_DBL_NEXT(1); a.y = TOKEN_DBL_NEXT(2); a.z = TOKEN_DBL_NEXT(3);
                              W_normalize_vect_3d (&a); },
                            { a.x = 1; a.y = 0; a.z = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("B", TRUE,
                            { b.x = TOKEN_DBL_NEXT(1); b.y = TOKEN_DBL_NEXT(2); b.z = TOKEN_DBL_NEXT(3);
                              W_normalize_vect_3d (&b); },
                            { b.x = 0; b.y = 1; b.z = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("C", FALSE,
                            { c.x = TOKEN_DBL_NEXT(1); c.y = TOKEN_DBL_NEXT(2); c.z = TOKEN_DBL_NEXT(3);
                              W_normalize_vect_3d (&c); },
                            { W_cross_prod_3d (&c, &a, &b); W_normalize_vect_3d (&c); }
			    );
      TOKENS_SET_BY_KEYWORD("Q", FALSE,
			    { q = TOKEN_DBL_NEXT(1); },
			    { q = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("H_0", TRUE,
			    { h_0 = TOKEN_DBL_NEXT(1); },
			    { h_0 = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("G_0", TRUE,
			    { g_0 = TOKEN_DBL_NEXT(1); },
			    { g_0 = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("G_INF", TRUE,
			    { g_inf = TOKEN_DBL_NEXT(1); },
			    { g_inf = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("GAMMA_0", TRUE,
			    { gamma_0 = TOKEN_DBL_NEXT(1); },
			    { gamma_0 = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("M_SRS", TRUE,
			    { m_srs = TOKEN_DBL_NEXT(1); },
			    { m_srs = 0.02; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("DEBUG", FALSE,
			    { debug = TOKEN_INT_NEXT(1); },
                            { debug = 0; /* no debugging */ }
			    );
      TOKENS_SET_BY_KEYWORD("MAX_INNER_ITER", FALSE,
			    { max_inner_iter = TOKEN_INT_NEXT(1); },
                            { max_inner_iter = 20;  }
			    );
      TOKENS_SET_BY_KEYWORD("E_INCR", FALSE,
			    { e_incr = TOKEN_DBL_NEXT(1); },
			    { e_incr = 0; }/* numerical approx of tangent */
			    );
      TOKENS_SET_BY_KEYWORD("ACTIVITY_TOL", FALSE,
			    { activity_tol = TOKEN_DBL_NEXT(1); },
			    { activity_tol = 0.01; }
			    );
      TOKENS_SET_BY_KEYWORD("ELASTIC", FALSE,
			    { elastic = TRUE; },
			    { elastic = FALSE; }
			    );
      return new_mat_3d_cryspl (c11, c12, c44, &a, &b, &c,
                                gamma_0,  q,  h_0, m_srs, g_0, g_inf, debug,
                                max_inner_iter, activity_tol,
                                e_incr, elastic,
                                rho);
    }
  }
  return NULL;
}
