/* Face centered cryspls under plane strain deformation */
#include "Wdomain.h"
#include "Wmat3dcrysplfccps.h"
    
#define N_SLIP_SYSTEMS 4
    
#define ISQRT2 0.70710678118654752440
#define ISQRT3 0.57735026918962576451
#define S(x,y,z) x*ISQRT2,y*ISQRT2,z*ISQRT2
#define M(x,y,z) x*ISQRT3,y*ISQRT3,z*ISQRT3
    
/* slip directions expressed in the a,b,c vectors */
static double ref_s[N_SLIP_SYSTEMS][3] = {
  {S(0,-1,1)},   /* B2 */
  {S(1,0,-1)},   /* B4 */
  {S(0,1,1)},    /* C1 */
  {S(-1,0,-1)}  /* C3 */
};
/* slip plane normals expressed in the a,b,c vectors */
static double ref_m[N_SLIP_SYSTEMS][3] = {
  {M(1,1,1)},   /* B2 */
  {M(1,1,1)},   /* B4 */
  {M(-1,-1,1)}, /* C1 */
  {M(-1,-1,1)} /* C3 */
};
    
static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_CRYSPL_FCC_PS;
}
    

#include "mat3dcrysplc.c"
    
    
W_mat_3d_t *
read_mat_3d_cryspl_fcc_ps (W_domain_t *d)
{
  const char *input_line;
  double c11=0, c12=0, c44=0, rho = 1;
  double q, h_0, gamma_0, m_srs, g_0, g_inf;
  double J, K, G, mb, const_a, const_k, D;/* For solving the dislocation density */
  W_vect_t a={1,0,0};
  W_vect_t b={0,1,0};
  W_vect_t c={0,0,1};
  int debug = 0;
  int max_inner_iter = 20;
  double e_incr = 0;
  double activity_tol = 0.01;
  W_bool_t elastic = 0;
  W_mat_3d_t *mat = 0;
    
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("C11", TRUE,
    			    { c11 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c11 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("C12", TRUE,
    			    { c12 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c12 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("C44", TRUE,
    			    { c44 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c44 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("A", TRUE,
                            { a.x = TOKEN_DBL_NEXT(1); a.y = TOKEN_DBL_NEXT(2); a.z = TOKEN_DBL_NEXT(3);
                              W_normalize_vect_3d (&a); },
                            { a.x = 1; a.y = 0; a.z = 0; }
    			    );
      TOKENS_SET_BY_KEYWORD("B", TRUE,
                            { b.x = TOKEN_DBL_NEXT(1); b.y = TOKEN_DBL_NEXT(2); b.z = TOKEN_DBL_NEXT(3);
                              W_normalize_vect_3d (&b); },
                            { b.x = 0; b.y = 1; b.z = 0; }
    			    );
      TOKENS_SET_BY_KEYWORD("C", FALSE,
                            { c.x = TOKEN_DBL_NEXT(1); c.y = TOKEN_DBL_NEXT(2); c.z = TOKEN_DBL_NEXT(3);
                              W_normalize_vect_3d (&c); },
                            { W_cross_prod_3d (&c, &a, &b); W_normalize_vect_3d (&c); }
    			    );
      TOKENS_SET_BY_KEYWORD("Q", FALSE,
    			    { q = TOKEN_DBL_NEXT(1); },
    			    { q = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("H_0", TRUE,
    			    { h_0 = TOKEN_DBL_NEXT(1); },
    			    { h_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("G_0", TRUE,
    			    { g_0 = TOKEN_DBL_NEXT(1); },
    			    { g_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("G_INF", TRUE,
    			    { g_inf = TOKEN_DBL_NEXT(1); },
    			    { g_inf = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("GAMMA_0", TRUE,
    			    { gamma_0 = TOKEN_DBL_NEXT(1); },
    			    { gamma_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("M_SRS", TRUE,
    			    { m_srs = TOKEN_DBL_NEXT(1); },
    			    { m_srs = 0.02; }
    			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
    			    { rho = TOKEN_DBL_NEXT(1); },
    			    { rho = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("DEBUG", FALSE,
    			    { debug = TOKEN_INT_NEXT(1); },
                            { debug = 0; /* no debugging */ }
    			    );
      TOKENS_SET_BY_KEYWORD("MAX_INNER_ITER", FALSE,
    			    { max_inner_iter = TOKEN_INT_NEXT(1); },
                            { max_inner_iter = 20;  }
    			    );
      TOKENS_SET_BY_KEYWORD("E_INCR", FALSE,
    			    { e_incr = TOKEN_DBL_NEXT(1); },
    			    { e_incr = 0; }/* numerical approx of tangent */
    			    );
      TOKENS_SET_BY_KEYWORD("ACTIVITY_TOL", FALSE,
    			    { activity_tol = TOKEN_DBL_NEXT(1); },
    			    { activity_tol = 0.01; }
    			    );

      /** 
	 J: athermal dislocation storage term -- constant 
	 K: thermally activated recovery term -- temperature and strain rate dependent 
	 G: shear modulus
	 const_a: numerical constant
	 const_k: constant 
	 mb: magnitude of Burger's vector
       **/

      TOKENS_SET_BY_KEYWORD("J", FALSE, 
    			    { J = TOKEN_DBL_NEXT(1); },
    			    { J = 4e+6; }
    			    );
      TOKENS_SET_BY_KEYWORD("K", FALSE,
    			    { K = TOKEN_DBL_NEXT(1); },
    			    { K = 10; }
    			    );
      TOKENS_SET_BY_KEYWORD("SHEAR_MODULUS", FALSE,
    			    { G = TOKEN_DBL_NEXT(1); },
    			    { G = 75000; }
    			    );
      TOKENS_SET_BY_KEYWORD("CONST_A", FALSE,
    			    { const_a = TOKEN_DBL_NEXT(1);},
    			    { const_a = 0.3; }
    			    );
      TOKENS_SET_BY_KEYWORD("CONST_K", FALSE,
    			    { const_k = TOKEN_DBL_NEXT(1);},
    			    { const_k = 1e-8; }
    			    );
      TOKENS_SET_BY_KEYWORD("MAGN_BURGER", FALSE,
    			    { mb = TOKEN_DBL_NEXT(1);},
    			    { mb = 2.5e-7; }
    			    );
      {
        mat_t *u;
        const double eps = 1e-9;
        int i;
        
        u = w_new(mat_t);
        CHECK_MALLOC(u);
        u->c11 = c11; 
        u->c12 = c12; 
        u->c44 = c44; 
        if (u->c11 < 0) REPORT_FATAL_ERROR(NULL, "Bad c11 (should be > 0)");
        if (u->c12 < 0) REPORT_FATAL_ERROR(NULL, "Bad c12 (should be > 0)");
        if (u->c44 < 0) REPORT_FATAL_ERROR(NULL, "Bad c44 (should be > 0)");
        u->a = a;
        u->b = b;
        u->c = c;
        if (W_dot_prod_3d (&u->a, &u->b) > eps) {
          REPORT_FATAL_ERROR (NULL, "Non-orthogonal orientation vectors");
        }
        if (W_dot_prod_3d (&u->c, &u->b) > eps) {
          REPORT_FATAL_ERROR (NULL, "Non-orthogonal orientation vectors");
        }
        if (W_dot_prod_3d (&u->c, &u->a) > eps) {
          REPORT_FATAL_ERROR (NULL, "Non-orthogonal orientation vectors");
        }
        if (fabs (W_vector_3d_length (&u->a) - 1) > eps) {
          REPORT_FATAL_ERROR (NULL, "Non-unit orientation vector");
        }
        if (fabs (W_vector_3d_length (&u->b) - 1) > eps) {
          REPORT_FATAL_ERROR (NULL, "Non-unit orientation vector");
        }
        if (fabs (W_vector_3d_length (&u->c) - 1) > eps) {
          REPORT_FATAL_ERROR (NULL, "Non-unit orientation vector");
        }
        u->gamma_0 = gamma_0;
        u->q = q;
        u->h_0 = h_0;
        u->m_srs = m_srs;
        u->g_0 = g_0;
        u->g_inf = g_inf;
        if (u->g_inf < u->g_0) {
          REPORT_FATAL_ERROR (NULL, "g_inf should be greater than g_0");
        }
        u->debug = debug;
        u->max_inner_iter = max_inner_iter;
        u->activity_tol = activity_tol;
        u->e_incr = e_incr;
        u->elastic = elastic;
        u->J = J;
        u->K = K;
        u->mb = mb;
        u->G = G;
        u->const_a = const_a;
        u->const_k = const_k;
        
        /* Calculate slip systems */
        /* could be cached with the material */
        FOR (i, N_SLIP_SYSTEMS) { 
          u->s[i][0] = u->a.x * ref_s[i][0] + u->b.x * ref_s[i][1] + u->c.x * ref_s[i][2];
          u->s[i][1] = u->a.y * ref_s[i][0] + u->b.y * ref_s[i][1] + u->c.y * ref_s[i][2];
          u->s[i][2] = u->a.z * ref_s[i][0] + u->b.z * ref_s[i][1] + u->c.z * ref_s[i][2];
          u->m[i][0] = u->a.x * ref_m[i][0] + u->b.x * ref_m[i][1] + u->c.x * ref_m[i][2];
          u->m[i][1] = u->a.y * ref_m[i][0] + u->b.y * ref_m[i][1] + u->c.y * ref_m[i][2];
          u->m[i][2] = u->a.z * ref_m[i][0] + u->b.z * ref_m[i][1] + u->c.z * ref_m[i][2];
        }
    
    	/* also should set up u->d, i.e. the domain */
    
    	u->d = d;
    
        mat = WM3D_new_material(&methods, (void *)u, rho); 
    
    	/* initialize the hardening solver */
    
    	u->hsi = make_hardening_solver(mat);
    
    	W_domain_add_finalize_callback(d, u->hsi, (W_finalize_callback_func)solve_for_hardening);
    	return mat;
    
      }/* new_mat_3d_cryspl() */
    }
  }
  return NULL;
}

