
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dengj2.h"



static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_ENG_J2;
}

typedef struct mat_t {
  double E, nu;
  double lambda, bulk_mod, G;
  double R_y, k;
  double kinematicModuli, isotropicModuli;
  W_bool_t is_kinem_hard, is_iso_hard;
}              mat_t;


static W_bool_t
ComputeStressSpaceHardeningVars (GaussPoint *gp, FloatArray* strainSpaceHardeningVariables)
{
  // in full stress strain space
  
  int i;
  int count=0, size = this->giveSizeOfFullHardeningVarsVector(), isize, rSize;
  IntArray mask;
  
  if (!(is_iso_hard || is_kinem_hard)) return FALSE;

  FloatArray* answer = new FloatArray (size);
  this->giveStressStrainMask (mask,ReducedForm ,gp);
  isize = mask.giveSize();
  rSize = this->giveSizeOfReducedHardeningVarsVector (gp);
  
  /* kinematic hardening variables are first */
  if (this->is_kinem_hard) {
    for (i=1; i<= isize; i++)
      // to be consistent with equivalent plastic strain formulation
      // we multiply by (sqrt(2.)*2./3.)
      answer->at(mask.at(i)) = (sqrt(2.)*2./3.)*this->kinematicModuli * strainSpaceHardeningVariables->at(i);
    count =6;
  }
  
  if (this->is_iso_hard) {
    answer->at(count+1) = this->isotropicModuli *
      strainSpaceHardeningVariables->at(rSize) ;
  }
  
  //delete mask;
  return answer->negated();  
}

double 
J2plasticMaterial::computeYieldValueAt (GaussPoint *gp, FloatArray* stressVector, 
                                        FloatArray* stressSpaceHardeningVars) 
{
  double f;
  FloatArray helpVector, backStress;
  
  if (this->is_kinem_hard) {
    if (stressVector != NULL) {
      this->giveStressBackVector (backStress, *stressSpaceHardeningVars);
      helpVector = *stressVector;
      helpVector.add (backStress);
    } else return -k;
  } else helpVector = *stressVector;
  
  f = sqrt (this -> computeJ2InvariantAt (&helpVector));
  
  //if (this->is_kinem_hard) delete helpVector;
  
  return f + sqrt(1./3.)*this->giveIsotropicHardeningVar(stressSpaceHardeningVars) - this->k;
}


void
J2plasticMaterial::computeHardeningReducedModuli (FloatMatrix& answer,
                                                  GaussPoint *gp, 
                                                  FloatArray* strainSpaceHardeningVariables, 
                                                  TimeStep* atTime)
{
  /* computes hardening moduli in reduced stress strain space (for kinematic back-stress)*/
  
  int i;
  int ksize, size = this->giveSizeOfReducedHardeningVarsVector(gp);
  
  if (!hasHardening()) {answer.resize (0,0); return ;}
  //FloatMatrix* answer = new FloatMatrix (size, size);
  answer.resize (size, size);
  answer.zero();
  
  /* kinematic hardening variables are first */
  if (this->is_kinem_hard) {
    int ksize = this->giveSizeOfReducedStressStrainVector(gp->giveMaterialMode());
    for (i=1; i<= ksize; i++)
      answer.at(i,i) = this->kinematicModuli;
  }
  
  if (this->is_iso_hard) {
    answer.at(size,size) = this->isotropicModuli;
  }
  
  return ;
}



FloatArray* 
J2plasticMaterial::ComputeStressGradient (GaussPoint *gp, FloatArray* stressVector,
                                          FloatArray* stressSpaceHardeningVars) 
{
  /* stress gradient of yield function in full stress - strain space */
  
  double f,ax,ay,az,sx,sy,sz;
  FloatArray *answer = new FloatArray (6);
  FloatArray helpVector, backStress;
  
  if (this->is_kinem_hard) {
    if (stressVector != NULL) {
      this->giveStressBackVector (backStress, *stressSpaceHardeningVars);
      helpVector = *stressVector;
      helpVector.add (backStress);
    } else return answer;
  } else helpVector = *stressVector;
  
  f = sqrt (this -> computeJ2InvariantAt (&helpVector));
  // check for yield value zero value
  if (fabs (f) < 1.e-6) return answer;
  
  ax = helpVector[0];
  ay = helpVector[1];
  az = helpVector[2];
  
  sx = (2./3.)*ax - (1./3.)*ay - (1./3.)*az;
  sy = (2./3.)*ay - (1./3.)*ax - (1./3.)*az;
  sz = (2./3.)*az - (1./3.)*ay - (1./3.)*ax;
  
  answer[0] = 0.5 * sx / f;
  answer[1] = 0.5 * sy / f;
  answer[2] = 0.5 * sz / f;
  answer[3] = helpVector[3] / f;
  answer[4] = helpVector[4] / f;
  answer[5] = helpVector[5] / f;
  
  //if (this->is_kinem_hard) delete helpVector;
  
  return answer;
}

FloatArray*
J2plasticMaterial:: ComputeStressSpaceHardeningVarsReducedGradient (GaussPoint *gp, FloatArray* stressVector,
                                                                    FloatArray* stressSpaceHardeningVars) 
{
  /* computes stress space hardening gradient in reduced stress-strain space */
  
  int i, kcount=0, size = this->giveSizeOfReducedHardeningVarsVector(gp);
  double f,ax,ay,az,sx,sy,sz;
  FloatArray *answer;
  FloatArray *fullKinematicGradient, reducedKinematicGrad;
  StructuralCrossSection *crossSection = (StructuralCrossSection*)
    (gp -> giveElement()->giveCrossSection());
  
  if (!hasHardening()) return NULL;
  answer = new FloatArray (size);
  
  /* kinematic hardening variables first */
  if (this->is_kinem_hard) {
    fullKinematicGradient = this->ComputeStressGradient (gp, stressVector, stressSpaceHardeningVars);
    crossSection->giveReducedCharacteristicVector(reducedKinematicGrad, gp, *fullKinematicGradient);  
    delete fullKinematicGradient;
    
    kcount = reducedKinematicGrad.giveSize();
  }
  
  if (this->is_kinem_hard) {
    for (i=1; i<=kcount; i++) answer->at(i) = reducedKinematicGrad.at(i);
    //delete reducedKinematicGrad;
  }
  
  if (this->is_iso_hard) {
    answer->at(size) = sqrt(1./3.);
  }
  
  return answer;
}




int
J2plasticMaterial:: hasHardening ()
{
  return (this->is_kinem_hard || this->is_iso_hard);
}


void
J2plasticMaterial:: computeReducedGradientMatrix (FloatMatrix& answer, 
                                                  GaussPoint *gp,
                                                  const FloatArray& stressVector, 
                                                  const FloatArray& stressSpaceHardeningVars) 
{
  int i,j,size;
  int imask, jmask;
  FloatArray helpVector, backStress, df(6) ;
  IntArray mask;
  double f, f32, f12, ax, ay, az ;
  
  this-> giveStressStrainMask(mask, FullForm, gp);
  size = giveSizeOfReducedStressStrainVector(gp->giveMaterialMode()) + 
    this->giveSizeOfReducedHardeningVarsVector(gp);
  
  //FloatMatrix* answer = new FloatMatrix(size,size);
  answer.resize (size,size);
  answer.zero();
  
  
  if (stressVector.giveSize() != 0) {
    /* kinematic hardening variables first */
    if (this->is_kinem_hard) {
      this->giveStressBackVector (backStress, stressSpaceHardeningVars);
      helpVector = stressVector;
      helpVector.add (backStress);
      //delete backStress;
    } else helpVector = stressVector;
    
    f = this -> computeJ2InvariantAt (&helpVector);
    f12 = pow (f,1./2.);
    f32 = pow (f,3./2.);
    
    ax = helpVector[0];
    ay = helpVector[1];
    az = helpVector[2];
    
    df[0] = (2./3.)*ax - (1./3.)*ay - (1./3.)*az;
    df[1] = (2./3.)*ay - (1./3.)*ax - (1./3.)*az;
    df[2] = (2./3.)*az - (1./3.)*ay - (1./3.)*ax;
    df[3] = 2.* helpVector[3];
    df[4] = 2.* helpVector[4];
    df[5] = 2.* helpVector[5];
    
    //delete helpVector;
    
    for (i=1; i<=3; i++) {
      if ((imask = mask.at(i)) == 0) continue;
      for (j=i; j<=3; j++) {
        if ((jmask = mask.at(j)) == 0) continue;
        if (i==j) {
          answer.at(imask,jmask) = -(1./4.)*(1./f32)*df.at(i)*df.at(j) + 0.5 * (1./f12)* (4./6);
        } else{
          answer.at(imask,jmask) = -(1./4.)*(1./f32)*df.at(i)*df.at(j) + 0.5 * (1./f12)* (-2./6);
          answer.at(jmask,imask) = -(1./4.)*(1./f32)*df.at(i)*df.at(j) + 0.5 * (1./f12)* (-2./6);
        }
      }
    }
    
    for (i=1; i<=3; i++) {
      if ((imask = mask.at(i)) == 0) continue;
      for (j=4; j<=6; j++) {
        if ((jmask = mask.at(j)) == 0) continue;
        answer.at(imask,jmask) = -(1./4.)*(1./f32)*df.at(i)*df.at(j);
        answer.at(jmask,imask) = -(1./4.)*(1./f32)*df.at(i)*df.at(j);
      }
    }
    
    for (i=4; i<=6; i++) {
      if ((imask = mask.at(i)) == 0) continue;
      for (j=i; j<=6; j++) {
        if ((jmask = mask.at(j)) == 0) continue;
        if (i==j) {
          answer.at(imask,jmask) = -(1./4.)*(1./f32)*df.at(i)*df.at(j) + 0.5 * (1./f12)* 2.;
        } else {
          answer.at(imask,jmask) = -(1./4.)*(1./f32)*df.at(i)*df.at(j);
          answer.at(jmask,imask) = -(1./4.)*(1./f32)*df.at(i)*df.at(j);
        }
      }
    }
  }
  
  //delete df;
  //delete mask;
  /* for isotropic hardening: the corresponding part of gradient matrix is zero valued */
  
  return ;
}


void
J2plasticMaterial:: computeTrialStressIncrement (FloatArray& answer, GaussPoint *gp,
                                                 const FloatArray& strainIncrement, 
                                                 TimeStep* atTime)
{ /* Computes the full trial elastic stress vector */
  FloatArray   reducedAnswer;
  FloatMatrix  elastic3dModuli;
  FloatMatrix  reducedModuli;
  StructuralCrossSection *crossSection = (StructuralCrossSection*)
    (gp -> giveElement()->giveCrossSection());
  
  this->compute3dElasticModuli (elastic3dModuli, gp, atTime);
  this->reduceStiffMtrx3d (reducedModuli, ReducedForm , gp,
                           elastic3dModuli);
  //delete elastic3dModuli;
  reducedAnswer.beProductOf (reducedModuli, strainIncrement);
  crossSection->giveFullCharacteristicVector(answer, gp, reducedAnswer);
  //delete reducedAnswer;
  //delete reducedModuli;
  
  return ;
}


void
J2plasticMaterial:: compute3dElasticModuli(FloatMatrix& answer,
                                           GaussPoint *gp,
                                           TimeStep *atTime)
{
  /* Returns 3d elastic moduli */
  //FloatMatrix *FullConstitutiveMatrix = new FloatMatrix (6,6);
  double e, ee, nu;
  
  answer.resize (6,6);
  answer.zero();
  
  e  = this -> E ;
  nu = this -> nu;
  
  ee = e/((1.+nu)*(1.-2.*nu));
  
  answer.at(1,1) =  1.-nu;
  answer.at(1,2) =  nu;
  answer.at(1,3) =  nu; 
  answer.at(2,1) =  nu; 
  answer.at(2,2) =  1.-nu; 
  answer.at(2,3) =  nu; 
  answer.at(3,1) =  nu; 
  answer.at(3,2) =  nu; 
  answer.at(3,3) =  1.-nu; 
  
  answer.at(4,4) =  (1.-2.*nu)*0.5;
  answer.at(5,5) =  (1.-2.*nu)*0.5;
  answer.at(6,6) =  (1.-2.*nu)*0.5;
  
  answer.times(ee);
  return;
}


static double
computeJ2InvariantAt (double *stressVector) 
{
  double answer;
  double v1,v2,v3;
  
  v1 = ((stressVector[0]-stressVector[1])*(stressVector[0]-stressVector[1]));
  v2 = ((stressVector[1]-stressVector[2])*(stressVector[1]-stressVector[2]));
  v3 = ((stressVector[2]-stressVector[0])*(stressVector[2]-stressVector[0]));
  
  answer
    = (1./6.) * (v1 + v2 + v3)
    + stressVector[3]*stressVector[3]
    + stressVector[4]*stressVector[4]
    + stressVector[5]*stressVector[5];
  
  return answer;
}


int 
J2plasticMaterial:: giveSizeOfFullHardeningVarsVector() 
{
  /* Returns the size of hardening variables vector */
  int size = 0;
  
  if (is_kinem_hard) size+=6;  /* size of full stress vector */
  if (is_iso_hard) size+=1;  /* scalar value */
  
  return size;
}

int 
J2plasticMaterial:: giveSizeOfReducedHardeningVarsVector(GaussPoint* gp) 
{
  /* Returns the size of hardening variables vector */
  int size = 0;
  
  if (is_kinem_hard) size+=this->giveSizeOfReducedStressStrainVector(gp->giveMaterialMode());  
  if (is_iso_hard) size+=1;  /* scalar value */
  
  return size;
}


void
J2plasticMaterial :: giveStressBackVector (FloatArray& answer, 
                                           const FloatArray& stressSpaceHardeningVars)
{
  /* returns part of hardening vector corresponding to kinematic hardening */
  if (this->is_kinem_hard) {
    answer.resize (6);
    for (int i=1; i<=6; i++) answer.at(i) = stressSpaceHardeningVars.at(i);
    return ;
  }
  answer.resize (0);
  return ;
}


double
J2plasticMaterial :: giveIsotropicHardeningVar(FloatArray* stressSpaceHardeningVars)
{
  /* returns value in  hardening vector corresponding to isotropic hardening */
  if (!is_iso_hard) {
    return 0.;
  } else if (this->is_kinem_hard) {
    return stressSpaceHardeningVars->at(7);
  } else {
    return stressSpaceHardeningVars->at(1);
  }
  
  return 0.;
}
