 
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dhalq.h"

/* RCSHEADER "$Id: mat3dhalq.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

/* The material defined here is based on the model described in the */
/* LS-DYNA3D theory manual.  It uses the Jaumann stress rate. */

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_RATE_HALLQUIST;
}

#define SQRT_2_OVER_3 0.81649658092772603272


typedef struct halq_mat_t {
  double E, nu;
  double lambda, bulk_mod, G;
  double sigma_y, hardening_param;
  W_bool_t is_elastoplastic;
}              halq_mat_t;


typedef struct matstate_isotr_hard_t {
  double   equiv_pl_def;
  double   cauchy[6];           /* stores current Cauchy stress: */
                                /* s_xx, s_xy, s_xz, s_yy, s_yz, s_zz */
  double   curr_sigma_y;
  double   deform_energy_density;
}              matstate_isotr_hard_t;

typedef struct matstate_elastic_t {
  double   cauchy[6];           /* stores current Cauchy stress: */
                                /* s_xx, s_xy, s_xz, s_yy, s_yz, s_zz */
  double   deform_energy_density;
}              matstate_elastic_t;

static W_bool_t
get_mat_const  (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  halq_mat_t *u =  (halq_mat_t *)m->data;

  switch  (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_TENSILE_STRENGTH:
    *c = u->E / 100; /* RAW need to read TENSILE_STRENGTH from the input */
    break;
  case MAT_3D_INITIAL_YIELD_STRESS:
    *c = u->sigma_y;
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

static W_mat_3d_state_p 
get_mat_state_elastic (W_mat_3d_t *m, W_loc_t *at_point)
{
  matstate_isotr_hard_t *ms;
  int i;

  ms = w_new (matstate_isotr_hard_t); CHECK_MALLOC (ms);

  FOR (i, 6) ms->cauchy[i] = 0;

  ms->deform_energy_density = 0;

  return  (W_mat_3d_state_p)ms;
}


static W_bool_t 
get_stress_elastic (W_mat_3d_t                  *m,
                    W_mat_3d_state_p             ms,
                    W_mat_3d_ip_ctx_t           *ip_ctx,
                    W_mat_3d_stress_type_enum    stress_type,
                    double                       stress[3][3],
                    W_mat_3d_stress_type_enum   *out_stress_type
                    );
static W_bool_t 
get_stress_isotr_hard (W_mat_3d_t                  *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_stress_type_enum    stress_type,
                       double                       stress[3][3],
                       W_mat_3d_stress_type_enum   *out_stress_type
                       );

static W_bool_t  
update_stress_ep_isotr_hard (W_mat_3d_t       *m,
                             W_mat_3d_state_p             ms,
                             W_mat_state_update_form_enum update_form,
                             W_mat_3d_ip_ctx_t           *ip_ctx,
                             W_mat_3d_stress_type_enum    stress_type,
                             double                       stress[3][3],
                             W_mat_3d_stress_type_enum   *out_stress_type
                             )
{
  halq_mat_t *u =  (halq_mat_t *)m->data;
  matstate_isotr_hard_t *matst =  (matstate_isotr_hard_t *)ms;
  double lambda = u->lambda, twomu = 2 * u->G;
  double dt = ip_ctx->time_step;
  double dtl = lambda * dt, dttm = twomu * dt;
  double tr_msD;
  double msD[3][3], dtjaum[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
  double msW[3][3], r1[3][3], r2[3][3], rot[3][3], sigma[3][3];
  double elsig[3][3], dev_elsig[3][3];
  double curr_press, effective_stress, Gamma, beta;
  double cauchy_prev[6];
  W_bool_t result = TRUE;
  matstate_isotr_hard_t input_matst;

  if  (!ip_ctx->get_kinem_var) return FALSE;  

  /* kinematic quantities */
  if  (!(*ip_ctx->get_kinem_var) (ip_ctx,
                                  MAT_3D_MIDSTEP_RATE_OF_DEF, msD))
    return FALSE;
  if  (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_MIDSTEP_SPIN, msW))
    return FALSE;

  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;
  
  tr_msD = msD[0][0] + msD[1][1] + msD[2][2];
  
  dtjaum[0][0] = dtl * tr_msD + dttm * msD[0][0];
  dtjaum[0][1] =                dttm * msD[0][1];
  dtjaum[0][2] =                dttm * msD[0][2];
  dtjaum[1][0] =                dttm * msD[1][0];
  dtjaum[1][1] = dtl * tr_msD + dttm * msD[1][1];
  dtjaum[1][2] =                dttm * msD[1][2];
  dtjaum[2][0] =                dttm * msD[2][0];
  dtjaum[2][1] =                dttm * msD[2][1];
  dtjaum[2][2] = dtl * tr_msD + dttm * msD[2][2];

  sigma[0][0]               = cauchy_prev[0] = matst->cauchy[0];
  sigma[0][1] = sigma[1][0] = cauchy_prev[1] = matst->cauchy[1];
  sigma[0][2] = sigma[2][0] = cauchy_prev[2] = matst->cauchy[2];
  sigma[1][1]               = cauchy_prev[3] = matst->cauchy[3];
  sigma[1][2] = sigma[2][1] = cauchy_prev[4] = matst->cauchy[4];
  sigma[2][2]               = cauchy_prev[5] = matst->cauchy[5];
  
  MAT3DUTIL_MULT_AB_3X3  (r1, msW, sigma); 
  MAT3DUTIL_MULT_ATB_3X3 (r2, msW, sigma);
  MAT3DUTIL_ADD_AB_3X3 (rot, r1, r2);
  
  MAT3DUTIL_ADD_AB_3X3 (elsig, sigma, rot);
  MAT3DUTIL_ADD_AB_3X3 (elsig, elsig, dtjaum);
  
  curr_press      = (MAT3DUTIL_TRACE_3X3 (elsig)) / 3;
  dev_elsig[0][0] = elsig[0][0] - curr_press;
  dev_elsig[0][1] = dev_elsig[1][0] = elsig[0][1];
  dev_elsig[0][2] = dev_elsig[2][0] = elsig[0][2];
  dev_elsig[1][1] = elsig[1][1] - curr_press;
  dev_elsig[1][2] = dev_elsig[2][1] = elsig[1][2];
  dev_elsig[2][2] = elsig[2][2] - curr_press;

  /* elastic predictor: after rotation */
  effective_stress = MAT3DUTIL_DBL_PROD_3X3 (dev_elsig, dev_elsig);
  effective_stress = sqrt (effective_stress);

  /* plasticity criterion */
  if  (effective_stress > SQRT_2_OVER_3 * matst->curr_sigma_y) {
    /* linear hardening */
    Gamma =  ( (effective_stress - SQRT_2_OVER_3 * matst->curr_sigma_y)
             /  (twomu *  (1 + u->hardening_param / 3 / u->G)));
    beta = 1 - twomu * Gamma / effective_stress;
    matst->equiv_pl_def += SQRT_2_OVER_3 * Gamma;
    matst->curr_sigma_y += SQRT_2_OVER_3 * u->hardening_param * Gamma;
    matst->cauchy[0] = curr_press + beta * dev_elsig[0][0];
    matst->cauchy[1] = beta * dev_elsig[0][1];
    matst->cauchy[2] = beta * dev_elsig[0][2];
    matst->cauchy[3] = curr_press + beta * dev_elsig[1][1];
    matst->cauchy[4] = beta * dev_elsig[1][2];
    matst->cauchy[5] = curr_press + beta * dev_elsig[2][2];
  } else {
    /* elastic state */
    matst->cauchy[0] = elsig[0][0];
    matst->cauchy[1] = elsig[0][1];
    matst->cauchy[2] = elsig[0][2];
    matst->cauchy[3] = elsig[1][1];
    matst->cauchy[4] = elsig[1][2]; 
    matst->cauchy[5] = elsig[2][2];
  }

  matst->deform_energy_density +=
    dt * (0.5 * ((cauchy_prev[0] + matst->cauchy[0]) * msD[0][0] +
                 (cauchy_prev[3] + matst->cauchy[3]) * msD[1][1] +
                 (cauchy_prev[5] + matst->cauchy[5]) * msD[2][2]) +
          ((cauchy_prev[1] + matst->cauchy[1]) * msD[0][1] +
           (cauchy_prev[2] + matst->cauchy[2]) * msD[0][2] +
           (cauchy_prev[4] + matst->cauchy[4]) * msD[1][2])
          );
  
  if  (stress_type == MAT_3D_1ST_PK) {
    double F[3][3], F_cof[3][3], sigma[3][3];
    if  (! (*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F)) {
      result = FALSE; goto exit;
    }
    sigma[0][0]               = matst->cauchy[0];
    sigma[0][1] = sigma[1][0] = matst->cauchy[1];
    sigma[0][2] = sigma[2][0] = matst->cauchy[2];
    sigma[1][1]               = matst->cauchy[3];
    sigma[1][2] = sigma[2][1] = matst->cauchy[4];
    sigma[2][2]               = matst->cauchy[5];
    MAT3DUTIL_COFACTOR_3X3 (F, F_cof);
    MAT3DUTIL_MULT_ABT_3X3 (stress, sigma, F_cof);
    *out_stress_type = MAT_3D_1ST_PK;
  } else 
    result = get_stress_isotr_hard (m, ms, ip_ctx,
                                    stress_type, stress, out_stress_type);
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

static W_bool_t  
update_stress_elastic (W_mat_3d_t                  *m,
                       W_mat_3d_state_p             ms,
                       W_mat_state_update_form_enum update_form,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_stress_type_enum    stress_type,
                       double                       stress[3][3],
                       W_mat_3d_stress_type_enum   *out_stress_type
                       )
{
  halq_mat_t *u =  (halq_mat_t *)m->data;
  matstate_elastic_t *matst =  (matstate_elastic_t *)ms;
  double lambda = u->lambda, twomu = 2 * u->G;
  double dt = ip_ctx->time_step;
  double dtl = lambda * dt, dttm = twomu * dt;
  double tr_msD;
  double msD[3][3], dtjaum[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
  double msW[3][3], sigma[3][3];
  double elsig[3][3];
  double cauchy_prev[6];
  W_bool_t result = TRUE;
  matstate_elastic_t input_matst;

  if  (!ip_ctx->get_kinem_var) return FALSE;  

  /* kinematic quantities */
  if  (!(*ip_ctx->get_kinem_var) (ip_ctx,
                                  MAT_3D_MIDSTEP_RATE_OF_DEF, msD))
    return FALSE;
  if  (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_MIDSTEP_SPIN, msW))
    return FALSE;

  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;
  
  tr_msD = msD[0][0] + msD[1][1] + msD[2][2];
  
  dtjaum[0][0] = dtl * tr_msD + dttm * msD[0][0];
  dtjaum[0][1] =                dttm * msD[0][1];
  dtjaum[0][2] =                dttm * msD[0][2];
  dtjaum[1][0] =                dttm * msD[1][0];
  dtjaum[1][1] = dtl * tr_msD + dttm * msD[1][1];
  dtjaum[1][2] =                dttm * msD[1][2];
  dtjaum[2][0] =                dttm * msD[2][0];
  dtjaum[2][1] =                dttm * msD[2][1];
  dtjaum[2][2] = dtl * tr_msD + dttm * msD[2][2];

  sigma[0][0]               = cauchy_prev[0] = matst->cauchy[0];
  sigma[0][1] = sigma[1][0] = cauchy_prev[1] = matst->cauchy[1];
  sigma[0][2] = sigma[2][0] = cauchy_prev[2] = matst->cauchy[2];
  sigma[1][1]               = cauchy_prev[3] = matst->cauchy[3];
  sigma[1][2] = sigma[2][1] = cauchy_prev[4] = matst->cauchy[4];
  sigma[2][2]               = cauchy_prev[5] = matst->cauchy[5];
  
  /* elastic predictor: corotational */
  MAT3DUTIL_ADD_AB_3X3 (elsig, sigma, dtjaum);
  
  matst->cauchy[0] = elsig[0][0];
  matst->cauchy[1] = elsig[0][1];
  matst->cauchy[2] = elsig[0][2];
  matst->cauchy[3] = elsig[1][1];
  matst->cauchy[4] = elsig[1][2]; 
  matst->cauchy[5] = elsig[2][2];

  matst->deform_energy_density +=
    dt * (0.5 * ((cauchy_prev[0] + matst->cauchy[0]) * msD[0][0] +
                 (cauchy_prev[3] + matst->cauchy[3]) * msD[1][1] +
                 (cauchy_prev[5] + matst->cauchy[5]) * msD[2][2]) +
          ((cauchy_prev[1] + matst->cauchy[1]) * msD[0][1] +
           (cauchy_prev[2] + matst->cauchy[2]) * msD[0][2] +
           (cauchy_prev[4] + matst->cauchy[4]) * msD[1][2])
          );
  
  if (matst->deform_energy_density < 0)
    REPORT_WARNING (NULL, "Non-positive deform_energy_density");
  if  (stress_type == MAT_3D_1ST_PK) {
    double F[3][3], F_cof[3][3], sigma[3][3];
    if  (! (*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F)) {
      result = FALSE; goto exit;
    }
    sigma[0][0]               = matst->cauchy[0];
    sigma[0][1] = sigma[1][0] = matst->cauchy[1];
    sigma[0][2] = sigma[2][0] = matst->cauchy[2];
    sigma[1][1]               = matst->cauchy[3];
    sigma[1][2] = sigma[2][1] = matst->cauchy[4];
    sigma[2][2]               = matst->cauchy[5];
    MAT3DUTIL_COFACTOR_3X3 (F, F_cof);
    MAT3DUTIL_MULT_ABT_3X3 (stress, sigma, F_cof);
    *out_stress_type = MAT_3D_1ST_PK;
  } else 
    result = get_stress_elastic (m, ms, ip_ctx,
                                 stress_type, stress, out_stress_type);
   
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

static W_mat_3d_state_p 
get_mat_state_isotr_hard (W_mat_3d_t *m, W_loc_t *at_point)
{
  halq_mat_t *u =  (halq_mat_t *)m->data;
  matstate_isotr_hard_t *ms;
  int i;

  ms = w_new (matstate_isotr_hard_t); CHECK_MALLOC (ms);

  ms->equiv_pl_def = 0;
  FOR (i, 6) ms->cauchy[i] = 0;
  ms->curr_sigma_y = u->sigma_y;

  ms->deform_energy_density = 0;

  return  (W_mat_3d_state_p) ms;
}

/* Function returns the stress saved in the material state  (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress_isotr_hard (W_mat_3d_t                  *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_stress_type_enum    stress_type,
                       double                       stress[3][3],
                       W_mat_3d_stress_type_enum   *out_stress_type
                       )
{
  matstate_isotr_hard_t *matst =  (matstate_isotr_hard_t *)ms;
  double sigma[3][3], F[3][3], inv_F[3][3], P[3][3], S[3][3];
  double J;
  
  sigma[0][0]               = matst->cauchy[0];
  sigma[0][1] = sigma[1][0] = matst->cauchy[1];
  sigma[0][2] = sigma[2][0] = matst->cauchy[2];
  sigma[1][1]               = matst->cauchy[3];
  sigma[1][2] = sigma[2][1] = matst->cauchy[4];
  sigma[2][2]               = matst->cauchy[5];
    
  *out_stress_type = stress_type;
  switch  (stress_type) {
  case MAT_3D_CAUCHY:
    MAT3DUTIL_COPY_3X3 (stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if  (! (*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3 (F);
    if  (!W_mat3dutil_invert_3x3 (F, inv_F))
      REPORT_FATAL_ERROR (NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3 (P, sigma, inv_F);
    MAT3DUTIL_SCALE_3X3 (P, J);
    MAT3DUTIL_COPY_3X3 (stress, P);
    break;
  case MAT_3D_2ND_PK:
    if  (! (*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3 (F);
    if  (!W_mat3dutil_invert_3x3 (F, inv_F))
      REPORT_FATAL_ERROR (NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3 (P, sigma, inv_F);
    MAT3DUTIL_MULT_AB_3X3 (S, inv_F, P);
    MAT3DUTIL_SCALE_3X3 (S, J);
    MAT3DUTIL_COPY_3X3 (stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3 (stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}

static W_bool_t 
get_stress_elastic (W_mat_3d_t                  *m,
                    W_mat_3d_state_p             ms,
                    W_mat_3d_ip_ctx_t           *ip_ctx,
                    W_mat_3d_stress_type_enum    stress_type,
                    double                       stress[3][3],
                    W_mat_3d_stress_type_enum   *out_stress_type
                    )
{
  matstate_elastic_t *matst =  (matstate_elastic_t *)ms;
  double sigma[3][3], F[3][3], inv_F[3][3], P[3][3], S[3][3];
  double J;
  
  sigma[0][0]               = matst->cauchy[0];
  sigma[0][1] = sigma[1][0] = matst->cauchy[1];
  sigma[0][2] = sigma[2][0] = matst->cauchy[2];
  sigma[1][1]               = matst->cauchy[3];
  sigma[1][2] = sigma[2][1] = matst->cauchy[4];
  sigma[2][2]               = matst->cauchy[5];
    
  *out_stress_type = stress_type;
  switch  (stress_type) {
  case MAT_3D_CAUCHY:
    MAT3DUTIL_COPY_3X3 (stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if  (! (*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3 (F);
    if  (!W_mat3dutil_invert_3x3 (F, inv_F))
      REPORT_FATAL_ERROR (NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3 (P, sigma, inv_F);
    MAT3DUTIL_SCALE_3X3 (P, J);
    MAT3DUTIL_COPY_3X3 (stress, P);
    break;
  case MAT_3D_2ND_PK:
    if  (! (*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3 (F);
    if  (!W_mat3dutil_invert_3x3 (F, inv_F))
      REPORT_FATAL_ERROR (NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3 (P, sigma, inv_F);
    MAT3DUTIL_MULT_AB_3X3 (S, inv_F, P);
    MAT3DUTIL_SCALE_3X3 (S, J);
    MAT3DUTIL_COPY_3X3 (stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3 (stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}

static W_bool_t 
get_mat_state_var_isotr_hard (W_mat_3d_t           *m,
                              W_mat_3d_state_p             ms,
                              W_mat_3d_ip_ctx_t           *ip_ctx,
                              W_mat_3d_mat_state_var_enum  var_type,
                              double                      *var
                              )
{
  matstate_isotr_hard_t *matst =  (matstate_isotr_hard_t *)ms;
  
  switch (var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = matst->equiv_pl_def; 
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;
  default:
    *var = 0;
    return FALSE;
  }
}

static W_bool_t 
get_mat_state_var_elastic (W_mat_3d_t           *m,
                           W_mat_3d_state_p             ms,
                           W_mat_3d_ip_ctx_t           *ip_ctx,
                           W_mat_3d_mat_state_var_enum  var_type,
                           double                      *var
                           )
{
  matstate_elastic_t *matst =  (matstate_elastic_t *)ms;
  switch (var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0;
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;
  default:
    *var = 0;
    return FALSE;
  }
}

static W_mat_3d_state_p 
interpolate_mat_state_elastic (W_mat_3d_t *m,
                               W_loc_t *at_point,
                               int npkgs,
                               WM3D_mat_state_pkg_t *pkgs)
{
  matstate_elastic_t *ms;
  int i, k;
  double weight_sum = 0;
  matstate_elastic_t *pms; 
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_elastic_t *) get_mat_state_elastic(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_elastic_t *)pkgs[k].ms;
    FOR(i, 6) {
      ms->cauchy[i] += pms->cauchy[i] * WGHT(k);
    }
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p) ms;
}

static W_mat_3d_state_p 
interpolate_mat_state_isotr_hard (W_mat_3d_t *m,
                                  W_loc_t *at_point,
                                  int npkgs,
                                  WM3D_mat_state_pkg_t *pkgs)
{
  matstate_isotr_hard_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_isotr_hard_t *) get_mat_state_isotr_hard(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_isotr_hard_t *)pkgs[k].ms;
    ms->equiv_pl_def += pms->equiv_pl_def * WGHT(k);
    FOR(i, 6) {
      ms->cauchy[i] += pms->cauchy[i] * WGHT(k);
    }
    ms->curr_sigma_y += pms->curr_sigma_y * WGHT(k);
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k); 
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}

static W_mat_3d_methods_t mat3dhalq_ep_isotr_hard_methods = {
  get_mcode,
  get_mat_state_isotr_hard,
  get_mat_const,
  update_stress_ep_isotr_hard,
  get_mat_state_var_isotr_hard,
  interpolate_mat_state_isotr_hard,
  NULL,
  NULL,
  NULL
};

static W_mat_3d_methods_t mat3dhalq_elastic_methods = {
  get_mcode,
  get_mat_state_elastic,
  get_mat_const,
  update_stress_elastic,
  get_mat_state_var_elastic,
  interpolate_mat_state_elastic,
  NULL,
  NULL,
  NULL
};

static W_mat_3d_t *
new_mat_3d_halq  (double E, double nu, double sigma_y, double hardening_param,
                 double delta, double rho)
{
  halq_mat_t *u;
  
  u = w_new  (halq_mat_t);
  CHECK_MALLOC (u);
  u->E               = E; 
  u->nu              = nu;
  if  (u->nu >= 0.5) REPORT_FATAL_ERROR  (NULL, "Bad nu  (should be 0 <= nu < 0.5)");
  u->lambda          = u->E * u->nu /  (1 + u->nu) /  (1 - 2* (u->nu));
  u->G               = u->E / 2 /  (1 + u->nu);
  u->bulk_mod        = u->E / 3 /  (1 - 2*u->nu);
  u->sigma_y         = sigma_y;
  if  (u->sigma_y == 0) u->is_elastoplastic = FALSE;
  else                  u->is_elastoplastic = TRUE;
  u->hardening_param = hardening_param;
  
  return WM3D_new_material  (
                             (u->is_elastoplastic
                              ? &mat3dhalq_ep_isotr_hard_methods
                              : &mat3dhalq_elastic_methods),
                             (void *)u, rho);

}


W_mat_3d_t *
read_mat_3d_halq  (W_domain_t *d)
{
  const char *input_line;
  double E, nu, rho = 1;
  double sigma_y = 0, hardening_param = 0, delta = 1;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { E = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { nu = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("SIGMA_Y", FALSE,
			    {
			      sigma_y = TOKEN_DBL_NEXT(1);
			    },
			    {
                              sigma_y = 0;
                            }
			    );
      TOKENS_SET_BY_KEYWORD("HARDENING_PARAM", FALSE,
			    { hardening_param = TOKEN_DBL_NEXT(1); },
			    { hardening_param = 0; }/* perfectly ep material */
			    );
      return new_mat_3d_halq (E, nu, sigma_y, hardening_param, delta, rho);
    }
  }
  return NULL;
}


