 
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3difr.h"

/* RCSHEADER "$Id: mat3difr.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */
  

typedef struct j2ep_mat_t {
  double E, nu;
  double bulk_mod, G;
  double sigma_y, hardening_param, delta;
  double dil_wave_speed_3d;
  double dil_wave_speed_pstrain;
  double dil_wave_speed_pstress;
  double shear_wave_speed;
  double rayleigh_pstrain_wave_speed;
  double rayleigh_pstress_wave_speed;
  double tensile_strength;
  double dyn_fracture_toughness;
  char   is_elastoplastic;      /* TRUE/FALSE */
  double e_incr; /* for numerical approximation of material tangent */
}              j2ep_mat_t; 


typedef struct matstate_mixed_hard_t {
  double   equiv_pl_def;
  double   cauchy[6];           /* stores current Cauchy stress: */
                                /* s_xx, s_xy, s_xz, s_yy, s_yz, s_zz */
  double   back_stress[6];      /* stores back-stress: */
                                /* alpha_xx, ..., alpha_zz */
  double   curr_sigma_y;
  double   deform_energy_density;
}              matstate_mixed_hard_t;


typedef struct matstate_isotr_hard_t {
  double   equiv_pl_def;
  double   cauchy[6];           /* stores current Cauchy stress: */
                                /* s_xx, s_xy, s_xz, s_yy, s_yz, s_zz */
  double   curr_sigma_y;
  double   deform_energy_density;
}              matstate_isotr_hard_t;


typedef struct matstate_elastic_t {
  double   cauchy[6];           /* stores current Cauchy stress: */
                                /* s_xx, s_xy, s_xz, s_yy, s_yz, s_zz */
  double   deform_energy_density;
}              matstate_elastic_t;









static W_bool_t
get_mat_const(W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  j2ep_mat_t *u = (j2ep_mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_TENSILE_STRENGTH:
    *c = u->tensile_strength; 
    break;
  case MAT_3D_DYN_FRACTURE_TOUGHNESS:
    *c = u->dyn_fracture_toughness; 
    break;
  case MAT_3D_DIL_WAVE_SPEED_3D:
    *c = u->dil_wave_speed_3d;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRAIN:
    *c = u->dil_wave_speed_pstrain;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRESS:
    *c = u->dil_wave_speed_pstress;
    break;
  case MAT_3D_SHEAR_WAVE_SPEED:
    *c = u->shear_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED:
    *c = u->rayleigh_pstrain_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRESS_WAVE_SPEED:
    *c = u->rayleigh_pstress_wave_speed;
    break;
  case MAT_3D_INITIAL_YIELD_STRESS:
    *c = u->sigma_y;
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}


static W_mat_3d_state_p 
get_mat_state_elastic(W_mat_3d_t *m, W_loc_t *at_point)
{
  matstate_elastic_t *ms;
  int i;

  ms = w_new(matstate_elastic_t); CHECK_MALLOC(ms);

  FOR(i, 6) ms->cauchy[i] = 0;

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

 

static W_mat_3d_state_p 
get_mat_state(W_mat_3d_t *m, W_loc_t *at_point)
{
  j2ep_mat_t *u = (j2ep_mat_t *)m->data;
  matstate_mixed_hard_t *ms;
  int i;

  ms = w_new(matstate_mixed_hard_t); CHECK_MALLOC(ms);

  ms->equiv_pl_def = 0;
  FOR(i, 6) ms->cauchy[i] = 0;
  FOR(i, 6) ms->back_stress[i] = 0;
  ms->curr_sigma_y = u->sigma_y;

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}


static W_mat_3d_state_p 
get_mat_state_isotr_hard(W_mat_3d_t *m, W_loc_t *at_point)
{
  j2ep_mat_t *u = (j2ep_mat_t *)m->data;
  matstate_isotr_hard_t *ms;
  int i;

  ms = w_new(matstate_isotr_hard_t); CHECK_MALLOC(ms);

  ms->equiv_pl_def = 0;
  FOR(i, 6) ms->cauchy[i] = 0;
  ms->curr_sigma_y = u->sigma_y;

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}


static W_bool_t 
get_stress_isotr_hard(W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           );
static W_bool_t 
get_stress_elastic(W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           );
static W_bool_t 
get_stress_ep(W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           );


static W_bool_t  
update_stress_elastic (W_mat_3d_t                  *m,
                       W_mat_3d_state_p             ms,
                       W_mat_state_update_form_enum update_form,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_stress_type_enum    stress_type,
                       double                       stress[3][3],
                       W_mat_3d_stress_type_enum   *out_stress_type
                       )
{
  double rel_F[3][3], rel_R[3][3], ln_rel_U[3][3];
  double rel_J, ln_rel_J, press_predictor, curr_press;
  j2ep_mat_t *u = (j2ep_mat_t *)m->data;
  matstate_elastic_t *matst = (matstate_elastic_t *)ms;
  double twomu = 2 * u->G;
  double dev_sig_corot[3][3], dev_sig_predictor[3][3], t[3][3];
  double msD[3][3], F[3][3];
  double tr_ln_rel_U_3;
  double dt = ip_ctx->time_step;
  double cauchy_prev[6];
#define SQRT_2_OVER_3 0.81649658092772603272
  W_bool_t result = TRUE;
  matstate_elastic_t input_matst;

  if (!ip_ctx->get_kinem_var) return FALSE;  

  /* kinematic quantities */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_REL_DEF_GRAD, rel_F))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_LN_POLAR_REL_U, ln_rel_U))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_POLAR_REL_R, rel_R))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
  
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  rel_J = MAT3DUTIL_DET_3X3(rel_F);
  LN_FUNC(ln_rel_J, /* = log */ rel_J);

  tr_ln_rel_U_3 = (ln_rel_U[0][0] + ln_rel_U[1][1] + ln_rel_U[2][2]) / 3;
  ln_rel_U[0][0] -= tr_ln_rel_U_3;
  ln_rel_U[1][1] -= tr_ln_rel_U_3;
  ln_rel_U[2][2] -= tr_ln_rel_U_3;

  cauchy_prev[0] = matst->cauchy[0];
  cauchy_prev[1] = matst->cauchy[1];
  cauchy_prev[2] = matst->cauchy[2];
  cauchy_prev[3] = matst->cauchy[3];
  cauchy_prev[4] = matst->cauchy[4];
  cauchy_prev[5] = matst->cauchy[5];
  
  /* elastic predictor: corotational */
  curr_press      = (matst->cauchy[0] + matst->cauchy[3] + matst->cauchy[5])/3;
  press_predictor =  curr_press + u->bulk_mod * ln_rel_J;
  dev_sig_corot[0][0]
    = matst->cauchy[0] - curr_press + twomu * ln_rel_U[0][0];
  dev_sig_corot[0][1] = dev_sig_corot[1][0]
    = matst->cauchy[1] + twomu * ln_rel_U[0][1];
  dev_sig_corot[0][2] = dev_sig_corot[2][0]
    = matst->cauchy[2] + twomu * ln_rel_U[0][2];
  dev_sig_corot[1][1]
    = matst->cauchy[3] - curr_press  + twomu * ln_rel_U[1][1];
  dev_sig_corot[1][2] = dev_sig_corot[2][1]
    = matst->cauchy[4] + twomu * ln_rel_U[1][2];
  dev_sig_corot[2][2]
    = matst->cauchy[5] - curr_press  + twomu * ln_rel_U[2][2];

  /* elastic predictor: after rotation */
  MAT3DUTIL_MULT_ABT_3X3(t, dev_sig_corot, rel_R);
  MAT3DUTIL_MULT_AB_3X3(dev_sig_predictor, rel_R, t);

  /* save state */
  matst->cauchy[0] = press_predictor + dev_sig_predictor[0][0];
  matst->cauchy[1] = dev_sig_predictor[0][1];
  matst->cauchy[2] = dev_sig_predictor[0][2];
  matst->cauchy[3] = press_predictor + dev_sig_predictor[1][1];
  matst->cauchy[4] = dev_sig_predictor[1][2];
  matst->cauchy[5] = press_predictor + dev_sig_predictor[2][2];

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_MIDSTEP_RATE_OF_DEF, msD)) {
    result = FALSE; goto exit;
  }
  matst->deform_energy_density +=
    dt * (0.5 * ((cauchy_prev[0] + matst->cauchy[0]) * msD[0][0] +
                 (cauchy_prev[3] + matst->cauchy[3]) * msD[1][1] +
                 (cauchy_prev[5] + matst->cauchy[5]) * msD[2][2]) +
          ((cauchy_prev[1] + matst->cauchy[1]) * msD[0][1] +
           (cauchy_prev[2] + matst->cauchy[2]) * msD[0][2] +
           (cauchy_prev[4] + matst->cauchy[4]) * msD[1][2])
          ) * MAT3DUTIL_DET_3X3(F);

  result = get_stress_elastic (m, ms, ip_ctx,
                               stress_type, stress, out_stress_type);
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}


static W_bool_t  
update_stress_ep (W_mat_3d_t                  *m,
                  W_mat_3d_state_p             ms,
                  W_mat_state_update_form_enum update_form,
                  W_mat_3d_ip_ctx_t           *ip_ctx,
                  W_mat_3d_stress_type_enum    stress_type,
                  double                       stress[3][3],
                  W_mat_3d_stress_type_enum   *out_stress_type
                  )
{
  double rel_F[3][3], rel_R[3][3], ln_rel_U[3][3];
  double rel_J, ln_rel_J, press_predictor, curr_press;
  j2ep_mat_t *u = (j2ep_mat_t *)m->data;
  matstate_mixed_hard_t *matst = (matstate_mixed_hard_t *)ms;
  double twomu = 2 * u->G;
  double dev_sig_corot[3][3], dev_sig_predictor[3][3], dev_sig[3][3], t[3][3];
  double alpha_corot[3][3], alpha_predictor[3][3], alpha[3][3];
  double N[3][3];
  double effective_stress, tr_ln_rel_U_3, fact, Gamma;
  double msD[3][3], F[3][3];
  double dt = ip_ctx->time_step;
  double cauchy_prev[6];
#define SQRT_2_OVER_3 0.81649658092772603272
  W_bool_t result = TRUE;
  matstate_mixed_hard_t input_matst;

  cauchy_prev[0] = matst->cauchy[0];
  cauchy_prev[1] = matst->cauchy[1];
  cauchy_prev[2] = matst->cauchy[2];
  cauchy_prev[3] = matst->cauchy[3];
  cauchy_prev[4] = matst->cauchy[4];
  cauchy_prev[5] = matst->cauchy[5];
  
  if (!ip_ctx->get_kinem_var) return FALSE;  

  /* kinematic quantities */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_REL_DEF_GRAD, rel_F))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_LN_POLAR_REL_U, ln_rel_U))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_POLAR_REL_R, rel_R))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
  
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  rel_J = MAT3DUTIL_DET_3X3(rel_F);
  LN_FUNC(ln_rel_J, /* = log */ rel_J);

  tr_ln_rel_U_3 = (ln_rel_U[0][0] + ln_rel_U[1][1] + ln_rel_U[2][2]) / 3;
  ln_rel_U[0][0] -= tr_ln_rel_U_3;
  ln_rel_U[1][1] -= tr_ln_rel_U_3;
  ln_rel_U[2][2] -= tr_ln_rel_U_3;
  
  /* elastic predictor: corotational */
  curr_press      = (matst->cauchy[0] + matst->cauchy[3] + matst->cauchy[5])/3;
  press_predictor = curr_press + u->bulk_mod * ln_rel_J;
  dev_sig_corot[0][0]
    = matst->cauchy[0] - curr_press + twomu * ln_rel_U[0][0];
  dev_sig_corot[0][1] = dev_sig_corot[1][0]
    = matst->cauchy[1] + twomu * ln_rel_U[0][1];
  dev_sig_corot[0][2] = dev_sig_corot[2][0]
    = matst->cauchy[2] + twomu * ln_rel_U[0][2];
  dev_sig_corot[1][1]
    = matst->cauchy[3] - curr_press  + twomu * ln_rel_U[1][1];
  dev_sig_corot[1][2] = dev_sig_corot[2][1]
    = matst->cauchy[4] + twomu * ln_rel_U[1][2];
  dev_sig_corot[2][2]
    = matst->cauchy[5] - curr_press  + twomu * ln_rel_U[2][2];

  alpha_corot[0][0] = matst->back_stress[0];
  alpha_corot[0][1] = alpha_corot[1][0] = matst->back_stress[1];
  alpha_corot[0][2] = alpha_corot[2][0] = matst->back_stress[2];
  alpha_corot[1][1] = matst->back_stress[3];
  alpha_corot[1][2] = alpha_corot[2][1] = matst->back_stress[4];
  alpha_corot[2][2] = matst->back_stress[5];

  /* elastic predictor: after rotation */
  MAT3DUTIL_MULT_ABT_3X3(t, dev_sig_corot, rel_R);
  MAT3DUTIL_MULT_AB_3X3(dev_sig_predictor, rel_R, t);
  MAT3DUTIL_MULT_ABT_3X3(t, alpha_corot, rel_R);
  MAT3DUTIL_MULT_AB_3X3(alpha_predictor, rel_R, t);
  MAT3DUTIL_SUB_AB_3X3(N, dev_sig_predictor, alpha_predictor);
  effective_stress = MAT3DUTIL_DBL_PROD_3X3(N, N);
  effective_stress = sqrt(effective_stress);

  /* plasticity criterion */
  if (effective_stress > SQRT_2_OVER_3 * matst->curr_sigma_y) {
    fact             = 1 / effective_stress;
    MAT3DUTIL_SCALE_3X3(N, fact);    /* normal unit */
    /* linear hardening */
    Gamma = ((effective_stress - SQRT_2_OVER_3 * matst->curr_sigma_y)
             / (twomu * (1 + u->hardening_param / 3 / u->G)));
    MAT3DUTIL_COPY_3X3(t, N);
    fact = -twomu * Gamma;
    MAT3DUTIL_SCALE_3X3(t, fact);
    MAT3DUTIL_ADD_AB_3X3(dev_sig, dev_sig_predictor, t);
    matst->equiv_pl_def += SQRT_2_OVER_3 * Gamma;
    MAT3DUTIL_COPY_3X3(t, N);
    fact = 2 / 3 * (1 - u->delta) * u->hardening_param * Gamma;
    MAT3DUTIL_SCALE_3X3(t, fact);
    MAT3DUTIL_ADD_AB_3X3(alpha, alpha_predictor, t);
    matst->curr_sigma_y
      += SQRT_2_OVER_3 * u->delta * u->hardening_param * Gamma;
  } else {
    /* elastic state */
    MAT3DUTIL_COPY_3X3(dev_sig, dev_sig_predictor);
    MAT3DUTIL_COPY_3X3(alpha, alpha_predictor);
  }

  /* save state */
  matst->cauchy[0] = press_predictor + dev_sig[0][0];
  matst->cauchy[1] = dev_sig[0][1];
  matst->cauchy[2] = dev_sig[0][2];
  matst->cauchy[3] = press_predictor + dev_sig[1][1];
  matst->cauchy[4] = dev_sig[1][2];
  matst->cauchy[5] = press_predictor + dev_sig[2][2];

  matst->back_stress[0] = alpha[0][0];
  matst->back_stress[1] = alpha[0][1];
  matst->back_stress[2] = alpha[0][2];
  matst->back_stress[3] = alpha[1][1];
  matst->back_stress[4] = alpha[1][2];
  matst->back_stress[5] = alpha[2][2];

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_MIDSTEP_RATE_OF_DEF, msD)) {
    result = FALSE; goto exit;
  }
  matst->deform_energy_density +=
    dt * (0.5 * ((cauchy_prev[0] + matst->cauchy[0]) * msD[0][0] +
                 (cauchy_prev[3] + matst->cauchy[3]) * msD[1][1] +
                 (cauchy_prev[5] + matst->cauchy[5]) * msD[2][2]) +
          ((cauchy_prev[1] + matst->cauchy[1]) * msD[0][1] +
           (cauchy_prev[2] + matst->cauchy[2]) * msD[0][2] +
           (cauchy_prev[4] + matst->cauchy[4]) * msD[1][2])
          ) * MAT3DUTIL_DET_3X3(F);

  result = get_stress_ep (m, ms, ip_ctx, stress_type, stress, out_stress_type);
   
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}







/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress_ep (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  matstate_mixed_hard_t *matst = (matstate_mixed_hard_t *)ms;
  double sigma[3][3], F[3][3], inv_F[3][3], P[3][3], S[3][3];
  double J;

  sigma[0][0]               = matst->cauchy[0];
  sigma[0][1] = sigma[1][0] = matst->cauchy[1];
  sigma[0][2] = sigma[2][0] = matst->cauchy[2];
  sigma[1][1]               = matst->cauchy[3];
  sigma[1][2] = sigma[2][1] = matst->cauchy[4];
  sigma[2][2]               = matst->cauchy[5];
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
    MAT3DUTIL_SCALE_3X3(P, J);
    MAT3DUTIL_COPY_3X3(stress, P);
    break;
  case MAT_3D_2ND_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
    MAT3DUTIL_MULT_AB_3X3(S, inv_F, P);
    MAT3DUTIL_SCALE_3X3(S, J);
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}



static W_bool_t 
get_mat_state_var_ep (W_mat_3d_t           *m,
                      W_mat_3d_state_p             ms,
                      W_mat_3d_ip_ctx_t           *ip_ctx,
                      W_mat_3d_mat_state_var_enum  var_type,
                      double                      *var
                      )
{
  matstate_mixed_hard_t *matst = (matstate_mixed_hard_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = matst->equiv_pl_def; 
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;
  default:
    return FALSE;
  }
}




























/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress_isotr_hard(W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           )
{
  matstate_isotr_hard_t *matst = (matstate_isotr_hard_t *)ms;
  double sigma[3][3], F[3][3], inv_F[3][3], P[3][3], S[3][3];
  double J;

  sigma[0][0]               = matst->cauchy[0];
  sigma[0][1] = sigma[1][0] = matst->cauchy[1];
  sigma[0][2] = sigma[2][0] = matst->cauchy[2];
  sigma[1][1]               = matst->cauchy[3];
  sigma[1][2] = sigma[2][1] = matst->cauchy[4];
  sigma[2][2]               = matst->cauchy[5];
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
    MAT3DUTIL_SCALE_3X3(P, J);
    MAT3DUTIL_COPY_3X3(stress, P);
    break;
  case MAT_3D_2ND_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
    MAT3DUTIL_MULT_AB_3X3(S, inv_F, P);
    MAT3DUTIL_SCALE_3X3(S, J);
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}



static W_bool_t 
get_mat_state_var_isotr_hard(W_mat_3d_t           *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_isotr_hard_t *matst = (matstate_isotr_hard_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = matst->equiv_pl_def; 
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}


static W_bool_t 
get_mat_state_var_elastic(W_mat_3d_t           *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_elastic_t *matst = (matstate_elastic_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0;
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}



/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress_elastic(W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           )
{
  matstate_elastic_t *matst = (matstate_elastic_t *)ms;
  double sigma[3][3], F[3][3], inv_F[3][3], P[3][3], S[3][3];
  double J;

  sigma[0][0]               = matst->cauchy[0];
  sigma[0][1] = sigma[1][0] = matst->cauchy[1];
  sigma[0][2] = sigma[2][0] = matst->cauchy[2];
  sigma[1][1]               = matst->cauchy[3];
  sigma[1][2] = sigma[2][1] = matst->cauchy[4];
  sigma[2][2]               = matst->cauchy[5];
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
    MAT3DUTIL_SCALE_3X3(P, J);
    MAT3DUTIL_COPY_3X3(stress, P);
    break;
  case MAT_3D_2ND_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
    MAT3DUTIL_MULT_AB_3X3(S, inv_F, P);
    MAT3DUTIL_SCALE_3X3(S, J);
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}






 













static W_bool_t  
update_stress_ep_isotr_hard (W_mat_3d_t                  *m,
                             W_mat_3d_state_p             ms,
                             W_mat_state_update_form_enum update_form,
                             W_mat_3d_ip_ctx_t           *ip_ctx,
                             W_mat_3d_stress_type_enum    stress_type,
                             double                       stress[3][3],
                             W_mat_3d_stress_type_enum   *out_stress_type
                             )
{
  double rel_F[3][3], rel_R[3][3], ln_rel_U[3][3];
  double rel_J, ln_rel_J, press_predictor, curr_press;
  j2ep_mat_t *u = (j2ep_mat_t *)m->data;
  matstate_isotr_hard_t *matst = (matstate_isotr_hard_t *)ms;
  double twomu = 2 * u->G;
  double dev_sig_corot[3][3], dev_sig_predictor[3][3], t[3][3];
  double effective_stress, tr_ln_rel_U_3, beta, Gamma;
  double msD[3][3], F[3][3];
  double dt = ip_ctx->time_step;
  double cauchy_prev[6];
#define SQRT_2_OVER_3 0.81649658092772603272
  W_bool_t result = TRUE;
  matstate_isotr_hard_t input_matst;

  cauchy_prev[0] = matst->cauchy[0];
  cauchy_prev[1] = matst->cauchy[1];
  cauchy_prev[2] = matst->cauchy[2];
  cauchy_prev[3] = matst->cauchy[3];
  cauchy_prev[4] = matst->cauchy[4];
  cauchy_prev[5] = matst->cauchy[5];
  
  if (!ip_ctx->get_kinem_var) return FALSE;  

  /* kinematic quantities */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_REL_DEF_GRAD, rel_F))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_LN_POLAR_REL_U, ln_rel_U))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_POLAR_REL_R, rel_R))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;

  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  rel_J = MAT3DUTIL_DET_3X3(rel_F);
  LN_FUNC(ln_rel_J, /* = log */ rel_J);

  tr_ln_rel_U_3 = (ln_rel_U[0][0] + ln_rel_U[1][1] + ln_rel_U[2][2]) / 3;
  ln_rel_U[0][0] -= tr_ln_rel_U_3;
  ln_rel_U[1][1] -= tr_ln_rel_U_3;
  ln_rel_U[2][2] -= tr_ln_rel_U_3;
  
  /* elastic predictor: corotational */
  curr_press      = (matst->cauchy[0] + matst->cauchy[3] + matst->cauchy[5])/3;
  press_predictor =  curr_press + u->bulk_mod * ln_rel_J;
  dev_sig_corot[0][0]
    = matst->cauchy[0] - curr_press + twomu * ln_rel_U[0][0];
  dev_sig_corot[0][1] = dev_sig_corot[1][0]
    = matst->cauchy[1] + twomu * ln_rel_U[0][1];
  dev_sig_corot[0][2] = dev_sig_corot[2][0]
    = matst->cauchy[2] + twomu * ln_rel_U[0][2];
  dev_sig_corot[1][1]
    = matst->cauchy[3] - curr_press  + twomu * ln_rel_U[1][1];
  dev_sig_corot[1][2] = dev_sig_corot[2][1]
    = matst->cauchy[4] + twomu * ln_rel_U[1][2];
  dev_sig_corot[2][2]
    = matst->cauchy[5] - curr_press  + twomu * ln_rel_U[2][2];

  /* elastic predictor: after rotation */
  MAT3DUTIL_MULT_ABT_3X3(t, dev_sig_corot, rel_R);
  MAT3DUTIL_MULT_AB_3X3(dev_sig_predictor, rel_R, t);
  effective_stress
    = MAT3DUTIL_DBL_PROD_3X3(dev_sig_predictor, dev_sig_predictor);
  effective_stress = sqrt(effective_stress);

  /* plasticity criterion */
  if (effective_stress > SQRT_2_OVER_3 * matst->curr_sigma_y) {
    /* linear hardening */
    Gamma = ((effective_stress - SQRT_2_OVER_3 * matst->curr_sigma_y)
             / (twomu * (1 + u->hardening_param / 3 / u->G)));
    beta = 1 - twomu * Gamma / effective_stress;
    matst->equiv_pl_def += SQRT_2_OVER_3 * Gamma;
    matst->curr_sigma_y
      += SQRT_2_OVER_3 * u->delta * u->hardening_param * Gamma;
    matst->cauchy[0] = press_predictor + beta * dev_sig_predictor[0][0];
    matst->cauchy[1] = beta * dev_sig_predictor[0][1];
    matst->cauchy[2] = beta * dev_sig_predictor[0][2];
    matst->cauchy[3] = press_predictor + beta * dev_sig_predictor[1][1];
    matst->cauchy[4] = beta * dev_sig_predictor[1][2];
    matst->cauchy[5] = press_predictor + beta * dev_sig_predictor[2][2];
  } else {
    /* elastic state */
    matst->cauchy[0] = press_predictor + dev_sig_predictor[0][0];
    matst->cauchy[1] = dev_sig_predictor[0][1];
    matst->cauchy[2] = dev_sig_predictor[0][2];
    matst->cauchy[3] = press_predictor + dev_sig_predictor[1][1];
    matst->cauchy[4] = dev_sig_predictor[1][2];
    matst->cauchy[5] = press_predictor + dev_sig_predictor[2][2];
  }

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_MIDSTEP_RATE_OF_DEF, msD)) {
    result = FALSE; goto exit;
  }
  matst->deform_energy_density +=
    dt * (0.5 * ((cauchy_prev[0] + matst->cauchy[0]) * msD[0][0] +
                 (cauchy_prev[3] + matst->cauchy[3]) * msD[1][1] +
                 (cauchy_prev[5] + matst->cauchy[5]) * msD[2][2]) +
          ((cauchy_prev[1] + matst->cauchy[1]) * msD[0][1] +
           (cauchy_prev[2] + matst->cauchy[2]) * msD[0][2] +
           (cauchy_prev[4] + matst->cauchy[4]) * msD[1][2])
          ) * MAT3DUTIL_DET_3X3(F);

  if (stress_type == MAT_3D_1ST_PK) {
    double F[3][3], F_cof[3][3], sigma[3][3];
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      result = FALSE; goto exit;
    }
    sigma[0][0]               = matst->cauchy[0];
    sigma[0][1] = sigma[1][0] = matst->cauchy[1];
    sigma[0][2] = sigma[2][0] = matst->cauchy[2];
    sigma[1][1]               = matst->cauchy[3];
    sigma[1][2] = sigma[2][1] = matst->cauchy[4];
    sigma[2][2]               = matst->cauchy[5];
    MAT3DUTIL_COFACTOR_3X3(F, F_cof);
    MAT3DUTIL_MULT_ABT_3X3(stress, sigma, F_cof);
    *out_stress_type = MAT_3D_1ST_PK;
  } else 
    result = get_stress_isotr_hard (m, ms, ip_ctx,
                                    stress_type, stress, out_stress_type);
  
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}














static W_mat_3d_state_p 
interpolate_mat_state_elastic(W_mat_3d_t *m,
			      W_loc_t *at_point,
			      int npkgs,
			      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_elastic_t *ms;
  int i, k;
  double weight_sum = 0;
  matstate_elastic_t *pms;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_elastic_t *) get_mat_state_elastic(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_elastic_t *)pkgs[k].ms;
    FOR(i, 6) {
      ms->cauchy[i] += pms->cauchy[i] * WGHT(k);
    }
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p) ms;
}


static W_mat_3d_state_p 
interpolate_mat_state(W_mat_3d_t *m,
		      W_loc_t *at_point,
		      int npkgs,
		      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_mixed_hard_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_mixed_hard_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_mixed_hard_t *)pkgs[k].ms;
    ms->equiv_pl_def += pms->equiv_pl_def * WGHT(k);
    FOR(i, 6) {
      ms->cauchy[i]      += pms->cauchy[i] * WGHT(k);
      ms->back_stress[i] += pms->back_stress[i] * WGHT(k);
    }
    ms->curr_sigma_y += pms->curr_sigma_y * WGHT(k);
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}



static W_mat_3d_state_p 
interpolate_mat_state_isotr_hard(W_mat_3d_t *m,
				 W_loc_t *at_point,
				 int npkgs,
				 WM3D_mat_state_pkg_t *pkgs)
{
  matstate_isotr_hard_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_isotr_hard_t *) get_mat_state_isotr_hard(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_isotr_hard_t *)pkgs[k].ms;
    ms->equiv_pl_def += pms->equiv_pl_def * WGHT(k);
    FOR(i, 6) {
      ms->cauchy[i] += pms->cauchy[i] * WGHT(k);
    }
    ms->curr_sigma_y += pms->curr_sigma_y * WGHT(k);
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k); 
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}



static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_RATE_INST_FINAL_ROT;
}


static W_bool_t
get_mat_stiff_elastic (W_mat_3d_t *m/* mat3d */,
                       W_mat_3d_state_p ms,/* material state */
                       W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                                    used to generate kinematic
                                                    quantities which the material state
                                                    might need in order to compute
                                                    the requested stress */
                       W_mat_3d_stiff_enum stiff_type, /* stiffness type */
                       double C_mat[6][6] /* material stiffness tensor in matrix form */
                       )
{
  j2ep_mat_t *u = (j2ep_mat_t *) m->data;
  double E = u->E, nu = u->nu, mu = u->G, lambda;
  lambda = E * nu / (1 + nu) / (1 - 2*(nu));
#undef C
#define C(I,J,V) { C_mat[I][J] = V; }
    C(0,0,lambda + 2*mu); C(0,1,lambda);        C(0,2,lambda);        C(0,3,0);  C(0,4,0);  C(0,5,0);
    C(1,0,lambda);        C(1,1,lambda + 2*mu); C(1,2,lambda);        C(1,3,0);  C(1,4,0);  C(1,5,0);
    C(2,0,lambda);        C(2,1,lambda);        C(2,2,lambda + 2*mu); C(2,3,0);  C(2,4,0);  C(2,5,0);
    C(3,0,0);             C(3,1,0);             C(3,2,0);             C(3,3,mu); C(3,4,0);  C(3,5,0); 
    C(4,0,0);             C(4,1,0);             C(4,2,0);             C(4,3,0);  C(4,4,mu); C(4,5,0); 
    C(5,0,0);             C(5,1,0);             C(5,2,0);             C(5,3,0);  C(5,4,0);  C(5,5,mu); 
#undef C

  /* I'm handling request for Lagrangian stiffness only;
     if Eulerian, should transform first.
  */
  if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) return TRUE;
  else                                       return FALSE;
}

typedef struct {
  double F[3][3], R[3][3], U[3][3];
  double midstep_D[3][3];
  double incr_E[3][3], incr_U[3][3], rel_F[3][3], rel_R[3][3], ln_rel_U[3][3];
} custom_data_t;

static W_bool_t 
custom_get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
               W_mat_3d_kinem_var_type_enum  type,
               double kinvar[3][3])
{
  custom_data_t *custom_data = (custom_data_t *) ctx->ctx_data;
  switch (type) {
  case MAT_3D_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, custom_data->F);
    break;
  case MAT_3D_REL_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, custom_data->incr_U);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    MAT3DUTIL_COPY_3X3(kinvar, custom_data->midstep_D);
    break;
  case MAT_3D_POLAR_REL_R:
    MAT3DUTIL_COPY_3X3(kinvar, custom_data->rel_R);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    MAT3DUTIL_COPY_3X3(kinvar, custom_data->ln_rel_U);
    break;
  default:
    REPORT_FATAL_ERROR(NULL, "Unknown kinematic tensor type");
    break;
  }
  return TRUE;
}

/*
  This routine computes the material tangent by numerical differention of the stress
  tensor.
 */
static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
               )
{
  j2ep_mat_t *u = (j2ep_mat_t *) m->data;
  int i, j, k, l;
  double diag_u, offdiag_u, e = u->e_incr;
  custom_data_t custom_data;
  double Sp[6], Sn[6];
  W_mat_3d_stress_type_enum   out_stress_type;
  int result = TRUE;
  int v_to_t[6][2] = {{0,0},{1,1},{2,2},{0,1},{0,2},{1,2}};
  int mult[6] = {1,1,1,2,2,2};
  double equiv_pl_def;
  W_mat_3d_ip_ctx_t custom_ip_ctx;

  if (u->e_incr == 0) /* numerical differentiation not allowed */
    return get_mat_stiff_elastic (m, ms, ip_ctx, stiff_type, C_mat);
  if (WM3D_get_mat_state_var (m, ms, ip_ctx, MAT_3D_EQUIV_PLASTIC_DEF, &equiv_pl_def)) {
  /* if material is virgin (elastic), don't bother with the numerical approximation */
    if (equiv_pl_def == 0)
      return get_mat_stiff_elastic (m, ms, ip_ctx, stiff_type, C_mat);
  }
  
  /* zero out */
  FOR(i, 6) FOR (j, 6) C_mat[i][j] = 0;
  FOR(i, 3) FOR (j, 3) custom_data.incr_E[i][j] = 0;

  /* here we access the current kinematic variables through the element defined
     kinematics callback
   */
  if (!ip_ctx->get_kinem_var) return FALSE;  
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, custom_data.F)) {
    result = FALSE; goto exit;
  }
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_POLAR_R, custom_data.R))  {
    result = FALSE; goto exit;
  } 
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_POLAR_U, custom_data.U))  {
    result = FALSE; goto exit;
  } 

  /* now replace the kinematic variable callback by our custom callback */
  custom_ip_ctx               = *ip_ctx;
  custom_ip_ctx.get_kinem_var = custom_get_kinem_var;
  custom_ip_ctx.ctx_data      = &custom_data;

#define GET_2ND_PK(S)                                                                   \
  {                                                                                     \
    double St[3][3];                                                                    \
    if (!WM3D_update_stress (m, ms, W_MAT_STATE_UPDATE_TEMP,                            \
                             &custom_ip_ctx, MAT_3D_2ND_PK, St, &out_stress_type)) {    \
      result = FALSE; goto exit;                                                        \
    }                                                                                   \
    if (out_stress_type != MAT_3D_2ND_PK) {result = FALSE; goto exit;}                  \
    S[0] = St[0][0];                                                                    \
    S[1] = St[1][1];                                                                    \
    S[2] = St[2][2];                                                                    \
    S[3] = St[0][1];                                                                    \
    S[4] = St[0][2];                                                                    \
    S[5] = St[1][2];                                                                    \
  }

  /*
   incr_U is primary here.  we compute its components so that incr_E is of the
   simple form custom_data.incr_E[I][J] = custom_data.incr_E[J][I] = INCR;
   ln_rel_U is approximated by the first term of the Taylor series.
   For real small a, we could replace the expressions for a by
   a = sqrt (2 * INCR + 1) - 1; ==>  a = INCR;
   a = 1./2*(sqrt(1+2*INCR)-sqrt(1-2*INCR));  ==>  a = INCR;
   */
#define COMP_KINEM_VARS(I,J,E,DIAG_U,OFFDIAG_U)                                 \
  {                                                                             \
    MAT3DUTIL_UNIT_3X3 (custom_data.incr_U);                                    \
    if (I == J) {                                                               \
      custom_data.incr_U[I][J] += DIAG_U;                                       \
    } else {                                                                    \
      custom_data.incr_U[I][J] = custom_data.incr_U[J][I] = OFFDIAG_U;          \
    }                                                                           \
    custom_data.incr_E[I][J] = custom_data.incr_E[J][I] = E;                    \
    MAT3DUTIL_COPY_3X3 (custom_data.rel_F, custom_data.incr_U);                 \
    MAT3DUTIL_COPY_3X3 (custom_data.ln_rel_U, custom_data.incr_U);              \
    custom_data.ln_rel_U[0][0] -= 1;                                            \
    custom_data.ln_rel_U[1][1] -= 1;                                            \
    custom_data.ln_rel_U[2][2] -= 1;                                            \
  }

  MAT3DUTIL_UNIT_3X3 (custom_data.rel_R);
  MAT3DUTIL_ZERO_3X3 (custom_data.midstep_D); /* we don't care about this tensor */

  if (fabs (e) > 0.001) diag_u = sqrt (2 * e + 1) - 1;                  
  else                  diag_u = e;                                     

  if (fabs (e) > 0.001) offdiag_u = 1./2*(sqrt(1+2*e)-sqrt(1-2*e));     
  else                  offdiag_u = e;                                     

  FOR (k, 6) {
    i = v_to_t[k][0];
    j = v_to_t[k][1];
    COMP_KINEM_VARS(i,j, (+e), (+diag_u), (+offdiag_u));
    GET_2ND_PK (Sp);
    COMP_KINEM_VARS(i,j, (-e), (-diag_u), (-offdiag_u));
    GET_2ND_PK (Sn);
    FOR (l, 6) {
      C_mat[k][l] += (Sp[l] - Sn[l]) / (2 * e) / mult[l];
    }
    custom_data.incr_E[i][j] = custom_data.incr_E[j][i] = 0;
  }

  /* symmetrize */
  FOR (k, 6) {
    FOR (l, 6) {
      if (k != l) {
        double a = (C_mat[k][l] + C_mat[l][k]) / 2;
        C_mat[k][l] = C_mat[l][k] = a;
      }
    }
  }

 exit:

  if (result) {
    /* I'm handling request for Lagrangian stiffness only;
     if Eulerian, should transform first. */
    if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) return TRUE;
    else                                       return FALSE;
  } else
    return FALSE; /* have nothing to show for my effort :( */
}


static W_mat_3d_methods_t mat3difr_elastic_methods = {
  get_mcode,
  get_mat_state_elastic,
  get_mat_const,
  update_stress_elastic,
  get_mat_state_var_elastic,
  interpolate_mat_state_elastic,
  get_mat_stiff_elastic,
  NULL,
  NULL,
};

static W_mat_3d_methods_t mat3difr_ep_methods = {
  get_mcode,
  get_mat_state,
  get_mat_const,
  update_stress_ep,
  get_mat_state_var_ep,
  interpolate_mat_state,
  get_mat_stiff,
  NULL,
  NULL
};
  
static W_mat_3d_methods_t mat3difr_ep_isotr_hard_methods = {
  get_mcode,
  get_mat_state_isotr_hard,
  get_mat_const, 
  update_stress_ep_isotr_hard,
  get_mat_state_var_isotr_hard,
  interpolate_mat_state_isotr_hard,
  get_mat_stiff,
  NULL,
  NULL
};

static W_mat_3d_t *
new_mat_3d_ifr (double E, double nu,
		 double tensile_strength, double dyn_fracture_toughness,
		 double sigma_y, double hardening_param,
                 double delta, double rho, double e_incr)
{
  j2ep_mat_t *u;
  
  u = w_new(j2ep_mat_t);
  CHECK_MALLOC(u);
  u->E                      = E; 
  u->nu                     = nu;
  u->tensile_strength       = tensile_strength;
  u->dyn_fracture_toughness = dyn_fracture_toughness;
  if (u->nu >= 0.5) REPORT_FATAL_ERROR(NULL, "Bad nu (should be 0 <= nu < 0.5)");
  u->bulk_mod               = u->E / 3 / (1 - 2*u->nu);
  u->G                      = u->E / 2 / (1 + u->nu);
  u->sigma_y                = sigma_y;
  if (u->sigma_y == 0) u->is_elastoplastic = FALSE;
  else                 u->is_elastoplastic = TRUE;
  u->hardening_param        = hardening_param;
  u->delta                  = delta;
  u->dil_wave_speed_3d
    = W_3D_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstrain
    = W_plstrain_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstress
    = W_plstress_dilatational_wave_speed (u->E, u->nu, rho);
  u->shear_wave_speed
    = W_shear_wave_speed (u->E, u->nu, rho);
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstress,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstress_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane stress speed does not exist");
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstrain,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstrain_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane strain speed does not exist");
  u->e_incr = e_incr;
  
  return WM3D_new_material(
                           (u->is_elastoplastic
                            ? (u->delta == 1
                               ? &mat3difr_ep_isotr_hard_methods
                               : &mat3difr_ep_methods)
                            : &mat3difr_elastic_methods),
                           (void *)u, rho);
}


W_mat_3d_t *
read_mat_3d_ifr(W_domain_t *d)
{
  const char *input_line;
  double E, nu, rho = 1;
  double sigma_y = 0, hardening_param = 0, delta = 1;
  double tensile_strength = 0;
  double dyn_fracture_toughness = 0;
  double e_incr = 0;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { E = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { nu = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { tensile_strength = TOKEN_DBL_NEXT(1); },
			    { tensile_strength = E / 100; }
			    );
      TOKENS_SET_BY_KEYWORD("DYN_FRACTURE_TOUGHNESS", FALSE,
			    { dyn_fracture_toughness = TOKEN_DBL_NEXT(1); },
			    { dyn_fracture_toughness = E / 4000; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("SIGMA_Y", FALSE,
			    {
			      sigma_y = TOKEN_DBL_NEXT(1);
			    },
			    {
                              sigma_y = 0;
                            }
			    );
      TOKENS_SET_BY_KEYWORD("DELTA", FALSE,
			    { delta = TOKEN_DBL_NEXT(1); },
			    { delta = 1; } /* isotropic hardening  */
			    );
      TOKENS_SET_BY_KEYWORD("HARDENING_PARAM", FALSE,
			    { hardening_param = TOKEN_DBL_NEXT(1); },
			    { hardening_param = 0; }/* perfectly ep material */
			    );
      TOKENS_SET_BY_KEYWORD("E_INCR", FALSE,
			    { e_incr = TOKEN_DBL_NEXT(1); },
			    { e_incr = 0; }/* numerical approx of tangent */
			    );
      return new_mat_3d_ifr (E, nu, tensile_strength, dyn_fracture_toughness,
			      sigma_y, hardening_param, delta, rho, e_incr);
    }
  }
  return NULL;
}


