#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Wtfuncs.h"
#include "Wmat3d.h"
#include "Wmat3dis.h"
#include "Wmat3dunrot.h"
#include "Wmat3dssj2ep.h"

/* RCSHEADER "$Id: mat3dis.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

typedef struct mat_t {
  W_time_func_t *tf;
  char          *imposed_isotropic_strain_expr;
  W_mat_3d_t    *ssmat; /* delegate material update to this material */
}              mat_t; 

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_IS;
}

static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{ 
  mat_t *u = (mat_t *)m->data;

  return WM3D_get_mat_const (u->ssmat, type, c);
}

typedef struct matstate_t {
  double         imposed_isotropic_strain_magnitude;
  double         last_tfmult;
  W_mat_3d_state_p ssms;
}              matstate_t;

static W_mat_3d_state_p 
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;
  int i;

  ms = w_new (matstate_t); CHECK_MALLOC(ms);

  ms->ssms = WM3D_get_mat_state (u->ssmat, at_point);
  if (u->imposed_isotropic_strain_expr) {
    int stat;                                                   
    double v;
    
    expeval_set_value("x", at_point->x);
    expeval_set_value("y", at_point->y);
    expeval_set_value("z", at_point->z);
    stat = expeval_eval(u->imposed_isotropic_strain_expr, &v);   
    if (stat != EXPEVAL_E_OK) {                                 
      sprintf(W_buf(), "%s", expeval_err_msgs[stat]);               
      REPORT_ERROR(NULL, W_buf());                                        
    }
    ms->imposed_isotropic_strain_magnitude = v;
    ms->last_tfmult = 0;
  }
  return (W_mat_3d_state_p)ms;
}

static W_mat_3d_state_p 
interpolate_mat_state (W_mat_3d_t *m,
                       W_loc_t *at_point,
                       int npkgs,
                       WM3D_mat_state_pkg_t *pkgs)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms = (matstate_t *) get_mat_state (m, at_point);
  
  return WM3D_interpolate_mat_state (u->ssmat, at_point, npkgs, pkgs);
}

typedef struct ip_ctx_data_t {
  W_mat_3d_ip_ctx_t *up_ip_ctx;
  double             d[3][3];
}              ip_ctx_data_t;

static W_bool_t 
is_get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
                    W_mat_3d_kinem_var_type_enum  type,
                    double kinvar[3][3])
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;
  switch (type) {
  case MAT_3D_INCR_DISPL_GRAD:
    MAT3DUTIL_COPY_3X3 (kinvar, ipctx->d);
    return TRUE;
  default:
    return FALSE;
  }
}

static W_bool_t 
dummy_get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
                     W_mat_3d_kinem_var_type_enum  type,
                     double kinvar[3][3])
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;
  return FALSE;
}

static W_bool_t 
get_mat_state_var (W_mat_3d_t *m,
                   W_mat_3d_state_p ms,
                   W_mat_3d_ip_ctx_t *ip_ctx,
                   W_mat_3d_mat_state_var_enum  var_type,
                   double *var
                   )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  W_mat_3d_ip_ctx_t is_ip_ctx;
  ip_ctx_data_t ipdata;

  is_ip_ctx               = *ip_ctx;
  is_ip_ctx.ctx_data      = &ipdata;
  is_ip_ctx.get_kinem_var = dummy_get_kinem_var;
  ipdata.up_ip_ctx = ip_ctx;

  return WM3D_get_mat_state_var (u->ssmat, matst->ssms, &is_ip_ctx, var_type, var); 
}

/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress (W_mat_3d_t *m,
            W_mat_3d_state_p ms,
            W_mat_3d_ip_ctx_t *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double sigma[3][3], F[3][3], R[3][3], U[3][3], t[3][3];
  W_mat_3d_stress_type_enum out_stress_type_internal;
  W_mat_3d_ip_ctx_t is_ip_ctx;
  ip_ctx_data_t ipdata;

  is_ip_ctx               = *ip_ctx;
  is_ip_ctx.ctx_data      = &ipdata;
  is_ip_ctx.get_kinem_var = dummy_get_kinem_var;
  ipdata.up_ip_ctx = ip_ctx;
  
  if (!WM3D_update_stress (u->ssmat, matst->ssms, W_MAT_STATE_UPDATE_TEMP,
                           &is_ip_ctx, MAT_3D_CAUCHY,
                           t, &out_stress_type_internal))
    return FALSE;
  if (out_stress_type_internal != MAT_3D_CAUCHY)
    return FALSE;
    
  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
  if (!W_polar_decomposition (F, R, U)) {
    return FALSE;
  }
  
  MAT3DUTIL_MULT_ABAT_3X3 (sigma, R, t);
  
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    MAT3DUTIL_COPY_3X3 (stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    {
      double P[3][3], J, inv_F[3][3];
      J = MAT3DUTIL_DET_3X3(F);
      if (!W_mat3dutil_invert_3x3(F, inv_F))
        REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
      MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
      MAT3DUTIL_SCALE_3X3(P, J);
      MAT3DUTIL_COPY_3X3(stress, P);
    }
    break;
  case MAT_3D_2ND_PK:
    {
      double P[3][3], S[3][3], J, inv_F[3][3];
      J = MAT3DUTIL_DET_3X3(F);
      if (!W_mat3dutil_invert_3x3(F, inv_F))
         REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
      MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
      MAT3DUTIL_MULT_AB_3X3(S, inv_F, P);
      MAT3DUTIL_SCALE_3X3(S, J);
      MAT3DUTIL_COPY_3X3(stress, S);
    }
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}

static W_bool_t  
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double t[3][3];
  double DU[3][3], R[3][3];
  double dt = ip_ctx->time_step;
  W_bool_t result = TRUE;
  W_mat_3d_ip_ctx_t is_ip_ctx;
  ip_ctx_data_t ipdata;
  W_mat_3d_stress_type_enum out_stress_type_internal;
  matstate_t input_matst;

  /* kinematic quantities */
  if (!ip_ctx->get_kinem_var) return FALSE;  

  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_INCR_DISPL_GRAD, DU)) {
    double F[3][3], prev_F[3][3];
    if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    } 
    if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_PREV_DEF_GRAD, prev_F)) {
      return FALSE;
    }
    MAT3DUTIL_3_MAT (DU, =, F, -, prev_F);
  }

  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_POLAR_R, R))
    return FALSE;

  /* save current material state */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  is_ip_ctx               = *ip_ctx;
  is_ip_ctx.ctx_data      = &ipdata;
  is_ip_ctx.get_kinem_var = is_get_kinem_var;
  ipdata.up_ip_ctx = ip_ctx;
 
  MAT3DUTIL_COMB_ATB_3X3 (ipdata.d, 0.5, DU, 0.5, DU);

  /* Add imposed-strain effect */
  if (matst->imposed_isotropic_strain_magnitude != 0) {
    double tfmult = WTF_func_val(u->tf, ip_ctx->time);
    double t = tfmult - matst->last_tfmult;
    double s = matst->imposed_isotropic_strain_magnitude * t;
    ipdata.d[0][0] -= s;
    ipdata.d[1][1] -= s;
    ipdata.d[2][2] -= s;
    matst->last_tfmult = tfmult;
  }

  if (!WM3D_update_stress (u->ssmat, matst->ssms, update_form, &is_ip_ctx,
                           MAT_3D_CAUCHY, t, &out_stress_type_internal))
    return FALSE;
  if (out_stress_type_internal != MAT_3D_CAUCHY) return FALSE;

  if (stress_type == MAT_3D_CAUCHY) { /* accelerated handling for what we've got ready */
  /* rotate to obtain the Cauchy stress */
    MAT3DUTIL_MULT_ABAT_3X3 (stress, R, t);
    *out_stress_type = MAT_3D_CAUCHY;
    result = TRUE;
  } else { /* generic case */
    result = get_stress (m, ms, ip_ctx, stress_type, stress, out_stress_type);
  }
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

#include "consttransf.c"

static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
               )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  W_mat_3d_ip_ctx_t is_ip_ctx;
  ip_ctx_data_t ipdata;

  is_ip_ctx               = *ip_ctx;
  is_ip_ctx.ctx_data      = &ipdata;
  is_ip_ctx.get_kinem_var = dummy_get_kinem_var;
  ipdata.up_ip_ctx = ip_ctx;

  if (!WM3D_get_mat_stiff (u->ssmat, matst->ssms, &is_ip_ctx,
                           stiff_type, C_mat))
    return FALSE;
  if (stiff_type == MAT_3D_STIFF_EULERIAN) return TRUE;
  else                                     return FALSE;
}

static W_mat_3d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const,
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  NULL,
  NULL
};

W_mat_3d_t *
new_mat_3d_is (double E, double nu,
		 double tensile_strength, double dyn_fracture_toughness,
                double sigma_y, double hardening_param,
                char *imposed_isotropic_strain_expr,
               W_time_func_t *tf, 
                 double delta, double rho, W_bool_t rot_C)
{
  mat_t *u;
  
  u = w_new(mat_t);
  CHECK_MALLOC(u);
  u->ssmat = new_mat_3d_ssj2ep (E, nu,
                                tensile_strength, dyn_fracture_toughness,
                                sigma_y, hardening_param,
                                delta, rho);
  u->tf = tf;
  if (imposed_isotropic_strain_expr != 0) {
    DUPLICATE_STRING (u->imposed_isotropic_strain_expr, imposed_isotropic_strain_expr);
  } else {
    u->imposed_isotropic_strain_expr = 0;
  }
  return WM3D_new_material (&methods, (void *)u, rho);
}


W_mat_3d_t *
read_mat_3d_is (W_domain_t *d)
{
  const char *input_line;
  double E, nu, rho = 1;
  double sigma_y = 0, hardening_param = 0, delta = 1;
  double tensile_strength = 0;
  double dyn_fracture_toughness = 0;
  W_bool_t rot_C = FALSE;
  W_bool_t imposed_set = FALSE;
  char  imposed_isotropic_strain_expr[1024];
  W_index_t time_func = 1;
  W_time_func_t *tf = NULL;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { E = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { nu = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { tensile_strength = TOKEN_DBL_NEXT(1); },
			    { tensile_strength = E / 100; }
			    );
      TOKENS_SET_BY_KEYWORD("DYN_FRACTURE_TOUGHNESS", FALSE,
			    { dyn_fracture_toughness = TOKEN_DBL_NEXT(1); },
			    { dyn_fracture_toughness = E / 4000; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("SIGMA_Y", FALSE,
			    {
			      sigma_y = TOKEN_DBL_NEXT(1);
			    },
			    {
                              sigma_y = 0;
                            }
			    );
      TOKENS_SET_BY_KEYWORD("DELTA", FALSE,
			    { delta = TOKEN_DBL_NEXT(1); },
			    { delta = 1; } /* isotropic hardening  */
			    );
      TOKENS_SET_BY_KEYWORD("HARDENING_PARAM", FALSE,
			    { hardening_param = TOKEN_DBL_NEXT(1); },
			    { hardening_param = 0; }/* perfectly ep material */
			    );
      TOKENS_SET_BY_KEYWORD("IMPOSED_ISOTROPIC_STRAIN", FALSE,
                            { sprintf (imposed_isotropic_strain_expr, "%s",
                                       TOKEN_STR_NEXT(1));
                              imposed_set = TRUE; },
                            { imposed_set = FALSE; } /* no strain */
			    );
      if (imposed_set) {
        TOKENS_SET_BY_KEYWORD("TIME_FUNC", FALSE,
                              { time_func = TOKEN_INT_NEXT(1); },
                              { time_func = 1; } 
                              );
        tf = WD_get_time_func_by_idx (d, time_func);
        if (tf == NULL) {
          sprintf (W_buf (), "Time function #%d not found", time_func);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
      }
      if (TOKENS_HAVE_KEYWORD ("ROT_C")) rot_C = TRUE;
       return new_mat_3d_is (E, nu, tensile_strength, dyn_fracture_toughness,
                               sigma_y, hardening_param,
                             (imposed_set
                              ? imposed_isotropic_strain_expr
                              : 0),
                              tf, delta, rho, rot_C);
    }
  }
  return NULL;
}


