#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dmodstvk.h"

/* RCSHEADER "$Id: mat3dmodstvk.c,v 1.1 2004/09/06 01:17:58 cvs Exp $" */

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_MOD_STV_K;
}


typedef struct mat_t {
  double E, nu;
}              mat_t;


typedef struct matstate_t {
  double   deform_energy_density;
}              matstate_t;


  
#if 0
#define DUMP(S) {                                               \
  printf("%g\t %g\t %g\n", S[0][0], S[0][1], S[0][2]);          \
  printf("%g\t %g\t %g\n", S[1][0], S[1][1], S[1][2]);          \
  printf("%g\t %g\t %g\n", S[2][0], S[2][1], S[2][2]); }

#define HAVENAN(S) (   isnan(S[0][0]) || isnan(S[0][1]) || isnan(S[0][2])  \
                    || isnan(S[1][0]) || isnan(S[1][1]) || isnan(S[1][2])  \
                    || isnan(S[2][0]) || isnan(S[2][1]) || isnan(S[2][2]))
#endif

/* Function updates the stress state, and returns the stress (with proper */
/* conversion into the requested type).  Note that the function get_stress() */
/* is identical with this function since no stress state is saved for this */
/* type of elastic material. */
static W_bool_t 
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  double S[3][3], E[3][3], F[3][3], C[3][3], invC[3][3];
  mat_t *u = (mat_t *)m->data;
  double twomu = u->E /  (1 + u->nu);
  double B = u->E / 3 / (1 - 2*u->nu);
  double J, pressure;
  W_bool_t result = TRUE;
  matstate_t *matst = (matstate_t *)ms;
  matstate_t input_matst;

  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_CAUCHY_GREEN, C))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_GREEN_LAGRANGE, E))
    return FALSE;

  /* save the material state if the update is temporary */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  J = sqrt(MAT3DUTIL_DET_3X3(C));
  
  MAT3DUTIL_2_MAT(S, =, twomu * E);

  if (!W_mat3dutil_invert_3x3 (C, invC)) { 
    result = FALSE; goto exit;
  }
  pressure =  (- log(J)) * B;
  MAT3DUTIL_2_MAT(S, -=, pressure * invC);
  
  matst->deform_energy_density = 0.5 * (MAT3DUTIL_DBL_PROD_3X3(S, E));

#if 0
  {
    double Psi = B / 2 * log(J) * log(J) + twomu / 2 * MAT3DUTIL_DBL_PROD_3X3(E, E);
    if (fabs(Psi - matst->deform_energy_density) > 1e-9)
      fprintf(stderr, "Psi=%g, Psi-S:E/2=%g\n", Psi, Psi - matst->deform_energy_density);
  }
#endif
  
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    {
      double P[3][3], sigma[3][3], invdetF;
      int i, j, k;
      
      if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
        result = FALSE; goto exit;
      }
      FOR(i, 3) {
        FOR(j, 3) {
          P[i][j] = 0; /* /F/ * /S/ */ 
          FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
        }
      }
      invdetF = 1 / MAT3DUTIL_DET_3X3(F);
      FOR(i, 3) {
        FOR(j, 3) {
          sigma[i][j] = 0; /* 1/detF /P/ * /F/^T */ 
          FOR(k, 3) sigma[i][j] += invdetF * P[i][k] * F[j][k];
        }
      }
      MAT3DUTIL_COPY_3X3(stress, sigma);
    }
    break;
  case MAT_3D_1ST_PK:
    {
      double P[3][3];
      int i, j, k;
      if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
        result = FALSE; goto exit;
      }
      FOR(i, 3) {
        FOR(j, 3) {
          P[i][j] = 0; /* /F/ * /S/ */ 
          FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
        }
      }
      MAT3DUTIL_COPY_3X3(stress, P);
    }
    break;
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, S);
    *out_stress_type = MAT_3D_2ND_PK;
    break;
  }
    
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->E / 2 / (1 + u->nu);
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->E / 3 / (1 - 2*u->nu);
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

static W_bool_t 
get_mat_state_named_var (W_mat_3d_t *m,
                         W_mat_3d_state_p ms,
                         W_mat_3d_ip_ctx_t *ip_ctx,
                         char *name,
                         double *var
                         )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

  if        (tokens_token_and_keyword_equiv (name, "ln_U_1") ||
             tokens_token_and_keyword_equiv (name, "ln_U_2") ||
             tokens_token_and_keyword_equiv (name, "ln_U_3") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U") 
             ) {
    double F[3][3], R[3][3], U[3][3], ln_U[3][3];
    double evals[3], evecs[3][3], lnevals[3];
    *var = 0;
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    if (W_polar_decomposition (F, R, U)) {
      int imin, imid, imax;
      double logevals[3];
      W_eigendecomposition (U, evecs, evals);
      logevals[0] = log(evals[0]);
      logevals[1] = log(evals[1]);
      logevals[2] = log(evals[2]);
      if (logevals[0] > logevals[1]) {
        if (logevals[1] > logevals[2]) {
          imax = 0; imid = 1; imin = 2; 
        } else {
          if (logevals[0] > logevals[2]) {
            imax = 0; imid = 2; imin = 1; 
          } else {
            imax = 2; imid = 0; imin = 1; 
          }          
        }
      } else {
        if (logevals[0] > logevals[2]) {
          imax = 1; imid = 0; imin = 2; 
        } else {
          if (logevals[1] > logevals[2]) {
            imax = 1; imid = 2; imin = 0; 
          } else {
            imax = 2; imid = 1; imin = 0; 
          }          
        }
      }

      if (tokens_token_and_keyword_equiv (name, "ln_U_1")) *var = log(evals[imax]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_2")) *var = log(evals[imid]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_3")) *var = log(evals[imin]);
      if (tokens_token_and_keyword_equiv (name, "max_ln_U")) {
        *var = fabs(log(evals[0]));
        if (fabs(log(evals[1])) > *var) *var = fabs(log(evals[1]));
        if (fabs(log(evals[2])) > *var) *var = fabs(log(evals[2]));
      }
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_x")) *var = evecs[imin][0];
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_y")) *var = evecs[imin][1];
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_z")) *var = evecs[imin][2];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_x")) *var = evecs[imid][0];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_y")) *var = evecs[imid][1];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_z")) *var = evecs[imid][2];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_x")) *var = evecs[imax][0];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_y")) *var = evecs[imax][1];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_z")) *var = evecs[imax][2];
      goto exit;
    }
  } else if (tokens_token_and_keyword_equiv (name, "J")) {
    double F[3][3];
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    *var = MAT3DUTIL_DET_3X3(F);
  } else {
    *var = 0;
    return FALSE;
  }
 exit:
  return TRUE;
}

static W_bool_t 
get_stress (W_mat_3d_t                  *m,
            W_mat_3d_state_p             ms,
            W_mat_3d_ip_ctx_t           *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double                       stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  return update_stress (m, ms, W_MAT_STATE_UPDATE_TEMP, ip_ctx,
                        stress_type, stress, out_stress_type);
}

static W_mat_3d_state_p 
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  matstate_t *ms;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_mat_state_var(W_mat_3d_t           *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0; /* elastic material */
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}




static W_mat_3d_state_p 
interpolate_mat_state(W_mat_3d_t *m,
		      W_loc_t *at_point,
		      int npkgs,
		      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}


static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* material stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
                   )
{
  mat_t *u = (mat_t *)m->data;
    /* RAW this is *not* correct: need to properly modify for the Fung pseudoelasticity */
    return FALSE;
}
	
static W_mat_3d_methods_t _methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  get_mat_state_named_var,
  NULL
};

static W_mat_3d_t *
new_mat  (double E, double nu, double rho)
{
  mat_t *u;
  
  u = w_new(mat_t); CHECK_MALLOC(u);

  u->E = E;
  u->nu = nu;
  return WM3D_new_material (&_methods, (void *)u, rho);
}

W_mat_3d_t *
read_mat_3d_mod_stvk (W_domain_t *d)
{
  const char *input_line;
  double E = 0, nu = 0, rho = 0;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      if  (nu >= 0.5 || nu < 0)
        REPORT_FATAL_ERROR  (d->errh, "Bad nu  (should be 0 <= nu < 0.5)");
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      return new_mat (E, nu, rho);
    }
  }
  return NULL;
}

  /*
  {
    double Psi = B / 2 * (J - 1) * J * 1./2 * (3 - MAT3DUTIL_TRACE_3X3(invC)) + twomu / 2 * MAT3DUTIL_DBL_PROD_3X3(E, E);
      fprintf(stderr, "Psi=%g, Psi-S:E/2=%g\n", Psi, Psi - matst->deform_energy_density);
  }
  */
