#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dneohk.h"

/* RCSHEADER "$Id: mat3dneohk.c,v 1.6 2004/09/06 01:17:58 cvs Exp $" */

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_NEOHOOKEAN;
}

typedef struct mat_t {
  double E, nu;
  double lambda, G, bulk_mod;
  double dil_wave_speed_3d;
  double dil_wave_speed_pstrain;
  double dil_wave_speed_pstress;
  double shear_wave_speed;
  double rayleigh_pstrain_wave_speed;
  double rayleigh_pstress_wave_speed;
}              mat_t;


typedef struct matstate_t {
  double   deform_energy_density;
}              matstate_t;


static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_TENSILE_STRENGTH:
    *c = 0;
    break;
  case MAT_3D_DYN_FRACTURE_TOUGHNESS:
    *c = 0;
    break;
  case MAT_3D_DIL_WAVE_SPEED_3D:
    *c = u->dil_wave_speed_3d;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRAIN:
    *c = u->dil_wave_speed_pstrain;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRESS:
    *c = u->dil_wave_speed_pstress;
    break;
  case MAT_3D_SHEAR_WAVE_SPEED:
    *c = u->shear_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED:
    *c = u->rayleigh_pstrain_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRESS_WAVE_SPEED:
    *c = u->rayleigh_pstress_wave_speed;
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

  
#if 0
#define DUMP(S) {                                               \
  printf("%g\t %g\t %g\n", S[0][0], S[0][1], S[0][2]);          \
  printf("%g\t %g\t %g\n", S[1][0], S[1][1], S[1][2]);          \
  printf("%g\t %g\t %g\n", S[2][0], S[2][1], S[2][2]); }

#define HAVENAN(S) (   isnan(S[0][0]) || isnan(S[0][1]) || isnan(S[0][2])  \
                    || isnan(S[1][0]) || isnan(S[1][1]) || isnan(S[1][2])  \
                    || isnan(S[2][0]) || isnan(S[2][1]) || isnan(S[2][2]))
#endif

/* Function updates the stress state, and returns the stress (with proper 
 conversion into the requested type).  Note that the function get_stress() 
 is identical with this function since no stress state is saved for this 
 type of elastic material. */
static W_bool_t 
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  double cauchy[3][3], b[3][3], F[3][3], C[3][3], invF[3][3];
  mat_t *u = (mat_t *)m->data;
  double lambda = u->lambda, mu = u->G, muJ, lJ, J, trC, logJ;
  W_bool_t result = TRUE;
  matstate_t *matst = (matstate_t *)ms;
  matstate_t input_matst;

  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;

  /* save the material state if the update is temporary */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;
  
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_CAUCHY_GREEN, C)) {
    result = FALSE; goto exit;
  }
  
  J = MAT3DUTIL_DET_3X3(F); 
  if (J <= 0) {
    static int reported = 0;
    if (!reported) {
      sprintf(W_buf(), "J <= 0: clamped. Watch the results closely!");
      REPORT_WARNING(0, W_buf());
      reported = 1;
    }
    J = 2e-16; /* clamp it */
  }

  /* compute the left Cauchy-Green tensor F.F^T */
  MAT3DUTIL_MULT_ABT_3X3 (b, F, F);

  muJ = mu / J;
  logJ = log (J);
  lJ  = lambda / J * logJ;
  cauchy[0][0] = muJ * (b[0][0] - 1) + lJ;
  cauchy[0][1] = muJ * b[0][1];
  cauchy[0][2] = muJ * b[0][2];
  cauchy[1][0] = muJ * b[1][0];
  cauchy[1][1] = muJ * (b[1][1] - 1) + lJ;
  cauchy[1][2] = muJ * b[1][2];
  cauchy[2][0] = muJ * b[2][0];
  cauchy[2][1] = muJ * b[2][1];
  cauchy[2][2] = muJ * (b[2][2] - 1) + lJ;

  if (!W_mat3dutil_invert_3x3 (F, invF)) {
    result = FALSE; goto exit;
  }

  trC = MAT3DUTIL_TRACE_3X3 (C);
  matst->deform_energy_density
    = (mu / 2 * (trC - 3) - mu * logJ + lambda / 2 * logJ * logJ);

  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_1ST_PK:
    MAT3DUTIL_MULT_ABT_3X3 (stress, cauchy, invF);
    MAT3DUTIL_SCALE_3X3 (stress, J);
    break;
  case MAT_3D_2ND_PK:
    {
      double S[3][3];
      MAT3DUTIL_MULT_ABT_3X3 (S, cauchy, invF);
      MAT3DUTIL_MULT_AB_3X3 (stress, invF, S);
      MAT3DUTIL_SCALE_3X3 (stress, J);
    }
    break;
  case MAT_3D_CAUCHY:
  default:
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

static W_bool_t 
get_stress (W_mat_3d_t                  *m,
            W_mat_3d_state_p             ms,
            W_mat_3d_ip_ctx_t           *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double                       stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  return update_stress (m, ms, W_MAT_STATE_UPDATE_TEMP, ip_ctx,
                        stress_type, stress, out_stress_type);
}

static W_mat_3d_state_p 
get_mat_state(W_mat_3d_t *m, W_loc_t *at_point)
{
  matstate_t *ms;
  int i;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_mat_state_var(W_mat_3d_t           *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0; /* elastic material */
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}

static W_bool_t 
get_mat_state_named_var (W_mat_3d_t *m,
                         W_mat_3d_state_p ms,
                         W_mat_3d_ip_ctx_t *ip_ctx,
                         char *name,
                         double *var
                         )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

  if        (tokens_token_and_keyword_equiv (name, "ln_U_1") ||
             tokens_token_and_keyword_equiv (name, "ln_U_2") ||
             tokens_token_and_keyword_equiv (name, "ln_U_3") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "min_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "mid_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_x") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_y") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U_z") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U") 
             ) {
    double F[3][3], R[3][3], U[3][3], ln_U[3][3];
    double evals[3], evecs[3][3], lnevals[3];
    *var = 0;
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    if (W_polar_decomposition (F, R, U)) {
      int imin, imid, imax;
      double logevals[3];
      W_eigendecomposition (U, evecs, evals);
      logevals[0] = log(evals[0]);
      logevals[1] = log(evals[1]);
      logevals[2] = log(evals[2]);
      if (fabs(logevals[0]) > fabs(logevals[1])) {
        if (fabs(logevals[1]) > fabs(logevals[2])) {
          imax = 0; imid = 1; imin = 2; 
        } else {
          if (fabs(logevals[0]) > fabs(logevals[2])) {
            imax = 0; imid = 2; imin = 1; 
          } else {
            imax = 2; imid = 0; imin = 1; 
          }          
        }
      } else {
        if (fabs(logevals[0]) > fabs(logevals[2])) {
          imax = 1; imid = 0; imin = 2; 
        } else {
          if (fabs(logevals[1]) > fabs(logevals[2])) {
            imax = 1; imid = 2; imin = 0; 
          } else {
            imax = 2; imid = 1; imin = 0; 
          }          
        }
      }

      if (tokens_token_and_keyword_equiv (name, "ln_U_1")) *var = log(evals[0]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_2")) *var = log(evals[1]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_3")) *var = log(evals[2]);
      if (tokens_token_and_keyword_equiv (name, "max_ln_U")) {
        *var = fabs(log(evals[0]));
        if (fabs(log(evals[1])) > *var) *var = fabs(log(evals[1]));
        if (fabs(log(evals[2])) > *var) *var = fabs(log(evals[2]));
      }
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_x")) *var = evecs[imin][0];
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_y")) *var = evecs[imin][1];
      if (tokens_token_and_keyword_equiv (name, "min_ln_U_z")) *var = evecs[imin][2];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_x")) *var = evecs[imid][0];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_y")) *var = evecs[imid][1];
      if (tokens_token_and_keyword_equiv (name, "mid_ln_U_z")) *var = evecs[imid][2];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_x")) *var = evecs[imax][0];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_y")) *var = evecs[imax][1];
      if (tokens_token_and_keyword_equiv (name, "max_ln_U_z")) *var = evecs[imax][2];
      goto exit;
    }
  } else if (tokens_token_and_keyword_equiv (name, "J")) {
    double F[3][3];
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    *var = MAT3DUTIL_DET_3X3(F);
  } else {
    *var = 0;
    return FALSE;
  }
 exit:
  return TRUE;
}

static W_mat_3d_state_p 
interpolate_mat_state (W_mat_3d_t *m,
                       W_loc_t *at_point,
                       int npkgs,
                       WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}

static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, 
               double C_mat[6][6] /* Eulerian material stiffness matrix */
               )
{
  mat_t *u = (mat_t *)m->data;
  double lambda = u->lambda, mu = u->G, lp, mp, J;
  double F[3][3];

  if (stiff_type == MAT_3D_STIFF_EULERIAN) {
    
    if (!ip_ctx->get_kinem_var) return FALSE;
    
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    
    J = MAT3DUTIL_DET_3X3(F);
    
    lp = lambda / J; 
    mp = (mu - lambda * log (J)) / J;
    
#undef C
#define C(I,J,V) { C_mat[I][J] = V; }
    C(0,0,lp+2*mp); C(0,1,lp);      C(0,2,lp);      C(0,3,0);  C(0,4,0);  C(0,5,0);
    C(1,0,lp);      C(1,1,lp+2*mp); C(1,2,lp);      C(1,3,0);  C(1,4,0);  C(1,5,0);
    C(2,0,lp);      C(2,1,lp);      C(2,2,lp+2*mp); C(2,3,0);  C(2,4,0);  C(2,5,0);
    C(3,0,0);       C(3,1,0);       C(3,2,0);       C(3,3,mp); C(3,4,0);  C(3,5,0); 
    C(4,0,0);       C(4,1,0);       C(4,2,0);       C(4,3,0);  C(4,4,mp); C(4,5,0); 
    C(5,0,0);       C(5,1,0);       C(5,2,0);       C(5,3,0);  C(5,4,0);  C(5,5,mp); 
#undef C
    return TRUE;
  } else 
    return FALSE;
}
	
static W_mat_3d_methods_t mat3dstvk_methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  get_mat_state_named_var,
  NULL
};


W_mat_3d_t *
read_mat_3d_neohk (W_domain_t *d)
{
  const char *input_line;
  double rho = 1;
  mat_t *u;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      u = w_new(mat_t); CHECK_MALLOC(u);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { u->E = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { u->nu = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      u->lambda           = u->E * u->nu / (1 + u->nu) / (1 - 2*(u->nu));
      u->G                = u->E / 2 / (1 + u->nu);
      u->bulk_mod         = u->E / 3 / (1 - 2*u->nu);
      u->dil_wave_speed_3d
        = W_3D_dilatational_wave_speed (u->E, u->nu, rho);
      u->dil_wave_speed_pstrain
        = W_plstrain_dilatational_wave_speed (u->E, u->nu, rho);
      u->dil_wave_speed_pstress
        = W_plstress_dilatational_wave_speed (u->E, u->nu, rho);
      u->shear_wave_speed
        = W_shear_wave_speed (u->E, u->nu, rho);
      if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstress,
                                     u->shear_wave_speed,
                                     0.01,
                                     &u->rayleigh_pstress_wave_speed))
        REPORT_FATAL_ERROR (d->errh, "Rayleigh plane stress speed does not exist");
      if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstrain,
                                     u->shear_wave_speed,
                                     0.01,
                                     &u->rayleigh_pstrain_wave_speed))
        REPORT_FATAL_ERROR (d->errh, "Rayleigh plane strain speed does not exist");
      
      return WM3D_new_material(&mat3dstvk_methods, (void *)u, rho);
    }
  }
  return NULL;
}

