 
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dpf.h"

/* RCSHEADER "$Id: mat3dpf.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

/* This file implements the perfect (inviscid compressible) fluid  */


static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_PERF_FLUID;
}

 

typedef struct pf_mat_t {
  double   bulk_mod;
  char    *initial_pressure_expr;
}              pf_mat_t;


typedef struct matstate_t {
  double   mean_stress;
  double   deform_energy_density;
}              matstate_t;



static W_bool_t
get_mat_const(W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  pf_mat_t *u = (pf_mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = 15 * u->bulk_mod;
    break;
  case MAT_3D_POISSON:
    *c = 0.4;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = 0;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_TENSILE_STRENGTH:
    *c = 0; /* RAW need to read TENSILE_STRENGTH from the input */
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

static W_mat_3d_state_p 
get_mat_state(W_mat_3d_t *m, W_loc_t *at_point)
{
  pf_mat_t *u = (pf_mat_t *)m->data;
  matstate_t *ms;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->mean_stress = 0;

  if (u->initial_pressure_expr) {
    int stat;                                                   
    double v;
    
    expeval_set_value("x", at_point->x);
    expeval_set_value("y", at_point->y);
    expeval_set_value("z", at_point->z);
    stat = expeval_eval(u->initial_pressure_expr, &v);   
    if (stat != EXPEVAL_E_OK) {                                 
      sprintf(W_buf(), "%s", expeval_err_msgs[stat]);               
      REPORT_ERROR(NULL, W_buf());                                        
    }
    ms->mean_stress = -v;
  }

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

 
static W_bool_t 
get_stress(W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           );


static W_bool_t  
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  double rel_F[3][3];
  double rel_J, ln_rel_J;
  pf_mat_t *u = (pf_mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  W_bool_t result = TRUE;
  matstate_t input_matst;

  if (!ip_ctx->get_kinem_var) return FALSE;  

  /* kinematic quantities */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_REL_DEF_GRAD, rel_F))
    return FALSE;
  
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  rel_J = MAT3DUTIL_DET_3X3(rel_F);
  LN_FUNC(ln_rel_J, /* = log */ rel_J);

  matst->mean_stress += u->bulk_mod * ln_rel_J;
  matst->deform_energy_density += 0.5 * matst->mean_stress * ln_rel_J;

  result = get_stress(m, ms, ip_ctx, stress_type, stress, out_stress_type);

 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}




static W_bool_t 
get_mat_state_var(W_mat_3d_t           *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0;
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}



/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress (W_mat_3d_t                  *m,
            W_mat_3d_state_p             ms,
            W_mat_3d_ip_ctx_t           *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double                       stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  matstate_t *matst = (matstate_t *)ms;
  double sigma[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
  double F[3][3], inv_F[3][3], S[3][3];
  double J, f;

    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    sigma[0][0] = sigma[1][1] = sigma[2][2] = matst->mean_stress;
    MAT3DUTIL_COPY_3X3(stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_COPY_TRANSP_3X3(stress, inv_F);
    f = J * matst->mean_stress;
    MAT3DUTIL_SCALE_3X3(stress, f);
    break;
  case MAT_3D_2ND_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_COPY_TRANSP_3X3(stress, inv_F);
    f = J * matst->mean_stress;
    MAT3DUTIL_SCALE_3X3(stress, f);
    MAT3DUTIL_MULT_AB_3X3(S, inv_F, stress);
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    sigma[0][0] = sigma[1][1] = sigma[2][2] = matst->mean_stress;
    MAT3DUTIL_COPY_3X3(stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}



static W_mat_3d_state_p 
interpolate_mat_state(W_mat_3d_t *m,
		      W_loc_t *at_point,
		      int npkgs,
		      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  ms = (matstate_t *) get_mat_state(m, at_point);
  
  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    ms->mean_stress += pms->mean_stress * WGHT(k);
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
}

static W_mat_3d_methods_t mat3dpf_methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  NULL,
  NULL,
  NULL
};


W_mat_3d_t *
read_mat_3d_perf_fluid(W_domain_t *d)
{
  const char *input_line;
  double rho = 1;
  pf_mat_t *u;
  int kp;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      u = w_new(pf_mat_t); CHECK_MALLOC(u);
      TOKENS_SET_BY_KEYWORD("BULK_MOD", TRUE,
			    { u->bulk_mod = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      u->initial_pressure_expr = NULL;
      if ((kp = tokens_have_keyword(d->parser, 1, "INITIAL_PRESSURE")) > 0) {
	const char *s;
	s = TOKEN_STR(kp+1);
	DUPLICATE_STRING(u->initial_pressure_expr, s);
      }
        
      return WM3D_new_material (&mat3dpf_methods, (void *)u, rho);
    }
  }
  return NULL;
}
