 
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3drel.h"

/* RCSHEADER "$Id: mat3drel.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

/* The rate-elastic material is defined here.  The model is based on */
/* the paper by P. Pinsky, M. Ortiz, and K. S. Pister, Numerical integration */
/* of rate constitutive equations in finite deformation analysis, CMAME 40 */
/* (1983), 137--158. */ 

typedef struct rel_mat_t {
  double E, nu;
  double lambda, G, bulk_mod;
  double tensile_strength;
  double dil_wave_speed_3d;
  double dil_wave_speed_pstrain;
  double dil_wave_speed_pstress;
  double shear_wave_speed;
  double rayleigh_pstrain_wave_speed;
  double rayleigh_pstress_wave_speed;
  double dyn_fracture_toughness;
}              rel_mat_t;



typedef struct matstate_t {
  double   cauchy[6];           /* stores current Cauchy stress: */
                                /* s_xx, s_xy, s_xz, s_yy, s_yz, s_zz */
  double deform_energy_density;
}              matstate_t;




static W_bool_t
get_mat_const(W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  rel_mat_t *u = (rel_mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_TENSILE_STRENGTH:
    *c = u->tensile_strength; 
    break;
  case MAT_3D_DYN_FRACTURE_TOUGHNESS:
    *c = u->dyn_fracture_toughness; 
    break;
  case MAT_3D_DIL_WAVE_SPEED_3D:
    *c = u->dil_wave_speed_3d;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRAIN:
    *c = u->dil_wave_speed_pstrain;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRESS:
    *c = u->dil_wave_speed_pstress;
    break;
  case MAT_3D_SHEAR_WAVE_SPEED:
    *c = u->shear_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED:
    *c = u->rayleigh_pstrain_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRESS_WAVE_SPEED:
    *c = u->rayleigh_pstress_wave_speed;
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

static W_bool_t 
get_stress(W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           );


/* Function updates the stress state, and returns the stress (with proper */
/* conversion into the requested type).   */
static W_bool_t 
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  double midstep_D[3][3], rel_F[3][3], midstep_F[3][3];
  double sig[3][3], newsig[3][3], t[3][3], rotated[3][3];
  double incr_sig[3][3], rot_incr_sig[3][3];
  double dt = ip_ctx->time_step, fact;
  rel_mat_t *u = (rel_mat_t *)m->data;
  double lambda = u->lambda, twomu = 2 * u->G;
  double trmidstep_D, inv_rel_J, midstep_J, F[3][3];
  matstate_t *matst = (matstate_t *)ms;
  W_bool_t result = TRUE;
  matstate_t input_matst;

  if (!ip_ctx->get_kinem_var) return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_REL_DEF_GRAD, rel_F))
    return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_MIDSTEP_DEF_GRAD, midstep_F))
    return FALSE;
  midstep_J = MAT3DUTIL_DET_3X3(midstep_F);

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_MIDSTEP_RATE_OF_DEF, midstep_D))
    return FALSE;

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;

  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  sig[0][0]             = matst->cauchy[0];
  sig[0][1] = sig[1][0] = matst->cauchy[1];
  sig[0][2] = sig[2][0] = matst->cauchy[2];
  sig[1][1]             = matst->cauchy[3];
  sig[1][2] = sig[2][1] = matst->cauchy[4];
  sig[2][2]             = matst->cauchy[5];

  MAT3DUTIL_MULT_ABT_3X3(t, sig, rel_F);
  MAT3DUTIL_MULT_AB_3X3(rotated, rel_F, t);
  inv_rel_J = 1 / MAT3DUTIL_DET_3X3(rel_F);
  MAT3DUTIL_SCALE_3X3(rotated, inv_rel_J);

  trmidstep_D = midstep_D[0][0] + midstep_D[1][1] + midstep_D[2][2];
  incr_sig[0][0] = lambda * trmidstep_D + twomu * midstep_D[0][0];
  incr_sig[0][1] =                        twomu * midstep_D[0][1];
  incr_sig[0][2] =                        twomu * midstep_D[0][2];
  incr_sig[1][0] =                        twomu * midstep_D[1][0];
  incr_sig[1][1] = lambda * trmidstep_D + twomu * midstep_D[1][1];
  incr_sig[1][2] =                        twomu * midstep_D[1][2];
  incr_sig[2][0] =                        twomu * midstep_D[2][0];
  incr_sig[2][1] =                        twomu * midstep_D[2][1];
  incr_sig[2][2] = lambda * trmidstep_D + twomu * midstep_D[2][2];

  MAT3DUTIL_MULT_ABT_3X3(t, incr_sig, midstep_F);
  MAT3DUTIL_MULT_AB_3X3(rot_incr_sig, midstep_F, t);
  fact = dt / midstep_J;
  MAT3DUTIL_SCALE_3X3(rot_incr_sig, fact);
  MAT3DUTIL_ADD_AB_3X3(newsig, rot_incr_sig, sig);

  matst->cauchy[0] = newsig[0][0];
  matst->cauchy[1] = newsig[0][1];
  matst->cauchy[2] = newsig[0][2];
  matst->cauchy[3] = newsig[1][1];
  matst->cauchy[4] = newsig[1][2];
  matst->cauchy[5] = newsig[2][2];

  {
    double ide =
      dt * ((sig[0][0] + newsig[0][0]) * midstep_D[0][0] +
            (sig[0][1] + newsig[0][1]) * midstep_D[0][1] +
            (sig[0][2] + newsig[0][2]) * midstep_D[0][2] +
            (sig[1][0] + newsig[1][0]) * midstep_D[1][0] +
            (sig[1][1] + newsig[1][1]) * midstep_D[1][1] +
            (sig[1][2] + newsig[1][2]) * midstep_D[1][2] +
            (sig[2][0] + newsig[2][0]) * midstep_D[2][0] +
            (sig[2][1] + newsig[2][1]) * midstep_D[2][1] +
            (sig[2][2] + newsig[2][2]) * midstep_D[2][2]) *
      MAT3DUTIL_DET_3X3(F) * 0.5;
    matst->deform_energy_density += ide;
  }

  result = get_stress (m, ms, ip_ctx, stress_type, stress, out_stress_type);
   
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}



static W_mat_3d_state_p 
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  matstate_t *ms;
  int i;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  FOR(i, 6) ms->cauchy[i] = 0;

  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}



/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress(W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           )
{
  matstate_t *matst = (matstate_t *)ms;
  double sigma[3][3], F[3][3], inv_F[3][3], P[3][3], S[3][3];
  double J;

  sigma[0][0]               = matst->cauchy[0];
  sigma[0][1] = sigma[1][0] = matst->cauchy[1];
  sigma[0][2] = sigma[2][0] = matst->cauchy[2];
  sigma[1][1]               = matst->cauchy[3];
  sigma[1][2] = sigma[2][1] = matst->cauchy[4];
  sigma[2][2]               = matst->cauchy[5];
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
    MAT3DUTIL_SCALE_3X3(P, J);
    MAT3DUTIL_COPY_3X3(stress, P);
    break;
  case MAT_3D_2ND_PK:
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
      return FALSE;
    J = MAT3DUTIL_DET_3X3(F);
    if (!W_mat3dutil_invert_3x3(F, inv_F))
      REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
    MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
    MAT3DUTIL_MULT_AB_3X3(S, inv_F, P);
    MAT3DUTIL_SCALE_3X3(S, J);
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}



static W_bool_t 
get_mat_state_var(W_mat_3d_t           *m,
                       W_mat_3d_state_p             ms,
                       W_mat_3d_ip_ctx_t           *ip_ctx,
                       W_mat_3d_mat_state_var_enum  var_type,
                       double                      *var
                       )
{
  matstate_t *matst = (matstate_t *)ms;
  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0; /* elastic material */
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;
  default:
    return FALSE;
  }
}




static W_mat_3d_state_p 
interpolate_mat_state (W_mat_3d_t *m,
		      W_loc_t *at_point,
		      int npkgs,
		      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    FOR(i, 6) {
      ms->cauchy[i] += pms->cauchy[i] * WGHT(k);
    }
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }
  
  return (W_mat_3d_state_p) ms;
}


static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_PINSKY_ETAL_EL;
}


static W_mat_3d_methods_t mat3drel_methods = {
  get_mcode,
  get_mat_state,
  get_mat_const, 
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  NULL,
  NULL,
  NULL
};

W_mat_3d_t *
read_mat_3d_rate_el(W_domain_t *d)
{
  const char *input_line;
  double rho = 1;
  rel_mat_t *u;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      u = w_new(rel_mat_t); CHECK_MALLOC(u);
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { u->E = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { u->nu = TOKEN_DBL_NEXT(1); },
			    { NOOP(); }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { u->tensile_strength = TOKEN_DBL_NEXT(1); },
			    { u->tensile_strength = u->E / 100; }
			    );
      TOKENS_SET_BY_KEYWORD("DYN_FRACTURE_TOUGHNESS", FALSE,
			    { u->dyn_fracture_toughness = TOKEN_DBL_NEXT(1); },
			    { u->dyn_fracture_toughness = u->E / 1000; }
			    );
      u->lambda   = u->E * u->nu / (1 + u->nu) / (1 - 2*(u->nu));
      u->G        = u->E / 2 / (1 + u->nu);
      u->bulk_mod = u->E / 3 / (1 - 2*u->nu);
      u->dil_wave_speed_3d
        = W_3D_dilatational_wave_speed (u->E, u->nu, rho);
      u->dil_wave_speed_pstrain
        = W_plstrain_dilatational_wave_speed (u->E, u->nu, rho);
      u->dil_wave_speed_pstress
        = W_plstress_dilatational_wave_speed (u->E, u->nu, rho);
      u->shear_wave_speed
        = W_shear_wave_speed (u->E, u->nu, rho);
      if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstress,
                                     u->shear_wave_speed,
                                     0.01,
                                     &u->rayleigh_pstress_wave_speed))
        REPORT_FATAL_ERROR (d->errh, "Rayleigh plane stress speed does not exist");
      if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstrain,
                                     u->shear_wave_speed,
                                     0.01,
                                     &u->rayleigh_pstrain_wave_speed))
        REPORT_FATAL_ERROR (d->errh, "Rayleigh plane strain speed does not exist");
         
      return WM3D_new_material (&mat3drel_methods, (void *)u, rho);
    }
  }
  return NULL;
}


