/* Face centered polycryspls */
#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wdomain.h"
#include "Wmat3drtpoly.h"
    
#define N_SLIP_SYSTEMS 12
#define LAYERS 2
#define NUM_NODES (LAYERS+1)*LAYERS/2
#define NUM_ORIENT 3*(LAYERS+1)*LAYERS/2
    
#define PI 3.14159265358979
#define ISQRT2 0.70710678118654752440
#define ISQRT3 0.57735026918962576451
#define S(x,y,z) x*ISQRT2,y*ISQRT2,z*ISQRT2
#define M(x,y,z) x*ISQRT3,y*ISQRT3,z*ISQRT3
    
/* slip directions expressed in the a,b,c vectors */
static double ref_s[N_SLIP_SYSTEMS][3] = {
  {S(0,-1,1)},   /* B2 */
  {S(1,0,-1)},   /* B4 */
  {S(-1,1,0)},   /* B5 */
  {S(1,0,1)},    /* A3 */
  {S(0,1,-1)},   /* A2 */
  {S(-1,-1,0)},  /* A6 */
  {S(0,1,1)},    /* C1 */
  {S(-1,0,-1)},  /* C3 */
  {S(1,-1,0)},   /* C5 */
  {S(-1,0,1)},   /* D4 */
  {S(0,-1,-1)},  /* D1 */
  {S(1,1,0)}     /* D6 */
};
/* slip plane normals expressed in the a,b,c vectors */
static double ref_m[N_SLIP_SYSTEMS][3] = {
  {M(1,1,1)},   /* B2 */
  {M(1,1,1)},   /* B4 */
  {M(1,1,1)},   /* B5 */
  {M(-1,1,1)},  /* A3 */
  {M(-1,1,1)},  /* A2 */
  {M(-1,1,1)},  /* A6 */
  {M(-1,-1,1)}, /* C1 */
  {M(-1,-1,1)}, /* C3 */
  {M(-1,-1,1)}, /* C5 */
  {M(1,-1,1)},  /* D4 */
  {M(1,-1,1)},  /* D1 */
  {M(1,-1,1)}   /* D6 */
};
    
static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_RT_POLY;
}

/* RCSHEADER "$Id: mat3drtpoly.c,v 1.1 2004/06/08 21:56:06 cvs Exp $" */
  
typedef struct mat_t {
  double c11, c44, c12; /* elastic relation S=C:E for a cubic symmetry */
  double gamma_0; /* reference shear strain rate */
  double q; /* hardening behaviour parameter; q=1 .. isotropic (Taylor);
                      1 <= q <= 1.4 Kocks (1970) */
  double h_0; /* initial hardening modulus */
  double m_srs; /* strain-rate sensitivity exponent, m->0, then nearly independent*/ 
  double g_0; /* initial flow stress */
  double g_inf; /* saturation flow stress [Asaro (1983)] */
  int max_inner_iter; /* max number of iterations allowed in the inner loop */
  double activity_tol; /* tolerance under which systems are considered inactive */
  int debug; /* debug level */
  double e_incr; /* increment of the Green-Lagrange tensor for the numerical
                    differentiation to obtain the constitutive stiffness matrix */
  W_bool_t elastic; /* this material is elastic; no plasticity is allowed */
  W_bool_t rate_dependent; /* the material is rate dependent */
  double *rand;
  int    layers; /* tot num of lattice orientations = 3*n(n+1)/2 */
}              mat_t; 

typedef double vector_6[6];
typedef double tensor_3x3[3][3];
typedef double slip[N_SLIP_SYSTEMS];

typedef struct matstate_t {
  vector_6     S[NUM_ORIENT];   /* stores current 2nd P-K/kirchhoff stress: */
  slip         tau[NUM_ORIENT]; /* resolved shear stress on all slip systems */
  tensor_3x3   Fp[NUM_ORIENT]; /* current plastic deformation gradient */
  slip         g[NUM_ORIENT]; /* current flow stress on a slip system */
  slip         gamma[NUM_ORIENT]; /* current shear strain on a slip system */
  slip         dgamma[NUM_ORIENT]; /* current shear strain increment on a slip system */
  double       accum_gamma[NUM_ORIENT]; /* accumulated plastic slip on all slip systems */

  double       deform_energy_density; /* deformation energy density */
  double       average_stress[6]; /* averaged 2nd PK/kirchhoff stress */
  double       average_Fp[3][3]; /* averaged stress 2nd PK stress */
  double       average_tau[N_SLIP_SYSTEMS];
  double       average_g[N_SLIP_SYSTEMS];
  double       average_gamma[N_SLIP_SYSTEMS];
}              matstate_t;

#define SIGN(a) (a > 0? +1 : -1)

static W_bool_t
get_mat_const(W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;
  double lambda = u->c12; /* this is a very crude approximation: only true for c11-c12=2*c44 */
  double G = 2 * u->c44; /* faked: this only makes sense for isotropic materials */
  double nu = lambda / 2 / (lambda + G);
  double E = 2 * (1 + nu) * G;
  double bulk_mod = E / 3 / (1 - 2*nu);

  switch (type) { 
  case MAT_3D_YOUNG_MODULUS:
    *c = E;
    break;
  case MAT_3D_POISSON:
    *c = nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = bulk_mod;
    break;
  default: /* quantities that are not supported */
    *c = 0;
    return FALSE;
  }
  return TRUE;
}


static W_mat_3d_state_p 
get_mat_state(W_mat_3d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;
  int i,orientation;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  FOR (i, 6) ms->average_stress[i] = 0; /* averaged stress */
  MAT3DUTIL_UNIT_3X3 (ms->average_Fp);
  FOR (i, N_SLIP_SYSTEMS) {
    ms->average_g[i] = u->g_0;
    ms->average_tau[i] = 0;
    ms->average_gamma[i] = 0;
  }
  FOR(orientation, NUM_ORIENT) {
    FOR (i, 6) ms->S[orientation][i] = 0;
    MAT3DUTIL_UNIT_3X3 (ms->Fp[orientation]);
    FOR (i, N_SLIP_SYSTEMS) {
      ms->g[orientation][i] = u->g_0;
      ms->gamma[orientation][i] = 0;
      ms->dgamma[orientation][i] = 0;
    }
    ms->accum_gamma[orientation] = 0; 
  }
  ms->deform_energy_density = 0;
  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_stress (W_mat_3d_t                  *m,
           W_mat_3d_state_p             ms,
           W_mat_3d_ip_ctx_t           *ip_ctx,
           W_mat_3d_stress_type_enum    stress_type,
           double                       stress[3][3],
           W_mat_3d_stress_type_enum   *out_stress_type
           );

static void
calc_Dbar (double c11, double c12, double c44,
           W_vect_t *a, W_vect_t *b, W_vect_t *c,
           double Dbar[6][6]
           )
{
  double a1 = (*a).x, a2 = (*a).y, a3 = (*a).z;
  double b1 = (*b).x, b2 = (*b).y, b3 = (*b).z;
  double c1 = (*c).x, c2 = (*c).y, c3 = (*c).z;
  double c11c122c44 = (c11 - c12 - 2*c44);
  
  Dbar[0][0] = c12 +  2*c44 + c11c122c44 * ((a1*a1*a1*a1 + b1*b1*b1*b1 + c1*c1*c1*c1));
  Dbar[0][1] = Dbar[1][0] = c12 +   c11c122c44 * ((a1*a1*a2*a2 + b1*b1*b2*b2 + c1*c1*c2*c2));
  Dbar[0][2] = Dbar[2][0] = c12 +   c11c122c44 * ((a1*a1*a3*a3 + b1*b1*b3*b3 + c1*c1*c3*c3));
  Dbar[0][3] = Dbar[3][0] = ((  c11c122c44 * ((a1*a1*a1*a2 + b1*b1*b1*b2 + c1*c1*c1*c2))) + (  c11c122c44 * ((a1*a1*a2*a1 + b1*b1*b2*b1 + c1*c1*c2*c1))))/2;
  Dbar[0][4] = Dbar[4][0] = ((  c11c122c44 * ((a1*a1*a1*a3 + b1*b1*b1*b3 + c1*c1*c1*c3))) + (  c11c122c44 * ((a1*a1*a3*a1 + b1*b1*b3*b1 + c1*c1*c3*c1))))/2;
  Dbar[0][5] = Dbar[5][0] = ((  c11c122c44 * ((a1*a1*a2*a3 + b1*b1*b2*b3 + c1*c1*c2*c3))) + (  c11c122c44 * ((a1*a1*a3*a2 + b1*b1*b3*b2 + c1*c1*c3*c2))))/2;
  Dbar[1][1] = c12 +  2*c44 + c11c122c44 * ((a2*a2*a2*a2 + b2*b2*b2*b2 + c2*c2*c2*c2));
  Dbar[1][2] = Dbar[2][1] = c12 +   c11c122c44 * ((a2*a2*a3*a3 + b2*b2*b3*b3 + c2*c2*c3*c3));
  Dbar[1][3] = Dbar[3][1] = ((  c11c122c44 * ((a2*a2*a1*a2 + b2*b2*b1*b2 + c2*c2*c1*c2))) + (  c11c122c44 * ((a2*a2*a2*a1 + b2*b2*b2*b1 + c2*c2*c2*c1))))/2;
  Dbar[1][4] = Dbar[4][1] = ((  c11c122c44 * ((a2*a2*a1*a3 + b2*b2*b1*b3 + c2*c2*c1*c3))) + (  c11c122c44 * ((a2*a2*a3*a1 + b2*b2*b3*b1 + c2*c2*c3*c1))))/2;
  Dbar[1][5] = Dbar[5][1] = ((  c11c122c44 * ((a2*a2*a2*a3 + b2*b2*b2*b3 + c2*c2*c2*c3))) + (  c11c122c44 * ((a2*a2*a3*a2 + b2*b2*b3*b2 + c2*c2*c3*c2))))/2;
  Dbar[2][2] = c12 +  2*c44 + c11c122c44 * ((a3*a3*a3*a3 + b3*b3*b3*b3 + c3*c3*c3*c3));
  Dbar[2][3] = Dbar[3][2] = ((  c11c122c44 * ((a3*a3*a1*a2 + b3*b3*b1*b2 + c3*c3*c1*c2))) + (  c11c122c44 * ((a3*a3*a2*a1 + b3*b3*b2*b1 + c3*c3*c2*c1))))/2;
  Dbar[2][4] = Dbar[4][2] = ((  c11c122c44 * ((a3*a3*a1*a3 + b3*b3*b1*b3 + c3*c3*c1*c3))) + (  c11c122c44 * ((a3*a3*a3*a1 + b3*b3*b3*b1 + c3*c3*c3*c1))))/2;
  Dbar[2][5] = Dbar[5][2] = ((  c11c122c44 * ((a3*a3*a2*a3 + b3*b3*b2*b3 + c3*c3*c2*c3))) + (  c11c122c44 * ((a3*a3*a3*a2 + b3*b3*b3*b2 + c3*c3*c3*c2))))/2;
  Dbar[3][3] = (( 2*c44*(0.5) + c11c122c44 * ((a1*a2*a1*a2 + b1*b2*b1*b2 + c1*c2*c1*c2))) + ( 2*c44*(0.5) + c11c122c44 * ((a1*a2*a2*a1 + b1*b2*b2*b1 + c1*c2*c2*c1))))/2;
  Dbar[3][4] = Dbar[4][3] = ((  c11c122c44 * ((a1*a2*a1*a3 + b1*b2*b1*b3 + c1*c2*c1*c3))) + (  c11c122c44 * ((a1*a2*a3*a1 + b1*b2*b3*b1 + c1*c2*c3*c1))))/2;
  Dbar[3][5] = Dbar[5][3] = ((  c11c122c44 * ((a1*a2*a2*a3 + b1*b2*b2*b3 + c1*c2*c2*c3))) + (  c11c122c44 * ((a1*a2*a3*a2 + b1*b2*b3*b2 + c1*c2*c3*c2))))/2;
  Dbar[4][4] = (( 2*c44*(0.5) + c11c122c44 * ((a1*a3*a1*a3 + b1*b3*b1*b3 + c1*c3*c1*c3))) + ( 2*c44*(0.5) + c11c122c44 * ((a1*a3*a3*a1 + b1*b3*b3*b1 + c1*c3*c3*c1))))/2;
  Dbar[4][5] = Dbar[5][4] = ((  c11c122c44 * ((a1*a3*a2*a3 + b1*b3*b2*b3 + c1*c3*c2*c3))) + (  c11c122c44 * ((a1*a3*a3*a2 + b1*b3*b3*b2 + c1*c3*c3*c2))))/2;
  Dbar[5][5] = (( 2*c44*(0.5) + c11c122c44 * ((a2*a3*a2*a3 + b2*b3*b2*b3 + c2*c3*c2*c3))) + ( 2*c44*(0.5) + c11c122c44 * ((a2*a3*a3*a2 + b2*b3*b3*b2 + c2*c3*c3*c2))))/2;
}

#define VOP1(dest, op1, a, numel)               \
{                                               \
  int _j__, _N__ = numel;                       \
  FOR (_j__, _N__) {                            \
    dest[_j__] op1 a[_j__];                     \
  }                                             \
}
#define VOP1C(dest, op1, C, numel)               \
{                                               \
  int _j__, _N__ = numel;                       \
  FOR (_j__, _N__) {                            \
    dest[_j__] op1 C;                     \
  }                                             \
}
#define VOP2(dest, op1, a, op2, b, numel)       \
{                                               \
  int _j__, _N__ = numel;                       \
  FOR (_j__, _N__) {                            \
    dest[_j__] op1 a[_j__] op2 b[_j__];         \
  }                                             \
}
#define MOP1(dest, op1, a, numel)               \
{                                               \
  int _i__, _j__, _N__ = numel;                 \
  FOR (_i__, _N__) {                            \
     FOR (_j__, _N__) {                         \
       dest[_i__][_j__] op1 a[_i__][_j__];      \
     }                                          \
  }                                             \
}
#define MOP2(dest, op1, a, op2, b, numel)                       \
{                                                               \
  int _i__, _j__, _N__ = numel;                                 \
  FOR (_i__, _N__) {                                            \
     FOR (_j__, _N__) {                                         \
       dest[_i__][_j__] op1 a[_i__][_j__] op2 b[_i__][_j_];     \
     }                                                          \
  }                                                             \
}

static void
calc_Cebar1 (double F1[3][3], /* F_n+1   total deformation gradient at t_n+1*/
             double Fp1[3][3], /* F^p_n+1  current plastic deformation gradient at t_n+1 */
             double Cebar1[3][3] /* F^p_n+1  current plastic deformation gradient at t_n+1 */
             )
{
  double Fe1[3][3];
  double invFp1[3][3];
  double detFp1 = MAT3DUTIL_DET_3X3 (Fp1);
  if (detFp1 == 0) {
    REPORT_FATAL_ERROR (NULL, "Zero det(Fp1)");
  }
  MAT3DUTIL_INVERT_3X3 (detFp1, Fp1, invFp1);
  MAT3DUTIL_MULT_AB_3X3 (Fe1, F1, invFp1);
  MAT3DUTIL_MULT_ATB_3X3 (Cebar1, Fe1, Fe1);
}

static void
calc_Sbar (double Dbar[6][6],/* Sbar=Dbar:Ebar: 2nd PK stress on the intermediate configuration  */
        double Cebar1[3][3],/* elastic Cauchy-Green tensor */
        double Fp1[3][3], /* F^p_n+1  current plastic deformation gradient at t_n+1 */
        double Sbar[3][3] /* OUT: PK2 stress tensor */
        )
{
  double E[3][3];
  double Ev[6];
  double Sv[6];
  MAT3DUTIL_2_MAT (E, = 0.5 *, Cebar1);
  MAT3DUTIL_MAT_MOD_DIAG (E, -=, 0.5);
  Ev[0] = E[0][0];
  Ev[1] = E[1][1];
  Ev[2] = E[2][2];
  Ev[3] = E[0][1]+E[1][0];
  Ev[4] = E[0][2]+E[2][0];
  Ev[5] = E[2][1]+E[1][2];
  MAT3DUTIL_S_C_E (Sv,Dbar,Ev);
  MAT3DUTIL_V_TO_T (Sv, Sbar);
}
      
static void
calc_S (double Sbar[3][3], double Fp[3][3], double S[3][3])
{
  double t[3][3], inv_Fp[3][3];
  double det_Fp = MAT3DUTIL_DET_3X3 (Fp);
  MAT3DUTIL_INVERT_3X3 (det_Fp, Fp, inv_Fp);
  MAT3DUTIL_MULT_ABT_3X3 (t, Sbar, inv_Fp);
  MAT3DUTIL_MULT_AB_3X3  (S, inv_Fp, t);
}

static double
l2_norm_squared (double f[N_SLIP_SYSTEMS])
{
  double d = 0;
  int i;
  FOR (i, N_SLIP_SYSTEMS) { d += (f[i])*(f[i]); }
  return d;
}

static double
linf_norm (double f[N_SLIP_SYSTEMS])
{
  double d = 0;
  int i;
  FOR (i, N_SLIP_SYSTEMS) { if ( fabs(f[i]) > d ) d = fabs(f[i]); }
  return d;
}

static void
calc_Fp1 (double Fp[3][3],
          double dgamma[N_SLIP_SYSTEMS],
          double s[N_SLIP_SYSTEMS][3],
          double m[N_SLIP_SYSTEMS][3], double Fp1[3][3])
{
#define IMPL_FP 0
#if IMPL_FP /* Implicit time stepping of the Fp */
  int i;
  double Id_dgsm_sum[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
  double inv_Id_dgsm_sum[3][3];
  FOR (i, N_SLIP_SYSTEMS) {
    double dgsm[3][3], minus_dg =  - dgamma[i];
    MAT3DUTIL_TENS_PROD_V (dgsm, s[i], m[i]);
    MAT3DUTIL_SCALE_3X3 (dgsm, minus_dg);
    MAT3DUTIL_INCR_A_BY_B_3X3 (Id_dgsm_sum, dgsm);
  }
  MAT3DUTIL_MAT_MOD_DIAG (Id_dgsm_sum, +=, 1.0);
  if (!W_mat3dutil_invert_3x3 (Id_dgsm_sum, inv_Id_dgsm_sum)) {
    REPORT_FATAL_ERROR (NULL, "Id_dgsm_sum is singular");
  }
  MAT3DUTIL_MULT_AB_3X3 (Fp1, inv_Id_dgsm_sum, Fp);
#else /* Explicit time stepping of the Fp */
  int i;
  double Id_dgsm_sum[3][3] = {{0,0,0},{0,0,0},{0,0,0}};
  FOR (i, N_SLIP_SYSTEMS) {
    double dgsm[3][3], minus_dg =  dgamma[i];
    MAT3DUTIL_TENS_PROD_V (dgsm, s[i], m[i]);
    MAT3DUTIL_SCALE_3X3 (dgsm, minus_dg);
    MAT3DUTIL_INCR_A_BY_B_3X3 (Id_dgsm_sum, dgsm);
  }
  MAT3DUTIL_MAT_MOD_DIAG (Id_dgsm_sum, +=, 1.0);
  MAT3DUTIL_MULT_AB_3X3 (Fp1, Id_dgsm_sum, Fp);
#endif
}

static void
calc_g1 (double q, double h_0, double g_0, double g_inf,
         double g[N_SLIP_SYSTEMS],
         double accum_gamma, 
         double dgamma[N_SLIP_SYSTEMS],
         double g1[N_SLIP_SYSTEMS])
{
#define DELTA(i,j) (i==j?1:0)
  int i, j;
  double hgamma = 0, accum_gamma1 = accum_gamma, ch;
  FOR (i, N_SLIP_SYSTEMS) { accum_gamma1 += fabs(dgamma[i]); }
  ch = cosh (h_0 * accum_gamma1 / (g_inf - g_0));
  hgamma = h_0 / (ch * ch); 
  FOR (i, N_SLIP_SYSTEMS) {
    double dg = 0;
    FOR (j, N_SLIP_SYSTEMS) {
      double h1ij = hgamma * (q + DELTA(i,j)*(1-q));
      dg += h1ij * fabs(dgamma[j]);
    }
    g1[i] = g[i] + dg;
  }
}
/* = sech^2(h_0*gamma_sum/(tau_s - tau_0)) */

#define DBGP_DGAMMA(tit,what)  {                                        \
    int I_I_;                                                           \
    fprintf (stderr, "%s = [ ", tit);                                   \
    FOR (I_I_, N_SLIP_SYSTEMS) { fprintf (stderr, "%g ", what[I_I_]); }	\
    fprintf (stderr, "]\n");                                            \
  }

#define DBGP(tit,what)  {                                                    \
    int I_I_;                                                           \
    fprintf (stderr, "%s = [ ", tit);                                   \
    FOR (I_I_, N_SLIP_SYSTEMS) { fprintf (stderr, "%g ", what[I_I_]); } \
    fprintf (stderr, "]\n");                                             \
  }

#define DBGPM(tit,what)  {                                   \
    int I_I_, J_J_;                                     \
    fprintf (stderr, "%s = [ ", tit);                    \
    FOR (I_I_, N_SLIP_SYSTEMS) {                        \
       FOR (J_J_, N_SLIP_SYSTEMS) {                     \
         fprintf (stderr, "%g ", what[I_I_][J_J_]);     \
       }                                                \
       fprintf (stderr, ";\n");                          \
    }                                                   \
    fprintf (stderr, "]\n");                             \
  }
#define DBGPM1(tit,what)  {                                   \
    int I_I_, J_J_;                                     \
    fprintf (stderr, "%s = [ ", tit);                    \
    FOR (I_I_, 3) {                        \
       FOR (J_J_, 3) {                     \
         fprintf (stderr, "%g ", what[I_I_][J_J_]);     \
       }                                                \
       fprintf (stderr, ";");                          \
    }                                                   \
    fprintf (stderr, "]\n");                             \
  }
#define DBGPMK(tit,what)  {                                   \
    int I_I_, J_J_;                                     \
    fprintf (stderr, "%s = [ ", tit);                    \
    FOR (I_I_, 24) {                        \
       FOR (J_J_, 24) {                     \
         fprintf (stderr, "%g ", what[I_I_][J_J_]);     \
       }                                                \
       fprintf (stderr, ";\n");                          \
    }                                                   \
    fprintf (stderr, "]\n");                             \
  }
/* -(exp(log(p)*m)-1)/m */

static void
get_elastic_moduli (double c11,
                    double c12,
                    double c44,
                    double Ce[6][6])
{
  int i, j;
  FOR(i, 6){FOR(j, 6){Ce[i][j] = 0.0;}}

  Ce[0][0] = c11; Ce[1][1] = c11; Ce[2][2] = c11;
  Ce[3][3] = c44; Ce[4][4] = c44; Ce[5][5] = c44;

  Ce[0][1] = c12; Ce[1][0] = c12;
  Ce[0][2] = c12; Ce[2][0] = c12;
  Ce[1][2] = c12; Ce[2][1] = c12;
}

static void
contract_v6_t6x6(double v1[6], double v[6], double t[6][6])
{
  int i, j;
  FOR(i, 6){
     v1[i] = 0.0;
     FOR(j, 3){
        v1[i] += v[j]*t[j][i];
     }
         v1[i] += 2 * (v[3]*t[3][i] + v[4]*t[4][i] + v[5]*t[5][i]);
  }
}

static void
contract_t6x6_v6 (double v1[6], double t[6][6], double v[6])
{
  int i, j;
  FOR(i, 6){
     v1[i] = 0.0;
     FOR(j, 6){ v1[i] += v[j]*t[i][j]; }
  }
}

/* hardening moduli */
static double 
get_hij (double q, 
         double h_0, 
         double g_0, 
         double g_inf,
         double accum_gamma, 
         int    isys, 
         int    jsys) 
{
#define DELTA(i,j) (i==j?1:0)
  double hgamma = 0, ch, h_ij;
  ch = cosh (h_0 * accum_gamma / (g_inf - g_0));
  hgamma = h_0 / (ch * ch); 
  h_ij = hgamma * (q + DELTA(isys,jsys)*(1-q));
  return h_ij;
}

static void
mirror (double C[6][6])
{
   int i, j;
   /* copy the upper to lower triangle */
   for(i=1; i<6; i++){
      for(j=0; j<i; j++){
         C[i][j]=C[j][i];
      }
   }
}

/* elastic moduli Dc*/
static void
calc_Dc (double Fp[3][3],
         double Ce[6][6],
         double Cbar[6][6])
{
  int indx[9][2] = {{0,0},{1,1},{2,2},
                    {0,1},{0,2},{1,2},
                    {1,0},{2,0},{2,1}};
  int i, j, k, l;
  int ix, jx, kx, lx, m, n, r, s;
  double Ce9x9[9][9];

  double det, invFp[3][3];
  det = MAT3DUTIL_DET_3X3(Fp);
  MAT3DUTIL_INVERT_3X3(det, Fp, invFp);

  FOR(i, 6){FOR(j, 6){Cbar[i][j] = 0.0;}}

  /* make Ce9x9 */
  FOR (i, 9) {
      FOR(j,9) {
        if (i <=5  && j <= 5) {
                Ce9x9[i][j] = Ce[i][j];
        } else {
                if (i >5 && j<=5 ) Ce9x9[i][j] = Ce[i-3][j];
                else {
                if (i <=5 && j>5 ) Ce9x9[i][j] = Ce[i][j-3];
                else           Ce9x9[i][j] = Ce[i-3][j-3];
                     }
               }
      }
  }

  /*Cbar must be symmetric
    upper triangle of Cbar */
  for(i=0; i<6; i++){
     ix = indx[i][0];jx = indx[i][1];
     for(j=i; j<6; j++){
         kx = indx[j][0];lx = indx[j][1];
         FOR(k, 9){
            m = indx[k][0]; n = indx[k][1];
            FOR(l, 9){
               r = indx[l][0]; s = indx[l][1];
               Cbar[i][j] += Ce9x9[k][l]*invFp[ix][m]*invFp[jx][n]*invFp[kx][r]*invFp[lx][s];
            }
         }
     }
  }
  /* mirror */
  mirror (Cbar);
}

static void 
calc_P(double s[3],
       double m[3],
       double F[3][3], 
       double Fp[3][3], 
       double P[3][3])
{
  int i, j, k;
  double geco[3][3];
  double det, Fe[3][3], Fp_inv[3][3];
  double tmp[3], s_m[3][3];

  /* Fe */
  det = MAT3DUTIL_DET_3X3(Fp);
  MAT3DUTIL_INVERT_3X3(det, Fp, Fp_inv);
  MAT3DUTIL_MULT_AB_3X3(Fe, F, Fp_inv);

  MAT3DUTIL_MULT_ATB_3X3(geco, Fe, Fe);

  MAT3DUTIL_MULT_A_3X3_V_3(tmp, geco, s);
  MAT3DUTIL_TENS_PROD_V(s_m, tmp, m);
  MAT3DUTIL_SYMM(P, s_m);
}

static void 
calc_W(double s[3],
       double m[3],
       double F[3][3], 
       double Fp[3][3], 
       double W[3][3])
{
  int i, j, k;
  double geco[3][3];
  double det, Fe[3][3], Fp_inv[3][3];
  double tmp[3], s_m[3][3];

  /* Fe */
  det = MAT3DUTIL_DET_3X3(Fp);
  MAT3DUTIL_INVERT_3X3(det, Fp, Fp_inv);
  MAT3DUTIL_MULT_AB_3X3(Fe, F, Fp_inv);

  MAT3DUTIL_MULT_ATB_3X3(geco, Fe, Fe);

  MAT3DUTIL_MULT_A_3X3_V_3(tmp, geco, s);
  MAT3DUTIL_TENS_PROD_V(s_m, tmp, m);
  MAT3DUTIL_ASYMM(W, s_m);
}

/* 
   beta = skewsm*ks- ks*skewm 
   on the star(intermediate) basis 
*/
static void
calc_beta(double F[3][3],
          double Fp[3][3],
          double s[3],
          double m[3],
          double S[6], 
          double beta[3][3] 
         )
{
  double gw[3][3], St[3][3], Se[3][3];
  double invFp[3][3], Fe[3][3], skewsm[3][3], tmp[3][3];
  double det, geco[3][3], gecontra[3][3];

  /* stress on the intermediate basis */
  MAT3DUTIL_V_TO_T(S, St);
  MAT3DUTIL_MULT_ABAT_3X3(Se, Fp, St);
  
  /* Fe */
  det = MAT3DUTIL_DET_3X3(Fp);
  MAT3DUTIL_INVERT_3X3(det, Fp, invFp);
  MAT3DUTIL_MULT_AB_3X3(Fe, F, invFp);

  /* metric tensors */
  MAT3DUTIL_MULT_ATB_3X3(geco, Fe, Fe);
  det = MAT3DUTIL_DET_3X3(geco);
  MAT3DUTIL_INVERT_3X3(det, geco, gecontra);
 
  calc_W (s, m, F, Fp, skewsm);

  MAT3DUTIL_MULT_AB_3X3(gw, gecontra, skewsm);
  MAT3DUTIL_MULT_AB_3X3(beta, gw, Se);
  MAT3DUTIL_COPY_TRANSP_3X3(tmp, beta);
  MAT3DUTIL_ADD_AB_3X3(beta, beta, tmp);
}

static double
mach_eps(void)
{
  double   epsilon;

  for (epsilon = 1e-10; 1.0 + epsilon>1.0; ) {epsilon = epsilon/2.0;}
  epsilon         *= 2.0;

  return epsilon;
}

static void
calc_Snew (W_bool_t elastic,
           double S[6],                /* n */
           double F[3][3],             /* n */
           double Fp[3][3],            /* n */
           double W[3][3],                /* n */ 
           double D[3][3],                /* n */
           double dgamma[N_SLIP_SYSTEMS], 
           double s[N_SLIP_SYSTEMS][3],
           double m[N_SLIP_SYSTEMS][3],
           double Dbar[6][6],
           double dt,
           double Snew[3][3],              /* n+1 */
           double F1[3][3]              /* n+1 */
          )
{
  double Sdot[3][3], St[3][3];
  double Ce_dot_D[6], Dv[6], tmp[3][3];
  double Dc[6][6];
  
  double symsm[3][3], skewsm[3][3];
  double R[3][3], Rv[6], Pv[6], beta[3][3];
  double Pbar[3][3], beta_bar[3][3];
 
  double de[3][3], dp[3][3];
  int i, j;
  double det, invFp[3][3];

  /* invFp */
  det = MAT3DUTIL_DET_3X3(Fp);
  MAT3DUTIL_INVERT_3X3(det, Fp, invFp);

  if(elastic) 
  {
    double C1[3][3], E1[3][3];
    double Ev[6];
    double Sv[6];
    MAT3DUTIL_MULT_ATB_3X3(C1, F1, F1);
    MAT3DUTIL_2_MAT (E1, = 0.5 *, C1);
    MAT3DUTIL_MAT_MOD_DIAG (E1, -=, 0.5);
    Ev[0] = E1[0][0];
    Ev[1] = E1[1][1];
    Ev[2] = E1[2][2];
    Ev[3] = E1[0][1]+E1[1][0];
    Ev[4] = E1[0][2]+E1[2][0];
    Ev[5] = E1[2][1]+E1[1][2];
    MAT3DUTIL_S_C_E (Sv,Dbar,Ev);
    MAT3DUTIL_V_TO_T (Sv, Snew);
    return;
  }

  /* zero out dp, which is the summation over dgamma*P */
  FOR(i, 3) FOR(j, 3) dp[i][j] = 0;

  Dv[0] = D[0][0];
  Dv[1] = D[1][1];
  Dv[2] = D[2][2];
  Dv[3] = D[0][1] + D[1][0];
  Dv[4] = D[0][2] + D[2][0];
  Dv[5] = D[1][2] + D[2][1];
  
  /* get correct elastic moduli */
  calc_Dc (Fp, Dbar, Dc); // this F should be at t_n

  MAT3DUTIL_S_C_E(Ce_dot_D, Dc, Dv);
  MAT3DUTIL_V_TO_T(Ce_dot_D, Sdot);

  /* plastic correction 
     - sum{dgamma*(Dbar:P + beta)} 
     then bring it to bar basis
     (1) P=sym{sm}, W=skew{sm}
     (2) beta = W*ks-ks*W = W*ks+transpose(W*ks)
     (3) ksold at t_n
     (4) Fp at t_n to bring it to bar basis
     (5) dgamma at t_n !!
   */
  if (!elastic){
     FOR (i, N_SLIP_SYSTEMS) {

      if (dgamma[i] !=0 ){
        /* get sym_sm, skew_sm */
        calc_P(s[i], m[i], F, Fp, symsm);

        /* bring symsm to bar basis */
        MAT3DUTIL_MULT_ATBA_3X3(Pbar, Fp, symsm);
    
        /* symsm--> vector */
        Pv[0] = Pbar[0][0];
        Pv[1] = Pbar[1][1];
        Pv[2] = Pbar[2][2];
        Pv[3] = Pbar[0][1] + Pbar[1][0];
        Pv[4] = Pbar[0][2] + Pbar[2][0];
        Pv[5] = Pbar[1][2] + Pbar[2][1];

        /* contract Dbar with symsm */
        MAT3DUTIL_S_C_E(Rv, Dc, Pv);
        MAT3DUTIL_V_TO_T(Rv, R);

        /* beta = skewsm*ksold - ksold*skewm */
        calc_beta(F, Fp, s[i], m[i], S, beta);

        /* bring beta to bar basis */
        MAT3DUTIL_MULT_ABAT_3X3(beta_bar, invFp, beta);

        /* complete R */
        MAT3DUTIL_ADD_AB_3X3(R, R, beta_bar);

        /* plastic correction */
        MAT3DUTIL_SCALE_3X3(R, -(dgamma[i]/dt));
        MAT3DUTIL_ADD_AB_3X3(Sdot, Sdot, R);

        /* dp */
        MAT3DUTIL_SCALE_3X3(Pbar, dgamma[i]);
        MAT3DUTIL_ADD_AB_3X3(dp, dp, Pbar);
     }
    }
  }

  /* de */
  FOR(i, 3)FOR(j, 3) de[i][j] = D[i][j] - dp[i][j];

  /* complete Sdot--material derivative */
  {
    double detgco, gco[3][3], gcontra[3][3]; 
    double gd[3][3];

    MAT3DUTIL_MULT_ATB_3X3(gco, F, F);
    detgco = MAT3DUTIL_DET_3X3(gco);
    MAT3DUTIL_INVERT_3X3(detgco, gco, gcontra);

    /* get stress tensor */
    MAT3DUTIL_V_TO_T(S, St);

    MAT3DUTIL_MULT_AB_3X3(gd, gcontra, dp);
    MAT3DUTIL_MULT_AB_3X3(tmp, gd, St);
    FOR(i, 3) FOR(j, 3) Sdot[i][j] = Sdot[i][j] - tmp[i][j] - tmp[j][i];
  }

  /* New stress */
  FOR(i, 3) FOR(j, 3) Snew[i][j] = St[i][j] + Sdot[i][j]*dt;
}

static void
calc_tau (double F[3][3],
          double Fp[3][3],
          double S[3][3],
          double s[N_SLIP_SYSTEMS][3],
          double m[N_SLIP_SYSTEMS][3],
          double tau[N_SLIP_SYSTEMS]
         )
{
  int i;
  double Se[3][3], t[3];
  double geco[3][3], gecos[3];
  double det, Fe[3][3], Fp_inv[3][3];

  MAT3DUTIL_MULT_ABAT_3X3(Se, Fp, S);

   /* Fe */
  det = MAT3DUTIL_DET_3X3(Fp);
  MAT3DUTIL_INVERT_3X3(det, Fp, Fp_inv);
  MAT3DUTIL_MULT_AB_3X3(Fe, F, Fp_inv);

  /* geco */
  MAT3DUTIL_MULT_ATB_3X3(geco, Fe, Fe);

  FOR(i, N_SLIP_SYSTEMS){
     MAT3DUTIL_MULT_A_3X3_V_3 (gecos, geco, s[i]);
     MAT3DUTIL_MULT_A_3X3_V_3 (t, Se, gecos);
     tau[i] = MAT3DUTIL_DOT_V (t, m[i]);
  }
}

/* calculate dgamma */
static void
get_dgamma (double F[3][3],
            double Fp[3][3],
	    double D[3][3],
            double S[6],
            double Dbar[6][6],
            double s[N_SLIP_SYSTEMS][3],
            double m[N_SLIP_SYSTEMS][3],
            double tau[N_SLIP_SYSTEMS],
            double g[N_SLIP_SYSTEMS],
            double dgamma[N_SLIP_SYSTEMS],
            double dt,
            double m_srs,
            double q, 
            double h_0, 
            double g_0, 
            double g_inf, 
            double gamma_0, /* reference shear strain rate */
            double accum_gamma
           )
{
  int i, j;
  double theta = 0.95;
  double beta[3][3], symsm[3][3];
  double D_star[3][3];
  double Dv[6], Pv[6], Rv[6];
  double det, invFp[3][3];

  double rhs[N_SLIP_SYSTEMS];
  GEP_MATRIX (jac);
  double rcond, z[N_SLIP_SYSTEMS];
  const double rcond_tol = 1e-20;
  int jacpvt[N_SLIP_SYSTEMS];

  GEP_STACKINIT(jac, N_SLIP_SYSTEMS);

  det = MAT3DUTIL_DET_3X3(Fp);
  MAT3DUTIL_INVERT_3X3(det, Fp, invFp);

  /* rate of deformation on the star basis(intermediate) */ 
  MAT3DUTIL_MULT_ATBA_3X3(D_star, invFp, D);

  /* LHS_ij = delta_ij+(theta*dt*gammadot_i/m_srs)*{R_i:P_j/tau_i + sgn(tau_j)*hij/g_i}
     RHS_i = dt * (gammadot_i + Q_i:D) 
   */
  FOR(i, N_SLIP_SYSTEMS) {

     double tau_over_g = fabs(tau[i]/g[i]);

     if (tau_over_g > 1.5) { REPORT_FATAL_ERROR (NULL, "tau is way too big!\n"); }
     if (tau_over_g != 0 ) {

       double gamma_incr = SIGN(tau[i])*dt*gamma_0*pow(tau_over_g, 1.0/m_srs);
       double factor = theta * gamma_incr/m_srs/tau[i];

       /* need R at i slip system on the intermediate basis
          R = D:P + beta 
        */
  
       calc_P (s[i], m[i], F, Fp, symsm);
       /* symsm--> vector */
       Pv[0] = symsm[0][0];
       Pv[1] = symsm[1][1];
       Pv[2] = symsm[2][2];
       Pv[3] = symsm[0][1] + symsm[1][0];
       Pv[4] = symsm[0][2] + symsm[2][0];
       Pv[5] = symsm[1][2] + symsm[2][1];
       MAT3DUTIL_S_C_E(Rv, Dbar, Pv);

       /* beta = skewsm*ks- ks*skewm */
       calc_beta(F, Fp, s[i], m[i], S, beta);
  
       /* complete Rv */
       Rv[0] += beta[0][0];
       Rv[1] += beta[1][1];
       Rv[2] += beta[2][2];
       Rv[3] += beta[0][1];
       Rv[4] += beta[0][2];
       Rv[5] += beta[1][2];
  
       MAT3DUTIL_T_TO_V(D_star, Dv);
       rhs[i] = gamma_incr + factor * dt * MAT3DUTIL_DOT_V6(Rv, Dv);

       FOR(j, N_SLIP_SYSTEMS) {
          double R_dot_P, hij;

          /* need P at j slip system */
          calc_P (s[j], m[j], F, Fp, symsm);

          /* contract Rv_i with P_j */
          R_dot_P = Rv[0]*symsm[0][0] + Rv[1]*symsm[1][1] + Rv[2]*symsm[2][2]
                  + 2.0*(Rv[3]*symsm[0][1] + Rv[4]*symsm[0][2] + Rv[5]*symsm[1][2]);

          hij = get_hij(q, h_0, g_0, g_inf, accum_gamma, i, j);

          GEP_ELEM(jac, i, j) = factor * (R_dot_P + SIGN(tau[j]) * hij * tau[i]/g[i]);
          if (i == j) GEP_ELEM(jac, i, j) += 1.0;
       }
     }
     else {  //tau[i] = 0
         rhs[i] = 0.0;
         FOR(j, N_SLIP_SYSTEMS) {
            GEP_ELEM(jac, i, j) = 0.0;
            if (i == j) GEP_ELEM(jac, i, j) += 1.0;
         }
     }
  } 

  /* solve the linear eqn sys N*dgamma = f for dgamma */
  gep_sgeco (&jac, jacpvt, &rcond, z);
  gep_sgesl (&jac, jacpvt, rhs, GEP_JOB_A_X_B);

  VOP1 (dgamma, =, rhs, N_SLIP_SYSTEMS); /* what sign?! POSITIVE incremental direction */
}

static void
upd (W_bool_t elastic,
     W_bool_t rate_dependent,
     W_bool_t must_be_accurate,
     double time,
     int debug, /* should I print debugging information, and how much? */
     int max_inner_iter,
     double activity_tol,
     double s[N_SLIP_SYSTEMS][3], /* slip directions */
     double m[N_SLIP_SYSTEMS][3], /* normals to slip planes */
     double g_0, /* initial flow stress */
     double g_inf, 
     double q, /* hardening behaviour parameter */
     double h_0, /* initial hardening rate */
     double gamma_0, /* reference shear strain rate */
     double m_srs, /* strain-rate sensitivity exponent */
     double Dbar[6][6], /* S=C:E */
     double F[3][3], /* F_n   total deformation gradient at t_n*/
     double F1[3][3], /* F_n+1   total deformation gradient at t_n+1*/
     double Fp[3][3], /* F^p_n  plastic deformation gradient at t_n */
     double g[N_SLIP_SYSTEMS], /* flow stress on slip system at t_n */
     double gamma[N_SLIP_SYSTEMS], /* shear strain on slip system at t_n */
     double accum_gamma, /* accumulated slip */
     double dt, /* time increment */
     double Sv_out[6], /* OUT: PK2 */
     double Fp1_out[3][3],          /* OUT: plastic deformation gradient */
     double g1_out[N_SLIP_SYSTEMS], /* OUT: flow stresses */
     double *accum_gamma_out, /* OUT: accumulated slip */
     double gamma1_out[N_SLIP_SYSTEMS], /* OUT: shear strain rates */
     double dgamma_out[N_SLIP_SYSTEMS], /* OUT: shear strain rates */
     double *deform_energy_density_out, /* OUT: deformation energy density */
     double tau[N_SLIP_SYSTEMS],    /* before updating, tau at t_n
                                           OUT: after upd,  tau at t_n+1 */
     double S[6],                    /* 2nd Pk at t_n */
     double W[3][3],            /* spin at t_n*/
     double D[3][3]             /* rate of def at t_n*/ 
     )
{
  int i;
  double dgamma[N_SLIP_SYSTEMS];
 
  VOP1C (dgamma, =, 0, N_SLIP_SYSTEMS);
  if (!elastic) {
     get_dgamma (F, Fp, D, S, Dbar, s, m, tau, g, dgamma, dt, m_srs, 
                 q, h_0, g_0, g_inf, gamma_0, accum_gamma);
  } /* plasticity treatment */
  
    /* Return the result */
  {
    double S1[3][3], Sbar1[3][3]; /* 2nd PK */
    double Cebar1[3][3]; /* Green-Lagrange strain */
    int j;
    double Snew[3][3];
    VOP2 (gamma1_out, =, gamma, +, dgamma, N_SLIP_SYSTEMS);
    *accum_gamma_out = accum_gamma;
    FOR (j, N_SLIP_SYSTEMS) { *accum_gamma_out += fabs(dgamma[j]); }
    FOR (j, N_SLIP_SYSTEMS) { dgamma_out[j] = dgamma[j]; }

    calc_Fp1 (Fp, dgamma, s, m, Fp1_out); 
    calc_Cebar1 (F1, Fp1_out, Cebar1); /* from old method */
    calc_Sbar (Dbar, Cebar1, Fp1_out, Sbar1); /* from old method */
//    calc_S (Sbar1, Fp1_out, S1); /* from old method */

    calc_Snew (elastic, S, F, Fp, W, D, 
               dgamma, s, m, Dbar, dt, Snew, F1);
    calc_tau (F1, Fp1_out, Snew, s, m, tau); /* save tau to matstate */
    calc_g1 (q, h_0, g_0, g_inf, g, *accum_gamma_out, dgamma, g1_out);
    MAT3DUTIL_T_TO_V (Snew, Sv_out);

    {
      double E1[3][3];
      MAT3DUTIL_2_MAT (E1, = 0.5 *, Cebar1);
      MAT3DUTIL_MAT_MOD_DIAG (E1, -=, 0.5);
      *deform_energy_density_out
        = 0.5 * MAT3DUTIL_DBL_PROD_3X3 (Sbar1, E1);
    }
  }
}

/* 
   calculate the transformation matrix 
   from lattice basis to lab basis 
 */
static void
get_phi (int      i,
         int      j,
         int      n,
         int      layers,
         double  *rand,
         double   phi[3][3])
{
  double x = 1.0;
  double y, z, dl;
  double a, b, c;
  double e2_norm, norm, dot, under_root;
  double e1[3],e2[3],e3[3];
  int idx = i*(i+1)/2 + j;
  int k;
  double r = rand[idx];
  double phi1[3][3], phi2[3][3];
  double theta;

  if (layers != 1 ) dl = 1.0/(double)(layers-1);
  
  if (i == 0){ // loading axis e3 is the same as a
     e3[0] = 1;
     e3[1] = 0.0;
     e3[2] = 0.0;

     e1[0] = 0.0;
     e1[1] = r;
     e1[2] = sqrt(1-r*r);

     e2[0] = 0.0;
     e2[1] = -sqrt(1-r*r);
     e2[2] = r;

  }
  else{
     y = dl* (double)i;
     z = dl* (double)j;
     norm = sqrt(x*x + y*y + z*z);

     /* Calculate the tensile direction t */
     e3[0] = x/norm;
     e3[1] = y/norm;
     e3[2] = z/norm;

     /* Calculate e1 */
       e1[0] = r;
       a = e3[1]*e3[1] + e3[2]*e3[2];
       b = 2.0 * e3[0] * e3[2] * e1[0];
       c = (e3[0]*e3[0] + e3[1]*e3[1]) * e1[0]*e1[0] - e3[1]*e3[1];
       under_root = b*b-4.0*a*c;
       if (under_root < 0){ 
          fprintf (stderr, "r(%ld, %ld) = %g\n", i, j, r);
          REPORT_FATAL_ERROR (NULL, "bad rand number?");
       }

    e1[2] = (-b - sqrt(under_root))/2.0/a;
    e1[1] = -sqrt(1 - e1[0]*e1[0] - e1[2]*e1[2]); 
    dot = MAT3DUTIL_DOT_V(e1,e3);
    if(fabs(dot) > 1e-10){e1[1] = -e1[1];}
 
    /* Calculate e2 by cross product of e1 and e3 */
    e2[0] = e1[2] * e3[1] - e1[1] * e3[2];
    e2[1] = e1[0] * e3[2] - e1[2] * e3[0];
    e2[2] = e1[1] * e3[0] - e1[0] * e3[1];

    /* Normalize e2 */
    e2_norm = sqrt(e2[0]*e2[0] + e2[1]*e2[1] + e2[2]*e2[2]);
    e2[0] = e2[0]/e2_norm;
    e2[1] = e2[1]/e2_norm;
    e2[2] = e2[2]/e2_norm;
  }

  /* Get the transformation matrices */
  for (k=0; k<3; k++){
      phi1[0][k] = e1[k];
      phi1[1][k] = e2[k];
      phi1[2][k] = e3[k];
  }

/*
  {
    double angle[N_SLIP_SYSTEMS];
    double s;
    
    fprintf(stderr, "angle = [");
    FOR(i, N_SLIP_SYSTEMS) {
      angle[i] = 0;
      FOR(k, 3) {
         s = ref_s[i][k];
         angle[i] += e3[k] * s;
      }
      fprintf(stderr, "%g;", angle[i]);  
    }
    fprintf(stderr, "]\n");
  }
*/  

  if (n == 0)      theta = 0;
  else if (n == 1) theta = 2.0*PI/3.0;
  else             theta = 4.0*PI/3.0;

  phi2[0][0] =  cos(theta);
  phi2[1][1] =  cos(theta);
  phi2[2][2] =  1.0;

  phi2[0][1] = -sin(theta);
  phi2[1][0] =  sin(theta);

  phi2[0][2] =  0;
  phi2[2][0] =  0;

  phi2[1][2] =  0;
  phi2[2][1] =  0;

  MAT3DUTIL_MULT_AB_3X3 (phi, phi2, phi1);

  {
    double R[3][3], U[3][3];
    double det1 = MAT3DUTIL_DET_3X3(phi1); 
    double det = MAT3DUTIL_DET_3X3(phi); 
/*
    fprintf(stderr, "theta = %g\n", theta);
*/
    MAT3DUTIL_MULT_ATB_3X3(R, phi, phi);

    if (det < 0.0){ /* is this alright? */}
    if (fabs(det-1.0) >= mach_eps()){
      if (!W_polar_decomposition (phi, R, U)) {
        REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
      }
      MAT3DUTIL_COPY_3X3(phi, R);
    }
    det = MAT3DUTIL_DET_3X3(phi); 
  }
}

static void
sum_over_orientations(double S[6],
                      double Sv[6],
                      double phi[3][3])
{
  double S1[3][3], tmp[3][3];
  MAT3DUTIL_V_TO_T(Sv, tmp);
  MAT3DUTIL_MULT_ABAT_3X3(S1, phi, tmp);

  S[0] += S1[0][0]; 
  S[1] += S1[1][1]; 
  S[2] += S1[2][2]; 
  S[3] += S1[0][1]; 
  S[4] += S1[0][2]; 
  S[5] += S1[1][2]; 
}

static void
get_rate_of_deformation(double F1[3][3],
                        double F0[3][3],
                        double dt,
                        double D[3][3],
                        double W[3][3])
{
  int i, j, k;
  double L[3][3], Fdot[3][3];
  FOR(i, 3)FOR(j, 3) Fdot[i][j] = (F1[i][j] - F0[i][j])/dt;
  MAT3DUTIL_MULT_ATB_3X3(L, F0, Fdot);
  FOR(i, 3)FOR(j, 3) D[i][j] = (L[i][j] + L[j][i])/2;
  FOR(i, 3)FOR(j, 3) W[i][j] = (L[i][j] - L[j][i])/2;
}

static W_bool_t  
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  int i, j, k, l, orient, n;

  double dt = ip_ctx->time_step;
  double time = ip_ctx->time;
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  W_bool_t result = TRUE;
  matstate_t input_matst;

  double gamma1[N_SLIP_SYSTEMS];
  double dgamma1[N_SLIP_SYSTEMS];
  double g1[N_SLIP_SYSTEMS];
  double accum_gamma;
  double deform_energy_density;
 
  double Fp1[3][3], Fp_lab[3][3];
  double F[3][3], F1[3][3];
  double Fold[3][3], Fold1[3][3];
  double W[3][3], W1[3][3];
  double D[3][3], D1[3][3];
  double S[6], Sv[6];
  
  double phi[3][3];
  double Ce[6][6];

  if (!ip_ctx->get_kinem_var) return FALSE;  

  /* kinematic quantities */
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_PREV_DEF_GRAD, Fold))
    return FALSE;
  
  get_rate_of_deformation(F, Fold, dt, D, W);
  
  /* save current material state */
  if (update_form == W_MAT_STATE_UPDATE_TEMP)  input_matst = *matst;
 
  get_elastic_moduli (u->c11, u->c12, u->c44, Ce);
 
  orient = 0;
  FOR(i, 6){S[i] = 0;}
  MAT3DUTIL_UNIT_3X3 (matst->average_Fp);
  FOR (i, N_SLIP_SYSTEMS) {
    matst->average_g[i] = 0;
    matst->average_tau[i] = 0;
    matst->average_gamma[i] = 0;
  }
  matst->deform_energy_density = 0;

  for(k = 0; k<u->layers; k++) {
    for(l = 0; l<=k; l++) {
       for(n = 0; n<3; n++) {
         /* set up the transformation matrix */
         get_phi (k, l, n, u->layers, u->rand, phi);
    
         /* transform to the lattice orientation */
         MAT3DUTIL_MULT_ATBA_3X3(F1, phi, F);
         MAT3DUTIL_MULT_ATBA_3X3(Fold1, phi, Fold);
         MAT3DUTIL_MULT_ATBA_3X3(W1, phi, W);
         MAT3DUTIL_MULT_ATBA_3X3(D1, phi, D);

         /* update */
         upd (u->elastic, u->rate_dependent, (update_form == W_MAT_STATE_UPDATE_FINAL),
           time, u->debug, u->max_inner_iter, u->activity_tol,
           ref_s, ref_m,  u->g_0, u->g_inf,
           u->q, u->h_0, u->gamma_0, u->m_srs,
           Ce, Fold1, F1, 
           matst->Fp[orient], matst->g[orient], 
           matst->gamma[orient], matst->accum_gamma[orient],
           dt, Sv, Fp1, g1, &accum_gamma, gamma1, dgamma1, &deform_energy_density,
           matst->tau[orient], matst->S[orient], 
           W1, D1);

 //        DBGP_DGAMMA("tau",matst->tau[orient]);
         /* save state */
         FOR (i, 6) { matst->S[orient][i] = Sv[i]; }

         MAT3DUTIL_2_MAT (matst->Fp[orient], =, Fp1);
         FOR (i, N_SLIP_SYSTEMS) {
           matst->g[orient][i] = g1[i];
           matst->gamma[orient][i] = gamma1[i];
           matst->dgamma[orient][i] = dgamma1[i];
         }
         matst->accum_gamma[orient] = accum_gamma;

         MAT3DUTIL_MULT_ABAT_3X3 (Fp_lab, phi, Fp1);
         MAT3DUTIL_ADD_AB_3X3(matst->average_Fp, matst->average_Fp, Fp_lab);
         matst->deform_energy_density += deform_energy_density;
         FOR(i, N_SLIP_SYSTEMS){
            matst->average_tau[i] += matst->tau[orient][i];
            matst->average_g[i] += g1[i];
            matst->average_gamma[i] += gamma1[i];
         }
         sum_over_orientations(S, Sv, phi);
         orient += 1; 
      } //n
    } //l 
  } //k

  /* average stress */
  FOR(i, 6) matst->average_stress[i] = S[i]/(double)orient;
  FOR(i, 3) {
     FOR(j, 3) {
        matst->average_Fp[i][j] = matst->average_Fp[i][j]/(double)orient;
     }
  }
  matst->deform_energy_density = matst->deform_energy_density/(double)orient;
  FOR(i, N_SLIP_SYSTEMS){
     matst->average_tau[i] = matst->average_tau[i]/(double)orient;
     matst->average_g[i] = matst->average_g[i]/(double)orient;
     matst->average_gamma[i] = matst->average_gamma[i]/(double)orient;
  }

//  DBGP_DGAMMA("average_tau",matst->average_tau);
  result = get_stress (m, ms, ip_ctx, stress_type, stress, out_stress_type);

 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}


/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress (W_mat_3d_t                  *m,
            W_mat_3d_state_p             ms,
            W_mat_3d_ip_ctx_t           *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double                       stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  matstate_t *matst = (matstate_t *)ms;
  double S[3][3];
  W_bool_t result = TRUE;

  MAT3DUTIL_V_TO_T (matst->average_stress, S);
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    {
      double F[3][3], P[3][3], invdetF, sigma[3][3];
      int i, j, k;
      if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
        result = FALSE; goto exit;
      }
      FOR(i, 3) {
        FOR(j, 3) {
          P[i][j] = 0; /* /F/ * /S/ */ 
          FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
        }
      }
      invdetF = 1 / MAT3DUTIL_DET_3X3(F);
      FOR(i, 3) {
        FOR(j, 3) {
          sigma[i][j] = 0; /* 1/detF /P/ * /F/^T */ 
          FOR(k, 3) sigma[i][j] += invdetF * P[i][k] * F[j][k];
        }
      }
      MAT3DUTIL_COPY_3X3(stress, sigma);
    }
    break;
  case MAT_3D_1ST_PK:
    {
      double F[3][3], P[3][3];
      int i, j, k;
      if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
        result = FALSE; goto exit;
      }
      FOR(i, 3) {
        FOR(j, 3) {
          P[i][j] = 0; /* /F/ * /S/ */ 
          FOR(k, 3) P[i][j] += F[i][k] * S[k][j];
        }
      }
      MAT3DUTIL_COPY_3X3(stress, P);
    }
    break;
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, S);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, S);
    *out_stress_type = MAT_3D_2ND_PK;
    break;
  }
    
 exit:
  return result;
}


static W_bool_t 
get_mat_state_var (W_mat_3d_t           *m,
                   W_mat_3d_state_p             ms,
                   W_mat_3d_ip_ctx_t           *ip_ctx,
                   W_mat_3d_mat_state_var_enum  var_type,
                   double                      *var
                   )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0; /* RAW */
    return TRUE;
  default:
    return FALSE;
  }
}

static W_bool_t 
get_mat_state_named_var (W_mat_3d_t *m,
                         W_mat_3d_state_p ms,
                         W_mat_3d_ip_ctx_t *ip_ctx,
                         char *name,
                         double *var
                         )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

  if        (tokens_token_and_keyword_equiv (name, "fp_xx")) {
    *var = matst->average_Fp[0][0];
  } else if (tokens_token_and_keyword_equiv (name, "fp_xy")) {
    *var = matst->average_Fp[0][1];
  } else if (tokens_token_and_keyword_equiv (name, "fp_xz")) {
    *var = matst->average_Fp[0][2];
  } else if (tokens_token_and_keyword_equiv (name, "fp_yx")) {
    *var = matst->average_Fp[1][0];
  } else if (tokens_token_and_keyword_equiv (name, "fp_yy")) {
    *var = matst->average_Fp[1][1];
  } else if (tokens_token_and_keyword_equiv (name, "fp_yz")) {
    *var = matst->average_Fp[1][2];
  } else if (tokens_token_and_keyword_equiv (name, "fp_zx")) {
    *var = matst->average_Fp[2][0];
  } else if (tokens_token_and_keyword_equiv (name, "fp_zy")) {
    *var = matst->average_Fp[2][1];
  } else if (tokens_token_and_keyword_equiv (name, "fp_zz")) {
    *var = matst->average_Fp[2][2];
  } else if (tokens_token_and_keyword_equiv (name, "ln_U_1") ||
             tokens_token_and_keyword_equiv (name, "ln_U_2") ||
             tokens_token_and_keyword_equiv (name, "ln_U_3") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U") 
             ) {
    double F[3][3], R[3][3], U[3][3], ln_U[3][3];
    double evals[3], evecs[3][3], lnevals[3];
    *var = 0;
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    if (W_polar_decomposition (F, R, U)) {
      W_eigendecomposition (U, evecs, evals);
      if (tokens_token_and_keyword_equiv (name, "ln_U_1")) *var = log(evals[0]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_2")) *var = log(evals[1]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_3")) *var = log(evals[2]);
      if (tokens_token_and_keyword_equiv (name, "max_ln_U")) {
        *var = fabs(log(evals[0]));
        if (fabs(log(evals[1])) > *var) *var = fabs(log(evals[1]));
        if (fabs(log(evals[2])) > *var) *var = fabs(log(evals[2]));
      }
      goto exit;
    }
  } else {
    int i;
    char buf[128];
    FOR (i, N_SLIP_SYSTEMS) {
      sprintf (buf, "gamma_%d", i+1);
      if (tokens_token_and_keyword_equiv (name, buf)) {
        *var = matst->average_gamma[i];
        goto exit;
      }
    }
    FOR (i, N_SLIP_SYSTEMS) {
      sprintf (buf, "g_%d", i+1);
      if (tokens_token_and_keyword_equiv (name, buf)) {
        *var = matst->average_g[i];
        goto exit;
      }
    }
    FOR (i, N_SLIP_SYSTEMS) {
      sprintf (buf, "tau_%d", i+1);
      if (tokens_token_and_keyword_equiv (name, buf)) {
        double tau[N_SLIP_SYSTEMS];
        double F[3][3], Cebar[3][3], S[3][3];
        if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
          return FALSE;
        }
        *var = matst->average_tau[i];
        goto exit;
      }
    }

    *var = 0;
    return FALSE;
  }
 exit:
  return TRUE;
}

static void
update_history (W_mat_3d_t *m,
                W_mat_3d_state_p ms,
                W_mat_3d_ip_ctx_t *ip_ctx,
                W_solution_history_t *h, /* solution history */
                W_index_t pntidx,
                W_float_t time
                )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
#define STORE(HW, NAME, TIME, I, VAL)           \
  {                                             \
    char buf[128];                              \
    sprintf (buf, "p%d%s_%d", pntidx, NAME, I); \
    WHR_store (HW, buf, TIME, VAL);             \
  }
  int i;
  double tau[N_SLIP_SYSTEMS];
  double F[3][3], Cebar[3][3], S[3][3];

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
    return;
  }
  FOR (i, N_SLIP_SYSTEMS) {
    STORE (h->writer, "g", time, i+1, matst->average_g[i]);
    STORE (h->writer, "gamma", time, i+1, matst->average_gamma[i]);
    STORE (h->writer, "tau", time, i+1, matst->average_tau[i]);
  }

  /* Mises effective stress-effective strain curves */
  { 
    char buf[128];                              

    double R[3][3], U[3][3], ln_U[3][3];
    double Ee, Se;

    W_polar_decomposition (F, R, U);
    W_ln_U (U, ln_U);
    Ee = 2.0 * fabs(ln_U[2][2])/sqrt(3.0);
    Se = sqrt(3) * fabs(matst->average_stress[2])/2.0;

    sprintf (buf, "p%d%s", pntidx, "mises"); 
    WHR_store (h->writer, buf, Ee, Se);             
  }
}

static W_mat_3d_state_p 
interpolate_mat_state(W_mat_3d_t *m,
		      W_loc_t *at_point,
		      int npkgs,
		      WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    FOR(i, 6) {
      ms->average_stress[i]      += pms->average_stress[i] * WGHT(k);
    }
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
#undef WGHT
}


static W_bool_t
get_mat_stiff_elastic (W_mat_3d_t *m/* mat3d */,
                       W_mat_3d_state_p ms,/* material state */
                       W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                                    used to generate kinematic
                                                    quantities which the material state
                                                    might need in order to compute
                                                    the requested stress */
                       W_mat_3d_stiff_enum stiff_type, /* stiffness type */
                       double D[6][6] /* material stiffness tensor in matrix form */
                       )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  double c11 = u->c11, c12 = u->c12, c44 = u->c44;
  double t0, s1, s2, Dbar[6][6];
  double invFp[3][3], detFp;
  double iFp00, iFp01, iFp02;
  double iFp10, iFp11, iFp12;
  double iFp20, iFp21, iFp22;
  double Dbar11, Dbar12, Dbar13, Dbar14, Dbar15, Dbar16, Dbar22, Dbar23,
    Dbar24, Dbar25, Dbar26, Dbar33, Dbar34, Dbar35, Dbar36, Dbar44,
    Dbar45, Dbar46, Dbar55, Dbar56, Dbar66;
  
  detFp = MAT3DUTIL_DET_3X3 (matst->average_Fp);
  if (detFp == 0) {
    REPORT_FATAL_ERROR (NULL, "Zero det(Fp)");
  }
  MAT3DUTIL_INVERT_3X3 (detFp, matst->average_Fp, invFp);
  iFp00 = invFp[0][0];    iFp01 = invFp[0][1];    iFp02 = invFp[0][2];
  iFp10 = invFp[1][0];    iFp11 = invFp[1][1];    iFp12 = invFp[1][2];
  iFp20 = invFp[2][0];    iFp21 = invFp[2][1];    iFp22 = invFp[2][2];
  
  /* Calculate Dbar such that Sbar = Dbar:Ebar */
//  calc_Dbar (c11, c12, c44, &u->a, &u->b, &u->c, Dbar);
  get_elastic_moduli (c11, c12, c44, Dbar);
  Dbar11=Dbar[0][0];
  Dbar12=Dbar[0][1];
  Dbar13=Dbar[0][2];
  Dbar14=Dbar[0][3];
  Dbar15=Dbar[0][4];
  Dbar16=Dbar[0][5];
  Dbar22=Dbar[1][1];
  Dbar23=Dbar[1][2];
  Dbar24=Dbar[1][3];
  Dbar25=Dbar[1][4];
  Dbar26=Dbar[1][5];
  Dbar33=Dbar[2][2];
  Dbar34=Dbar[2][3];
  Dbar35=Dbar[2][4];
  Dbar36=Dbar[2][5];
  Dbar44=Dbar[3][3];
  Dbar45=Dbar[3][4];
  Dbar46=Dbar[3][5];
  Dbar55=Dbar[4][4];
  Dbar56=Dbar[4][5];
  Dbar66=Dbar[5][5];

  /* Now pull back by Fp */;
  t0 = Dbar11*iFp00*iFp00*iFp00*iFp00+4.0*Dbar14*iFp00*iFp00*iFp00*iFp01+4.0*Dbar15*iFp00*iFp00*iFp00*iFp02+2.0*Dbar12*iFp00*iFp00*iFp01*iFp01+4.0*Dbar16*iFp00*iFp00*iFp01*iFp02+2.0*Dbar13*iFp00*iFp00*iFp02*iFp02+4.0*Dbar44*iFp00*iFp00*iFp01*iFp01+8.0*Dbar45*iFp00*iFp00*iFp01*iFp02+4.0*Dbar24*iFp00*iFp01*iFp01*iFp01+8.0*Dbar46*iFp00*iFp01*iFp01*iFp02+4.0*Dbar34*iFp00*iFp01*iFp02*iFp02+4.0*Dbar55*iFp02*iFp02*iFp00*iFp00+4.0*Dbar25*iFp02*iFp00*iFp01*iFp01+8.0*Dbar56*iFp02*iFp02*iFp00*iFp01+4.0*Dbar35*iFp02*iFp02*iFp02*iFp00+Dbar22*iFp01*iFp01*iFp01*iFp01+4.0*Dbar26*iFp01*iFp01*iFp01*iFp02+2.0*Dbar23*iFp01*iFp01*iFp02*iFp02+4.0*Dbar66*iFp02*iFp02*iFp01*iFp01+4.0*Dbar36*iFp01*iFp02*iFp02*iFp02+Dbar33*iFp02*iFp02*iFp02*iFp02;
  D[0][0] = t0;
  s1 = 2.0*Dbar25*iFp02*iFp00*iFp11*iFp11+2.0*Dbar34*iFp00*iFp01*iFp12*iFp12+2.0*Dbar25*iFp01*iFp01*iFp10*iFp12+2.0*Dbar16*iFp02*iFp01*iFp10*iFp10+2.0*Dbar16*iFp00*iFp00*iFp11*iFp12+2.0*Dbar34*iFp02*iFp02*iFp10*iFp11+Dbar13*iFp00*iFp00*iFp12*iFp12+2.0*Dbar36*iFp02*iFp02*iFp11*iFp12+2.0*Dbar35*iFp02*iFp02*iFp10*iFp12+Dbar12*iFp00*iFp00*iFp11*iFp11+2.0*Dbar24*iFp01*iFp01*iFp10*iFp11+2.0*Dbar35*iFp00*iFp12*iFp12*iFp02+2.0*Dbar36*iFp01*iFp12*iFp12*iFp02+2.0*Dbar24*iFp00*iFp11*iFp11*iFp01+2.0*Dbar15*iFp00*iFp00*iFp10*iFp12+2.0*Dbar14*iFp00*iFp10*iFp10*iFp01+2.0*Dbar15*iFp00*iFp10*iFp10*iFp02+2.0*Dbar14*iFp00*iFp00*iFp10*iFp11;
  t0 = s1+2.0*Dbar26*iFp01*iFp11*iFp11*iFp02+2.0*Dbar26*iFp01*iFp01*iFp11*iFp12+4.0*Dbar66*iFp01*iFp12*iFp11*iFp02+Dbar13*iFp02*iFp02*iFp10*iFp10+Dbar23*iFp01*iFp01*iFp12*iFp12+Dbar23*iFp02*iFp02*iFp11*iFp11+Dbar11*iFp00*iFp00*iFp10*iFp10+4.0*Dbar46*iFp01*iFp10*iFp11*iFp02+Dbar12*iFp01*iFp01*iFp10*iFp10+4.0*Dbar44*iFp00*iFp11*iFp10*iFp01+4.0*Dbar56*iFp01*iFp12*iFp10*iFp02+Dbar22*iFp01*iFp01*iFp11*iFp11+4.0*Dbar55*iFp00*iFp12*iFp10*iFp02+4.0*Dbar45*iFp00*iFp12*iFp10*iFp01+4.0*Dbar56*iFp00*iFp12*iFp11*iFp02+4.0*Dbar45*iFp00*iFp11*iFp10*iFp02+4.0*Dbar46*iFp00*iFp11*iFp12*iFp01+Dbar33*iFp02*iFp02*iFp12*iFp12;
  D[0][1] = D[1][0] = t0;
  s1 = 2.0*Dbar16*iFp01*iFp02*iFp20*iFp20+2.0*Dbar34*iFp00*iFp01*iFp22*iFp22+2.0*Dbar26*iFp01*iFp01*iFp21*iFp22+2.0*Dbar35*iFp00*iFp02*iFp22*iFp22+2.0*Dbar25*iFp01*iFp01*iFp20*iFp22+2.0*Dbar24*iFp01*iFp01*iFp20*iFp21+Dbar11*iFp00*iFp00*iFp20*iFp20+Dbar12*iFp00*iFp00*iFp21*iFp21+Dbar33*iFp02*iFp02*iFp22*iFp22+Dbar23*iFp02*iFp02*iFp21*iFp21+Dbar13*iFp02*iFp02*iFp20*iFp20+Dbar23*iFp01*iFp01*iFp22*iFp22+2.0*Dbar34*iFp02*iFp02*iFp20*iFp21+2.0*Dbar35*iFp02*iFp02*iFp20*iFp22+2.0*Dbar14*iFp00*iFp00*iFp20*iFp21+2.0*Dbar15*iFp00*iFp00*iFp20*iFp22+2.0*Dbar16*iFp00*iFp00*iFp21*iFp22+Dbar22*iFp01*iFp01*iFp21*iFp21;
  t0 = s1+Dbar12*iFp01*iFp01*iFp20*iFp20+2.0*Dbar25*iFp00*iFp02*iFp21*iFp21+Dbar13*iFp00*iFp00*iFp22*iFp22+2.0*Dbar36*iFp01*iFp02*iFp22*iFp22+2.0*Dbar26*iFp01*iFp02*iFp21*iFp21+2.0*Dbar36*iFp02*iFp02*iFp21*iFp22+2.0*Dbar24*iFp00*iFp01*iFp21*iFp21+2.0*Dbar14*iFp00*iFp01*iFp20*iFp20+4.0*Dbar45*iFp00*iFp02*iFp20*iFp21+4.0*Dbar46*iFp01*iFp02*iFp20*iFp21+4.0*Dbar55*iFp00*iFp02*iFp20*iFp22+4.0*Dbar66*iFp01*iFp02*iFp21*iFp22+4.0*Dbar56*iFp00*iFp02*iFp21*iFp22+2.0*Dbar15*iFp00*iFp02*iFp20*iFp20+4.0*Dbar45*iFp00*iFp01*iFp20*iFp22+4.0*Dbar44*iFp00*iFp01*iFp20*iFp21+4.0*Dbar46*iFp00*iFp01*iFp21*iFp22+4.0*Dbar56*iFp01*iFp02*iFp20*iFp22;
  D[0][2] = D[2][0] = t0;
  s1 = Dbar35*iFp02*iFp02*iFp02*iFp10+Dbar11*iFp00*iFp00*iFp00*iFp10+Dbar14*iFp00*iFp00*iFp00*iFp11+Dbar15*iFp00*iFp00*iFp00*iFp12+Dbar22*iFp01*iFp01*iFp01*iFp11+Dbar26*iFp01*iFp01*iFp01*iFp12+Dbar24*iFp01*iFp01*iFp01*iFp10+Dbar33*iFp02*iFp02*iFp02*iFp12+Dbar36*iFp02*iFp02*iFp02*iFp11+2.0*Dbar55*iFp02*iFp00*iFp00*iFp12+2.0*Dbar44*iFp00*iFp00*iFp01*iFp11+3.0*Dbar24*iFp00*iFp01*iFp01*iFp11+2.0*Dbar45*iFp00*iFp00*iFp01*iFp12+2.0*Dbar46*iFp00*iFp01*iFp01*iFp12+2.0*Dbar55*iFp02*iFp02*iFp00*iFp10+3.0*Dbar14*iFp00*iFp00*iFp10*iFp01+3.0*Dbar15*iFp00*iFp00*iFp10*iFp02+2.0*Dbar44*iFp00*iFp01*iFp01*iFp10+3.0*Dbar36*iFp01*iFp02*iFp02*iFp12+2.0*Dbar66*iFp01*iFp01*iFp02*iFp12+Dbar23*iFp02*iFp02*iFp11*iFp01+Dbar34*iFp02*iFp02*iFp10*iFp01;
  s2 = s1+Dbar16*iFp00*iFp00*iFp11*iFp02+Dbar12*iFp00*iFp00*iFp11*iFp01+Dbar13*iFp00*iFp00*iFp12*iFp02+Dbar13*iFp02*iFp02*iFp10*iFp00+Dbar16*iFp00*iFp00*iFp12*iFp01+Dbar25*iFp01*iFp01*iFp12*iFp00+Dbar12*iFp01*iFp01*iFp10*iFp00+Dbar25*iFp01*iFp01*iFp10*iFp02+Dbar23*iFp01*iFp01*iFp12*iFp02+Dbar34*iFp02*iFp02*iFp11*iFp00+2.0*Dbar56*iFp01*iFp02*iFp02*iFp10;
  t0 = s2+2.0*Dbar66*iFp01*iFp02*iFp02*iFp11+3.0*Dbar26*iFp01*iFp01*iFp11*iFp02+2.0*Dbar46*iFp01*iFp01*iFp02*iFp10+3.0*Dbar35*iFp02*iFp02*iFp00*iFp12+2.0*Dbar45*iFp02*iFp00*iFp00*iFp11+2.0*Dbar56*iFp02*iFp02*iFp00*iFp11+2.0*Dbar16*iFp01*iFp02*iFp10*iFp00+2.0*Dbar25*iFp02*iFp00*iFp11*iFp01+4.0*Dbar45*iFp00*iFp01*iFp10*iFp02+4.0*Dbar46*iFp00*iFp01*iFp11*iFp02+2.0*Dbar34*iFp01*iFp00*iFp12*iFp02+4.0*Dbar56*iFp02*iFp00*iFp12*iFp01;
  D[0][3] = D[3][0] = t0;
  s1 = 2.0*Dbar66*iFp01*iFp01*iFp02*iFp22+2.0*Dbar55*iFp00*iFp02*iFp02*iFp20+3.0*Dbar36*iFp01*iFp02*iFp02*iFp22+2.0*Dbar45*iFp00*iFp00*iFp02*iFp21+2.0*Dbar44*iFp00*iFp00*iFp01*iFp21+3.0*Dbar24*iFp00*iFp01*iFp01*iFp21+2.0*Dbar66*iFp01*iFp02*iFp02*iFp21+Dbar12*iFp00*iFp00*iFp21*iFp01+Dbar16*iFp00*iFp00*iFp22*iFp01+Dbar16*iFp00*iFp00*iFp21*iFp02+Dbar13*iFp00*iFp00*iFp22*iFp02+Dbar12*iFp01*iFp01*iFp20*iFp00+Dbar25*iFp01*iFp01*iFp20*iFp02+Dbar23*iFp01*iFp01*iFp22*iFp02+Dbar22*iFp01*iFp01*iFp01*iFp21+Dbar34*iFp02*iFp02*iFp20*iFp01+Dbar25*iFp01*iFp01*iFp22*iFp00+Dbar34*iFp02*iFp02*iFp21*iFp00+Dbar23*iFp02*iFp02*iFp21*iFp01+Dbar13*iFp02*iFp02*iFp20*iFp00+2.0*Dbar55*iFp00*iFp00*iFp02*iFp22+3.0*Dbar14*iFp00*iFp00*iFp20*iFp01;
  t0 = s1+3.0*Dbar26*iFp01*iFp01*iFp21*iFp02+3.0*Dbar35*iFp00*iFp02*iFp02*iFp22+2.0*Dbar56*iFp00*iFp02*iFp02*iFp21+2.0*Dbar45*iFp00*iFp00*iFp01*iFp22+2.0*Dbar46*iFp00*iFp01*iFp01*iFp22+2.0*Dbar44*iFp00*iFp01*iFp01*iFp20+2.0*Dbar56*iFp01*iFp02*iFp02*iFp20+Dbar35*iFp02*iFp02*iFp02*iFp20+Dbar33*iFp02*iFp02*iFp02*iFp22+2.0*Dbar46*iFp01*iFp01*iFp02*iFp20+Dbar36*iFp02*iFp02*iFp02*iFp21+3.0*Dbar15*iFp00*iFp00*iFp20*iFp02+Dbar26*iFp01*iFp01*iFp01*iFp22+Dbar11*iFp00*iFp00*iFp00*iFp20+Dbar24*iFp01*iFp01*iFp01*iFp20+2.0*Dbar34*iFp00*iFp01*iFp22*iFp02+4.0*Dbar45*iFp00*iFp01*iFp20*iFp02+2.0*Dbar25*iFp00*iFp02*iFp21*iFp01+2.0*Dbar16*iFp01*iFp02*iFp20*iFp00+Dbar15*iFp00*iFp00*iFp00*iFp22+4.0*Dbar46*iFp00*iFp01*iFp21*iFp02+4.0*Dbar56*iFp00*iFp02*iFp22*iFp01+Dbar14*iFp00*iFp00*iFp00*iFp21;
  D[0][4] = D[4][0] = t0;
  s2 = Dbar12*iFp00*iFp00*iFp21*iFp11+Dbar15*iFp00*iFp00*iFp20*iFp12+Dbar16*iFp00*iFp00*iFp21*iFp12+Dbar16*iFp00*iFp00*iFp22*iFp11+Dbar14*iFp00*iFp00*iFp21*iFp10+Dbar11*iFp00*iFp00*iFp20*iFp10+Dbar14*iFp00*iFp00*iFp20*iFp11+Dbar25*iFp01*iFp01*iFp20*iFp12+Dbar24*iFp01*iFp01*iFp21*iFp10+Dbar12*iFp01*iFp01*iFp20*iFp10+Dbar15*iFp00*iFp00*iFp22*iFp10+Dbar13*iFp00*iFp00*iFp22*iFp12+Dbar23*iFp01*iFp01*iFp22*iFp12;
  s1 = s2+Dbar25*iFp01*iFp01*iFp22*iFp10+Dbar26*iFp01*iFp01*iFp22*iFp11+Dbar22*iFp01*iFp01*iFp21*iFp11+Dbar26*iFp01*iFp01*iFp21*iFp12+Dbar24*iFp01*iFp01*iFp20*iFp11+Dbar35*iFp02*iFp02*iFp20*iFp12+Dbar13*iFp02*iFp02*iFp20*iFp10+Dbar34*iFp02*iFp02*iFp20*iFp11+Dbar34*iFp02*iFp02*iFp21*iFp10+Dbar36*iFp02*iFp02*iFp22*iFp11+Dbar33*iFp02*iFp02*iFp22*iFp12+Dbar23*iFp02*iFp02*iFp21*iFp11+Dbar36*iFp02*iFp02*iFp21*iFp12+Dbar35*iFp02*iFp02*iFp22*iFp10;
  s2 = s1+2.0*Dbar36*iFp01*iFp02*iFp22*iFp12+2.0*Dbar56*iFp00*iFp02*iFp22*iFp11+2.0*Dbar44*iFp00*iFp01*iFp21*iFp10+2.0*Dbar24*iFp00*iFp01*iFp21*iFp11+2.0*Dbar45*iFp00*iFp02*iFp21*iFp10+2.0*Dbar44*iFp00*iFp01*iFp20*iFp11+2.0*Dbar25*iFp00*iFp02*iFp21*iFp11+2.0*Dbar66*iFp01*iFp02*iFp22*iFp11+2.0*Dbar15*iFp00*iFp02*iFp20*iFp10+2.0*Dbar55*iFp00*iFp02*iFp20*iFp12+2.0*Dbar45*iFp00*iFp02*iFp20*iFp11+2.0*Dbar56*iFp01*iFp02*iFp22*iFp10+2.0*Dbar46*iFp00*iFp01*iFp22*iFp11;
  t0 = s2+2.0*Dbar66*iFp01*iFp02*iFp21*iFp12+2.0*Dbar46*iFp01*iFp02*iFp21*iFp10+2.0*Dbar34*iFp00*iFp01*iFp22*iFp12+2.0*Dbar14*iFp00*iFp01*iFp20*iFp10+2.0*Dbar45*iFp00*iFp01*iFp20*iFp12+2.0*Dbar56*iFp00*iFp02*iFp21*iFp12+2.0*Dbar55*iFp00*iFp02*iFp22*iFp10+2.0*Dbar16*iFp01*iFp02*iFp20*iFp10+2.0*Dbar35*iFp00*iFp02*iFp22*iFp12+2.0*Dbar46*iFp00*iFp01*iFp21*iFp12+2.0*Dbar45*iFp00*iFp01*iFp22*iFp10+2.0*Dbar56*iFp01*iFp02*iFp20*iFp12+2.0*Dbar46*iFp01*iFp02*iFp20*iFp11+2.0*Dbar26*iFp01*iFp02*iFp21*iFp11;
  D[0][5] = D[5][0] = t0;
  t0 = Dbar11*iFp10*iFp10*iFp10*iFp10+4.0*Dbar14*iFp10*iFp10*iFp10*iFp11+4.0*Dbar15*iFp10*iFp10*iFp10*iFp12+2.0*Dbar12*iFp10*iFp10*iFp11*iFp11+4.0*Dbar16*iFp10*iFp10*iFp11*iFp12+2.0*Dbar13*iFp10*iFp10*iFp12*iFp12+4.0*Dbar44*iFp11*iFp11*iFp10*iFp10+8.0*Dbar45*iFp10*iFp10*iFp11*iFp12+4.0*Dbar24*iFp11*iFp11*iFp11*iFp10+8.0*Dbar46*iFp10*iFp11*iFp11*iFp12+4.0*Dbar34*iFp10*iFp11*iFp12*iFp12+4.0*Dbar55*iFp10*iFp10*iFp12*iFp12+4.0*Dbar25*iFp10*iFp12*iFp11*iFp11+8.0*Dbar56*iFp10*iFp12*iFp12*iFp11+4.0*Dbar35*iFp10*iFp12*iFp12*iFp12+Dbar22*iFp11*iFp11*iFp11*iFp11+4.0*Dbar26*iFp11*iFp11*iFp11*iFp12+2.0*Dbar23*iFp11*iFp11*iFp12*iFp12+4.0*Dbar66*iFp11*iFp11*iFp12*iFp12+4.0*Dbar36*iFp11*iFp12*iFp12*iFp12+Dbar33*iFp12*iFp12*iFp12*iFp12;
  D[1][1] = t0;
  s1 = Dbar12*iFp10*iFp10*iFp21*iFp21+Dbar13*iFp10*iFp10*iFp22*iFp22+2.0*Dbar16*iFp10*iFp10*iFp21*iFp22+Dbar22*iFp11*iFp11*iFp21*iFp21+2.0*Dbar14*iFp10*iFp11*iFp20*iFp20+2.0*Dbar24*iFp10*iFp11*iFp21*iFp21+2.0*Dbar15*iFp10*iFp12*iFp20*iFp20+2.0*Dbar34*iFp10*iFp11*iFp22*iFp22+Dbar11*iFp10*iFp10*iFp20*iFp20+2.0*Dbar15*iFp10*iFp10*iFp20*iFp22+2.0*Dbar14*iFp10*iFp10*iFp20*iFp21+2.0*Dbar25*iFp11*iFp11*iFp20*iFp22+Dbar33*iFp12*iFp12*iFp22*iFp22+Dbar23*iFp12*iFp12*iFp21*iFp21+Dbar12*iFp11*iFp11*iFp20*iFp20+Dbar13*iFp12*iFp12*iFp20*iFp20+2.0*Dbar36*iFp12*iFp12*iFp21*iFp22+2.0*Dbar24*iFp11*iFp11*iFp20*iFp21;
  t0 = s1+2.0*Dbar36*iFp11*iFp12*iFp22*iFp22+2.0*Dbar26*iFp11*iFp12*iFp21*iFp21+2.0*Dbar34*iFp12*iFp12*iFp20*iFp21+2.0*Dbar35*iFp12*iFp12*iFp20*iFp22+2.0*Dbar26*iFp11*iFp11*iFp21*iFp22+2.0*Dbar35*iFp10*iFp12*iFp22*iFp22+2.0*Dbar16*iFp11*iFp12*iFp20*iFp20+4.0*Dbar46*iFp11*iFp12*iFp20*iFp21+4.0*Dbar56*iFp11*iFp12*iFp20*iFp22+4.0*Dbar66*iFp11*iFp12*iFp21*iFp22+2.0*Dbar25*iFp10*iFp12*iFp21*iFp21+4.0*Dbar44*iFp10*iFp11*iFp20*iFp21+4.0*Dbar56*iFp10*iFp12*iFp21*iFp22+4.0*Dbar55*iFp10*iFp12*iFp20*iFp22+4.0*Dbar45*iFp10*iFp12*iFp20*iFp21+4.0*Dbar46*iFp10*iFp11*iFp21*iFp22+4.0*Dbar45*iFp10*iFp11*iFp20*iFp22+Dbar23*iFp11*iFp11*iFp22*iFp22;
  D[1][2] = D[2][1] = t0;
  s1 = Dbar12*iFp10*iFp10*iFp11*iFp01+Dbar13*iFp10*iFp10*iFp12*iFp02+Dbar25*iFp11*iFp11*iFp12*iFp00+Dbar25*iFp11*iFp11*iFp10*iFp02+Dbar23*iFp11*iFp11*iFp12*iFp02+Dbar34*iFp12*iFp12*iFp11*iFp00+Dbar23*iFp12*iFp12*iFp11*iFp01+Dbar12*iFp11*iFp11*iFp10*iFp00+Dbar16*iFp10*iFp10*iFp12*iFp01+Dbar36*iFp12*iFp12*iFp12*iFp01+Dbar35*iFp12*iFp12*iFp12*iFp00+Dbar33*iFp12*iFp12*iFp12*iFp02+2.0*Dbar45*iFp10*iFp10*iFp11*iFp02+Dbar26*iFp11*iFp11*iFp11*iFp02+Dbar24*iFp11*iFp11*iFp11*iFp00+Dbar22*iFp11*iFp11*iFp11*iFp01+Dbar11*iFp10*iFp10*iFp10*iFp00+Dbar14*iFp10*iFp10*iFp10*iFp01+Dbar15*iFp10*iFp10*iFp10*iFp02+2.0*Dbar44*iFp10*iFp11*iFp11*iFp00+3.0*Dbar24*iFp10*iFp11*iFp11*iFp01+2.0*Dbar46*iFp10*iFp11*iFp11*iFp02;
  s2 = s1+3.0*Dbar35*iFp10*iFp12*iFp12*iFp02+2.0*Dbar56*iFp10*iFp12*iFp12*iFp01+2.0*Dbar66*iFp11*iFp12*iFp12*iFp01+2.0*Dbar55*iFp10*iFp12*iFp12*iFp00+3.0*Dbar36*iFp11*iFp12*iFp12*iFp02+Dbar16*iFp10*iFp10*iFp11*iFp02+Dbar34*iFp12*iFp12*iFp10*iFp01+Dbar13*iFp12*iFp12*iFp10*iFp00+2.0*Dbar56*iFp11*iFp12*iFp12*iFp00+2.0*Dbar46*iFp11*iFp11*iFp12*iFp00+3.0*Dbar26*iFp11*iFp11*iFp12*iFp01;
  t0 = s2+2.0*Dbar55*iFp10*iFp10*iFp12*iFp02+2.0*Dbar45*iFp10*iFp10*iFp12*iFp01+3.0*Dbar15*iFp10*iFp10*iFp12*iFp00+3.0*Dbar14*iFp10*iFp10*iFp11*iFp00+2.0*Dbar44*iFp10*iFp10*iFp11*iFp01+4.0*Dbar45*iFp10*iFp11*iFp12*iFp00+4.0*Dbar46*iFp10*iFp11*iFp12*iFp01+4.0*Dbar56*iFp10*iFp12*iFp11*iFp02+2.0*Dbar34*iFp10*iFp11*iFp12*iFp02+2.0*Dbar25*iFp10*iFp12*iFp11*iFp01+2.0*Dbar16*iFp11*iFp12*iFp10*iFp00+2.0*Dbar66*iFp11*iFp11*iFp12*iFp02;
  D[1][3] = D[3][1] = t0;
  s2 = Dbar35*iFp12*iFp12*iFp22*iFp00+Dbar26*iFp11*iFp11*iFp21*iFp02+Dbar23*iFp11*iFp11*iFp22*iFp02+Dbar26*iFp11*iFp11*iFp22*iFp01+Dbar13*iFp12*iFp12*iFp20*iFp00+Dbar34*iFp12*iFp12*iFp21*iFp00+Dbar14*iFp10*iFp10*iFp20*iFp01+Dbar36*iFp12*iFp12*iFp21*iFp02+Dbar25*iFp11*iFp11*iFp20*iFp02+Dbar23*iFp12*iFp12*iFp21*iFp01+Dbar15*iFp10*iFp10*iFp20*iFp02+Dbar12*iFp10*iFp10*iFp21*iFp01+Dbar14*iFp10*iFp10*iFp21*iFp00;
  s1 = s2+Dbar16*iFp10*iFp10*iFp21*iFp02+Dbar12*iFp11*iFp11*iFp20*iFp00+Dbar35*iFp12*iFp12*iFp20*iFp02+Dbar22*iFp11*iFp11*iFp21*iFp01+Dbar33*iFp12*iFp12*iFp22*iFp02+Dbar36*iFp12*iFp12*iFp22*iFp01+Dbar11*iFp10*iFp10*iFp20*iFp00+Dbar15*iFp10*iFp10*iFp22*iFp00+Dbar25*iFp11*iFp11*iFp22*iFp00+Dbar24*iFp11*iFp11*iFp20*iFp01+Dbar13*iFp10*iFp10*iFp22*iFp02+Dbar16*iFp10*iFp10*iFp22*iFp01+Dbar24*iFp11*iFp11*iFp21*iFp00+Dbar34*iFp12*iFp12*iFp20*iFp01;
  s2 = s1+2.0*Dbar45*iFp10*iFp11*iFp22*iFp00+2.0*Dbar66*iFp11*iFp12*iFp21*iFp02+2.0*Dbar14*iFp10*iFp11*iFp20*iFp00+2.0*Dbar34*iFp10*iFp11*iFp22*iFp02+2.0*Dbar15*iFp10*iFp12*iFp20*iFp00+2.0*Dbar55*iFp10*iFp12*iFp20*iFp02+2.0*Dbar45*iFp10*iFp12*iFp21*iFp00+2.0*Dbar56*iFp10*iFp12*iFp21*iFp02+2.0*Dbar56*iFp11*iFp12*iFp22*iFp00+2.0*Dbar36*iFp11*iFp12*iFp22*iFp02+2.0*Dbar25*iFp10*iFp12*iFp21*iFp01+2.0*Dbar66*iFp11*iFp12*iFp22*iFp01+2.0*Dbar35*iFp10*iFp12*iFp22*iFp02;
  t0 = s2+2.0*Dbar45*iFp10*iFp12*iFp20*iFp01+2.0*Dbar56*iFp10*iFp12*iFp22*iFp01+2.0*Dbar55*iFp10*iFp12*iFp22*iFp00+2.0*Dbar44*iFp10*iFp11*iFp20*iFp01+2.0*Dbar45*iFp10*iFp11*iFp20*iFp02+2.0*Dbar16*iFp11*iFp12*iFp20*iFp00+2.0*Dbar46*iFp11*iFp12*iFp21*iFp00+2.0*Dbar44*iFp10*iFp11*iFp21*iFp00+2.0*Dbar46*iFp11*iFp12*iFp20*iFp01+2.0*Dbar56*iFp11*iFp12*iFp20*iFp02+2.0*Dbar26*iFp11*iFp12*iFp21*iFp01+2.0*Dbar24*iFp10*iFp11*iFp21*iFp01+2.0*Dbar46*iFp10*iFp11*iFp22*iFp01+2.0*Dbar46*iFp10*iFp11*iFp21*iFp02;
  D[1][4] = D[4][1] = t0;
  s1 = Dbar22*iFp11*iFp11*iFp11*iFp21+Dbar24*iFp11*iFp11*iFp11*iFp20+Dbar35*iFp12*iFp12*iFp12*iFp20+Dbar15*iFp10*iFp10*iFp10*iFp22+Dbar14*iFp10*iFp10*iFp10*iFp21+Dbar33*iFp12*iFp12*iFp12*iFp22+Dbar11*iFp10*iFp10*iFp10*iFp20+Dbar36*iFp12*iFp12*iFp12*iFp21+Dbar26*iFp11*iFp11*iFp11*iFp22+Dbar16*iFp10*iFp10*iFp21*iFp12+Dbar13*iFp10*iFp10*iFp22*iFp12+Dbar12*iFp10*iFp10*iFp21*iFp11+Dbar34*iFp12*iFp12*iFp21*iFp10+Dbar34*iFp12*iFp12*iFp20*iFp11+3.0*Dbar24*iFp10*iFp11*iFp11*iFp21+2.0*Dbar46*iFp11*iFp11*iFp12*iFp20+3.0*Dbar26*iFp11*iFp11*iFp21*iFp12+2.0*Dbar44*iFp10*iFp10*iFp11*iFp21+2.0*Dbar44*iFp10*iFp11*iFp11*iFp20+3.0*Dbar35*iFp10*iFp12*iFp12*iFp22+2.0*Dbar56*iFp10*iFp12*iFp12*iFp21+2.0*Dbar55*iFp10*iFp10*iFp12*iFp22;
  s2 = s1+2.0*Dbar45*iFp10*iFp10*iFp12*iFp21+Dbar23*iFp12*iFp12*iFp21*iFp11+Dbar25*iFp11*iFp11*iFp20*iFp12+Dbar25*iFp11*iFp11*iFp22*iFp10+Dbar13*iFp12*iFp12*iFp20*iFp10+Dbar23*iFp11*iFp11*iFp22*iFp12+Dbar16*iFp10*iFp10*iFp22*iFp11+Dbar12*iFp11*iFp11*iFp20*iFp10+2.0*Dbar55*iFp10*iFp12*iFp12*iFp20+2.0*Dbar66*iFp11*iFp11*iFp12*iFp22+3.0*Dbar36*iFp11*iFp12*iFp12*iFp22;
  t0 = s2+3.0*Dbar15*iFp10*iFp10*iFp20*iFp12+2.0*Dbar45*iFp10*iFp10*iFp11*iFp22+2.0*Dbar66*iFp11*iFp12*iFp12*iFp21+3.0*Dbar14*iFp10*iFp10*iFp20*iFp11+2.0*Dbar46*iFp10*iFp11*iFp11*iFp22+2.0*Dbar56*iFp11*iFp12*iFp12*iFp20+2.0*Dbar16*iFp11*iFp12*iFp20*iFp10+4.0*Dbar46*iFp10*iFp11*iFp21*iFp12+2.0*Dbar34*iFp10*iFp11*iFp22*iFp12+4.0*Dbar56*iFp10*iFp12*iFp22*iFp11+4.0*Dbar45*iFp10*iFp11*iFp20*iFp12+2.0*Dbar25*iFp10*iFp12*iFp21*iFp11;
  D[1][5] = D[5][1] = t0;
  t0 = Dbar11*iFp20*iFp20*iFp20*iFp20+4.0*Dbar14*iFp20*iFp20*iFp20*iFp21+4.0*Dbar15*iFp20*iFp20*iFp20*iFp22+2.0*Dbar12*iFp20*iFp20*iFp21*iFp21+4.0*Dbar16*iFp20*iFp20*iFp21*iFp22+2.0*Dbar13*iFp20*iFp20*iFp22*iFp22+4.0*Dbar44*iFp20*iFp20*iFp21*iFp21+8.0*Dbar45*iFp20*iFp20*iFp21*iFp22+4.0*Dbar24*iFp20*iFp21*iFp21*iFp21+8.0*Dbar46*iFp20*iFp21*iFp21*iFp22+4.0*Dbar34*iFp20*iFp21*iFp22*iFp22+4.0*Dbar55*iFp20*iFp20*iFp22*iFp22+4.0*Dbar25*iFp20*iFp22*iFp21*iFp21+8.0*Dbar56*iFp20*iFp22*iFp22*iFp21+4.0*Dbar35*iFp20*iFp22*iFp22*iFp22+Dbar22*iFp21*iFp21*iFp21*iFp21+4.0*Dbar26*iFp21*iFp21*iFp21*iFp22+2.0*Dbar23*iFp21*iFp21*iFp22*iFp22+4.0*Dbar66*iFp21*iFp21*iFp22*iFp22+4.0*Dbar36*iFp22*iFp22*iFp22*iFp21+Dbar33*iFp22*iFp22*iFp22*iFp22;
  D[2][2] = t0;
  s2 = Dbar23*iFp21*iFp21*iFp12*iFp02+Dbar16*iFp20*iFp20*iFp02*iFp11+Dbar13*iFp20*iFp20*iFp02*iFp12+Dbar14*iFp20*iFp20*iFp00*iFp11+Dbar15*iFp20*iFp20*iFp00*iFp12+Dbar11*iFp20*iFp20*iFp00*iFp10+Dbar14*iFp20*iFp20*iFp01*iFp10+Dbar24*iFp21*iFp21*iFp10*iFp01+Dbar35*iFp22*iFp22*iFp12*iFp00+Dbar34*iFp22*iFp22*iFp11*iFp00+Dbar15*iFp20*iFp20*iFp02*iFp10+Dbar23*iFp22*iFp22*iFp11*iFp01+Dbar35*iFp22*iFp22*iFp10*iFp02;
  s1 = s2+Dbar22*iFp21*iFp21*iFp11*iFp01+Dbar12*iFp20*iFp20*iFp01*iFp11+Dbar16*iFp20*iFp20*iFp01*iFp12+Dbar25*iFp21*iFp21*iFp10*iFp02+Dbar26*iFp21*iFp21*iFp12*iFp01+Dbar34*iFp22*iFp22*iFp10*iFp01+Dbar13*iFp22*iFp22*iFp10*iFp00+Dbar36*iFp22*iFp22*iFp11*iFp02+Dbar12*iFp21*iFp21*iFp10*iFp00+2.0*Dbar45*iFp20*iFp21*iFp02*iFp10+2.0*Dbar46*iFp20*iFp21*iFp01*iFp12+2.0*Dbar15*iFp20*iFp22*iFp00*iFp10+2.0*Dbar34*iFp20*iFp21*iFp02*iFp12+2.0*Dbar55*iFp20*iFp22*iFp00*iFp12;
  s2 = s1+2.0*Dbar45*iFp20*iFp22*iFp00*iFp11+Dbar36*iFp22*iFp22*iFp12*iFp01+Dbar33*iFp22*iFp22*iFp12*iFp02+Dbar26*iFp21*iFp21*iFp11*iFp02+Dbar25*iFp21*iFp21*iFp12*iFp00+Dbar24*iFp21*iFp21*iFp11*iFp00+2.0*Dbar26*iFp21*iFp22*iFp11*iFp01+2.0*Dbar45*iFp20*iFp22*iFp01*iFp10+2.0*Dbar56*iFp21*iFp22*iFp12*iFp00+2.0*Dbar36*iFp21*iFp22*iFp12*iFp02+2.0*Dbar66*iFp21*iFp22*iFp12*iFp01+2.0*Dbar46*iFp21*iFp22*iFp11*iFp00+2.0*Dbar46*iFp20*iFp21*iFp02*iFp11;
  t0 = s2+2.0*Dbar56*iFp21*iFp22*iFp10*iFp02+2.0*Dbar66*iFp21*iFp22*iFp11*iFp02+2.0*Dbar46*iFp21*iFp22*iFp10*iFp01+2.0*Dbar16*iFp21*iFp22*iFp10*iFp00+2.0*Dbar35*iFp20*iFp22*iFp02*iFp12+2.0*Dbar25*iFp20*iFp22*iFp01*iFp11+2.0*Dbar14*iFp20*iFp21*iFp10*iFp00+2.0*Dbar56*iFp20*iFp22*iFp02*iFp11+2.0*Dbar56*iFp20*iFp22*iFp01*iFp12+2.0*Dbar55*iFp20*iFp22*iFp02*iFp10+2.0*Dbar45*iFp20*iFp21*iFp00*iFp12+2.0*Dbar44*iFp20*iFp21*iFp00*iFp11+2.0*Dbar24*iFp20*iFp21*iFp01*iFp11+2.0*Dbar44*iFp20*iFp21*iFp01*iFp10;
  D[2][3] = D[3][2] = t0;
  s1 = Dbar26*iFp21*iFp21*iFp21*iFp02+Dbar22*iFp21*iFp21*iFp21*iFp01+Dbar33*iFp22*iFp22*iFp22*iFp02+Dbar35*iFp22*iFp22*iFp22*iFp00+Dbar36*iFp22*iFp22*iFp22*iFp01+Dbar11*iFp20*iFp20*iFp20*iFp00+Dbar15*iFp20*iFp20*iFp20*iFp02+Dbar14*iFp20*iFp20*iFp20*iFp01+Dbar12*iFp20*iFp20*iFp21*iFp01+Dbar16*iFp20*iFp20*iFp21*iFp02+Dbar25*iFp21*iFp21*iFp22*iFp00+Dbar23*iFp21*iFp21*iFp22*iFp02+Dbar34*iFp22*iFp22*iFp21*iFp00+Dbar12*iFp21*iFp21*iFp20*iFp00+3.0*Dbar36*iFp21*iFp22*iFp22*iFp02+2.0*Dbar46*iFp22*iFp21*iFp21*iFp00+2.0*Dbar56*iFp22*iFp22*iFp21*iFp00+2.0*Dbar66*iFp22*iFp21*iFp21*iFp02+3.0*Dbar26*iFp21*iFp21*iFp22*iFp01+3.0*Dbar35*iFp22*iFp22*iFp20*iFp02+2.0*Dbar56*iFp22*iFp22*iFp20*iFp01+2.0*Dbar55*iFp22*iFp22*iFp20*iFp00;
  t0 = s1+3.0*Dbar15*iFp20*iFp20*iFp22*iFp00+2.0*Dbar55*iFp22*iFp20*iFp20*iFp02+2.0*Dbar45*iFp22*iFp20*iFp20*iFp01+Dbar23*iFp22*iFp22*iFp21*iFp01+Dbar16*iFp20*iFp20*iFp22*iFp01+Dbar13*iFp20*iFp20*iFp22*iFp02+Dbar25*iFp21*iFp21*iFp20*iFp02+Dbar34*iFp22*iFp22*iFp20*iFp01+Dbar13*iFp22*iFp22*iFp20*iFp00+2.0*Dbar44*iFp20*iFp20*iFp21*iFp01+2.0*Dbar45*iFp20*iFp20*iFp21*iFp02+2.0*Dbar44*iFp20*iFp21*iFp21*iFp00+3.0*Dbar24*iFp20*iFp21*iFp21*iFp01+2.0*Dbar46*iFp20*iFp21*iFp21*iFp02+3.0*Dbar14*iFp20*iFp20*iFp21*iFp00+2.0*Dbar66*iFp22*iFp22*iFp21*iFp01+Dbar24*iFp21*iFp21*iFp21*iFp00+4.0*Dbar45*iFp20*iFp21*iFp22*iFp00+4.0*Dbar56*iFp22*iFp20*iFp21*iFp02+2.0*Dbar25*iFp22*iFp20*iFp21*iFp01+2.0*Dbar34*iFp20*iFp21*iFp22*iFp02+2.0*Dbar16*iFp22*iFp21*iFp20*iFp00+4.0*Dbar46*iFp20*iFp21*iFp22*iFp01;
  D[2][4] = D[4][2] = t0;
  s1 = Dbar24*iFp21*iFp21*iFp21*iFp10+Dbar11*iFp20*iFp20*iFp20*iFp10+Dbar14*iFp20*iFp20*iFp20*iFp11+Dbar15*iFp20*iFp20*iFp20*iFp12+Dbar26*iFp21*iFp21*iFp21*iFp12+Dbar36*iFp22*iFp22*iFp22*iFp11+Dbar34*iFp22*iFp22*iFp20*iFp11+Dbar25*iFp21*iFp21*iFp20*iFp12+Dbar34*iFp22*iFp22*iFp21*iFp10+Dbar23*iFp22*iFp22*iFp21*iFp11+Dbar12*iFp21*iFp21*iFp20*iFp10+Dbar13*iFp22*iFp22*iFp20*iFp10+2.0*Dbar56*iFp20*iFp22*iFp22*iFp11+3.0*Dbar35*iFp22*iFp22*iFp20*iFp12+2.0*Dbar55*iFp20*iFp22*iFp22*iFp10+2.0*Dbar45*iFp20*iFp20*iFp22*iFp11+2.0*Dbar55*iFp20*iFp20*iFp22*iFp12+3.0*Dbar24*iFp20*iFp21*iFp21*iFp11+2.0*Dbar56*iFp22*iFp22*iFp21*iFp10+2.0*Dbar46*iFp20*iFp21*iFp21*iFp12+2.0*Dbar66*iFp22*iFp22*iFp21*iFp11+3.0*Dbar36*iFp22*iFp22*iFp21*iFp12;
  t0 = s1+Dbar16*iFp20*iFp20*iFp21*iFp12+3.0*Dbar15*iFp20*iFp20*iFp22*iFp10+Dbar16*iFp20*iFp20*iFp22*iFp11+Dbar13*iFp20*iFp20*iFp22*iFp12+2.0*Dbar46*iFp22*iFp21*iFp21*iFp10+2.0*Dbar44*iFp20*iFp20*iFp21*iFp11+2.0*Dbar45*iFp20*iFp20*iFp21*iFp12+2.0*Dbar66*iFp22*iFp21*iFp21*iFp12+2.0*Dbar44*iFp20*iFp21*iFp21*iFp10+Dbar23*iFp21*iFp21*iFp22*iFp12+3.0*Dbar14*iFp20*iFp20*iFp21*iFp10+Dbar12*iFp20*iFp20*iFp21*iFp11+3.0*Dbar26*iFp22*iFp21*iFp21*iFp11+Dbar25*iFp21*iFp21*iFp22*iFp10+2.0*Dbar34*iFp20*iFp21*iFp22*iFp12+2.0*Dbar16*iFp21*iFp22*iFp20*iFp10+Dbar35*iFp22*iFp22*iFp22*iFp10+4.0*Dbar45*iFp20*iFp21*iFp22*iFp10+Dbar22*iFp21*iFp21*iFp21*iFp11+4.0*Dbar56*iFp20*iFp22*iFp21*iFp12+2.0*Dbar25*iFp20*iFp22*iFp21*iFp11+Dbar33*iFp22*iFp22*iFp22*iFp12+4.0*Dbar46*iFp20*iFp21*iFp22*iFp11;
  D[2][5] = D[5][2] = t0;
  s1 = Dbar44*iFp01*iFp01*iFp10*iFp10+Dbar55*iFp00*iFp00*iFp12*iFp12+Dbar55*iFp02*iFp02*iFp10*iFp10+Dbar44*iFp00*iFp00*iFp11*iFp11+Dbar66*iFp01*iFp01*iFp12*iFp12+Dbar11*iFp00*iFp00*iFp10*iFp10+Dbar33*iFp02*iFp02*iFp12*iFp12+Dbar22*iFp01*iFp01*iFp11*iFp11+2.0*Dbar36*iFp01*iFp12*iFp12*iFp02+2.0*Dbar24*iFp00*iFp11*iFp11*iFp01+2.0*Dbar46*iFp00*iFp11*iFp11*iFp02+2.0*Dbar45*iFp00*iFp00*iFp11*iFp12+2.0*Dbar15*iFp00*iFp00*iFp10*iFp12+2.0*Dbar14*iFp00*iFp10*iFp10*iFp01+2.0*Dbar15*iFp00*iFp10*iFp10*iFp02+2.0*Dbar14*iFp00*iFp00*iFp10*iFp11+2.0*Dbar26*iFp01*iFp11*iFp11*iFp02+2.0*Dbar26*iFp01*iFp01*iFp11*iFp12+2.0*Dbar36*iFp02*iFp02*iFp11*iFp12+2.0*Dbar46*iFp01*iFp01*iFp10*iFp12+2.0*Dbar35*iFp02*iFp02*iFp10*iFp12+2.0*Dbar45*iFp01*iFp10*iFp10*iFp02;
  s2 = s1+2.0*Dbar24*iFp01*iFp01*iFp10*iFp11+2.0*Dbar56*iFp00*iFp12*iFp12*iFp01+2.0*Dbar56*iFp02*iFp02*iFp10*iFp11+2.0*Dbar35*iFp00*iFp12*iFp12*iFp02+2.0*Dbar44*iFp00*iFp11*iFp10*iFp01+2.0*Dbar13*iFp00*iFp10*iFp12*iFp02+2.0*Dbar16*iFp00*iFp10*iFp12*iFp01+2.0*Dbar55*iFp00*iFp12*iFp10*iFp02+2.0*Dbar45*iFp00*iFp12*iFp10*iFp01+2.0*Dbar46*iFp01*iFp10*iFp11*iFp02+2.0*Dbar34*iFp01*iFp10*iFp12*iFp02;
  t0 = s2+2.0*Dbar34*iFp00*iFp11*iFp12*iFp02+2.0*Dbar46*iFp00*iFp11*iFp12*iFp01+2.0*Dbar45*iFp00*iFp11*iFp10*iFp02+2.0*Dbar56*iFp00*iFp12*iFp11*iFp02+2.0*Dbar66*iFp01*iFp12*iFp11*iFp02+2.0*Dbar25*iFp00*iFp12*iFp11*iFp01+2.0*Dbar25*iFp01*iFp11*iFp10*iFp02+2.0*Dbar56*iFp01*iFp12*iFp10*iFp02+2.0*Dbar23*iFp01*iFp11*iFp12*iFp02+2.0*Dbar12*iFp00*iFp10*iFp11*iFp01+2.0*Dbar16*iFp00*iFp10*iFp11*iFp02+Dbar66*iFp02*iFp02*iFp11*iFp11;
  D[3][3] = t0;
  s2 = Dbar11*iFp00*iFp00*iFp20*iFp10+Dbar14*iFp00*iFp00*iFp20*iFp11+Dbar24*iFp01*iFp01*iFp21*iFp10+Dbar15*iFp00*iFp00*iFp22*iFp10+Dbar26*iFp01*iFp01*iFp22*iFp11+Dbar22*iFp01*iFp01*iFp21*iFp11+Dbar26*iFp01*iFp01*iFp21*iFp12+Dbar24*iFp01*iFp01*iFp20*iFp11+Dbar35*iFp02*iFp02*iFp20*iFp12+Dbar36*iFp02*iFp02*iFp22*iFp11+Dbar15*iFp00*iFp00*iFp20*iFp12+Dbar14*iFp00*iFp00*iFp21*iFp10+Dbar56*iFp10*iFp02*iFp02*iFp21+Dbar46*iFp12*iFp01*iFp01*iFp20+Dbar55*iFp12*iFp00*iFp00*iFp22+Dbar66*iFp12*iFp01*iFp01*iFp22+Dbar46*iFp10*iFp01*iFp01*iFp22+Dbar55*iFp10*iFp02*iFp02*iFp20;
  s1 = s2+Dbar45*iFp12*iFp00*iFp00*iFp21+Dbar45*iFp11*iFp00*iFp00*iFp22+Dbar44*iFp11*iFp00*iFp00*iFp21+Dbar66*iFp11*iFp02*iFp02*iFp21+Dbar56*iFp11*iFp02*iFp02*iFp20+Dbar44*iFp10*iFp01*iFp01*iFp20+Dbar35*iFp02*iFp02*iFp22*iFp10+Dbar56*iFp00*iFp02*iFp22*iFp11+2.0*Dbar24*iFp00*iFp01*iFp21*iFp11+Dbar44*iFp00*iFp01*iFp21*iFp10+Dbar55*iFp00*iFp02*iFp22*iFp10+Dbar56*iFp00*iFp02*iFp21*iFp12+Dbar45*iFp00*iFp01*iFp20*iFp12+Dbar44*iFp00*iFp01*iFp20*iFp11+Dbar45*iFp00*iFp02*iFp21*iFp10+2.0*Dbar14*iFp00*iFp01*iFp20*iFp10+Dbar45*iFp00*iFp02*iFp20*iFp11+Dbar55*iFp00*iFp02*iFp20*iFp12;
  s2 = 2.0*Dbar15*iFp00*iFp02*iFp20*iFp10+Dbar66*iFp01*iFp02*iFp22*iFp11+Dbar66*iFp01*iFp02*iFp21*iFp12+Dbar46*iFp00*iFp01*iFp22*iFp11+Dbar56*iFp01*iFp02*iFp22*iFp10+Dbar46*iFp01*iFp02*iFp21*iFp10+2.0*Dbar26*iFp01*iFp02*iFp21*iFp11+Dbar46*iFp01*iFp02*iFp20*iFp11+Dbar56*iFp01*iFp02*iFp20*iFp12+Dbar33*iFp02*iFp02*iFp22*iFp12+Dbar45*iFp00*iFp01*iFp22*iFp10+Dbar46*iFp00*iFp01*iFp21*iFp12+2.0*Dbar35*iFp00*iFp02*iFp22*iFp12+2.0*Dbar36*iFp01*iFp02*iFp22*iFp12+Dbar34*iFp12*iFp02*iFp20*iFp01+Dbar34*iFp12*iFp02*iFp21*iFp00+Dbar16*iFp12*iFp01*iFp20*iFp00+Dbar13*iFp12*iFp02*iFp20*iFp00;
  t0 = s2+Dbar34*iFp10*iFp01*iFp22*iFp02+Dbar25*iFp10*iFp02*iFp21*iFp01+Dbar25*iFp12*iFp00*iFp21*iFp01+Dbar36*iFp02*iFp02*iFp21*iFp12+Dbar25*iFp11*iFp01*iFp22*iFp00+Dbar23*iFp11*iFp01*iFp22*iFp02+Dbar12*iFp11*iFp01*iFp20*iFp00+Dbar23*iFp12*iFp02*iFp21*iFp01+Dbar34*iFp11*iFp00*iFp22*iFp02+Dbar16*iFp11*iFp02*iFp20*iFp00+Dbar16*iFp10*iFp00*iFp21*iFp02+Dbar16*iFp10*iFp00*iFp22*iFp01+Dbar13*iFp10*iFp00*iFp22*iFp02+Dbar25*iFp11*iFp01*iFp20*iFp02+Dbar12*iFp10*iFp00*iFp21*iFp01+2.0*Dbar45*iFp10*iFp01*iFp20*iFp02+2.0*Dbar56*iFp12*iFp00*iFp22*iFp01+2.0*Dbar46*iFp11*iFp00*iFp21*iFp02+s1;
  D[3][4] = D[4][3] = t0;
  s2 = Dbar24*iFp11*iFp11*iFp20*iFp01+Dbar56*iFp12*iFp12*iFp01*iFp20+Dbar46*iFp11*iFp11*iFp00*iFp22+Dbar46*iFp11*iFp11*iFp02*iFp20+Dbar44*iFp10*iFp10*iFp01*iFp21+Dbar44*iFp11*iFp11*iFp00*iFp20+Dbar66*iFp12*iFp12*iFp01*iFp21+Dbar45*iFp10*iFp10*iFp02*iFp21+2.0*Dbar46*iFp10*iFp01*iFp21*iFp12+Dbar56*iFp11*iFp12*iFp22*iFp00+Dbar55*iFp10*iFp12*iFp20*iFp02+Dbar45*iFp10*iFp12*iFp21*iFp00+Dbar45*iFp10*iFp11*iFp22*iFp00+Dbar66*iFp11*iFp12*iFp22*iFp01+2.0*Dbar35*iFp10*iFp12*iFp22*iFp02+2.0*Dbar36*iFp11*iFp12*iFp22*iFp02+Dbar45*iFp10*iFp12*iFp20*iFp01+Dbar56*iFp10*iFp12*iFp21*iFp02;
  s1 = s2+2.0*Dbar15*iFp10*iFp12*iFp20*iFp00+Dbar66*iFp11*iFp12*iFp21*iFp02+2.0*Dbar45*iFp11*iFp00*iFp20*iFp12+2.0*Dbar56*iFp10*iFp02*iFp22*iFp11+Dbar44*iFp10*iFp11*iFp20*iFp01+2.0*Dbar14*iFp10*iFp11*iFp20*iFp00+Dbar55*iFp10*iFp12*iFp22*iFp00+Dbar56*iFp10*iFp12*iFp22*iFp01+Dbar56*iFp11*iFp12*iFp20*iFp02+Dbar46*iFp11*iFp12*iFp21*iFp00+Dbar44*iFp10*iFp11*iFp21*iFp00+Dbar46*iFp11*iFp12*iFp20*iFp01+Dbar45*iFp10*iFp11*iFp20*iFp02+Dbar34*iFp12*iFp02*iFp20*iFp11+Dbar13*iFp12*iFp02*iFp20*iFp10+Dbar46*iFp10*iFp11*iFp22*iFp01+Dbar46*iFp10*iFp11*iFp21*iFp02+2.0*Dbar26*iFp11*iFp12*iFp21*iFp01;
  s2 = 2.0*Dbar24*iFp10*iFp11*iFp21*iFp01+Dbar23*iFp11*iFp01*iFp22*iFp12+Dbar25*iFp11*iFp01*iFp22*iFp10+Dbar34*iFp11*iFp00*iFp22*iFp12+Dbar23*iFp12*iFp02*iFp21*iFp11+Dbar34*iFp12*iFp02*iFp21*iFp10+Dbar16*iFp12*iFp01*iFp20*iFp10+Dbar25*iFp11*iFp01*iFp20*iFp12+Dbar16*iFp10*iFp00*iFp21*iFp12+Dbar12*iFp10*iFp00*iFp21*iFp11+Dbar12*iFp11*iFp01*iFp20*iFp10+Dbar16*iFp11*iFp02*iFp20*iFp10+Dbar25*iFp12*iFp00*iFp21*iFp11+Dbar13*iFp10*iFp00*iFp22*iFp12+Dbar25*iFp10*iFp02*iFp21*iFp11+Dbar24*iFp11*iFp11*iFp21*iFp00+Dbar45*iFp10*iFp10*iFp01*iFp22+Dbar56*iFp12*iFp12*iFp00*iFp21;
  t0 = s2+Dbar55*iFp12*iFp12*iFp00*iFp20+Dbar15*iFp10*iFp10*iFp20*iFp02+Dbar55*iFp10*iFp10*iFp02*iFp22+s1+Dbar16*iFp10*iFp00*iFp22*iFp11+Dbar35*iFp12*iFp12*iFp22*iFp00+Dbar66*iFp11*iFp11*iFp02*iFp22+Dbar15*iFp10*iFp10*iFp22*iFp00+Dbar33*iFp12*iFp12*iFp22*iFp02+Dbar26*iFp11*iFp11*iFp21*iFp02+Dbar35*iFp12*iFp12*iFp20*iFp02+Dbar14*iFp10*iFp10*iFp21*iFp00+Dbar34*iFp10*iFp01*iFp22*iFp12+Dbar26*iFp11*iFp11*iFp22*iFp01+Dbar36*iFp12*iFp12*iFp22*iFp01+Dbar36*iFp12*iFp12*iFp21*iFp02+Dbar11*iFp10*iFp10*iFp20*iFp00+Dbar22*iFp11*iFp11*iFp21*iFp01+Dbar14*iFp10*iFp10*iFp20*iFp01;
  D[3][5] = D[5][3] = t0;
  s1 = Dbar44*iFp21*iFp21*iFp00*iFp00+Dbar66*iFp22*iFp22*iFp01*iFp01+Dbar55*iFp22*iFp22*iFp00*iFp00+2.0*Dbar46*iFp20*iFp01*iFp01*iFp22+2.0*Dbar45*iFp20*iFp20*iFp01*iFp02+2.0*Dbar45*iFp21*iFp00*iFp00*iFp22+2.0*Dbar46*iFp21*iFp21*iFp00*iFp02+2.0*Dbar56*iFp22*iFp22*iFp00*iFp01+2.0*Dbar35*iFp02*iFp02*iFp20*iFp22+2.0*Dbar14*iFp00*iFp00*iFp20*iFp21+2.0*Dbar15*iFp00*iFp00*iFp20*iFp22+2.0*Dbar36*iFp01*iFp02*iFp22*iFp22+2.0*Dbar26*iFp01*iFp02*iFp21*iFp21+2.0*Dbar36*iFp02*iFp02*iFp21*iFp22+2.0*Dbar15*iFp00*iFp02*iFp20*iFp20+2.0*Dbar24*iFp00*iFp01*iFp21*iFp21+2.0*Dbar14*iFp00*iFp01*iFp20*iFp20+2.0*Dbar26*iFp01*iFp01*iFp21*iFp22+2.0*Dbar35*iFp00*iFp02*iFp22*iFp22+2.0*Dbar24*iFp01*iFp01*iFp20*iFp21+2.0*Dbar56*iFp20*iFp02*iFp02*iFp21+2.0*Dbar45*iFp00*iFp02*iFp20*iFp21;
  s2 = s1+2.0*Dbar25*iFp21*iFp01*iFp22*iFp00+2.0*Dbar34*iFp20*iFp01*iFp22*iFp02+2.0*Dbar34*iFp21*iFp00*iFp22*iFp02+2.0*Dbar16*iFp20*iFp00*iFp21*iFp02+2.0*Dbar13*iFp20*iFp00*iFp22*iFp02+2.0*Dbar16*iFp20*iFp00*iFp22*iFp01+2.0*Dbar46*iFp01*iFp02*iFp20*iFp21+2.0*Dbar25*iFp20*iFp02*iFp21*iFp01+2.0*Dbar46*iFp00*iFp01*iFp21*iFp22+2.0*Dbar56*iFp01*iFp02*iFp20*iFp22+2.0*Dbar44*iFp00*iFp01*iFp20*iFp21;
  t0 = s2+2.0*Dbar56*iFp00*iFp02*iFp21*iFp22+2.0*Dbar45*iFp00*iFp01*iFp20*iFp22+2.0*Dbar55*iFp00*iFp02*iFp20*iFp22+2.0*Dbar66*iFp01*iFp02*iFp21*iFp22+Dbar22*iFp01*iFp01*iFp21*iFp21+Dbar33*iFp02*iFp02*iFp22*iFp22+Dbar11*iFp00*iFp00*iFp20*iFp20+Dbar66*iFp21*iFp21*iFp02*iFp02+Dbar55*iFp20*iFp20*iFp02*iFp02+2.0*Dbar12*iFp20*iFp00*iFp21*iFp01+Dbar44*iFp20*iFp20*iFp01*iFp01+2.0*Dbar23*iFp21*iFp01*iFp22*iFp02;
  D[4][4] = t0;
  s2 = Dbar11*iFp20*iFp20*iFp00*iFp10+Dbar14*iFp20*iFp20*iFp01*iFp10+Dbar24*iFp21*iFp21*iFp10*iFp01+Dbar35*iFp22*iFp22*iFp12*iFp00+Dbar22*iFp21*iFp21*iFp11*iFp01+Dbar55*iFp20*iFp20*iFp02*iFp12+Dbar55*iFp22*iFp22*iFp00*iFp10+Dbar56*iFp22*iFp22*iFp00*iFp11+Dbar44*iFp21*iFp21*iFp00*iFp10+Dbar44*iFp20*iFp20*iFp01*iFp11+Dbar45*iFp20*iFp20*iFp01*iFp12+Dbar46*iFp21*iFp21*iFp02*iFp10+Dbar66*iFp21*iFp21*iFp02*iFp12+Dbar56*iFp22*iFp22*iFp01*iFp10+Dbar46*iFp21*iFp21*iFp00*iFp12+Dbar14*iFp20*iFp20*iFp00*iFp11+Dbar15*iFp20*iFp20*iFp00*iFp12+Dbar34*iFp22*iFp02*iFp21*iFp10;
  s1 = s2+Dbar23*iFp22*iFp02*iFp21*iFp11+Dbar34*iFp21*iFp00*iFp22*iFp12+Dbar36*iFp22*iFp22*iFp12*iFp01+Dbar33*iFp22*iFp22*iFp12*iFp02+Dbar35*iFp22*iFp22*iFp10*iFp02+2.0*Dbar56*iFp20*iFp02*iFp21*iFp12+2.0*Dbar45*iFp21*iFp00*iFp22*iFp10+2.0*Dbar36*iFp21*iFp22*iFp12*iFp02+Dbar45*iFp20*iFp22*iFp01*iFp10+2.0*Dbar46*iFp22*iFp01*iFp20*iFp11+Dbar56*iFp21*iFp22*iFp10*iFp02+Dbar66*iFp21*iFp22*iFp11*iFp02+2.0*Dbar26*iFp21*iFp22*iFp11*iFp01+Dbar55*iFp20*iFp22*iFp02*iFp10+Dbar56*iFp20*iFp22*iFp02*iFp11+2.0*Dbar14*iFp20*iFp21*iFp10*iFp00+2.0*Dbar35*iFp20*iFp22*iFp02*iFp12+Dbar44*iFp20*iFp21*iFp01*iFp10;
  s2 = Dbar45*iFp20*iFp21*iFp00*iFp12+Dbar44*iFp20*iFp21*iFp00*iFp11+2.0*Dbar24*iFp20*iFp21*iFp01*iFp11+Dbar56*iFp20*iFp22*iFp01*iFp12+Dbar46*iFp21*iFp22*iFp10*iFp01+Dbar46*iFp21*iFp22*iFp11*iFp00+Dbar66*iFp21*iFp22*iFp12*iFp01+Dbar56*iFp21*iFp22*iFp12*iFp00+Dbar25*iFp21*iFp01*iFp20*iFp12+Dbar12*iFp20*iFp00*iFp21*iFp11+Dbar16*iFp20*iFp00*iFp21*iFp12+Dbar25*iFp20*iFp02*iFp21*iFp11+Dbar16*iFp22*iFp01*iFp20*iFp10+Dbar66*iFp22*iFp22*iFp01*iFp11+Dbar25*iFp21*iFp01*iFp22*iFp10+Dbar55*iFp20*iFp22*iFp00*iFp12+Dbar45*iFp20*iFp22*iFp00*iFp11+2.0*Dbar15*iFp20*iFp22*iFp00*iFp10;
  t0 = s2+Dbar46*iFp20*iFp21*iFp02*iFp11+Dbar45*iFp20*iFp21*iFp02*iFp10+Dbar46*iFp20*iFp21*iFp01*iFp12+Dbar16*iFp21*iFp02*iFp20*iFp10+Dbar23*iFp21*iFp01*iFp22*iFp12+Dbar16*iFp20*iFp00*iFp22*iFp11+Dbar25*iFp22*iFp00*iFp21*iFp11+Dbar34*iFp20*iFp01*iFp22*iFp12+s1+Dbar12*iFp21*iFp01*iFp20*iFp10+Dbar13*iFp22*iFp02*iFp20*iFp10+Dbar34*iFp22*iFp02*iFp20*iFp11+Dbar13*iFp20*iFp00*iFp22*iFp12+Dbar45*iFp20*iFp20*iFp02*iFp11+Dbar24*iFp21*iFp21*iFp11*iFp00+Dbar26*iFp21*iFp21*iFp12*iFp01+Dbar15*iFp20*iFp20*iFp02*iFp10+Dbar36*iFp22*iFp22*iFp11*iFp02+Dbar26*iFp21*iFp21*iFp11*iFp02;
  D[4][5] = D[5][4] = t0;
  s1 = Dbar66*iFp21*iFp21*iFp12*iFp12+Dbar44*iFp21*iFp21*iFp10*iFp10+Dbar66*iFp22*iFp22*iFp11*iFp11+Dbar55*iFp20*iFp20*iFp12*iFp12+2.0*Dbar14*iFp10*iFp11*iFp20*iFp20+2.0*Dbar24*iFp10*iFp11*iFp21*iFp21+2.0*Dbar15*iFp10*iFp12*iFp20*iFp20+2.0*Dbar15*iFp10*iFp10*iFp20*iFp22+2.0*Dbar14*iFp10*iFp10*iFp20*iFp21+2.0*Dbar36*iFp12*iFp12*iFp21*iFp22+2.0*Dbar24*iFp11*iFp11*iFp20*iFp21+2.0*Dbar36*iFp11*iFp12*iFp22*iFp22+2.0*Dbar26*iFp11*iFp12*iFp21*iFp21+2.0*Dbar35*iFp12*iFp12*iFp20*iFp22+2.0*Dbar56*iFp22*iFp22*iFp10*iFp11+2.0*Dbar46*iFp21*iFp21*iFp10*iFp12+2.0*Dbar45*iFp22*iFp10*iFp10*iFp21+2.0*Dbar56*iFp21*iFp12*iFp12*iFp20+Dbar55*iFp22*iFp22*iFp10*iFp10+Dbar44*iFp20*iFp20*iFp11*iFp11+2.0*Dbar16*iFp20*iFp10*iFp22*iFp11+2.0*Dbar16*iFp20*iFp10*iFp21*iFp12;
  s2 = s1+2.0*Dbar13*iFp20*iFp10*iFp22*iFp12+2.0*Dbar23*iFp21*iFp11*iFp22*iFp12+2.0*Dbar25*iFp21*iFp11*iFp22*iFp10+2.0*Dbar46*iFp22*iFp11*iFp11*iFp20+2.0*Dbar34*iFp20*iFp11*iFp22*iFp12+2.0*Dbar34*iFp21*iFp10*iFp22*iFp12+Dbar11*iFp10*iFp10*iFp20*iFp20+2.0*Dbar12*iFp20*iFp10*iFp21*iFp11+2.0*Dbar25*iFp20*iFp12*iFp21*iFp11+2.0*Dbar45*iFp20*iFp20*iFp11*iFp12+2.0*Dbar56*iFp10*iFp12*iFp21*iFp22;
  t0 = s2+2.0*Dbar44*iFp10*iFp11*iFp20*iFp21+2.0*Dbar45*iFp10*iFp11*iFp20*iFp22+2.0*Dbar45*iFp10*iFp12*iFp20*iFp21+2.0*Dbar46*iFp10*iFp11*iFp21*iFp22+2.0*Dbar56*iFp11*iFp12*iFp20*iFp22+2.0*Dbar35*iFp10*iFp12*iFp22*iFp22+2.0*Dbar46*iFp11*iFp12*iFp20*iFp21+2.0*Dbar66*iFp11*iFp12*iFp21*iFp22+2.0*Dbar55*iFp10*iFp12*iFp20*iFp22+2.0*Dbar26*iFp11*iFp11*iFp21*iFp22+Dbar33*iFp12*iFp12*iFp22*iFp22+Dbar22*iFp11*iFp11*iFp21*iFp21;
  D[5][5] = t0;

  /* I'm handling request for Lagrangian stiffness only;
     if Eulerian, should transform first.
  */
  if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) return TRUE;
  else                                       return FALSE;
}

static W_bool_t
get_mat_stiff_elastic_9x9 (W_mat_3d_t *m /* mat3d */,
                       W_mat_3d_state_p ms,/* material state */
                       W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                                    used to generate kinematic
                                                    quantities which the material state
                                                    might need in order to compute
                                                    the requested stress */
                       W_mat_3d_stiff_enum stiff_type, /* stiffness type */
                       double D[9][9] /* material stiffness tensor in matrix form */
                       )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  double C[6][6];
  int i, j; 
/* get C[6][6] first, then expand to D[9][9] */
  get_mat_stiff_elastic(m, ms, ip_ctx, stiff_type, C);
  FOR (i, 9) {
      FOR(j,9) {
	if (i <=5  && j <= 5) { 
		D[i][j] = C[i][j];
	} else {
		if (i >5 && j<=5 ) D[i][j] = C[i-3][j];
		else {
		if (i <=5 && j>5 ) D[i][j] = C[i][j-3];
		else	       D[i][j] = C[i-3][j-3];
		     }	
	       }
      }
  }
  /* I'm handling request for Lagrangian stiffness only;
     if Eulerian, should transform first.
  */
  if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) return TRUE;
  else                                       return FALSE;
}

typedef struct {
  double F[3][3];
} custom_data_t;

static W_bool_t 
custom_get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
               W_mat_3d_kinem_var_type_enum  type,
               double kinvar[3][3])
{
  custom_data_t *custom_data = (custom_data_t *) ctx->ctx_data;
  switch (type) {
  case MAT_3D_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, custom_data->F);
    break;
  default:
    REPORT_FATAL_ERROR(NULL, "Unknown kinematic tensor type");
    break;
  }
  return TRUE;
}

static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* stiffness type */
               double C[6][6] /* material stiffness tensor in matrix form */
               )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  int i, j, k, l;
  double e = u->e_incr;
  double U[3][3], R[3][3], dU[3][3], U1[3][3];
  custom_data_t custom_data;
  double Sp[6], Sn[6];
  W_mat_3d_stress_type_enum   out_stress_type;
  int result = TRUE;
  int v_to_t[6][2] = {{0,0},{1,1},{2,2},{0,1},{0,2},{1,2}};
  int mult[6] = {1,1,1,2,2,2};
  double equiv_pl_def;
  W_mat_3d_ip_ctx_t custom_ip_ctx;
  GEP_MATRIX (um);
  int umpvt[6];
  double *pd[6];
  double um_[6][6];
  GEP_STACKINIT(um, 6);
  

  if (u->e_incr == 0) /* numerical differentiation not allowed */ {
    W_bool_t result =  get_mat_stiff_elastic (m, ms, ip_ctx, stiff_type, C);
    return result;
  }
  
  /* zero out */
  FOR(i, 6) FOR (j, 6) C[i][j] = 0;

  /* here we access the current kinematic variables through the element defined
     kinematics callback
   */
  if (!ip_ctx->get_kinem_var) return FALSE;  
  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, custom_data.F)) {
    result = FALSE; goto exit;
  }

  /* now replace the kinematic variable callback by our custom callback */
  custom_ip_ctx               = *ip_ctx;
  custom_ip_ctx.get_kinem_var = custom_get_kinem_var;
  custom_ip_ctx.ctx_data      = &custom_data;

  /* compute the polar decomposition of the elastic deformation gradient */
  if (!W_polar_decomposition (custom_data.F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }
#define UM(i,j,v) {um_[i][j] = v;}
  UM(0,0,U[0][0]); UM(0,1,0);       UM(0,2,0);       UM(0,3,U[1][0]);             UM(0,4,U[2][0]);             UM(0,5,0);
  UM(1,0,0);       UM(1,1,U[1][1]); UM(1,2,0);       UM(1,3,U[0][1]);             UM(1,4,0);                   UM(1,5,U[2][1]);
  UM(2,0,0);       UM(2,1,0);       UM(2,2,U[2][2]); UM(2,3,0);                   UM(2,4,U[0][2]);             UM(2,5,U[1][2]);
  UM(3,0,U[0][1]); UM(3,1,U[1][0]); UM(3,2,0);       UM(3,3,(U[0][0]+U[1][1])/2); UM(3,4,U[2][1]/2);           UM(3,5,U[2][0]/2);
  UM(4,0,U[0][2]); UM(4,1,0);       UM(4,2,U[2][0]); UM(4,3,U[1][2]/2);           UM(4,4,(U[0][0]+U[2][2])/2); UM(4,5,U[1][0]/2);
  UM(5,0,0);       UM(5,1,U[1][2]); UM(5,2,U[2][1]); UM(5,3,U[0][2]/2);           UM(5,4,U[0][1]/2);           UM(5,5,(U[1][1]+U[2][2])/2);
  FOR(i, 6){FOR(j, 6){GEP_ELEM(um, i, j) = um_[i][j];}}
  gep_sgefa (&um, umpvt);
  
#define GET_2ND_PK(S)                                                                   \
  {                                                                                     \
    double St[3][3];                                                                    \
    if (!WM3D_update_stress (m, ms, W_MAT_STATE_UPDATE_TEMP,                            \
                             &custom_ip_ctx, MAT_3D_2ND_PK, St, &out_stress_type)) {    \
      result = FALSE; goto exit;                                                        \
    }                                                                                   \
    if (out_stress_type != MAT_3D_2ND_PK) {result = FALSE; goto exit;}                  \
    S[0] = St[0][0];                                                                    \
    S[1] = St[1][1];                                                                    \
    S[2] = St[2][2];                                                                    \
    S[3] = St[0][1];                                                                    \
    S[4] = St[0][2];                                                                    \
    S[5] = St[1][2];                                                                    \
  }
#define COMP_DU(K,delta_E)                      \
  {                                             \
    double dUv[6]={0,0,0,0,0,0};                \
    dUv[K] = delta_E;                           \
    gep_sgesl (&um, umpvt, dUv, GEP_JOB_A_X_B); \
    MAT3DUTIL_V_TO_T (dUv, dU);                 \
  }

  FOR (k, 6) {
    COMP_DU(k, (+e));
    MAT3DUTIL_2_MAT (U1, =, U);
    MAT3DUTIL_2_MAT (U1, +=, dU);
    MAT3DUTIL_MULT_AB_3X3 (custom_data.F, R, U1);
    GET_2ND_PK (Sp);
    MAT3DUTIL_2_MAT (U1, =, U);
    MAT3DUTIL_2_MAT (U1, -=, dU);
    MAT3DUTIL_MULT_AB_3X3 (custom_data.F, R, U1);
    GET_2ND_PK (Sn);
    FOR (l, 6) {
      C[k][l] = (Sp[l] - Sn[l]) / (2 * e) / mult[l];
    }
  }

#if 1
  /* symmetrize */
  FOR (k, 6) {
    FOR (l, 6) {
      if (k != l) {
        double a = (C[k][l] + C[l][k]) / 2;
        C[k][l] = C[l][k] = a;
      }
    }
  }
#endif

 exit:

  if (result) {
    /* I'm handling request for Lagrangian stiffness only;
     if Eulerian, should transform first. */
    if (stiff_type == MAT_3D_STIFF_LAGRANGIAN) return TRUE;
    else                                       return FALSE;
  } else
    return FALSE; /* have nothing to show for my effort :( */
}

static W_bool_t
get_mat_stiff_9x9 (W_mat_3d_t           *m/* mat3d */,
                   W_mat_3d_state_p     ms,/* material state */
                   W_mat_3d_ip_ctx_t    *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
                   W_mat_3d_stiff_enum  stiff_type, /* stiffness type */
                   /* double               Kmat[9][9] */
                   double               K[9][9]
                   )
{

}

static W_mat_3d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const,
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  get_mat_state_named_var,
  update_history,
  get_mat_stiff_9x9,
};
    
    
W_mat_3d_t *
read_mat_3d_rt_poly (W_domain_t *d)
{
  const char *input_line;
  double c11=0, c12=0, c44=0, rho = 1;
  double q, h_0, gamma_0, m_srs, g_0, g_inf;
  int debug = 0;
  int max_inner_iter = 20;
  double e_incr = 0;
  double activity_tol = 0.01;
  W_bool_t elastic = 0;
  W_bool_t rate_dependent = 0;
  W_mat_3d_t *mat = 0;
  int n = 1; /* single crystal */
    
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("C11", TRUE,
    			    { c11 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c11 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("C12", TRUE,
    			    { c12 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c12 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("C44", TRUE,
    			    { c44 = TOKEN_DBL_NEXT(1); },
    			    { REPORT_FATAL_ERROR (NULL, "c44 must be given"); }
    			    );
      TOKENS_SET_BY_KEYWORD("Q", FALSE,
    			    { q = TOKEN_DBL_NEXT(1); },
    			    { q = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("H_0", TRUE,
    			    { h_0 = TOKEN_DBL_NEXT(1); },
    			    { h_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("G_0", TRUE,
    			    { g_0 = TOKEN_DBL_NEXT(1); },
    			    { g_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("G_INF", TRUE,
    			    { g_inf = TOKEN_DBL_NEXT(1); },
    			    { g_inf = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("GAMMA_0", TRUE,
    			    { gamma_0 = TOKEN_DBL_NEXT(1); },
    			    { gamma_0 = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("M_SRS", TRUE,
    			    { m_srs = TOKEN_DBL_NEXT(1); },
    			    { m_srs = 0.02; }
    			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
    			    { rho = TOKEN_DBL_NEXT(1); },
    			    { rho = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("DEBUG", FALSE,
    			    { debug = TOKEN_INT_NEXT(1); },
                            { debug = 0; /* no debugging */ }
    			    );
      TOKENS_SET_BY_KEYWORD("MAX_INNER_ITER", FALSE,
    			    { max_inner_iter = TOKEN_INT_NEXT(1); },
                            { max_inner_iter = 20;  }
    			    );
      TOKENS_SET_BY_KEYWORD("E_INCR", FALSE,
    			    { e_incr = TOKEN_DBL_NEXT(1); },
    			    { e_incr = 0; }/* numerical approx of tangent */
    			    );
      TOKENS_SET_BY_KEYWORD("ACTIVITY_TOL", FALSE,
    			    { activity_tol = TOKEN_DBL_NEXT(1); },
    			    { activity_tol = 0.01; }
    			    );
      TOKENS_SET_BY_KEYWORD("N", FALSE,
    			    { n = TOKEN_INT_NEXT(1); },
    			    { n = 1; }
    			    );
      TOKENS_SET_BY_KEYWORD("ELASTIC", FALSE,
			    { elastic = TRUE; },
			    { elastic = FALSE; }
			    );
      TOKENS_SET_BY_KEYWORD("RATE_DEPENDENT", FALSE,
			    { rate_dependent = TRUE; },
			    { rate_dependent = FALSE; }
			    );

      {
        mat_t *u;
        const double eps = 1e-9;
        int i, j;
        
        u = w_new(mat_t);
        CHECK_MALLOC(u);

        u->rand = w_new_array(NUM_ORIENT, double);
        CHECK_MALLOC(u->rand);

        u->c11 = c11; 
        u->c12 = c12; 
        u->c44 = c44; 
        if (u->c11 < 0) REPORT_FATAL_ERROR(NULL, "Bad c11 (should be > 0)");
        if (u->c12 < 0) REPORT_FATAL_ERROR(NULL, "Bad c12 (should be > 0)");
        if (u->c44 < 0) REPORT_FATAL_ERROR(NULL, "Bad c44 (should be > 0)");

        u->gamma_0 = gamma_0;
        u->q = q;
        u->h_0 = h_0;
        u->m_srs = m_srs;
        u->g_0 = g_0;
        u->g_inf = g_inf;
        if (u->g_inf < u->g_0) {
          REPORT_FATAL_ERROR (NULL, "g_inf should be greater than g_0");
        }
        u->debug = debug;
        u->max_inner_iter = max_inner_iter;
        u->activity_tol = activity_tol;
        u->e_incr = e_incr;
        u->elastic = elastic;
        u->rate_dependent = rate_dependent;
        u->layers = n;
        
        /* generate the random numbers for each orientation
           check if the random number is good */
        {
          double under_root, dl = 1/(double)(u->layers - 1);
          double x, y, r, a, b, c;
          int counter = 0;
          
          FOR(i, u->layers) {
             for(j=0; j<=i; j++ ){
                x = (double)i * dl;
                y = (double)j * dl;
                under_root = -1.0; 
                while (under_root < 0){
                  r = rand_0_to_1();
                  a = x*x + y*y;
                  b = 2*x*r;
                  c = r*r*(1 + y*y) - y*y;
                  under_root = b*b -4*a*c;
                }
                u->rand[counter] = r;
                counter ++;
             }
          }          
        }

        mat = WM3D_new_material(&methods, (void *)u, rho); 
    
    	return mat;
    
      }/* new_mat_3d_rt_poly() */
    }
  }
  return NULL;
}
