#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Wtfuncs.h"
#include "Wmat3d.h"
#include "Wmat3dsmass1.h"

/* RCSHEADER "$Id: mat3dsmass1.c,v 1.4 2004/04/27 20:10:11 cvs Exp $" */

/*
  Small strain shape memory material.
  Version: 1
 */

typedef struct mat_t {
  double E, nu;
  double bulk_mod, G;

  double g0; /*param on the extent to which M_transformation strain is aligned*/
  double geff; /*detwinning "yield" stress */
  double M0f; /*M_transformation finish temperature (stress free) */
  double M0s; /*M_transformation start temperature (stress free)*/
  double A0f; /*A_transformation finish temperature (stress free)*/
  double A0s; /*A_transformation start temperature (stress free)*/
  double alpha; /*thermal expansion coefficient*/
  double sigma_Ay;
  double sigma_My;
  
  double lamda0, m, cm, ca;

  W_time_func_t *tf; /* time function for the temperature:
		      thermal strain = value(tf) * magnitude_of_isotropic_strain */
  char          *imposed_isotropic_strain_expr;  /* spatial variation
						    of the magnitude_of_isotropic_strain */
  double dil_wave_speed_3d;
  double dil_wave_speed_pstrain;
  double dil_wave_speed_pstress;
  double shear_wave_speed;
  double rayleigh_pstrain_wave_speed;
  double rayleigh_pstress_wave_speed;
  char   is_elastoplastic;      /* TRUE/FALSE */
}              mat_t; 

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_SMASS1;
}

static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;

  switch (type) {
    case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_TENSILE_STRENGTH:
    *c = 0;
    break;


  case MAT_3D_DYN_FRACTURE_TOUGHNESS:
    *c = 0;
    break;
  case MAT_3D_DIL_WAVE_SPEED_3D:
    *c = u->dil_wave_speed_3d;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRAIN:
    *c = u->dil_wave_speed_pstrain;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRESS:
    *c = u->dil_wave_speed_pstress;
    break;
  case MAT_3D_SHEAR_WAVE_SPEED:
    *c = u->shear_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED:
    *c = u->rayleigh_pstrain_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRESS_WAVE_SPEED:
    *c = u->rayleigh_pstress_wave_speed;
    break;

  case MAT_3D_SMA_MARTENSITE_INITIAL_YIELD_STRESS:
    *c = u->sigma_My;
    break;
  case MAT_3D_SMA_AUSTENITE_INITIAL_YIELD_STRESS: 
    *c = u->sigma_Ay;
    break;

  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

/* ============================================================== */

/* RAW transformation strain, ... */
typedef struct matstate_t {
  double   T_range;
  double   last_tfmult;
  double   equiv_pl_def;
  double   cauchy[6];
  double   equiv_stress;
  double   incr_equiv_stress;
  double   phase_strain_hat[6]; 
  double   sigma_y;
  double   Gamma; /*what's Gamma?*/
  double   deform_energy_density;
}              matstate_t;

static W_mat_3d_state_p 
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;
  int i;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  if (u->imposed_isotropic_strain_expr) {
    int stat;                                                   
    double v;
    
    expeval_set_value("x", at_point->x);
    expeval_set_value("y", at_point->y);
    expeval_set_value("z", at_point->z);
    stat = expeval_eval(u->imposed_isotropic_strain_expr, &v);   
    if (stat != EXPEVAL_E_OK) {                                 
      sprintf(W_buf(), "%s", expeval_err_msgs[stat]);               
      REPORT_ERROR(NULL, W_buf());                                        
    }
    ms->T_range = v;
    ms->last_tfmult = 0;
  }
 /* ms->equiv_pl_def = 0; */

  FOR(i, 6) { ms->cauchy[i] = 0; }

  ms->sigma_y = 0;

  ms->Gamma = 0;
  ms->deform_energy_density = 0;
  ms->equiv_stress = 0;

  return (W_mat_3d_state_p)ms;
}

static W_mat_3d_state_p 
interpolate_mat_state (W_mat_3d_t *m,
                       W_loc_t *at_point,
                       int npkgs,
                       WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    ms->equiv_pl_def += pms->equiv_pl_def * WGHT(k);
    FOR(i, 6) {
      ms->cauchy[i]      += pms->cauchy[i] * WGHT(k);
    }
    ms->sigma_y += pms->sigma_y * WGHT(k);
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_mat_state_var (W_mat_3d_t *m,
                   W_mat_3d_state_p ms,
                   W_mat_3d_ip_ctx_t *ip_ctx,
                   W_mat_3d_mat_state_var_enum  var_type,
                   double *var
                   )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = matst->equiv_pl_def; 
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}

static W_bool_t 
get_mat_state_named_var (W_mat_3d_t *m,
                         W_mat_3d_state_p ms,
                         W_mat_3d_ip_ctx_t *ip_ctx,
                         char *name,
                         double *var
                         )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

 
    *var = 0;
    return FALSE;

 exit:
  return TRUE;
}

/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress (W_mat_3d_t *m,
            W_mat_3d_state_p ms,
            W_mat_3d_ip_ctx_t *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  matstate_t *matst = (matstate_t *)ms;
  double sigma[3][3], cauchy[3][3];

  MAT3DUTIL_V_TO_T (matst->cauchy, cauchy);
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
  case MAT_3D_1ST_PK:
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}

typedef enum W_mat_3d_sma_phtrans_state_enum{
    MAT_3D_SMA_A_TO_M     = 0,
    MAT_3D_SMA_M_TO_A     = 1
}           W_mat_3d_sma_phtrans_state_enum;


static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
               )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  double G = u->G, K = u->bulk_mod;
#define SQRT_2_OVER_3 0.81649658092772603272  
 
  if (matst->Gamma == 0) {
    C_mat[0][0] = K +2*G*(0.666667);
    C_mat[0][1] = C_mat[1][0] = K +2*G*(-0.333333);
    C_mat[0][2] = C_mat[2][0] = K +2*G*(-0.333333);
    C_mat[0][3] = C_mat[3][0] = 0;
    C_mat[0][4] = C_mat[4][0] = 0;
    C_mat[0][5] = C_mat[5][0] = 0;
    C_mat[1][1] = K +2*G*(0.666667);
    C_mat[1][2] = C_mat[2][1] = K +2*G*(-0.333333);
    C_mat[1][3] = C_mat[3][1] = 0;
    C_mat[1][4] = C_mat[4][1] = 0;
    C_mat[1][5] = C_mat[5][1] = 0;
    C_mat[2][2] = K +2*G*(0.666667);
    C_mat[2][3] = C_mat[3][2] = 0;
    C_mat[2][4] = C_mat[4][2] = 0;
    C_mat[2][5] = C_mat[5][2] = 0;
    C_mat[3][3] = G;
    C_mat[3][4] = C_mat[4][3] = 0;
    C_mat[3][5] = C_mat[5][3] = 0;
    C_mat[4][4] = G;
    C_mat[4][5] = C_mat[5][4] = 0;
    C_mat[5][5] = G;
  } else {
    double Gamma = matst->Gamma;
    double sigma_y = matst->sigma_y;
    double B, gb, strdevn, pressure;
    double cauchydev[3][3], cauchy[3][3], alpha[3][3], N[3][3];
 
    MAT3DUTIL_V_TO_T (matst->cauchy, cauchy);
    pressure = MAT3DUTIL_TRACE_3X3 (cauchy) / 3;
    MAT3DUTIL_COPY_3X3 (cauchydev, cauchy);
    MAT3DUTIL_MAT_MOD_DIAG (cauchydev, -=, pressure);
    MAT3DUTIL_COPY_3X3 (N, cauchydev);
    strdevn = sqrt (MAT3DUTIL_DBL_PROD_3X3 (N, N));
    MAT3DUTIL_SCALE_3X3 (N, (1/strdevn));    /* normal unit */
 
    B = (SQRT_2_OVER_3 * sigma_y) / strdevn;
    gb = B;
 
    C_mat[0][0] = K +2*G*B*(0.666667) -2*G*gb*N[0][0]*N[0][0];
    C_mat[0][1] = C_mat[1][0] = K +2*G*B*(-0.333333) -2*G*gb*N[0][0]*N[1][1];
    C_mat[0][2] = C_mat[2][0] = K +2*G*B*(-0.333333) -2*G*gb*N[0][0]*N[2][2];
    C_mat[0][3] = C_mat[3][0] = ((-2*G*gb*N[0][0]*N[0][1]) + (-2*G*gb*N[0][0]*N[1][0]))/2;
    C_mat[0][4] = C_mat[4][0] = ((-2*G*gb*N[0][0]*N[0][2]) + (-2*G*gb*N[0][0]*N[2][0]))/2;
    C_mat[0][5] = C_mat[5][0] = ((-2*G*gb*N[0][0]*N[1][2]) + (-2*G*gb*N[0][0]*N[2][1]))/2;
    C_mat[1][1] = K +2*G*B*(0.666667) -2*G*gb*N[1][1]*N[1][1];
    C_mat[1][2] = C_mat[2][1] = K +2*G*B*(-0.333333) -2*G*gb*N[1][1]*N[2][2];
    C_mat[1][3] = C_mat[3][1] = ((-2*G*gb*N[1][1]*N[0][1]) + (-2*G*gb*N[1][1]*N[1][0]))/2;
    C_mat[1][4] = C_mat[4][1] = ((-2*G*gb*N[1][1]*N[0][2]) + (-2*G*gb*N[1][1]*N[2][0]))/2;
    C_mat[1][5] = C_mat[5][1] = ((-2*G*gb*N[1][1]*N[1][2]) + (-2*G*gb*N[1][1]*N[2][1]))/2;
    C_mat[2][2] = K +2*G*B*(0.666667) -2*G*gb*N[2][2]*N[2][2];
    C_mat[2][3] = C_mat[3][2] = ((-2*G*gb*N[2][2]*N[0][1]) + (-2*G*gb*N[2][2]*N[1][0]))/2;
    C_mat[2][4] = C_mat[4][2] = ((-2*G*gb*N[2][2]*N[0][2]) + (-2*G*gb*N[2][2]*N[2][0]))/2;
    C_mat[2][5] = C_mat[5][2] = ((-2*G*gb*N[2][2]*N[1][2]) + (-2*G*gb*N[2][2]*N[2][1]))/2;
    C_mat[3][3] = G*B + ((-2*G*gb*N[0][1]*N[0][1]) + (-2*G*gb*N[0][1]*N[1][0]))/2;
    C_mat[3][4] = C_mat[4][3] = ((-2*G*gb*N[0][1]*N[0][2]) + (-2*G*gb*N[0][1]*N[2][0]))/2;
    C_mat[3][5] = C_mat[5][3] = ((-2*G*gb*N[0][1]*N[1][2]) + (-2*G*gb*N[0][1]*N[2][1]))/2;
    C_mat[4][4] = G*B + ((-2*G*gb*N[0][2]*N[0][2]) + (-2*G*gb*N[0][2]*N[2][0]))/2;
    C_mat[4][5] = C_mat[5][4] = ((-2*G*gb*N[0][2]*N[1][2]) + (-2*G*gb*N[0][2]*N[2][1]))/2;
    C_mat[5][5] = G*B + ((-2*G*gb*N[1][2]*N[1][2]) + (-2*G*gb*N[1][2]*N[2][1]))/2;
  }
 
  return TRUE;
}


static W_bool_t  
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type,
               W_mat_3d_sma_phtrans_state_enum  phtrans_state 
               )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double lambda = u->E * u->nu / (1 + u->nu) / (1 - 2*(u->nu));
  double mu = u->G, lfact, mfact, trd;
  double prevcauchy[3][3], cauchy[3][3], incrcauchy[3][3];
  double incrcauchyv[6]; 
  double DU[3][3], DUT[3][3], deps[3][3];
  double deps_Th[3][3], deps_Pl[3][3];
#define SQRT_2_OVER_3 0.81649658092772603272
#define PI 3.141592654
  W_bool_t result = TRUE;
  matstate_t input_matst;

  double delta[6] = {1, 1, 1, 0, 0, 0};
  double cauchydev_predictor[3][3], N[3][3];        
  double n[6];
  double pressure;
  double incr_T;       

/* Quantities about M volume fraction */
  double f;
  double T;
  double f1,f2,f3,f4;
  double F_M, F_A;
  double es;
  double incr_lm, incr_la;
  double Mf,Ms,As,Af;
  double coeffa, coeffm;  
  double sigma_y;   
  /* g */
  double g;
  double g0 = 300;
  double a = -15;
  double b = 2;
  double g_der; 

  double E_TRIP;
  double E_TWIN;
  double E_Mph[6];
  double E_sigma[6];
  double E_T[6];
 
  double unit[6][6];
  double temp[3][3];    
  double tempv[6];
  double mat_stiff[6][6];
  W_mat_3d_stiff_enum stiff_type;

#define STRAIN_EQUIV_T 0.08
#define STRAIN_V_T 0.003
 
#define N_SYSTEMS 6

  double rcond, z[N_SYSTEMS];
  double *pd[N_SYSTEMS];
  double A[N_SYSTEMS][N_SYSTEMS];
  int Apvt[N_SYSTEMS];
  double depsv[6];
  double deps_Thv[6];
  double deps_Plv[6];
  double R[6];
 
  double cauchydev[3][3];
  double phase_strain_hat_tensor[3][3];
    double tfmult = 0;
    double t = 0;
    double s = 0;
    double incr_lamda; 
  int i;
 
  /* kinematic quantities */
  if (!ip_ctx->get_kinem_var) return FALSE;  

  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_INCR_DISPL_GRAD, DU)) {
    double F[3][3], prev_F[3][3];
    if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    } 
    if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_PREV_DEF_GRAD, prev_F)) {
      return FALSE;
    }
    MAT3DUTIL_3_MAT (DU, =, F, -, prev_F);
  }

  /* save current material state */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  /* increment of total strain */
  MAT3DUTIL_COPY_TRANSP_3X3 (DUT, DU);
  MAT3DUTIL_3_MAT (deps, =, 0.5 * DU, +, 0.5 * DUT);

  /* previous stress */ 
  MAT3DUTIL_V_TO_T (matst->cauchy, cauchy); 
  MAT3DUTIL_COPY_3X3 (prevcauchy, cauchy);

  /* N, n */
   pressure = MAT3DUTIL_TRACE_3X3 (cauchy) / 3;
   MAT3DUTIL_COPY_3X3 (cauchydev_predictor, cauchy);
   MAT3DUTIL_MAT_MOD_DIAG (cauchydev_predictor, -=, pressure);
   MAT3DUTIL_COPY_3X3 (N, cauchydev_predictor);
   MAT3DUTIL_SCALE_3X3(N, 3/(2*matst->equiv_stress));
 
  /* n1 = N11, ... , n4 = 2*N12, ... */
   (N, n);

  /* Thermal strain increment, plastic strain increment */
  MAT3DUTIL_ZERO_3X3(deps_Th);
  
/* if (matst->T_range != 0) { */
    tfmult = WTF_func_val(u->tf, ip_ctx->time);
    t = tfmult - matst->last_tfmult;
    s = matst->T_range * t;
    
    incr_T = s;
    deps_Th[0][0] -= s * u->alpha;
    deps_Th[1][1] -= s * u->alpha;
    deps_Th[2][2] -= s * u->alpha;
    
    incr_lamda = t * u->lamda0 * pow(matst->equiv_stress/matst->sigma_y, u->m);
    /* increment of plastic strain */
    MAT3DUTIL_2_MAT(deps_Pl, = , incr_lamda * N);
    
    matst->last_tfmult = tfmult;
/*  }*/

  	/* lfact = lambda;
	  mfact = 2 * mu;
	  trd = MAT3DUTIL_TRACE_3X3 (deps);
	  MAT3DUTIL_2_MAT (incrcauchy, =, mfact * deps);
	  MAT3DUTIL_MAT_MOD_DIAG (incrcauchy, +=, lfact * trd);
	*/

	/* total stress */
	/*  MAT3DUTIL_2_MAT (cauchy, +=, incrcauchy);*/

 /* increment of phase transformation strain */  

#define STRAIN_EQUIV_T 0.08
#define STRAIN_V_T 0.003

  /* Martensite volume fraction */
  
  T = matst->T_range*matst->last_tfmult;  
  es = matst->equiv_stress;
  f1 = es-u->cm*(T-u->M0f);
  f2 = es-u->cm*(T-u->M0s);
  f3 = es-u->ca*(T-u->A0s);
  f4 = es-u->ca*(T-u->A0f);
  
  Mf = u->M0f + es/u->cm;
  Ms = u->M0s + es/u->cm; 
  As = u->A0s + es/u->ca; 
  Af = u->A0f + es/u->ca; 

  coeffm = PI/(Ms-Mf);
  coeffa = PI/(Af-As);

  switch (phtrans_state) {
  case MAT_3D_SMA_A_TO_M:
    f = 0.5*cos(coeffm*(T-Mf)+1);
  break;
  case MAT_3D_SMA_M_TO_A:
    f = 0.5*cos(coeffa*(T-As)+1);  
  break;
  }

/* Save current sigma_y */
  sigma_y = (1-f) * pow(u->sigma_Ay, - u->m) + f * pow(u->sigma_My, - u->m);
  matst->sigma_y = pow(sigma_y,- 1/u->m);

  /* g */
  g = 1 - exp(a*pow(matst->equiv_stress/g0, b));

  /* g' */
  g_der = -g*a*b*pow(matst->equiv_stress/g0, b-1)  ;

  /* E_TRIP */
  E_TRIP = g*STRAIN_EQUIV_T;

  /* E_TWIN */
/* (1) !!!! only if increment of equiv_stress is positive and g is greater than geff!!! */
  if (matst->incr_equiv_stress>0 && (g - u->geff)>0)
    E_TWIN = f*g_der*STRAIN_EQUIV_T/g0;
 
  /* E_Mph */
  for ( i = 0; i < 6; i++ ){
    E_Mph[i] = matst->phase_strain_hat[i]; 
  }

/* (2) need to know what is incr_la and incr_lm */
  incr_lm = - matst->incr_equiv_stress/u->cm;
  incr_la = - matst->incr_equiv_stress/u->ca; 

  if((1-f)>0 && -f1*f2>0 && incr_lm>0)
    F_M = -coeffm*sin(coeffm*(T-Mf))/2;
  else
    F_M = 0;

  if(f>0 && -f3*f4>0 && incr_la>0)
    F_A = -coeffa*sin(coeffa*(T-As))/2;
  else
    F_A = 0;

  
  /* E_sigma--vector */
  for ( i = 0; i < 6; i++ ){
  E_sigma[i] = (E_TWIN - E_TRIP*F_M/u->cm)*n[i];
  E_sigma[i] -=  F_A*E_Mph[i]/u->ca;
  E_sigma[i] -=  STRAIN_V_T*delta[i]*F_M/u->cm; 
  }

  /* E_T--vector */
  for ( i = 0; i<6; i++){
  E_T[i] = E_TRIP*F_M*n[i]+E_Mph[i]*F_A+STRAIN_V_T*F_M*delta[i];
  } 
  
  get_mat_stiff(m, ms, ip_ctx, stiff_type, mat_stiff);
  MAT3DUTIL_MULT_A_6X6_V_6(tempv, mat_stiff, E_sigma);
  MAT3DUTIL_UNIT_6X6(unit);
  {  int j;
  for (i = 0; i<6; i++){
    for (j = 0; j<6; j++){
      A[i][j] = unit[i][j] + tempv[i] * n[j];
    }
  }
  }
  /* incrcauchyv = inv_A * mat_stiff * (deps - deps_Th - deps_Pl - E_T * incr_T) */


  {
    GEP_MATRIX (Am);
    
    GEP_STACKINIT(Am, N_SYSTEMS);
    
    { int j;    FOR(i, 6) FOR(j, 6) GEP_ELEM(Am, i, j) = A[i][j]; }
    gep_sgeco (&Am, Apvt, &rcond, z);
    
    /* Need to calculate Right hand side R first, R = mat_stiff * (deps - deps_Th - deps_Pl - E_T * incr_T) */
    /* What's incr_T */
    
    MAT3DUTIL_T_TO_V1(deps_Th, deps_Thv);                
    MAT3DUTIL_T_TO_V1(deps_Pl, deps_Plv);
    MAT3DUTIL_2_V(R, =, depsv);
    MAT3DUTIL_2_V(R, -=, deps_Thv);
    MAT3DUTIL_2_V(R, -=, deps_Plv);
    MAT3DUTIL_2_V(R, -=, incr_T * E_T);
    
    gep_sgesl (&Am, Apvt, R, GEP_JOB_A_X_B);
  }
  
  /* update and save new stress */
  
  MAT3DUTIL_V_TO_T(incrcauchyv, incrcauchy);
  MAT3DUTIL_2_MAT(cauchy, +=, incrcauchy);
  MAT3DUTIL_T_TO_V(cauchy, matst->cauchy);
  
  /* update and save incr_equiv_stress and equiv_stress */
  pressure = MAT3DUTIL_TRACE_3X3 (cauchy) / 3;
  MAT3DUTIL_COPY_3X3 (cauchydev, cauchy);
  MAT3DUTIL_MAT_MOD_DIAG (cauchydev, -=, pressure);
  es = sqrt (MAT3DUTIL_DBL_PROD_3X3 (cauchydev, cauchydev) * 3/2);
  matst->incr_equiv_stress = es - matst->equiv_stress;
  matst->equiv_stress = es;
  
  /* store phase_strain_hat */
  g = 1 - exp(a*pow(matst->equiv_stress/g0, b)); 
  MAT3DUTIL_2_MAT(phase_strain_hat_tensor, = , cauchydev);
  MAT3DUTIL_SCALE_3X3(phase_strain_hat_tensor, g * STRAIN_EQUIV_T * 3/2/matst->equiv_stress);
  MAT3DUTIL_UNIT_3X3(temp);
  MAT3DUTIL_SCALE_3X3(temp, STRAIN_V_T);
  MAT3DUTIL_INCR_A_BY_B_3X3(phase_strain_hat_tensor, temp);
  MAT3DUTIL_T_TO_V(phase_strain_hat_tensor, matst->phase_strain_hat);
  
  
  matst->deform_energy_density += (MAT3DUTIL_DBL_PROD_3X3 (prevcauchy, deps) +
                                   MAT3DUTIL_DBL_PROD_3X3 (cauchy, deps)
                                   ) / 2;
  
  if (stress_type == MAT_3D_CAUCHY) { /* accelerated handling for what we've got ready */
    /* rotate to obtain the Cauchy stress */
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    *out_stress_type = MAT_3D_CAUCHY;
    result = TRUE;
  } else { /* generic case */
    result = get_stress (m, ms, ip_ctx, stress_type, stress, out_stress_type);
  }
  exit:
    if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

#include "consttransf.c"

static void
update_history (W_mat_3d_t *m,
                W_mat_3d_state_p ms,
                W_mat_3d_ip_ctx_t *ip_ctx,
                W_solution_history_t *h, /* solution history */
                W_index_t pntidx,
                W_float_t time
                )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
#define STORE(HW, NAME, TIME, I, VAL)           \
  {                                             \
    char buf[128];                              \
    sprintf (buf, "p%d%s_%d", pntidx, NAME, I); \
    WHR_store (HW, buf, TIME, VAL);             \
  }
  int i;
#if 0
  double tau[N_SLIP_SYSTEMS];
  double F[3][3], Cebar[3][3], S[3][3];

  if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
    return;
  }
  calc_Cebar1 (F, matst->Fp, Cebar);
  MAT3DUTIL_V_TO_T (matst->S, S);
  calc_tau1 (Cebar, S, u->s, u->m, tau);

  FOR (i, N_SLIP_SYSTEMS) {
    STORE (h->writer, "g", time, i+1, matst->g[i]);
    STORE (h->writer, "gamma", time, i+1, matst->gamma[i]);
    STORE (h->writer, "tau", time, i+1, tau[i]);
  }
#endif
}

static W_mat_3d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const,
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  get_mat_state_named_var,
  update_history
};

/*
  Create a new material.
 */
W_mat_3d_t *
new_mat_3d_smass1 (double E, double nu, double lamda0, double m, double alpha, 
                   double sigma_Ay, double sigma_My, 
                   double g0, double geff, double cm, double ca,
                   double M0f, double M0s, double A0s, double A0f,
                   char *imposed_isotropic_strain_expr, W_time_func_t *tf, 
                   double rho)
{
  mat_t *u;
  
  u = w_new(mat_t);
  CHECK_MALLOC(u);
  u->E                      = E; 
  u->nu                     = nu;
  if (u->nu >= 0.5) REPORT_FATAL_ERROR(NULL, "Bad nu (should be 0 <= nu < 0.5)");
  u->bulk_mod               = u->E / 3 / (1 - 2*u->nu);
  u->G                      = u->E / 2 / (1 + u->nu);
  
  u->M0f		    = M0f;
  u->M0s		    = M0s;
  u->A0s		    = A0s;
  u->A0f		    = A0f;
  u->sigma_Ay               = sigma_Ay;
  u->sigma_My		    = sigma_My;
  
/*  if (u->sigma_y == 0) u->is_elastoplastic = FALSE;
  else                  u->is_elastoplastic = TRUE;*/

  u->dil_wave_speed_3d
    = W_3D_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstrain
    = W_plstrain_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstress
    = W_plstress_dilatational_wave_speed (u->E, u->nu, rho);
  u->shear_wave_speed
    = W_shear_wave_speed (u->E, u->nu, rho);
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstress,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstress_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane stress speed does not exist");
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstrain,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstrain_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane strain speed does not exist");
  
  return WM3D_new_material (&methods, (void *)u, rho);
}


W_mat_3d_t *
read_mat_3d_smass1 (W_domain_t *d)
{
  const char *input_line;
  double E, nu, rho = 1;
  double sigma_y = 0;
  double tensile_strength = 0;
  double g0;
  double geff;
  double M0f;
  double M0s;
  double A0s;
  double A0f;
  double cm, ca;
  double sigma_My, sigma_Ay;
  double alpha, m, lamda0;
  double dyn_fracture_toughness = 0;
  W_bool_t imposed_set = FALSE;
  char  imposed_isotropic_strain_expr[1024];
  W_index_t time_func = 1;
  W_time_func_t *tf = NULL;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { E = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { nu = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { tensile_strength = TOKEN_DBL_NEXT(1); },
			    { tensile_strength = E / 100; }
			    );

      TOKENS_SET_BY_KEYWORD("G0", TRUE,
                            { g0 = TOKEN_DBL_NEXT(1); },
                            { g0 = 0; }
                            );

      TOKENS_SET_BY_KEYWORD("GEFF", TRUE,
                            { geff = TOKEN_DBL_NEXT(1); },
                            { geff = 0; }
                            );

      TOKENS_SET_BY_KEYWORD("MF0", TRUE,
                            { M0f = TOKEN_DBL_NEXT(1); },
                            { M0f = 0; }
                            );
      TOKENS_SET_BY_KEYWORD("MS0", TRUE,
                            { M0s = TOKEN_DBL_NEXT(1); },
                            { M0s = 0; }
                            ); 
      TOKENS_SET_BY_KEYWORD("AS0", TRUE,
                            { A0s = TOKEN_DBL_NEXT(1); },
                            { A0s = 0; }
                            );
      TOKENS_SET_BY_KEYWORD("AF0", TRUE,
                            { A0f = TOKEN_DBL_NEXT(1); },
                            { A0f = 0; }
                            );


      TOKENS_SET_BY_KEYWORD("DYN_FRACTURE_TOUGHNESS", FALSE,
			    { dyn_fracture_toughness = TOKEN_DBL_NEXT(1); },
			    { dyn_fracture_toughness = E / 4000; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );

      TOKENS_SET_BY_KEYWORD("CM", TRUE,
                            { cm = TOKEN_DBL_NEXT(1); },
                            { cm = 1; }
                            );

      TOKENS_SET_BY_KEYWORD("CA", TRUE,
                            { ca = TOKEN_DBL_NEXT(1); },
                            { ca = 1; }
                            );

      TOKENS_SET_BY_KEYWORD("SIGMA_MY", TRUE,
                            { sigma_My = TOKEN_DBL_NEXT(1); },
                            { sigma_My = 0; }
                            );

      TOKENS_SET_BY_KEYWORD("SIGMA_AY", TRUE,
                            { sigma_Ay = TOKEN_DBL_NEXT(1); },
                            { sigma_Ay = 0; }
                            );

      TOKENS_SET_BY_KEYWORD("THERM_EXPANSION", TRUE,
                            { alpha = TOKEN_DBL_NEXT(1); },
                            { alpha = 1; }
                            );

      TOKENS_SET_BY_KEYWORD("INIT_EQUIV_PLS_RATE", TRUE,
                            { lamda0 = TOKEN_DBL_NEXT(1); },
                            { lamda0 = 0; }
                            );
      TOKENS_SET_BY_KEYWORD("POWER", TRUE,

                            { m = TOKEN_DBL_NEXT(1); },
                            { m = 0; }
                            );

      TOKENS_SET_BY_KEYWORD("IMPOSED_ISOTROPIC_STRAIN", FALSE,
                            { sprintf (imposed_isotropic_strain_expr, "%s",
                                       TOKEN_STR_NEXT(1));
                              imposed_set = TRUE; },
                            { imposed_set = FALSE; } /* no strain */
			    );
      if (imposed_set) {
        TOKENS_SET_BY_KEYWORD("TIME_FUNC", FALSE,
                              { time_func = TOKEN_INT_NEXT(1); },
                              { time_func = 1; } 
                              );
        tf = WD_get_time_func_by_idx (d, time_func);
        if (tf == NULL) {
          sprintf (W_buf (), "Time function #%d not found", time_func);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
      }

      return new_mat_3d_smass1 (E, nu, lamda0, m, alpha,
                                sigma_Ay, sigma_My,
                                g0, geff, cm, ca,
                                M0f, M0s, A0s, A0f,
                                (imposed_set ? imposed_isotropic_strain_expr : 0), tf, rho);
    }
  }
  return NULL;
}
