#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dssj2ep.h"

/* RCSHEADER "$Id: mat3dssj2ep.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

/*
  Small strain J-2 elastoplastic material.
 */

typedef struct mat_t {
  double E, nu;
  double bulk_mod, G;
  double sigma_y, hardening_param, delta;
  double dil_wave_speed_3d;
  double dil_wave_speed_pstrain;
  double dil_wave_speed_pstress;
  double shear_wave_speed;
  double rayleigh_pstrain_wave_speed;
  double rayleigh_pstress_wave_speed;
  double tensile_strength;
  double dyn_fracture_toughness;
  char   is_elastoplastic;      /* TRUE/FALSE */
}              mat_t; 

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_SSJ2EP;
}

static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_TENSILE_STRENGTH:
    *c = u->tensile_strength; 
    break;
  case MAT_3D_DYN_FRACTURE_TOUGHNESS:
    *c = u->dyn_fracture_toughness; 
    break;
  case MAT_3D_DIL_WAVE_SPEED_3D:
    *c = u->dil_wave_speed_3d;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRAIN:
    *c = u->dil_wave_speed_pstrain;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRESS:
    *c = u->dil_wave_speed_pstress;
    break;
  case MAT_3D_SHEAR_WAVE_SPEED:
    *c = u->shear_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED:
    *c = u->rayleigh_pstrain_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRESS_WAVE_SPEED:
    *c = u->rayleigh_pstress_wave_speed;
    break;
  case MAT_3D_INITIAL_YIELD_STRESS:
    *c = u->sigma_y;
    break;
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

/* ============================================================== */

typedef struct matstate_t {
  double   equiv_pl_def;
  double   cauchy[6]; 
  double   back_stress[6];      
  double   sigma_y;
  double   Gamma;
  double   deform_energy_density;
}              matstate_t;

static W_mat_3d_state_p 
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;
  int i;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  ms->equiv_pl_def = 0;
  FOR(i, 6) { ms->cauchy[i] = 0; ms->back_stress[i] = 0; }
  ms->sigma_y = u->sigma_y;
  ms->Gamma = 0;
  ms->deform_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

static W_mat_3d_state_p 
interpolate_mat_state (W_mat_3d_t *m,
                       W_loc_t *at_point,
                       int npkgs,
                       WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    ms->equiv_pl_def += pms->equiv_pl_def * WGHT(k);
    FOR(i, 6) {
      ms->cauchy[i]      += pms->cauchy[i] * WGHT(k);
      ms->back_stress[i] += pms->back_stress[i] * WGHT(k);
    }
    ms->sigma_y += pms->sigma_y * WGHT(k);
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_mat_state_var (W_mat_3d_t *m,
                   W_mat_3d_state_p ms,
                   W_mat_3d_ip_ctx_t *ip_ctx,
                   W_mat_3d_mat_state_var_enum  var_type,
                   double *var
                   )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = matst->equiv_pl_def; 
    return TRUE;
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  default:
    return FALSE;
  }
}
/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress (W_mat_3d_t *m,
            W_mat_3d_state_p ms,
            W_mat_3d_ip_ctx_t *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  matstate_t *matst = (matstate_t *)ms;
  double sigma[3][3], cauchy[3][3];

  MAT3DUTIL_V_TO_T (matst->cauchy, cauchy);
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
  case MAT_3D_1ST_PK:
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}

static W_bool_t  
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double lambda = u->E * u->nu / (1 + u->nu) / (1 - 2*(u->nu));
  double mu = u->G, lfact, mfact, trd;
  double prevcauchy[3][3], cauchy[3][3], incrcauchy[3][3];
  double DU[3][3], DUT[3][3], d[3][3];
#define SQRT_2_OVER_3 0.81649658092772603272
  W_bool_t result = TRUE;
  matstate_t input_matst;
 
  /* kinematic quantities */
  if (!ip_ctx->get_kinem_var) return FALSE;  

  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_INCR_DISPL_GRAD, DU)) {
    double F[3][3], prev_F[3][3];
    if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    } 
    if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_PREV_DEF_GRAD, prev_F)) {
      return FALSE;
    }
    MAT3DUTIL_3_MAT (DU, =, F, -, prev_F);
  }

  /* save current material state */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  MAT3DUTIL_V_TO_T (matst->cauchy, cauchy); /* previous stress */
  MAT3DUTIL_COPY_3X3 (prevcauchy, cauchy);

  /* compute the increment of stress */
  MAT3DUTIL_COPY_TRANSP_3X3 (DUT, DU);
  MAT3DUTIL_3_MAT (d, =, 0.5 * DU, +, 0.5 * DUT);

  lfact = lambda;
  mfact = 2 * mu;
  trd = MAT3DUTIL_TRACE_3X3 (d);
  MAT3DUTIL_2_MAT (incrcauchy, =, mfact * d);
  MAT3DUTIL_MAT_MOD_DIAG (incrcauchy, +=, lfact * trd);

  /* total stress */
  MAT3DUTIL_2_MAT (cauchy, +=, incrcauchy);
 
  if (u->is_elastoplastic) { /* ie. if material may plasticize */
    double pressure, strdevn, Hp = u->hardening_param;
    double cauchydev_predictor[3][3], N[3][3];

    /* elastic predictor */
    pressure = MAT3DUTIL_TRACE_3X3 (cauchy) / 3;
    MAT3DUTIL_COPY_3X3 (cauchydev_predictor, cauchy);
    MAT3DUTIL_MAT_MOD_DIAG (cauchydev_predictor, -=, pressure);
    MAT3DUTIL_COPY_3X3 (N, cauchydev_predictor);
    if (u->delta != 1.0) {
      double alpha_predictor[3][3];
      MAT3DUTIL_V_TO_T (matst->back_stress, alpha_predictor);
      MAT3DUTIL_2_MAT (N, -=, alpha_predictor);
    }
    matst->Gamma = 0; /* elastic predictor */
    strdevn = sqrt (MAT3DUTIL_DBL_PROD_3X3 (N, N));
   
    if (strdevn > SQRT_2_OVER_3 * matst->sigma_y) { /* if plastic, correct */
      double sc;
      MAT3DUTIL_SCALE_3X3 (N, (1/strdevn));    /* normal unit */
      /* correct t deviator */
      matst->Gamma = ((strdevn - SQRT_2_OVER_3 * matst->sigma_y)
                      / (2 * mu * (1 + Hp / 3 / mu)));
      sc = -2 * mu * matst->Gamma;
      MAT3DUTIL_3_MAT (cauchy, =, cauchydev_predictor, +, sc * N);
      MAT3DUTIL_MAT_MOD_DIAG (cauchy, +=, pressure); /* add pressure */
      /* update */
      matst->equiv_pl_def += SQRT_2_OVER_3 * matst->Gamma;
      matst->sigma_y += SQRT_2_OVER_3 * u->delta * Hp * matst->Gamma;
      if (u->delta != 1.0) {
        double Nv[6], f = 2 / 3 * (1 - u->delta) * Hp * matst->Gamma;
        MAT3DUTIL_T_TO_V (N, Nv);
        MAT3DUTIL_2_V (matst->back_stress, +=, f * Nv);
      }      
    }
  } /* if the material is elastoplastic */
 
  /* save state: new stress */
  MAT3DUTIL_T_TO_V (cauchy, matst->cauchy);

  matst->deform_energy_density += (MAT3DUTIL_DBL_PROD_3X3 (prevcauchy, d) +
                                   MAT3DUTIL_DBL_PROD_3X3 (cauchy, d)
                                   ) / 2;

  if (stress_type == MAT_3D_CAUCHY) { /* accelerated handling for what we've got ready */
  /* rotate to obtain the Cauchy stress */
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    *out_stress_type = MAT_3D_CAUCHY;
    result = TRUE;
  } else { /* generic case */
    result = get_stress (m, ms, ip_ctx, stress_type, stress, out_stress_type);
  }
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

#include "consttransf.c"

static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
               )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  double G = u->G, K = u->bulk_mod;

  if (matst->Gamma == 0) {
    C_mat[0][0] = K +2*G*(0.666667);
    C_mat[0][1] = C_mat[1][0] = K +2*G*(-0.333333);
    C_mat[0][2] = C_mat[2][0] = K +2*G*(-0.333333);
    C_mat[0][3] = C_mat[3][0] = 0;
    C_mat[0][4] = C_mat[4][0] = 0;
    C_mat[0][5] = C_mat[5][0] = 0;
    C_mat[1][1] = K +2*G*(0.666667);
    C_mat[1][2] = C_mat[2][1] = K +2*G*(-0.333333);
    C_mat[1][3] = C_mat[3][1] = 0;
    C_mat[1][4] = C_mat[4][1] = 0;
    C_mat[1][5] = C_mat[5][1] = 0;
    C_mat[2][2] = K +2*G*(0.666667);
    C_mat[2][3] = C_mat[3][2] = 0;
    C_mat[2][4] = C_mat[4][2] = 0;
    C_mat[2][5] = C_mat[5][2] = 0;
    C_mat[3][3] = G;
    C_mat[3][4] = C_mat[4][3] = 0;
    C_mat[3][5] = C_mat[5][3] = 0;
    C_mat[4][4] = G;
    C_mat[4][5] = C_mat[5][4] = 0;
    C_mat[5][5] = G;
  } else {
    double Hp = u->hardening_param;
    double Gamma = matst->Gamma;
    double sigma_y = matst->sigma_y;
    double B, gb, strdevn, pressure;
    double cauchydev[3][3], cauchy[3][3], alpha[3][3], N[3][3];
    
    MAT3DUTIL_V_TO_T (matst->cauchy, cauchy);
    pressure = MAT3DUTIL_TRACE_3X3 (cauchy) / 3;
    MAT3DUTIL_COPY_3X3 (cauchydev, cauchy);
    MAT3DUTIL_MAT_MOD_DIAG (cauchydev, -=, pressure); 
    MAT3DUTIL_COPY_3X3 (N, cauchydev);
    if (u->delta != 1) { /* mixed hardening */
      MAT3DUTIL_V_TO_T (matst->back_stress, alpha);
      MAT3DUTIL_2_MAT (N, -=, alpha);
    }
    strdevn = sqrt (MAT3DUTIL_DBL_PROD_3X3 (N, N));
    MAT3DUTIL_SCALE_3X3 (N, (1/strdevn));    /* normal unit */
    
    B = (SQRT_2_OVER_3 * sigma_y + 2.0/3*(1 - u->delta)*Hp*Gamma) / strdevn;
    gb = 1 / (1 + Hp/3/G) + B - 1;
    
    C_mat[0][0] = K +2*G*B*(0.666667) -2*G*gb*N[0][0]*N[0][0];
    C_mat[0][1] = C_mat[1][0] = K +2*G*B*(-0.333333) -2*G*gb*N[0][0]*N[1][1];
    C_mat[0][2] = C_mat[2][0] = K +2*G*B*(-0.333333) -2*G*gb*N[0][0]*N[2][2];
    C_mat[0][3] = C_mat[3][0] = ((-2*G*gb*N[0][0]*N[0][1]) + (-2*G*gb*N[0][0]*N[1][0]))/2;
    C_mat[0][4] = C_mat[4][0] = ((-2*G*gb*N[0][0]*N[0][2]) + (-2*G*gb*N[0][0]*N[2][0]))/2;
    C_mat[0][5] = C_mat[5][0] = ((-2*G*gb*N[0][0]*N[1][2]) + (-2*G*gb*N[0][0]*N[2][1]))/2;
    C_mat[1][1] = K +2*G*B*(0.666667) -2*G*gb*N[1][1]*N[1][1];
    C_mat[1][2] = C_mat[2][1] = K +2*G*B*(-0.333333) -2*G*gb*N[1][1]*N[2][2];
    C_mat[1][3] = C_mat[3][1] = ((-2*G*gb*N[1][1]*N[0][1]) + (-2*G*gb*N[1][1]*N[1][0]))/2;
    C_mat[1][4] = C_mat[4][1] = ((-2*G*gb*N[1][1]*N[0][2]) + (-2*G*gb*N[1][1]*N[2][0]))/2;
    C_mat[1][5] = C_mat[5][1] = ((-2*G*gb*N[1][1]*N[1][2]) + (-2*G*gb*N[1][1]*N[2][1]))/2;
    C_mat[2][2] = K +2*G*B*(0.666667) -2*G*gb*N[2][2]*N[2][2];
    C_mat[2][3] = C_mat[3][2] = ((-2*G*gb*N[2][2]*N[0][1]) + (-2*G*gb*N[2][2]*N[1][0]))/2;
    C_mat[2][4] = C_mat[4][2] = ((-2*G*gb*N[2][2]*N[0][2]) + (-2*G*gb*N[2][2]*N[2][0]))/2;
    C_mat[2][5] = C_mat[5][2] = ((-2*G*gb*N[2][2]*N[1][2]) + (-2*G*gb*N[2][2]*N[2][1]))/2;
    C_mat[3][3] = G*B + ((-2*G*gb*N[0][1]*N[0][1]) + (-2*G*gb*N[0][1]*N[1][0]))/2;
    C_mat[3][4] = C_mat[4][3] = ((-2*G*gb*N[0][1]*N[0][2]) + (-2*G*gb*N[0][1]*N[2][0]))/2;
    C_mat[3][5] = C_mat[5][3] = ((-2*G*gb*N[0][1]*N[1][2]) + (-2*G*gb*N[0][1]*N[2][1]))/2;
    C_mat[4][4] = G*B + ((-2*G*gb*N[0][2]*N[0][2]) + (-2*G*gb*N[0][2]*N[2][0]))/2;
    C_mat[4][5] = C_mat[5][4] = ((-2*G*gb*N[0][2]*N[1][2]) + (-2*G*gb*N[0][2]*N[2][1]))/2;
    C_mat[5][5] = G*B + ((-2*G*gb*N[1][2]*N[1][2]) + (-2*G*gb*N[1][2]*N[2][1]))/2;
  }
  
  return TRUE;
}

static W_mat_3d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const,
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  NULL,
  NULL
};

/*
  Create a new material.
 */
W_mat_3d_t *
new_mat_3d_ssj2ep (double E, double nu,
                   double tensile_strength, double dyn_fracture_toughness,
                   double sigma_y, double hardening_param,
                   double delta, double rho)
{
  mat_t *u;
  
  u = w_new(mat_t);
  CHECK_MALLOC(u);
  u->E                      = E; 
  u->nu                     = nu;
  u->tensile_strength       = tensile_strength;
  u->dyn_fracture_toughness = dyn_fracture_toughness;
  if (u->nu >= 0.5) REPORT_FATAL_ERROR(NULL, "Bad nu (should be 0 <= nu < 0.5)");
  u->bulk_mod               = u->E / 3 / (1 - 2*u->nu);
  u->G                      = u->E / 2 / (1 + u->nu);
  u->sigma_y                = sigma_y;
  if (u->sigma_y == 0) u->is_elastoplastic = FALSE;
  else                 u->is_elastoplastic = TRUE;
  u->hardening_param        = hardening_param;
  u->delta                  = delta;
  u->dil_wave_speed_3d
    = W_3D_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstrain
    = W_plstrain_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstress
    = W_plstress_dilatational_wave_speed (u->E, u->nu, rho);
  u->shear_wave_speed
    = W_shear_wave_speed (u->E, u->nu, rho);
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstress,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstress_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane stress speed does not exist");
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstrain,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstrain_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane strain speed does not exist");
  
  return WM3D_new_material (&methods, (void *)u, rho);
}


W_mat_3d_t *
read_mat_3d_ssj2ep (W_domain_t *d)
{
  const char *input_line;
  double E, nu, rho = 1;
  double sigma_y = 0, hardening_param = 0, delta = 1;
  double tensile_strength = 0;
  double dyn_fracture_toughness = 0;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { E = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { nu = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { tensile_strength = TOKEN_DBL_NEXT(1); },
			    { tensile_strength = E / 100; }
			    );
      TOKENS_SET_BY_KEYWORD("DYN_FRACTURE_TOUGHNESS", FALSE,
			    { dyn_fracture_toughness = TOKEN_DBL_NEXT(1); },
			    { dyn_fracture_toughness = E / 4000; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("SIGMA_Y", FALSE,
			    {
			      sigma_y = TOKEN_DBL_NEXT(1);
			    },
			    {
                              sigma_y = 0;
                            }
			    );
      TOKENS_SET_BY_KEYWORD("DELTA", FALSE,
			    { delta = TOKEN_DBL_NEXT(1); },
			    { delta = 1; } /* isotropic hardening  */
			    );
      TOKENS_SET_BY_KEYWORD("HARDENING_PARAM", FALSE,
			    { hardening_param = TOKEN_DBL_NEXT(1); },
			    { hardening_param = 0; }/* perfectly ep material */
			    );
      return new_mat_3d_ssj2ep (E, nu, tensile_strength, dyn_fracture_toughness,
                                sigma_y, hardening_param, delta, rho);
    }
  }
  return NULL;
}


