#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dred2d.h"
#include "Wmat3dunrot.h"
#include "Wmat3dssj2ep.h"

/* RCSHEADER "$Id: mat3durd2d.c,v 1.2 2002/03/08 20:04:20 cvs Exp $" */

typedef struct mat_t {
  W_mat_3d_t *ssmat; /* delegate material update to this material */
}              mat_t; 

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_RED_2D;
}

static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{ 
  mat_t *u = (mat_t *)m->data;

  return WM3D_get_mat_const (u->ssmat, type, c);
}

typedef struct matstate_t {
  W_mat_3d_state_t  ssms;
  double            transverse_stretch;
}              matstate_t;

static W_mat_3d_state_p
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;
  int i;

  ms = w_new (matstate_t); CHECK_MALLOC(ms);

  ms->ssms = WM3D_get_mat_state (u->ssmat, at_point);

  return (W_mat_3d_state_p)ms;
}

static W_mat_3d_state_t 
interpolate_mat_state (W_mat_3d_t *m,
                       W_loc_t *at_point,
                       int npkgs,
                       WM3D_mat_state_pkg_t *pkgs)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms = (matstate_t *) get_mat_state (m, at_point);
  
  return WM3D_interpolate_mat_state (u->ssmat, at_point, npkgs, pkgs);
}

typedef struct ip_ctx_data_t {
  matstate_t        *ms;
  W_mat_3d_ip_ctx_t *up_ip_ctx;
  /* kinematic variables */
  double F[3][3];
  double prev_F[3][3];
  double rel_F[3][3];
  double midstep_F[3][3];
  double C[3][3];
  double E[3][3];
  double midstep_D[3][3];
  double midstep_W[3][3];
  double R[3][3];
  double U[3][3];
  double rel_R[3][3];
  double rel_U[3][3];
  double ln_rel_U[3][3];
  double F_dot[3][3];
}              ip_ctx_data_t;

/*
  NOTE: this routine does not cache a validity flag with
  each kinematic tensor as the element routines do.  It should
  be implemented to compute quantities just-in-time if this
  proves a computational bottleneck.
 */
static void
compute_kinem_quantities (W_mat_3d_ip_ctx_t *ctx)
{
  ip_ctx_data_t *ipdata = (ip_ctx_data_t *)ctx->ctx_data;
  double t = ctx->time;
  double dt = ctx->time_step;

  /* Cauchy-Green tensor */
  MAT3DUTIL_MULT_ATB_3X3 (ipdata->C, ipdata->F, ipdata->F);

  /* Green-Lagrange strain tensor */
  {
    double unit[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
    MAT3DUTIL_COMB_AB_3X3 (ipdata->E, 0.5, ipdata->C, - 0.5, unit);
  }

  /* relative deformation gradient */
  {
    double inv_prev_F[3][3];
    double det_prev_F = MAT3DUTIL_DET_3X3(ipdata->prev_F);
    if (det_prev_F <= 0) {
      sprintf(W_buf(), "Failed to invert def grad");
      REPORT_FATAL_ERROR(NULL, W_buf());
    }
    
    MAT3DUTIL_INVERT_3X3(det_prev_F, ipdata->prev_F, inv_prev_F);
    
    MAT3DUTIL_MULT_AB_3X3(ipdata->rel_F, ipdata->F, inv_prev_F);
  }

  /* midstep deformation gradient */
  MAT3DUTIL_COMB_AB_3X3 (ipdata->midstep_F, 0.5, ipdata->F, 0.5, ipdata->prev_F);

  /* midstep rate of deformation and spin */
  {
    int i, j;
    double m[3][3] = { {1, 0, 0}, {0, 1, 0}, {0, 0, 1} };
    double M[3][3] = { {-1, 0, 0}, {0, -1, 0}, {0, 0, -1} };
    double s[3][3], inv_s[3][3], q[3][3], r[3][3];
    double inv_dt = 1 / dt;
    
    MAT3DUTIL_ADD_AB_3X3 (s, m, ipdata->rel_F);
    MAT3DUTIL_SCALE_3X3 (s, 0.5);
    if (!W_mat3dutil_invert_3x3(s, inv_s)) {
      sprintf(W_buf(), "Failed to invert def grad");
      REPORT_FATAL_ERROR(NULL, W_buf());
    }
    MAT3DUTIL_ADD_AB_3X3 (q, ipdata->rel_F, M);
    MAT3DUTIL_MULT_AB_3X3 (r, inv_s, q);
    
    MAT3DUTIL_SYMM  (ipdata->midstep_D, r);
    MAT3DUTIL_ASYMM (ipdata->midstep_W, r);
    
    MAT3DUTIL_SCALE_3X3 (ipdata->midstep_D, inv_dt);
    MAT3DUTIL_SCALE_3X3 (ipdata->midstep_W, inv_dt);
  }

  /* polar decomposition */
  if (!W_polar_decomposition (ipdata->F, ipdata->R, ipdata->U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }

  /* relative polar decomposition */
  if (!W_polar_decomposition (ipdata->rel_F, ipdata->rel_R, ipdata->rel_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }

  /* logarithm of the relative deformation tensor */
  W_ln_U (ipdata->rel_U, ipdata->ln_rel_U);
}

static W_bool_t 
get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
               W_mat_3d_kinem_var_type_enum  type,
               double kinvar[3][3])
{
  ip_ctx_data_t *ipdata = (ip_ctx_data_t *)ctx->ctx_data;

  compute_kinem_quantities (ctx);

  switch (type) {
  case MAT_3D_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->F);
    break;
  case MAT_3D_DEF_GRAD_DOT:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->F_dot);
    break;
  case MAT_3D_REL_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->rel_F);
    break;
  case MAT_3D_PREV_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->prev_F);
    break;
  case MAT_3D_MIDSTEP_DEF_GRAD:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->midstep_F);
    break;
  case MAT_3D_CAUCHY_GREEN:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->C);
    break;
  case MAT_3D_MIDSTEP_RATE_OF_DEF:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->midstep_D);
    break;
  case MAT_3D_MIDSTEP_SPIN:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->midstep_W);
    break;
  case MAT_3D_GREEN_LAGRANGE:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->E);
    break;
  case MAT_3D_POLAR_R:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->R);
    break;
  case MAT_3D_POLAR_U:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->U);
    break;
  case MAT_3D_POLAR_REL_R:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->rel_R);
    break;
  case MAT_3D_POLAR_REL_U:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->rel_U);
    break;
  case MAT_3D_LN_POLAR_REL_U:
    MAT3DUTIL_COPY_3X3(kinvar, ipdata->ln_rel_U);
    break;
  default:
    {
      static W_bool_t complained = FALSE;
      if (!complained) {
        sprintf (W_buf(), "Unknown kinematic tensor type: %d",     (int)type);
        REPORT_WARNING (NULL, W_buf());
        complained = TRUE;
      }
    }
    return FALSE;
  }

  return TRUE;
}

static void
setup_deformation_gradients (ip_ctx_data_t *ipdata,
                             double prev_dudX, double prev_stretch,
                             double dudX, double stretch)
{
  MAT3DUTIL_ZERO_3X3 (ipdata->F);
  ipdata->F[0][0] = 1 + dudX;
  ipdata->F[1][1] = ipdata->F[2][2] = stretch;
  MAT3DUTIL_ZERO_3X3 (ipdata->prev_F);
  ipdata->prev_F[0][0] = 1 + prev_dudX;
  ipdata->prev_F[1][1] = ipdata->prev_F[2][2] = prev_stretch;
}

typedef struct ckit_fminbr_client_t {
  W_mat_3d_t               *m;
  W_mat_3d_state_p          matst;
  W_mat_3d_ip_ctx_t        *ip_ctx;
  double                    s_scale;
  W_mat_3d_stress_type_enum stress_type;
}              ckit_fminbr_client_t;

static double
eval_constraint (void *client, double stretch)
{
  ckit_fminbr_client_t *c = (ckit_fminbr_client_t *) client;
  ip_ctx_data_t *ipdata = (ip_ctx_data_t *)c->ip_ctx->ctx_data;
  double stress[3][3];
  W_mat_3d_stress_type_enum out_stress_type;
  double fval;
  /* modify deformation gradient */
  ipdata->F[2][2] = stretch;
  /* temporarily update the material state */
  if (!WM3D_update_stress (c->m, c->matst, W_MAT_STATE_UPDATE_TEMP,
                           c->ip_ctx, c->stress_type, stress, &out_stress_type))
    REPORT_FATAL_ERROR (NULL, "Update failed");
  /* evaluate the zero-in-plane-stress constraint */
  if (c->s_scale == 0) {
    double E, sigma_y;
    if (WM3D_get_mat_const (c->m, MAT_3D_YOUNG_MODULUS, &E)) {
      if (WM3D_get_mat_const (c->m, MAT_3D_INITIAL_YIELD_STRESS, &sigma_y)) {
        c->s_scale = sigma_y;
      } else {
        c->s_scale = fabs (E * (ipdata->F[0][0] - 1));
      }
    } else {
      REPORT_FATAL_ERROR (NULL, "No Young's modulus");
    }
  }
  fval = powi ((stress[1][1] + stress[2][2])/2 / c->s_scale, 2);
  return fval;
}

static W_bool_t 
dummy_get_kinem_var (W_mat_3d_ip_ctx_t *ctx, 
                     W_mat_3d_kinem_var_type_enum  type,
                     double kinvar[3][3])
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;
  return FALSE;
}

static W_bool_t 
get_mat_state_var (W_mat_3d_t *m,
                   W_mat_3d_state_t ms,
                   W_mat_3d_ip_ctx_t *ip_ctx,
                   W_mat_3d_mat_state_var_enum  var_type,
                   double *var
                   )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  W_mat_3d_ip_ctx_t red_ip_ctx;
  ip_ctx_data_t ipdata;

  red_ip_ctx               = *ip_ctx;
  red_ip_ctx.ctx_data      = &ipdata;
  red_ip_ctx.get_kinem_var = dummy_get_kinem_var;
  ipdata.up_ip_ctx = ip_ctx;
  ipdata.ms        = matst;

  return WM3D_get_mat_state_var (u->ssmat, matst->ssms, &red_ip_ctx, var_type, var); 
}

/* Function returns the stress saved in the material state (with proper 
   conversion into the requested type).  NO UPDATE is performed on the 
   stress. */
static W_bool_t 
get_stress (W_mat_3d_t *m,
            W_mat_3d_state_t ms,
            W_mat_3d_ip_ctx_t *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double sigma[3][3], F[3][3], R[3][3], U[3][3], t[3][3];
  W_mat_3d_stress_type_enum out_stress_type_internal;
  W_mat_3d_ip_ctx_t red_ip_ctx;
  ip_ctx_data_t ipdata;

  red_ip_ctx               = *ip_ctx;
  red_ip_ctx.ctx_data      = &ipdata;
  red_ip_ctx.get_kinem_var = dummy_get_kinem_var;
  ipdata.up_ip_ctx = ip_ctx;
  ipdata.ms        = matst;
  
  if (!WM3D_get_stress (u->ssmat, matst->ssms, &red_ip_ctx, MAT_3D_CAUCHY,
                        t, &out_stress_type_internal))
    return FALSE;
  if (out_stress_type_internal != MAT_3D_CAUCHY)
    return FALSE;
    
  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
  if (!W_polar_decomposition (F, R, U)) {
    return FALSE;
  }
  
  MAT3DUTIL_MULT_ABAT_3X3 (sigma, R, t);
  
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
    MAT3DUTIL_COPY_3X3 (stress, sigma);
    break;
  case MAT_3D_1ST_PK:
    {
      double P[3][3], J, inv_F[3][3];
      J = MAT3DUTIL_DET_3X3(F);
      if (!W_mat3dutil_invert_3x3(F, inv_F))
        REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
      MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
      MAT3DUTIL_SCALE_3X3(P, J);
      MAT3DUTIL_COPY_3X3(stress, P);
    }
    break;
  case MAT_3D_2ND_PK:
    {
      double P[3][3], S[3][3], J, inv_F[3][3];
      J = MAT3DUTIL_DET_3X3(F);
      if (!W_mat3dutil_invert_3x3(F, inv_F))
         REPORT_FATAL_ERROR(NULL, "Failed inversion of F");
      MAT3DUTIL_MULT_ABT_3X3(P, sigma, inv_F);
      MAT3DUTIL_MULT_AB_3X3(S, inv_F, P);
      MAT3DUTIL_SCALE_3X3(S, J);
      MAT3DUTIL_COPY_3X3(stress, S);
    }
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, sigma);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}

static W_bool_t  
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double t[3][3];
  double F[3][3], U[3][3], R[3][3];
  double midstep_F[3][3], midstep_U[3][3], midstep_R[3][3], midstep_D[3][3];
  double dt = ip_ctx->time_step;
  W_bool_t result = TRUE;
  W_mat_3d_ip_ctx_t red_ip_ctx;
  ip_ctx_data_t ipdata;
  W_mat_3d_stress_type_enum out_stress_type_internal;

  /* kinematic quantities */
  if (!ip_ctx->get_kinem_var) return FALSE;  

  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_MIDSTEP_DEF_GRAD, midstep_F))
    return FALSE;
  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_MIDSTEP_RATE_OF_DEF, midstep_D))
    return FALSE;

  if (!W_polar_decomposition (F, R, U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }

  if (!W_polar_decomposition (midstep_F, midstep_R, midstep_U)) {
    REPORT_FATAL_ERROR (NULL, "Failed in W_polar_decomposition()");
  }

  red_ip_ctx               = *ip_ctx;
  red_ip_ctx.ctx_data      = &ipdata;
  red_ip_ctx.get_kinem_var = red_get_kinem_var;
  ipdata.up_ip_ctx = ip_ctx;
  ipdata.ms        = matst;
 
  MAT3DUTIL_MULT_ATBA_3X3 (ipdata.d, midstep_R, midstep_D);
  MAT3DUTIL_SCALE_3X3 (ipdata.d, dt);

  if (!WM3D_update_stress (u->ssmat, matst->ssms, update_form, &red_ip_ctx,
                           MAT_3D_CAUCHY, t, &out_stress_type_internal))
    return FALSE;
  if (out_stress_type_internal != MAT_3D_CAUCHY) return FALSE;

  if (stress_type == MAT_3D_CAUCHY) { /* accelerated handling for what we've got ready */
  /* rotate to obtain the Cauchy stress */
    MAT3DUTIL_MULT_ABAT_3X3 (stress, R, t);
    *out_stress_type = MAT_3D_CAUCHY;
    result = TRUE;
  } else { /* generic case */
    result = get_stress (m, ms, ip_ctx, stress_type, stress, out_stress_type);
  }
 exit:
  return result;
}

#include "consttransf.c"

static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_t ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
               )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  W_mat_3d_ip_ctx_t red_ip_ctx;
  ip_ctx_data_t ipdata;

  red_ip_ctx               = *ip_ctx;
  red_ip_ctx.ctx_data      = &ipdata;
  red_ip_ctx.get_kinem_var = dummy_get_kinem_var;
  ipdata.up_ip_ctx = ip_ctx;
  ipdata.ms        = matst;

  if (!WM3D_get_mat_stiff (u->ssmat, matst->ssms, &red_ip_ctx,
                           stiff_type, C_mat))
    return FALSE;
  if (stiff_type == MAT_3D_STIFF_EULERIAN) return TRUE;
  else                                     return FALSE;
}

static W_mat_3d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const,
  update_stress,
  get_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  NULL,
  NULL
};

static W_mat_3d_t *
new_mat_3d_red2d (double E, double nu,
		 double tensile_strength, double dyn_fracture_toughness,
		 double sigma_y, double hardening_param,
                 double delta, double rho, W_bool_t rot_C)
{
  mat_t *u;
  
  u = w_new(mat_t);
  CHECK_MALLOC(u);
  u->ssmat = new_mat_3d_ssj2ep (E, nu,
                                tensile_strength, dyn_fracture_toughness,
                                sigma_y, hardening_param,
                                delta, rho);
  
  return WM3D_new_material (&methods, (void *)u, rho);
}


W_mat_3d_t *
read_mat_3d_red2d (W_domain_t *d)
{
  const char *input_line;
  double E, nu, rho = 1;
  double sigma_y = 0, hardening_param = 0, delta = 1;
  double tensile_strength = 0;
  double dyn_fracture_toughness = 0;
  W_bool_t rot_C = FALSE;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { E = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { nu = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("TENSILE_STRENGTH", FALSE,
			    { tensile_strength = TOKEN_DBL_NEXT(1); },
			    { tensile_strength = E / 100; }
			    );
      TOKENS_SET_BY_KEYWORD("DYN_FRACTURE_TOUGHNESS", FALSE,
			    { dyn_fracture_toughness = TOKEN_DBL_NEXT(1); },
			    { dyn_fracture_toughness = E / 4000; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("SIGMA_Y", FALSE,
			    {
			      sigma_y = TOKEN_DBL_NEXT(1);
			    },
			    {
                              sigma_y = 0;
                            }
			    );
      TOKENS_SET_BY_KEYWORD("DELTA", FALSE,
			    { delta = TOKEN_DBL_NEXT(1); },
			    { delta = 1; } /* isotropic hardening  */
			    );
      TOKENS_SET_BY_KEYWORD("HARDENING_PARAM", FALSE,
			    { hardening_param = TOKEN_DBL_NEXT(1); },
			    { hardening_param = 0; }/* perfectly ep material */
			    );
      if (TOKENS_HAVE_KEYWORD ("ROT_C")) rot_C = TRUE;
      return new_mat_3d_red2d (E, nu, tensile_strength, dyn_fracture_toughness,
                               sigma_y, hardening_param, delta, rho, rot_C);
    }
  }
  return NULL;
}


