#include "Wtypes.h"
#include "Wutil.h"
#include "Wxf.h"
#include "Wmat3d.h"
#include "Wmat3dviscel.h"

/* RCSHEADER "$Id: mat3dviscel.c,v 1.8 2005/02/16 04:34:54 cvs Exp $" */

/*
  Small strain visco-elastic material. Small-strain elasticity + Newton-like viscosity.
 */

typedef struct mat_t {
  double E, nu;
  double bulk_mod, G;
  double dil_wave_speed_3d;
  double dil_wave_speed_pstrain;
  double dil_wave_speed_pstress;
  double shear_wave_speed;
  double rayleigh_pstrain_wave_speed;
  double rayleigh_pstress_wave_speed;
  double viscosity;
}              mat_t; 

/*
  Create a new material.
 */
W_mat_3d_t *
new_mat_3d_viscel (double E, double nu,
                   double viscosity, double rho);

static int
get_mcode (W_mat_3d_t *m)
{
  return MAT_3D_VISCEL;
}

static W_bool_t
get_mat_const (W_mat_3d_t *m, W_mat_3d_mat_const_enum type, double *c)
{
  mat_t *u = (mat_t *)m->data;

  switch (type) {
  case MAT_3D_YOUNG_MODULUS:
    *c = u->E;
    break;
  case MAT_3D_POISSON:
    *c = u->nu;
    break;
  case MAT_3D_SHEAR_MODULUS:
    *c = u->G;
    break;
  case MAT_3D_BULK_MODULUS:
    *c = u->bulk_mod;
    break;
  case MAT_3D_DIL_WAVE_SPEED_3D:
    *c = u->dil_wave_speed_3d;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRAIN:
    *c = u->dil_wave_speed_pstrain;
    break;
  case MAT_3D_DIL_WAVE_SPEED_PSTRESS:
    *c = u->dil_wave_speed_pstress;
    break;
  case MAT_3D_SHEAR_WAVE_SPEED:
    *c = u->shear_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRAIN_WAVE_SPEED:
    *c = u->rayleigh_pstrain_wave_speed;
    break;
  case MAT_3D_RAYLEIGH_PSTRESS_WAVE_SPEED:
    *c = u->rayleigh_pstress_wave_speed;
    break;
    /*
  case MAT_3D_NEWTONIAN_VISCOSITY:
    *c = u->sigma_y;
    break;
    */
  default:
    *c = 0;
    return FALSE;
  }
  return TRUE;
}

/* ============================================================== */

typedef struct matstate_t {
  double   cauchy[6]; 
  double   deform_energy_density;
  double   dissip_energy_density;
}              matstate_t;

static W_mat_3d_state_p 
get_mat_state (W_mat_3d_t *m, W_loc_t *at_point)
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *ms;
  int i;

  ms = w_new(matstate_t); CHECK_MALLOC(ms);

  FOR(i, 6) { ms->cauchy[i] = 0; }
  ms->deform_energy_density = 0;
  ms->dissip_energy_density = 0;

  return (W_mat_3d_state_p)ms;
}

static W_mat_3d_state_p 
interpolate_mat_state (W_mat_3d_t *m,
                       W_loc_t *at_point,
                       int npkgs,
                       WM3D_mat_state_pkg_t *pkgs)
{
  matstate_t *ms, *pms;
  int i, k;
  double weight_sum = 0;
#define WGHT(K) (pkgs[k].weight / weight_sum)

  FOR(k, npkgs) {weight_sum += pkgs[k].weight;}

  ms = (matstate_t *) get_mat_state(m, at_point);

  FOR(k, npkgs) {
    pms = (matstate_t *)pkgs[k].ms;
    FOR(i, 6) {
      ms->cauchy[i]      += pms->cauchy[i] * WGHT(k);
    }
    ms->deform_energy_density += pms->deform_energy_density * WGHT(k);
    ms->dissip_energy_density += pms->dissip_energy_density * WGHT(k);
  }

  return (W_mat_3d_state_p)ms;
}

static W_bool_t 
get_mat_state_var (W_mat_3d_t *m,
                   W_mat_3d_state_p ms,
                   W_mat_3d_ip_ctx_t *ip_ctx,
                   W_mat_3d_mat_state_var_enum  var_type,
                   double *var
                   )
{
  matstate_t *matst = (matstate_t *)ms;

  switch(var_type) {
  case MAT_3D_DEFORM_ENERGY_DENSITY:
    *var = matst->deform_energy_density;
    return TRUE;  
  case MAT_3D_EQUIV_PLASTIC_DEF:
    *var = 0;
    return TRUE;  
  default:
    return FALSE;
  }
}
/* Function returns the stress saved in the material state (with proper */
/* conversion into the requested type).  NO UPDATE is performed on the */
/* stress. */
static W_bool_t 
get_stress (W_mat_3d_t *m,
            W_mat_3d_state_p ms,
            W_mat_3d_ip_ctx_t *ip_ctx,
            W_mat_3d_stress_type_enum    stress_type,
            double stress[3][3],
            W_mat_3d_stress_type_enum   *out_stress_type
            )
{
  matstate_t *matst = (matstate_t *)ms;
  double sigma[3][3], cauchy[3][3];

  MAT3DUTIL_V_TO_T (matst->cauchy, cauchy);
    
  *out_stress_type = stress_type;
  switch (stress_type) {
  case MAT_3D_CAUCHY:
  case MAT_3D_1ST_PK:
  case MAT_3D_2ND_PK:
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    break;
  default:
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    *out_stress_type = MAT_3D_CAUCHY;
    break;
  }
    
  return TRUE;
}

static W_bool_t  
update_stress (W_mat_3d_t                  *m,
               W_mat_3d_state_p             ms,
               W_mat_state_update_form_enum update_form,
               W_mat_3d_ip_ctx_t           *ip_ctx,
               W_mat_3d_stress_type_enum    stress_type,
               double                       stress[3][3],
               W_mat_3d_stress_type_enum   *out_stress_type
               )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;
  double dt = ip_ctx->time_step;
  double lambda = u->E * u->nu / (1 + u->nu) / (1 - 2*(u->nu));
  double mu = u->G, lfact, mfact, trd;
  double viscosity = u->viscosity;
  double prevcauchy[3][3], cauchy[3][3], cauchyn[3][3];
  double F[3][3], DU[3][3], DUT[3][3], d[3][3], devd[3][3];
  double midD[3][3]; 
#define SQRT_2_OVER_3 0.81649658092772603272
  W_bool_t result = TRUE;
  matstate_t input_matst;
 
  /* kinematic quantities */
  if (!ip_ctx->get_kinem_var) return FALSE;  

  if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_DEF_GRAD, F)) return FALSE;

  /* save current material state */
  if (update_form == W_MAT_STATE_UPDATE_TEMP) input_matst = *matst;

  MAT3DUTIL_V_TO_T (matst->cauchy, cauchy); /* previous stress */
  MAT3DUTIL_COPY_3X3 (prevcauchy, cauchy);

  /* compute the strain */
  MAT3DUTIL_COPY_3X3(DU, F);
  MAT3DUTIL_MAT_MOD_DIAG (DU, -=, 1);
  MAT3DUTIL_COPY_TRANSP_3X3 (DUT, DU);
  MAT3DUTIL_3_MAT (d, =, 0.5 * DU, +, 0.5 * DUT);

  /* elastic stress */
  lfact = lambda;
  mfact = 2 * mu;
  trd = MAT3DUTIL_TRACE_3X3 (d);
  MAT3DUTIL_2_MAT (cauchyn, =, mfact * d);
  MAT3DUTIL_MAT_MOD_DIAG (cauchyn, +=, lfact * trd);
  MAT3DUTIL_2_MAT (cauchy, =, cauchyn);
 
  /* viscous stress */
  if (viscosity != 0) {
    double dil=0;
    double visccauchy[3][3];
    /*
      MAT3DUTIL_COPY_3X3 (devd, d);
      MAT3DUTIL_MAT_MOD_DIAG (devd, -=, trd/3); 
      mfact = 2 * viscosity / dt;
      MAT3DUTIL_2_MAT (cauchy, +=, mfact*devd); 
    */
    
    if (!(*ip_ctx->get_kinem_var) (ip_ctx, MAT_3D_MIDSTEP_RATE_OF_DEF, midD)) return FALSE;
    
    dil = MAT3DUTIL_TRACE_3X3 (midD) / 3;
    MAT3DUTIL_MAT_MOD_DIAG (midD, -=, dil); 
    mfact = 2 * viscosity;
    MAT3DUTIL_2_MAT (visccauchy, =, mfact * midD);
    MAT3DUTIL_2_MAT (cauchy, +=, visccauchy);
    matst->dissip_energy_density += (MAT3DUTIL_DBL_PROD_3X3(visccauchy, midD)) * dt;
  }
  
  /* save state: new stress */
  MAT3DUTIL_T_TO_V (cauchy, matst->cauchy);

  matst->deform_energy_density = (MAT3DUTIL_DBL_PROD_3X3 (cauchy, d)) / 2;

  if (stress_type == MAT_3D_CAUCHY) { /* accelerated handling for what we've got ready */
  /* rotate to obtain the Cauchy stress */
    MAT3DUTIL_COPY_3X3(stress, cauchy);
    *out_stress_type = MAT_3D_CAUCHY;
    result = TRUE;
  } else { /* generic case */
    result = get_stress (m, ms, ip_ctx, stress_type, stress, out_stress_type);
  }
 exit:
  if (update_form == W_MAT_STATE_UPDATE_TEMP) *matst = input_matst;
  return result;
}

#include "consttransf.c"

static W_bool_t
get_mat_stiff (W_mat_3d_t *m/* mat3d */,
               W_mat_3d_state_p ms,/* material state */
               W_mat_3d_ip_ctx_t *ip_ctx,/* integration point context;
                                            used to generate kinematic
                                            quantities which the material state
                                            might need in order to compute
                                            the requested stress */
               W_mat_3d_stiff_enum stiff_type, /* stiffness type */
               double C_mat[6][6] /* material stiffness tensor in matrix form */
               )
{
  mat_t *u = (mat_t *) m->data;
  matstate_t *matst = (matstate_t *)ms;
  double G = u->G, K = u->bulk_mod;

  /* RAW only the elastic part is included */
    C_mat[0][0] = K +2*G*(0.666667);
    C_mat[0][1] = C_mat[1][0] = K +2*G*(-0.333333);
    C_mat[0][2] = C_mat[2][0] = K +2*G*(-0.333333);
    C_mat[0][3] = C_mat[3][0] = 0;
    C_mat[0][4] = C_mat[4][0] = 0;
    C_mat[0][5] = C_mat[5][0] = 0;
    C_mat[1][1] = K +2*G*(0.666667);
    C_mat[1][2] = C_mat[2][1] = K +2*G*(-0.333333);
    C_mat[1][3] = C_mat[3][1] = 0;
    C_mat[1][4] = C_mat[4][1] = 0;
    C_mat[1][5] = C_mat[5][1] = 0;
    C_mat[2][2] = K +2*G*(0.666667);
    C_mat[2][3] = C_mat[3][2] = 0;
    C_mat[2][4] = C_mat[4][2] = 0;
    C_mat[2][5] = C_mat[5][2] = 0;
    C_mat[3][3] = G;
    C_mat[3][4] = C_mat[4][3] = 0;
    C_mat[3][5] = C_mat[5][3] = 0;
    C_mat[4][4] = G;
    C_mat[4][5] = C_mat[5][4] = 0;
    C_mat[5][5] = G;
  
  return TRUE;
}

static W_bool_t 
get_mat_state_named_var (W_mat_3d_t *m,
                         W_mat_3d_state_p ms,
                         W_mat_3d_ip_ctx_t *ip_ctx,
                         char *name,
                         double *var
                         )
{
  mat_t *u = (mat_t *)m->data;
  matstate_t *matst = (matstate_t *)ms;

  if        (tokens_token_and_keyword_equiv (name, "ln_U_1") ||
             tokens_token_and_keyword_equiv (name, "ln_U_2") ||
             tokens_token_and_keyword_equiv (name, "ln_U_3") ||
             tokens_token_and_keyword_equiv (name, "max_ln_U") 
             ) {
    double F[3][3], R[3][3], U[3][3], ln_U[3][3];
    double evals[3], evecs[3][3], lnevals[3];
    *var = 0;
    if (!(*ip_ctx->get_kinem_var)(ip_ctx, MAT_3D_DEF_GRAD, F)) {
      return FALSE;
    }
    if (W_polar_decomposition (F, R, U)) {
      W_eigendecomposition (U, evecs, evals);
      if (tokens_token_and_keyword_equiv (name, "ln_U_1")) *var = log(evals[0]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_2")) *var = log(evals[1]);
      if (tokens_token_and_keyword_equiv (name, "ln_U_3")) *var = log(evals[2]);
      if (tokens_token_and_keyword_equiv (name, "max_ln_U")) {
        *var = fabs(log(evals[0]));
        if (fabs(log(evals[1])) > *var) *var = fabs(log(evals[1]));
        if (fabs(log(evals[2])) > *var) *var = fabs(log(evals[2]));
      }
      goto exit;
    }

    *var = 0;
    return FALSE;
  } else if (tokens_token_and_keyword_equiv (name, "dissip_energy_density")) {
    *var = matst->dissip_energy_density;
  }
 exit:
  return TRUE;
}

static W_mat_3d_methods_t methods = {
  get_mcode,
  get_mat_state,
  get_mat_const,
  update_stress,
  get_mat_state_var,
  interpolate_mat_state,
  get_mat_stiff,
  get_mat_state_named_var,
  NULL
};

/*
  Create a new material.
 */
W_mat_3d_t *
new_mat_3d_viscel (double E, double nu,
                   double viscosity, double rho)
{
  mat_t *u;
  
  u = w_new(mat_t);
  CHECK_MALLOC(u);
  u->E                      = E; 
  u->nu                     = nu;
  u->viscosity              = viscosity;
  if (u->nu >= 0.5) REPORT_FATAL_ERROR(NULL, "Bad nu (should be 0 <= nu < 0.5)");
  u->bulk_mod               = u->E / 3 / (1 - 2*u->nu);
  u->G                      = u->E / 2 / (1 + u->nu);
  u->dil_wave_speed_3d
    = W_3D_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstrain
    = W_plstrain_dilatational_wave_speed (u->E, u->nu, rho);
  u->dil_wave_speed_pstress
    = W_plstress_dilatational_wave_speed (u->E, u->nu, rho);
  u->shear_wave_speed
    = W_shear_wave_speed (u->E, u->nu, rho);
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstress,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstress_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane stress speed does not exist");
  if (!W_compute_Rayleigh_speed (u->dil_wave_speed_pstrain,
                                 u->shear_wave_speed,
                                 0.01,
                                 &u->rayleigh_pstrain_wave_speed))
    REPORT_FATAL_ERROR (NULL, "Rayleigh plane strain speed does not exist");
  
  return WM3D_new_material (&methods, (void *)u, rho);
}


W_mat_3d_t *
read_mat_3d_viscel (W_domain_t *d)
{
  const char *input_line;
  double E, nu, rho = 1;
  double viscosity = 0;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return NULL;
    else {
      TOKENS_PARSE(input_line);
      TOKENS_SET_BY_KEYWORD("E", TRUE,
			    { E = TOKEN_DBL_NEXT(1); },
			    { E = 1; }
			    );
      TOKENS_SET_BY_KEYWORD("NU", TRUE,
			    { nu = TOKEN_DBL_NEXT(1); },
			    { nu = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("VISCOSITY", FALSE,
			    { viscosity = TOKEN_DBL_NEXT(1); },
			    { viscosity = 0; }
			    );
      TOKENS_SET_BY_KEYWORD("RHO", TRUE,
			    { rho = TOKEN_DBL_NEXT(1); },
			    { rho = 1; }
			    );
      return new_mat_3d_viscel (E, nu, viscosity, rho);
    }
  }
  return NULL;
}


