#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wdispbc.h"
#include "Wivcond.h"
#include "Wmats.h"
#include "Wcontact.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "Wmpi.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wmodrr.h"
#include "Wmodrc.h"
#include "meschach.h"

/* RCSHEADER "$Id: modrc.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

#ifdef REPLACE
# undef REPLACE
#endif
#ifdef ADD
# undef ADD
#endif
#define REPLACE FALSE
#define ADD     TRUE

typedef void (*compute_modes_func) (W_modr_collector_p modr);

typedef struct W_modr_collector_t {
  W_domain_t         *domain;        /* domain */
  /* configuration parameters */
  double              dt;     /* time interval between data collections */
  double              from_t; /* collect data from this time ... */
  double              up_to_t;/* ... up to this time */
  W_modr_method_t     method; /* collection method */
  double              cutoff_ev;
  W_bool_t            dont_subtract_mean;
  /* internal variables */
  double              last_t; /* last time the data was collected */
  W_index_t           toteqn; /* dimension of the below array */
  LIST_OF (double *)  ul;            /* list of arrays; one array per snapshot */
  double             *umean;
  MAT                *evecs;         /* eigen-modes of R */
  VEC                *evals;         /* eigenvalues of R */
  compute_modes_func  compute_modes;
}              W_modr_collector_t;

/*
  New model-reduction data collector.
  Create a new collector.
 */
W_modr_collector_p /* collector */
W_modr_collector_new (W_domain_t *d /* domain (client) */
                     )
{
  W_modr_collector_p modr = w_new (W_modr_collector_t);
  CHECK_MALLOC (modr);
  modr->domain = d;
  modr->dt = 0;
  modr->from_t = 0;
  modr->up_to_t = FLT_MAX;
  modr->method = MODR_SNAPSHOTS;
  modr->cutoff_ev = 0;
  modr->dont_subtract_mean = FALSE;
  modr->last_t = 0;
  modr->toteqn = 0;
  modr->ul = NULL;
  modr->umean = NULL;
  modr->evecs = NULL;
  modr->evals = NULL;
  modr->compute_modes = NULL;
  return modr;
}

static double
dot (double *a, double *b, int M)
{
  int k;
  double accum = 0;
  FOR (k, M) { accum += a[k] * b[k]; }
  return accum;
}

static void
accumulate (W_modr_collector_p modr, W_modr_get_node_data_proc get_node_data, void *client)
{
  W_node_t *n;
  W_vect_t d;
  int i;
  W_index_t toteqn = WD_tot_of_equations (modr->domain);
  double *v;

  if (modr->toteqn == 0) modr->toteqn = toteqn;
  if (modr->toteqn != toteqn) REPORT_FATAL_ERROR (modr->domain->errh, "Change in the number of nodes");

  if (modr->ul == NULL) {
    modr->ul = make_list (); CHECK_MALLOC (modr->ul);
  }
  
  /* accumulate array at this snapshot */
  v = w_new_array (modr->toteqn, double); CHECK_MALLOC (v);
  add_to_tail (modr->ul, v);
  i = 0;
  n = NULL;
  while (n = WD_get_next_node (modr->domain, n)) {
    if (i+3 > modr->toteqn) REPORT_FATAL_ERROR (modr->domain->errh, "Wrong numbering");
    (*get_node_data) (client, n, &d.x, &d.y, &d.z);
    v[i] = d.x; i++;
    v[i] = d.y; i++;
    v[i] = d.z; i++;
  }
}

static void
get_displ_wrapper (void *client, W_node_t *n, double *x, double *y, double *z)
{
  WN_get_displ (n, x, y, z);
}

/* 
   Accumulate data for model reduction.
   Give the collector chance to accumulate data for the construction
   of the POD (proper orthogonal decomposition) modes.  The collector actually gathers the data if
      (i) the time (as supplied by the client in WD_time())
          is in the <'from time'; 'up-to time'> interval;
      (ii) and the last data had been gathered later than `time+dt' before.
   W_modr_collector_accumulate() uses WN_get_displ() to access the data
   to collect.
 */
void
W_modr_collector_accumulate (W_modr_collector_p modr/* model reduction data collector */
                            )
{
  double t = WD_time (modr->domain);
  if (t >= modr->from_t && t <= modr->up_to_t &&
      t >= modr->last_t + modr->dt 
      ) {
    accumulate (modr, get_displ_wrapper, NULL);
    modr->last_t = t;
  }
}

/*
  Accumulate data for model reduction.
   Give the collector chance to accumulate data for the construction
   of the POD (proper orthogonal decomposition) modes.  Works similarly as the
   W_modr_collector_accumulate() function, but gives the additional
   flexibility to use a specified function that supplies
   the vector to collect at each point.
 */
void
W_modr_collector_accumulate_with (W_modr_collector_p modr,/* model reduction data collector */
                                 void *client, /* collect nodal data from client (may be different from the domain client) */
                                 W_modr_get_node_data_proc get_node_data /* accessor for nodal data to collect */
                                 )
{
  double t = WD_time (modr->domain);
  if (t >= modr->from_t && t <= modr->up_to_t &&
      t >= modr->last_t + modr->dt 
      ) {
    accumulate (modr, get_node_data, client);
    modr->last_t = t;
  }
}


typedef struct modr_mode_data_t {
  int     mode;
  double  eval;
} modr_mode_data_t;

static int
compare_modes_from_largest (modr_mode_data_t *pm0, modr_mode_data_t *pm1)
{
  return (pm0->eval >= pm1->eval);
}

static int
compare_modes_from_smallest (modr_mode_data_t *pm0, modr_mode_data_t *pm1)
{
  return (pm0->eval <= pm1->eval);
}

static void
subtract_mean (W_modr_collector_p modr)
{
  int i;
  double one_over_M;
  double *u;
  int toteqn = modr->toteqn;
  int M = list_length (modr->ul);

  one_over_M = 1. / M;

  modr->umean = w_new_array (toteqn, double); CHECK_MALLOC (modr->umean);
  FOR (i, toteqn) { modr->umean[i] = 0; }

  loop_through_list (modr->ul, u, double *) {
    FOR (i, toteqn) { modr->umean[i] += one_over_M * u[i]; }
  }

  loop_through_list (modr->ul, u, double *) {
    FOR (i, toteqn) { u[i] -= modr->umean[i]; }
  }
}

static void
make_mode_MoS (W_modr_collector_p modr, int mode, double *u)
{
  int i, j, M = modr->evals->dim;
  double aj, *uj;
  double sum = 0;

  FOR (i, modr->toteqn) { u[i] = 0; }
  
  FOR (j, M) {
    uj = (double *) get_list_pos (modr->ul, j+1);
    if (!uj) REPORT_FATAL_ERROR (NULL, "No data");
    aj = modr->evecs->me[j][mode];
    FOR (i, modr->toteqn) {
      u[i] += aj * uj[i];
    }
  }
  
  sum = 0;
  FOR (i, modr->toteqn) {
    sum += u[i] * u[i];
  }
  sum = sqrt (sum);
  if (sum <= 0 || isnan (sum)) sum = 1;
  FOR (i, modr->toteqn) {
    u[i] /= sum;
  }
}

#if 0
#define CHECK_ORTHONORMAL(i,j,d)                                        \
  if ((i == j) && (fabs (d - 1) > 1e-6)) {                              \
    max_d = max (max_d, fabs (d - 1));                                  \
    all_orthonormal = FALSE;                                            \
    sprintf (W_buf (), "Mode %d not normalized: %g", i, d);             \
    REPORT_WARNING (modr->domain->errh, W_buf ());                       \
  }                                                                     \
  if ((i != j) && (fabs (d) > 1e-6)) {                                  \
    max_d = max (max_d, fabs (d));                                      \
    all_orthonormal = FALSE;                                            \
    sprintf (W_buf (), "Modes %d and %d not orthonormal: %g", i, j, d); \
    REPORT_WARNING (modr->domain->errh, W_buf ());                       \
  }
#else
#define CHECK_ORTHONORMAL(i,j,d)                                        \
  if ((i == j) && (fabs (d - 1) > 1e-6)) {                              \
    max_d = max (max_d, fabs (d - 1));                                  \
    all_orthonormal = FALSE;                                            \
  }                                                                     \
  if ((i != j) && (fabs (d) > 1e-6)) {                                  \
    max_d = max (max_d, fabs (d));                                      \
    all_orthonormal = FALSE;                                            \
  }
#endif

static void
check_orthonorm_c (W_modr_collector_p modr)
{
  int i, j, k, M = modr->evals->dim;
  MAT *evecs = modr->evecs;
  double d;
  W_bool_t all_orthonormal = TRUE;
  double max_d = 0;
  
  FOR (i, M) {
    FOR (j, M) {
      if (modr->evals->ve[i] > modr->cutoff_ev && modr->evals->ve[j] > modr->cutoff_ev) {
        d = 0;
        FOR (k, modr->toteqn) {
          d += evecs->me[k][i] * evecs->me[k][j];
        }
        CHECK_ORTHONORMAL (i, j, d);
      }
    }
  }
  if (!all_orthonormal) {
    sprintf (W_buf (), "Some modes not orthonormal; max mismatch %g", max_d);
    REPORT_WARNING (modr->domain->errh, W_buf ());
  }
}

static void
order_modes_MoS (W_modr_collector_p modr)
{
  LIST ml = make_list ();
  modr_mode_data_t *pm;
  int i, j;
  MAT *new_evecs;
  int M = modr->evals->dim;
  double *u;

  FOR (i, M) {
    pm = w_new (modr_mode_data_t); CHECK_MALLOC (pm);
    pm->mode = i;
    pm->eval = modr->evals->ve[i];
    add_to_tail (ml, pm);
  }

  sort_list_func (ml, compare_modes_from_largest);

  W_log_update (modr->domain->logh, "Method of Snaphots POD");
  new_evecs = m_get (modr->toteqn, M);
  m_zero (new_evecs);
  u = w_new_array (modr->toteqn, double); CHECK_MALLOC (u);
  i = 0;
  loop_through_list (ml, pm, modr_mode_data_t *) {
    sprintf (W_buf (), "   Mode %d, eigenvalue %g", i, pm->eval);
    W_log_update (modr->domain->logh, W_buf ());
    make_mode_MoS (modr, pm->mode, u);
    FOR (j, modr->toteqn) { /* set eigenvector */
      new_evecs->me[j][i] = u[j];
    }
    modr->evals->ve[i] = pm->eval; /* eigenvalue */
    i++;
#   if 0 /* RAW check of orthonormality */
    {
      int k, m, n;
      double d;
      d = 0;

      FOR (m, i) {
        FOR (n, i) {
          d = 0;
          FOR (k, modr->toteqn) {
            d += new_evecs->me[k][m] * new_evecs->me[k][n];
          }
          if (m == n && fabs (d - 1) > 1e-6 ||
              m != n && fabs (d) > 1e-6)
            printf ("m=%d, n=%d, d=%g\n", m, n, d);
        }
      }
    }
#   endif
  }
  w_free (u);

  m_free (modr->evecs);
  modr->evecs = new_evecs;

  free_list (ml, AND_NODES);
}


static void
compute_modes_MoS (W_modr_collector_p modr)
{
  int i, j, n;
  double one_over_M;
  double *ui, *uj;
  int M = list_length (modr->ul);
  MAT *C;

  fprintf (stderr, "compute_modes_MoS\n");
  
  if (M == 0) { REPORT_FATAL_ERROR (modr->domain->errh, "No collected data"); }

  one_over_M = 1.0 / (double) M;
  
  C = m_get (M, M);
  m_zero (C);

  n = modr->toteqn;
  FOR (i, M) {
    ui = (double *) get_list_pos (modr->ul, i+1);
    if (ui == NULL) REPORT_FATAL_ERROR (modr->domain->errh, "No such data");
    FOR (j, M) {
      uj = (double *) get_list_pos (modr->ul, j+1);
      if (uj == NULL) REPORT_FATAL_ERROR (modr->domain->errh, "No such data");
      C->me[i][j] += one_over_M * dot (ui, uj, n);
    }
  }

  modr->evecs = m_resize (modr->evecs, M, M); CHECK_MALLOC (modr->evecs);
  modr->evals = v_resize (modr->evals, M); CHECK_MALLOC (modr->evals);
#if 0
  {
    MAT *V;
    V = m_get (M, M);
    m_ident (V);
    m_ident (modr->evecs);
    svd (C, V, modr->evecs, modr->evals);
    m_transp (modr->evecs, modr->evecs);
    m_free (V);
  }
#else
  symmeig (C, modr->evecs, modr->evals);
#endif
  m_free (C);

  order_modes_MoS (modr);
}

static void
order_modes_DM_orig (W_modr_collector_p modr)
{
  LIST ml = make_list ();
  modr_mode_data_t *pm;
  int i, j;
  int M = modr->evals->dim;
  MAT *new_evecs;

  FOR (i, M) {
    pm = w_new (modr_mode_data_t); CHECK_MALLOC (pm);
    pm->mode = i;
    pm->eval = modr->evals->ve[i];
    add_to_tail (ml, pm);
  }

  sort_list_func (ml, compare_modes_from_largest);

  W_log_update (modr->domain->logh, "Direct Method of POD");
  new_evecs = m_get (modr->evecs->m, modr->evecs->m);
  i = 0;
  loop_through_list (ml, pm, modr_mode_data_t *) {
    sprintf (W_buf (), "   Mode %d, eigenvalue %g", i, pm->eval);
    W_log_update (modr->domain->logh, W_buf ());
    FOR (j, M) {
      new_evecs->me[j][i] = modr->evecs->me[j][pm->mode];
    }
    modr->evals->ve[i] = pm->eval;
    i++;
  }

  m_free (modr->evecs);
  modr->evecs = new_evecs;

  free_list (ml, AND_NODES);
}


static void
compute_modes_DM_orig (W_modr_collector_p modr)
{
  int i, j, k;
  double one_over_M;
  double *uk;
  int toteqn = modr->toteqn;
  int M = list_length (modr->ul);
  MAT *C;

  if (M == 0) { REPORT_FATAL_ERROR (modr->domain->errh, "No collected data"); }

  fprintf (stderr, "compute_modes_DM_orig\n");
  
  one_over_M = 1.0 / (double) M;
  
  C = m_get (toteqn, toteqn);
  m_zero (C);
  
  FOR (k, M) {
    uk = (double *) get_list_pos (modr->ul, k+1);
    if (uk == NULL) REPORT_FATAL_ERROR (modr->domain->errh, "No such data");
    FOR (i, toteqn) {
      FOR (j, toteqn) {
        C->me[i][j] += one_over_M * uk[i] * uk[j];
      }
    }
  }

  modr->evecs = m_resize (modr->evecs, toteqn, toteqn); CHECK_MALLOC (modr->evecs);
  modr->evals = v_resize (modr->evals, toteqn); CHECK_MALLOC (modr->evals);
  symmeig (C, modr->evecs, modr->evals);
  m_free (C);

  FOR (i, toteqn) { /* normalize */
    double sum = 0;
    FOR (j, toteqn) {
      sum += modr->evecs->me[j][i]*modr->evecs->me[j][i];
    }
    sum = sqrt (sum);
    if (sum <= 0 || isnan (sum)) sum = 1;
    FOR (j, toteqn) {
      modr->evecs->me[j][i] /= sum;
    }      
  }

  order_modes_DM_orig (modr);
}


static double
uk_dot_squared (double *uk, int n)
{
  double v = 0;
  int i;
  FOR (i, n) { v += uk[i] * uk[i]; }
  return (v);
}

/*
  Compute modes.
  Call this function to compute the modes before writing them
  out to disk.
 */
void
W_modr_collector_compute_modes (W_modr_collector_p modr/* model reduction data collector */)
{
  if (!(modr->dont_subtract_mean)) subtract_mean (modr);
  modr->compute_modes (modr);
  check_orthonorm_c (modr);
}

static void
set_mean (W_modr_collector_p modr, W_bool_t add, double scale)
{
  W_node_t *n;
  W_vect_t d;
  W_bool_t apply[3] = {TRUE, TRUE, TRUE};
  int i;

  if (modr->umean == NULL){REPORT_WARNING (modr->domain->errh, "No mean");return;}

  i = 0;
  n = NULL;
  while (n = WD_get_next_node (modr->domain, n)) {
    VECMAC_SCALAR (d, =, 0);
    if (add) {
      WN_get_displ (n, &d.x, &d.y, &d.z);
      d.x  += scale * modr->umean[i+0];
      d.y  += scale * modr->umean[i+1];
      d.z  += scale * modr->umean[i+2];
    } else {
      d.x  = scale * modr->umean[i+0];
      d.y  = scale * modr->umean[i+1];
      d.z  = scale * modr->umean[i+2];
    }
    i += 3;
    WN_set_displ (n, &d, apply);
  }
}


static void
set_mode (W_modr_collector_p modr, int mode, W_bool_t add, double scale)
{
  W_node_t *n;
  W_vect_t d;
  W_bool_t apply[3] = {TRUE, TRUE, TRUE};
  int i, j;

  if (modr->evecs == NULL){
    REPORT_WARNING (modr->domain->errh, "No component modes matrix"); return;
  }
  if (mode >= modr->evals->dim)  {
    REPORT_WARNING (modr->domain->errh, "No such mode"); return;
  }
  if (mode < 0) mode = 0;

  i = 0;
  n = NULL;
  while (n = WD_get_next_node (modr->domain, n)) {
    if (i+3 > modr->toteqn) REPORT_FATAL_ERROR (modr->domain->errh, "Wrong numbering");
    if (add) {
      WN_get_displ (n, &d.x, &d.y, &d.z);
      d.x += scale * modr->evecs->me[i+0][mode];
      d.y += scale * modr->evecs->me[i+1][mode];
      d.z += scale * modr->evecs->me[i+2][mode];
    } else {
      d.x = scale * modr->evecs->me[i+0][mode];
      d.y = scale * modr->evecs->me[i+1][mode];
      d.z = scale * modr->evecs->me[i+2][mode];
    }
    WN_set_displ (n, &d, apply);
    i += 3;
  }
}


#ifdef X_XFACE
void
W_modr_collector_show_mode (W_modr_collector_p modr, int mode)
{
  double displ_scale = 1;
  if (modr == NULL) { REPORT_WARNING (modr->domain->errh, "Not in collect mode"); return; }
  displ_scale = WOPT_OPT (modr->domain->options, displ_scale);
  if (!(modr->dont_subtract_mean)) {
    set_mean (modr, REPLACE, (displ_scale > 0 ? 1 / displ_scale : 1)); /*  */
    set_mode (modr, mode, ADD, 1);
  } else {
    set_mode (modr, mode, REPLACE, 1);
  }
  WD_build_grep (modr->domain, TRUE);
}
#endif

/*
  Get POD mode eigenvalue.
  Get the eigenvalue associated with mode /mode/.
 */
double /* eigenvalue */
W_modr_collector_eigenvalue (W_modr_collector_p modr, /* model reduction data collector */
                            int mode /* mode */
                            )
{
  if (modr->evecs == NULL) {
    REPORT_WARNING (modr->domain->errh, "No component modes"); return 0;
  }
  if (mode >= modr->evals->dim)  {
    REPORT_WARNING (modr->domain->errh, "No such mode"); return 0;
  }
  if (mode < 0) mode = 0;
  return modr->evals->ve[mode];
}

/*
  Get the total number of modes computed.
 */
int
W_modr_collector_tot_modes (W_modr_collector_p modr/* model reduction data collector */)
{
  if (modr == NULL) {
    REPORT_WARNING (modr->domain->errh, "Not in collect mode"); return 0;
  }
  if (modr->evecs == NULL) {
    REPORT_WARNING (modr->domain->errh, "No component modes matrix"); return 0;
  }
  return modr->evals->dim;
}

static double
ev_sum (W_modr_collector_p modr)
{
  double sum = 0;
  int i;
  FOR (i, modr->evals->dim) { sum += modr->evals->ve[i]; }
  return sum;
}

/*
  Write the mean of the POD modes.
  Write the mean of the POD modes.
 */
int
W_modr_collector_write_mean (W_modr_collector_p modr, /* model reduction data collector */
                            FILE *fp /* write to stream */
                            )
{
  W_node_t *n;
  W_vect_t d;

  if (modr == NULL) { REPORT_WARNING (modr->domain->errh, "Not in collect mode"); return FALSE; }
  if (modr->umean == NULL)     { REPORT_WARNING (modr->domain->errh, "No mean data"); return FALSE; }

  set_mean (modr, REPLACE, 1);

  fprintf (fp, "!\n");
  fprintf (fp, "MEAN\n");
  fprintf (fp, "! node_id u_x u_y u_z\n");
  n = NULL;
  while (n = WD_get_next_node (modr->domain, n)) {
    WN_get_displ (n, &d.x, &d.y, &d.z);
    fprintf (fp, "%d %g %g %g\n", (int) n->id, d.x, d.y, d.z);
  }
  fprintf (fp, "END MEAN\n");
  fprintf (fp, "!\n");
  return TRUE;
}

/*
  Write POD mode.
  Write POD mode onto given stream.
 */
int /* ignore */
W_modr_collector_write_mode (W_modr_collector_p modr, /* model reduction data collector */
                            int mode, /* 0 <= mode < max_dof */
                            FILE *fp /* write to this file */
                            )
{
  W_node_t *n;
  W_vect_t d;
  double evsum = ev_sum (modr);

  if (modr == NULL) {
    REPORT_WARNING (modr->domain->errh, "Not in collect mode"); return FALSE;
  }
  if (modr->evecs == NULL) {
    REPORT_WARNING (modr->domain->errh, "No modes"); return FALSE;
  }
  if (mode >= modr->evals->dim)  {
    REPORT_WARNING (modr->domain->errh, "No such mode"); return FALSE;
  }
  if (mode < 0) mode = 0;

  set_mode (modr, mode, REPLACE, 1);

  fprintf (fp, "!\n");
  fprintf (fp, "MODE %d ", mode);
  fprintf (fp, "EIGENVALUE %g (%g%%)\n", W_modr_collector_eigenvalue (modr, mode),
           W_modr_collector_eigenvalue (modr, mode) / evsum * 100);
  fprintf (fp, "! node_id u_x u_y u_z\n");
  n = NULL;
  while (n = WD_get_next_node (modr->domain, n)) {
    WN_get_displ (n, &d.x, &d.y, &d.z);
    fprintf (fp, "%d %g %g %g\n", (int) n->id, d.x, d.y, d.z);
  }
  fprintf (fp, "END MODE %d\n", mode);
  fprintf (fp, "!\n");
  return TRUE;
}

/*
  Write collected data.
  Write all collected data (not the modes, just the raw data).
 */
int
W_modr_collector_save_data (W_modr_collector_p modr,  /* model reduction data collector */
                           FILE *fp /* write to stream */
                           ) 
{
  int toteqn = modr->toteqn;
  double *v;
  int sample, j;
  W_node_t *n;
  W_index_t ens[W_NDOF_PER_NODE];
  W_bool_t constrained[W_NDOF_PER_NODE];
  int neqn;

  fprintf (fp, "!     \n");
  fprintf (fp, "! POD sampled data, %d nodes (3 displacements each), %d samples\n",
           dict_count (modr->domain->nodes->d), list_length (modr->ul));
  fprintf (fp, "!     \n");
  fprintf (fp, "! Mapping of equations to node numbers:\n");
  fprintf (fp, "!   node #   equation #s\n");
  n = NULL;
  while (n = WD_get_next_node (modr->domain, n)) {
    neqn = WN_eqn_numbers_and_constraints (n, ens, constrained);
    fprintf (fp, "!    %d: %d %d %d\n", n->id, ens[0], ens[1], ens[2]);
  }
  fprintf (fp, "!   node #   equation #s\n");
  fprintf (fp, "!   \n");
  fprintf (fp, "!   displacements given on separate lines for each equation\n");
  fprintf (fp, "!     \n");
  sample = 0;
  loop_through_list (modr->ul, v, double *) {
    fprintf (fp, "! start sample %d\n", sample);
    FOR (j, toteqn) {
      fprintf (fp, "%g\n", v[j]);
    }
    fprintf (fp, "! end sample %d\n", sample);
    sample++;
  }
  return TRUE;
}


static void
write_header (W_modr_collector_p modr, FILE *fp)
{
  fprintf (fp, "! POD modes\n");
  fprintf (fp, "!   Computed %d modes\n", modr->evals->dim);
  fprintf (fp, "!   Collected %d samples\n", list_length (modr->ul));
  fprintf (fp, "!   From time %g\n", modr->from_t);
  fprintf (fp, "!   Up to time %g\n", modr->up_to_t);
  switch (modr->method) {
  case MODR_DIRECT:
    fprintf (fp, "!   Used the direct method\n");
    break;
  case MODR_SNAPSHOTS:
    fprintf (fp, "!   Used the snapshots method\n");
    break;
  default:
    fprintf (fp, "!   Used the unknown method\n");
    break;
  }
  if (modr->cutoff_ev < 0)
    fprintf (fp, "!   Writing out all computed modes\n");
  else
    fprintf (fp, "!   Writing out modes with eigenvalue above %g\n", modr->cutoff_ev);
  fprintf (fp, "! \n");
}

/*
  Write POD data.
  Write selected POD data.  Configure first the collector
  by giving a cutoff eigenvalue.
 */
int /* ignore */
W_modr_collector_write_modr (W_modr_collector_p modr,  /* model reduction data collector */
                           FILE *fp /* write to stream */
                           )
{
  int mode;
  write_header (modr, fp);
  FOR (mode, W_modr_collector_tot_modes (modr)) {
    if (W_modr_collector_eigenvalue (modr, mode) > modr->cutoff_ev) 
      W_modr_collector_write_mode (modr, mode, fp);
  }
  if (!(modr->dont_subtract_mean))
    W_modr_collector_write_mean (modr, fp);
  return TRUE;
}

/*
  Get collection time increment.
  Get time increment.  
 */
double
W_modr_collector_get_dt (W_modr_collector_p modr /* model reduction data collector */)
{
  return modr->dt;
}

/*
  Set collection time increment.
  Set collection time increment.
 */
double /* previous value */
W_modr_collector_set_dt (W_modr_collector_p modr,  /* model reduction data collector */
                        double dt /* time increment */
                        )
{
  double prev_dt = modr->dt;
  modr->dt = dt;
  return prev_dt;
}

/*
  Set `from' time.
  Set the time from  which the modes should be collected.
 */
double
W_modr_collector_set_from_t (W_modr_collector_p modr,  /* model reduction data collector */
                            double from_t
                            )
{
  double prev_from_t = modr->from_t;
  modr->from_t = from_t;
  return prev_from_t;
}

/*
  Set `up to' time.
  Set time up to which the modes should be collected.
 */
double
W_modr_collector_set_up_to_t (W_modr_collector_p modr,  /* model reduction data collector */
                             double up_to_t /* time */
                             )
{
  double prev_up_to_t = modr->up_to_t;
  modr->up_to_t = up_to_t;
  return prev_up_to_t;
}

/*
  Set cutoff eigenvalue.
  Set cutoff eigenvalue.  Modes with eigenvalues smaller
  than the cutoff value are dropped when being written to file.
 */
double /* previous value */
W_modr_collector_set_cutoff_ev (W_modr_collector_p modr, /* model reduction data collector */
                               double cutoff_ev /* cutoff eigenvalue */
                               )
{
  double prev_cutoff_ev = modr->cutoff_ev;
  modr->cutoff_ev = cutoff_ev;
  return prev_cutoff_ev;
}

/*
  Set `don't subtract mean' flag.
  Configure the collector not to subtract the mean
  when writing the modes to file.
 */
W_bool_t /* previous value */
W_modr_collector_set_dont_subtract_mean (W_modr_collector_p modr,  /* model reduction data collector */
                                        W_bool_t dont_subtract_mean /* boolean */
                                        )
{
  W_bool_t prev_dont_subtract_mean = modr->dont_subtract_mean;
  modr->dont_subtract_mean = dont_subtract_mean;
  return prev_dont_subtract_mean;
}

/*
  Set collection method.
  Set method which is used by the collector
  to compute the POD modes.
 */
W_modr_method_t /* previous method */
W_modr_collector_set_method (W_modr_collector_p modr,  /* model reduction data collector */
                            W_modr_method_t method /* method */
                            )
{
  W_modr_method_t prev_method = modr->method;
  modr->method = method;
  switch (modr->method) {
  case MODR_DIRECT:
    fprintf (stderr, "Method: direct\n");
    modr->compute_modes = compute_modes_DM_orig;
    break;
  case MODR_SNAPSHOTS:
  default:
    modr->compute_modes = compute_modes_MoS;
    break;
  }
  return prev_method;
}

