#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wdispbc.h"
#include "Wivcond.h"
#include "Wmats.h"
#include "Wcontact.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "Wmpi.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wmodrr.h"
#include "Wmodru.h"
#include "Wmodrc.h"

/* RCSHEADER "$Id: modrr.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

/* This file provides one essential component of the implementation:
  the reader.  Otherwise, the code in here serves as a sort of
  convenient wrapper around the functionality of POD mode collector /
  user container.  In fact, when the model reduction code was originally written,
  the functionality of both collector/user container was bundled in
  one file, so this represents an effort not to break existing code in
  the WASP library.  Evidently, the collector/user container routines
  may be called directly without this front-end.

 */

typedef struct W_modr_t {
  W_domain_t         *domain;        /* domain */
  W_modr_collector_p   collector;
  W_modr_user_p        user;
  W_bool_t             should_save_data;
}              W_modr_t;

void
W_modr_accumulate (W_modr_p modr)
{
  if (modr->collector) {
    W_modr_collector_accumulate (modr->collector);
  }
}

void
W_modr_compute_modes (W_modr_p modr)
{
  if (!modr->collector) REPORT_FATAL_ERROR (modr->domain->errh, "No collector");
  W_modr_collector_compute_modes (modr->collector);
}

#ifdef X_XFACE
void
W_modr_show_mode (W_modr_p modr, int mode)
{
  if (modr->collector) W_modr_collector_show_mode (modr->collector, mode);
  else if (modr->user) W_modr_user_show_mode (modr->user, mode);
}
#endif

double
W_modr_eigenvalue (W_modr_p modr, int mode)
{
  if (modr->collector) return  W_modr_collector_eigenvalue (modr->collector, mode);
  else                 return  0;
}

int
W_modr_tot_modes (W_modr_p modr)
{
  if (modr->collector) return W_modr_collector_tot_modes (modr->collector);
  else if (modr->user) return W_modr_user_get_nmodes (modr->user);
  else                 return 0;
}

int
W_modr_write_mean (W_modr_p modr, FILE *fp)
{
  if (!modr->collector) REPORT_FATAL_ERROR (modr->domain->errh, "No collector");
  return W_modr_collector_write_mean (modr->collector, fp);
}


int
W_modr_write_mode (W_modr_p modr, int mode, FILE *fp)
{
  if (!modr->collector) REPORT_FATAL_ERROR (modr->domain->errh, "No collector");
  return W_modr_collector_write_mode (modr->collector, mode, fp);
}

int
W_modr_get_nmodes (W_modr_p modr)
{
  if (!modr->user) REPORT_FATAL_ERROR (modr->domain->errh, "No user");
  return W_modr_user_get_nmodes (modr->user);
}

void
W_modr_dof_modes (W_modr_p modr, W_index_t dof, double *v, double *mean)
{
  if (!modr->user) REPORT_FATAL_ERROR (modr->domain->errh, "No user");
  W_modr_user_dof_modes (modr->user, dof, v, mean);
}

static void
skip_input (W_domain_t *d)
{
  const char *input_line;
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      REPORT_FATAL_ERROR (d->errh, "Missing END");
    TOKENS_PARSE(input_line);
    if (TOKENS_HAVE_KEYWORD ("END")) {
      break;
    }
  }
}

static W_bool_t use_modr = FALSE;
static W_bool_t collect_modr = FALSE;
static int modr_nm = 0;
static int modr_max_nm = 0;
static LIST_OF (int) modenl = NULL;
static double modr_dt = 0;
static double modr_up_to_t = FLT_MAX;
static double modr_from_t = 0;
static W_modr_method_t method = MODR_DIRECT;
static double modr_cutoff_ev = -1; /* no constraint */
static W_bool_t modr_dont_subtract_mean = FALSE;
static char *modr_file = NULL;

static void
parse_command_line_options (W_domain_t *d)
{
  if (WOPT_OPT (d->options, modr_options)) {
    int pos;
    tokens_parser parser = tokens_new_parser ();
    if (tokens_parse_line (parser, WOPT_OPT (d->options, modr_options))) {
      if ((pos = tokens_have_keyword (parser, 1, "USE")) > 0) {
        use_modr = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "COLLECT")) > 0) {
        collect_modr = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "METHOD")) > 0) {
        if (tokens_have_keyword (parser, pos, "DIRECT"))
          method = MODR_DIRECT;
        else
          method = MODR_SNAPSHOTS;
      }
      if ((pos = tokens_have_keyword (parser, 1, "DONT_SUBTRACT_MEAN")) > 0) {
        modr_dont_subtract_mean = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "CUTOFF_EV")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modr_cutoff_ev = tokens_token_as_double (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "CUTOFF_EV option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "NM")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modr_nm = tokens_token_as_int (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "NM option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "MAX_NM")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modr_nm = tokens_token_as_int (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "MAX_NM option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "NMODES")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modr_nm = modr_max_nm = tokens_token_as_int (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "NMODES option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "MODENL")) > 0) {
        int i = 1, m;
        modenl = make_list (); CHECK_MALLOC (modenl);
        while (tokens_total_of_tokens (parser) >= pos+i &&
               tokens_type_of_token (parser, pos+i) == TOKENS_NUMBER) {
          m = tokens_token_as_int (parser, pos+i);
          add_to_tail (modenl, (void *) m);
          i++;
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "FILE")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1 &&
            tokens_type_of_token (parser, pos+1) == TOKENS_STRING) {
          const char *s = tokens_token_as_string (parser, pos+1);
          NEWP_DUPL_STRING (modr_file, s,
                            { REPORT_FATAL_ERROR (d->errh, "Failed in string copy"); });
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "DT")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modr_dt = tokens_token_as_double (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "DT option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "UP_TO_T")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modr_up_to_t = tokens_token_as_double (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "UP_TO_T option requires a number");
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "FROM_T")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modr_from_t = tokens_token_as_double (parser, pos+1);
        } else {
          REPORT_FATAL_ERROR (d->errh, "FROM_T option requires a number");
        }
      }
     }
    tokens_delete_parser (parser);
  }
}

/*
  Reduction reader.
  Reader of configuration data for model reduction through POD (Proper Orthogonal Decomposition).
 */

W_bool_t 
W_modr_reader (W_domain_t *d)
{
  const char *input_line;
  const char *s;
  double inflate = 0;

  parse_command_line_options (d);

  if (!(use_modr || collect_modr)) /* disregard MODR */
    return W_read_and_discard_block (d);

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      TOKENS_PARSE(input_line);
      if        (TOKENS_HAVE_KEYWORD ("COLLECT")) {
        if (collect_modr) {
          d->modr = w_new (W_modr_t); CHECK_MALLOC (d->modr);
          d->modr->domain = d;
          d->modr->user   = NULL;
          d->modr->should_save_data = FALSE;
          d->modr->collector = W_modr_collector_new (d);
          W_modr_collector_set_dt (d->modr->collector, modr_dt);
          W_modr_collector_set_up_to_t (d->modr->collector, modr_up_to_t);
          W_modr_collector_set_from_t (d->modr->collector, modr_from_t);
          W_modr_collector_set_cutoff_ev (d->modr->collector, modr_cutoff_ev);
          W_modr_collector_set_method (d->modr->collector, method);
          W_modr_collector_set_dont_subtract_mean (d->modr->collector, modr_dont_subtract_mean);
          while ((input_line = W_get_input_line(d)) != NULL) {
            TOKENS_PARSE (input_line);
            if        (TOKENS_HAVE_KEYWORD ("DT")) {
              if (W_modr_collector_get_dt (d->modr->collector) <= 0) /* command line arguments override input file */
                W_modr_collector_set_dt (d->modr->collector, TOKEN_DBL (2));
            } else if (TOKENS_HAVE_KEYWORD ("SNAPSHOTS")) {
              W_modr_collector_set_method (d->modr->collector, MODR_SNAPSHOTS);
            } else if (TOKENS_HAVE_KEYWORD ("DIRECT")) {
              W_modr_collector_set_method (d->modr->collector, MODR_DIRECT);
            } else if (TOKENS_HAVE_KEYWORD ("DONT_SUBTRACT_MEAN")) {
              W_modr_collector_set_dont_subtract_mean (d->modr->collector, TRUE);
            } else if (TOKENS_HAVE_KEYWORD ("SAVE_DATA")) {
              d->modr->should_save_data = TRUE;
            } else if (TOKENS_HAVE_KEYWORD ("END")) {
              break;
            }
          }
        } else {
          skip_input (d);
        }
      } else if (TOKENS_HAVE_KEYWORD ("USE")) {
        if (use_modr) {
          d->modr = w_new (W_modr_t); CHECK_MALLOC (d->modr);
          d->modr->domain = d;
          d->modr->collector = NULL;
          d->modr->user      = W_modr_user_new (d);
          d->modr->should_save_data = FALSE;
          if (modenl) {
            W_modr_user_set_max_nm (d->modr->user, list_length (modenl));
            W_modr_user_set_modenl (d->modr->user, copy_list (modenl, NOT_NODES));
          } else {
            if (modr_max_nm <= 0) modr_max_nm = modr_nm;
            W_modr_user_set_max_nm (d->modr->user, modr_max_nm);
          }
          {
            char *file = NULL;
            while ((input_line = W_get_input_line(d)) != NULL) {
              TOKENS_PARSE(input_line);
              if        (TOKENS_HAVE_KEYWORD ("FILE")) {
                const char *s = TOKEN_STR (2);
                NEWP_DUPL_STRING (file, s,
                                  { REPORT_FATAL_ERROR (d->errh, "Failed in string copy"); });
              } else if (TOKENS_HAVE_KEYWORD ("NMODES")) {
                if (W_modr_user_get_max_nm (d->modr->user) == 0) /* only if not given on command line*/
                  W_modr_user_set_max_nm (d->modr->user, TOKEN_INT (2));
              } else if (TOKENS_HAVE_KEYWORD ("END")) {
                if (modr_file != NULL) {
                  NEWP_DUPL_STRING (file, modr_file,
                                    { REPORT_FATAL_ERROR (d->errh, "Failed in string copy"); });
                } else {
                  if (file == NULL) {
                    sprintf (W_buf (), "%s.pod", WOPT_OPT (d->options, in_file_basename));
                    NEWP_DUPL_STRING (file, W_buf (),
                                      { REPORT_FATAL_ERROR (d->errh, "Failed in string copy"); });
                  }
                }
                W_modr_user_read_modes (d->modr->user, file);
                break;
              }
            }
            if (file) w_free (file);
          }
        } else {
          skip_input (d);
        }
      } else {
        REPORT_FATAL_ERROR (d->errh, "Unknown keyword");
      }
    }
  }
  return FALSE;
}

int
W_modr_is_in_collect_mode (W_modr_p modr)
{
  return (modr->collector != NULL);
}

int
W_modr_is_in_use_mode (W_modr_p modr)
{
  return (modr->user != NULL);
}

int
W_modr_save_collected_data (W_modr_p modr, FILE *fp) 
{
  if (!modr->collector) REPORT_FATAL_ERROR (modr->domain->errh, "No collector");
  return W_modr_collector_save_data (modr->collector, fp);
}

int
W_modr_write_modr (W_modr_p modr, FILE *fp)
{
  if (!modr->collector) REPORT_FATAL_ERROR (modr->domain->errh, "No collector");
  return W_modr_collector_write_modr (modr->collector, fp);
}

int
W_modr_revert_to_orig_modes (W_modr_p modr)
{
  if (!modr->user) REPORT_FATAL_ERROR (modr->domain->errh, "No user");
  return W_modr_user_revert_to_orig_modes (modr->user);
}

int
W_modr_include_mode (W_modr_p modr, int to_replace, double *data)
{
  if (!modr->user) REPORT_FATAL_ERROR (modr->domain->errh, "No user");
  return W_modr_user_include_mode (modr->user, to_replace, data);
}

int
W_modr_should_save_data (W_modr_p modr)
{
  return ( W_modr_is_in_collect_mode (modr) && modr->should_save_data);
}
