#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Woptions.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wloads.h"
#include "Wdispbc.h"
#include "Wivcond.h"
#include "Wmats.h"
#include "Wcontact.h"
#include "Wvset.h"
#include "Wpvm.h"
#include "Wmpi.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wmodrr.h"
#include "Wmodru.h"
#include "meschach.h"

/* RCSHEADER "$Id: modru.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

typedef struct dof_modes_array_t {
  double *a;
}              dof_modes_array_t;

typedef struct mode_mat_t {
  W_modr_user_p        modr;
  int                 nm;
  dof_modes_array_t   *dof_modes;
}              mode_mat_t;

typedef struct W_modr_user_t {
  W_domain_t         *domain;
  /* configuration parameters */
  int                 max_nm;
  LIST_OF (int)       modenl;
  /* internal variables */
  int                 dof_modes_dim;
  int                 curr_m;
  dict_dictionary     node_transf_matd;
  mode_mat_t         *mm;
  mode_mat_t         *mm_orig;
}              W_modr_user_t;

/*
  New model reduction user data.
  Create a new user data container.
 */
W_modr_user_p /*  user data container  */
W_modr_user_new (W_domain_t *d /* domain (client) */
                )
{
  W_modr_user_p modr = w_new (W_modr_user_t); CHECK_MALLOC(modr);
  modr->domain = d;
  modr->max_nm = 0;
  modr->dof_modes_dim = 0;
  modr->curr_m = 0;
  modr->modenl = NULL;
  modr->node_transf_matd = NULL;
  modr->mm = NULL;
  modr->mm_orig = NULL;
  return modr;
}

static mode_mat_t *
mm_new (W_modr_user_p modr)
{
  mode_mat_t *mm;
  int i, j;
  int mnm1 = modr->max_nm+1;
  
  mm = w_new (mode_mat_t); CHECK_MALLOC (mm);
  mm->modr = modr;
  mm->nm  = modr->max_nm;
  mm->dof_modes = w_new_array (modr->dof_modes_dim, dof_modes_array_t);
  CHECK_MALLOC (mm->dof_modes);
  FOR (i, modr->dof_modes_dim) {
    mm->dof_modes[i].a = w_new_array (mnm1, double); 
    CHECK_MALLOC (mm->dof_modes[i].a);
    FOR (j, mnm1) {
      /*      printf ("i=%d, j=%d, mm->dof_modes=%g\n", i, j, mm->dof_modes[i].a[j]);*/
      mm->dof_modes[i].a[j] = 0;
    }
  }

  return mm;
}

static void
mm_orthonormalize (mode_mat_t *mm);

static void
mm_init (mode_mat_t *mm)
{
  int i, j, k;
  W_node_t *n;
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  MAT *ntm;
  
  n = NULL;
  while (n = WD_get_next_node (mm->modr->domain, n)) {
    WN_eqn_numbers_and_constraints (n, ens, constrained);
    ntm = dict_find_obj_with_int_name (mm->modr->node_transf_matd, n->id);
    if (!ntm) {
      REPORT_FATAL_ERROR (mm->modr->domain->errh,
                          "Nodal transformation matrix not found");
    }
    FOR (k, 3) {
      i = ens[k];
      FOR (j, mm->modr->max_nm+1) { mm->dof_modes[i].a[j] = ntm->me[k][j]; }
    }
  }
  mm->nm = mm->modr->max_nm;
  mm_orthonormalize (mm);
}

static void
mm_copy (mode_mat_t *mm_dest, mode_mat_t *mm_src)
{
  int i, j;
  int n, m;
  n = mm_dest->modr->dof_modes_dim;
  m = mm_dest->modr->max_nm + 1;
  FOR (i, n) {
    FOR (j, m) { mm_dest->dof_modes[i].a[j] = mm_src->dof_modes[i].a[j]; }
  }
}

static void
mm_set (mode_mat_t *mm, int which, double *v)
{
  int i, n;
  dof_modes_array_t *dof_modes = mm->dof_modes;
  n = mm->modr->dof_modes_dim;
  FOR (i, n) {
    dof_modes[i].a[which] = v[i];
  }
}

static double
mm_dot (mode_mat_t *mm, int m0, int m1)
{
  int i, n;
  double d = 0;
  dof_modes_array_t *dof_modes = mm->dof_modes;
  n = mm->modr->dof_modes_dim;
  FOR (i, n) {
    d += dof_modes[i].a[m0] * dof_modes[i].a[m1];
  }
  return d;
}

static void
mm_subtract (mode_mat_t *mm, int which, double d, int from)
{
  int i, n;
  dof_modes_array_t *dof_modes = mm->dof_modes;
  n = mm->modr->dof_modes_dim;
  FOR (i, n) {
    dof_modes[i].a[from] -= d * dof_modes[i].a[which];
  }
}

static void
mm_normalize (mode_mat_t *mm, int which)
{
  int i, n;
  double d = 0;
  dof_modes_array_t *dof_modes = mm->dof_modes;
  n = mm->modr->dof_modes_dim;
  d = mm_dot (mm, which, which);
  d = sqrt (d);
  FOR (i, n) {
    dof_modes[i].a[which] /= d;
  }
}

static void
mm_orthonormalize (mode_mat_t *mm)
{
  int i, j, k, nm = mm->nm;
  double d;
  j = 0;
  mm_normalize (mm, 0);
  for (i = 1; i < nm; i++) {
    for (k = i; k < nm; k++) {
      d = mm_dot (mm, k, j);
      mm_subtract (mm, j, d, k);
    }
    j++;
    mm_normalize (mm, j);
  }
}

static double
dot (double *a, double *b, int M)
{
  int k;
  double accum = 0;
  FOR (k, M) { accum += a[k] * b[k]; }
  return accum;
}


static MAT *
new_ntm (W_modr_user_p modr, W_index_t id)
{
  MAT *ntm;
  ntm = m_get (3, modr->max_nm+1); CHECK_MALLOC (ntm);
  dict_add_obj_int_name (modr->node_transf_matd, ntm, id);
  m_zero (ntm);
  return ntm;
}

/*
  Read mean vector.
  Read mean vector.  Use the parser which has to be set up to
  read and parse the input file.
 */
void
W_modr_user_read_mean (W_modr_user_p modr,/* model reduction user data container  */
                      tokens_parser parser, /* parser */
                      W_domain_t *d /* domain (client) */
                      )
{
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) { BAD_INPUT (); }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      return;
    } else {
      int id;
      double ux, uy, uz;
      MAT *ntm;
      
      if (tokens_total_of_tokens (parser) < 4) { BAD_INPUT (); }
      id = tokens_token_as_int (parser, 1);
      ux = tokens_token_as_double (parser, 2);
      uy = tokens_token_as_double (parser, 3);
      uz = tokens_token_as_double (parser, 4);
      /* store the mode vector in the node matrix */
      ntm = dict_find_obj_with_int_name (modr->node_transf_matd, id);
      if (!ntm) ntm = new_ntm (modr, id);
      ntm->me[0][modr->max_nm] = ux;
      ntm->me[1][modr->max_nm] = uy;
      ntm->me[2][modr->max_nm] = uz;
    }
  }
}

/*
  Read mode vector.
  Read mode vector.  Use the parser which has to be set up to
  read and parse the input file.
 */
void
W_modr_user_read_mode (W_modr_user_p modr,/* model reduction user data container  */
                      tokens_parser parser, /* parser */
                      W_domain_t *d /* domain (client) */
                      )
{
  double ev = 0;
  int mode = modr->curr_m;
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) { BAD_INPUT (); }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      modr->curr_m++;
      return;
    } else {
      int id;
      double ux, uy, uz;
      MAT *ntm;
      
      if (tokens_total_of_tokens (parser) < 4) { BAD_INPUT (); }
      id = tokens_token_as_int (parser, 1);
      ux = tokens_token_as_double (parser, 2);
      uy = tokens_token_as_double (parser, 3);
      uz = tokens_token_as_double (parser, 4);
      /* store the mode vector in the node matrix */
      ntm = dict_find_obj_with_int_name (modr->node_transf_matd, id);
      if (!ntm) ntm = new_ntm (modr, id);
      ntm->me[0][mode] = ux;
      ntm->me[1][mode] = uy;
      ntm->me[2][mode] = uz;
    }
  }
}

static void
skip_till_END (tokens_parser parser, W_domain_t *d)
{
  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) { BAD_INPUT (); }
    key = tokens_token_as_string (parser, 1);
    if (tokens_token_and_keyword_equiv (key, "END")) {
      return;
    } 
  }
}

static int
to_be_loaded (W_modr_user_p modr, int mode)
{
  if (modr->modenl) {
    return in_list (modr->modenl, (void *)mode);
  } else {
    return (mode >= 0 && mode < modr->max_nm);
  }
}

/*
  Read POD modes.
  Read POD data from the named file.
 */
void
W_modr_user_read_modes (W_modr_user_p modr,/*  user data container  */
                       const char *file /* POD file written in the format
                                           described in modrc.c */
                       )
{
  tokens_parser parser;
  int mode = -1;
  W_domain_t *d = modr->domain;
  
  if (modr->max_nm <= 0)
    REPORT_FATAL_ERROR (modr->domain->errh, "Wrong number of modes");

  parser  = tokens_new_parser ();

  if (!tokens_open_file (parser, (char *) file)) {
    sprintf (W_buf (), "Could not open POD file \"%s\"", file);
    REPORT_FATAL_ERROR (modr->domain->errh, W_buf ());
  }

  modr->node_transf_matd   = dict_create_dictionary ();
  CHECK_MALLOC (modr->node_transf_matd);
  modr->dof_modes_dim = WD_tot_of_equations (modr->domain);
  modr->mm = modr->mm_orig = NULL;
  
  modr->curr_m = 0;

  while (tokens_next_line (parser)) {
    const char *key, *input_line = tokens_curr_line_contents (parser);
    if (tokens_total_of_tokens (parser) < 1) { BAD_INPUT (); }
    key = tokens_token_as_string (parser, 1);
    if        (tokens_token_and_keyword_equiv (key, "MODE")) {
      if (tokens_total_of_tokens (parser) < 2) { BAD_INPUT (); }
      mode = tokens_token_as_int (parser, 2);
      if (to_be_loaded (modr, mode)) {
        W_modr_user_read_mode (modr, parser, modr->domain);
      } else {
        skip_till_END (parser, modr->domain);
      }
    } else if (tokens_token_and_keyword_equiv (key, "MEAN")) {
      W_modr_user_read_mean (modr, parser, modr->domain);
    } else {
      sprintf (W_buf (), "Unknown keyword \"%s\"", key);
      REPORT_FATAL_ERROR (modr->domain->errh, W_buf ());
    }
  }
  
  tokens_close_curr_file (parser);
  tokens_delete_parser (parser);
}

#define CHECK_ORTHONORMAL(i,j,d)                                        \
  if ((i == j) && (fabs (d - 1) > 1e-6)) {                              \
    max_d = max (max_d, fabs (d - 1));                                  \
    all_orthonormal = FALSE;                                            \
  }                                                                     \
  if ((i != j) && (fabs (d) > 1e-6)) {                                  \
    max_d = max (max_d, fabs (d));                                      \
    all_orthonormal = FALSE;                                            \
  }

static void
check_orthonorm_u (W_modr_user_p modr)
{
  W_index_t toteqn = WD_tot_of_equations (modr->domain);
  int i, j, k;
  int M = modr->max_nm;
  double d;
  W_bool_t all_orthonormal = TRUE;
  double max_d = 0;

  FOR (i, M) {
    FOR (j, M) {
      d = 0;
      FOR (k, toteqn) {
        d += modr->mm->dof_modes[k].a[i] * modr->mm->dof_modes[k].a[j];
      }
      CHECK_ORTHONORMAL (i, j, d);
    }
  }
  if (!all_orthonormal) {
    sprintf (W_buf (), "Some modes not orthonormal; max mismatch %g", max_d);
    REPORT_WARNING (modr->domain->errh, W_buf ());
  }
}

static void
build_dof_modes (W_modr_user_p modr)
{
  modr->mm_orig = mm_new (modr);
  mm_init (modr->mm_orig);
  modr->mm = mm_new (modr);
  mm_copy (modr->mm, modr->mm_orig);
  check_orthonorm_u (modr);
}


/*
  Get number of modes.
  Get number of modes available in the user data container.
 */
int /* number of modes available in the container */
W_modr_user_get_nmodes (W_modr_user_p modr)
{
  if (!modr->mm)
    return modr->max_nm;
  else
    return modr->mm->nm;
}

/*
  Get the model-reduction modes.
  Get the amplitudes of the model reduction modes for a given degree of freedom.
  
 */
void
W_modr_user_dof_modes (W_modr_user_p modr, /*  user data container  */
                      W_index_t dof, /* degree of freedom;
                                        integer >= 0 && < max_dof, where max_dof
                                        is the total number of degrees of freedom
                                      */
                      double v[], /* output: amplitudes of modes */
                      double *mean /* output: value of mean  */
                      )
{
  dof_modes_array_t *dof_modes;
  int nm;
  int i;
  if (modr->mm == NULL) build_dof_modes (modr);
  dof_modes = modr->mm->dof_modes;
  nm = modr->mm->nm;
  FOR (i, nm) { v[i] = dof_modes[dof].a[i]; }
  *mean = dof_modes[dof].a[nm];
}

/*
  Revert to original modes.
  Reset the user data container to use
  only the original modes as read from the input file.
 */
int /* ignore */
W_modr_user_revert_to_orig_modes (W_modr_user_p modr/*  user data container  */
                                 )
{
  if (modr->mm == NULL) return FALSE;
  mm_copy (modr->mm, modr->mm_orig);
  return TRUE;
}

/*
  Include a mode.
  Include the mode on input among the model-reduction modes.
  It is to replace one of the modes, to_replace, so
  if this replaced mode is to be made available again,
  W_modr_user_revert_to_orig_modes() needs to be called.
 */
int /* ignore */
W_modr_user_include_mode (W_modr_user_p modr,/*  user data container  */
                         int to_replace, /* replace this mode */
                         double *data /* modal vector */
                         )
{
  mm_set (modr->mm, to_replace, data);
  mm_orthonormalize (modr->mm);
#if 0
  check_orthonorm_u (modr);      /* RAW unnecessary when debugged */
#endif
  return TRUE;
}

/*
  Set maximum of modes.
  Set maximum of modes.
 */
int /* previous max_nm */
W_modr_user_set_max_nm (W_modr_user_p modr, /*  user data container  */
                       int max_nm /* maximum number of modes */
                       )
{
  int prev_max_nm = modr->max_nm;
  modr->max_nm = max_nm;
  return prev_max_nm;
}

/*
  Get maximum number of modes.
  Get the maximum number of modes.
 */
int /* maximum number of modes  */
W_modr_user_get_max_nm (W_modr_user_p modr/*  user data container  */) 
{
  return modr->max_nm;
}

/*
  Get a list of modes.
 */
LIST_OF (int) /* list of modes */
W_modr_user_set_modenl (W_modr_user_p modr, /*  user data container  */
                       LIST_OF (int)  modenl /* list of modes */
                       )
{
  LIST_OF (int)  prev_modenl = modr->modenl;
  modr->modenl = modenl;
  return prev_modenl;
}


#ifdef X_XFACE

#ifdef REPLACE
# undef REPLACE
#endif
#ifdef ADD
# undef ADD
#endif
#define REPLACE FALSE
#define ADD     TRUE

static void
set_mode (W_modr_user_p modr, int mode, W_bool_t add, double scale)
{
  W_node_t *n;
  W_vect_t d;
  W_bool_t apply[3] = {TRUE, TRUE, TRUE};
#if W_ROT_DOFS
  W_index_t ens[6];
  W_bool_t constrained[6];
#else
  W_index_t ens[3];
  W_bool_t constrained[3];
#endif
  double *v, mean;
  int dofpn;
  int nm = W_modr_user_get_nmodes (modr);

  if (mode < 0) mode = 0;
  v = w_new_array (nm, double); CHECK_MALLOC (v);

  n = NULL;
  while (n = WD_get_next_node (modr->domain, n)) {
    dofpn = WN_eqn_numbers_and_constraints (n, ens, constrained);
    if (add) {
      WN_get_displ (n, &d.x, &d.y, &d.z);
      W_modr_user_dof_modes (modr, ens[0], v, &mean);
      d.x += scale * v[mode];
      W_modr_user_dof_modes (modr, ens[1], v, &mean);
      d.y += scale * v[mode];
      W_modr_user_dof_modes (modr, ens[2], v, &mean);
      d.z += scale * v[mode];
    } else {
      W_modr_user_dof_modes (modr, ens[0], v, &mean);
      d.x = scale * v[mode];
      W_modr_user_dof_modes (modr, ens[1], v, &mean);
      d.y = scale * v[mode];
      W_modr_user_dof_modes (modr, ens[2], v, &mean);
      d.z = scale * v[mode];
    }
    WN_set_displ (n, &d, apply);
  }

  w_free (v);
}

void
W_modr_user_show_mode (W_modr_user_p modr, int mode)
{
  if (modr == NULL) { REPORT_WARNING (modr->domain->errh, "Not in user mode"); return; }
  set_mode (modr, mode, REPLACE, 1);
  WD_build_grep (modr->domain, TRUE);
}
#endif
