#include "Wconfig.h"

#if defined(MPI_WASP) && MPI_WASP
 
#include "Wxf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wvset.h"
#include "EID.h"

#include "Wmpi.h"

/* RCSHEADER "$Id: mpi.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

void
W_mpi_realloc_sbuf (W_mpi_comm_buff_t *comm_buff
                    )
{
  if (comm_buff->sbuf_dim < comm_buff->count) {
    comm_buff->sbuf_dim = comm_buff->count;
    comm_buff->sbuf
      = w_realloc_array (comm_buff->sbuf, comm_buff->sbuf_dim, double);
    CHECK_MALLOC (comm_buff->sbuf);
  }
}

void
W_mpi_realloc_rbuf (W_mpi_comm_buff_t *comm_buff
                    )
{
  if (comm_buff->rbuf_dim < comm_buff->count) {
    comm_buff->rbuf_dim = comm_buff->count;
    comm_buff->rbuf
      = w_realloc_array (comm_buff->rbuf, comm_buff->rbuf_dim, double);
    CHECK_MALLOC (comm_buff->rbuf);
  }
}

static void
isend (W_mpi_process_t *p,    /* IN: process */
       int to,                /* IN: destination partition */
       int num_vals_per_node, /* IN: number of values per node */
       int tag,               /* IN: message tag */
       W_mpi_pack_func pack   /* IN: pack function */
       )
{
  W_mpi_comm_buff_t *cb = &p->comm_buffs[to];
  if (cb->num_to_send > 0) { 
    MPI_Request req;
    cb->count = num_vals_per_node * cb->num_to_send;
    W_mpi_realloc_sbuf (cb); /* re-size */
    pack (p, cb);    /* pack into buffer */
    MPI_Isend (cb->sbuf, cb->count, MPI_DOUBLE,                 
               cb->rank, tag, MPI_COMM_WORLD, &req);
  } 
}

static void
irecv (W_mpi_process_t *p,    /* IN: process */
       int from,                /* IN: source partition */
       int num_vals_per_node, /* IN: number of values per node */
       int tag                  /* IN: message tag */
  )
{
  W_mpi_comm_buff_t *cb = &p->comm_buffs[from];
  if (cb->num_to_recv > 0) { /* receive expected */
    cb->count = num_vals_per_node * cb->num_to_recv;
    W_mpi_realloc_rbuf (cb); /* re-size */
    MPI_Irecv (cb->rbuf, cb->count, MPI_DOUBLE,                 
               cb->rank, tag, MPI_COMM_WORLD,
               &p->recv_requests[from]);
  } else { /* nothing to receive; set request to NULL */
    p->recv_requests[from] = MPI_REQUEST_NULL;
  }
} 

static void
wait_all (W_mpi_process_t *p,      /* IN: process */
          W_mpi_unpack_func unpack /* IN: unpack function */
         )
{
  MPI_Status status;
  int i, num_comm, idx;

  num_comm = 0; /* count receives */
  FOR (i, p->size) {
    if (p->comm_buffs[i].num_to_recv > 0) num_comm++;
  }
  
  while (num_comm-- > 0) { /* wait for all receives */
    MPI_Waitany (p->size, p->recv_requests, &idx, &status);
    unpack (p, &p->comm_buffs[idx]); /* unpack */
  }

  MPI_Barrier (MPI_COMM_WORLD); /* synchronize */
}

void
W_mpi_exchange (W_domain_t *d         /* IN: domain */,
                int num_vals_per_node, /* IN: number of double values per node */
                W_mpi_pack_func pack, /* IN: pack function */
                W_mpi_unpack_func unpack, /* IN: unpack function */
                int tag               /* IN: message tag */
                )
{
  W_mpi_process_t *p = &d->mpi_process; /* process */
  int i;

#if TIMED_WASP
  W_timer_restart (d->comm_timer);
#endif

  FOR (i, p->size) { /* initialize sends and receives */
    isend (p, i, num_vals_per_node, tag, pack);
    irecv (p, i, num_vals_per_node, tag);
  }

  wait_all (p, unpack);

#if TIMED_WASP
  W_timer_accumulate (d->comm_timer);
#endif 
}

#endif /* only for MPI WASP: common code for both node- and element-cut */
