#include "Wconfig.h"

#if defined(MPI_WASP) && MPI_WASP && defined (MPI_WASP_ELEM_CUT) && MPI_WASP_ELEM_CUT
 
#include "Wxf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wvset.h"
#include "EID.h"

#include "Wmpi.h"

/* RCSHEADER "$Id: mpi_ec.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

static void
build_node_lists (W_mpi_process_t *p)
{
  W_mpi_comm_buff_t *cb;
  W_mpi_npmap_pair_t *pair, *pairs = p->node_part_map.pairs;
  int i, part, j, k, np = p->node_part_map.num_pairs;

  FOR (part, p->size) {
    if (part != p->rank) {      /* not self */
      cb = &p->comm_buffs[part];
      if (cb->num_to_send > 0) { /* anything to send? */
        cb->to_send = w_new_array (cb->num_to_send, W_node_t *);
        CHECK_MALLOC (cb->to_send);
        k = 0; 
        FOR (i, np) {
          pair = &pairs[i];
          if (cb->send_flag[i]) { /* To be sent? */
            cb->to_send[k++] = pair->n;
          }
        } /* FOR (i, np) */
        if (k != cb->num_to_send) REPORT_FATAL_ERROR ("mismatch");
      } /* if (cb->num_to_send > 0) */
      if (cb->num_to_recv > 0) { /* anything to receive? */
        cb->to_recv = w_new_array (cb->num_to_recv, W_node_t *);
        CHECK_MALLOC (cb->to_recv);
        k = 0; 
        FOR (i, np) {
          pair = &pairs[i];
          if (pair->n &&              /* Defined (own or remote)?*/
              pair->rank == cb->rank  /* Owned by /cb->rank/? */
              ) {
            cb->to_recv[k++] = pair->n;
          }
        } /* FOR (i, np) */
        if (k != cb->num_to_recv) REPORT_FATAL_ERROR ("mismatch");
      } /* if (cb->num_to_recv > 0) */
    } /* not self */
  } /* FOR (part, p->size) */
}


static void
def_comm_buff (W_mpi_process_t *p, int to, int *procranks)
{
  int i, nn;

  p->comm_buffs[to].num_to_send = 0;
  nn = p->node_part_map.num_pairs;

  p->comm_buffs[to].send_flag = w_new_array (nn, W_short_bool_t);
  CHECK_MALLOC (p->comm_buffs[to].send_flag);

  FOR (i, nn) {
    if (procranks[i] == p->rank) { /* wishes to receive from me? */
      p->comm_buffs[to].send_flag[i] = TRUE; /* send this node */
      p->comm_buffs[to].num_to_send++;
    }
  }
}

/*
  Build communication buffers.
  Build communication buffers by exchanging information with
  all the peers.
 */
void
W_mpi_build_comm_buffs (W_mpi_process_t *p)
{
  int i, np;
  int *sproc, *rproc;
  W_mpi_npmap_pair_t *pairs;

  if (p->rank < 0 || p->size < 1) {
      REPORT_FATAL_ERROR ("Parallel data not initialized");
  }
  
  sproc = w_new_array (p->node_part_map.num_pairs, int);
  CHECK_MALLOC (sproc);
  rproc = w_new_array (p->node_part_map.num_pairs, int);
  CHECK_MALLOC (rproc);

  /* build communication requests list */
  FOR (i, p->size) { p->comm_buffs[i].num_to_recv = 0; }
  pairs = p->node_part_map.pairs;
  np    = p->node_part_map.num_pairs;
  FOR (i, np) {
    if (pairs[i].n) { /* if node included in my partition */
      if (pairs[i].rank == p->rank) {
        sproc[i] = -1; /* owned by p */
      } else {
        sproc[i] = pairs[i].rank; /* should be sending to me */
        p->comm_buffs[pairs[i].rank].num_to_recv++;
      }
    } else
      sproc[i] = -1; /* not used by p */
  }

  /* exchange communication lists with all processes */
  np = p->node_part_map.num_pairs;
  FOR (i, p->size) {
    if (i == p->rank) {         /* sending */
      MPI_Bcast (sproc, np, MPI_INT, i, MPI_COMM_WORLD);
    } else {                    /* receiving */
      MPI_Bcast (rproc, np, MPI_INT, i, MPI_COMM_WORLD);
      def_comm_buff (p, i, rproc);
    }
    MPI_Barrier (MPI_COMM_WORLD); /* rendez-vous */
  }
  
  W_update_log ("Communication pattern (EC):");
  FOR (i, p->size) {
    if (p->comm_buffs[i].num_to_send > 0) {
      sprintf (W_buf(), "   Process %d sends data for %d nodes to %d",
               p->rank, p->comm_buffs[i].num_to_send, i);
      W_update_log (W_buf ());
      sprintf (W_buf(), "   Process %d receives data for %d nodes from %d",
               p->rank, p->comm_buffs[i].num_to_recv, i);
      W_update_log (W_buf ());
    }
  }

  build_node_lists (p);

  w_free (sproc);  /* no longer needed */
  w_free (rproc); /* no longer needed */
}

/*  
! MPI SETUP
PARALLEL
num_node_proc_pairs
n0 py
n1 px
...
nN pz
&

or

PARALLEL
MAP_FILE aName
&

*/

W_bool_t
W_mpi_run_reader (W_domain_t *d, FILE *stream)
{
  const char *input_line;
  W_index_t idx;
  int i, j;
  W_mpi_npmap_pair_t *pairs;
  W_mpi_comm_buff_t *comm_buffs;
  FILE *orig_stream = stream;

  input_line = W_get_input_line (stream);
  TOKENS_PARSE (input_line);

  /* Create process */
  d->mpi_process.domain        = d;
  MPI_Comm_rank (MPI_COMM_WORLD, &d->mpi_process.rank);
  MPI_Comm_size (MPI_COMM_WORLD, &d->mpi_process.size);
  if (TOKEN_IS_STRING (1)) {
    const char *k = TOKEN_STR (1);
    char *map_file;
    if (tokens_token_and_keyword_equiv (k, "MAP_FILE")) {
    } else {
      REPORT_FATAL_ERROR ("Expecting \"MAP_FILE aName\"");
    }
    if (TOKENS_TOTAL () < 2) {
      REPORT_FATAL_ERROR ("Expecting \"MAP_FILE aName\"");
    }
    map_file = (char *) TOKEN_STR (2);
    stream = fopen (map_file, "r");
    if (!stream) {
      sprintf (W_buf (), "Could not open node/partition map file \"%s\"",
               map_file);
      REPORT_FATAL_ERROR (W_buf ());
    }
    input_line = W_get_input_line (stream);
    TOKENS_PARSE (input_line);
  }
  d->mpi_process.node_part_map.num_pairs = TOKEN_INT (1);
  d->mpi_process.node_part_map.pairs      =
    w_new_array (d->mpi_process.node_part_map.num_pairs,
                 W_mpi_npmap_pair_t);
  CHECK_MALLOC (d->mpi_process.node_part_map.pairs);
  d->mpi_process.recv_requests =
    w_new_array (d->mpi_process.size, MPI_Request);
  CHECK_MALLOC (d->mpi_process.recv_requests);
  FOR (i, d->mpi_process.size) {
    d->mpi_process.recv_requests[i] = MPI_REQUEST_NULL;
  }
  
  /* Set up the node/processor map */
  pairs = d->mpi_process.node_part_map.pairs;
  FOR (i, d->mpi_process.node_part_map.num_pairs) {
    input_line = W_get_input_line (stream);
    TOKENS_PARSE (input_line);
    idx           = TOKEN_INT (1);
    pairs[i].n    = WD_get_node_by_idx (d, idx);
    pairs[i].rank = TOKEN_INT (2);
    if (pairs[i].rank == d->mpi_process.rank) { /* own this node */
      if (!pairs[i].n) { /* must have this node! */
        sprintf (W_buf (), "Node %d owned by proc %d not found",
                 (int)idx, d->mpi_process.rank);
        REPORT_FATAL_ERROR (W_buf ());
      }
    }
  }

  /* Initialize the communication buffers */
  d->mpi_process.comm_buffs
    = w_new_array (d->mpi_process.size, W_mpi_comm_buff_t);
  CHECK_MALLOC (d->mpi_process.comm_buffs);
  comm_buffs = d->mpi_process.comm_buffs;

  FOR (i, d->mpi_process.size) {
    comm_buffs[i].rank                = i;
    comm_buffs[i].num_to_send         = 0;
    comm_buffs[i].num_to_recv         = 0;
    comm_buffs[i].send_flag           = NULL;
    comm_buffs[i].sbuf_dim            = 0;
    comm_buffs[i].count               = 0;
    comm_buffs[i].sbuf                = NULL;
    comm_buffs[i].rbuf_dim            = 0;
    comm_buffs[i].rbuf                = NULL;
  }

  if (stream != orig_stream) {
    fclose (stream);
    stream = orig_stream;
  }
  
  input_line = W_get_input_line (stream);
  if (input_line[0] != END_OF_BLOCK)
    REPORT_FATAL_ERROR ("Parallel block does not end in '&'");

  return TRUE;
}

/*
  Pack data.
  Pack data into the send buffer of process /to/.
  Reallocate buffer if needed.  Node accelerations are
  packed in the order of the node/ processor map, x-, y-, and z-component.
 */
static void
pack_accels (W_mpi_process_t *p, /* IN: process */
             W_mpi_comm_buff_t *cb
             )
{
  int i, j, nn = cb->num_to_send;
  W_node_t *n, **npa = cb->to_send;
  double *sbp;
  
  sbp       = cb->sbuf;
  FOR (i, nn) {                 /* packing loop */
    n    = npa[i];
    *sbp = n->force_t.x; sbp++;
    *sbp = n->force_t.y; sbp++;
    *sbp = n->force_t.z; sbp++;
  }
}

static void
unpack_accels  (W_mpi_process_t *p, /* IN: process */
                W_mpi_comm_buff_t *cb /* IN: communication buffer */
                )
{
  int i, j, nn = cb->num_to_recv;
  W_node_t *n, **npa = cb->to_recv;
  double *rbp;
  
  rbp   = cb->rbuf;
  FOR (i, nn) {                 /* unpacking loop */
    n    = npa[i];
    n->force_t.x = *rbp; rbp++; /* increment force */
    n->force_t.y = *rbp; rbp++;
    n->force_t.z = *rbp; rbp++;
  }
}


  /* Communicate:
     (i) Export acceleration of nodes belonging to domain,
     (ii) Import acceleration of nodes belonging to neighbors.
     NOTE: The action solve_for leaves the computed accelerations
           in the force_t field of the node data, hence the routines
           in this file access this field to transfer the new
           (computed) accelerations.

     To communicate, the routine uses the communication buffers
     and the node/partition map.
   */
void
W_mpi_exchange_accels (W_domain_t *d /* domain */)
{
  static int tag = 0;
  W_mpi_exchange (d, 3, pack_accels, unpack_accels, ++tag);
}

#endif  /* only for MPI WASP and element-cut strategy */
