#include "Wconfig.h"

#if defined(MPI_WASP) && MPI_WASP && defined (MPI_WASP_NODE_CUT) && MPI_WASP_NODE_CUT
 
#include "Wxf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wvset.h"
#include "EID.h"

#include "Wmpi.h"
 
/* RCSHEADER "$Id: mpi_nc.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

static void
build_node_lists (W_mpi_process_t *p)
{
  W_mpi_comm_buff_t *cb;
  W_mpi_npmap_pair_t *pair, *pairs = p->node_part_map.pairs;
  int i, part, j, k, np = p->node_part_map.num_pairs;

  FOR (part, p->size) {
    if (part != p->rank) {      /* not self */
      cb = &p->comm_buffs[part];
      if (cb->num_to_send > 0) { /* anything to send/ receive? */
        cb->to_send = cb->to_recv = w_new_array (cb->num_to_send, W_node_t *);
        CHECK_MALLOC (cb->to_send);
        k = 0; 
        FOR (i, np) {
          pair = &pairs[i];
          if (pair->n && pair->ranks_dim > 1) { /* Present? Shared? */
            FOR (j, pair->ranks_dim) {
              if (pair->ranks[j] == cb->rank) { /* Shared with /cb->rank/? */
                cb->to_send[k++] = pairs[i].n;
                break; /* /cb->rank/ can be present only once */
              }
            } /* FOR (j, pair->ranks_dim) */
          }
        } /* FOR (i, np) */
        if (k != cb->num_to_send) REPORT_FATAL_ERROR ("mismatch");
      } /* if (cb->num_to_send > 0) */
    } /* not self */
  } /* FOR (part, p->size) */
}

/*
  Build communication buffer structures.
  Build communication buffers by exchanging information with
  all the peers.
 */
void
W_mpi_build_comm_buffs (W_mpi_process_t *p)
{
  int i, nn;
  W_mpi_comm_buff_t *comm_buffs = p->comm_buffs;
  W_mpi_npmap_pair_t *pairs;

  if (p->rank < 0 || p->size < 1) {
      REPORT_FATAL_ERROR ("Parallel data not initialized");
  }

  FOR (i, p->size) {
    comm_buffs[i].num_to_send = 0;
    comm_buffs[i].to_send     = NULL;
    comm_buffs[i].num_to_recv = 0;
    comm_buffs[i].to_recv     = NULL;
  }
  
  /* build communication requests list */
  pairs = p->node_part_map.pairs;
  nn    = p->node_part_map.num_pairs;
  FOR (i, nn) {
    if (pairs[i].n && pairs[i].ranks_dim > 1) { /* shared node */
      int j, nr;
      FOR (j, pairs[i].ranks_dim) {
        nr = pairs[i].ranks[j];
        if (nr != p->rank) { /* not self */
          comm_buffs[nr].num_to_send++;
        }
      }
    }
  }

  FOR (i, p->size) { /* the same number of receives as sends */
    comm_buffs[i].num_to_recv = comm_buffs[i].num_to_send;
  }

  build_node_lists (p);

  /* A good opportunity to synchronize; all processes have to have
     the communication buffers ready */
  MPI_Barrier (MPI_COMM_WORLD);
  
  W_update_log ("Communication pattern (NC):");
  FOR (i, p->size) {
    if (p->comm_buffs[i].num_to_send > 0) {
      sprintf (W_buf(), "   Process %d sends data for %d nodes to %d",
               p->rank, p->comm_buffs[i].num_to_send, i);
      W_update_log (W_buf ());
    }
  }
}

static W_bool_t
node_on_proc (W_mpi_npmap_pair_t *pairs, int rank)
{
  if        (pairs->ranks_dim == 1) {
    return ((int) (pairs->ranks) == rank); /* non-shared node */
  } else {
    return (pairs->ranks[0] == rank); /* ranks[0] is the owner */
  }
}


/*  
! MPI SETUP
PARALLEL
num_node_proc_pairs
n0 p0_0 ... p0_m0 ! the owning processor is the first rank
n1 p1_0 ... p1_m1
...
nN pN_0 ... pN_mN
&

or

PARALLEL
MAP_FILE aName
&

*/

W_bool_t
W_mpi_run_reader (W_domain_t *d, FILE *stream)
{
  const char *input_line;
  W_index_t idx;
  int i, j, nt;
  W_mpi_npmap_pair_t *pairs;
  W_mpi_comm_buff_t *comm_buffs;
  FILE *orig_stream = stream;

  input_line = W_get_input_line (stream);
  TOKENS_PARSE (input_line);

  /* Create process */
  d->mpi_process.domain        = d;
  MPI_Comm_rank (MPI_COMM_WORLD, &d->mpi_process.rank);
  MPI_Comm_size (MPI_COMM_WORLD, &d->mpi_process.size);
  if (TOKEN_IS_STRING (1)) {
    const char *k = TOKEN_STR (1);
    char *map_file;
    if (tokens_token_and_keyword_equiv (k, "MAP_FILE")) {
    } else {
      REPORT_FATAL_ERROR ("Expecting \"MAP_FILE aName\"");
    }
    if (TOKENS_TOTAL () < 2) {
      REPORT_FATAL_ERROR ("Expecting \"MAP_FILE aName\"");
    }
    map_file = (char *) TOKEN_STR (2);
    stream = fopen (map_file, "r");
    if (!stream) {
      sprintf (W_buf (), "Could not open node/partition map file \"%s\"",
               map_file);
      REPORT_FATAL_ERROR (W_buf ());
    }
    input_line = W_get_input_line (stream);
    TOKENS_PARSE (input_line);
  }
  d->mpi_process.node_part_map.num_pairs = TOKEN_INT (1);
  d->mpi_process.node_part_map.pairs =
    w_new_array (d->mpi_process.node_part_map.num_pairs,
                 W_mpi_npmap_pair_t);
  CHECK_MALLOC (d->mpi_process.node_part_map.pairs);
  d->mpi_process.recv_requests =
    w_new_array (d->mpi_process.size, MPI_Request);
  CHECK_MALLOC (d->mpi_process.recv_requests);
  FOR (i, d->mpi_process.size) {
    d->mpi_process.recv_requests[i] = MPI_REQUEST_NULL;
  }

  /* Set up the node/processor map */
  pairs = d->mpi_process.node_part_map.pairs;
  FOR (i, d->mpi_process.node_part_map.num_pairs) {
    input_line = W_get_input_line (stream);
    TOKENS_PARSE (input_line);
    idx                = TOKEN_INT (1);
    pairs[i].n         = WD_get_node_by_idx (d, idx);
    nt                 = TOKENS_TOTAL () - 1; 
    pairs[i].ranks_dim = nt;
    if (nt > 1) { /* shared node */
      pairs[i].ranks = w_new_array (pairs[i].ranks_dim, W_mpi_rank_t);
      FOR (j, pairs[i].ranks_dim) {
        pairs[i].ranks[j] = TOKEN_INT (2+j);
        if ((pairs[i].ranks[j] < 0) ||
            (pairs[i].ranks[j] >= d->mpi_process.size)) {
          BAD_INPUT ();
        }
      }
    } else {
      pairs[i].ranks = (void *) TOKEN_INT (2); /* store directly the rank */
    }
    if (pairs[i].ranks_dim == 1 &&
        pairs[i].ranks == NULL) { /* own this node */
      if (node_on_proc (&pairs[i], d->mpi_process.rank) &&
          pairs[i].n == NULL) {
        sprintf (W_buf (), "Node %d owned by proc %d not found",
                 (int)idx, d->mpi_process.rank);
        REPORT_FATAL_ERROR (W_buf ());
      }
    }
  }

  /* Initialize the communication buffs */
  d->mpi_process.comm_buffs
    = w_new_array (d->mpi_process.size, W_mpi_comm_buff_t);
  CHECK_MALLOC (d->mpi_process.comm_buffs);
  comm_buffs = d->mpi_process.comm_buffs;

  FOR (i, d->mpi_process.size) {
    comm_buffs[i].rank                = i;
    comm_buffs[i].num_to_send         = 0;
    comm_buffs[i].count               = 0;
    comm_buffs[i].sbuf_dim            = 0;
    comm_buffs[i].sbuf                = NULL;
    comm_buffs[i].rbuf_dim            = 0;
    comm_buffs[i].rbuf                = NULL;
  }

  if (stream != orig_stream) {
    fclose (stream);
    stream = orig_stream;
  }

  input_line = W_get_input_line (stream);
  if (input_line[0] != END_OF_BLOCK)
    REPORT_FATAL_ERROR ("Parallel block does not end in '&'");

  return TRUE;
}

static void
pack_forces  (W_mpi_process_t *p, /* IN: process */
              W_mpi_comm_buff_t *cb
              )
{
  int i, j, nn = cb->num_to_send;
  W_node_t *n, **npa = cb->to_send;
  double *sbp;
  
  sbp       = cb->sbuf;
  FOR (i, nn) {                 /* packing loop */
    n    = npa[i];
    *sbp = n->force_t.x; sbp++;
    *sbp = n->force_t.y; sbp++;
    *sbp = n->force_t.z; sbp++;
  }
}

static void
unpack_forces (W_mpi_process_t *p, /* IN: MPI process */
               W_mpi_comm_buff_t *cb /* IN: communication buffer */
               )
{
  int i, j, nn = cb->num_to_recv;
  W_node_t *n, **npa = cb->to_recv;
  double *rbp;
  
  rbp   = cb->rbuf;
  FOR (i, nn) {                 /* unpacking loop */
    n    = npa[i];
    n->force_t.x += *rbp; rbp++; /* increment force */
    n->force_t.y += *rbp; rbp++;
    n->force_t.z += *rbp; rbp++;
  }
}


static void
pack_masses  (W_mpi_process_t *p, /* IN: process */
              W_mpi_comm_buff_t *cb
              )
{
  int i, j, nn = cb->num_to_send;
  W_node_t *n, **npa = cb->to_send;
  double *sbp;
  
  sbp       = cb->sbuf;
  FOR (i, nn) {                 /* packing loop */
    n    = npa[i];
    *sbp = n->mass; sbp++;
  }
}

static void
unpack_masses (W_mpi_process_t *p, /* IN: MPI process */
               W_mpi_comm_buff_t *cb /* IN: communication buffer */
               )
{
  int i, j, nn = cb->num_to_recv;
  W_node_t *n, **npa = cb->to_recv;
  double *rbp;
  
  rbp   = cb->rbuf;
  FOR (i, nn) {                 /* unpacking loop */
    n    = npa[i];
    n->mass += *rbp; rbp++; /* increment mass */
  }
}

  /* Communicate forces (node-cut version):
     (i) Export force on nodes belonging to neighbors;
     (ii) Import force acting on nodes belonging to you, but shared
     with neighbors, and add it up to your own copy of the node force.

     To communicate, the routine uses the communication buffers
     and the node/processor map.
   */
void
W_mpi_exchange_forces (W_domain_t *d /* domain */
                   )
{
  static int tag = 0;
  W_mpi_exchange (d, 3, pack_forces, unpack_forces, ++tag);
}

  /* Communicate mass (node-cut version only):
     (i) Export mass of nodes belonging to neighbors;
     (ii) Import mass of nodes belonging to you, but shared
     with neighbors, and add it up to your own copy of the node force.

     To communicate, the routine uses the communication comm_buffs
     and the node/processor map.
   */
void
W_mpi_exchange_masses (W_domain_t *d /* domain */
                        )
{
  static int tag = 1303;
  W_mpi_exchange (d, 1, pack_masses, unpack_masses, tag);
}

#endif /* only for MPI WASP and node-cut strategy */
