#include "Wconfig.h"
#include "Wxf.h"
#include "Wdomain.h"
#include "Whistry.h"
#include "Welem.h"
#include "Wtfuncs.h"
#include "Wnode.h"
#include "Wvset.h"
#include "EID.h"

#if defined(MPI_WASP) 
 
#include "Wmpi.h"

extern W_options_t options;

#define TIMED 0
#if TIMED
#include "timeP.h"
#endif

#define PRINT 0
#if PRINT
static char pb[1024];
# define LOGIT { if (!options.no_output) { W_update_log (pb); } }
#endif

static int tag = 0;

static void
def_comm_mapper (W_mpi_process_t *p, int to, int *procranks)
{
  int i, nn;

  p->comm_mapper[to].num_to_send = 0;
  nn = p->node_proc_map.num_nodes;
  FOR (i, nn) {
    if (procranks[i] == p->rank) { /* wishes to receive from me? */
      p->comm_mapper[to].send_flag[i] = TRUE; /* send this node */
      p->comm_mapper[to].num_to_send++;
    }
  }
}

/*
  Build communication mappers.
  Build communication mappers by exchanging information with
  all the peers.
 */
void
W_mpi_build_comm_mappers (W_mpi_process_t *p)
{
  int i, n, nn;
  int *sproc, *rproc;
  W_mpi_comm_mapper_t *mapper = p->comm_mapper;
  W_mpi_node_processor_pair_t *pair;

  if (p->rank < 0 || p->size < 1) {
      REPORT_FATAL_ERROR ("Parallel data not initialized");
  }
  
  sproc = w_new_array (p->node_proc_map.num_nodes, int);
  CHECK_MALLOC (sproc);
  rproc = w_new_array (p->node_proc_map.num_nodes, int);
  CHECK_MALLOC (rproc);

  /* build communication requests list */
  pair = p->node_proc_map.pair;
  nn   = p->node_proc_map.num_nodes;
  FOR (i, nn) {
    if (pair[i].n) { /* if node included in my partition */
      if (pair[i].rank == p->rank) {
        sproc[i] = -1; /* owned by p */
      } else {
        sproc[i] = pair[i].rank; /* should be sending to me */
        mapper[pair[i].rank].should_receive_from = TRUE;
      }
    } else
      sproc[i] = -1; /* not used by p */
  }

  /* exchange communication lists with all processes */
  nn = p->node_proc_map.num_nodes;
  FOR (i, p->size) {
    if (i == p->rank) {         /* sending */
      MPI_Bcast (sproc, nn, MPI_INT, i, MPI_COMM_WORLD);
    } else {                    /* receiving */
      MPI_Bcast (rproc, nn, MPI_INT, i, MPI_COMM_WORLD);
      def_comm_mapper (p, i, rproc);
    }
    MPI_Barrier (MPI_COMM_WORLD); /* rendez-vous */
  }
  
#if PRINT
  FOR (i, p->size) {
    if (p->comm_mapper[i].num_to_send > 0) {
      sprintf (pb, "process %d should send data for %d nodes to %d",
               p->rank, p->comm_mapper[i].num_to_send, i); LOGIT;
    }
    if (p->comm_mapper[i].should_receive_from) {
      sprintf (pb, "process %d should receive data from %d",
               p->rank, i); LOGIT;
    }
  }
#endif

  w_free (sproc);  /* no longer needed */
  w_free (rproc); /* no longer needed */
}

/*  
! MPI SETUP
PARALLEL
num_node_proc_pairs
n0 py
n1 px
...
nN pz
&
*/

W_bool_t
W_mpi_run_reader (W_domain_t *d, FILE *stream)
{
  const char *input_line;
  W_index_t idx;
  int i, j;
  W_mpi_node_processor_pair_t *pair;
  W_mpi_comm_mapper_t *mapper;

  input_line = W_get_input_line (stream);
  TOKENS_PARSE (input_line);

  /* Create process */
  d->mpi_process.domain        = d;
  MPI_Comm_rank (MPI_COMM_WORLD, &d->mpi_process.rank);
  MPI_Comm_size (MPI_COMM_WORLD, &d->mpi_process.size);
  d->mpi_process.node_proc_map.num_nodes = TOKEN_INT (1);
  d->mpi_process.node_proc_map.pair      =
    w_new_array (d->mpi_process.node_proc_map.num_nodes,
                 W_mpi_node_processor_pair_t);
  CHECK_MALLOC (d->mpi_process.node_proc_map.pair);

  /* Set up the node/processor map */
  pair = d->mpi_process.node_proc_map.pair;
  FOR (i, d->mpi_process.node_proc_map.num_nodes) {
    input_line = W_get_input_line (stream);
    TOKENS_PARSE (input_line);
    idx          = TOKEN_INT (1);
    pair[i].n    = WD_get_node_by_idx (d, idx);
    pair[i].rank = TOKEN_INT (2);
    if (pair[i].rank == d->mpi_process.rank) { /* own this node */
      if (!pair[i].n) { /* must have this node! */
        sprintf (W_buf (), "Node %d owned by proc %d not found",
                 (int)idx, d->mpi_process.rank);
        REPORT_FATAL_ERROR (W_buf ());
      }
    }
  }

  /* Initialize the communication mappers */
  d->mpi_process.comm_mapper
    = w_new_array (d->mpi_process.size, W_mpi_comm_mapper_t);
  CHECK_MALLOC (d->mpi_process.comm_mapper);
  mapper = d->mpi_process.comm_mapper;

  FOR (i, d->mpi_process.size) {
    mapper[i].rank = i;
    mapper[i].send_flag
      = w_new_array (d->mpi_process.node_proc_map.num_nodes, W_short_bool_t);
    CHECK_MALLOC (mapper[i].send_flag);
    mapper[i].num_to_send         = 0;
    mapper[i].should_receive_from = FALSE;
    mapper[i].received            = FALSE;
    FOR (j, d->mpi_process.node_proc_map.num_nodes)
      mapper[i].send_flag[j] = FALSE;
    mapper[i].sbuf_dim            = 0;
    mapper[i].sbuf                = NULL;
    mapper[i].rbuf_dim            = 0;
    mapper[i].rbuf                = NULL;
  }

  input_line = W_get_input_line (stream);
  if (input_line[0] != END_OF_BLOCK)
    REPORT_FATAL_ERROR ("Parallel block does not end in '&'");

  return TRUE;
}

/*
  Pack data.
  Pack data into the send buffer of process /to/.
  Reallocate buffer if needed.  Node accelerations are
  packed in the order of the node/ processor map, x-, y-, and z-component.
 */
static void
pack_and_send (W_domain_t *d, /* IN: domain */
               int to /* IN: pack into buffer of this process */
  )
{
  int i, dim, nn;
  W_node_t *n;
  double *sbp;
  W_short_bool_t *send_flag;
  W_mpi_node_processor_pair_t *pair;
  W_mpi_process_t *p = &d->mpi_process;
  W_mpi_comm_mapper_t *mapper = &p->comm_mapper[to];
#if TIMED
  static ckit_timer_t *timer = NULL;
  
  if (!timer) timer = ckit_timer_new ();
  ckit_timer_restart (timer);
#endif
  
  /* reallocate buffer for packing ... */
  dim = mapper->num_to_send * 3;/* sending array 3-D vectors */
  if (mapper->sbuf_dim < dim) {
    mapper->sbuf_dim = dim;
    mapper->sbuf = w_realloc_array (mapper->sbuf, mapper->sbuf_dim, double);
    CHECK_MALLOC (mapper->sbuf);
  }

  /* pack data */
  sbp       = mapper->sbuf;
  send_flag = mapper->send_flag;
  pair      = p->node_proc_map.pair;
  nn        = p->node_proc_map.num_nodes;
  FOR (i, nn) {
    if (send_flag[i]) {
      n    = pair[i].n;
      *sbp = n->force_t.x; sbp++;
      *sbp = n->force_t.y; sbp++;
      *sbp = n->force_t.z; sbp++;
    }
  }

   /* immediate (non-blocking) send */
  MPI_Isend (mapper->sbuf, dim, MPI_DOUBLE,
             to, tag, MPI_COMM_WORLD,
             &mapper->comm_request);

  /* done with sending  */
  mapper->packed = TRUE;
#if TIMED
  sprintf (W_buf(), "send %g", ckit_timer_current (timer));
  W_update_log (W_buf ());
#endif
#if PRINT
  sprintf (pb, "process %d packed and initiated send to %d", p->rank, to); LOGIT;
#endif
}

static void
receive_and_unpack (W_domain_t *d, int from, MPI_Status *status)
{
  int i, c, nn;
  W_node_t *n;
  double *rbp;
  W_mpi_process_t *p = &d->mpi_process;
  W_mpi_comm_mapper_t *mapper = &p->comm_mapper[from];
  W_mpi_node_processor_pair_t *pair = p->node_proc_map.pair;
  MPI_Status stat;
#if TIMED
  static ckit_timer_t *timer = NULL;
  
  if (!timer) timer = ckit_timer_new ();
  ckit_timer_restart (timer);
#endif
  
  /* Is the receive buffer big enough? */
  MPI_Get_count (status, MPI_DOUBLE, &c); 

  if (mapper->rbuf_dim < c) {
    mapper->rbuf_dim = c;
    mapper->rbuf = w_realloc_array (mapper->rbuf, mapper->rbuf_dim, double);
    CHECK_MALLOC (mapper->rbuf);
  }

  /* receive */
  MPI_Recv (mapper->rbuf, c, MPI_DOUBLE, from, tag, MPI_COMM_WORLD, &stat);
  
  /* unpack */
  nn    = p->node_proc_map.num_nodes;
  rbp   = mapper->rbuf;
  FOR (i, nn) {
    if (pair[i].n &&          /* defined for domain and ... */
        pair[i].rank == from  /* ... owned by process /from/ */
      ) {
      n    = pair[i].n;
      n->force_t.x = *rbp; rbp++;
      n->force_t.y = *rbp; rbp++;
      n->force_t.z = *rbp; rbp++;
    }
  }
  
  /* done with receiving */
  mapper->received = TRUE;
#if TIMED
  sprintf (W_buf(), "recv %g", ckit_timer_current (timer));
  W_update_log (W_buf ());
#endif
#if PRINT
  sprintf (pb, "process %d received from %d", p->rank, from); LOGIT;
#endif   
}
static int /* TRUE if message from processor
              /from/ ready; FALSE otherwise */
receive_pending (W_mpi_process_t *p, /* IN: process */
                 int from, /* IN: processor from
                              which receive expected */
                 MPI_Status *status /* OUT: status of message */
  )
{
  int pending;
  MPI_Iprobe (from, tag, MPI_COMM_WORLD, &pending, status);
  return pending;
}

  /* Communicate:
     (i) Export acceleration of nodes belonging to domain,
     (ii) Import acceleration of nodes belonging to neighbors.
     NOTE: The action solve_for leaves the computed accelerations
           in the force_t field of the node data, hence the routines
           in this file access this field to transfer the new
           (computed) accelerations.

     To communicate, the routine uses the communication mappers
     and the node/processor map.
   */
void
W_mpi_communicate (W_domain_t *d /* domain */)
{
  int i, r;
  static int start_r = -1;
  int done;
  MPI_Status status;
  W_mpi_process_t *p = &d->mpi_process;
  W_mpi_comm_mapper_t *mapper = p->comm_mapper;
  int size = p->size;
  int num_receive, tot_receive;
  int ntries;
#define RECEIVE_NOT_DONE(MAPPER) \
  ((MAPPER).should_receive_from && (!(MAPPER).received))
#define SEND_NOT_DONE(MAPPER) \
  ((MAPPER).num_to_send > 0 && (!(MAPPER).sent))
#define PACKING_NOT_DONE(MAPPER) \
  ((MAPPER).num_to_send > 0 && (!(MAPPER).packed))
#if TIMED
  static ckit_timer_t *timer = NULL;
  if (!timer) timer = ckit_timer_new ();
  ckit_timer_restart (timer);
#endif

  tag++; /* the tag is incremented for each time step; hence, any
            lack of synchronization would be displayed by a program hang
          */
  
  /* clear flags */
  tot_receive = 0;
  FOR (i, size) {
    mapper[i].received = mapper[i].packed = mapper[i].sent = FALSE;
    if (mapper[i].should_receive_from) tot_receive++;
  }

  /* initialize sends */
  if (start_r < 0) start_r = p->rank;
  FOR (i, size) {
    r = (start_r + i) % size; /* produces permutations to enhance fairness */
    if (r != p->rank && mapper[r].num_to_send > 0)
      pack_and_send (d, r);
  }
  start_r++;

  num_receive = 0;
  ntries = 0;
  do {
    done = TRUE; /* assume we're done */
    FOR (i, p->size) {
      if (i != p->rank) { /* no communication with self */
        if (RECEIVE_NOT_DONE (mapper[i])) {
          if (receive_pending (p, i, &status)) {
            receive_and_unpack (d, i, &status); num_receive++;
          } 
          ntries++;
          done = FALSE; /* allow for additional loop */
        }
      } /* if (i != p->rank) */
    } /* FOR */
  } while (!done);
#if TIMED
  sprintf (W_buf(), "comm %g (%d tries)",
           ckit_timer_current (timer), ntries);
  W_update_log (W_buf ());
#endif 
  if (num_receive != tot_receive) REPORT_FATAL_ERROR ("Failed to receive");
  MPI_Barrier (MPI_COMM_WORLD);
}

#endif /* defined(MPI_WASP) */
