#include "Wconfig.h"
#include "Wxf.h"
#include "Wnewmarki.h"
#include "Woptions.h"
#include "Wnode.h"
#include "Whistry.h"
#include "Wutil.h"
#include "Wnewmfend.h"
#include "Wnewmdflt.h"
#include "fles.h"

/* RCSHEADER "$Id: newmdflt.c,v 1.5 2004/04/02 19:39:59 cvs Exp $" */

#define DBGPRINT(stuff) 

#define SOLV_SPETSC    0
#define SOLV_DMESCHACH 1
#define SOLV_SMESCHACH 2
#define SOLV_GEP       3

#define DELEGATE_COEFFS_TO_CLIENT -1.0

typedef struct history_client_data_t {
  double                            residual_norm_0;
  double                            residual_norm;
  double                            displ_incr_norm_0;
  double                            displ_incr_norm;
  double                            extf_norm_0;
  double                            inertf_norm_0;
  double                            extf_norm;
  double                            inertf_norm;
  int                               niter;
  W_newmark_predictor_enum          pred;
  double                            resid_work;
  double                            resid_work_incr;
  double                            resid_work_bound;
  double                            resid_work_incr_bound;
}              history_client_data_t;

typedef enum nonlinear_solver_type_enum {
  NEWTON_SOLVER = 0,
  NCG_SOLVER    = 1
}            nonlinear_solver_type_enum;

struct nonlinear_solver_t;

typedef struct newton_solver_t {
  struct nonlinear_solver_t     *nles;
  W_bool_t                       full_newton;
  int                            recomp_keff_after;
  int                            modnewton_after;
  W_bool_t                       initial_stiff;
  fles_solver_ctx_t              slesctx;
}              newton_solver_t;

typedef enum {
  NCG_PRECOND_IDENTITY = 1,
  NCG_PRECOND_SCALED_MASS = 2
}    ncg_precond_type_enum;

typedef enum {
  NCG_FLETCHER_REEVES     = 1,
  NCG_POLAK_RIBIERE       = 2
}    ncg_variant_enum;

typedef struct ncg_solver_t {
  struct nonlinear_solver_t     *nles;
  ncg_variant_enum               variant;
  double                         u_perturbation_mag;
  double                         eps;
  double                        *p, *z, *dR, *prev_z;
  ncg_precond_type_enum          precond_type;
  W_bool_t                       precond_is_set_up;
  fles_solver_ctx_t              precond;
}              ncg_solver_t;

/*
  Different nonlinear solvers can be easily accomodated by
  (i) creating a type, (ii) listing the solver in the two switch statements
  in this file (to create the solver and to call it), and (iii) implementing
  the solver itself.
 */
typedef struct nonlinear_solver_t {
  struct W_newmark_integrator_implem_t  *ii;
  nonlinear_solver_type_enum             type;
  fles_tot_of_equations_callback         tot_of_equations_callback;
  newton_solver_t                        newton;
  ncg_solver_t                           ncg;
  /* list additional solver here */
}              nonlinear_solver_t;

typedef struct W_newmark_integrator_implem_t {
  /* client and client callbacks */
  W_newmark_client_t                            client;/* client of the integrator implementation */
  W_newmark_finalize_step_client_proc           finalize_step;
  W_newmark_assemble_eff_stiff_client_proc      assemble_eff_stiff;
  W_newmark_assemble_residual_force_client_proc residual_force;
  W_newmark_start_step_client_proc              start_step;
  W_newmark_update_config_client_proc           update_config;
  /* configuration / options */
  double                     conv_ratio;
  double                     conv_res_mag;
  int                        max_iter;
  W_bool_t                   report_iter;
  W_bool_t                   pass_on_max;
  W_bool_t                   conv_in_equil;
  W_bool_t                   conv_in_displ;
  /* internal variables */
  W_domain_t                *domain; /* domain provides some services */
  W_float_t                  t;
  W_float_t                  dt;
  int                        step;
  int                        iter;
  double                     gamma;
  double                     beta;
  W_newmark_predictor_enum   predictor;
  W_newmark_predictor_enum   curr_predictor;
  nonlinear_solver_t         nles;
  int                        num_eqns;
  double                    *Delta_u;
  double                    *R;
  double                    *prev_R;
  int                       *eqnnums;
  double                     residual_norm_0, residual_norm;
  double                     displ_incr_norm_0, displ_incr_norm;
  double                     extf_norm_0, inertf_norm_0, extf_norm, residf_norm, inertf_norm;
  double                     resid_work, resid_work_incr, resid_work_bound, resid_work_incr_bound;
  /* timers */
  W_timer_t                 *ass_rhs_timer; 
  W_timer_t                 *ass_lhs_timer; 
  W_timer_t                 *solve_timer; 
  W_timer_t                 *updt_timer;
  W_timer_t                 *start_timer;
  /* history */
  history_client_data_t      hist_buf;
}              W_newmark_integrator_implem_t;

static newton_solver_t *
newton_hndl (struct W_newmark_integrator_implem_t *i)
{
  if (i->nles.type == NEWTON_SOLVER) return &i->nles.newton;
  else                               return NULL;
}
static ncg_solver_t *
ncg_hndl (struct W_newmark_integrator_implem_t *i)
{
  if (i->nles.type == NCG_SOLVER) return &i->nles.ncg;
  else                            return NULL;
}
#  define NEWTON_HNDL(i)  newton_hndl (i)
#  define NCG_HNDL(i)     ncg_hndl (i)

static double
vdot (double *a, double *b, int n)
{
  int i;
  double r = 0;
  FOR (i, n) { r += a[i] * b[i]; }
  return r;
}

static void
vzero (double *c, int n)
{
  int i;
  FOR (i, n) { c[i] = 0; }
}

#define VOP1(dest, op1, a, numel)               \
{                                               \
  int _j__, _N__ = numel;                       \
  FOR (_j__, _N__) {                            \
    dest[_j__] op1 a[_j__];                     \
  }                                             \
}
#define VOP2(dest, op1, a, op2, b, numel)       \
{                                               \
  int _j__, _N__ = numel;                       \
  FOR (_j__, _N__) {                            \
    dest[_j__] op1 a[_j__] op2 b[_j__];         \
  }                                             \
}

static void
vcomb (double *c, double alpha, double *a, double beta, double *b, int n)
{
  VOP2 (c, =, alpha * a, +, beta * b, n);
}

static double
vnorm (double *fr, int n)
{
  return sqrt (vdot (fr, fr, n));
}

static double
ventrymaxabs (double *c, int n)
{
  int i;
  double m = 0;
  FOR (i, n) { m = max (fabs (c[i]), m); }
  return m;
}

#define  VPRINT(array, n)                               \
{                                                       \
  int _i_;                                              \
  printf ("%s, l%d: %s\n", __FILE__, __LINE__, #array); \
  FOR (_i_, n) {                                        \
    printf ("%g ", array[_i_]);                         \
  }                                                     \
  printf ("\n");                                        \
}

static int
error_notify (fles_error_enum error_num,
              int line,
              char *file,
              char *detail,
              void *iip)
{
  char buf[1024];
  W_newmark_integrator_implem_t *ii = (W_newmark_integrator_implem_t *)iip;
  W_domain_t *d = ii->domain;
  sprintf (buf, "%s (l.%d), %s", file, line, detail);
  REPORT_FATAL_ERROR (d->errh, buf);
  return 0;
}

static void
save_hist (W_newmark_integrator_implem_t *i)
{
  i->hist_buf.niter                        = i->iter;
  i->hist_buf.pred                         = i->curr_predictor;
  i->hist_buf.residual_norm_0              = i->residual_norm_0;
  i->hist_buf.residual_norm                = i->residual_norm;
  i->hist_buf.displ_incr_norm_0            = i->displ_incr_norm_0;
  i->hist_buf.displ_incr_norm              = i->displ_incr_norm;
  i->hist_buf.inertf_norm_0                = i->inertf_norm_0;
  i->hist_buf.extf_norm_0                  = i->extf_norm_0;
  i->hist_buf.inertf_norm                  = i->inertf_norm;
  i->hist_buf.extf_norm                    = i->extf_norm;
  i->hist_buf.resid_work                   = i->resid_work;
  i->hist_buf.resid_work_bound             = i->resid_work_bound;
  i->hist_buf.resid_work_incr              = i->resid_work_incr;
  i->hist_buf.resid_work_incr_bound        = i->resid_work_incr_bound;
}

static int
converged (W_newmark_integrator_implem_t *i)
{
  if (i->iter == 1) {
    i->residual_norm_0           = i->residual_norm;
    i->displ_incr_norm_0         = i->displ_incr_norm;
  }
  
  if (i->report_iter) {
    double rrat = (i->residual_norm_0 > 0
                   ? i->residual_norm / i->residual_norm_0
                   : (i->residual_norm == 0
                      ? 1
                      : FLT_MAX
                      ));
    double drat = (i->displ_incr_norm_0 > 0
                   ? i->displ_incr_norm / i->displ_incr_norm_0
                   : (i->displ_incr_norm == 0
                      ? 1
                      : FLT_MAX
                      ));

    if (i->conv_in_displ)
      sprintf (W_buf (), "   %d: |du|/|du0|=%g",
               i->iter, drat);
    else {
      if (i->conv_res_mag > 0)
        sprintf (W_buf (), "   %d: |R|=%g", i->iter, i->residual_norm);
      else
        sprintf (W_buf (), "   %d: |R|/|R0|=%g", i->iter, rrat);
    }

    W_log_update (i->domain->logh, W_buf ());
  }
  
  if (i->iter >= i->max_iter) {
    if (i->pass_on_max) return TRUE;
    else {
      REPORT_FATAL_ERROR (i->domain->errh, "Divergent in equilibrium iteration");
      return FALSE;
    }
  } else {
    if (i->conv_in_displ)
      return ((i->displ_incr_norm == 0 && i->displ_incr_norm_0 == 0) ||
              i->displ_incr_norm / i->displ_incr_norm_0 <= i->conv_ratio
              );
    else
      if (i->conv_res_mag > 0)
        return (i->residual_norm <= i->conv_res_mag);
      else
        return ((i->residual_norm == 0 && i->residual_norm_0 == 0) ||
                i->residual_norm / i->residual_norm_0 <= i->conv_ratio
                );
  }
}

static W_bool_t 
should_reform_k_eff (newton_solver_t *newton)
{
  int first_n = newton->modnewton_after;
  int fn1 = first_n - 1;
  int ra = newton->recomp_keff_after;
  int iter = newton->nles->ii->iter;
  int step = newton->nles->ii->step;
  
  /* for full Newton */
  if (newton->full_newton) {
    return TRUE; 
  }

  /* initial stiffness iteration */
  if (newton->initial_stiff) {
    if (iter < 2) {
      if (step == 1) {
        return TRUE; /* first step */
      }
      if (step / ra * ra == step) {
        return TRUE; /* should reform in this step */
      }
    }
  } else {
    /* for modified Newton */
    if (iter < fn1 ||           /* first time around */
        ((iter >= fn1) &&         /* first n-iterations with full Newton */
         (iter  - fn1) / ra * ra == (iter - fn1)) /* every m iterations should reform */ 
        ) {
      return TRUE;
    }
  }
  
  /* none of the above; no need to reform */
  return FALSE;
}

static int 
get_incr (struct W_newmark_integrator_t *i,
          int eqns[], /* for equation numbers */
          double values[], /* return values in this array */
          int dim /* dimension of preceding arrays */
          )
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  if (dim > 1) {
    double *Delta_u = ii->Delta_u;
    int j;
    FOR (j, dim) {
      values[j] = Delta_u[eqns[j]];
    }
    return dim;
  } else {
    values[0] = ii->Delta_u[eqns[0]];
    return 1;
  }
}

static void
assemble_rhs_callback (struct W_newmark_integrator_t *i, /* Newmark integrator */
                       fles_update_mode_enum update_mode, /* update mode */
                       int nr,/* number of rows in vector to be assembled */
                       int *ri,/* array of equations numbers */
                       double *values /* vector to be assembled */
                       )
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  double *R = ii->R;
  int k;

  FOR (k, nr) { 
    R[ri[k]] += values[k];
  }
}

static void
assemble_R (W_newmark_integrator_implem_t *ii,
            W_newmark_get_incr_proc get_incr,
            W_newmark_predictor_enum predictor,
            W_mat_state_update_form_enum update_form
            )
{
  W_timer_restart (ii->ass_rhs_timer);

  vzero (ii->R, ii->num_eqns);
  ii->residual_force (ii->client, assemble_rhs_callback,
                     update_form, get_incr, predictor,
                     &ii->extf_norm, &ii->inertf_norm, &ii->residf_norm);
  if (ii->iter == 1) {
    ii->extf_norm_0 = ii->extf_norm;
    ii->inertf_norm_0 = ii->inertf_norm;
  }

  W_timer_accumulate (ii->ass_rhs_timer);
}

/*
  Assemble force residual for the Newton algorithm.
  Use the generic assemble_R() + zeroing of context rhs,
  assembly into context rhs, and finishing of context rhs.
 */
static void
assemble_R_newton (newton_solver_t *newton,
                   W_newmark_get_incr_proc get_incr,
                   W_newmark_predictor_enum predictor,
                   W_mat_state_update_form_enum update_form
                   )
{
  W_newmark_integrator_implem_t *i = newton->nles->ii;

  newton->slesctx->zero_rhs (newton->slesctx);

  assemble_R (i, get_incr, predictor, update_form);

  newton->slesctx->assemble_vect (newton->slesctx, FLES_ADD,
                                  i->num_eqns, i->eqnnums, i->R);
  newton->slesctx->finish_rhs (newton->slesctx);

}

/*
  First update material state of all quadrature points in the domain
  by assembling the residual for the last time (final assembly, with
  update_form == W_MAT_STATE_UPDATE_FINAL).
  Then update configuration of nodes.
 */
static void
update_config (W_newmark_integrator_implem_t *ii)
{
  W_timer_restart (ii->updt_timer);
  /* update material state */
  assemble_R (ii, get_incr, ii->curr_predictor, W_MAT_STATE_UPDATE_FINAL);
  /* update geometric configuration */
  ii->update_config (ii->client, get_incr, ii->curr_predictor);
  W_timer_accumulate (ii->updt_timer);
}

static int 
dummy_get_incr (struct W_newmark_integrator_t *i,
          int eqns[], /* for equation numbers */
          double values[], /* return values in this array */
          int dim /* dimension of preceding arrays */
          )
{
  if (dim > 1) {
    int j;
    FOR (j, dim) {
      values[j] = 0;
    }
    return dim;
  } else {
    values[0] = 0;
    return 1;
  }
}

static W_newmark_predictor_enum
optimal_predictor (W_newmark_integrator_implem_t *i)
{
#define MAXPRED 5
  double residnorm[MAXPRED] = {0,0,0,0,0};
  double efnorm, ifnorm;
  W_newmark_predictor_enum pred[MAXPRED] = {
    W_NEWMARK_PRED_SAME_U,
    W_NEWMARK_PRED_SAME_V,
    W_NEWMARK_PRED_SAME_A,
    W_NEWMARK_PRED_ZERO_A,
    W_NEWMARK_PRED_PREV_INCR
  };
  int j, minidx;
  FOR (j, MAXPRED) {
    assemble_R (i, dummy_get_incr, pred[j], W_MAT_STATE_UPDATE_TEMP);
    residnorm[j] = i->residf_norm;
  }
  minidx = 0;
  FOR (j, MAXPRED) {
    if (residnorm[j] <= residnorm[minidx]) {
      minidx = j;
    }
  }
  return pred[minidx];
}

static W_newmark_predictor_enum
choose_predictor (W_newmark_integrator_implem_t *i)
{
  if (i->predictor == W_NEWMARK_PRED_ADAPTIVE)
    return optimal_predictor (i);
  else
    return i->predictor;
}

static void
start_step (W_newmark_integrator_implem_t *i)
{
  W_timer_restart (i->start_timer);

  i->step++;
  i->iter = 0;
  i->t += i->dt;

  vzero (i->Delta_u, i->num_eqns);

  i->start_step (i->client);

  i->curr_predictor = choose_predictor (i);

  W_timer_accumulate (i->start_timer);
}

static void
give_me_disp_incr (struct W_newmark_integrator_t *i, int nr, int *ri, double *du)
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  int j, k;
  double *fr = ii->R;
  double *pfr = ii->prev_R;
  double du_norm = 0;
  double fr_norm = 0;
  FOR (j, nr) {
    k = ri[j];
    du_norm += du[j] * du[j];
    fr_norm += ((fr[k] + pfr[k]) / 2) * ((fr[k] + pfr[k]) / 2);
    ii->resid_work_incr += (fr[k] + pfr[k]) / 2 * du[j];
  }
  ii->resid_work_incr_bound += sqrt (du_norm * fr_norm);
}

static void
finalize_step (W_newmark_integrator_implem_t *ii)
{
  ii->resid_work_incr = 0;
  ii->resid_work_incr_bound = 0;
  
  if (ii->finalize_step != NULL)
    ii->dt = ii->finalize_step (ii->client, give_me_disp_incr);

  ii->resid_work += ii->resid_work_incr;
  ii->resid_work_bound += ii->resid_work_incr_bound;

  VOP1 (ii->prev_R, =, ii->R, ii->num_eqns);
  
  save_hist (ii);
}

static int 
get_perturbed_incr (struct W_newmark_integrator_t *i,
                    int eqns[], /* for equation numbers */
                    double values[], /* return values in this array */
                    int dim /* dimension of preceding arrays */
                    )
{
  int k, j;
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  ncg_solver_t *ncg = NCG_HNDL (ii);
  double *Delta_u = ii->Delta_u;
  double *p = ncg->p;
  double eps = ncg->eps;

  FOR (j, dim) {
    k = eqns[j];
    values[j] = Delta_u[k] + eps * p[k];
  }
  return dim;
}
/*
  Assemble submatrix of the LHS.
  Assemble submatrix of the LHS 
 */
static void
ncg_assemble_lhs_callback (struct W_newmark_integrator_t *i, /* Newmark integrator */
                           fles_update_mode_enum update_mode, /* update mode;  */
                           int nr, /* number of rows and columns in square matrix to be assembled */
                           int *ri, /* array of equations numbers, which apply to both rows and columns
                                       of the matrix to be assembled */
                           double *values, /* square matrix to be assembled, passed
                                              as a linear array of nr*nr values */
                           W_bool_t *constrained /* array of nr elements; TRUE if degree-of-freedom
                                                    constrained, FALSE otherwise */
                           )
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  ncg_solver_t *ncg = NCG_HNDL (ii);
  int j, k, l;
  W_bool_t need_to_rewrite = FALSE; /* do I need to modify the element matrix
                                       because of some constraints?
                                     */
  W_bool_t is_diagonal = TRUE;  /* if the global matrix is diagonal,
                                   we could save a lot of storage by
                                   assembling *only* the diagonal terms
                                 */

  FOR (j, nr) { if (constrained[j]) { need_to_rewrite = TRUE; break; } }

  if (need_to_rewrite) {
    l = 0;
    FOR (j, nr) {
      FOR (k, nr) {
        if (j != k && (constrained[j] || constrained[k])) {
          values[l] = 0;
        }
        l++;
      }
    }
  }

  /* figure out if the matrix is diagonal */
  l = 0;
  FOR (j, nr) {
    FOR (k, nr) {
      if (j != k) {
        if (values[l] != 0) {
          is_diagonal = FALSE;
          goto now_i_know;
        }
      }
      l++;
    }
  }

  /* now we know what form may be assembled */
 now_i_know:
  if (is_diagonal) { /* assemble diagonal terms only */
    l = 0;
    FOR (j, nr) {
      FOR (k, nr) {
        if (j == k) {
          ncg->precond->assemble_mat (ncg->precond, FLES_FALSE,
                                      update_mode, 1, &ri[j], 1, &ri[j], &values[l]);
        }
        l++;
      }
    }
  } else { /* general assembly */
    ncg->precond->assemble_mat (ncg->precond, FLES_FALSE, update_mode,
                                nr, ri, nr, ri, values);
  }
}

static void
ncg_set_up_scaled_mass_precond (ncg_solver_t *ncg)
{
  W_newmark_integrator_implem_t *ii = ncg->nles->ii;
  ncg->precond->zero_lhs (ncg->precond);
  ii->assemble_eff_stiff (ii->client, ncg_assemble_lhs_callback,
                          get_incr, ii->curr_predictor, 
                          0, 1.0 / (ii->beta * ii->dt * ii->dt));
  ncg->precond->finish_lhs (ncg->precond, FLES_FINAL_FINISH);
  ncg->precond_is_set_up = TRUE;
}

static void
ncg_precond_solve (ncg_solver_t *ncg)
{
  W_newmark_integrator_implem_t *ii = ncg->nles->ii;
  fles_solver_ctx_t precond = ncg->precond;
  fles_get_from_sol_func get_from_sol = precond->get_from_sol;
  int j, num_eqns = ncg->nles->ii->num_eqns;
  double *R = ii->R;
  double *z = ncg->z;

  precond->zero_rhs (precond);
  precond->assemble_vect (precond, FLES_ADD, num_eqns, ii->eqnnums, R);
  precond->finish_rhs (precond);
  
  if (!precond->solve (precond))
    REPORT_FATAL_ERROR (ii->domain->errh, "Failed in ncg_precond_solve()");

  FOR (j, num_eqns) {
    z[j] = get_from_sol (precond, j);
  }
}

static void
ncg_precond (ncg_solver_t *ncg)
{
  W_newmark_integrator_implem_t *ii = ncg->nles->ii;
  switch (ncg->precond_type) {
  case NCG_PRECOND_SCALED_MASS:
    if (!ncg->precond_is_set_up) ncg_set_up_scaled_mass_precond (ncg);
    ncg_precond_solve (ncg);
    break;
  case NCG_PRECOND_IDENTITY:
  default:
    VOP1 (ncg->z, =, ii->R, ii->num_eqns); /* trivial preconditioner */
    break;
  } 
}

static void
compute_dR (ncg_solver_t *ncg)
{
  W_newmark_integrator_implem_t *i = ncg->nles->ii;
  double *R = i->R;
  double maxp = 0;
  VOP1 (ncg->dR, =, R, i->num_eqns);
  maxp = ventrymaxabs (ncg->p, i->num_eqns);
  ncg->eps = ncg->u_perturbation_mag / maxp;/* should compute u_perturbation_mag from the actual solution */
  assemble_R (i, get_perturbed_incr, i->curr_predictor, W_MAT_STATE_UPDATE_TEMP);
  vcomb (ncg->dR, -1.0 / ncg->eps, ncg->dR, 1.0 / ncg->eps, R, i->num_eqns);
}

/*
  Solve R(u) = 0 by nonlinear preconditioned CG
  =============================================

  R_0 = R(u_0)
  k = 0
  while (||R_k|| > tol) {
     Solve M(z_k) = R_k
     k++
     if (k == 1) {
        p_1 = z_0
     } else {
        eta = (R_k-1 . z_k-1) / (R_k-2 . z_k-2)
        p_k  = z_k-1 + eta p_k-1
     }
     lambda = - (R_k-1 . z_k-1) / (p_k . delta R(u_k-1, p_k))
     u_k = u_k-1 + lambda p_k
     R_k = R (u_k)
  }

  Here delta R(u_k-1, p_k) = [ \partial R / \partial u @ u_k-1 ] . p_k
  Approximation of delta R(u_k-1, p_k):
  delta R(u_k-1, p_k) ~ R(u_k-1+p_k) - R(u_k-1)
 */
static void
ncg_solve_fletcher_reeves (ncg_solver_t *ncg)
{
  W_newmark_integrator_implem_t *i = ncg->nles->ii;
  int num_eqns = i->num_eqns;
  double pdr;
  double eta, lambda, ip_1_back, ip_2_back = 0;
  double *p = ncg->p;
  double *z = ncg->z;
  double *dR = ncg->dR;
  double *R = i->R;

  i->iter = 0;
  assemble_R (i, get_incr, i->curr_predictor, W_MAT_STATE_UPDATE_TEMP);
  do {
    i->iter++;
    ncg_precond (ncg);
    ip_1_back = vdot (R, z, num_eqns);
    if (i->iter == 1) {
      VOP1 (p, =, z, num_eqns);
    } else {
      eta = ip_1_back / ip_2_back;
      vcomb (p, 1.0, z, eta, p, num_eqns);
    }
    ip_2_back = ip_1_back;
    compute_dR (ncg);
    pdr = vdot (p, dR, num_eqns);
    lambda = - vdot (z, R, num_eqns) / pdr;
    vcomb (i->Delta_u, 1.0, i->Delta_u, lambda, p, num_eqns);
    assemble_R (i, get_incr, i->curr_predictor, W_MAT_STATE_UPDATE_TEMP);
    i->residual_norm = vnorm (R, num_eqns);
    i->displ_incr_norm = lambda * vnorm (p, num_eqns);
  } while (!converged (i));
}

static void
ncg_solve_polak_ribiere (ncg_solver_t *ncg)
{
  W_newmark_integrator_implem_t *i = ncg->nles->ii;
  int num_eqns = i->num_eqns;
  double pdr;
  double *p = ncg->p;
  double *z = ncg->z;
  double *R = i->R;
  double *dR = ncg->dR;
  double *prev_z = ncg->prev_z;
  double eta, lambda, Rz, Rz_1_back = 0;
  double *Delta_u = i->Delta_u;

  i->iter = 0;
  assemble_R (i, get_incr, i->curr_predictor, W_MAT_STATE_UPDATE_TEMP);
  do {
    i->iter++;
    ncg_precond (ncg);
    Rz = vdot (R, z, num_eqns);
    if (i->iter == 1) {
      VOP1 (p, =, -z, num_eqns);
    } else {
      eta = (Rz - vdot (R, prev_z, num_eqns)) / Rz_1_back;
      VOP2 (p, =, -z, +, eta * p, num_eqns);
    }
    Rz_1_back = Rz;
    compute_dR (ncg);
    pdr = vdot (p, dR, num_eqns);
    lambda = Rz / pdr;
    VOP1 (Delta_u, +=, lambda * p, num_eqns);
    assemble_R (i, get_incr, i->curr_predictor, W_MAT_STATE_UPDATE_TEMP);
    i->residual_norm = vnorm (R, num_eqns);
    i->displ_incr_norm = lambda * vnorm (p, num_eqns);
    VOP1 (prev_z, =, z, num_eqns);
  } while (!converged (i));
}

static void
ncg_solve (ncg_solver_t *ncg)
{
  switch (ncg->variant) {
  case NCG_FLETCHER_REEVES:
    ncg_solve_fletcher_reeves (ncg);
    break;
  case NCG_POLAK_RIBIERE:
    ncg_solve_polak_ribiere (ncg);
    break;
  }
}

/*
  Assemble submatrix of the LHS.
  Assemble submatrix of the LHS 
 */
static void
newton_assemble_lhs_callback (struct W_newmark_integrator_t *i, /* Newmark integrator */
                        fles_update_mode_enum update_mode, /* update mode;  */
                        int nr, /* number of rows and columns in square matrix to be assembled */
                        int *ri, /* array of equations numbers, which apply to both rows and columns
                                  of the matrix to be assembled */
                        double *values, /* square matrix to be assembled, passed
                                           as a linear array of nr*nr values */
                        W_bool_t *constrained /* array of nr elements; TRUE if degree-of-freedom
                                               constrained, FALSE otherwise */
                        )
{
  W_newmark_integrator_implem_t *ii = newmark_implem (i);
  newton_solver_t *newton = NEWTON_HNDL (ii);
  int j, k, l;
  W_bool_t need_to_rewrite = FALSE; /* do I need to modify the element matrix
                                       because of some constraints?
                                     */

  FOR (j, nr) { if (constrained[j]) { need_to_rewrite = TRUE; break; } }

  if (need_to_rewrite) {
    l = 0;
    FOR (j, nr) {
      FOR (k, nr) {
        if (j != k && (constrained[j] || constrained[k])) {
          values[l] = 0;
        }
        l++;
      }
    }
  }
  
  newton->slesctx->assemble_mat (newton->slesctx,
                               FLES_FALSE, update_mode, nr, ri, nr, ri, values);
}

static void
assemble_k_eff (newton_solver_t *newton)
{
  W_newmark_integrator_implem_t *ii = newton->nles->ii;
  if (should_reform_k_eff (newton)) {
    if (ii->report_iter) {
      sprintf (W_buf (), "   Reforming stiffness");
      W_log_update (ii->domain->logh, W_buf ());
    }
    W_timer_restart (ii->ass_lhs_timer);
    newton->slesctx->zero_lhs (newton->slesctx);
    ii->assemble_eff_stiff (ii->client, newton_assemble_lhs_callback,
                            get_incr, ii->curr_predictor, 
                            DELEGATE_COEFFS_TO_CLIENT, DELEGATE_COEFFS_TO_CLIENT);
    newton->slesctx->finish_lhs (newton->slesctx, FLES_FINAL_FINISH);
    W_timer_accumulate (ii->ass_lhs_timer);
  }
}

static void
accumulate_Delta_u (newton_solver_t *newton)
{
  int j, num_eqns = newton->nles->ii->num_eqns;
  fles_solver_ctx_t slesctx = newton->slesctx;
  fles_get_from_sol_func get_from_sol = slesctx->get_from_sol;
  double *Delta_u = newton->nles->ii->Delta_u;
  FOR (j, num_eqns) {
    Delta_u[j] += get_from_sol (slesctx, j);
  }
}

static void
solve_for_incr_u_corr (newton_solver_t *newton)
{
  W_newmark_integrator_implem_t *i = newton->nles->ii;
  fles_solver_ctx_t slesctx = newton->slesctx;
  W_timer_restart (i->solve_timer);

  if (!slesctx->solve (slesctx))
    REPORT_FATAL_ERROR (i->domain->errh, "Failed in solve_for_incr_u_corr()");

  i->residual_norm   = slesctx->rhs_norm (slesctx, FLES_NORM_2);
  i->displ_incr_norm = slesctx->sol_norm (slesctx, FLES_NORM_2);
  
  accumulate_Delta_u (newton);
    
  W_timer_accumulate (i->solve_timer);
}

static void
newton_solve (newton_solver_t *newton)
{
  W_newmark_integrator_implem_t *i = newton->nles->ii;
  do {
    i->iter++;
    assemble_R_newton (newton, get_incr, i->curr_predictor, W_MAT_STATE_UPDATE_TEMP);
    assemble_k_eff (newton);
    solve_for_incr_u_corr (newton);
  } while (!converged (i));
}

static void
solve_for_u_corr (W_newmark_integrator_implem_t *i)
{
  switch (i->nles.type) {
  /* list additional solver(s) here */
  case NCG_SOLVER:
    ncg_solve (NCG_HNDL (i));
    break;
  case NEWTON_SOLVER:
  default:
    newton_solve (NEWTON_HNDL (i));
    break;
  }
}

static int
tot_of_equations (void *iip)
{
  W_newmark_integrator_implem_t *ii = (W_newmark_integrator_implem_t *)iip;
  return ii->nles.tot_of_equations_callback (ii->client);
}

static void
config_ncg (W_newmark_integrator_implem_t *ii,
            char *optstring /* options string; specify option+value pairs */
            )
{  
  ncg_solver_t *ncg = NCG_HNDL (ii);
  ncg_precond_type_enum precond_type = NCG_PRECOND_IDENTITY;
  double nz_per_row = 0.01;
  int solver = SOLV_GEP;
  ncg_variant_enum variant = NCG_POLAK_RIBIERE;

  ncg->nles = &ii->nles;

  if (optstring) {
    int pos;
    tokens_parser parser = tokens_new_parser ();
    if (tokens_parse_line (parser, optstring)) {
      if ((pos = tokens_have_keyword (parser, 1, "IDENTITY_PRECOND")) > 0) {
        precond_type  = NCG_PRECOND_IDENTITY;
      }
      if ((pos = tokens_have_keyword (parser, 1, "MASS_PRECOND")) > 0) {
        precond_type  = NCG_PRECOND_SCALED_MASS;
      }
      if ((pos = tokens_have_keyword (parser, 1, "FLETCHER_REEVES")) > 0) {
        variant = NCG_FLETCHER_REEVES;
      }
      if ((pos = tokens_have_keyword (parser, 1, "POLAK_RIBIERE")) > 0) {
        variant = NCG_POLAK_RIBIERE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "NZ_PER_ROW")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          nz_per_row = tokens_token_as_double (parser, pos+1);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "SOLVER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          const char *solvname;
          solvname = tokens_token_as_string (parser, pos+1);
          if        (tokens_token_and_keyword_equiv (solvname, "SMESCHACH")) {
            solver = SOLV_SMESCHACH;
          } else if (tokens_token_and_keyword_equiv (solvname, "DMESCHACH")) {
            solver = SOLV_DMESCHACH;
          } else if (tokens_token_and_keyword_equiv (solvname, "SPETSC")) {
            solver = SOLV_SPETSC;
          } else if (tokens_token_and_keyword_equiv (solvname, "GEP")) {
            solver = SOLV_GEP;
          }
        }
      }
    }
    tokens_delete_parser (parser);
  }

  ncg->variant = variant;
  ncg->u_perturbation_mag = 1e-6;
  ncg->p  = w_new_array (ii->num_eqns, double); CHECK_MALLOC (ncg->p);
  ncg->z  = w_new_array (ii->num_eqns, double); CHECK_MALLOC (ncg->z);
  ncg->dR = w_new_array (ii->num_eqns, double); CHECK_MALLOC (ncg->dR);
  ncg->prev_z = w_new_array (ii->num_eqns, double); CHECK_MALLOC (ncg->prev_z);

  ncg->precond_type       = precond_type;
  ncg->precond_is_set_up  = FALSE;
  
  switch (ncg->precond_type) {
  case NCG_PRECOND_IDENTITY:
    ncg->precond = NULL;
    ncg->precond_is_set_up = TRUE;
    break;
  case NCG_PRECOND_SCALED_MASS:
    {
      /* The tot_of_equations callback is a little bit tricky here.
         First note that the client of the FLES library is the implementation /ii/.
         Hence, the FLES library needs a tot_of_equations() callback to call
         on the implementation, which in turns calls the tot_of_equations_callback() defined
         on *its* client, ii->client.
      */
      switch (solver) {
      case SOLV_SMESCHACH:
        {
          int method = FLES_SMESCHACH_SPCG;
          ncg->precond
            = fles_new_sparse_meschach_solver (ii, tot_of_equations, error_notify);
          ncg->precond->set_option (ncg->precond, FLES_SMESCHACH_METHOD, (void *)method);
        } 
        break;
#     if   SOLVER_PETSC
      case SOLV_SPETSC:
        ncg->precond = fles_new_spetsc_solver (ii, tot_of_equations, error_notify);
        ncg->precond->set_option (ncg->precond, PETSC_OPTS_NZ_PER_ROW, (void *)&nz_per_row);
        break;
#     endif
      case SOLV_DMESCHACH:
        ncg->precond = fles_new_dense_meschach_solver (ii, tot_of_equations, error_notify);
        break;
      case SOLV_GEP:
      default:
        ncg->precond = fles_new_gep_solver (ii, tot_of_equations, error_notify);
        break;
      } /* switch (solver) */
      if (ncg->precond == NULL) REPORT_FATAL_ERROR (ii->domain->errh, "No solver");
      break;
    }
  } /* switch (precond_type) */
}


/** 
 **  Configure the newton nonlinear solver 
 **/

static void
config_newton (W_newmark_integrator_implem_t *ii,
               char *optstring /* options string; specify option+value pairs */
               )
{  
  int solver = SOLV_GEP;
  double nz_per_row = 0.01;
  W_bool_t full_newton = TRUE;
  W_bool_t initial_stiff = FALSE;
  int recomp_keff_after = 1;
  int modnewton_after = 2;
  newton_solver_t *newton = NEWTON_HNDL (ii);

  newton->nles = &ii->nles;

  if (optstring) {
    int pos;
    tokens_parser parser = tokens_new_parser ();
    if (tokens_parse_line (parser, optstring)) {
      if ((pos = tokens_have_keyword (parser, 1, "MODNEWTON")) > 0) {
        full_newton = FALSE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "INITIAL_STIFF")) > 0) {
        initial_stiff = TRUE;
        full_newton = FALSE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "RECOMP_KEFF_AFTER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          recomp_keff_after = tokens_token_as_int (parser, pos+1);
        }
        if (recomp_keff_after < 1) recomp_keff_after = 1;
      }
      if ((pos = tokens_have_keyword (parser, 1, "MODNEWTON_AFTER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          modnewton_after = tokens_token_as_int (parser, pos+1);
        }
        if (modnewton_after < 1) modnewton_after = 1;
      }
      if ((pos = tokens_have_keyword (parser, 1, "NZ_PER_ROW")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          nz_per_row = tokens_token_as_double (parser, pos+1);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "SOLVER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          const char *solvname;
          solvname = tokens_token_as_string (parser, pos+1);
          if        (tokens_token_and_keyword_equiv (solvname, "SMESCHACH")) {
            solver = SOLV_SMESCHACH;
          } else if (tokens_token_and_keyword_equiv (solvname, "DMESCHACH")) {
            solver = SOLV_DMESCHACH;
          } else if (tokens_token_and_keyword_equiv (solvname, "GEP")) {
            solver = SOLV_GEP;
          } else if (tokens_token_and_keyword_equiv (solvname, "SPETSC")) {
            solver = SOLV_SPETSC;
          }
        }
      }
    }
    tokens_delete_parser (parser);
  }

  /* Initialize control parameters */
  newton->full_newton         = full_newton;
  newton->initial_stiff       = initial_stiff;
  newton->recomp_keff_after   = recomp_keff_after;
  newton->modnewton_after     = modnewton_after;
  newton->slesctx             = NULL;

  /* The tot_of_equations callback is a little bit tricky here.
     First note that the client of the FLES library is the implementation /ii/.
     Hence, the FLES library needs a tot_of_equations() callback to call
     on the implementation, which in turns calls the tot_of_equations_callback() defined
     on *its* client, ii->client.
   */
  switch (solver) {
  case SOLV_SMESCHACH:
    {
      int method = FLES_SMESCHACH_SPCG;
      newton->slesctx
        = fles_new_sparse_meschach_solver (ii, tot_of_equations, error_notify);
      newton->slesctx->set_option (newton->slesctx, FLES_SMESCHACH_METHOD, (void *)method);
    } 
    break;
#if SOLVER_PETSC
  case SOLV_SPETSC:
    newton->slesctx = fles_new_spetsc_solver (ii, tot_of_equations, error_notify);
    newton->slesctx->set_option (newton->slesctx, PETSC_OPTS_NZ_PER_ROW, (void *)&nz_per_row);
    break;
#endif
  case SOLV_DMESCHACH:
    newton->slesctx
      = fles_new_dense_meschach_solver (ii, tot_of_equations, error_notify);
    break;
  case SOLV_GEP:
  default:
    newton->slesctx
      = fles_new_gep_solver (ii, tot_of_equations, error_notify);
    break;
  }
  if (newton->slesctx == NULL) REPORT_FATAL_ERROR (ii->domain->errh, "No solver");
}

static void
advance (W_newmark_integrator_implem_t *ii)
{
  start_step (ii);
  solve_for_u_corr (ii);
  update_config (ii);
  finalize_step (ii);
}

static W_float_t
get_time (W_newmark_integrator_implem_t *ii)
{
  return ii->t;
}

static W_float_t
get_dt (W_newmark_integrator_implem_t *ii)
{
  return ii->dt;
}

static void
update_history (W_newmark_integrator_implem_t *ii, W_history_writer_t *hw)
{
  history_client_data_t *cd = &ii->hist_buf;
  double time = ii->t;
  WHR_store (hw, "newmdflt-R0", time, cd->residual_norm_0);
  WHR_store (hw, "newmdflt-R", time, cd->residual_norm);
  WHR_store (hw, "newmdflt-d0", time, cd->displ_incr_norm_0);
  WHR_store (hw, "newmdflt-d", time, cd->displ_incr_norm);
  WHR_store (hw, "newmdflt-niter", time, cd->niter);
  WHR_store (hw, "newmdflt-pred", time, cd->pred);
  WHR_store (hw, "newmdflt-extf", time, cd->extf_norm);
  WHR_store (hw, "newmdflt-inertf", time, cd->inertf_norm);
  WHR_store (hw, "newmdflt-resid_work", time, cd->resid_work);
  WHR_store (hw, "newmdflt-resid_work_bound", time, cd->resid_work_bound);
  WHR_store (hw, "newmdflt-resid_work_incr", time, cd->resid_work_incr);
  WHR_store (hw, "newmdflt-resid_work_incr_bound", time, cd->resid_work_incr_bound);
  if (cd->resid_work_incr_bound != 0) {
    WHR_store (hw, "newmdflt-resid_work_cos", time,
               fabs (cd->resid_work_incr) / cd->resid_work_incr_bound);
  }
} 

static int 
done (W_newmark_integrator_implem_t *i)
{
  return TRUE;
}

static void
get_consts (W_newmark_integrator_implem_t *ii, double *beta, double *gamma)
{
  *beta = ii->beta;
  *gamma = ii->gamma;
}


/*
  Get integrator timing statistics.
  Get integrator timing statistics.
 */
static void
get_stat (W_newmark_integrator_implem_t *ii,
                double *ass_rhs_time, 
                double *ass_lhs_time, 
                double *solve_time,  
                double *start_time, 
                double *updt_time
                )
{
  *ass_rhs_time = W_timer_get_time (ii->ass_rhs_timer);
  *ass_lhs_time = W_timer_get_time (ii->ass_lhs_timer);
  *solve_time   = W_timer_get_time (ii->solve_timer);
  *start_time   = W_timer_get_time (ii->start_timer);
  *updt_time    = W_timer_get_time (ii->updt_timer);
}

static W_newmark_integrator_methods_t methods = {
  advance,
  get_time,
  get_dt,
  get_consts,
  get_stat,
  update_history,
  done
};

/* ######################################################################### */
/*              Newmark public interface:                                    */
/* ######################################################################### */

/*
   New default Newmark integrator.
   Create a new  Newmark integrator implementation.
   This is the default (or inner) layer of Newmark integrator implementation.
   The plan is to allow integrator implementation(s) layered on top of this one
   to do model reduction and so on.
 */
struct W_newmark_integrator_t *
W_newmark_new_dflt_integrator (W_newmark_client_t     client, /* client (may be, but needn't be, the domain) */
                               W_domain_t *domain, /* domain */
                               W_float_t               t0, /* initial time */
                               W_float_t               dt, /* time step */
                               W_float_t               beta, /* Newmark beta */
                               W_float_t               gamma, /* Newmark gamma */
                               fles_tot_of_equations_callback
                               tot_of_equations_callback, /* total-of-equations callback; this function is defined
                                                             on the client of the implementation */
                               W_newmark_assemble_eff_stiff_client_proc
                               assemble_eff_stiff, /* procedure to assemble
                                                      effective stiffness */
                               W_newmark_assemble_residual_force_client_proc
                               residual_force, /* procedure to assemble
                                                  residual force */
                               W_newmark_start_step_client_proc
                               start_step, /* procedure to start a new step */
                               W_newmark_update_config_client_proc
                               update_config, /* procedure to update configuration */
                               W_newmark_finalize_step_client_proc
                               finalize_step, /* procedure to finalize step */
                               char *optstring /* options string; specify option+value pairs */
                               )
{  
  struct W_newmark_integrator_t *i = NULL;
  W_newmark_integrator_implem_t *ii;
  int max_iter = 12;
  double conv_ratio = 0.0001;
  double nz_per_row = 0.01;
  double conv_res_mag = 0;
  W_bool_t report_iter = FALSE;
  W_bool_t pass_on_max = FALSE;
  nonlinear_solver_type_enum nles_type = NEWTON_SOLVER;
  
  /* Create integrator implementation */
  ii = w_new  (W_newmark_integrator_implem_t); CHECK_MALLOC(ii);
  
  /* Client */
  ii->client = client;
  
  /* Other parameters & variables */
  ii->domain = domain;
  ii->t  = t0;
  ii->dt = dt;
  if (gamma <= 0) ii->gamma = 0.5;
  else            ii->gamma = gamma;
  if (beta <= 0)  ii->beta  = 0.25 * (0.5 + ii->gamma) * (0.5 + ii->gamma);
  else            ii->beta  = beta;
  ii->iter = 0;
  ii->step = 0;
  ii->conv_in_equil = TRUE;
  ii->conv_in_displ = FALSE;
  ii->residual_norm_0 = 0;
  ii->residual_norm = 0;
  ii->displ_incr_norm_0 = 0;
  ii->displ_incr_norm = 0;
  ii->extf_norm = ii->extf_norm_0 = 0;
  ii->inertf_norm = ii->inertf_norm_0 = 0;
  ii->resid_work = 0;
  ii->resid_work_bound = 0;
  ii->resid_work_incr = 0;
  ii->resid_work_incr_bound = 0;
  {
    int id = WOPT_OPT (ii->domain->options, integrator_detail);
    if (id == W_NEWMARK_PRED_UNKNOWN) id = W_NEWMARK_PRED_DEFAULT;
    if (!(id == W_NEWMARK_PRED_SAME_U ||
          id == W_NEWMARK_PRED_SAME_V ||
          id == W_NEWMARK_PRED_SAME_A ||
          id == W_NEWMARK_PRED_ZERO_A ||
          id == W_NEWMARK_PRED_PREV_INCR ||
          id == W_NEWMARK_PRED_ADAPTIVE)) {
      REPORT_FATAL_ERROR (ii->domain->errh, "Unknown Newmark predictor");
    }
    ii->predictor = (W_newmark_predictor_enum)id;
  }
  ii->curr_predictor = ii->predictor;
  
  if (optstring) {
    int pos;
    tokens_parser parser = tokens_new_parser ();
    if (tokens_parse_line (parser, optstring)) {
      if ((pos = tokens_have_keyword (parser, 1, "NEWTON")) > 0) {
        nles_type = NEWTON_SOLVER;
      }
      if ((pos = tokens_have_keyword (parser, 1, "NCG")) > 0) {
        nles_type = NCG_SOLVER;
      }
      if ((pos = tokens_have_keyword (parser, 1, "REPORT_ITER")) > 0) {
        report_iter = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "PASS_ON_MAX")) > 0) {
        pass_on_max = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "CONV_IN_DISPL")) > 0) {
        ii->conv_in_displ = TRUE;
        ii->conv_in_equil = FALSE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "CONV_IN_EQUIL")) > 0) {
        ii->conv_in_displ = FALSE;
        ii->conv_in_equil = TRUE;
      }
      if ((pos = tokens_have_keyword (parser, 1, "MAX_ITER")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          max_iter = tokens_token_as_int (parser, pos+1);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "GAMMA")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          ii->gamma = tokens_token_as_double (parser, pos+1);
          ii->beta  = 0.25 * (0.5 + ii->gamma) * (0.5 + ii->gamma);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "CONV_RATIO")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          conv_ratio = tokens_token_as_double (parser, pos+1);
        }
      }
      if ((pos = tokens_have_keyword (parser, 1, "CONV_RES_MAG")) > 0) {
        if (tokens_total_of_tokens (parser) >= pos+1) {
          conv_res_mag = tokens_token_as_double (parser, pos+1);
        }
      }
    }
    tokens_delete_parser (parser);
  }

  /* Initialize procedures */
  ii->assemble_eff_stiff  = assemble_eff_stiff;
  ii->update_config       = update_config;
  ii->start_step          = start_step;
  ii->residual_force      = residual_force;
  ii->finalize_step       = finalize_step;

  /* Initialize control parameters */
  ii->max_iter            = max_iter;
  ii->conv_ratio          = conv_ratio;
  ii->conv_res_mag        = conv_res_mag;
  ii->report_iter         = report_iter;
  ii->pass_on_max         = pass_on_max;

  /* nonlinear solver */
  ii->nles.ii   = ii;
  ii->nles.type = nles_type;
  ii->nles.tot_of_equations_callback = tot_of_equations_callback;

  /* need this for local arrays */
  ii->num_eqns = ii->nles.tot_of_equations_callback (ii->client);
  if (ii->num_eqns <= 0) {
    REPORT_FATAL_ERROR (ii->domain->errh, "Number of equations non-positive");
  }

  /* Initialize nonlinear solver */
  switch (ii->nles.type) {
  /* list additional solver(s) here */
  case NCG_SOLVER:
    config_ncg (ii, optstring);
    break;
  case NEWTON_SOLVER:
  default:
    config_newton (ii, optstring);
    break;
  }

  ii->Delta_u = w_new_array (ii->num_eqns, double); CHECK_MALLOC (ii->Delta_u);
  ii->R = w_new_array (ii->num_eqns, double); CHECK_MALLOC (ii->R);
  vzero (ii->R, ii->num_eqns);
  ii->prev_R = w_new_array (ii->num_eqns, double); CHECK_MALLOC (ii->prev_R);
  vzero (ii->prev_R, ii->num_eqns);
  ii->eqnnums = w_new_array (ii->num_eqns, int); CHECK_MALLOC (ii->eqnnums);
  { int j; FOR (j, ii->num_eqns) { ii->eqnnums[j] = j; } }
  
  ii->ass_rhs_timer = W_timer_new ();
  ii->ass_lhs_timer = W_timer_new ();
  ii->solve_timer   = W_timer_new ();
  ii->updt_timer    = W_timer_new ();
  ii->start_timer   = W_timer_new ();

  save_hist (ii);

  i = newmark_new_front_end (ii, &methods);

  return i;
}
