#include "Wconfig.h"
#include "Wxf.h"
#include "Wnewmarki.h"
#include "Woptions.h"
#include "Wnode.h"
#include "Whistry.h"
#include "Wutil.h"
#include "Wnewmfend.h"
#include "Wnewmdflt.h"
#include "Wnewmmodr.h"

/* RCSHEADER "$Id: newmfend.c,v 1.1.1.1 2001/07/24 02:42:51 pkrysl Exp $" */

/*
  This is the Newmark integrator front-end.
 */

typedef struct W_newmark_integrator_t {
  struct W_newmark_integrator_implem_t         *ii; /* implementation */
  W_newmark_integrator_methods_t                methods; /* methods used by the implementation */
}              W_newmark_integrator_t;

/* ######################################################################### */
/*              Public interface functions follow:                           */
/* ######################################################################### */

/*
  Advance Newmark integrator.
   Advance the integrator in time by a single time step.
 */
void
W_newmark_advance (W_newmark_integrator_t *i /* integrator */)
{
  i->methods.advance (i->ii);
}

/*
  Get current time.
   Get the current time from the integrator.
 */
W_float_t /* time */
W_newmark_get_time (W_newmark_integrator_t *i/* integrator */)
{
  return i->methods.time (i->ii);
}

/*
   Get the time step.
   Get the time step from the integrator.
 */
W_float_t
W_newmark_get_dt (W_newmark_integrator_t *i/* integrator */)
{
  return i->methods.dt (i->ii);
}

/*
 Update history.
 Update history by storing the current values of integrator data.
 */
void
W_newmark_update_history (W_newmark_integrator_t *i, /* integrator */
                          W_history_writer_t *hw /* history writer */
                          )
{
  if (i->methods.update_history)
    i->methods.update_history (i->ii, hw);
} 

/*
  Finish step.
  Finish step by giving the integrator a chance to perform whatever actions
  it deems necessary.
 */
int /* ignore */
W_newmark_done (W_newmark_integrator_t *i /* integrator */)
{
  if (i->methods.done)
    return i->methods.done (i->ii);
  else
    return TRUE;
}

/*
  Get constant from integrator.
  Get the values of the Newmark parameters from the integrator.
 */
void
W_newmark_get_consts (W_newmark_integrator_t *i, /* integrator */
                      double *beta, /* Newmark beta */
                      double *gamma /* Newmark gamma */
                      )
{
  i->methods.get_consts (i->ii, beta, gamma);
}


/*
  Get integrator timing statistics.
  Get integrator timing statistics.
 */
void
W_newmark_get_stat (W_newmark_integrator_t *i,
                    double *ass_rhs_time, /* CPU time to assemble rhs */
                    double *ass_lhs_time,  /* CPU time to assemble lhs */
                    double *solve_time,  /* CPU time to solve */
                    double *start_time, /* CPU time to start step */
                    double *updt_time /* CPU time to update configuration */
                    )
{
  if (i->methods.stat) {
    i->methods.stat (i->ii, ass_rhs_time, ass_lhs_time, solve_time, start_time, updt_time);
  } else {
    *ass_rhs_time = 0;
    *ass_lhs_time = 0;
    *solve_time   = 0;
    *start_time   = 0;
    *updt_time    = 0;
  }
}


/*
   New Newmark integrator.
   Create a new Newmark integrator.
   Called by implementations only.
 */
struct W_newmark_integrator_t * /* integrator or NULL if required methods not defined */
newmark_new_front_end (
                       struct W_newmark_integrator_implem_t *ii,
                       W_newmark_integrator_methods_t *methods
                       )
{
  W_newmark_integrator_t *i;
  
  if (methods->advance    == NULL ||
      methods->time       == NULL ||
      methods->dt         == NULL ||
      methods->get_consts == NULL
      ) {
    return NULL;
  }

  i = w_new (W_newmark_integrator_t); CHECK_MALLOC (i);
  
  i->ii      = ii;
  i->methods = *methods;

  return i;
}

/*
  Get integrator implementation.
  Called by implementations only.
 */
struct W_newmark_integrator_implem_t * /* implementation */
newmark_implem (struct W_newmark_integrator_t *i)
{
  return i->ii;
}
