#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Welem.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wmat1d.h"
#include "Wmat1deng.h"
#include "Wmat1dstvk.h"
#include "Wcdi.h"
#include "Wnewmarki.h"
#include "Wntruss.h"
#include "Whistry.h"
#include "meschach.h"

/* RCSHEADER "$Id: ntruss.c,v 1.3 2002/03/21 01:32:35 cvs Exp $" */

/*
  The type of a ntruss element.  
 */
typedef struct priv_elem_t {
  W_node_t          *node[2]; /* node ID's */
  W_float_t          A;		  /* cross-sectional area */
  W_float_t          L0;		  /* length -- cached */
  W_material_t      *material;
  W_mat_1d_state_t   matst;
}              priv_elem_t;


/* history client data */
typedef struct history_client_data_t {
  W_elem_t          *el; /* FE */
  W_index_t          pntidx; /* unique integer  */
}              history_client_data_t;

static int
elem_elcode (W_elem_t *el)
{
  return NTRUSS_ELCODE;
}


#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

/*
  Compute the element length.
  */
static double
length (priv_elem_t *t)
{
  W_coords_3d_t c[2];
  W_coords_3d_t N;

  GET_FROM_NODE (t->node[0], WN_get_ref_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, c[1]);
 
  VECMAC_3_VEC(N, =, c[1], -, c[0]);
  return VECMAC_NORM (N);
} 

/*
  Returns a unmodifiable string with the element name.
 */
static char *
type_name (W_elem_t *el)
{
  static char buf[] = "ntruss";
  return buf;
}

typedef struct ip_ctx_data_t {
  priv_elem_t *t;
  double lt, ltdt, l0;
}              ip_ctx_data_t;

static W_bool_t 
get_kinem_var (W_mat_1d_ip_ctx_t *ctx, 
               W_mat_1d_kinem_var_type_enum  type,
               double *kinvar
               )
{
  ip_ctx_data_t *ipctx = (ip_ctx_data_t *)ctx->ctx_data;

  switch (type) {
  case MAT_1D_ENG_STRAIN:
    *kinvar = ipctx->lt / ipctx->l0 - 1;
    break;
  case MAT_1D_GREEN_LAGRANGE:
    {
      double du = ipctx->lt - ipctx->l0;
      double dul = du / ipctx->l0;
      *kinvar = dul + dul*dul / 2;
    }
    break;
  case MAT_1D_PREV_DISPL_GRAD:
    *kinvar = ipctx->ltdt / ipctx->l0 - 1;
    break;
  case MAT_1D_DISPL_GRAD:
    *kinvar = ipctx->lt / ipctx->l0 - 1;
    break;
  default:
    REPORT_FATAL_ERROR(NULL, "Unknown kinematic tensor type");
    break;
  }

  return TRUE;
}

#ifdef X_XFACE
/*
  Modify the graphic representation of the element (a 3D line).
 */
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  WCRec pnts[2];
  W_coords_3d_t refc[2], u[2];
  priv_elem_t *t = (priv_elem_t *)el->data;

  GET_FROM_NODE (t->node[0], WN_get_ref_coords, refc[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, refc[1]);
  GET_FROM_NODE (t->node[0], WN_get_displ, u[0]);
  GET_FROM_NODE (t->node[1], WN_get_displ, u[1]);

  VECMAC_3_VEC (pnts[0], =, refc[0], +, displ_scale * u[0]);
  VECMAC_3_VEC (pnts[1], =, refc[1], +, displ_scale * u[1]);

  if (el->grep == NULL) {
    el->grep = CreateLine3D(pnts);
    EASValsSetColor(WM_get_color(t->material));
    EGWithMaskChangeAttributes(COLOR_MASK, el->grep);
    EGAttachObject(el->grep, (EObjectP)el);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
  } else {
    if (WOPT_OPT (el->domain->options, color_encoded_forces)) {
        W_material_t *m = t->material;
        W_mat_1d_t *m1d = (W_mat_1d_t *)m->data;
        double eps_pl = 0;
        EPixel color;
        EFringeTable ft = EMGetAssocFringeTable (ESIModel());
        ip_ctx_data_t ipctx_data;
        W_mat_1d_ip_ctx_t ip_ctx;
        ipctx_data.t         = t;
        ip_ctx.time          = WD_time (el->domain);
        ip_ctx.time_step     = WD_time_step (el->domain);
        ip_ctx.ctx_data      = &ipctx_data;
        ip_ctx.get_kinem_var = get_kinem_var;
        WM1D_get_mat_state_var (m1d, t->matst, &ip_ctx,
                                MAT_1D_EQUIV_PLASTIC_DEF, &eps_pl);
        color = ColorFringeRangeToColor (ColorFringeValueToRange (ft, eps_pl));
        EASValsSetColor(color);
        EGWithMaskChangeAttributes(COLOR_MASK, el->grep);
    }  
    EGModifyGraphicsByHandle(el->grep, 1, &pnts[0]);
    EGModifyGraphicsByHandle(el->grep, 2, &pnts[1]);
  }
}

#endif


/*
  Assembles nodal forces by which the ntruss element acts on its nodes.
  All effects (restoring and damping forces) must be included.
 */
static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_float_t nforce, L;
  double ded;
  W_coords_3d_t c[2], cdt[2], N, pN, force;
  W_float_t stress;
  ip_ctx_data_t ipctx_data;
  W_mat_1d_t *m = (W_mat_1d_t *)t->material->data;
  W_mat_1d_ip_ctx_t ip_ctx;
  W_mat_3d_stress_type_enum out_stress_type;
  
  ipctx_data.t         = t;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ipctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;

  /* kinematic quantities */
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  GET_FROM_NODE (t->node[0], WN_get_prev_coords, cdt[0]);
  GET_FROM_NODE (t->node[1], WN_get_prev_coords, cdt[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  ipctx_data.lt = L = VECMAC_NORM (N);
  VECMAC_3_VEC (pN, =, cdt[1], -, cdt[0]);
  ipctx_data.ltdt = VECMAC_NORM (pN);
  ipctx_data.l0 = t->L0;
  VECMAC_SCALAR (N, /=, L);
   
  /* obtain stress */
  if (!WM1D_update_stress (m, t->matst, update_form, &ip_ctx,
                           MAT_3D_CAUCHY, &stress, &out_stress_type))
    REPORT_FATAL_ERROR (el->domain->errh, "Failed to compute stress");

  nforce = stress * t->A;

  VECMAC_3_SCALAR (force, =, nforce * N.x, nforce * N.y, nforce * N.z);
  WN_assemble_force (t->node[0], &force);
  VECMAC_3_SCALAR (force, = , -force.x, -force.y, -force.z);
  WN_assemble_force (t->node[1], &force);

  if (!WM1D_get_mat_state_var (m, t->matst, &ip_ctx,
                               MAT_1D_DEFORM_ENERGY_DENSITY, &ded))
    REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
  
  el->deformation_energy = (t->A * t->L0 * ded);
}


/*
  Compute the total mass of the element.
 */
static W_float_t
total_mass(W_elem_t *el)
{
  W_float_t volume_mass;
  priv_elem_t *t= (priv_elem_t *)el->data;
  W_mat_1d_t *m = (W_mat_1d_t *)t->material->data;

  volume_mass = WM1D_get_volume_mass (m);
  
  return volume_mass * t->A * t->L0;
}


/*
  Assemble mass to the nodes.
 */
static void
assemble_mass(W_elem_t *el)
{
  W_float_t mass2;
  priv_elem_t *t = (priv_elem_t *)el->data;
  
  mass2 = total_mass(el) / 2;

  WN_assemble_mass (t->node[0], mass2);
  WN_assemble_mass (t->node[1], mass2);
}


/*
  Compute the suggested time step.
 */
static W_float_t 
suggested_time_step(W_elem_t *el)
{
  W_float_t c_d;
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_mat_1d_t *m = (W_mat_1d_t *)t->material->data;

  if (!WM1D_get_mat_const (m, MAT_1D_DIL_WAVE_SPEED_1D, &c_d))
    REPORT_FATAL_ERROR(el->domain->errh, "No speed");
  
  return t->L0 / c_d;
}



#define EFFICIENT_ELMAT_TRANSF 1

#if EFFICIENT_ELMAT_TRANSF
static void
transf_matrix_eff (W_vect_t *t0, W_vect_t *t1, W_vect_t *t2,
               W_coords_3d_t *refci,
               W_coords_3d_t *refcj,
               W_coords_3d_t *ci,
               W_coords_3d_t *cj
               )
{
  W_vect_t vt, s = {0,0,0}, t = {0,0,0};
  
  VECMAC_3_VEC (vt, =, (*cj), -, (*ci));

  VECMAC_NORMALIZE (vt);
  if (fabs (vt.z) > 0.95) {     /* element almost parallel to the Z-axis */
    W_vect_t x = {1,0,0};
    s = x;
  } else {
    W_vect_t z = {0,0,1};
    s = z;
  }
  VECMAC_CROSS (t, =, vt, x, s);
  VECMAC_CROSS (s, =, t,  x, vt); 
  VECMAC_NORMALIZE (s);
  VECMAC_NORMALIZE (t);

  t0->x = vt.x;
  t0->y = vt.y;
  t0->z = vt.z;

  t1->x = s.x;
  t1->y = s.y;
  t1->z = s.z;

  t2->x = t.x;
  t2->y = t.y;
  t2->z = t.z;
}
#endif

static void
transf_matrix (MAT *tm,
               W_coords_3d_t *refci,
               W_coords_3d_t *refcj,
               W_coords_3d_t *ci,
               W_coords_3d_t *cj
               )
{
  W_vect_t vt, s = {0,0,0}, t = {0,0,0};
  
  VECMAC_3_VEC (vt, =, (*cj), -, (*ci));

  VECMAC_NORMALIZE (vt);
  if (fabs (vt.z) > 0.95) {     /* element almost parallel to the Z-axis */
    W_vect_t x = {1,0,0};
    s = x;
  } else {
    W_vect_t z = {0,0,1};
    s = z;
  }
  VECMAC_CROSS (t, =, vt, x, s);
  VECMAC_CROSS (s, =, t,  x, vt); 
  VECMAC_NORMALIZE (s);
  VECMAC_NORMALIZE (t);

  m_zero (tm);
  
  tm->me[0][0] = vt.x;
  tm->me[1][0] = vt.y;
  tm->me[2][0] = vt.z;

  tm->me[0][1] = s.x;
  tm->me[1][1] = s.y;
  tm->me[2][1] = s.z;

  tm->me[0][2] = t.x;
  tm->me[1][2] = t.y;
  tm->me[2][2] = t.z;

  tm->me[3][3] = vt.x;
  tm->me[4][3] = vt.y;
  tm->me[5][3] = vt.z;

  tm->me[3][4] = s.x;
  tm->me[4][4] = s.y;
  tm->me[5][4] = s.z;

  tm->me[3][5] = t.x;
  tm->me[4][5] = t.y;
  tm->me[5][5] = t.z;
}

static void
copy_to_k (MAT *sm, double m[36])
{
  int i, j, k;
  k = 0;
  FOR (i, 6) {
    FOR (j, 6) {
      m[k] = sm->me[i][j];
      k++;
    }
  }
}




static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                            W_mass_type_enum mass_type
                        )
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_float_t E, nu;
  double ni;
  static MAT *sm = NULL;
  static MAT *smt = NULL;
  static MAT *tm = NULL;
  W_coords_3d_t refc[2], c[2], cdt[2], N, pN;
  double mass2;
  double CA, lt;
  double k[36], tk[36];
  W_index_t idcs[6];
  W_bool_t constrained[6];
  double nforce, stress, strain, strain_rate;
  ip_ctx_data_t ipctx_data;
  W_mat_1d_t *m = (W_mat_1d_t *)t->material->data;
  W_mat_1d_ip_ctx_t ip_ctx;
  int mcode;
  W_mat_3d_stress_type_enum out_stress_type;

  ipctx_data.t         = t;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ipctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;

  /* kinematic quantities */
  GET_FROM_NODE (t->node[0], WN_get_ref_coords, refc[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, refc[1]);
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  GET_FROM_NODE (t->node[0], WN_get_prev_coords, cdt[0]);
  GET_FROM_NODE (t->node[1], WN_get_prev_coords, cdt[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  ipctx_data.lt = lt = VECMAC_NORM (N);
  VECMAC_3_VEC (pN, =, cdt[1], -, cdt[0]);
  ipctx_data.ltdt = VECMAC_NORM (pN);
  ipctx_data.l0 = t->L0;
  VECMAC_SCALAR (N, /=, lt);
   
  if (!WM1D_get_mat_const (m, MAT_1D_YOUNG_MODULUS, &E))
    REPORT_FATAL_ERROR (el->domain->errh, "No Young's modulus");

  if (!sm)  { CATCH(E_MEM, { sm  = m_get (6, 6); }, { sm = NULL; CHECK_MALLOC (sm); }); }
  if (!smt) { CATCH(E_MEM, { smt = m_get (6, 6); }, { smt = NULL; CHECK_MALLOC (smt); }); } 
  if (!tm)  { CATCH(E_MEM, { tm  = m_get (6, 6); }, { tm = NULL; CHECK_MALLOC (tm); }); } 

  m_zero (sm);
  
  /* elastic stiffness matrix */
  /* RAW: we should use a method defined on the material instead of this switch;
     see the way this is done in mat3d materials   */
  mcode = WM1D_get_mcode (m);
  switch (mcode) {
  case MAT_1D_STV_K:
    {
      double du = lt - t->L0;
      ni = stiff_mult * E * t->A *
        (1 + du / t->L0 + du * du / t->L0 / t->L0) / t->L0;
    }
    break;
  case MAT_1D_ENG:
  default:
    ni = stiff_mult * E * t->A / lt;
    break;
  }
  sm->me[0][0] = sm->me[0][0] + ni;
  sm->me[0][3] = sm->me[0][3] - ni;
  sm->me[3][0] = sm->me[0][3];
  sm->me[3][3] = sm->me[3][3] + ni;

  /* stress matrix  */

  if (!WM1D_update_stress (m, t->matst, W_MAT_STATE_UPDATE_TEMP,
                           &ip_ctx, MAT_3D_CAUCHY, &stress, &out_stress_type))
    REPORT_FATAL_ERROR (el->domain->errh, "Failed to compute stress");
  nforce = stress * t->A;

  CA = stiff_mult * nforce / lt;
  sm->me[1][1]  += +CA;
  sm->me[2][2]  += +CA;
  sm->me[1][4]  += -CA;
  sm->me[4][1]  += -CA;
  sm->me[2][5]  += -CA;
  sm->me[5][2]  += -CA;
  sm->me[4][4]  += +CA;
  sm->me[5][5]  += +CA;
  
  /* mass matrix */
  mass2 = total_mass(el) / 2;
  switch (mass_type) {
  case W_MASS_CONSISTENT:
    {
      double diag = 2./3 * mass_mult * mass2;
      double offdiag = 1./2 * mass_mult * mass2;
      sm->me[0][0] += diag;
      sm->me[1][1] += diag;
      sm->me[2][2] += diag;
      sm->me[3][3] += diag;
      sm->me[4][4] += diag;
      sm->me[5][5] += diag;
      sm->me[0][3] += offdiag;
      sm->me[1][4] += offdiag;
      sm->me[2][5] += offdiag;
      sm->me[3][0] += offdiag;
      sm->me[4][1] += offdiag;
      sm->me[5][2] += offdiag;
    }
    break;
  case W_MASS_DIAG: 
  default:
    CA    = mass_mult * mass2;
    sm->me[0][0] += CA;
    sm->me[1][1] += CA;
    sm->me[2][2] += CA;
    sm->me[3][3] += CA;
    sm->me[4][4] += CA;
    sm->me[5][5] += CA;
    break;
  }

  transf_matrix (tm, &refc[0], &refc[1], &c[0], &c[1]);
  m_mlt (tm, sm, smt);
  mmtr_mlt (smt, tm, sm);
  copy_to_k (sm, k);

  WN_eqn_numbers_and_constraints (t->node[0], &idcs[0], &constrained[0]);
  WN_eqn_numbers_and_constraints (t->node[1], &idcs[3], &constrained[3]);
  (*assemble_eff_stiff_callback) (el->domain->newmark_integrator,
                                  FLES_ADD, 6, idcs, k, constrained);
}


/*
  Assembles nodal forces by which the ntruss element acts on its nodes.
  All effects (restoring and damping forces) must be included.
 */
static void
assemble_residual_forces (W_elem_t *el,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_float_t nforce, L;
  double mass2, ded;
  W_coords_3d_t c[2], cdt[2], N, pN, force;
  W_acceleration_3d_t a[2];
  W_float_t stress;
  ip_ctx_data_t ipctx_data;
  W_mat_1d_t *m = (W_mat_1d_t *)t->material->data;
  W_mat_1d_ip_ctx_t ip_ctx;
  W_mat_3d_stress_type_enum out_stress_type;

  ipctx_data.t         = t;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ipctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;

  /* kinematic quantities */
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  GET_FROM_NODE (t->node[0], WN_get_prev_coords, cdt[0]);
  GET_FROM_NODE (t->node[1], WN_get_prev_coords, cdt[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  ipctx_data.lt = L = VECMAC_NORM (N);
  VECMAC_3_VEC (pN, =, cdt[1], -, cdt[0]);
  ipctx_data.ltdt = VECMAC_NORM (pN);
  ipctx_data.l0 = t->L0;
  VECMAC_SCALAR (N, /=, L);
   
  /* obtain stress */
  if (!WM1D_update_stress (m, t->matst, update_form, &ip_ctx,
                           MAT_3D_CAUCHY, &stress, &out_stress_type))
    REPORT_FATAL_ERROR (el->domain->errh, "Failed to compute stress");

  nforce = stress * t->A;

  VECMAC_3_SCALAR (force, =, nforce * N.x, nforce * N.y, nforce * N.z);
  WN_assemble_force (t->node[0], &force);
  VECMAC_3_SCALAR (force, = , -force.x, -force.y, -force.z);
  WN_assemble_force (t->node[1], &force);
   
  GET_FROM_NODE (t->node[0], WN_get_acceleration, a[0]);
  GET_FROM_NODE (t->node[1], WN_get_acceleration, a[1]);
  mass2 = total_mass (el) / 2;
  switch (mass_type) {
  case W_MASS_CONSISTENT:
    VECMAC_3_VEC (force, =, (-2. / 3 * mass2) * a[0], +, (-1. / 2 * mass2) * a[1]);
    WN_assemble_force (t->node[0], &force);
    VECMAC_3_VEC (force, =, (-1. / 2 * mass2) * a[0], +, (-2. / 3 * mass2) * a[1]);
    WN_assemble_force (t->node[1], &force);
    break;
  case W_MASS_DIAG:
  default:
    VECMAC_2_VEC (force, =, (-mass2) * a[0]);
    WN_assemble_force (t->node[0], &force);
    VECMAC_2_VEC (force, =, (-mass2) * a[1]);
    WN_assemble_force (t->node[1], &force);
    break;
  }

  if (!WM1D_get_mat_state_var (m, t->matst, &ip_ctx,
                               MAT_1D_DEFORM_ENERGY_DENSITY, &ded))
    REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
  
  el->deformation_energy = (t->A * t->L0 * ded);
}


static void *
new_history_client_data (W_elem_t *e,
                         W_index_t idx, /* unique integer  */
                         W_history_writer_t *r, /* history writer */
                         char *input_line /* input line (to allow
                                             for additional parsing of
                                             the line contents by this
                                             function) */
                         )
{
  priv_elem_t *t = (priv_elem_t *)e->data;
  history_client_data_t *cd;
  W_domain_t *d = e->domain;
  cd = w_new (history_client_data_t);
  CHECK_MALLOC(cd);
  cd->el     = e;
  TOKENS_PARSE(input_line);
  cd->pntidx = TOKEN_INT(3);
  return (NODE)cd;
}

static void
update_history (W_elem_t *el, /* generic element */
                W_solution_history_t *h, /* solution history */
                W_float_t time
                )
{
  history_client_data_t *cd =
    (history_client_data_t *)h->client_data;
  W_domain_t *d = el->domain;
  priv_elem_t *t = (priv_elem_t *)el->data;
#define STORE(CD, HW, NAME, TIME, VAL)          \
  {                                             \
    char buf[128];                              \
    sprintf (buf, "p%d%s", CD->pntidx, NAME);   \
    WHR_store (HW, buf, TIME, VAL);             \
  }
  W_float_t L;
  W_coords_3d_t c[2], cdt[2], N, pN;
  W_velocity_3d_t vel[2], v;
  W_float_t stress, eps_pl, de;
  ip_ctx_data_t ipctx_data;
  W_mat_1d_t *m = (W_mat_1d_t *)t->material->data;
  W_mat_1d_ip_ctx_t ip_ctx;
  W_mat_3d_stress_type_enum out_stress_type;

  ipctx_data.t         = t;
  ip_ctx.time          = WD_time (el->domain);
  ip_ctx.time_step     = WD_time_step (el->domain);
  ip_ctx.ctx_data      = &ipctx_data;
  ip_ctx.get_kinem_var = get_kinem_var;
 
  /* kinematic quantities */
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  GET_FROM_NODE (t->node[0], WN_get_prev_coords, cdt[0]);
  GET_FROM_NODE (t->node[1], WN_get_prev_coords, cdt[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  ipctx_data.lt = L = VECMAC_NORM (N);
  VECMAC_3_VEC (pN, =, cdt[1], -, cdt[0]);
  ipctx_data.ltdt = VECMAC_NORM (pN);
  ipctx_data.l0 = t->L0;
  VECMAC_SCALAR (N, /=, L);
   
  /* obtain stress */
  if (!WM1D_update_stress (m, t->matst, W_MAT_STATE_UPDATE_TEMP,
                           &ip_ctx, MAT_3D_CAUCHY, &stress, &out_stress_type))
    REPORT_FATAL_ERROR (el->domain->errh, "Failed to compute stress");
  STORE (cd, h->writer, "stress", time, stress);

  if (!WM1D_get_mat_state_var (m, t->matst, &ip_ctx,
                               MAT_1D_DEFORM_ENERGY_DENSITY, &de))
    REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
  STORE (cd, h->writer, "DeformE", time, t->A * t->L0 * de);

  if (!WM1D_get_mat_state_var (m, t->matst, &ip_ctx,
                               MAT_1D_EQUIV_PLASTIC_DEF, &eps_pl))
    REPORT_ERROR(el->domain->errh, "Failed to retrieve work of deformation");
  STORE (cd, h->writer, "EPlS", time, eps_pl);
}


static W_float_t
kinetic_energy_consistent (W_elem_t *el)
{
  priv_elem_t *e = (priv_elem_t *)el->data;
  W_vect_t v[2];
  int i;
  double mass2, ke = 0;
  FOR (i, 2) { GET_FROM_NODE (e->node[i], WN_get_velocity, v[i]); }
  mass2 = total_mass (el) / 2;
  ke += (2. / 3 * mass2) * (VECMAC_DOT (v[0], v[0]) + VECMAC_DOT (v[1], v[1]));
  ke += 2 * (1. / 2 * mass2) * VECMAC_DOT (v[0], v[1]);
  return ke / 2;
}

static W_float_t
kinetic_energy_diag (W_elem_t *el)
{
  priv_elem_t *e = (priv_elem_t *)el->data;
  W_vect_t v;
  int i;
  double ke = 0;
  double m = total_mass (el) / 2;

  FOR (i, 2) {
    GET_FROM_NODE (e->node[i], WN_get_velocity, v);
    ke += VECMAC_DOT (v, v);
  }
  return (m * ke) / 2;
}

static W_float_t
kinetic_energy (W_elem_t *el, W_mass_type_enum mass_type)
{
  switch (mass_type) {
  case W_MASS_CONSISTENT:
    return kinetic_energy_consistent (el);
  default:
    return kinetic_energy_diag (el);
  }
}

/*
  Element method dispatch table.
 */
static W_elem_methods_t ntruss_methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  new_history_client_data,
  update_history,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  assemble_residual_forces,
  assemble_eff_stiffness,
  kinetic_energy
};

#define CENTROID(loc, t)                             \
  {                                                     \
    VECMAC_2_VEC (loc,  =, t->node[0]->ref_coords);  \
    VECMAC_2_VEC (loc, +=, t->node[1]->ref_coords);  \
    VECMAC_SCALAR (loc, *=, 0.5);                      \
  }

/*
  Reads ntruss element data from stream as long as there is no 
  END_OF_BLOCK tag in the input line.  Creates elements. 
  */
W_bool_t 
ntruss_reader(W_domain_t *d)
{
  const char *input_line;
  long elem_num, nodeid;
  priv_elem_t *t;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_vect_t loc;
  W_index_t matid;
  W_material_t *mat;
  W_mat_1d_t *m;
 
  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      t = w_new(priv_elem_t);
      CHECK_MALLOC(t);
      TOKENS_PARSE(input_line);
      elem_num    = TOKEN_INT(1);
      nodeid      = TOKEN_INT(2);
      t->node[0] = WD_get_node_by_idx (d, nodeid);
      if (t->node[0] == NULL) {
        sprintf (W_buf (), "No such node: %ld", nodeid);
        REPORT_FATAL_ERROR (d->errh, W_buf ());
      }
      nodeid      = TOKEN_INT(3);
      t->node[1] = WD_get_node_by_idx (d, nodeid);
      if (t->node[1] == NULL) {
        sprintf (W_buf (), "No such node: %ld", nodeid);
        REPORT_FATAL_ERROR (d->errh, W_buf ());
      }
      
      matid = TOKEN_INT(4);
      mat = WD_get_material_by_idx(d, matid);
      if (mat == NULL) {
        sprintf(W_buf(), "No such material: %ld", matid);
        REPORT_ERROR(d->errh, W_buf());
      }
      t->material = mat;
      t->A        = TOKEN_DBL(5);
      t->L0       = length (t);
      if (t->L0 <= 0) REPORT_FATAL_ERROR (d->errh, "Zero-length element");
      m = (W_mat_1d_t *) t->material->data;
      CENTROID (loc, t);
      t->matst = WM1D_get_mat_state (m, &loc); CHECK_MALLOC (t->matst);
#ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif
      WD_add_element(d, elem_num,
                     WE_new_element(d, &ntruss_methods, t
#                    ifdef PVM_NEED_GHOST_ELEMENTS
                                    , is_ghost
#                    endif
                     ));
    }
  }
  return FALSE;
}
