#include "Woctree.h"


#define LEAF_IS_TERMINAL(L) ((L)->leaves == NULL)
#define LEAF_CONTAINS_CRACK(L) ((L)->contains_crack)

#ifdef X_XFACE
static void
new_g_rep(octree_leaf_t *l);
#endif

static void
leaf_geom(octree_root_t *root, int level, loc_t *param_center,
          loc_t verts[8]);

#define FOR(_Counter_, _Up_)                                      \
   for (_Counter_ = 0; _Counter_ < _Up_; _Counter_++)

static octree_leaf_t *
new_leaf(octree_root_t *root, octree_leaf_t *parent, int which);
static octree_leaf_t *
top_leaf(octree_root_t *root);
static double
leaf_radius(octree_leaf_t *l);



static loc_t offsets[8] = {
  {-1, -1, -1},
  { 1, -1, -1},
  { 1,  1, -1},
  {-1,  1, -1},
  {-1, -1,  1},
  { 1, -1,  1},
  { 1,  1,  1},
  {-1,  1,  1}
};


/* pow2[i] = pow(2, i) */
static double pow2[] = {
1,
2,
4,
8,
16,
32,
64,
128,
256,
512,
1024,
2048,
4096,
8192,
16384,
32768,
65536,
131072,
262144,
524288,
1048576,
2097152,
4194304,
8388608,
16777216,
33554432,
67108864,
134217728,
268435456,
536870912,
1073741824,
};




octree_root_t *
octree_new_root(void *owner, owner_get_vertex_loc_proc get_vertex_loc)
{
  octree_root_t *root;
  double size;

  root = new_instance(octree_root_t); CHECK_MALLOC(root);

  root->owner          = owner;
  root->get_vertex_loc = get_vertex_loc;
  root->top_leaf       = NULL;
  root->contains_crack = FALSE;

  return root;
}





static void
leaf_geom(octree_root_t *root, int level, loc_t *param_center,
          loc_t verts[8])
{
  loc_t top_verts[8];
  int i;
  double n[8];
  double size = 2.0 / pow2[level];
  double hsize = size / 2;
#define N_LOC1(v, DIR, n)                  \
   v.DIR = (  n[0] * top_verts[0].DIR      \
            + n[1] * top_verts[1].DIR      \
            + n[2] * top_verts[2].DIR      \
            + n[3] * top_verts[3].DIR      \
            + n[4] * top_verts[4].DIR      \
            + n[5] * top_verts[5].DIR      \
            + n[6] * top_verts[6].DIR      \
            + n[7] * top_verts[7].DIR)
#define N_LOC(v) { N_LOC1(v, x, n); N_LOC1(v, y, n); N_LOC1(v, z, n); }

  FOR(i, 8) { root->get_vertex_loc(root->owner, i, &top_verts[i]); }
  FOR(i, 8) {
    G3D_hex_N(n,
              param_center->x + offsets[i].x * hsize,
              param_center->y + offsets[i].y * hsize,
              param_center->z + offsets[i].z * hsize);
    N_LOC(verts[i]);
  }
}





static octree_leaf_t *
new_leaf(octree_root_t *root, octree_leaf_t *parent, int which)
{
  octree_leaf_t *nl;
  double size;
  double hsize;

  nl = new_instance(octree_leaf_t); CHECK_MALLOC(nl);

  nl->root = root;
  nl->level = parent->level + 1;

  size  = 2.0 / pow2[nl->level];
  hsize = size / 2;

  nl->param_center.x = parent->param_center.x + offsets[which].x * hsize;
  nl->param_center.y = parent->param_center.y + offsets[which].y * hsize;
  nl->param_center.z = parent->param_center.z + offsets[which].z * hsize;

  nl->radius = leaf_radius(nl);

  nl->leaves = NULL;
  
  nl->contains_crack = FALSE;

  return nl;
}






static octree_leaf_t *
top_leaf(octree_root_t *root)
{
  octree_leaf_t *nl;
  double size = 2;
  double hsize = size / 2;

  nl = new_instance(octree_leaf_t); CHECK_MALLOC(nl);

  nl->root = root;
  nl->level = 0;

  nl->param_center.x = 0; 
  nl->param_center.y = 0;
  nl->param_center.z = 0;
 
  nl->radius = leaf_radius(nl);

  nl->leaves = NULL;
  
  nl->contains_crack = FALSE;

  return nl;
}




static double
leaf_radius(octree_leaf_t *l)
{
  loc_t verts[8];
  double r;

  leaf_geom(l->root, l->level, &l->param_center, verts);
  r = v3_distance(&verts[0], &verts[6]);
  r = max(r, v3_distance(&verts[1], &verts[7]));
  r = max(r, v3_distance(&verts[2], &verts[4]));
  r = max(r, v3_distance(&verts[3], &verts[5]));

  return r/2;
}



#ifdef X_XFACE
static void
new_g_rep(octree_leaf_t *l);
#endif

#ifdef X_XFACE
static void
new_g_rep(octree_leaf_t *l)
{
  WCRec pts[8];
  loc_t verts[8];
  int i;

  leaf_geom(l->root, l->level, &l->param_center, verts);

  FOR(i, 8) {
    pts[i].x = verts[i].x;
    pts[i].y = verts[i].y;
    pts[i].z = verts[i].z;
  }

  l->g_rep = CreateHexahedron(pts);
  EASValsSetLayer(MAX_LAYER - l->level - 1);
  EASValsSetColor(ColorFringeRangeToColor(l->level));
  EGWithMaskChangeAttributes(COLOR_MASK | LAYER_MASK, l->g_rep);
  EMNoDrawAddGraphicsToModel(ESIModel(), l->g_rep);
}
#endif



static void
extend_crack(octree_leaf_t *l, G3D_poly_t *poly, double terminal_size);
static int
leaf_is_to_be_cut(loc_t verts[8], G3D_poly_t *poly, double tol);


void
octree_extend_crack(octree_root_t *root,
                    G3D_poly_t *poly,
                    double terminal_size)
{
  loc_t verts[8];
  
  if (root->top_leaf == NULL) root->top_leaf = top_leaf(root);
  
  leaf_geom(root, root->top_leaf->level, &root->top_leaf->param_center, verts);
  if (leaf_is_to_be_cut(verts, poly, root->top_leaf->radius/1000)) {
    extend_crack(root->top_leaf, poly, terminal_size);
  }
}




static octree_leaf_t **
new_leaves(void);


static void
extend_crack(octree_leaf_t *l, G3D_poly_t *poly, double terminal_size)
{
  loc_t verts[8];
  int i;
  loc_t param_center;
  double size = 2.0 / pow2[l->level+1];
  double hsize = size / 2;

  if (l->radius > terminal_size) {
    FOR(i, 8) {
      param_center.x = l->param_center.x + offsets[i].x * hsize;
      param_center.y = l->param_center.y + offsets[i].y * hsize;
      param_center.z = l->param_center.z + offsets[i].z * hsize;
      leaf_geom(l->root, l->level+1, &param_center, verts);
      if (leaf_is_to_be_cut(verts, poly, l->radius/2000)) {
        if (l->leaves == NULL) l->leaves = new_leaves();
        if (l->leaves[i] == NULL) l->leaves[i] = new_leaf(l->root, l, i);
        extend_crack(l->leaves[i], poly, terminal_size);
      }
    }
  } else {
    l->contains_crack = TRUE; 
#   ifdef X_XFACE
    new_g_rep(l);
#   endif
  }
}



static octree_leaf_t **
new_leaves(void)
{
  octree_leaf_t **la;
  la = new_array(8, octree_leaf_t *);
  CHECK_MALLOC(la);
  return la;
}




static int
poly_inside_hex(loc_t verts[8], G3D_poly_t *poly);
static void
get_face_vert_loc(G3D_vertex_t v, double *x, double *y, double *z);

static G3D_poly_t *face = NULL;
static G3D_vertex_t vertices[4];

static int
leaf_is_to_be_cut(loc_t verts[8], G3D_poly_t *poly, double tol)
{
  int i, k, j;
  G3D_segment_t seg;

  /* first check for intersections of the hex edges with the poly */
  FOR(i, 12) {
    seg.start = verts[g3dhex_vertex_on_edge[i][0]];
    seg.end   = verts[g3dhex_vertex_on_edge[i][1]];
    if (G3D_segment_intersects_poly(&seg, poly, tol))
      return TRUE;
  }

  /* now check for intersections of the poly edges with the hex faces */
  if (face == NULL) {
    FOR(k, 4) vertices[k] = (G3D_vertex_t)&verts[k];
    face = G3D_new_poly(4, vertices, get_face_vert_loc);
  }
  FOR(i, poly->num_vert-1) {
    G3D_get_vert_loc(poly, i, &seg.start);
    G3D_get_vert_loc(poly, i+1, &seg.end);
    FOR(j, 6) {
      FOR(k, 4)
        vertices[k] = (G3D_vertex_t)&verts[g3dhex_vertex_on_face[j][k]];
      face = G3D_modify_poly(face, 4, vertices, get_face_vert_loc);
      if (G3D_segment_intersects_poly(&seg, face, tol))
        return TRUE;
    }
  }
  G3D_get_vert_loc(poly, poly->num_vert-1, &seg.start);
  G3D_get_vert_loc(poly, 0, &seg.end);
  FOR(j, 6) {
    FOR(k, 4) vertices[k] = (G3D_vertex_t)&verts[g3dhex_vertex_on_face[j][k]];
    face = G3D_modify_poly(face, 4, vertices, get_face_vert_loc);
    if (G3D_segment_intersects_poly(&seg, face, tol))
      return TRUE;
  }

  /* finally check whether the poly is enclosed by the hex */
  return poly_inside_hex(verts, poly);
}


static void
get_face_vert_loc(G3D_vertex_t v, double *x, double *y, double *z)
{
  loc_t *loc = (loc_t *)v;
  *x = loc->x;
  *y = loc->y;
  *z = loc->z;
}



static int
poly_inside_hex(loc_t verts[8], G3D_poly_t *poly)
{
  G3D_box_t hexbb;
  G3D_box_t polybb;
  int i;

  hexbb.lower.x = verts[7].x;
  hexbb.upper.x = verts[7].x;
  hexbb.lower.y = verts[7].y;
  hexbb.upper.y = verts[7].y;
  hexbb.lower.z = verts[7].z;
  hexbb.upper.z = verts[7].z;
  FOR(i, 7) {
    hexbb.lower.x = min(hexbb.lower.x, verts[i].x);
    hexbb.upper.x = max(hexbb.upper.x, verts[i].x);
    hexbb.lower.y = min(hexbb.lower.y, verts[i].y);
    hexbb.upper.y = max(hexbb.upper.y, verts[i].y);
    hexbb.lower.z = min(hexbb.lower.z, verts[i].z);
    hexbb.upper.z = max(hexbb.upper.z, verts[i].z);
  }
  G3D_poly_bbox(poly, &polybb);
  return
    (   (hexbb.lower.x <= polybb.lower.x && hexbb.upper.x >= polybb.upper.x)
     && (hexbb.lower.y <= polybb.lower.y && hexbb.upper.y >= polybb.upper.y)
     && (hexbb.lower.z <= polybb.lower.z && hexbb.upper.z >= polybb.upper.z));
}







