#include "Wconfig.h"


#include "Wtypes.h"
#include "Wpcont4.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wutil.h"
#include "EID.h"
#include "g3d.h"
#include "Wmat3d.h"
#include "Wcontact.h"

typedef struct contact_point_t {
  double                       param_coords[2];
  struct pcont4_t             *in_contact_cf;
#ifdef X_XFACE
  GraphicObj                  *go;
#endif
}              contact_point_t;

typedef struct pcont4_t {
  W_contact_surf_t             *cs;
  W_node_t                     *n[3];
  int                           ncp;
  struct contact_point_t       *cp;
  G3D_plane_t                   plane; /* cached plane */
  G3D_box_t                     box; /* cached box */
  double                        cp_area;
#ifdef X_XFACE
  GraphicObj                   *go;
#endif
}              pcont4_t;

typedef struct pcont_t {
  W_node_t          *node[3];
  W_material_t      *material;
  W_mat_3d_state_t  *matst;
  W_bool_t           use_2pt_rule;
}              pcont_t;

static double subt_mult[6/*vertex*/][3/*mult*/] = {
  {1,   0,   0  },
  {0,   1,   0  },
  {0,   0,   1  },
  {0.5, 0.5, 0  },
  {0  , 0.5, 0.5},
  {0.5, 0  , 0.5}
};

static int subt_idx[4 /*subtriangle*/][3 /*vertex*/] = {
  {0,3,5},
  {3,1,4},
  {4,2,5},
  {3,4,5}
};

static void
init_cp (pcont4_t *cf,
         double ac[3][3],
         int level,
         int subdiv_level)
{
  if (level == subdiv_level) {
    int j, i;
    i = cf->ncp;
    FOR (j, 3) {
      cf->cp[i].area_coords[j] = (ac[0][j] + ac[1][j] + ac[2][j]) / 3;
    }
    cf->cp[i].in_contact_cf    = NULL;
#ifdef X_XFACE
    cf->cp[i].go               = NULL;
#endif
    cf->ncp++;
  } else {
    double nlac[3][3];
    int subt, v, c;

    FOR (subt, 4) {
      FOR (v, 3) {
        FOR (c, 3) {
          nlac[v][c]
            = (subt_mult[subt_idx[subt][v]][0] * ac[0][c] +
               subt_mult[subt_idx[subt][v]][1] * ac[1][c] +
               subt_mult[subt_idx[subt][v]][2] * ac[2][c]);
        }
      }
      init_cp (cf, nlac, level+1, subdiv_level);
    }
  }
}

static  double area_coords[3][3] = {
    {1,0,0},
    {0,1,0},
    {0,0,1}
  };

static double
face_area (pcont4_t *cf)
{ 
  int i;
  CkitPoint3 l[3], v0, v1, c={0,0,0};
  FOR (i, 3) {
    WN_get_coords (cf->n[i], &l[i].x, &l[i].y, &l[i].z);
  }
  VECMAC_3_VEC (v0, =, l[1], -, l[0]);
  VECMAC_3_VEC (v1, =, l[2], -, l[0]);
  VECMAC_CROSS (c, =, v0, x, v1);
  return VECMAC_NORM (c) / 2;
}

static void
subdiv_contact_face (pcont4_t *cf, int subdiv_level)
{
  cf->ncp = (int) pow (4, subdiv_level);
  cf->cp  = w_realloc_array (cf->cp, cf->ncp, contact_point_t);
  CHECK_MALLOC (cf->cp);
  cf->ncp = 0; /* so that we can use it as a counter */
  init_cp (cf, area_coords, 0, subdiv_level);
  cf->cp_area = face_area (cf) / cf->ncp;
}


static pcont4_t *
new_contact_face (W_node_t *n0, W_node_t *n1, W_node_t *n2,
                      int subdiv_level)
{
  pcont4_t *cf;
  W_coords_3d_t zerov = {0,0,0};
  
  cf = w_new (pcont4_t); CHECK_MALLOC (cf);
  cf->n[0]                 = n0;
  cf->n[1]                 = n1;
  cf->n[2]                 = n2;
  if (subdiv_level <= 0) {
    /* subdivision level unspecified */
    cf->cp = NULL;
    cf->ncp = 0;
  } else {
    cf->ncp = (int) pow (4, subdiv_level);
    cf->cp = w_new_array (cf->ncp, contact_point_t); CHECK_MALLOC (cf->cp);
    cf->ncp = 0; /* so that we can use it as a counter */
    init_cp (cf, area_coords, 0, subdiv_level);
  }
  cf->plane.n              = zerov;
  cf->plane.u              = zerov;
  cf->plane.v              = zerov;
  cf->plane.c              = 0;
  cf->cp_area              = 0;
#ifdef X_XFACE
  cf->go                   = NULL;
#endif

  return cf;
}

static int
elem_elcode (W_elem_t *el)
{
  return PCONT4_ELCODE;
}


static char *
type_name (W_elem_t *el)
{
  static char buf[] = "pcont4";
  return buf;
}

#define GET_COORD(n, vec) WN_get_coords (n, &(vec.x), &(vec.y), &(vec.z))
#define GET_PREV_COORD(n, vec) WN_get_prev_coords (n, &(vec.x), &(vec.y), &(vec.z))
#define GET_VELOC(n, vec) WN_get_velocity (n, &(vec.x), &(vec.y), &(vec.z))


static void
get_cp_coord (contact_point_t *cp,
              pcont4_t *cf,
              W_coords_3d_t *loc)
{
  W_coords_3d_t nl[3];
  GET_COORD (cf->n[0], nl[0]);
  GET_COORD (cf->n[1], nl[1]);
  GET_COORD (cf->n[2], nl[2]);
  VECMAC_2_VEC ((*loc),  =, cp->area_coords[0] * nl[0]);
  VECMAC_2_VEC ((*loc), +=, cp->area_coords[1] * nl[1]);
  VECMAC_2_VEC ((*loc), +=, cp->area_coords[2] * nl[2]);
}

static void
get_cp_prev_coord (contact_point_t *cp,
                   pcont4_t *cf,
                   W_coords_3d_t *loc)
{
  W_coords_3d_t nl[3];
  GET_PREV_COORD (cf->n[0], nl[0]);
  GET_PREV_COORD (cf->n[1], nl[1]);
  GET_PREV_COORD (cf->n[2], nl[2]);
  VECMAC_2_VEC ((*loc),  =, cp->area_coords[0] * nl[0]);
  VECMAC_2_VEC ((*loc), +=, cp->area_coords[1] * nl[1]);
  VECMAC_2_VEC ((*loc), +=, cp->area_coords[2] * nl[2]);
}

static void
get_cp_vel (contact_point_t *cp,
              pcont4_t *cf,
              W_velocity_3d_t *vel)
{
  W_velocity_3d_t vl[3];
  GET_VELOC (cf->n[0], vl[0]);
  GET_VELOC (cf->n[1], vl[1]);
  GET_VELOC (cf->n[2], vl[2]);
  VECMAC_2_VEC ((*vel),  =, cp->area_coords[0] * vl[0]);
  VECMAC_2_VEC ((*vel), +=, cp->area_coords[1] * vl[1]);
  VECMAC_2_VEC ((*vel), +=, cp->area_coords[2] * vl[2]);
}

static int
candidate_faces (WPC_penalty_contact_t *pc,
                 pcont4_t *cf,
                 pcont4_t *candcf[],
                 int maxcf
                 )
{
  int i;
  pcont4_t *acf;

  lobs_3d_search (pc->lobs,
                  cf->box.lower.x, cf->box.lower.y, cf->box.lower.z,
                  cf->box.upper.x, cf->box.upper.y, cf->box.upper.z);

  acf = (pcont4_t *) lobs_3d_first_overlapping_object (pc->lobs);
  i = 0;
  while (acf) {
    if (acf != cf) {     /* not the oneself */
      if (i == maxcf) break;
      candcf[i++] = acf;
    }
    acf = (pcont4_t *) lobs_3d_next_overlapping_object (pc->lobs);
  }
  
  return i; 
}


static void
compute_area_coords (W_coords_3d_t cfl[3],
                     W_coords_3d_t *nl,
                     G3D_plane_t *plane,
                     double L[3])
{
  double x0, y0, x1, y1, x2, y2, x, y;
  
  x0 = VECMAC_DOT (plane->u, cfl[0]);
  y0 = VECMAC_DOT (plane->v, cfl[0]);
  x1 = VECMAC_DOT (plane->u, cfl[1]);
  y1 = VECMAC_DOT (plane->v, cfl[1]);
  x2 = VECMAC_DOT (plane->u, cfl[2]);
  y2 = VECMAC_DOT (plane->v, cfl[2]);

  x  = VECMAC_DOT (plane->u, (*nl));
  y  = VECMAC_DOT (plane->v, (*nl));

  G3D_tri_inv_map (x, y, x0, y0, x1, y1, x2, y2, &L[0], &L[1], &L[2]);
}


typedef struct proj_t {
  W_bool_t  projection_is_in;
  double    dist_to_plane;
}              proj_t;

static void
push (contact_point_t *cp,
      W_coords_3d_t *loc,
      pcont4_t *owning_cf,
      double overlap, W_gap_penalty_type force_type, double avgnmass,
      double overlap_penalty, double overlap_rate_penalty)
{
  W_vect_t of = {0,0,0}, orf = {0,0,0}, f, cff;
  W_velocity_3d_t v, cfv[3], xv;
  W_coords_3d_t cfl[3];
  double L[3];
  double overlap_rate;
  double massr; 
  double angle_factor;
  double total_factor;

  GET_COORD (cp->in_contact_cf->n[0], cfl[0]);
  GET_COORD (cp->in_contact_cf->n[1], cfl[1]);
  GET_COORD (cp->in_contact_cf->n[2], cfl[2]);
  
  compute_area_coords (cfl, loc, &cp->in_contact_cf->plane, L);
  
  if (force_type == W_PENALTY_OVERLAP_RATE ||
      force_type == W_PENALTY_OVERLAP_COMB) {
    get_cp_vel (cp, owning_cf, &v);
    GET_VELOC (cp->in_contact_cf->n[0], cfv[0]);
    GET_VELOC (cp->in_contact_cf->n[1], cfv[1]);
    GET_VELOC (cp->in_contact_cf->n[2], cfv[2]);
    VECMAC_2_VEC (xv,  =, L[0] * cfv[0]);
    VECMAC_2_VEC (xv, +=, L[1] * cfv[1]);
    VECMAC_2_VEC (xv, +=, L[2] * cfv[2]);
    overlap_rate
      = (- (VECMAC_DOT (v, cp->in_contact_cf->plane.n))
         + (VECMAC_DOT (xv, cp->in_contact_cf->plane.n)));
    VECMAC_2_VEC (orf, =, overlap_rate * overlap_rate_penalty *
                           cp->in_contact_cf->plane.n);
  }

  if (force_type == W_PENALTY_OVERLAP ||
      force_type == W_PENALTY_OVERLAP_COMB) {
    VECMAC_2_VEC (of, =, (overlap) * overlap_penalty *
                         cp->in_contact_cf->plane.n);
  }

  /* force */
  VECMAC_3_VEC (f, =, of, +, orf);

  /* mass ratio */
  massr = (WN_nodal_mass (owning_cf->n[0]) +
           WN_nodal_mass (owning_cf->n[1]) +
           WN_nodal_mass (owning_cf->n[2])) / 3 / avgnmass;

  /* angle factor */
  angle_factor
    = fabs (VECMAC_DOT (owning_cf->plane.n, cp->in_contact_cf->plane.n));

  /* scale by mass ratio, angle factor and area corresponding to the cp */
  total_factor = massr * angle_factor * owning_cf->cp_area;
  
  /* scale force by total factor */
  VECMAC_SCALAR (f, *=, total_factor);
  fprintf (stderr, "f={%g,%g,%g}\n", f.x, f.y, f.z);
  
  /* assemble to nodes of the owning face */
  VECMAC_2_VEC (cff, =, cp->area_coords[0] * f);
  WN_assemble_force (owning_cf->n[0], &cff);
  VECMAC_2_VEC (cff, =, cp->area_coords[1] * f);
  WN_assemble_force (owning_cf->n[1], &cff);
  VECMAC_2_VEC (cff, =, cp->area_coords[2] * f);
  WN_assemble_force (owning_cf->n[2], &cff);

  /* assemble to nodes of the contacting face */
  VECMAC_2_VEC (cff, =, (-L[0]) * f);
  WN_assemble_force (cp->in_contact_cf->n[0], &cff);
  VECMAC_2_VEC (cff, =, (-L[1]) * f);
  WN_assemble_force (cp->in_contact_cf->n[1], &cff);
  VECMAC_2_VEC (cff, =, (-L[2]) * f);
  WN_assemble_force (cp->in_contact_cf->n[2], &cff);
}


static void
project_on_cf (W_coords_3d_t *nl,
                         pcont4_t *cf,
                         double eps, proj_t *proj_data)
{
  G3D_loc_t intersection;
  double lambda;
  G3D_tri_t tri;
  W_loc_t d;
  
  WN_get_coords (cf->n[0], &(tri.loc[0].x), &(tri.loc[0].y), &(tri.loc[0].z));
  WN_get_coords (cf->n[1], &(tri.loc[1].x), &(tri.loc[1].y), &(tri.loc[1].z));
  WN_get_coords (cf->n[2], &(tri.loc[2].x), &(tri.loc[2].y), &(tri.loc[2].z));

  if (G3D_intersect_tri_and_line (&intersection, &lambda,
                                  nl, &cf->plane.n, &tri, eps)) {
    VECMAC_3_VEC (d, =, (*nl), -, intersection);
    proj_data->dist_to_plane    = VECMAC_DOT (d, cf->plane.n);
    proj_data->projection_is_in = TRUE;
  } else {
    proj_data->dist_to_plane    = 0;
    proj_data->projection_is_in = FALSE;
  }
}

static int
intersect_path_with_cf (W_loc_t *loc, W_loc_t *prevloc, pcont4_t *cf,
                        double eps,
                        proj_t *proj_data)
{
  G3D_loc_t intersection;
  double lambda;
  G3D_tri_t tri;
  G3D_loc_t prev_nl[3];
  G3D_loc_t nl[3];
  W_loc_t d;
  G3D_plane_t plane;
  G3D_loc_t l_prev_hat, l_hat;
  int i;
  
  /* transform the previous location into the coordinate system carried
     along by the face at the previous time
   */
  WN_get_prev_coords (cf->n[0], &(prev_nl[0].x), &(prev_nl[0].y), &(prev_nl[0].z));
  WN_get_prev_coords (cf->n[1], &(prev_nl[1].x), &(prev_nl[1].y), &(prev_nl[1].z));
  WN_get_prev_coords (cf->n[2], &(prev_nl[2].x), &(prev_nl[2].y), &(prev_nl[2].z));

  G3D_plane_consts_3_alt (&plane, &(prev_nl[0]), &(prev_nl[1]), &(prev_nl[2]));
  VECMAC_3_VEC (d, =, (*prevloc), -, prev_nl[0]);
  l_prev_hat.x = VECMAC_DOT (d, plane.u);
  l_prev_hat.y = VECMAC_DOT (d, plane.v);
  l_prev_hat.z = VECMAC_DOT (d, plane.n);
  
  /* the same transformation at the current time */
  WN_get_coords (cf->n[0], &(nl[0].x), &(nl[0].y), &(nl[0].z));
  WN_get_coords (cf->n[1], &(nl[1].x), &(nl[1].y), &(nl[1].z));
  WN_get_coords (cf->n[2], &(nl[2].x), &(nl[2].y), &(nl[2].z));

  G3D_plane_consts_3_alt (&plane, &(nl[0]), &(nl[1]), &(nl[2]));
  VECMAC_3_VEC (d, =, (*loc), -, nl[0]);
  l_hat.x = VECMAC_DOT (d, plane.u);
  l_hat.y = VECMAC_DOT (d, plane.v);
  l_hat.z = VECMAC_DOT (d, plane.n);
  
  /* fing intersection of the segment <l_hat ; l_prev_hat> with the current face;
     to do that we need to tranform the triangle vertices to the same coordinate
     system
  */
  tri.loc[0].x = /* the first vertex is in the origin */
  tri.loc[0].y = 
  tri.loc[0].z = 0;
  for (i = 1; i < 3; i++) {
    VECMAC_3_VEC (d, =, nl[i], -, nl[0]);
    tri.loc[i].x = VECMAC_DOT (d, plane.u);
    tri.loc[i].y = VECMAC_DOT (d, plane.v);
    tri.loc[i].z = VECMAC_DOT (d, plane.n);
  }

  /* now find an intersection of the segment with the face */
  VECMAC_3_VEC (d, =, l_hat, -, l_prev_hat);
  if (G3D_intersect_tri_and_seg (&intersection, &lambda, &l_prev_hat, &d, &tri, eps)) {
    proj_data->dist_to_plane    = l_hat.z;
    proj_data->projection_is_in = TRUE;
  } else {
    proj_data->dist_to_plane    = 0;
    proj_data->projection_is_in = FALSE;
  }
  return TRUE;
}

static int
left_cf (W_loc_t *loc, W_loc_t *prevloc, pcont4_t *cf,
            double eps,
            proj_t *proj_data)
{
  intersect_path_with_cf (loc, prevloc, cf, eps, proj_data);
  return (proj_data->projection_is_in && proj_data->dist_to_plane >= 0);
}

static int
entered_cf (W_loc_t *loc, W_loc_t *prevloc, pcont4_t *cf,
            double eps, 
            proj_t *proj_data)
{
  intersect_path_with_cf (loc, prevloc, cf, eps, proj_data);
  return (proj_data->projection_is_in && proj_data->dist_to_plane < 0);
}

static int
share_node (pcont4_t *in_contact_cf,
            pcont4_t *cf)
{
  return (in_contact_cf->n[0] == cf->n[0] ||
          in_contact_cf->n[0] == cf->n[1] ||
          in_contact_cf->n[0] == cf->n[2] ||
          in_contact_cf->n[1] == cf->n[0] ||
          in_contact_cf->n[1] == cf->n[1] ||
          in_contact_cf->n[1] == cf->n[2] ||
          in_contact_cf->n[2] == cf->n[0] ||
          in_contact_cf->n[2] == cf->n[1] ||
          in_contact_cf->n[2] == cf->n[2]);
}

static void
cp_on_cf (contact_point_t *cp,
          pcont4_t *cf,
          pcont4_t *candcf[],
          proj_t   projection[],
          int nf,
          WPC_penalty_contact_t *pc)
{
 
  int i;
  W_coords_3d_t loc, prevloc;
  int hit;
  W_bool_t do_search = TRUE;
  double mindist;
  int mini;
#define ARE_NEIGHBORS(cf0, cf1) share_node(cf0, cf1)

  get_cp_coord (cp, cf, &loc);
  get_cp_prev_coord (cp, cf, &prevloc);

  hit       = -1; /* assume no hit */
  do_search = TRUE; /* assume we need to search */

  if (cp->in_contact_cf) {      /* if previously in contact */
    project_on_cf (&loc, cp->in_contact_cf, MACHEPS, &projection[0]);
    if (projection[0].projection_is_in) {
      if (projection[0].dist_to_plane < 0) {
        /* contact with the same face */
        candcf[0]     = cp->in_contact_cf;
        nf            = 1;
        hit           = 0;        /* hit with the previous face */
        do_search     = FALSE;
        fprintf (stderr, "hit on previous\n");
      } else {
        if (left_cf (&loc, &prevloc, cp->in_contact_cf,
                     MACHEPS, &projection[0])) {
          cp->in_contact_cf = NULL;
          fprintf (stderr, "left\n");
        } 
      }
    }
  }

  if (do_search) {
    if (cp->in_contact_cf) { /* was in contact, but not now */
      mindist = -FLT_MAX;
      mini    = -1;
      FOR (i, nf) {
        project_on_cf (&loc, candcf[i], MACHEPS, &projection[i]);
        if (projection[i].projection_is_in) { /* point projects inside ... */
          if (projection[i].dist_to_plane < 0) {/* and it is an overlap ... */
            if (projection[i].dist_to_plane > mindist) {
              /* and it is a smaller overlap... */
              if (ARE_NEIGHBORS (cp->in_contact_cf, candcf[i])) {
                /* and the hit face is adjacent to previous contact face */
                hit = i; /* hit */
                fprintf (stderr, "neighbor\n");
                mindist = projection[i].dist_to_plane;
              }
            }
          } else if (projection[i].dist_to_plane >= 0) { /* gap */
            hit = -1; /* there is a gap with one face; no contact possible */
            break;
          }
        } /* if projection in */
      } /* FOR */
    } else { /* not in contact previously */
      FOR (i, nf) {
        if (entered_cf (&loc, &prevloc, candcf[i], MACHEPS, &projection[i])) {
          hit = i; /* hit */
          break;
        } 
      }
    }
  }

  if (hit >= 0 && hit < nf) {
    double overlap = fabs (projection[hit].dist_to_plane);
    cp->in_contact_cf = candcf[hit];
    push (cp, &loc, cf, overlap, pc->force_type, pc->avgnmass,
          pc->overlap_penalty, pc->overlap_rate_penalty);
  } else {
    cp->in_contact_cf = NULL;   /* no contact */
  }
}


static void
cf_on_cf (pcont4_t *cf, WPC_penalty_contact_t *pc)
{
  #define MAX_CF 1000
  pcont4_t *candcf[MAX_CF];
  proj_t projection[MAX_CF];
  int i, nf;

  nf = candidate_faces (pc, cf, candcf, MAX_CF);
  FOR (i, cf->ncp) {
    cp_on_cf (&cf->cp[i], cf, candcf, projection, nf, pc);
  }
}


static void
geom (pcont4_t *cf)
{
  W_loc_t loc[3];
  WN_get_coords (cf->n[0], &(loc[0].x), &(loc[0].y), &(loc[0].z));
  WN_get_coords (cf->n[1], &(loc[1].x), &(loc[1].y), &(loc[1].z));
  WN_get_coords (cf->n[2], &(loc[2].x), &(loc[2].y), &(loc[2].z));
  G3D_plane_consts_3 (&cf->plane, &loc[0], &loc[1], &loc[2]);
  G3D_INIT_BOX (cf->box, loc[0]);
  G3D_UPDT_BOX (cf->box, loc[1]);
  G3D_UPDT_BOX (cf->box, loc[2]);
  WN_get_prev_coords (cf->n[0], &(loc[0].x), &(loc[0].y), &(loc[0].z));
  WN_get_prev_coords (cf->n[1], &(loc[1].x), &(loc[1].y), &(loc[1].z));
  WN_get_prev_coords (cf->n[2], &(loc[2].x), &(loc[2].y), &(loc[2].z));
  G3D_UPDT_BOX (cf->box, loc[0]);
  G3D_UPDT_BOX (cf->box, loc[1]);
  G3D_UPDT_BOX (cf->box, loc[2]);
}

static void
assemble_restoring_forces (W_elem_t *el)
{
  pcont4_t *cf = (pcont4_t *) el->data;
  WPC_penalty_contact_t *pc = cf->cs->data.penalty;
  geom (cf);
  cf_on_cf (cf, pc);
}

#ifdef X_XFACE
static EPixel white, slateblue, red;
static W_bool_t first = TRUE;
static BOOLEAN suc;

static void
boot_grep ()
{
  slateblue = ColorGetPixelFromString ("slateblue", &suc);
  if (!suc) REPORT_FATAL_ERROR ("No such color");
  white     = ColorGetPixelFromString ("white", &suc);
  if (!suc) REPORT_FATAL_ERROR ("No such color");
  red       = ColorGetPixelFromString ("red", &suc);
  if (!suc) REPORT_FATAL_ERROR ("No such color");
  first = FALSE;
}

static void
modify_elem_grep (W_elem_t *el, W_bool_t immed_draw)
{
  pcont4_t *cf = (pcont4_t *) el->data;
  W_coords_3d_t loc[3];
  WCRec pts[3];
  W_bool_t is_rigid = !(cf->n[0]->can_move_in_dir[0] &&
                        cf->n[0]->can_move_in_dir[1] &&
                        cf->n[0]->can_move_in_dir[2]);
  int i;

  if (first) boot_grep ();

  WN_get_coords (cf->n[0], &(loc[0].x), &(loc[0].y), &(loc[0].z));
  WN_get_coords (cf->n[1], &(loc[1].x), &(loc[1].y), &(loc[1].z));
  WN_get_coords (cf->n[2], &(loc[2].x), &(loc[2].y), &(loc[2].z));
  pts[0] = loc[0];
  pts[1] = loc[1];
  pts[2] = loc[2];

  if (!cf->go) {
    cf->go = CreateTriangle3D (pts);
    if (is_rigid) EASValsSetColor (slateblue);
    else          EASValsSetColor (white);
    if (is_rigid) EASValsSetLayer (WX_RIGID_CF_LAYER);  
    else          EASValsSetLayer (WX_NATURAL_CF_LAYER);  
    EASValsSetFillStyle (FILL_SOLID);
    EGWithMaskChangeAttributes (COLOR_MASK | FILL_MASK | LAYER_MASK, cf->go);
    EMNoDrawAddGraphicsToModel (ESIModel (), cf->go);
  } else {
    EGModifyGraphicsGeometry (cf->go, pts);
  }

  FOR (i, cf->ncp) {
    WCRec mpts[1];

    VECMAC_2_VEC (mpts[0],  =, cf->cp[i].area_coords[0] * pts[0]);
    VECMAC_2_VEC (mpts[0], +=, cf->cp[i].area_coords[1] * pts[1]);
    VECMAC_2_VEC (mpts[0], +=, cf->cp[i].area_coords[2] * pts[2]);
    if (!cf->cp[i].go) {
      cf->cp[i].go = CreateMarker3D (mpts);
      if (is_rigid) EASValsSetColor (slateblue);
      else          EASValsSetColor (white);

      if (is_rigid) EASValsSetLayer (WX_RIGID_CF_LAYER);  
      else          EASValsSetLayer (WX_NATURAL_CF_LAYER);  
      EASValsSetMSize (6);
      EASValsSetMType (X_LETTER_MARKER);
      EGWithMaskChangeAttributes (COLOR_MASK |
                                  LAYER_MASK |
                                  MSIZE_MASK |
                                  MTYPE_MASK, cf->cp[i].go);
      EMNoDrawAddGraphicsToModel (ESIModel (), cf->cp[i].go);
    } else {
      if (cf->cp[i].in_contact_cf) {
        EASValsSetColor (red);
      } else {
        if (is_rigid) EASValsSetColor (slateblue);
        else          EASValsSetColor (white);
      }
      EGWithMaskChangeAttributes (COLOR_MASK, cf->cp[i].go);
      EGModifyGraphicsGeometry (cf->cp[i].go, mpts);
    }
  }
}
#endif


static int 
num_of_faces (W_elem_t *el)
{
  return 1;
}

static int 
get_face_vertex_nodes (W_elem_t *el, int nface, W_node_t *n[4])
{
  pcont4_t *c = (pcont4_t *) el->data;
  int i;
  FOR (i, 3) {
    n[i] = c->n[i];
  }
  return 3;
}


static W_elem_methods_t pcont4_methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  num_of_faces,
  get_face_vertex_nodes,
  NULL
};

W_bool_t 
pcont4_reader (W_domain_t *d, FILE *stream)
{
  const char *input_line;
  long elem_num;
  W_node_t *n[3];
  pcont4_t *c;
  W_elem_t *el;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  int i;
  W_index_t nodeid, cont_surf_id;
  W_contact_surf_t *cs;

  while ((input_line = W_get_input_line(stream)) != NULL) {
    if (input_line[0] == END_OF_BLOCK) {
      return TRUE;
    } else {
      TOKENS_PARSE(input_line);
      elem_num = TOKEN_INT(1);
      FOR (i, 3) {
        nodeid = TOKEN_INT(2+i);
        n[i] = WD_get_node_by_idx (d, nodeid);
        if (n[i] == NULL) {
          sprintf(W_buf(), "No such node: %ld", nodeid);
          REPORT_ERROR(W_buf());
        }
      }
      c = new_contact_face (n[0], n[1], n[2], 0);
      cont_surf_id = TOKEN_INT (5);
      cs = WD_get_cont_surf_by_idx (d, cont_surf_id);
      if (cs == NULL) {
        sprintf(W_buf(), "No such contact surface: %ld", cont_surf_id);
        REPORT_ERROR(W_buf());
      }
      c->cs = cs;
      if (cs->type != W_PENALTY_CONTACT_SURF) {
        sprintf(W_buf(), "Element %d must belong to a penalty contact",
                (int) elem_num);
        REPORT_FATAL_ERROR (W_buf ());
      }
      subdiv_contact_face (c, cs->data.penalty->subdiv_level);
      WD_add_element (d, elem_num,
                      el = WE_new_element (d, &pcont4_methods, c
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                           , is_ghost
#                                   endif
                                           ));
      WC_add_cont_elem (cs, el);
    } /* closing else */ 
  }
  return FALSE;
}
