#include "Wconfig.h"
#include "Wtypes.h" 
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wpss.h"
#include "Wgaussp.h"
#include "Wxf.h"
#include "Wcontface.h"
#include "Wtfuncs.h"
#include "Wq4press.h"

/* RCSHEADER "$Id: q4press.c,v 1.1.1.1 2001/07/24 02:42:52 pkrysl Exp $" */

#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

typedef struct priv_elem_t {
  W_node_t         *node[4];
  double            pressure;
  W_time_func_t    *tf;
}              priv_elem_t;

static int
elem_elcode (W_elem_t *el)
{
  return Q4PRESS_ELCODE;
}

static char *
type_name(W_elem_t *el)
{
  static char buf[] = "q4press";
  return buf;
}

static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  W_coords_3d_t c[4];
  priv_elem_t *t = (priv_elem_t *)el->data;
  double mult;
  W_vect_t f, txi, teta, n={0,0,0};
  double tfv = WTF_func_val (t->tf, WD_time(el->domain));
  double n_xi[4], n_eta[4];
  int i;

  /* kinematic quantities */
#define GET_FROM_NODE_ALL(i) { GET_FROM_NODE (t->node[i], WN_get_coords, c[i]); }
  GET_FROM_NODE_ALL(0);
  GET_FROM_NODE_ALL(1);
  GET_FROM_NODE_ALL(2);
  GET_FROM_NODE_ALL(3);
#undef GET_FROM_NODE_ALL

  /* compute normal */
  G3D_q4_N_xi  ( n_xi, 0, 0);
  G3D_q4_N_eta (n_eta, 0, 0);
  VECMAC_SCALAR (txi, =, 0); 
  VECMAC_SCALAR (teta, =, 0);
  FOR (i, 4) {
    VECMAC_2_VEC ( txi, +=,  n_xi[i] * c[i]);
    VECMAC_2_VEC (teta, +=, n_eta[i] * c[i]);
  }
  VECMAC_CROSS (n, =, txi, x, teta); /* vector of magnitude equal to 1/4 of the area */
  
  /* compute force and assemble */
  mult = tfv * 4 * t->pressure / 4; /* first 4: see above;
                                       second 4: each node gets one fourth */
  VECMAC_2_VEC (f, =, mult * n);
  WN_assemble_force (t->node[0], &f);
  WN_assemble_force (t->node[1], &f);
  WN_assemble_force (t->node[2], &f);
  WN_assemble_force (t->node[3], &f);
  
  /* deformation energy */
  el->deformation_energy = 0;   /* RAW */
}

#ifdef X_XFACE
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  priv_elem_t *h = (priv_elem_t *)el->data;
  int i;
  W_vect_t refc, u;
  WCRec pnts[4];
 
  for (i = 0; i < 4; i++) {
    GET_FROM_NODE (h->node[i], WN_get_ref_coords, refc);
    GET_FROM_NODE (h->node[i], WN_get_displ, u);
    pnts[i].x = refc.x + displ_scale * u.x;
    pnts[i].y = refc.y + displ_scale * u.y;
    pnts[i].z = refc.z + displ_scale * u.z;
  }

  if (el->grep == NULL) {
    EPixel white;
    BOOLEAN suc;
    white = ColorGetPixelFromString ("white", &suc);
    if (!suc) white = 1;
    el->grep = CreateQuad3D (pnts);
    EASValsSetColor (white);
    EASValsSetLayer (WX_ELEM_LAYER);
    EASValsSetFillStyle (FILL_HOLLOW);
    EASValsSetShrink (0.9);
    EGWithMaskChangeAttributes(COLOR_MASK | FILL_MASK | LAYER_MASK | SHRINK_MASK,
                               el->grep);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
  } else {
    EGModifyGraphicsGeometry(el->grep, pnts);
  }
}
#endif

static void
assemble_residual_forces (W_elem_t *el,
                          W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  assemble_restoring_forces (el, update_form);
}

static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                        W_mass_type_enum mass_type
                        )
{
  /* RAW need to code the pressure stiffness */
}

static W_elem_methods_t methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  assemble_residual_forces,
  assemble_eff_stiffness
};

/* Reads element data from stream as long as there is no 
   END_OF_BLOCK tag in the input line. Creates elements. */
W_bool_t 
q4press_reader (W_domain_t *d)
{
  const char *input_line;
  long elem_num;
  priv_elem_t *t;
  int i, nodeid;
  int tfnum;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      t = w_new (priv_elem_t);
      CHECK_MALLOC(t);
      TOKENS_PARSE(input_line);
      elem_num             = TOKEN_INT(1);
      FOR (i, 4) {
        nodeid = TOKEN_INT(2+i);
        t->node[i] = WD_get_node_by_idx (d, nodeid);
        if (t->node[i] == NULL) {
          sprintf (W_buf (), "No such node: %d", (int)nodeid);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
      }
      tfnum       = TOKEN_INT (6);
      t->tf       = WD_get_time_func_by_idx(d, tfnum);
      if (!t->tf) {
        sprintf (W_buf (), "No such time function: %d", (int)tfnum);
        REPORT_FATAL_ERROR (d->errh, W_buf ());
      }
      t->pressure = TOKEN_DBL(7);
#ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif      
      WD_add_element(d, elem_num,
                     WE_new_element(d, &methods, t
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                    , is_ghost
#                                   endif
                                    ));
    }
  }
  return FALSE;
}

