#include "Wconfig.h"
#include "Wxf.h"
#include "Wqmem4c.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wgaussp.h"
#include "Wpss.h"
#include "Wmat3dutil.h"

/* RCSHEADER "$Id: qmem4c.c,v 1.1.1.1 2001/07/24 02:42:52 pkrysl Exp $" */

#define INTEGRATION_ORDER 2

#define SHOW_AXES 0 /* define to TRUE for graphic debugging */

typedef struct priv_elem_t {
  W_node_t         *node[4];
  W_float_t         thickness;
  W_index_t         material;
  W_matstate_t      matstate[INTEGRATION_ORDER][INTEGRATION_ORDER];
#if SHOW_AXES
  GraphicObj *conv_axes[2];
  GraphicObj *rot_axes[2];
#endif
}              priv_elem_t;

static int
elem_elcode (W_elem_t *el)
{
  return QMEM4C_ELCODE;
}

static char *
type_name(W_elem_t *el)
{
  static char buf[] = "qmem4c";
  return buf;
}

#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

static W_float_t
total_mass (W_elem_t *el)
{
  W_float_t a1, a2;
  W_float_t volume_mass;
  W_coords_3d_t c[4];
  priv_elem_t *t;
  W_material_t *m;
  int i;

  t   = (priv_elem_t *)el->data;

  FOR (i, 4) {
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);
  }
  
  m = WD_get_material_by_idx(el->domain, t->material);
  volume_mass = WM_get_volume_mass(m);

  a1 = G3D_tri2d_area (c[0].x, c[0].y, c[1].x, c[1].y, c[2].x, c[2].y);
  a2 = G3D_tri2d_area (c[0].x, c[0].y, c[2].x, c[2].y, c[3].x, c[3].y);
  if (a1 < 0 || a2 < 0) {
    REPORT_FATAL_ERROR (el->domain->errh, "bad area");
  }

  return (a1 + a2) * volume_mass * t->thickness;
}

static W_float_t 
suggested_time_step (W_elem_t *el)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_coords_3d_t c[4];
  W_float_t y24, y31, x24, x31, BiIBiI, E, nu, G, A;
  W_float_t lambda, rho, tot_mass;
  W_material_t *m;
  int i;

  FOR (i, 4) {
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);
  }

  x31 = c[2].x - c[0].x;
  y31 = c[2].y - c[0].y;
  x24 = c[1].x - c[3].x;
  y24 = c[1].y - c[3].y;
  BiIBiI = (x31*x31 + y31*y31 + x24*x24 + y24*y24) / 2; 

  m = WD_get_material_by_idx(el->domain, t->material);
  if (!WM_get_uniax_lin_el_consts(m, &E, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No uniax_lin_el_consts");
  tot_mass = total_mass (el);
  rho      = WM_get_volume_mass(m);
  A        = tot_mass / rho / t->thickness;
  lambda   = nu * E / ((1 + nu)*(1 - 2*nu));
  G        = E / (2 * (1 + nu));

  /* Formula identical to: Flanagan, Belytschko 1981 */
  return A * sqrt(rho / ((lambda + 2*G)*BiIBiI));
}

static void
assemble_mass (W_elem_t *el)
{
  W_float_t m;
  priv_elem_t *t = (priv_elem_t *)el->data;
  
  m = total_mass(el) / 4;

  WN_assemble_mass (t->node[0], m);
  WN_assemble_mass (t->node[1], m);
  WN_assemble_mass (t->node[2], m);
  WN_assemble_mass (t->node[3], m);
}

typedef struct geom_desc_t {
  W_vect_t at;
  W_vect_t covariant[3];          /* Gaussian intrinsic frame */
  W_vect_t contravariant[3];      /* dual frame */
  double j0_bar;
  G3D_plane_t plane;
  W_vect_t corot_axes[2];
}              geom_desc_t;

static void
def_grad (geom_desc_t *gd, double F[3][3])
{
  int alpha, i, j;
  double *covv;
  FOR (j, 3) {
    covv = &gd->covariant[j].x;
    FOR (i, 3) {
      F[i][j] = covv[i];
    }
  }
}

static void
corot_axes (geom_desc_t *gd)
{
  double F[3][3], R[3][3], U[3][3];
  def_grad (gd, F);
  if (!W_polar_decomposition (F, R, U))
    REPORT_FATAL_ERROR (NULL, "Polar decomposition routine failed");
  VECMAC_3_SCALAR (gd->corot_axes[0], =, R[0][0], R[1][0], R[2][0]);
  VECMAC_3_SCALAR (gd->corot_axes[1], =, R[0][1], R[1][1], R[2][1]);
}

static void
compute_geom_desc (W_vect_t c[4], double xi, double eta, geom_desc_t *gd)
{
  double n[4], n_xi[4], n_eta[4];
  double j0_bar_inv;
  int i;

  G3D_q4_N     (n,  xi, eta);
  G3D_q4_N_xi  (n_xi,  xi, eta);
  G3D_q4_N_eta (n_eta, xi, eta);

  /* tangent vectors (Gaussian intrinsic frame) */
  VECMAC_SCALAR (gd->at, =, 0);
  VECMAC_SCALAR (gd->covariant[0], =, 0);
  VECMAC_SCALAR (gd->covariant[1], =, 0);
  FOR (i, 4) {
    VECMAC_2_VEC (gd->at, +=,  n[i] * c[i]);
    VECMAC_2_VEC (gd->covariant[0], +=,  n_xi[i] * c[i]);
    VECMAC_2_VEC (gd->covariant[1], +=, n_eta[i] * c[i]);
  }

  /* Jacobian */
  {
    W_vect_t n={0,0,0}, p, q, r;
    VECMAC_CROSS (n, =, gd->covariant[0], x, gd->covariant[1]);
    gd->j0_bar = VECMAC_NORM (n); 
    /* tangent plane */
    VECMAC_2_VEC (q, =, gd->at);
    VECMAC_3_VEC (p, =, q, +, gd->covariant[0]);
    VECMAC_3_VEC (r, =, p, -, gd->covariant[1]);
    G3D_plane_consts_3_alt (&gd->plane, &p, &q, &r);
  }
  gd->covariant[2] = gd->plane.n;

  /* contravariant vectors */
  j0_bar_inv = 1 / gd->j0_bar;
  VECMAC_CROSS (gd->contravariant[0], =, gd->covariant[1], x, gd->plane.n);
  VECMAC_SCALAR (gd->contravariant[0], *=, j0_bar_inv);
  VECMAC_CROSS (gd->contravariant[1], =, gd->plane.n, x, gd->covariant[0]);
  VECMAC_SCALAR (gd->contravariant[1], *=, j0_bar_inv);
  gd->contravariant[2] = gd->covariant[2];

  /* co-rotation axes */
  corot_axes (gd);
}

#if 0
static void
def (W_elem_t *el)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  geom_desc_t gd;
  double cC[2][2];
  int i, j;
  W_vect_t c[4];
 
  FOR (i, 4) { GET_FROM_NODE (pe->node[i], WN_get_coords, c[i]); }
  compute_geom_desc (c, 0, 0, &gd);
  /* classical Cauchy-Green deformation */
  FOR (i, 2) {
    FOR (j, 2) {
      cC[i][j] = VECMAC_DOT (gd.covariant[i], gd.covariant[j]);
    }
  }
  /* deformation gradient in the co-rotational basis */
  
}
#endif

static void
proj (W_vect_t c[4], geom_desc_t *gd, double x[4], double y[4])
{
  W_vect_t v;
  int i;
  FOR (i, 4) {
    VECMAC_3_VEC (v, =, c[i], -, gd->at);
    x[i] = VECMAC_DOT (v, gd->corot_axes[0]);
    y[i] = VECMAC_DOT (v, gd->corot_axes[1]);
  }
}

static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  W_coords_3d_t c[4], force_vect, refc[4];
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_material_t *m;
  W_float_t th, eps[3] = {0, 0, 0};
  W_float_t J[2][2], inv_J[2][2], inv_J00[2][2], u[8], B[3][8];
  W_float_t jdetw, eps1_rate = 0, eps2_rate = 0, gamma_rate = 0;
  W_float_t sigma[3], F[8] = {0, 0, 0, 0, 0, 0, 0, 0};
  W_gauss_integration_rule_t *gr;
  W_gauss_integration_point_t *ip;
  int i, j;
  double x[4], y[4], X[4], Y[4];
  geom_desc_t cgd, rgd;
#define B_TIMES_U(theE, theB, theU)      \
  {                                      \
    int nn;                              \
    theE[0] = theE[1] = theE[2] = 0;     \
    for (nn = 0; nn < 8; nn++) {         \
      theE[0] += theB[0][nn] * theU[nn]; \
      theE[1] += theB[1][nn] * theU[nn]; \
      theE[2] += theB[2][nn] * theU[nn]; \
    }                                    \
  }

#define ADD_BT_TIMES_SIGMA(theF, theB, theS)                                  \
  {                                                                           \
    int nn;                                                                   \
    for (nn = 0; nn < 8; nn++) {                                              \
      theF[nn] +=                                                             \
        (theB[0][nn]*theS[0]+theB[1][nn]*theS[1]+theB[2][nn]*theS[2]);        \
    }                                                                         \
  }

  gr = WG_get_integration_rule(INTEGRATION_ORDER);
  ip = gr->ip;
  
  m = WD_get_material_by_idx(el->domain, t->material);
  if (t->matstate[0][0] == NULL) {
    for (i = 0; i < INTEGRATION_ORDER; i++) 
      for (j = 0; j < INTEGRATION_ORDER; j++) 
        if (t->matstate[i][j] == NULL)
          t->matstate[i][j] = WM_get_mat_state(m);
  }
  
  /* kinematic quantities */
  FOR (i, 4) {
    GET_FROM_NODE (t->node[i], WN_get_coords, c[i]);            
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]);     
  }

  compute_geom_desc (c, 0, 0, &cgd);
  compute_geom_desc (refc, 0, 0, &rgd);
  proj (c, &cgd, x, y);
  proj (refc, &rgd, X, Y);
  /* displacements in the local frame */
  FOR (i, 4) { u[2*i] = x[i] - X[i]; u[2*i+1] = y[i] - Y[i]; }

  /* strain */
  pss_jacobian_matrix (J, 0.0, 0.0, x[0], y[0], x[1], y[1], x[2], y[2], x[3], y[3]);
  pss_invert_jacobian_matrix (J, inv_J00);
  for (i = 0; i < INTEGRATION_ORDER; i++) {
    for (j = 0; j < INTEGRATION_ORDER; j++) {
      pss_jacobian_matrix(J, ip[i].Loc, ip[j].Loc,
                          x[0], y[0], x[1], y[1], x[2], y[2], x[3], y[3]);
      pss_invert_jacobian_matrix (J, inv_J);
      pss_compute_B_matrix (B, ip[i].Loc, ip[j].Loc, inv_J, inv_J00);
      jdetw = ip[i].Wgt * ip[j].Wgt * pss_jacobian(J);
      B_TIMES_U (eps, B, u);     /* strain */
      eps[0]  *= jdetw; eps[1]  *= jdetw; eps[2]  *= jdetw;
      WM_get_pss_stress (m, t->matstate[i][j], /* RAW need to use update_form */
                         eps[0], eps[1], eps[2],
                         eps1_rate, eps2_rate, gamma_rate,
                         &sigma[0], &sigma[1], &sigma[2]);
      ADD_BT_TIMES_SIGMA(F, B, sigma); /* nodal forces */
    }
  }

  /* nodal forces */
  th = t->thickness;  
  FOR (i, 4) {
    VECMAC_3_VEC (force_vect, =, F[2*i] * cgd.corot_axes[0], +, F[2*i+1] * cgd.corot_axes[1]);
    VECMAC_SCALAR (force_vect, *=, (-th));
    WN_assemble_force (t->node[i], &force_vect);
  }

  /* deformation energy */
  el->deformation_energy =
    th * 0.5 * (  F[0] * u[0] + F[1] * u[1] + F[2] * u[2] + F[3] * u[3]
                + F[4] * u[4] + F[5] * u[5] + F[6] * u[6] + F[7] * u[7]);
}


#ifdef X_XFACE
#if SHOW_AXES
static void
show_axes (W_elem_t *el)
{
  priv_elem_t *pe = (priv_elem_t *)el->data;
  WCRec pnts[2];
  W_coords_3d_t c[4];
  geom_desc_t gd;
  int i;
  static EPixel conv_color[2], rot_color[2];
  
  FOR (i, 4) {GET_FROM_NODE (pe->node[i], WN_get_coords, c[i]);}

  compute_geom_desc (c, 0, 0, &gd);
  
  if (pe->conv_axes[0] == NULL) {
    BOOLEAN suc;
    conv_color[0] = ColorGetPixelFromString("red", &suc);
    conv_color[1] = ColorGetPixelFromString("green", &suc);
    rot_color[0] = ColorGetPixelFromString("orange", &suc);
    rot_color[1] = ColorGetPixelFromString("cyan", &suc);
    FOR (i, 2) {
      /* convected axes */
      VECMAC_3_VEC (pnts[0], =, gd.at, +, gd.covariant[i]);
      VECMAC_2_VEC (pnts[1], =, gd.covariant[i]);
      pe->conv_axes[i] = CreateVector3D (pnts);
      EASValsSetColor (conv_color[i]);
      EASValsSetLayer (WX_NODE_LAYER);
      EGWithMaskChangeAttributes(COLOR_MASK | LAYER_MASK, pe->conv_axes[i]);
      EMNoDrawAddGraphicsToModel (ESIModel (), pe->conv_axes[i]);
      /* co-rotated axes */
      VECMAC_3_VEC (pnts[0], =, gd.at, +, gd.corot_axes[i]);
      VECMAC_2_VEC (pnts[1], =, gd.corot_axes[i]);
      pe->rot_axes[i] = CreateVector3D (pnts);
      EASValsSetColor (rot_color[i]);
      EASValsSetLayer (WX_NODE_LAYER);
      EGWithMaskChangeAttributes(COLOR_MASK | LAYER_MASK, pe->rot_axes[i]);
      EMNoDrawAddGraphicsToModel (ESIModel (), pe->rot_axes[i]);
    }
  } else {
    FOR (i, 2) {
      /* convected axes */
      VECMAC_3_VEC (pnts[0], =, gd.at, +, gd.covariant[i]);
      VECMAC_2_VEC (pnts[1], =, gd.covariant[i]);
      EGModifyGraphicsGeometry (pe->conv_axes[i], pnts);
      /* co-rotated axes */
      VECMAC_3_VEC (pnts[0], =, gd.at, +, gd.corot_axes[i]);
      VECMAC_2_VEC (pnts[1], =, gd.corot_axes[i]);
      EGModifyGraphicsGeometry (pe->rot_axes[i], pnts);
    }
  }
}
#endif

static void
modify_elem_grep (W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  WCRec pnts[4];
  W_coords_3d_t refc[4], u[4];
  W_material_t *m;
  priv_elem_t *t = (priv_elem_t *)el->data;
  const FPNum zero = 0;
  int i;

  FOR (i, 4) {
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]);
    GET_FROM_NODE (t->node[i], WN_get_displ, u[i]);
    VECMAC_3_VEC (pnts[i], =, refc[i], +, displ_scale * u[i]);
  }

  m = WD_get_material_by_idx(el->domain, t->material);
  if (el->grep == NULL) {
    el->grep = CreateQuadWD3D (pnts, zero, zero, zero, zero);
    EASValsSetColor (WM_get_color(m));
    EASValsSetFillStyle (FILL_HOLLOW);
    EGWithMaskChangeAttributes (COLOR_MASK | FILL_MASK, el->grep);
    EMNoDrawAddGraphicsToModel (ESIModel (), el->grep);
  } else {
    EGModifyGraphicsGeometry (el->grep, pnts);
  }

#if SHOW_AXES
  show_axes (el);
#endif
}
#endif

static W_elem_methods_t methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL
};

static void
check_orientation (W_index_t elem_num, priv_elem_t *t)
{
  W_loc_t v0, v, cp={0,0,0};
  W_loc_t refc[4];
  int i, o0, o1;

  FOR (i, 4) {
    GET_FROM_NODE (t->node[i], WN_get_ref_coords, refc[i]);
  }
  
  VECMAC_3_VEC (v0, =, refc[1], -, refc[0]);
  
  VECMAC_3_VEC (v, =, refc[2], -, refc[0]);
  VECMAC_CROSS (cp, =, v0, x, v);
  o0 = (cp.z <= 0) ? -1 : +1;
  
  VECMAC_3_VEC (v, =, refc[3], -, refc[0]);
  VECMAC_CROSS (cp, =, v0, x, v);
  o1 = (cp.z <= 0) ? -1 : +1;

  if (o0 * o1 < 0) {
    sprintf (W_buf (), "Element %d: bad orientation (bow-tie)",
             (int)elem_num);
    REPORT_FATAL_ERROR (NULL, W_buf ());
  } else if (o0 < 0) {
#define swap(n0, n1) \
    {W_node_t *n=t->node[n0]; t->node[n0]=t->node[n1]; t->node[n1]=n;}
    swap (0, 1);
    swap (3, 2);
  }
}


/* Reads qmem4c element data from stream as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates elements. */

W_bool_t 
qmem4c_reader(W_domain_t *d)
{
  const char *input_line;
  long elem_num;
  priv_elem_t *t;
  int i, j, nodeid;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      t = w_new(priv_elem_t);
      CHECK_MALLOC(t);
      TOKENS_PARSE(input_line);
      elem_num             = TOKEN_INT(1);
      FOR (i, 4) {
        nodeid = TOKEN_INT(2+i);
        t->node[i] = WD_get_node_by_idx (d, nodeid);
        if (t->node[i] == NULL) {
          sprintf (W_buf (), "No such node: %d", (int)nodeid);
          REPORT_FATAL_ERROR (d->errh, W_buf ());
        }
      }
      check_orientation (elem_num, t);
      t->material          = TOKEN_INT(6);
      t->thickness         = TOKEN_DBL(7);
      for (i = 0; i < INTEGRATION_ORDER; i++) 
        for (j = 0; j < INTEGRATION_ORDER; j++) 
          t->matstate[i][j] = NULL;
#ifdef X_XFACE
# if SHOW_AXES
      FOR (i, 2) {
        t->conv_axes[i] = t->rot_axes[i] = NULL;
      }
# endif
#endif
#ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif      
      WD_add_element(d, elem_num,
                     WE_new_element(d, &methods, t
#                                   ifdef PVM_NEED_GHOST_ELEMENTS
                                    , is_ghost
#                                   endif
                                    ));
    }
  }
  return FALSE;
}
