#include "Wtypes.h"
#include "getoptP.h"
#include "tokensP.h"


W_loc_t crack_tip;
W_loc_t crack_end;

W_loc_t eval_pnt_loc;
W_loc_t efg_node_loc;

int have_crack;
int have_3d_case;

W_loc_t domain_ll_corner;
W_loc_t domain_ur_corner;
int n_intervals;

double d_m;
double kappa;

#undef X
#undef Y
#undef Z
#define X 0
#define Y 1
#define Z 2


static void
write_header(void);
static void
write_data(void);
static void
write_footer(void);

/* Returns FALSE when the segments are either parallel or identical; */
/* otherwise TRUE is returned.  The parameters /t1/ and /t2/ are calculated */
/* from the equation p1 + t1(q1 - p1) = p2 + t2(q2 - p2) */
int
intersect_segments(double *t1,  /* OUT parameter on the 1st segment */
                   double *t2,  /* OUT parameter on the 2nd segment */
                   double p1[], double p2[],
                   double q1[], double q2[],
                   double tol
                   );


double /* OUT weight function value */
qs_wf(double r               /* IN the w.f. parameter */
      );
double           
qs_Dwf_Dr(double r              /* IN the w.f. parameter */
          );


static double
m_func(W_loc_t *loc);


/* This is the quartic spline weight function.  Definition: */
/* w(r) = 1 - 6r^2 + 8r^3 - 3r^4, with r = sqrt(dksi^2 + deta^2)/dist_max */

double /* OUT weight function value */
qs_wf(double r               /* IN the w.f. parameter */
      )
{
  double tmp;

  if (r >= 1)
    return 0;
  else {
    tmp = 1 + r * r * (-6 + r * (8 - 3 * r));
    return tmp > 0? tmp: 0;     /* catch round-off errors */
  }
}


double           
qs_Dwf_Dr(double r              /* IN the w.f. parameter */
          )
{
  if (r >= 1) {
    return 0;
  } else {
    return (-12 * r * (1 + r * (r - 2)));
  }
}


static void 
print_usage_and_exit(int argc, char **argv);
static void 
get_options(int argc, char **argv);


int
main(int argc, char *argv[])
{
  get_options(argc, argv);
  write_header();
  write_data();
  write_footer();
}





static void 
print_usage_and_exit(int argc, char **argv)
{
  printf("\nUsage: %s [ options ]\n", argv[0]);
#define PL(line) printf("\n%s", line)
#define PL_WITH_DEFAULT_NUMBER(line,n) \
                      printf("\n%s (default is %f)", line, (n))
  PL("   Options -->");
  PL("-? or -h this help on the usage");
  PL("-c #,#,#,# crack segment (tip first)");
  PL("-D #,#,#,# domain lower-left and upper-right corner");
  PL("-I # number of intervals along domain box sides");
  PL("-E #,# evaluation point location");
  PL("-N #,# EFG node location");
  PL("-d # d_m");
  PL("-k # kappa (= dist from crack tip / d_m)");
  printf("\n"); 

#define OPTIONS "?hc:D:I:E:N:d:k:" 

#undef PL
  exit(0);
} 



static void 
get_options(int argc, char **argv) 
{
  int  ret;

  domain_ll_corner.x = -1;
  domain_ll_corner.y = -1;
  domain_ll_corner.z =  0;
  domain_ur_corner.x =  1;
  domain_ur_corner.y =  1;
  domain_ur_corner.z =  0;

  n_intervals = 10;

  crack_tip.x = 0;
  crack_tip.y = 0;
  crack_tip.z = 0;
  crack_end.x = 0;
  crack_end.y = 0;
  crack_end.z = 0;

  have_crack  = FALSE;

  have_3d_case = FALSE;

  eval_pnt_loc.x = 0;
  eval_pnt_loc.y = 0;
  eval_pnt_loc.z = 0;

  efg_node_loc.x = 0;
  efg_node_loc.y = 0;
  efg_node_loc.z = 0;

  d_m = 1;

  kappa = 0.1;

  optind = 0; opterr = 1;
  while((ret = get_cl_options(argc, argv, OPTIONS)) != EOF) {
    switch(ret) {
    case '?':
    case 'h':
      print_usage_and_exit(argc, argv);
      break;
    case 'c':
      {   
        int char_p = 0;
        char token[132];
        int tok_type;
        
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          crack_tip.x = strtod(token, NULL);
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == DELIMITER_TOKEN)
          tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          crack_tip.y = strtod(token, NULL);
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == DELIMITER_TOKEN)
          tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          crack_end.x = strtod(token, NULL);
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == DELIMITER_TOKEN)
          tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          crack_end.y = strtod(token, NULL);
        have_crack = TRUE;
      }
      break;
    case 'D':
      {
        int char_p = 0;
        char token[132];
        int tok_type;
        
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          domain_ll_corner.x = strtod(token, NULL);
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == DELIMITER_TOKEN)
          tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          domain_ll_corner.y = strtod(token, NULL);
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == DELIMITER_TOKEN)
          tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          domain_ur_corner.x = strtod(token, NULL);
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == DELIMITER_TOKEN)
          tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          domain_ur_corner.y = strtod(token, NULL);
      }  
      break;
    case 'I':
      n_intervals = (int)strtod(optarg, NULL);
      break;
    case 'd':
      d_m = strtod(optarg, NULL);
      break;
    case 'k':
      kappa = strtod(optarg, NULL);
      break;   
    case 'E':
      {
        int char_p = 0;
        char token[132];
        int tok_type;
         
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          eval_pnt_loc.x = strtod(token, NULL);
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == DELIMITER_TOKEN)
          tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          eval_pnt_loc.y = strtod(token, NULL);
      }  
      break;
    case 'N':
      {
        int char_p = 0;
        char token[132];
        int tok_type;
        
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          efg_node_loc.x = strtod(token, NULL);
        tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == DELIMITER_TOKEN)
          tok_type = tokens_get_token(optarg, &char_p, token, SKIP_SPACES);
        if (tok_type == NUMBER_TOKEN) 
          efg_node_loc.y = strtod(token, NULL);
      }  
      break;  
    }
  }
}


static void
write_data(void)
{
  double dx, dy, x, y;
  W_loc_t loc;
  double zero = 0;
  int i, j;
#define SETLOC(X, Y) { loc.x = X; loc.y = Y; loc.z = 0; }

  dx = (domain_ur_corner.x - domain_ll_corner.x) / (double)n_intervals;
  dy = (domain_ur_corner.y - domain_ll_corner.y) / (double)n_intervals;

  x = domain_ll_corner.x;
  for (i = 0; i < n_intervals; i++) {
    y = domain_ll_corner.y;
    for (j = 0; j < n_intervals; j++) {
      printf("Q4\n");
      
      SETLOC(x, y);
      printf("%g %g %g\n", loc.x, loc.y, zero);
      printf("%g\n", qs_wf(m_func(&loc)));
      
      SETLOC(x+dx, y);
      printf("%g %g %g\n", loc.x, loc.y, zero);
      printf("%g\n", qs_wf(m_func(&loc)));
      
      SETLOC(x+dx, y+dy);
      printf("%g %g %g\n", loc.x, loc.y, zero);  
      printf("%g\n", qs_wf(m_func(&loc)));
      
      SETLOC(x, y+dy);
      printf("%g %g %g\n", loc.x, loc.y, zero); 
      printf("%g\n", qs_wf(m_func(&loc)));
      
      y += dy;
    }
    x += dx;
  }
}



static void
write_header(void)
{
  printf("! HEXVUE file.  Generated by showwf.\n");
  printf("Weight function.\n");
  printf(" 1 \n");
  printf("Weight function  value.\n");
}


static void
write_footer(void)
{
  printf("END ! HEXVUE file.\n");
}


static double
m_func_crack(W_loc_t *loc);


static double
m_func(W_loc_t *loc)
{
  double d;

  d = v3_distance(loc, &efg_node_loc);

  if (!have_crack) {
    return d / d_m;
  } else {
    return m_func_crack(loc);
  }
}



#undef X
#undef Y
#undef Z
#define X 0
#define Y 1
#define Z 2
#define is_zero(v) (fabs(v) < tol)
#define parallel_z(v) (is_zero((v)[X]) && is_zero((v)[Y]))
#define linear_comb(r, p, pq, l)                \
  {                                             \
    (r)[X] = (p)[X] + l * (pq)[X];              \
    (r)[Y] = (p)[Y] + l * (pq)[Y];              \
    (r)[Z] = (p)[Z] + l * (pq)[Z];              \
  }
#define linear_comb2(r, p, pq, l)               \
  {                                             \
    (r)[X] = (p)[X] + l * (pq)[X];              \
    (r)[Y] = (p)[Y] + l * (pq)[Y];              \
  }
#define normal_vec2(n, v) { (n)[X] = -(v)[Y]; (n)[Y] = (v)[X]; }
#define dot_p2(a, b) ((a)[X] * (b)[X] + (a)[Y] * (b)[Y])
#define subt_vec2(ab, a, b)                              \
  { (ab)[X] = (b)[X] - (a)[X]; (ab)[Y] = (b)[Y] - (a)[Y]; }

static double 
l_func(double l);


static double
m_func_crack(W_loc_t *loc)
{
  double d;
  double p1[2], q1[2], p2[2], q2[2];
  double t1, t2;
  W_loc_t s;
  double l;
#define T_INSIDE(T) ((T >= 0) && (T <= 1))

  d = v3_distance(loc, &efg_node_loc);

  p1[X] = loc->x;         p1[Y] = loc->y;
  q1[X] = efg_node_loc.x; q1[Y] = efg_node_loc.y;
  
  p2[X] = crack_tip.x;    p2[Y] = crack_tip.y;
  q2[X] = crack_end.x;    q2[Y] = crack_end.y;

  if (intersect_segments(&t1, &t2, p1, q1, p2, q2, 1.e-6)) {
    if (T_INSIDE(t1) && T_INSIDE(t2)) {
      s.x = p2[X] + t2 * (q2[X] - p2[X]);
      s.y = p2[Y] + t2 * (q2[Y] - p2[Y]);
      s.z = 0;
      l = v3_distance(&s, &crack_tip);
      return (d + l_func(l)) / d_m;
    } else {
      return d / d_m;
    }
  } else {
    return d / d_m;
  }
}


static double 
l_func(double l)
{
#if 0 
  return l / kappa;/* linear function */
#else
#  if 1
  return l * l / d_m / (kappa * kappa); /* quadratic */
#  else
  return sqrt(l) * sqrt(d_m / kappa); /* square root */
#  endif
#endif
}


static void
intersect_lines(double s[],     /* OUT the intersection point (either */
                                /* s[2] for dim == 2, or s[3] for dim == 3 */
                double *lambda, /* OUT parameter of the intersection point */
                                /* of the lines */
                                /* x = a + lambda * d = b + mu * e */
                int *parallel,  /* OUT are the lines parallel? */
                double a[], double d[], double b[], double e[], /* IN see */
                                                                /* above */
                int dim,        /* IN see above */
                double tol,     /* IN tolerance for computation */
                double infinity /* IN parameter */
                );
 

static void
intersect_lines(double s[],     /* OUT the intersection point (either */
                                /* s[2] for dim == 2, or s[3] for dim == 3 */
                double *lambda, /* OUT parameter of the intersection point */
                                /* of the lines */
                                /* x = a + lambda * d = b + mu * e */
                int *parallel,  /* OUT are the lines parallel? */
                double a[], double d[], double b[], double e[], /* IN see */
                                                                /* above */
                int dim,        /* IN see above */
                double tol,     /* IN tolerance for computation */
                double infinity /* IN parameter */
                )
{
  double ne[2], ab[2];
  double d_ne, ab_ne;


  if (dim == 3 && parallel_z(e)) {
    if ((*parallel = parallel_z(d))) {
      if (a[X] == b[X] && a[Y] == b[Y]) *lambda = 0;
      else                              *lambda = infinity;/* lines parallel */
    } else if (!is_zero(d[X])) *lambda = (b[X] - a[X]) / d[X];
    else if (!is_zero(d[Y]))   *lambda = (b[Y] - a[Y]) / d[Y];
  } else { /* /e/ not parallel to z */
    normal_vec2(ne, e);
    d_ne = dot_p2(d, ne);
    subt_vec2(ab, a, b);
    ab_ne = dot_p2(ab, ne);
    if ((*parallel = is_zero(d_ne))) {
      if (is_zero(ab_ne)) *lambda = 0;
      else                *lambda = infinity;
    } else { /* the general case */
      *lambda = ab_ne / d_ne;
    } /* if ((*parallel = is_zero(d_ne))) */
  } /* if (dim == 3 && parallel_z(e)) */
  if (dim == 2) {
    linear_comb2(s, a, d, *lambda);
  } else {
    linear_comb(s, a, d, *lambda);
  }
}


/* Returns FALSE when the lines are either parallel or identical; */
/* otherwise TRUE is returned.  The parameters /t1/ and /t2/ are calculated */
/* from the equation p1 + t1(q1 - p1) = p2 + t2(q2 - p2). */
/* NOTE: The parameters are NOT computed for PARALLEL lines. */
int
intersect_segments(double *t1,  /* OUT parameter on the 1st segment */
                   double *t2,  /* OUT parameter on the 2nd segment */
                   double p1[], double q1[], /* IN 1st segment */
                   double p2[], double q2[], /* IN 2nd segment */
                   double tol   /* IN tolerance */
                   )
{
  double dir1[2], dir2[2], p1p2[2];
  double n1[2], n2[2];
  double det;

  subt_vec2(dir1, p1, q1); 
  subt_vec2(dir2, p2, q2); 
  normal_vec2(n1, dir1);
  normal_vec2(n2, dir2);
  det = dot_p2(dir1, n2);
  if (fabs(det) < tol) return FALSE; /* parallel */
  subt_vec2(p1p2, p1, p2);
  *t1 = dot_p2(p1p2, n2) / det;
  *t2 = dot_p2(p1p2, n1) / det;
  return TRUE;
}
