
#include "Wconfig.h"
#ifdef X_XFACE
#   include "xiface.h"
#endif
#define WMAT003_PRIVATE_HEADER
#include "Wmat003.h"
#include "Wdomain.h"
#include "Wmats.h"

#define EM          2.1e9 
#define NU          0.15
#define YIELD_UPP   200000
#define YIELD_LOW  -200000
#define HTYPE       MAT003_ISOTROPIC_HARDENING 
#define HMUPP       1e9
#define HMLOW       1e9
#define MAX_EPS     (5 * YIELD_UPP / EM)
#define PHI_STEP    0.1
#define MAX_PHI     8

typedef struct mat003_matstate_t {
  W_bool_t    is_plastic;
  W_float_t   stress;
  W_float_t   strain;
  W_float_t   dstrain;
  W_float_t   yield_stress_upp;
  W_float_t   yield_stress_low;
}              mat003_matstate_t;


int main()
{
  W_float_t eps, phi, stress;
  W_material_t *m;
  mat003_matstate_t *ms;
  W_mat003_t *u;

  u = (W_mat003_t *)make_node(sizeof(W_mat003_t));
  if (u == NULL) W_make_node_failure("mat003_reader");
  u->E                     = EM; 
  u->nu                    = NU;
  u->yield_stress_upp      = YIELD_UPP;
  u->yield_stress_low      = YIELD_LOW;
  u->hardening_type        = HTYPE;
  u->hardening_modulus_upp = HMUPP;
  u->hardening_modulus_low = HMLOW;

  m = (W_material_t *)make_node(sizeof(W_material_t));
  if (m == NULL) W_make_node_failure("WE_new_material");

  m->mcode   = MCODE_UNIAX_HEP;
  m->data    = (NODE)u;
  m->methods = NULL;

  ms = (mat003_matstate_t *)make_node(sizeof(mat003_matstate_t));
  ms->stress = ms->strain = 0;
  ms->is_plastic = FALSE;
  ms->yield_stress_upp = u->yield_stress_upp;
  ms->yield_stress_low = u->yield_stress_low;
  
  for (phi = 0; phi < MAX_PHI; phi += PHI_STEP) {
    mat003_get_uniaxial_stress(m, (NODE)ms,
                               (eps = MAX_EPS * sin(phi)), cos(phi), &stress);
    printf("%f %f\n", eps, stress);
  }
  return 1;
}




W_bool_t
mat003_get_uniaxial_stress(W_material_t *m,
                           W_matstate_t ms,
                           W_float_t strain,
                           W_float_t strain_rate,
                           W_float_t *stress
                           )
{
  W_mat003_t *u = (W_mat003_t *)m->data;
  mat003_matstate_t *m003ms = (mat003_matstate_t *)ms;
  W_float_t dsigma, sigma, dsigmay, deps;

  deps = (strain - m003ms->strain);
  if (!m003ms->is_plastic) { /* elastic state */
    dsigma  = deps * u->E;
    if (deps > 0) { /* extending */
      dsigmay = (m003ms->yield_stress_upp - m003ms->stress);
      if (dsigma / dsigmay > 1) { /* yielding */
        dsigma  = (deps - dsigmay / u->E) * u->hardening_modulus_upp;
        sigma   = m003ms->yield_stress_upp + dsigma;
        m003ms->is_plastic = +1;
      } else {
        sigma   = m003ms->stress + deps * u->E;
      }
    } else {        /* shortening */
      dsigmay = (m003ms->yield_stress_low - m003ms->stress);
      if (dsigma / dsigmay > 1) { /* yielding */
        dsigma  = (deps - dsigmay / u->E) * u->hardening_modulus_low;
        sigma   = m003ms->yield_stress_low + dsigma;
        m003ms->is_plastic = -1;
      } else {
        sigma   = m003ms->stress + deps * u->E;
      }
    }
  } else {                    /* plastic state */
    if (m003ms->dstrain * m003ms->is_plastic > 0) /* loading */
      if (m003ms->stress > m003ms->yield_stress_upp)
        sigma = m003ms->stress + deps * u->hardening_modulus_upp;
      else
        sigma = m003ms->stress + deps * u->hardening_modulus_low;
    else {                      /* unloading */
      if (u->hardening_type == MAT003_ISOTROPIC_HARDENING) {
        if (m003ms->stress > m003ms->yield_stress_upp)
          m003ms->yield_stress_upp = m003ms->stress;
        else 
          m003ms->yield_stress_low = m003ms->stress;
      } else if (u->hardening_type == MAT003_KINEMATIC_HARDENING) {
        double d = m003ms->yield_stress_upp - m003ms->yield_stress_low;
        
        if (m003ms->stress > m003ms->yield_stress_upp) {
          m003ms->yield_stress_upp = m003ms->stress;
          m003ms->yield_stress_low = m003ms->yield_stress_upp - d;
        } else { 
          m003ms->yield_stress_low = m003ms->stress;
          m003ms->yield_stress_upp = m003ms->yield_stress_low + d;
        }
      }
      dsigma  = deps * u->E;
      sigma   = m003ms->stress + dsigma;
      m003ms->is_plastic = FALSE;
    }
  }
  /* update material state */
  m003ms->dstrain = strain - m003ms->strain;
  m003ms->strain = strain;
  *stress = m003ms->stress = sigma;
  return TRUE;
}

