
#include "Wconfig.h"

#if W_ROT_DOFS
#include "Wxf.h"
#include "Wtnbeam.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"

/* RCSHEADER "$Id: tnbeam.c,v 1.1.1.1 2001/07/24 02:42:52 pkrysl Exp $" */

#define ONE_TWELVETH 0.083333333333333333


typedef struct W_tnbeam_t {
  W_node_t         *node[3];
  W_float_t         A;          /* area */
  W_float_t         It;         /* torsional moment of inertia */
  W_float_t         I2, I3;/* moments of inertia about local X2, X3 axes */
  W_float_t         B2, B3;/* reduction coeff. for shear in direction X2, X3 */
  W_float_t         L0;
  W_triad_3d_t      initial_elem_triad;
  W_index_t         material;
  W_float_t         nforce;     /* normal force: cached value */
  W_float_t         sforce2;    /* shear force along X2: cached value */
  W_float_t         sforce3;    /* shear force along X2: cached value */
  W_float_t         torsm;      /* torsional moment about X1: cached value */
  W_float_t         bendm2;     /* bending moment about X2: cached value */
  W_float_t         bendm3;     /* bending moment about X3: cached value */
}              W_tnbeam_t;


static void
calc_initial_elem_triad (W_elem_t *el);
static void
current_elem_triad (W_coords_3d_t *c1, W_coords_3d_t *c2,
                          W_triad_3d_t *eti, W_triad_3d_t *etj,
                          W_triad_3d_t *current_elem_triad);
static void
net_def_and_transform (W_elem_t *el, W_vector_t net_def, W_triad_3d_t *transf);
static void
compute_mass_properties (W_elem_t *el,
                         W_float_t *mass_per_node,
                         W_float_t *I11_per_node,
                         W_float_t *I22_per_node,
                         W_float_t *I33_per_node);


static int
elem_elcode (W_elem_t *el)
{
  return TNBEAM_ELCODE;
}



#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)

static void
calc_initial_elem_triad (W_elem_t *el)
{
  W_tnbeam_t *t = (W_tnbeam_t *)el->data;
  W_coords_3d_t c[3];
  W_coords_3d_t W;

  GET_FROM_NODE (t->node[0], WN_get_ref_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, c[1]);
  GET_FROM_NODE (t->node[2], WN_get_ref_coords, c[2]);

  VECMAC_3_VEC (t->initial_elem_triad.vect1, =, c[1], -, c[0]);
  VECMAC_NORMALIZE (t->initial_elem_triad.vect1);

  VECMAC_3_VEC (W, =, c[2], -, c[0]);
  VECMAC_CROSS (t->initial_elem_triad.vect3,
                =, t->initial_elem_triad.vect1, x, W);
  VECMAC_NORMALIZE (t->initial_elem_triad.vect3);
  VECMAC_CROSS (t->initial_elem_triad.vect2,
                =, t->initial_elem_triad.vect3, x, t->initial_elem_triad.vect1
                );
}



static void
net_def_and_transform (W_elem_t *el, W_vector_t net_def,
                       W_triad_3d_t *transf)
{
  W_tnbeam_t *t = (W_tnbeam_t *)el->data;
  W_coords_3d_t c[2], d;
  W_triad_3d_t eti, etj, Leti, Letj;
  W_float_t L;

  /* Element triad rotated by current nodal rotation matrix */
  WN_transform_by_rot_matrix_t (t->node[0], &t->initial_elem_triad, &eti);
  WN_transform_by_rot_matrix_t (t->node[1], &t->initial_elem_triad, &etj);

  /* Current length */
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  VECMAC_3_VEC (d, =, c[1], -, c[0]);
  L = VECMAC_NORM (d);
  
  /* Current element triad */
  current_elem_triad (&c[0], &c[1], &eti, &etj, transf);

  /* Terms of eti, etj in local (convected) coordinates (i.e. in transf) */
  W_transp_L_shift_triad (&Leti, transf, &eti);
  W_transp_L_shift_triad (&Letj, transf, &etj);

  /* Compute net deformations- Delta L, Th1ij, Th2i, Th2j, Th3i, Th3j  */
  /* Change in axis' length */
  net_def[0] = L - t->L0;
  /* Torsion angle */
  net_def[1] = (  (Letj.vect2.z - Leti.vect2.z) / Letj.vect2.y
                + (Leti.vect3.y - Letj.vect3.y) / Letj.vect3.z) * 0.5;
  /* Bending */
  net_def[2] = - Leti.vect1.z / Leti.vect1.x;
  net_def[3] = - Letj.vect1.z / Letj.vect1.x;
  net_def[4] =   Leti.vect1.y / Leti.vect1.x;
  net_def[5] =   Letj.vect1.y / Letj.vect1.x;
}


  
  

static void
current_elem_triad (W_coords_3d_t *c1, W_coords_3d_t *c2,
                    W_triad_3d_t *eti, W_triad_3d_t *etj,
                    W_triad_3d_t *current_elem_triad)
{
  W_triad_3d_t dR;
  W_coords_3d_t drv;
  W_coords_3d_t ij;
  double sinphi, phi;

  /* Extract dR as difference rotation matrix */
  W_transp_R_shift_triad (&dR, etj, eti);
  /* Compute difference rotation vector from dR */
  W_extract_rot_vector_from_triad (&dR, &drv);
  /* Rotate eti by half the amount to get element triad */
  drv.x *= 0.5;
  drv.y *= 0.5;
  drv.z *= 0.5;
  W_extract_triad_from_rot_vector (&drv, &dR);
  W_L_shift_triad (current_elem_triad, &dR, eti);
  /* Now rotate the element triad to point in the line of nodes I, J */
  ij.x = c2->x - c1->x;  ij.y = c2->y - c1->y;  ij.z = c2->z - c1->z;
  W_normalize_vect_3d (&ij);
  W_cross_prod_3d (&drv, &current_elem_triad->vect1, &ij);
  sinphi = W_vector_3d_length (&drv);
  phi    = asin (sinphi);
  drv.x *= phi / sinphi; drv.y *= phi / sinphi; drv.z *= phi / sinphi;
  W_rotate_triad_by_vector (current_elem_triad, &drv);
}


static void
compute_mass_properties (W_elem_t *el,
                         W_float_t *mass_per_node,
                         W_float_t *I11_per_node,
                         W_float_t *I22_per_node,
                         W_float_t *I33_per_node)
{
  W_tnbeam_t *t = (W_tnbeam_t *)el->data;
  W_float_t volume_mass, mass2;
  W_material_t *m;
  
  m = WD_get_material_by_idx(el->domain, t->material);
  volume_mass = WM_get_volume_mass(m);

  *mass_per_node = mass2 = (t->A * t->L0 * volume_mass / 2);
  *I11_per_node = volume_mass * (t->I2 + t->I3) * t->L0 / 2;
  *I22_per_node = (volume_mass * t->I2 * t->L0 / 2
                   + mass2 * t->L0 * t->L0 * t->L0 / 18);
  *I33_per_node = (volume_mass * t->I3 * t->L0 / 2
                   + mass2 * t->L0 * t->L0 * t->L0 / 18);
}
                               

#define CHKL0(t)                                                \
  {                                                             \
    if (t->L0 == 0) {                                           \
      GET_FROM_NODE (t->node[0], WN_get_ref_coords, c[0]);      \
      GET_FROM_NODE (t->node[1], WN_get_ref_coords, c[1]);      \
      VECMAC_3_VEC (V, =, c[1], -, c[0]);                       \
      t->L0 = VECMAC_NORM (V);                                  \
      calc_initial_elem_triad (el);                             \
    }                                                           \
  }

static void
assemble_mass (W_elem_t *el)
{
  W_tnbeam_t *t = (W_tnbeam_t *)el->data;
  W_triad_3d_t transf;
  W_float_t netdef[6], mass2, I11, I22, I33;
  W_triad_3d_t I, IG;
  W_coords_3d_t c[2], V;

  CHKL0 (t);
  
  compute_mass_properties (el, &mass2, &I11, &I22, &I33);

  WN_assemble_mass (t->node[0], mass2);
  WN_assemble_mass (t->node[1], mass2);

  net_def_and_transform (el, netdef, &transf);

  IG.vect1.x = I11;
  IG.vect1.y = 0;
  IG.vect1.z = 0;
  
  IG.vect2.x = 0;
  IG.vect2.y = I22;
  IG.vect2.z = 0;
  
  IG.vect3.x = 0;
  IG.vect3.y = 0;
  IG.vect3.z = I33;
  
  W_L_shift_triad (&I, &transf, &IG);
  W_transp_R_shift_triad (&IG, &I, &transf);
  WN_assemble_tensor_of_inertia (t->node[0], &IG);
  WN_assemble_tensor_of_inertia (t->node[1], &IG);
}




static W_float_t
total_mass (W_elem_t *el)
{
  W_float_t mass2, I11, I22, I33;
  W_coords_3d_t c[2], V;
  W_tnbeam_t *t = (W_tnbeam_t *)el->data;
  
  CHKL0 (t);
  compute_mass_properties (el, &mass2, &I11, &I22, &I33);
  
  return mass2 + mass2;
}



static W_float_t 
suggested_time_step (W_elem_t *el)
{
  W_float_t E, nu, G, dtN, dtS, dtB2, dtB3, dtT, dt, volume_mass;
  W_tnbeam_t *t = (W_tnbeam_t *)el->data;
  W_material_t *m;
  W_float_t mass2, I11, I22, I33;
  W_float_t L0, TAU, d2, d3, d21d2, d31d3, GAMMA2, GAMMA3, KSI2, KSI3;
  W_coords_3d_t c[2], V;
  
  CHKL0 (t);
  m = WD_get_material_by_idx(el->domain, t->material);
  if (!WM_get_uniax_lin_el_consts(m, &E, &nu))
    REPORT_FATAL_ERROR("No uniax_lin_el_consts");
  G = E / (2 * (1 + nu));
  volume_mass = WM_get_volume_mass(m);

  compute_mass_properties(el, &mass2, &I11, &I22, &I33);

  L0     = t->L0;
  TAU    = G * t->It / L0;
  d2     = 12 * E * t->I2 / (G * t->B2 * t->A) / L0 / L0;
  d3     = 12 * E * t->I3 / (G * t->B3 * t->A) / L0 / L0;
  d21d2  = d2 / (1 + d2);
  d31d3  = d3 / (1 + d3);
  GAMMA2 = d31d3 * (G * t->B2 * t->A);
  GAMMA3 = d21d2 * (G * t->B3 * t->A);
  KSI2   = d21d2 * E * t->I2 / L0;
  KSI3   = d31d3 * E * t->I3 / L0;
  
  dt = FLT_MAX;
  /* Axial stiffness */
  dtN = sqrt(mass2 / (t->A * E / t->L0));
  dt = min(dt, dtN);
  /* Shear stiffness */
  dtS = sqrt(mass2 / (GAMMA2 * 0.5));
  dt = min(dt, dtS);
  dtS = sqrt(mass2 / (GAMMA3 * 0.5));
  dt = min(dt, dtS);
  /* Bending stiffness  */ 
  dtB2 = sqrt(I22 / (KSI2 + GAMMA3 * L0 * ONE_TWELVETH));
  dtB3 = sqrt(I33 / (KSI3 + GAMMA2 * L0 * ONE_TWELVETH));
  dt = min(dt, dtB2);  dt = min(dt, dtB3);
  /* Torsional stiffness */
  dtT  = sqrt(I11 / TAU);
  dt = min(dt, dtT);
  
  return dt; /* dt^2 = 4 / lambda, lambda = highest circular frequency */
}


static char *
type_name(W_elem_t *el)
{
  static char buf[] = "tnbeam";
  return buf;
}


/* Assembles nodal forces by which the tnbeam element acts on its nodal points. */

static void
arf (W_elem_t *el)
{
  W_coords_3d_t c[2], fv, v, V;
  W_tnbeam_t *t = (W_tnbeam_t *)el->data;
  W_material_t *m;
  W_float_t netdef[6], E, nu, G, L0;
  W_float_t TAU, d2, d3, d21d2, d31d3, GAMMA2, GAMMA3, KSI2, KSI3;
  W_triad_3d_t transf;
#define TRANSF(F_rom, I_nto)                          \
  {                                                   \
     (I_nto).x = (  (F_rom).x * transf.vect1.x        \
                  + (F_rom).y * transf.vect2.x        \
                  + (F_rom).z * transf.vect3.x);      \
     (I_nto).y = (  (F_rom).x * transf.vect1.y        \
                  + (F_rom).y * transf.vect2.y        \
                  + (F_rom).z * transf.vect3.y);      \
     (I_nto).z = (  (F_rom).x * transf.vect1.z        \
                  + (F_rom).y * transf.vect2.z        \
                  + (F_rom).z * transf.vect3.z);      \
   }

  CHKL0 (t);
  L0 = t->L0;

  /* kinematic quantities */
  net_def_and_transform (el, netdef, &transf);
   
  /* obtain stress */
  m = WD_get_material_by_idx (el->domain, t->material);
  if (!WM_get_uniax_lin_el_consts (m, &E, &nu))
    REPORT_FATAL_ERROR("No uniax_lin_el_consts");
  G    = E / (2 * (1 + nu));

  TAU    = G * t->It / L0;
  d2     = 12 * E * t->I2 / (G * t->B2 * t->A) / L0 / L0;
  d3     = 12 * E * t->I3 / (G * t->B3 * t->A) / L0 / L0;
  d21d2  = d2 / (1 + d2);
  d31d3  = d3 / (1 + d3);
  GAMMA2 = d31d3 * (G * t->B2 * t->A);
  GAMMA3 = d21d2 * (G * t->B3 * t->A);
  KSI2   = d21d2 * E * t->I2 / L0;
  KSI3   = d31d3 * E * t->I3 / L0;

  t->nforce  = E * t->A * netdef[0] / L0;
  t->sforce2 = GAMMA2 * 0.5 * (netdef[4] + netdef[5]);
  t->sforce3 = GAMMA3 * 0.5 * (netdef[2] + netdef[3]);
  t->torsm   = TAU * netdef[1];
  t->bendm2  = (KSI2 + GAMMA3 * L0 * ONE_TWELVETH) * (netdef[3] - netdef[2]);
  t->bendm3  = (KSI3 + GAMMA2 * L0 * ONE_TWELVETH) * (netdef[5] - netdef[4]);
  
  v.x =   t->nforce;
  v.y = - t->sforce2;
  v.z =   t->sforce3;
  TRANSF(v, fv);
  WN_assemble_force (t->node[0], &fv);
  fv.x = - fv.x; fv.y = - fv.y; fv.z = - fv.z;
  WN_assemble_force (t->node[1], &fv);
  v.x = t->torsm;
  v.y = t->bendm2 - (t->sforce3 * L0 / 2);
  v.z = t->bendm3 - (t->sforce2 * L0 / 2);
  TRANSF(v, fv);
  WN_assemble_moment(t->node[0], &fv);
  v.x = - t->torsm;
  v.y = - t->bendm2 - (t->sforce3 * L0 / 2);
  v.z = - t->bendm3 - (t->sforce2 * L0 / 2);
  TRANSF(v, fv);
  WN_assemble_moment(t->node[1], &fv);

  el->deformation_energy =
    0.5 * (  t->nforce * netdef[0]
           + t->sforce2 * (netdef[4] + netdef[5]) * L0
           + t->sforce3 * (netdef[2] + netdef[3]) * L0
           + t->torsm * netdef[1]
           + t->bendm2 * (netdef[3] - netdef[2])
           + t->bendm3 * (netdef[5] - netdef[4])
           ); 
#undef TRANSF  
}


#ifdef X_XFACE
static void
modify_elem_grep (W_elem_t *el, W_bool_t immed_draw)
{
  WCRec pnts[2];
  W_coords_3d_t refc[2], u[2];
  W_material_t *m;
  W_tnbeam_t *t = (W_tnbeam_t *)el->data;

  GET_FROM_NODE (t->node[0], WN_get_ref_coords, refc[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, refc[1]);
  GET_FROM_NODE (t->node[0], WN_get_displ, u[0]);
  GET_FROM_NODE (t->node[1], WN_get_displ, u[1]);

  VECMAC_3_VEC (pnts[0], =, refc[0], +, options.displ_scale * u[0]);
  VECMAC_3_VEC (pnts[1], =, refc[1], +, options.displ_scale * u[1]);

  m = WD_get_material_by_idx(el->domain, t->material);
  if (el->grep == NULL) {
    el->grep = CreateLine3D(pnts);
    EASValsSetColor(WM_get_color(m));
    EGWithMaskChangeAttributes(COLOR_MASK, el->grep);
    EGAttachObject(el->grep, (EObjectP)el);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
  } else {
    EGModifyGraphicsByHandle(el->grep, 1, &pnts[0]);
    EGModifyGraphicsByHandle(el->grep, 2, &pnts[1]);
  }
}
#endif /* #ifdef X_XFACE */

/*
 COMPUTES ELASTIC STIFFNESS MATRIX OF TWO-NODE C-0 BEAM IN LOCAL ELASTIC COORDINATES

  VERSION  92/12/08

  PROCEDURE CALCULATES ELASTIC COEFFICIENTS OF STIFFNESS MATRIX
  OF TWO-NODE BEAM ELEMENT
     L.........LENGTH OF BEAM AXIS
     IN2, IN3..MOMENTS OF INERTIA, BENDING ABOUT THE X2-AXIS (IN2)
               OR X3-AXIS (IN3)
     GB2,GB3...SHEAR STIFFNESS FOR SHEAR IN X2-DIRECTION (G/BETA2),
                                      OR IN X3-DIRECTION (G/BETA3)
     TST.......IT*G
     SM ... FIELD ZEROED OR NOT; DEPENDING ON THE DEMANDS
            OF THE CALLING PROCEDURE

*/
static void
elstiff (MAT *SM,
         double EY,
         double CSA,
         double TST, double IN2, double IN3,
         double GB2, double GB3, double L)
{
  double NI, GAMA2, GAMA3, TAU, KSI2, KSI3;
  
  NI    = EY*CSA/L;
  GAMA2 = GB2*CSA;
  GAMA3 = GB3*CSA;
  TAU   = TST/L;
  KSI2  = EY*IN2/L;
  KSI3  = EY*IN3/L;

  /*  ADD INTO THE FIELD SM (ZEROED OR NOT; THAT IS ON THE CALLING PROCEDURE) */
  
  SM->me[0][0] +=  + NI;
  SM->me[0][6] +=  - NI;
  SM->me[6][0]  = SM->me[0][6];
  
  SM->me[1][1] +=  + GAMA2/L;
  SM->me[1][5] +=  + GAMA2*0.5;
  SM->me[5][1]  = SM->me[1][5];
  SM->me[1][7] +=  -  GAMA2/L;
  SM->me[7][1]  = SM->me[1][7];
  SM->me[1][11] +=  + GAMA2*0.5;
  SM->me[11][1] = SM->me[1][11];
  
  SM->me[2][2] +=  + GAMA3/L;
  SM->me[2][4] +=  -  GAMA3/2.;
  SM->me[4][2]  = SM->me[2][4];
  SM->me[2][8] +=  -  GAMA3/L;
  SM->me[8][2]  = SM->me[2][8];
  SM->me[2][10] +=  -  GAMA3/2.;
  SM->me[10][2] = SM->me[2][10];
  
  SM->me[3][3] +=  + TAU;
  SM->me[3][9] +=  - TAU;
  SM->me[9][3] = SM->me[3][9];
  
  SM->me[4][4] +=  + KSI2+GAMA3*L/4.;
  SM->me[4][8] +=  + GAMA3*2.;
  SM->me[8][4]  = SM->me[4][8];
  SM->me[4][10] +=  - KSI2+GAMA3*L/4.;
  SM->me[10][4] = SM->me[4][10];
  
  SM->me[5][5] +=  + KSI3+GAMA2*L/4.;
  SM->me[5][7] +=  - GAMA2/2.;
  SM->me[7][5]  = SM->me[5][7];
  SM->me[5][11] +=  - KSI3+GAMA2*L/4.;
  SM->me[11][5] = SM->me[5][11];
  
  SM->me[6][6] +=  + NI;
  
  SM->me[7][7] +=  + GAMA2/L;
  SM->me[7][11] +=  - GAMA2/2.;
  SM->me[11][7] = SM->me[7][11];
  
  SM->me[8][8] +=  + GAMA3/L;
  SM->me[8][10] +=  + GAMA3/2.;
  SM->me[10][8] = SM->me[8][10];
  
  SM->me[9][9] +=  + TAU;
  
  SM->me[10][10] +=  + KSI2+GAMA3*L/4.;
  
  SM->me[11][11] +=  + KSI3+GAMA2*L/4.;
}

static W_elem_methods_t tnbeam_methods = {
  elem_elcode,
  type_name,
  arf,
  NULL,
  NULL,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL
};



/* Reads tnbeam element data from stream as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates elements. */

W_bool_t 
tnbeam_reader (W_domain_t *d, FILE *stream)
{
  const char *input_line;
  long elem_num, i, nodeid;
  W_tnbeam_t *t;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif

  while ((input_line = W_get_input_line(stream)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      t = w_new(W_tnbeam_t);
      CHECK_MALLOC(t);
      TOKENS_PARSE(input_line);
      elem_num    = TOKEN_INT(1);
      FOR (i, 3) {
        nodeid = TOKEN_INT(2+i);
        t->node[i] = WD_get_node_by_idx (d, nodeid);
        if (t->node[i] == NULL) {
          sprintf (W_buf (), "No such node: %ld", nodeid);
          REPORT_FATAL_ERROR (W_buf ());
        }
      }
      t->material = TOKEN_INT(5);
      t->A        = TOKEN_DBL(6);
      t->It       = TOKEN_DBL(7);
      t->I2       = TOKEN_DBL(8);
      t->I3       = TOKEN_DBL(9);
      t->B2       = TOKEN_DBL(10);
      t->B3       = TOKEN_DBL(11);
#ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif
      WD_add_element(d, elem_num,
                     WE_new_element(d, &tnbeam_methods, t
#                    ifdef PVM_NEED_GHOST_ELEMENTS
                     , is_ghost
#                    endif
                     ));
    }
  }
  return FALSE;
}

#endif /* W_ROT_DOFS */
