#include "Wconfig.h"
#include "Wxf.h"
#include "Wtypes.h"
#include "Wutil.h"
#include "Welem.h"
#include "Wdomain.h"
#include "Wnode.h"
#include "Wmats.h"
#include "Wnewmarki.h"
#include "Wcdi.h"
#include "Wtruss.h"
#include "Whistry.h"
#include "meschach.h"

/* RCSHEADER "$Id: truss.c,v 1.2 2001/08/16 22:07:14 pkrysl Exp $" */

/*
  The type of a truss element.  
 */
typedef struct priv_elem_t {
  W_node_t         *node[2]; /* node ID's */
  W_float_t         A;		  /* cross-sectional area */
  W_float_t         L0;		  /* length -- cached */
  W_index_t         material;	  /* material ID */
  W_matstate_t      matstate;	  /* material state */
  W_float_t         nforce;	  /* force -- cached */
}              priv_elem_t;


/* history client data */
typedef struct history_client_data_t {
  W_elem_t          *el; /* FE */
  W_index_t          pntidx; /* unique integer  */
}              history_client_data_t;


#define GET_FROM_NODE(node, func, tovect)         \
   func (node, &tovect.x, &tovect.y, &tovect.z)


/*
  Declarations of element methods.
 */
static int
elem_elcode (W_elem_t *el);
static char *
type_name(W_elem_t *el);
static void
assemble_restoring_forces(W_elem_t *el, W_mat_state_update_form_enum update_form);
static void
assemble_mass(W_elem_t *el);
static W_float_t 
suggested_time_step(W_elem_t *el);
#ifdef X_XFACE
/*
  Modify the graphic representation of the element (a 3D line).
 */
static void
modify_elem_grep(W_elem_t *el, double displ_scale, W_bool_t immed_draw)
{
  WCRec pnts[2];
  W_coords_3d_t refc[2], u[2];
  W_material_t *m;
  priv_elem_t *t = (priv_elem_t *)el->data;

  GET_FROM_NODE (t->node[0], WN_get_ref_coords, refc[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, refc[1]);
  GET_FROM_NODE (t->node[0], WN_get_displ, u[0]);
  GET_FROM_NODE (t->node[1], WN_get_displ, u[1]);

  VECMAC_3_VEC (pnts[0], =, refc[0], +, displ_scale * u[0]);
  VECMAC_3_VEC (pnts[1], =, refc[1], +, displ_scale * u[1]);

  m = WD_get_material_by_idx(el->domain, t->material);
  if (el->grep == NULL) {
    el->grep = CreateLine3D(pnts);
    EASValsSetColor(WM_get_color(m));
    EASValsSetLayer (WX_ELEM_LAYER);
    EGWithMaskChangeAttributes(COLOR_MASK | LAYER_MASK, el->grep);
    EMNoDrawAddGraphicsToModel(ESIModel (), el->grep);
  } else {
    EGModifyGraphicsByHandle(el->grep, 1, &pnts[0]);
    EGModifyGraphicsByHandle(el->grep, 2, &pnts[1]);
  }
}

#endif


static W_float_t
total_mass(W_elem_t *el);


static int
elem_elcode (W_elem_t *el)
{
  return TRUSS_ELCODE;
}



/*
  Compute the element length.
  */
static double
length (priv_elem_t *t)
{
  W_coords_3d_t c[2];
  W_coords_3d_t N;

  GET_FROM_NODE (t->node[0], WN_get_ref_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, c[1]);
 
  VECMAC_3_VEC(N, =, c[1], -, c[0]);
  return VECMAC_NORM (N);
} 





/*
  Returns a unmodifiable string with the element name.
 */
static char *
type_name(W_elem_t *el)
{
  static char buf[] = "truss";
  return buf;
}




/*
  Assembles nodal forces by which the truss element acts on its nodes.
  All effects (restoring and damping forces) must be included.
 */
static void
assemble_restoring_forces (W_elem_t *el, W_mat_state_update_form_enum update_form)
{
  W_float_t nforce;
  W_float_t L;
  W_coords_3d_t c[2], N, force;
  W_velocity_3d_t vel[2], v;
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_material_t *m;
  W_float_t stress, strain, strain_rate;

  /* kinematic quantities */
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  L   = VECMAC_NORM (N);
  VECMAC_SCALAR (N, /=, L);
  GET_FROM_NODE (t->node[0], WN_get_velocity, vel[0]);
  GET_FROM_NODE (t->node[1], WN_get_velocity, vel[1]);
  VECMAC_3_VEC(v, =, vel[1], -, vel[0]);
  strain_rate = VECMAC_DOT (v, N);
  strain      = (L - t->L0) / t->L0;
   
  /* obtain stress */
  m = WD_get_material_by_idx (el->domain, t->material);
  if (t->matstate == NULL)
    t->matstate = WM_get_mat_state (m); /* RAW need to use update_form  */
  if (!WM_get_uniax_stress (m, t->matstate, strain, strain_rate, &stress)) {
    sprintf (W_buf (), "Material %d does not support WM_get_uniax_stress",
             t->material);
    REPORT_FATAL_ERROR (el->domain->errh, W_buf ());
  }
  t->nforce = nforce = stress * t->A;

  VECMAC_3_SCALAR (force, =, nforce * N.x, nforce * N.y, nforce * N.z);
  WN_assemble_force (t->node[0], &force);
  VECMAC_3_SCALAR (force, = , -force.x, -force.y, -force.z);
  WN_assemble_force (t->node[1], &force);
   
  el->deformation_energy =
    (t->A * t->L0 * WM_get_uniax_def_energy(m, t->matstate, strain));
}


/*
  Compute the total mass of the element.
 */
static W_float_t
total_mass(W_elem_t *el)
{
  W_float_t volume_mass;
  priv_elem_t *t= (priv_elem_t *)el->data;
  W_material_t *m;

  m = WD_get_material_by_idx (el->domain, t->material);
  volume_mass = WM_get_volume_mass (m);
  
  return volume_mass * t->A * t->L0;
}


/*
  Assemble mass to the nodes.
 */
static void
assemble_mass (W_elem_t *el)
{
  W_float_t mass2;
  priv_elem_t *t = (priv_elem_t *)el->data;
  
  mass2 = total_mass(el) / 2;

  WN_assemble_mass (t->node[0], mass2);
  WN_assemble_mass (t->node[1], mass2);
}


/*
  Compute the suggested time step.
 */
static W_float_t 
suggested_time_step(W_elem_t *el)
{
  W_float_t E, nu;
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_material_t *m;

  m = WD_get_material_by_idx(el->domain, t->material);
  if (!WM_get_uniax_lin_el_consts(m, &E, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No uniax_lin_el_consts");

  return t->L0 / sqrt(E / WM_get_volume_mass(m));
}


#define EFFICIENT_ELMAT_TRANSF 0

#if EFFICIENT_ELMAT_TRANSF
static void
transf_matrix_eff (W_vect_t *t0, W_vect_t *t1, W_vect_t *t2,
               W_coords_3d_t *refci,
               W_coords_3d_t *refcj,
               W_coords_3d_t *ci,
               W_coords_3d_t *cj
               )
{
  W_vect_t vt, s = {0,0,0}, t = {0,0,0};
  
  VECMAC_3_VEC (vt, =, (*cj), -, (*ci));

  VECMAC_NORMALIZE (vt);
  if (fabs (vt.z) > 0.95) {     /* element almost parallel to the Z-axis */
    W_vect_t x = {1,0,0};
    s = x;
  } else {
    W_vect_t z = {0,0,1};
    s = z;
  }
  VECMAC_CROSS (t, =, vt, x, s);
  VECMAC_CROSS (s, =, t,  x, vt); 
  VECMAC_NORMALIZE (s);
  VECMAC_NORMALIZE (t);

  t0->x = vt.x;
  t0->y = vt.y;
  t0->z = vt.z;

  t1->x = s.x;
  t1->y = s.y;
  t1->z = s.z;

  t2->x = t.x;
  t2->y = t.y;
  t2->z = t.z;
}
#endif

static void
transf_matrix (MAT *tm,
               W_coords_3d_t *refci,
               W_coords_3d_t *refcj,
               W_coords_3d_t *ci,
               W_coords_3d_t *cj
               )
{
  W_vect_t vt, s = {0,0,0}, t = {0,0,0};
  
  VECMAC_3_VEC (vt, =, (*cj), -, (*ci));

  VECMAC_NORMALIZE (vt);
  if (fabs (vt.z) > 0.95) {     /* element almost parallel to the Z-axis */
    W_vect_t x = {1,0,0};
    s = x;
  } else {
    W_vect_t z = {0,0,1};
    s = z;
  }
  VECMAC_CROSS (t, =, vt, x, s);
  VECMAC_CROSS (s, =, t,  x, vt); 
  VECMAC_NORMALIZE (s);
  VECMAC_NORMALIZE (t);

  m_zero (tm);
  
  tm->me[0][0] = vt.x;
  tm->me[1][0] = vt.y;
  tm->me[2][0] = vt.z;

  tm->me[0][1] = s.x;
  tm->me[1][1] = s.y;
  tm->me[2][1] = s.z;

  tm->me[0][2] = t.x;
  tm->me[1][2] = t.y;
  tm->me[2][2] = t.z;

  tm->me[3][3] = vt.x;
  tm->me[4][3] = vt.y;
  tm->me[5][3] = vt.z;

  tm->me[3][4] = s.x;
  tm->me[4][4] = s.y;
  tm->me[5][4] = s.z;

  tm->me[3][5] = t.x;
  tm->me[4][5] = t.y;
  tm->me[5][5] = t.z;
}

static void
copy_to_k (MAT *sm, double m[36])
{
  int i, j, k;
  k = 0;
  FOR (i, 6) {
    FOR (j, 6) {
      m[k] = sm->me[i][j];
      k++;
    }
  }
}



#if EFFICIENT_ELMAT_TRANSF
 
static void
eff_stiffness_eff (W_elem_t *el,
                   
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                            double tk[36]
                        )
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_float_t E, nu;
  double ni;
  W_material_t *m;
  W_coords_3d_t refc[2], c[2], N;
  double mass2;
  double CA, lt;
  double k[36];
  W_vect_t t0, t1, t2;
  W_index_t idcs[6];
  double nforce, stress, strain, strain_rate;
#define K(I,J) k[I*6+J]

  m = WD_get_material_by_idx (el->domain, t->material);
  if (!WM_get_uniax_lin_el_consts (m, &E, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No uniax_lin_el_consts");

  GET_FROM_NODE (t->node[0], WN_get_ref_coords, refc[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, refc[1]);
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  lt = VECMAC_NORM (N);
  VECMAC_SCALAR (N, /=, lt);
 
  { int m; FOR (m, 36) k[m] = 0; } /* zero stiffness matrix */
  
  /* elastic stiffness matrix */
  ni = stiff_mult * E * t->A / lt;
  K(0,0) = +ni;
  K(0,3) = (-ni);
  K(3,0) = (-ni);
  K(3,3) = +ni;

  /* stress matrix  */
  strain_rate = 0;
  strain      = (lt - t->L0) / t->L0;
  if (t->matstate == NULL) t->matstate = WM_get_mat_state (m);
  WM_get_uniax_stress (m, t->matstate, strain, strain_rate, &stress);
  t->nforce = nforce = stress * t->A;
  ni = stiff_mult * nforce / lt;
  K(1,1)  += +ni;
  K(2,2)  += +ni;
  K(1,4)  += -ni;
  K(4,1)  += -ni;
  K(2,5)  += -ni;
  K(5,2)  += -ni;
  K(4,4)  += +ni;
  K(5,5)  += +ni;
  
  /* mass matrix */
  mass2 = total_mass(el) / 2;
  CA    = mass_mult * mass2;
  K(0,0) += CA;
  K(1,1) += CA;
  K(2,2) += CA;
  K(3,3) += CA;
  K(4,4) += CA;
  K(5,5) += CA;

  transf_matrix_eff (&t0, &t1, &t2, &refc[0], &refc[1], &c[0], &c[1]);

  tk[0] =
    + (t0.x * k[0] * t0.x)
    + (t0.y * k[7] * t0.y)
    + (t0.z * k[14] * t0.z);
  
  tk[1] =
    + (t0.x * k[0] * t1.x)
    + (t0.y * k[7] * t1.y)
    + (t0.z * k[14] * t1.z);
  
  tk[2] =
    + (t0.x * k[0] * t2.x)
    + (t0.y * k[7] * t2.y)
    + (t0.z * k[14] * t2.z);
  
  tk[3] =
    + (t0.x * k[3] * t0.x)
    + (t0.y * k[10] * t0.y)
    + (t0.z * k[17] * t0.z);
  
  tk[4] =
    + (t0.x * k[3] * t1.x)
    + (t0.y * k[10] * t1.y)
    + (t0.z * k[17] * t1.z);
  
  tk[5] =
    + (t0.x * k[3] * t2.x)
    + (t0.y * k[10] * t2.y)
    + (t0.z * k[17] * t2.z);
  
  tk[6] =
    + (t1.x * k[0] * t0.x)
    + (t1.y * k[7] * t0.y)
    + (t1.z * k[14] * t0.z);
  
  tk[7] =
    + (t1.x * k[0] * t1.x)
    + (t1.y * k[7] * t1.y)
    + (t1.z * k[14] * t1.z);
  
  tk[8] =
    + (t1.x * k[0] * t2.x)
    + (t1.y * k[7] * t2.y)
    + (t1.z * k[14] * t2.z);
  
  tk[9] =
    + (t1.x * k[3] * t0.x)
    + (t1.y * k[10] * t0.y)
    + (t1.z * k[17] * t0.z);
  
  tk[10] =
    + (t1.x * k[3] * t1.x)
    + (t1.y * k[10] * t1.y)
    + (t1.z * k[17] * t1.z);
  
  tk[11] =
    + (t1.x * k[3] * t2.x)
    + (t1.y * k[10] * t2.y)
    + (t1.z * k[17] * t2.z);
  
  tk[12] =
    + (t2.x * k[0] * t0.x)
    + (t2.y * k[7] * t0.y)
    + (t2.z * k[14] * t0.z);
  
  tk[13] =
    + (t2.x * k[0] * t1.x)
    + (t2.y * k[7] * t1.y)
    + (t2.z * k[14] * t1.z);
  
  tk[14] =
    + (t2.x * k[0] * t2.x)
    + (t2.y * k[7] * t2.y)
    + (t2.z * k[14] * t2.z);
  
  tk[15] =
    + (t2.x * k[3] * t0.x)
    + (t2.y * k[10] * t0.y)
    + (t2.z * k[17] * t0.z);
  
  tk[16] =
    + (t2.x * k[3] * t1.x)
    + (t2.y * k[10] * t1.y)
    + (t2.z * k[17] * t1.z);
  
  tk[17] =
    + (t2.x * k[3] * t2.x)
    + (t2.y * k[10] * t2.y)
    + (t2.z * k[17] * t2.z);
  
  tk[18] =
    + (t0.x * k[18] * t0.x)
    + (t0.y * k[25] * t0.y)
    + (t0.z * k[32] * t0.z);
  
  tk[19] =
    + (t0.x * k[18] * t1.x)
    + (t0.y * k[25] * t1.y)
    + (t0.z * k[32] * t1.z);
  
  tk[20] =
    + (t0.x * k[18] * t2.x)
    + (t0.y * k[25] * t2.y)
    + (t0.z * k[32] * t2.z);
  
  tk[21] =
    + (t0.x * k[21] * t0.x)
    + (t0.y * k[28] * t0.y)
    + (t0.z * k[35] * t0.z);
  
  tk[22] =
    + (t0.x * k[21] * t1.x)
    + (t0.y * k[28] * t1.y)
    + (t0.z * k[35] * t1.z);
  
  tk[23] =
    + (t0.x * k[21] * t2.x)
    + (t0.y * k[28] * t2.y)
    + (t0.z * k[35] * t2.z);
  
  tk[24] =
    + (t1.x * k[18] * t0.x)
    + (t1.y * k[25] * t0.y)
    + (t1.z * k[32] * t0.z);
  
  tk[25] =
    + (t1.x * k[18] * t1.x)
    + (t1.y * k[25] * t1.y)
    + (t1.z * k[32] * t1.z);
  
  tk[26] =
    + (t1.x * k[18] * t2.x)
    + (t1.y * k[25] * t2.y)
    + (t1.z * k[32] * t2.z);
  
  tk[27] =
    + (t1.x * k[21] * t0.x)
    + (t1.y * k[28] * t0.y)
    + (t1.z * k[35] * t0.z);
  
  tk[28] =
    + (t1.x * k[21] * t1.x)
    + (t1.y * k[28] * t1.y)
    + (t1.z * k[35] * t1.z);
  
  tk[29] =
    + (t1.x * k[21] * t2.x)
    + (t1.y * k[28] * t2.y)
    + (t1.z * k[35] * t2.z);
  
  tk[30] =
    + (t2.x * k[18] * t0.x)
    + (t2.y * k[25] * t0.y)
    + (t2.z * k[32] * t0.z);
  
  tk[31] =
    + (t2.x * k[18] * t1.x)
    + (t2.y * k[25] * t1.y)
    + (t2.z * k[32] * t1.z);
  
  tk[32] =
    + (t2.x * k[18] * t2.x)
    + (t2.y * k[25] * t2.y)
    + (t2.z * k[32] * t2.z);
  
  tk[33] =
    + (t2.x * k[21] * t0.x)
    + (t2.y * k[28] * t0.y)
    + (t2.z * k[35] * t0.z);
  
  tk[34] =
    + (t2.x * k[21] * t1.x)
    + (t2.y * k[28] * t1.y)
    + (t2.z * k[35] * t1.z);
  
  tk[35] =
    + (t2.x * k[21] * t2.x)
    + (t2.y * k[28] * t2.y)
    + (t2.z * k[35] * t2.z);
  
#if 0
  WN_eqn_numbers (t->node[0], &idcs[0], 3);
  WN_eqn_numbers (t->node[1], &idcs[3], 3);
  (*assemble_eff_stiff_callback) (el->domain->newmark_integrator, FLES_ADD, 6, idcs, tk);
#endif
}

static void
assemble_eff_stiffness_eff (W_elem_t *el,
                            W_float_t stiff_mult,
                            W_float_t mass_mult,
                            W_mass_type_enum mass_type)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  double tk[36];
  int idcs[6];
  
  eff_stiffness_eff (el, stiff_mult, mass_mult, tk);

  WN_eqn_numbers (t->node[0], &idcs[0], 3);
  WN_eqn_numbers (t->node[1], &idcs[3], 3);
  (*assemble_eff_stiff_callback) (el->domain->newmark_integrator, FLES_ADD, 6, idcs, tk);
}
#endif

static void
compare (double k[36], double ek[36])
{
  int i;
  FOR (i, 36) {
    if (fabs (k[i] - ek[i]) > 1e-6) {
      fprintf (stderr, "mismatch %g vs. %g\n", k[i], ek[i]);
    }
  }
}

static void
assemble_eff_stiffness (W_elem_t *el,
                        W_newmark_assemble_eff_stiff_proc assemble_eff_stiff_callback,
                        W_float_t stiff_mult,
                        W_float_t mass_mult,
                        W_mass_type_enum mass_type
                        )
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_float_t E, nu;
  double ni;
  static MAT *sm = NULL;
  static MAT *smt = NULL;
  static MAT *tm = NULL;
  W_material_t *m;
  W_coords_3d_t refc[2], c[2], N;
  double mass2;
  double CA, lt;
  double k[36], tk[36];
  W_bool_t constrained[6];
  W_index_t idcs[6];
  double nforce, stress, strain, strain_rate;

  m = WD_get_material_by_idx (el->domain, t->material);
  if (!WM_get_uniax_lin_el_consts (m, &E, &nu))
    REPORT_FATAL_ERROR(el->domain->errh, "No uniax_lin_el_consts");

  if (!sm)  { CATCH(E_MEM, { sm  = m_get (6, 6); }, { sm = NULL; CHECK_MALLOC (sm); }); }
  if (!smt) { CATCH(E_MEM, { smt = m_get (6, 6); }, { smt = NULL; CHECK_MALLOC (smt); }); } 
  if (!tm)  { CATCH(E_MEM, { tm  = m_get (6, 6); }, { tm = NULL; CHECK_MALLOC (tm); }); } 
  
  GET_FROM_NODE (t->node[0], WN_get_ref_coords, refc[0]);
  GET_FROM_NODE (t->node[1], WN_get_ref_coords, refc[1]);
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  lt = VECMAC_NORM (N);
  VECMAC_SCALAR (N, /=, lt);
 
  m_zero (sm);
  
  /* elastic stiffness matrix */
  ni = stiff_mult * E * t->A / lt;
  sm->me[0][0] = sm->me[0][0] + ni;
  sm->me[0][3] = sm->me[0][3] - ni;
  sm->me[3][0] = sm->me[0][3];
  sm->me[3][3] = sm->me[3][3] + ni;

  /* stress matrix  */
  strain_rate = 0;
  strain      = (lt - t->L0) / t->L0;
  if (t->matstate == NULL) t->matstate = WM_get_mat_state (m);
  WM_get_uniax_stress (m, t->matstate, strain, strain_rate, &stress);
  t->nforce = nforce = stress * t->A;
  CA = stiff_mult * nforce / lt;
  sm->me[1][1]  += +CA;
  sm->me[2][2]  += +CA;
  sm->me[1][4]  += -CA;
  sm->me[4][1]  += -CA;
  sm->me[2][5]  += -CA;
  sm->me[5][2]  += -CA;
  sm->me[4][4]  += +CA;
  sm->me[5][5]  += +CA;
  
  /* mass matrix */
  mass2 = total_mass(el) / 2;
  switch (mass_type) {
  case W_MASS_CONSISTENT:
    {
      double diag = 2./3 * mass_mult * mass2;
      double offdiag = 1./2 * mass_mult * mass2;
      sm->me[0][0] += diag;
      sm->me[1][1] += diag;
      sm->me[2][2] += diag;
      sm->me[3][3] += diag;
      sm->me[4][4] += diag;
      sm->me[5][5] += diag;
      sm->me[0][3] += offdiag;
      sm->me[1][4] += offdiag;
      sm->me[2][5] += offdiag;
      sm->me[3][0] += offdiag;
      sm->me[4][1] += offdiag;
      sm->me[5][2] += offdiag;
    }
    break;
  case W_MASS_DIAG: 
  default:
    CA    = mass_mult * mass2;
    sm->me[0][0] += CA;
    sm->me[1][1] += CA;
    sm->me[2][2] += CA;
    sm->me[3][3] += CA;
    sm->me[4][4] += CA;
    sm->me[5][5] += CA;
    break;
  }

  transf_matrix (tm, &refc[0], &refc[1], &c[0], &c[1]);
  m_mlt (tm, sm, smt);
  mmtr_mlt (smt, tm, sm);
  copy_to_k (sm, k);
#if 0
  eff_stiffness_eff (el, stiff_mult, mass_mult, tk); /* RAW */
  compare (k, tk);              /* RAW */
#endif

  WN_eqn_numbers_and_constraints (t->node[0], &idcs[0], &constrained[0]);
  WN_eqn_numbers_and_constraints (t->node[1], &idcs[3], &constrained[3]);
  (*assemble_eff_stiff_callback) (el->domain->newmark_integrator,
                                  FLES_ADD, 6, idcs, k, constrained);
}


/*
  Assembles nodal forces by which the truss element acts on its nodes.
  All effects (restoring and damping forces) must be included.
 */
static void
assemble_residual_forces (W_elem_t *el, W_mat_state_update_form_enum update_form,
                          W_mass_type_enum mass_type)
{
  W_float_t nforce;
  W_float_t L;
  W_coords_3d_t c[2], N, force, a[2];
  W_velocity_3d_t vel[2], v;
  priv_elem_t *t = (priv_elem_t *)el->data;
  W_material_t *m;
  W_float_t stress, strain, strain_rate;
  double mass2;

  /* kinematic quantities */
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  L   = VECMAC_NORM (N);
  VECMAC_SCALAR (N, /=, L);
  GET_FROM_NODE (t->node[0], WN_get_velocity, vel[0]);
  GET_FROM_NODE (t->node[1], WN_get_velocity, vel[1]);
  VECMAC_3_VEC(v, =, vel[1], -, vel[0]);
  strain_rate = VECMAC_DOT (v, N);
#if 1
  strain      = L / t->L0 - 1;  /* engineering strain */
#else
  strain      = (1 - (t->L0*t->L0) / (L*L)); /* Almansi */
#endif
   
  /* obtain stress */
  m = WD_get_material_by_idx (el->domain, t->material);
  if (t->matstate == NULL)
    t->matstate = WM_get_mat_state (m);
  /* RAW need to use update_form */
  WM_get_uniax_stress (m, t->matstate, strain, strain_rate, &stress);
  t->nforce = nforce = stress * t->A;

  VECMAC_3_SCALAR (force, =, nforce * N.x, nforce * N.y, nforce * N.z);
  WN_assemble_force (t->node[0], &force);
  VECMAC_3_SCALAR (force, = , -force.x, -force.y, -force.z);
  WN_assemble_force (t->node[1], &force);
   
  GET_FROM_NODE (t->node[0], WN_get_acceleration, a[0]);
  GET_FROM_NODE (t->node[1], WN_get_acceleration, a[1]);
  mass2 = total_mass (el) / 2;
  switch (mass_type) {
  case W_MASS_CONSISTENT:
    VECMAC_3_VEC (force, =, (-2. / 3 * mass2) * a[0], +, (-1. / 2 * mass2) * a[1]);
    WN_assemble_force (t->node[0], &force);
    VECMAC_3_VEC (force, =, (-1. / 2 * mass2) * a[0], +, (-2. / 3 * mass2) * a[1]);
    WN_assemble_force (t->node[1], &force);
    break;
  case W_MASS_DIAG:
  default:
    VECMAC_2_VEC (force, =, (-mass2) * a[0]);
    WN_assemble_force (t->node[0], &force);
    VECMAC_2_VEC (force, =, (-mass2) * a[1]);
    WN_assemble_force (t->node[1], &force);
    break;
  }

  el->deformation_energy =
    (t->A * t->L0 * WM_get_uniax_def_energy(m, t->matstate, strain));
}



static W_bool_t
get_elem_prop (W_elem_t *el, const char *prop, void *value)
{
  priv_elem_t *t = (priv_elem_t *)el->data;
  
  if (!value)
    return FALSE;
  
  if (strcmp (prop, "A") == 0) {
    *((double *) value) = t->A;
    return TRUE;
  } else {
    return FALSE;
  }
}


static void *
new_history_client_data (W_elem_t *e,
                         W_index_t idx, /* unique integer  */
                         W_history_writer_t *r, /* history writer */
                         char *input_line /* input line (to allow
                                             for additional parsing of
                                             the line contents by this
                                             function) */
                         )
{
  priv_elem_t *t = (priv_elem_t *)e->data;
  history_client_data_t *cd;
  W_domain_t *d = e->domain;
  cd = w_new (history_client_data_t);
  CHECK_MALLOC(cd);
  cd->el     = e;
  TOKENS_PARSE(input_line);
  cd->pntidx = TOKEN_INT(3);
  return (NODE)cd;
}

static void
update_history (W_elem_t *el, /* generic element */
                W_solution_history_t *h, /* solution history */
                W_float_t time
                )
{
  history_client_data_t *cd =
    (history_client_data_t *)h->client_data;
  W_domain_t *d = el->domain;
  priv_elem_t *t = (priv_elem_t *)el->data;
#define STORE(CD, HW, NAME, TIME, VAL)          \
  {                                             \
    char buf[128];                              \
    sprintf (buf, "p%d%s", CD->pntidx, NAME);   \
    WHR_store (HW, buf, TIME, VAL);             \
  }
  W_float_t L;
  W_coords_3d_t c[2], N;
  W_velocity_3d_t vel[2], v;
  W_material_t *m;
  W_float_t stress, strain, strain_rate, de;

  /* kinematic quantities */
  GET_FROM_NODE (t->node[0], WN_get_coords, c[0]);
  GET_FROM_NODE (t->node[1], WN_get_coords, c[1]);
  VECMAC_3_VEC (N, =, c[1], -, c[0]);
  L   = VECMAC_NORM (N);
  VECMAC_SCALAR (N, /=, L);
  GET_FROM_NODE (t->node[0], WN_get_velocity, vel[0]);
  GET_FROM_NODE (t->node[1], WN_get_velocity, vel[1]);
  VECMAC_3_VEC(v, =, vel[1], -, vel[0]);
  strain_rate = VECMAC_DOT (v, N);
  strain      = (L - t->L0) / t->L0;
  STORE (cd, h->writer, "strain", WD_time (d), strain);
  STORE (cd, h->writer, "strain_rate", WD_time (d), strain_rate);
   
  /* obtain stress */
  m = WD_get_material_by_idx (el->domain, t->material);
  if (t->matstate == NULL)
    t->matstate = WM_get_mat_state (m);
  WM_get_uniax_stress (m, t->matstate, strain, strain_rate, &stress);
  STORE (cd, h->writer, "stress", WD_time (d), stress);

  de = (t->A * t->L0 * WM_get_uniax_def_energy (m, t->matstate, strain));
  STORE (cd, h->writer, "DeformE", WD_time (d), de);
}


static W_float_t
kinetic_energy_consistent (W_elem_t *el)
{
  priv_elem_t *e = (priv_elem_t *)el->data;
  W_vect_t v[2];
  int i;
  double mass2, ke = 0;
  FOR (i, 2) { GET_FROM_NODE (e->node[i], WN_get_velocity, v[i]); }
  mass2 = total_mass (el) / 2;
  ke += (2. / 3 * mass2) * (VECMAC_DOT (v[0], v[0]) + VECMAC_DOT (v[1], v[1]));
  ke += 2 * (1. / 2 * mass2) * VECMAC_DOT (v[0], v[1]);
  return ke / 2;
}

static W_float_t
kinetic_energy_diag (W_elem_t *el)
{
  priv_elem_t *e = (priv_elem_t *)el->data;
  W_vect_t v;
  int i;
  double ke = 0;
  double m = total_mass (el) / 2;

  FOR (i, 2) {
    GET_FROM_NODE (e->node[i], WN_get_velocity, v);
    ke += VECMAC_DOT (v, v);
  }
  return (m * ke) / 2;
}

static W_float_t
kinetic_energy (W_elem_t *el, W_mass_type_enum mass_type)
{
  switch (mass_type) {
  case W_MASS_CONSISTENT:
    return kinetic_energy_consistent (el);
  default:
    return kinetic_energy_diag (el);
  }
}

/*
  Element method dispatch table.
 */
static W_elem_methods_t truss_methods = {
  elem_elcode,
  type_name,
  assemble_restoring_forces,
  assemble_mass,
  total_mass,
  NULL,
  suggested_time_step,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  new_history_client_data,
  update_history,
  NULL,
  NULL,
#ifdef X_XFACE
  modify_elem_grep,
#endif
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  NULL,
  assemble_residual_forces,
  assemble_eff_stiffness,
  kinetic_energy,
  NULL,
  get_elem_prop 
};

/*
  Reads truss element data from stream as long as there is no 
  END_OF_BLOCK tag in the input line.  Creates elements. 
  */
W_bool_t 
truss_reader(W_domain_t *d)
{
  const char *input_line;
  long elem_num, nodeid;
  priv_elem_t *t;
#ifdef PVM_NEED_GHOST_ELEMENTS
  int is_ghost = FALSE;
#endif
  W_material_t *m;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      t = w_new(priv_elem_t);
      CHECK_MALLOC(t);
      TOKENS_PARSE(input_line);
      elem_num    = TOKEN_INT(1);
      nodeid      = TOKEN_INT(2);
      t->node[0] = WD_get_node_by_idx (d, nodeid);
      if (t->node[0] == NULL) {
        sprintf (W_buf (), "No such node: %ld", nodeid);
        REPORT_FATAL_ERROR (d->errh, W_buf ());
      }
      nodeid      = TOKEN_INT(3);
      t->node[1] = WD_get_node_by_idx (d, nodeid);
      if (t->node[1] == NULL) {
        sprintf (W_buf (), "No such node: %ld", nodeid);
        REPORT_FATAL_ERROR (d->errh, W_buf ());
      }
      
      t->material = TOKEN_INT(4);
      t->A        = TOKEN_DBL(5);
      t->L0       = length (t);
      if (t->L0 <= 0) REPORT_FATAL_ERROR (d->errh, "Zero-length element");
      m = WD_get_material_by_idx (d, t->material);
      if (m == NULL)
        REPORT_FATAL_ERROR (d->errh, "Material not found");
      t->matstate = WM_get_mat_state (m);
#ifdef PVM_NEED_GHOST_ELEMENTS
      is_ghost = FALSE;
      if (TOKENS_HAVE_KEYWORD("GHOST"))
        is_ghost = TRUE;
#endif
      WD_add_element(d, elem_num,
                     WE_new_element(d, &truss_methods, t
#                    ifdef PVM_NEED_GHOST_ELEMENTS
                                    , is_ghost
#                    endif
                     ));
    }
  }
  return FALSE;
}
