#include "Wconfig.h"
#include "Wxf.h"
#include "Wvset.h"

/* RCSHEADER "$Id: vset.c,v 1.1.1.1 2001/07/24 02:42:52 pkrysl Exp $" */

/* Value set table is a two-dimensional table. Indexing is effected */
/* either by row and column number or by searching on the key (the 1st */
/* column). 


Named-item table
Table 1
======================
A       0.1
I2      0.0001
I3      0.0001
It      0.0002
Beta2   1.2
Beta3   1.2


Array table
Table 2
======================
   0.1  0.2  0.12
  -0.1  0.2  0.12
  -0.1 -0.2  0.12
   0.1 -0.2  0.12

*/

static HASH table_of_vset_array_tables = NULL;
static HASH table_of_vset_named_item_tables = NULL;

static void
add_vset_table(W_vset_table_t *t);



W_vset_table_t *
WVS_new_vset_named_item_table(void)
{
  W_vset_table_t *t = NULL;

  RAW;
  add_vset_table(t);
  return t;
}
 


W_vset_table_t *
WVS_new_vset_array_table(int table_number, int n_rows, int n_cols)
{
  W_vset_table_t *t = NULL;
  int row, col;

  if (n_rows * n_cols <= 0)
    REPORT_FATAL_ERROR(NULL, "Bad array dimensions");
  
  t = w_new(W_vset_table_t);
  CHECK_MALLOC(t);
  
  t->table_type                   = VSET_ARRAY_TABLE;
  t->table_number                 = table_number;
  t->data.array_table.num_of_rows = n_rows;
  t->data.array_table.num_of_cols = n_cols;
  t->data.array_table.item        = w_new_array(n_rows * n_cols, W_float_t);
  CHECK_MALLOC(t->data.array_table.item);
  for (row = 0; row < t->data.array_table.num_of_rows; row++)
    for (col = 0; col < t->data.array_table.num_of_cols; col++)
      WVS_put_into_array_table(t, row, col, (W_float_t)0);
  
  add_vset_table(t);
  return t;
}


int
WVS_get_vset_table_type(W_vset_table_t *t)
{
  assert(t != NULL);
  
  return t->table_type;
}


int
WVS_n_of_rows_in_array_table(W_vset_table_t *t)
{
  assert(t != NULL);

  return t->data.array_table.num_of_rows;
}



int
WVS_n_of_cols_in_array_table(W_vset_table_t *t)
{
  assert(t != NULL);

  return t->data.array_table.num_of_cols;
}



void
WVS_put_into_array_table(W_vset_table_t *t, int row, int col,
                         W_float_t item)
{
  assert(t != NULL);
  assert(row >= 0 && row < t->data.array_table.num_of_rows);
  assert(col >= 0 && col < t->data.array_table.num_of_cols);
  assert(t->data.array_table.item != NULL);
  t->data.array_table.item[row*t->data.array_table.num_of_cols + col]
    = item;
}


void
WVS_get_from_array_table(W_vset_table_t *t, int row, int col,
                         W_float_t *item_return)
{
  assert(t != NULL);
  assert(row >= 0 && row < t->data.array_table.num_of_rows);
  assert(col >= 0 && col < t->data.array_table.num_of_cols);
  assert(t->data.array_table.item != NULL);
  *item_return =
    t->data.array_table.item[row*t->data.array_table.num_of_cols + col];
}



W_vset_table_t *
WVS_get_table_by_number(int table_number, int table_type)
{
  W_vset_table_t *t = NULL;

  if        (table_type == VSET_NAMED_ITEM_TABLE) {
    t = (W_vset_table_t *)
      get_from_hashtable(table_of_vset_named_item_tables, table_number);
  } else if (table_type == VSET_ARRAY_TABLE) {
    t = (W_vset_table_t *)
      get_from_hashtable(table_of_vset_array_tables, table_number);
  } else {
    REPORT_FATAL_ERROR(NULL, "Unknown table type");
  }
  return t;
}

static int
vset_hash(int table_number);
static int
vset_match(int to_match, W_vset_table_t *t);


static void
add_vset_table(W_vset_table_t *t)
{
  void *result = NULL;
  
  assert(t != NULL);

  if (table_of_vset_array_tables == NULL) {
    table_of_vset_array_tables
         = make_hashtable(10, (PFI)vset_hash, (PFI)vset_match);
    CHECK_MALLOC(table_of_vset_array_tables);
  }
  
  if (table_of_vset_named_item_tables == NULL) {
    table_of_vset_named_item_tables
         = make_hashtable(10, (PFI)vset_hash, (PFI)vset_match);
    CHECK_MALLOC(table_of_vset_named_item_tables);
  }

  if        (t->table_type == VSET_NAMED_ITEM_TABLE) {
    result = add_unique_to_hashtable(table_of_vset_named_item_tables, t,
                                     t->table_number);
  } else if (t->table_type == VSET_ARRAY_TABLE) {
    result = add_unique_to_hashtable(table_of_vset_array_tables, t,
                                     t->table_number);
  } 
  if (result == NULL)
    REPORT_FATAL_ERROR(NULL, "Non-unique table number.");
}


static int
vset_hash(int table_number)
{
  return (table_number % HASH_TABLE_DIM);
}


static int
vset_match(int to_match, W_vset_table_t *t)
{
  return (to_match == t->table_number);
}


static W_bool_t 
vset_array_table_reader(W_domain_t *d, int table_number)
{
  char *input_line;
  long nrows, ncols, row, col;
  double val;
  char *p;
  W_vset_table_t *t = NULL;

  if ((input_line = (char *)W_get_input_line(d)) == NULL)
    REPORT_FATAL_ERROR(d->errh, "No data");
  if (sscanf(input_line, "%ld %ld", &nrows, &ncols) != 2) {
    BAD_INPUT();
  } else {
    t = WVS_new_vset_array_table(table_number, nrows, ncols);
    for (row = 0; row < nrows; row++) {
      if ((input_line = (char *)W_get_input_line(d)) == NULL)
        REPORT_FATAL_ERROR(d->errh, "No data");
      p = input_line;
      for (col = 0; col < ncols; col++) {
        val = strtod(p, &p);
        if (p == input_line)
          BAD_INPUT();
        WVS_put_into_array_table(t, row, col, (W_float_t)val);
      }
    }
  }
  /* test_read_table(table_number); */
  return TRUE;
}


static W_bool_t 
vset_named_item_table_reader(W_domain_t *d, int table_number)
{
  RAW
  return TRUE;
}




/* Reads tnbeam element data from stream as long as there is no */
/* END_OF_BLOCK tag in the input line. Creates elements. */
/* VSET TABLE DATA:                                    */
/* Line     Contents                                   */
/* 1.       table_type                                 */
/*          Array table             Named item table   */
/* 2.       n_rows n_cols              type of values  */ 
/* 3.       1st row of data           1st row of data  */ 
/* ...                                                 */
/* N.       last row of data          last row of data */



W_bool_t 
W_vset_table_reader(W_domain_t *d)
{
  const char *input_line;
  long table_type, table_number;

  while ((input_line = W_get_input_line(d)) != NULL) {
    if (W_is_end_of_block (input_line))
      return TRUE;
    else {
      if (sscanf(input_line, "%ld %ld", &table_type, &table_number) != 2) {
        BAD_INPUT();
      } else {
        switch (table_type) {
        case VSET_NAMED_ITEM_TABLE:
          vset_named_item_table_reader(d, table_number);
          break;
        case VSET_ARRAY_TABLE:
          vset_array_table_reader(d, table_number);
          break;
        default:
          REPORT_FATAL_ERROR(d->errh, "Unknown table type");
          break;
        }
      }
    }
  }
  return FALSE;
}


#if 0 /* TEST THE VSET ARRAY TABLE  */
void
test_read_table(int table_number)
{
  W_vset_table_t *t;
  int nrows, ncols, row, col;
  W_float_t item;

  t = WVS_get_table_by_number(table_number, VSET_ARRAY_TABLE);
  assert(t != NULL);
  printf("type %d\n", WVS_get_vset_table_type(t));
  printf("nrows %d, ncols %d\n",
         (nrows = WVS_n_of_rows_in_array_table(t)), 
         (ncols = WVS_n_of_cols_in_array_table(t)));
  for (row = 0; row < nrows; row++)
    for (col = 0; col < ncols; col++) {
      WVS_get_from_array_table(t, row, col, &item);
      printf("at %d, %d = %f\n", row, col, item);
    }
  for (row = 0; row < nrows; row++)
    for (col = 0; col < ncols; col++) {
      WVS_put_into_array_table(t, row, col, (W_float_t)row*col);
    }
  for (row = 0; row < nrows; row++)
    for (col = 0; col < ncols; col++) {
      WVS_get_from_array_table(t, row, col, &item);
      printf("at %d, %d = %f\n", row, col, item);
    }
}
#endif /* TEST THE VSET ARRAY TABLE */




