#include "Wconfig.h"
#include "Wxf.h"

#ifdef X_XFACE

#include "Wtypes.h"
#include "Wdomain.h"
#include "Wutil.h"
#include "Woptions.h"
#include "Wmodr.h"
#ifdef EFG_WASP
#  include "WEFG.h"
#endif

/* RCSHEADER "$Id: xiface.c,v 1.4 2004/09/06 01:17:58 cvs Exp $" */

static int xiface_started = FALSE;

static struct W_domain_t *
wasp_domain (void)
{
  EModel *em = ESIModel ();
  struct W_domain_t *d;
  if (!em) {
    REPORT_FATAL_ERROR (NULL, "No graphic model");
  }
  d = EMGetClient (em);
  if (!d) {
    REPORT_FATAL_ERROR (NULL, "No domain");
  }
  return d;
}


/*
  Process Elixir commands.
  Process Elixir commands contained in the file /name/.
  This would be used before a computation starts to set up the graphic interface, or
  load files with additional graphics, etc.
 */
void
WXF_process_elixir_cmds (char *name /* file with Elixir commands */
                            ) 
{
#define MAX_LINE_LEN MAXFILENAMECHAR
  char linebuf[MAX_LINE_LEN];
  FILE *stream;
  
  stream = fopen (name, "r");
  if (stream) {
    REPORT_WARNING (wasp_domain ()->errh, "Executing commands from an Elixir command file");
    while (fgets (linebuf, MAX_LINE_LEN-1, stream) != NULL) {
      ESIHandleCmd (linebuf);
    }
    if (!feof(stream)) {
      sprintf (W_buf (), "Elixir command file \"%s\" read failure", name);
      REPORT_FATAL_ERROR (NULL, W_buf ());
    }
    fclose(stream);
  } else {
    sprintf (W_buf (), "Elixir command file \"%s\" could not be opened", name);
    REPORT_FATAL_ERROR (NULL, W_buf ());
  }
}

/*
  Initialize graphics subsystem. No windows or other interface components are
  created.
*/
int
WXF_init (int argc, /* number of arguments; these are the command-line arguments */
                        char *argv[] /* array of arguments */
          )
{
  ESIBuildInterface (ESI_GRAPHIC_EDITOR_MASK, argc, argv);
  
  EMSetAssocFringeTable (ESIModel (), ColorCreateFringeTable ());
  ColorSetupFringeTableByMinMax (EMGetAssocFringeTable(ESIModel()), -1, 1);

  xiface_started = TRUE;

  return 1;
}

/*
  Build graphic interface.
  Build Elixir graphic interface.
  Command-line arguments are passed into this routine to allow
  for any customization.
 */
int /* ignore */
WXF_build_interface (int argc, /* number of arguments; these are the command-line arguments */
                        char *argv[], /* array of arguments */
                        void *client /* client linked to the graphics model (the domain) */
                        )
{
  if (!xiface_started) WXF_init (argc, argv);
  
  ESIPopup ();

  EMSetClient (ESIModel (), client);

  if (WOPT_OPT (wasp_domain ()->options, elixir_cmd_file) != NULL)
    WXF_process_elixir_cmds (WOPT_OPT (wasp_domain ()->options, elixir_cmd_file));
  else {
    ESIHandleCmd ("NEWFRAME WIRE name mesh_view");
    ESIHandleCmd ("view bground black");
  }
  
  return 1;
}

/*
  Build graphic interface.
  Build Elixir graphic interface. Don't popup any windows.
  Command-line arguments are passed into this routine to allow
  for any customization.
 */
int /* ignore */
WXF_build_interface_no_popup (int argc, /* number of arguments; these are the command-line arguments */
                              char *argv[], /* array of arguments */
                              void *client /* client linked to the graphics model (the domain) */
                              )
{
  if (!xiface_started) WXF_init (argc, argv);

  EMSetClient (ESIModel (), client);

  return 1;
}


static void 
StopComputation(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
Proceed(Widget w, XtPointer text_ptr, XtPointer call_data);
static void
toggle_layer_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void 
set_target_time(Widget w, XtPointer client_ptr, XtPointer call_data);
static void 
set_time_between_outputs(Widget w, XtPointer client_ptr, XtPointer call_data);
static void 
set_displ_scale(Widget w, XtPointer client_ptr, XtPointer call_data);
static void 
set_force_scale(Widget w, XtPointer client_ptr, XtPointer call_data);
static void
write_hexvue_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
write_ve3d_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
write_tecplot_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
print_elem_info_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void
print_node_info_CB(Widget w, XtPointer ptr, XtPointer call_data);
static void 
probe_CB(Widget w, XtPointer text_ptr, XtPointer call_data);
static void 
set_probe_option(Widget w, XtPointer client_ptr, XtPointer call_data);
static void 
set_probe_loc(Widget w, XtPointer client_ptr, XtPointer call_data);
static void 
set_probe_type(Widget w, XtPointer client_ptr, XtPointer call_data);
static void 
show_annotation_CB(Widget w, XtPointer client_ptr, XtPointer call_data);



static void
redisplay_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  if (wasp_domain () != NULL) {
    WD_build_grep (wasp_domain (), FALSE);
  }
}

static int modr_mode = 0;

static void 
show_modr_mode_CB (Widget w, XtPointer client_ptr, XtPointer call_data)
{
  char buf[1024];
  sprintf (buf, " Mode %d, eigenvalue %g",
           modr_mode, W_modr_eigenvalue (wasp_domain ()->modr, modr_mode));
  ERptStatusMessage (1, 1, buf);
  W_modr_show_mode (wasp_domain ()->modr, modr_mode);
  ESIHandleCmd ("view redraw");
}

static void 
show_next_modr_mode_CB (Widget w, XtPointer client_ptr, XtPointer call_data)
{
  char buf[1024];
  modr_mode++;
  if (modr_mode >= W_modr_tot_modes (wasp_domain ()->modr)) modr_mode = 0;
  sprintf (buf, " Mode %d, eigenvalue %g",
           modr_mode, W_modr_eigenvalue (wasp_domain ()->modr, modr_mode));
  ERptStatusMessage (1, 1, buf);
  W_modr_show_mode (wasp_domain ()->modr, modr_mode);
  ESIHandleCmd ("view redraw");
}

static void 
show_prev_modr_mode_CB (Widget w, XtPointer client_ptr, XtPointer call_data)
{
  char buf[1024];
  modr_mode--;
  if (modr_mode < 0) modr_mode = W_modr_tot_modes (wasp_domain ()->modr) - 1;
  sprintf (buf, " Mode %d, eigenvalue %g",
           modr_mode, W_modr_eigenvalue (wasp_domain ()->modr, modr_mode));
  ERptStatusMessage (1, 1, buf);
  W_modr_show_mode (wasp_domain ()->modr, modr_mode);
  ESIHandleCmd ("view redraw");
}


static void
set_modr_mode_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0 &&
      TypeInGetTokenType(1) == NUMBER) {
    modr_mode = TypeInGetTokenAsInt(1);
  }
}

static void
save_modr_modes_CB (Widget w, XtPointer ptr, XtPointer call_data)
{
  char *name;
  FILE *fp;
  
  fp = ESISelFile ("Select file", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp != NULL) {
    int mode;
    FOR (mode, W_modr_tot_modes (wasp_domain ()->modr)) {
      W_modr_write_mode (wasp_domain ()->modr, mode, fp);
    }
    fclose(fp);
  }
}

static void
save_modr_data_CB (Widget w, XtPointer ptr, XtPointer call_data)
{
  char *name;
  FILE *fp;
  
  fp = ESISelFile ("Select file", "OK", "CANCEL", NULL, NULL, "w", &name);
  if (fp != NULL) {
    W_modr_save_collected_data (wasp_domain ()->modr, fp);
    fclose(fp);
  }
}

/*
  Standard customization routine.
  Standard Elixir customization routine.  This needs to be defined
  by the calling program to fill the customization menu.
 */
void
ESICustomize(Widget parent_pane
             )
{
  Widget settings, graphic_control, probe, modr;
  Arg args[13];
  int argn;
  Boolean b;

  argn = 0;
  ESIAddButton("proceed", "Proceed", commandWidgetClass,
                   parent_pane, NULL, 0, Proceed, NULL);
  argn = 0;
  ESIAddButton("stop", "Stop", commandWidgetClass,
                   parent_pane, NULL, 0, StopComputation, NULL);
  argn = 0;
  ESIAddButton("write_hexvue_file",
                   "Write HEXVUE file",
                   commandWidgetClass, parent_pane,
                   NULL, 0, write_hexvue_CB, (XtPointer)NULL);
  argn = 0;
  ESIAddButton("write_ve3d_file",
                   "Write VE3D file",
                   commandWidgetClass, parent_pane,
                   NULL, 0, write_ve3d_CB, (XtPointer)NULL);
  argn = 0;
  ESIAddButton("write_tecplot_file",
                   "Write TECPLOT file",
                   commandWidgetClass, parent_pane,
                   NULL, 0, write_tecplot_CB, (XtPointer)NULL);
  argn = 0;
  ESIAddPopupDialog("print_elem_info", "Print element info",
                    "Input element number",
                    "1", parent_pane, args, argn,
                    print_elem_info_CB, NULL,
                    ESIDialogValueNumber, NULL);
  argn = 0;
  ESIAddPopupDialog("print_node_info", "Print node info",
                    "Input node number",
                    "1", parent_pane, args, argn,
                    print_node_info_CB, NULL,
                    ESIDialogValueNumber, NULL);
  argn = 0;
  ESIAddPalette("probe", "Probe ...", parent_pane,
		    args, argn, &probe);
  argn = 0;
  ESIAddPopupDialog("probe_option", "Options",
                    "Input options as a string",
                    "", probe, args, argn,
                    set_probe_option, NULL,
                    ESIDialogValueString, NULL);
  argn = 0;
  ESIAddPopupDialog("probe_loc", "Location",
                    "Input location as x,y,z",
                    "0 0 0", probe, args, argn,
                    set_probe_loc, NULL,
                    ESIDialogValueNumber, NULL);
  argn = 0;
  ESIAddPopupDialog("probe_type", "Type",
                    "One of D (displacement vector), S (stress tensor), E (equivalent plastic strain), V (von Mises stress)",
                    "D", probe, args, argn,
                    set_probe_type, NULL,
                    ESIDialogValueString, NULL);
  ESIAddButton("probe",
                   "Probe",
                   commandWidgetClass, probe,
                   NULL, 0, probe_CB, (XtPointer)NULL);
  

  ESIAddPalette("settings", "Settings ...", parent_pane,
                    NULL, 0, &settings);
  argn = 0;
  ESIAddPopupDialog("target_time", "Target time",
                    "Input the target time",
                    "1.0", settings, args, argn,
                    set_target_time, NULL,
                    ESIDialogValueNumber, NULL);
  argn = 0;
  ESIAddPopupDialog("time_between_outputs", "Time between outputs",
                    "Input the time",
                    "0", settings, args, argn,
                    set_time_between_outputs, NULL,
                    ESIDialogValueNumber, NULL);
  
  ESIAddPalette("graphic_control", "Graphic control ...", parent_pane,
                    NULL, 0, &graphic_control);
  argn = 0;
  ESIAddPopupDialog("displ_scale", "Displacement scaling",
                    "Input the displacement scale",
                    "1.0", graphic_control, args, argn,
                    set_displ_scale, NULL,
                    ESIDialogValueNumber, NULL);
  argn = 0;
  ESIAddPopupDialog("force_scale", "Force scaling",
                    "Input the force scale",
                    "1.0", graphic_control, args, argn,
                    set_force_scale, NULL,
                    ESIDialogValueNumber, NULL);
  b = TRUE;
  argn = 0;
  XtSetArg (args[argn], XtNstate, b); argn++;
  ESIAddButton("node_visibility",
                   "Toggle node visibility",
                   toggleWidgetClass, graphic_control,
                   args, argn, toggle_layer_CB, (XtPointer)WX_NODE_LAYER);
  b = TRUE;
  argn = 0;
  XtSetArg (args[argn], XtNstate, b); argn++;
  ESIAddButton("elem_visibility",
                   "Toggle element visibility",
                   toggleWidgetClass, graphic_control,
                   args, argn, toggle_layer_CB, (XtPointer)WX_ELEM_LAYER);
  b = FALSE;
  argn = 0;
  XtSetArg (args[argn], XtNstate, b); argn++;
  ESIAddButton("natural_contact_surface_visibility",
                   "Toggle natural contact face visibility",
                   toggleWidgetClass, graphic_control,
                   args, argn, toggle_layer_CB, (XtPointer)WX_NATURAL_CF_LAYER);
  b = TRUE;
  argn = 0;
  XtSetArg (args[argn], XtNstate, b); argn++;
  ESIAddButton("rigid_contact_surface_visibility",
                   "Toggle rigid contact face visibility",
                   toggleWidgetClass, graphic_control,
                   args, argn, toggle_layer_CB, (XtPointer)WX_RIGID_CF_LAYER);
 b = FALSE;
  argn = 0;
  XtSetArg (args[argn], XtNstate, b); argn++;
  ESIAddButton("natural_contact_surface_visibility",
                   "Toggle natural contact edge visibility",
                   toggleWidgetClass, graphic_control,
                   args, argn, toggle_layer_CB, (XtPointer)WX_NATURAL_CE_LAYER);
  b = TRUE;
  argn = 0;
  XtSetArg (args[argn], XtNstate, b); argn++;
  ESIAddButton("rigid_contact_surface_visibility",
                   "Toggle rigid contact edge visibility",
                   toggleWidgetClass, graphic_control,
                   args, argn, toggle_layer_CB, (XtPointer)WX_RIGID_CE_LAYER);
  b = FALSE;
  argn = 0;
  XtSetArg (args[argn], XtNstate, b); argn++;
  ESIAddButton("annotation_visibility",
                   "Toggle annotation visibility",
                   toggleWidgetClass, graphic_control,
                   args, argn, toggle_layer_CB, (XtPointer)WX_ANNOT_LAYER);
  ESIAddButton("show_annotation",
                   "Show annotation",
                   commandWidgetClass, graphic_control,
                   NULL, 0, show_annotation_CB, (XtPointer)NULL);
  
  ESIAddPalette("modr", "Model reduction ...", graphic_control,
                    NULL, 0, &modr);
  argn = 0;
  ESIAddPopupDialog("set_modr_mode", "Set mode number",
                    "Mode number",
                    "0", modr, args, argn,
                    set_modr_mode_CB, NULL,
                    ESIDialogValueNumber, NULL);
  ESIAddButton("show_modr_mode",
                   "Show mode",
                   commandWidgetClass, modr,
                   NULL, 0, show_modr_mode_CB, (XtPointer)NULL);
  ESIAddButton("show_prev_modr_mode",
                   "Show previous mode",
                   commandWidgetClass, modr,
                   NULL, 0, show_prev_modr_mode_CB, (XtPointer)NULL);
  ESIAddButton("show_next_modr_mode",
                   "Show next mode",
                   commandWidgetClass, modr,
                   NULL, 0, show_next_modr_mode_CB, (XtPointer)NULL);
  ESIAddButton("save_modr_modes",
                   "Save modes",
                   commandWidgetClass, modr,
                   NULL, 0, save_modr_modes_CB, (XtPointer)NULL);
  ESIAddButton("save_modr_data",
                   "Save data",
                   commandWidgetClass, modr,
                   NULL, 0, save_modr_data_CB, (XtPointer)NULL);
  ESIAddButton("redisplay",
                   "Redisplay",
                   commandWidgetClass, graphic_control,
                   NULL, 0, redisplay_CB, (XtPointer)NULL);
}


static void
write_hexvue_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  char *name;
  FILE *fp;
  
  fp = ESISelFile("Select HEXVUE file name", "OK", "CANCEL",
                  NULL, NULL, "w", &name);
  if (fp != NULL) {
    fclose(fp);
    WD_write_named_postfile (wasp_domain (), name, HEXVUE_POSTPROC_FILE);
  }
}

static void
write_ve3d_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  char *name;
  FILE *fp;
  
  fp = ESISelFile("Select VE3D file name", "OK", "CANCEL",
                  NULL, NULL, "w", &name);
  if (fp != NULL) {
    fclose(fp);
    WD_write_named_postfile (wasp_domain (), name, VE3D_POSTPROC_FILE);
  }
}

static void
write_tecplot_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  char *name;
  FILE *fp;
  
  fp = ESISelFile("Select TECPLOT file name", "OK", "CANCEL",
                  NULL, NULL, "w", &name);
  if (fp != NULL) {
    fclose(fp);
    WD_write_named_postfile (wasp_domain (), name, TECPLOT_I_POSTPROC_FILE);
  }
}


static void
print_elem_info_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  char buf[] = "./tmp.XXXXXX";
  FILE *fp;

  if (TypeInParseLine(call_data) > 0 &&
      TypeInGetTokenType(1) == NUMBER) {
    W_index_t id;

    id = TypeInGetTokenAsInt(1);
  
    if (mkstemp (buf) == -1) sprintf (buf, "./junk");;
    fp = fopen(buf, "w");

    WD_print_elem_info(wasp_domain (), id, W_INFO_FULL, fp);

    fclose(fp);
    ESIPopupInfoFromFile(buf);
  }
}



static void
print_node_info_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  char buf[] = "./tmp.XXXXXX";
  FILE *fp;

  if (TypeInParseLine(call_data) > 0 &&
      TypeInGetTokenType(1) == NUMBER) {
    W_index_t id;

    id = TypeInGetTokenAsInt(1);
  
    if (mkstemp (buf) == -1) sprintf (buf, "./junk");;
    fp = fopen(buf, "w");

    WD_print_node_info(wasp_domain (), id, W_INFO_FULL, fp);

    fclose(fp);
    ESIPopupInfoFromFile(buf);
  }
}


static void
toggle_layer_CB(Widget w, XtPointer ptr, XtPointer call_data)
{
  Boolean s;
  Arg al[1];
  int ac;
  char buf[132];
  int layer = (int)ptr;
  
  ac = 0;
  XtSetArg(al[ac], XtNstate, &s); ac++;
  XtGetValues(w, al, ac);
  if (s)
    sprintf(buf, "LAYER ON %d", layer);
  else
    sprintf(buf, "LAYER OFF %d", layer);
  ESIHandleCmd(buf);
}
  

static void 
set_target_time(Widget w, XtPointer client_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(1) == NUMBER) {
    WOPT_SET_OPT (wasp_domain ()->options, target_time, TypeInGetTokenAsDouble(1));
  }
}



static void 
set_time_between_outputs (Widget w, XtPointer client_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(1) == NUMBER) {
    WOPT_SET_OPT (wasp_domain ()->options, time_between_outputs, TypeInGetTokenAsDouble(1));
  } else {
  }
}

static void 
set_displ_scale(Widget w, XtPointer client_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(1) == NUMBER) {
    WOPT_SET_OPT (wasp_domain ()->options, displ_scale, TypeInGetTokenAsDouble(1));
  }
}


static void 
set_force_scale(Widget w, XtPointer client_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0 && TypeInGetTokenType(1) == NUMBER) {
    WOPT_SET_OPT (wasp_domain ()->options, force_scale, TypeInGetTokenAsDouble(1));
  }
}



static void 
StopComputation(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  ESIEventLoop(1, " ");
} 


static void 
Proceed(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  ESIEventLoopProceed();
}

/*
  Pop up error dialog.
  Pop up error dialog.
 */
void
WXF_error(char *errmsg, /* error message */
             int kind /* kind as defined by Elixir */
             )
{
  if (xiface_started) {
    if (kind == E_K_WARNING) {
      ERptStatusMessage (1, 1, errmsg);
    } else {
      ERptErrMessage (1, 1, errmsg, WARNING_GRADE);
      ESIEventLoop (1, "Error or warning");
    }
  } else {
    fprintf (stderr, "%s\n", errmsg);
    exit (EXIT_FAILURE);
  }
}



static char *probe_option = NULL;
static unsigned long probe_spec = 0;
static W_loc_t probe_loc = {0,0,0};


static void 
set_probe_option(Widget w, XtPointer client_ptr, XtPointer call_data)
{
  if (TypeInParseLine(call_data) > 0) {
    if (probe_option != NULL) free(probe_option);
    DUPLICATE_STRING(probe_option, call_data);
  }
}



static void 
set_probe_type(Widget w, XtPointer client_ptr, XtPointer call_data)
{
  switch (((char *)call_data)[0]) {
  case 'd':
  case 'D':
    probe_spec = W_PROBE_DISPL_MASK;
    break;
  case 's':
  case 'S':
    probe_spec = W_PROBE_STRESS_MASK;
    break;
  case 'e':
  case 'E':
    probe_spec = W_PROBE_EQPLS_MASK;
    break;
  case 'v':
  case 'V':
    probe_spec = W_PROBE_VON_MISES_MASK;
    break;
  default:
    probe_spec = 0;
    break;
  }
}

static void
set_probe_loc(Widget w, XtPointer client_data, XtPointer call_data)
{
  if (   TypeInParseLine(call_data) > 0
      && TypeInGetTokenType(1) == NUMBER
      && TypeInGetTokenType(2) == NUMBER
      && TypeInGetTokenType(3) == NUMBER
      ) {
    probe_loc.x = TypeInGetTokenAsDouble(1);
    probe_loc.y = TypeInGetTokenAsDouble(2);
    probe_loc.z = TypeInGetTokenAsDouble(3);
  } else {
    ERptErrMessage(1, 1, "Bad vector components", ERROR_GRADE);
  }
}


static void 
probe_CB(Widget w, XtPointer text_ptr, XtPointer call_data)
{
  W_probe_package_t result;
  char buf[] = "/tmp.XXXXXX";
  FILE *fp;
  
  if (wasp_domain () == NULL) return;
  
  if (mkstemp (buf) == -1) sprintf (buf, "./junk");;
  fp = fopen(buf, "w");

  fprintf(fp, "Probe: {%g,%g,%g} ", probe_loc.x, probe_loc.y, probe_loc.z); 
  if (probe_option != NULL)
    fprintf(fp, "option: %s\n", probe_option);
  else
    fprintf(fp, "\n");

  if (WD_probe(wasp_domain (), probe_option, &probe_loc, probe_spec, &result)) {
    switch (probe_spec) {
    case W_PROBE_DISPL_MASK:
      fprintf(fp, "Displacement {%g,%g,%g}\n",
	      result.value.vector[0],
	      result.value.vector[1],
	      result.value.vector[2]);
      break;
    case W_PROBE_STRESS_MASK:
      break;
    case W_PROBE_EQPLS_MASK:
      break;
    case W_PROBE_VON_MISES_MASK:
      break;
    default:
      break;
    }
  }
#if 1
  if (WD_probe(wasp_domain (), "interpolate", &probe_loc, probe_spec, &result)) {
    switch (probe_spec) {
    case W_PROBE_DISPL_MASK:
      fprintf(fp, "Displacement (interp. from ip's) {%g,%g,%g}\n",
	      result.value.vector[0],
	      result.value.vector[1],
	      result.value.vector[2]);
      break;
    case W_PROBE_STRESS_MASK:
      break;
    case W_PROBE_EQPLS_MASK:
      break;
    case W_PROBE_VON_MISES_MASK:
      break;
    default:
      break;
    }
  }
#endif

  fclose(fp);
  ESIPopupInfoFromFile(buf);  
}


static void 
show_annotation_CB(Widget w, XtPointer client_ptr, XtPointer call_data)
{
  WOPT_SET_OPT (wasp_domain ()->options, draw_numbers, TRUE);
  WD_build_grep (wasp_domain (), TRUE);
  WOPT_SET_OPT (wasp_domain ()->options, draw_numbers, FALSE);
}


static double last_frame_time = -FLT_MAX;
static double frame_dt = 0;

static void
make_frame (double time)
{
  char *default_frame_format = "jpg";
  char *frame_format = WOPT_OPT (wasp_domain ()->options, frame_format);
  static int frame_num = 0;
  char buf[1024];
  if (frame_dt == 0) {
    if ( WOPT_OPT (wasp_domain ()->options, frame_dt) > 0)
      frame_dt =  WOPT_OPT (wasp_domain ()->options, frame_dt);
    else
      frame_dt =  WOPT_OPT (wasp_domain ()->options, time_between_outputs);
  }
  if (time - last_frame_time >= frame_dt) {
    if ( WOPT_OPT (wasp_domain ()->options, frame_format) == NULL)
      frame_format = default_frame_format;
    else
      frame_format =  WOPT_OPT (wasp_domain ()->options, frame_format);
    if (strcmp(frame_format, "egf") == 0) {
      sprintf (buf, "%s-%05d.%s",
               WOPT_OPT (wasp_domain ()->options, hist_file_name), frame_num, frame_format);
      OnDiskSetFileName(buf);
      EMSaveVisibleGraphics(ESIModel());
    } else {
      sprintf (buf, "image \"%s-%05d.%s\"",
               WOPT_OPT (wasp_domain ()->options, hist_file_name), frame_num, frame_format);
      ESIHandleCmd (buf);
    }
    frame_num++;
    last_frame_time = time;
  }
}

/*
  Show graphics.
  Show graphics by refreshing any windows.
  ESIEventLoop() is called for this purpose.
  If the domain requests movie frames to be gathered,
  and if a frame should be created at time /time/, this routine takes care of it.
 */
void
WXF_show_graphics (int stay, /* TRUE means stay in the graphics interaction loop until released by the user; FALSE means refresh and return immediately */
                      char *title, /* display title on the gui panel */
                      double time /* called for problem time /time/ */
                      )
{
  if ( WOPT_OPT (wasp_domain ()->options, gather_frames)) make_frame (time);
  ERptStatusMessage (1, 1, title);
  ESIEventLoop (stay, title);
}

#endif /* X_XFACE defined */
